
#ifndef TARG_ATTRIBUTETRAITS_H
#define TARG_ATTRIBUTETRAITS_H

/**
 *  @file attributetraits.H
 *
 *  @brief Templates which map attributes to their type/properties
 *
 *  This header file contains templates which map attributes to their
 *  type/properties.  This file is autogenerated and should not be altered.
 */

//******************************************************************************
// Includes
//******************************************************************************

// STD
#include <stdint.h>
#include <stdlib.h>
#ifdef __HOSTBOOT_MODULE
#include <array>
#endif
#include <targeting/common/entitypath.H>

namespace TARGETING
{

//******************************************************************************
// Attribute Property Mappings
//******************************************************************************

/**
 *  @brief Template associating a specific attribute with a type and other
 *      properties, such as whether it is readable/writable
 *
 *      This is automatically generated
 *
 *      enum {
 *          disabled = Special value for the basic, unused wildcard attribute
 *          readable = Attribute is readable
 *          writable = Attribute is writable
 *          hasStringConversion = Attribute has debug string conversion
 *      }
 *
 *      typedef <type> TYPE // <type> is the Attribute's valid type
 */
template<const ATTRIBUTE_ID A>
class AttributeTraits
{
    private:
        enum { disabled };
        typedef void* Type;
};

template<>
class AttributeTraits<ATTR_SBE_IMAGE_MINIMUM_VALID_ECS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXTERNAL_VRM_STEPSIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXTERNAL_VRM_TRANSITION_START_NS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXTERNAL_VRM_TRANSITION_RATE_INC_UV_PER_US>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXTERNAL_VRM_TRANSITION_RATE_DEC_UV_PER_US>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXTERNAL_VRM_TRANSITION_STABILIZATION_TIME_NS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SAFE_MODE_NOVDM_UPLIFT_MV>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SAFE_MODE_VOLTAGE_MV>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SAFE_MODE_FREQUENCY_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_AVSBUS_FREQUENCY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_VDD_AVSBUS_BUSNUM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDN_AVSBUS_BUSNUM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VCS_AVSBUS_BUSNUM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDD_AVSBUS_RAIL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDN_AVSBUS_RAIL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VCS_AVSBUS_RAIL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VCS_I2C_BUSNUM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDD_BOOT_VOLTAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_VDN_BOOT_VOLTAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_VCS_BOOT_VOLTAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SPIPSS_FREQUENCY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PM_APSS_CHIP_SELECT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_R_LOADLINE_VDD_UOHM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_R_DISTLOSS_VDD_UOHM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_VRM_VOFFSET_VDD_UV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_R_LOADLINE_VDN_UOHM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_R_DISTLOSS_VDN_UOHM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_VRM_VOFFSET_VDN_UV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_R_LOADLINE_VCS_UOHM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_R_DISTLOSS_VCS_UOHM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_VRM_VOFFSET_VCS_UV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_BIAS_ULTRATURBO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_BIAS_TURBO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_BIAS_NOMINAL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_BIAS_POWERSAVE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_EXT_VDD_BIAS_ULTRATURBO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_EXT_VDD_BIAS_TURBO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_EXT_VDD_BIAS_NOMINAL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_EXT_VDD_BIAS_POWERSAVE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_EXT_VCS_BIAS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_EXT_VDN_BIAS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_INT_VDD_BIAS_ULTRATURBO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_INT_VDD_BIAS_TURBO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_INT_VDD_BIAS_NOMINAL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_VOLTAGE_INT_VDD_BIAS_POWERSAVE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_STOP4_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_STOP8_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_STOP11_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_CORE_PERIODIC_QUIESCE_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_POUNDV_VALIDITY_HALT_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_WOF_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_WOV_UNDERV_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_FORCE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_WOV_OVERV_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_IVRM_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOF_ENABLE_FRATIO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOF_ENABLE_VRATIO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOF_VRATIO_SELECT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_SAMPLE_125US>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_MAX_DROOP_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_VMIN_MV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_LIMIT_ALGORITHM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_LIMIT_MV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_LIMIT_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_OVERV_VMAX_SETPOINT_MV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_STEP_INCR_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_STEP_DECR_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_MAX_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_PERF_LOSS_THRESH_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_OVERV_STEP_INCR_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_OVERV_STEP_DECR_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_OVERV_MAX_10THPCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PBAX_GROUPID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PBAX_CHIPID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PBAX_BRDCST_ID_VECTOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_POUNDV_BUCKET_NUM_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_POUNDV_BUCKET_NUM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DUMP_STOP_INFO_SUPPRESS_ERROR_TRACE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DUMP_STOP_INFO_ENABLE_ERRORLOG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_VDM_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDM_DROOP_SMALL_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_DROOP_LARGE_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_DROOP_EXTREME_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_OVERVOLT_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_SMALL_FREQ_DROP_N_S_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_LARGE_FREQ_DROP_N_L_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_FREQ_RETURN_L_S_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_FREQ_RETURN_S_N_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_EXTREME_THOTTLE_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDM_FMAX_OVERRIDE_KHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_FMIN_OVERRIDE_KHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_VID_COMPARE_OVERRIDE_MV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VDM_VID_COMPARE_BIAS_0P5PCT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<int8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_DPLL_VDM_RESPONSE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IVRM_DEADZONE_MV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IVRM_STRENGTH_LOOKUP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[64];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 64> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_IVRM_VIN_MULTIPLIER>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[64];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 64> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_IVRM_VIN_MAX_MV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_IVRM_STEP_DELAY_NS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_IVRM_STABILIZATION_DELAY_NS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_TDP_RDP_CURRENT_FACTOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_RESCLK_DISABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_RESCLK_STEP_DELAY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_RESCLK_FREQ_REGIONS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SYSTEM_RESCLK_FREQ_REGION_INDEX>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SYSTEM_RESCLK_VALUE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[64];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 64> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SYSTEM_RESCLK_L3_VALUE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SYSTEM_RESCLK_L3_VOLTAGE_THRESHOLD_MV>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_RING_DBG_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PERF_24x7_INVOCATION_TIME_MS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CME_INSTRUCTION_TRACE_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CME_CHTM_TRACE_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CME_CHTM_TRACE_MEMORY_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PGPE_HCODE_FUNCTION_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_POUND_W_STATIC_DATA_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_AUX_FUNC_INVOCATION_TIME_MS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYS_VFRT_STATIC_DATA_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_LEAKAGE_PERCENT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_PSTATES_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CORE_THROTTLE_ASSERT_COUNT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CORE_THROTTLE_DEASSERT_COUNT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PM_GLOBAL_FIR_TRACE_EN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_PB_MHZ_POUNDV_FALLBACK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SKIP_WAKEUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_MALF_ALERT_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_RESET_FFDC_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_24x7_PERF_ACCUM_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_UNSECURE_HOMER_ADDRESS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_UNSECURE_HOMER_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PMCR_MOST_RECENT_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ENABLED_OBUS_BRICKS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_IO_XBUS_DCCAL_FLAGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_X_DEBUG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_X_MFG_CHK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_X_MFG_MIN_EYE_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_XBUS_MASTER_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_XBUS_TX_MARGIN_RATIO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_XBUS_TX_FFE_PRECURSOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_XBUS_CHAN_EQ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_XBUS_GRP0_PRE_BAD_LANE_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_IO_XBUS_GRP1_PRE_BAD_LANE_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DO_MSS_WR_VREF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SKIP_HW_VREF_CAL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SKIP_RD_VREF_VREFSENSE_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DO_MSS_VREF_DAC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DO_MSS_TRAINING_BAD_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_RUN_DCD_CALIBRATION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OMI_INBAND_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_NVDIMM_SLOT_POSITION>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_UNSUPPORTED_RANK_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_MT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[255];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 255> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_MT_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_MR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[255];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 255> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_MR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_DQ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[160];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 160> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_DQ_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_CK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_CK_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_MW>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[255];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 255> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_VPD_OVERRIDE_MW_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OVERRIDE_MEMDIAGS_BCMODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_GEN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_HYBRID_MEMORY_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_HYBRID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_DENSITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_BANK_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_BANK_GROUP_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_COLUMN_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_ROW_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_PRIM_STACK_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_PPR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_SOFT_PPR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRCD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRAS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRFC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TFAW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRRD_S>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRRD_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TCCD_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TWR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TWTR_S>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TWTR_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TMAW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_WIDTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_RANK_MIX>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_NUM_RANKS_PER_DIMM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_REGISTER_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_MFG_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_RCD_MFG_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_REGISTER_REV>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_PACKAGE_RANK_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][18];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 18>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_NIBBLE_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][18];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 18>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_ROW_REPAIR_SUPPORTED_MRW>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EI_BUS_TX_MSBSWAP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU_REGION_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CLOCK_PLL_MUX_TOD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CLOCK_PLL_MUX>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CLOCK_PLL_MUX0>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_I2C_BUS_DIV_REF>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_EQ_GARD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EC_GARD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_RUNTIME_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IS_SP_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_FFDC_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_INTERNAL_FFDC_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_PLL_BUCKET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FILTER_PLL_BUCKET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_FILTER_PLL_BUCKET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FORCE_SYNC_SS_PLL_SPREAD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_BOOT_FREQ_MULT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_RISK_LEVEL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DISABLE_HBBL_VECTORS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_BACKUP_SEEPROM_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_BOOT_FLAGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_BOOT_FREQ_MHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_BRANCH_PIBMEM_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CHIP_REGIONS_TO_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DEVICE_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ECID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_I2C_BUS_DIV_NEST>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_LEN_OF_SEEPROM_DATA>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MB_BIT_RATE_DIVISOR_PLL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MB_BIT_RATE_DIVISOR_REFCLK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MC_SYNC_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PB_BNDY_DMIPLL_DATA>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PB_BNDY_DMIPLL_FOR_DCCAL_DATA>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PERV_BNDY_PLL_DATA>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_SBE_MASTER_CHIP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_SEEPROM_I2C_DEVICE_ADDRESS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_SEEPROM_I2C_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_START_PIBMEM_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_START_SEEPROM_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WAIT_N0>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WAIT_N1>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WAIT_N2>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WAIT_N3>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYS_FORCE_ALL_CORES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MASTER_CORE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MASTER_EX>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SECURITY_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PFET_OFF_CONTROLS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_OBUS_RATIO_VALUE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PIBMEM_REPAIR0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PIBMEM_REPAIR1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PIBMEM_REPAIR2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SENSEADJ_STEP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CP_FILTER_BYPASS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SS_FILTER_BYPASS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_FILTER_BYPASS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DPLL_BYPASS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_MEM_X_O_PCI_BYPASS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TARGET_HAS_POWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TARGET_HAS_CLOCK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TARGET_IS_SCOMMABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_SYS_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_CP_REFCLOCK_RCVR_TERM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_REFCLOCK_RCVR_TERM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SECTOR_BUFFER_STRENGTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PULSE_MODE_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PULSE_MODE_VALUE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NDL_MESHCTRL_SETUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_START_CBS_FIFO_RESET_SKIP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_CORECACHE_SKEWADJ_DISABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_CORECACHE_DCADJ_DISABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_MEM_TO_USE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ORIG_FIR_SETTINGS_ACTION0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ORIG_FIR_SETTINGS_ACTION1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_LPC_CONSOLE_CNFG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_X_EREPAIR_THRESHOLD_FIELD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_X_EREPAIR_THRESHOLD_MNFG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_O_EREPAIR_THRESHOLD_FIELD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_O_EREPAIR_THRESHOLD_MNFG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DMI_EREPAIR_THRESHOLD_FIELD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DMI_EREPAIR_THRESHOLD_MNFG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_CKE_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_IO_OBUS_DCCAL_FLAGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OBUS_PAT_A_DETECT_RUN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OBUS_PAT_A_DETECT_RX_AMP_VALUE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OBUS_PAT_A_CAPTURE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[24];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 24> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_IO_OBUS_TRAIN_FOR_RECOVERY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OBUS_LINK_SPARE_MARK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_IO_OBUS_LANE_PDWN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_IO_O_DEBUG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_O_MFG_CHK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_O_MFG_MIN_EYE_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OBUS_TX_MARGIN_RATIO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OBUS_TX_FFE_PRECURSOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OBUS_TX_FFE_POSTCURSOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OLLFIR_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_IO_PB_IOOFIR_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_IO_O_CHANNEL_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_O_MFG_STRESS_PR_OFFSET_EVEN>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_O_MFG_STRESS_PR_OFFSET_ODD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_O_MNFG_ERROR_THRESHOLD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_0_VERSION_LAYOUT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_1_VERSION_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_2_SIGNATURE_HASH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CKE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DIMM_RCD_IBT_ODT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DRAM_RTT_NOM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DRAM_RTT_PARK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_DRAM_RTT_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_BIAS_TRIM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DQ_CTLE_CAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DQ_CTLE_RES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DRV_IMP_CLK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DRV_IMP_CNTL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DRV_IMP_CSCID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_DRV_IMP_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 5>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_MC_RCV_IMP_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 5>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_ODT_RD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_ODT_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_VREF_DRAM_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_VREF_MC_RD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MT_WINDAGE_RD_CTR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<int16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_DSTLCFG_MMIO_ADDRBIT_POS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_ENABLE_DL_TMPL_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_ENABLE_DL_TMPL_4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_ENABLE_DL_TMPL_7>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_ENABLE_DL_TMPL_A>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_TMPL_0_PACING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_TMPL_1_PACING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_TMPL_4_PACING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_TMPL_7_PACING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_TMPL_A_PACING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_OMI_OC_MAJOR_VER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_OMI_OC_MINOR_VER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_HRMOR_BYTES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_PROC_REFCLOCK_KHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_DPLL_REFCLOCK_KHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_MEM_REFCLOCK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_ALLOWED_DIMM_FREQ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEMVPD_POS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_REQUIRED_SYNCH_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_TRACE_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CHTM_TRACE_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[24];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 24> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HTMSC_TTYPEFILT_PAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_TSIZEFILT_PAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_TTYPEFILT_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_TSIZEFILT_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_TTYPEFILT_INVERT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CRESPFILT_INVERT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_FILT_PAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_FILT_CRESP_PAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_FILT_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_FILT_CRESP_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_CONTENT_SEL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_CAPTURE_GENERATED_WRITES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_CAPTURE_ENABLE_FILTER_ALL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_CAPTURE_PRECISE_CRESP_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_CAPTURE_LIMIT_MEM_ALLOCATION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_CAPTURE_PMISC_ONLY_CMD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_SYNC_STAMP_FORCE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_HTMSC_MODE_WRITETOIO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CHTM_HTMSC_MODE_CONTENT_SEL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CHTM_HTMSC_MODE_CAPTURE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CHTM_HTMSC_MODE_CORE_INSTR_STALL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MODE_WRAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MODE_DIS_TSTAMP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MODE_SINGLE_TSTAMP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MODE_MARKERS_ONLY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MODE_DIS_FORCE_GROUP_SCOPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MODE_VGTARGET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MEM_SCOPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_MEM_PRIORITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_CTRL_TRIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NHTM_CTRL_MARK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CHTM_CTRL_TRIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CHTM_CTRL_MARK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CTRL_DBG0_STOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CTRL_DBG1_STOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CTRL_RUN_STOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CTRL_OTHER_DBG0_STOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CTRL_XSTOP_STOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CTRL_CHIP0_STOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_CTRL_CHIP1_STOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_IMA_PDBAR_SPLIT_CORE_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_IMA_PDBAR_SCOPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMSC_IMA_PDBAR_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_PB_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_OMI_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_OMI_PLL_VCO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_O_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_FREQ_A_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_X_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SAFE_FREQUENCY_MHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SAFE_VOLTAGE_MV>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_PCIE_MHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DD1_SLOW_PCI_REF_CLOCK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_ASYNC_SAFE_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_BUS_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_BUS_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_CORE_FLOOR_RATIO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_CORE_CEILING_RATIO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_PUMP_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_CCSM_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPTICS_CONFIG_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_SMP_OPTICS_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_OPTICS_CONFIG_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_CAPI_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FABRIC_PRESENT_GROUPS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_SYSTEM_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FABRIC_ADDR_EXTENSION_GROUP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FABRIC_ADDR_EXTENSION_CHIP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EFF_FABRIC_GROUP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EFF_FABRIC_CHIP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_SYSTEM_MASTER_CHIP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_GROUP_MASTER_CHIP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_LINK_ACTIVE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_LINK_TRAIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_ATTACHED_CHIP_CNFG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[7];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 7> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_LINKS_CNFG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_ATTACHED_CHIP_CNFG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_LINKS_CNFG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_ATTACHED_CHIP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[7];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 7> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_ATTACHED_LINK_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[7];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 7> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_ATTACHED_CHIP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_ATTACHED_LINK_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_AGGREGATE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_ADDR_DIS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[7];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 7> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_X_LINK_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[7];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 7> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_AGGREGATE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_ADDR_DIS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_FABRIC_A_LINK_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_EPS_GB_PERCENTAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EPS_TABLE_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EPS_READ_CYCLES_T0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EPS_READ_CYCLES_T1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EPS_READ_CYCLES_T2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EPS_WRITE_CYCLES_T1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_EPS_WRITE_CYCLES_T2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DMI_REFCLOCK_SWIZZLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OMI_REFCLOCK_SWIZZLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_IPL_PHASE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_XSCOM_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_LPC_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_INTERLEAVE_GROUP_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_MIRROR_PLACEMENT_POLICY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_MEM_BASES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_MEM_SIZES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_MIRROR_BASES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_MIRROR_SIZES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_INTERLEAVE_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_INTERLEAVE_GRANULARITY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MEM_MC_IN_GROUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MCS_GROUP_32>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[16][21];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 21>, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MEM_IPL_COMPLETE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_HW_MIRRORING_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NHTM_BAR_BASE_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NHTM_BAR_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_CHTM_BAR_BASE_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[24];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 24> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_CHTM_BAR_SIZES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[24];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 24> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_SMF_BAR_BASE_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_SMF_BAR_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SMF_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SMF_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_OCC_SANDBOX_BASE_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_OCC_SANDBOX_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_MEM_BASES_ACK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_MEM_SIZES_ACK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_MIRROR_BASES_ACK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_MIRROR_SIZES_ACK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HTM_QUEUES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_ENABLE_MEM_EARLY_DATA_SCOM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_DQ_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][72];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 72>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_0_VERSION_LAYOUT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_1_VERSION_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_2_SIGNATURE_HASH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_DPHY_GPO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_DPHY_RLO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_DPHY_WLO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_2N_MODE_AUTOSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_TSYS_ADR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPD_MR_TSYS_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FSP_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FSP_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FSP_BAR_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_FSP_MMIO_MASK_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU_PHY0_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU_PHY0_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU_PHY1_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU_PHY1_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU_MMIO_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU_MMIO_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU0_MMIO_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU0_MMIO_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU1_MMIO_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU1_MMIO_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU2_MMIO_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU2_MMIO_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NPU0_PRI_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_NPU1_PRI_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_NPU2_PRI_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PSI_BRIDGE_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PSI_BRIDGE_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NX_RNG_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NX_RNG_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NX_RNG_FAILED_INT_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_NX_RNG_FAILED_INT_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_PC_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_MASK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_VC_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_MASK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_TM1_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_TM1_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_TM1_BAR_PAGE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_IC_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_IC_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_INT_CQ_IC_BAR_PAGE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_DCCAL_FLAGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_DEBUG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_MFG_CHK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_MFG_MIN_EYE_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_TX_MARGIN_RATIO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_TX_FFE_PRECURSOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_TX_FFE_POSTCURSOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_EVEN>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_ODD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_IOP_SWAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[3];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 3> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[6];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 6> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[6];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 6> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[6];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 6> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_BAR_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[3];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 3> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_CDR_GAIN>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_PK_INIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4][16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 16>, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_INIT_GAIN>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4][16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 16>, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_SIGDET_LVL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_ROT_CDR_LOOKAHEAD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_ROT_CDR_SSC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_ROT_EXTEL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_ROT_RST_FW>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLA>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLB>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_TX_DCLCK_ROT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG1>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG2>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_TX_POWER_SEQ_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG1>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG2>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_DFE_FDDC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_M_CNTL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_CACHE_INJ_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_CACHE_INJ_THROTTLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OMI_DL_LN_REV_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_DPLL_DIVIDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_STOPGPE_BOOT_COPIER_IVPR_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PSTATEGPE_BOOT_COPIER_IVPR_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SPWUP_IGNORE_XSTOP_FLAG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OCC_LFIR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PBA_LFIR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PM_FIRINIT_DONE_ONCE_FLAG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_QUAD_PPM_ERRMASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CORE_PPM_ERRMASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CME_LOCAL_FIRMASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_L2_HASCLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_L3_HASCLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_C0_EXEC_HASCLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_C1_EXEC_HASCLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_C0_PC_HASCLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_C1_PC_HASCLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_L2_HASPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_L3_HASPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_C0_HASPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_C1_HASPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SPIPSS_FRAME_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SPIPSS_IN_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SPIPSS_CLOCK_POLARITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SPIPSS_CLOCK_PHASE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SPIPSS_CLOCK_DIVIDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PM_SPIPSS_INTER_FRAME_DELAY_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PSTATES_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_RESCLK_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDM_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IVRM_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOF_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_OVERV_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOV_UNDERV_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SGPE_PHANTOM_HALT_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PGPE_PHANTOM_HALT_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CME_STOP_PHANTOM_HALT_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CME_PSTATE_PHANTOM_HALT_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CORE_INSIDE_SPECIAL_WAKEUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EX_INSIDE_SPECIAL_WAKEUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EQ_INSIDE_SPECIAL_WAKEUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_INITIATED_PM_RESET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_RESET_PHASE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PM_MALF_CYCLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OMI_DL_GROUP_POS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OMI_DL_NUM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_DMI_PROC_DCCAL_FLAGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_DMI_PROC_DEBUG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_DMI_PROC_TX_MARGIN_RATIO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_DMI_PROC_TX_FFE_PRECURSOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DMI_INBAND_BAR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DMI_INBAND_BAR_BASE_ADDR_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_FREQ_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_FREQ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_FREQ_BIAS_PERCENTAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_SPARE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_WR_VREF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_WR_VREF_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_WRDDR4_VREF_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_CL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_AL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_CWL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_RBT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_DLL_RESET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_DLL_PPD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_DLL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_WR_LVL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_OUTPUT_BUFFER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_PASR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_ASR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_SRT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_MPR_LOC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_MPR_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC00>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC01>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC02>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC06_07>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC08>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC09>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC0A>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC0B>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC0C>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC0D>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC0E>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC0F>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_1x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_2x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_3x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_4x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_5x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_6x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_7x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_8x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_9x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_Ax>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_RC_Bx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_RCD_MIRROR_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_ADDR_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_TEST_VALID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_PARAM_VALID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_WR_EYE_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_RD_EYE_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_DQS_CLK_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_RD_GATE_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_MEMCAL_INTERVAL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_ZQCAL_INTERVAL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_IBM_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_NUM_MASTER_RANKS_PER_DIMM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_RANKS_CONFIGED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_PRIM_DIE_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MEM_M_DRAM_CLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MEM_WATT_TARGET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_TOTAL_PWR_SLOPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_TOTAL_PWR_INTERCEPT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_DIMM_MAXBANDWIDTH_GBS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_DIMM_MAXBANDWIDTH_MRS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_DIMM_MAXPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_DATABUS_UTIL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_PORT_MAXPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_DIMM_THERMAL_LIMIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_RUNTIME_MEM_M_DRAM_CLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EFF_DIMM_FUNCTIONAL_VECTOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_LPASR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_MPR_PAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_GEARDOWN_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_PER_DRAM_ACCESS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_TEMP_READOUT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_CRC_WR_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_MPR_RD_FORMAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_MAX_POWERDOWN_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_INTERNAL_VREF_MONITOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_CS_CMD_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_SELF_REF_ABORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_RD_PREAMBLE_TRAIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_RD_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_WR_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_CA_PARITY_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_CRC_ERROR_CLEAR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_CA_PARITY_ERROR_STATUS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_ODT_INPUT_BUFF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_CA_PARITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DATA_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_WRITE_DBI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_READ_DBI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_VREF_DQ_TRAIN_VALUE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_VREF_DQ_TRAIN_RANGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_VREF_DQ_TRAIN_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_WRITE_CRC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CAL_STEP_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CUSTOM_TRAINING_ADV_PATTERNS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CUSTOM_TRAINING_ADV_WR_PATTERN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VREF_CAL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_RDVREF_CAL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_CAL_ABORT_ON_ERROR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_SLEW_RATE_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_SLEW_RATE_ADR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCHMOO_MULTIPLE_SETUP_CALL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_BUFFER_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_LRDIMM_WORD_X>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint64_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_LRDIMM_MR12_REG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_LRDIMM_ADDITIONAL_CNTL_WORDS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint64_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_LRDIMM_RANK_MULT_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EFF_VPD_VERSION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VDDR_OVERIDE_SPD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VMEM_REGULATOR_MAX_DIMM_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VREF_DAC_NIBBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VCCD_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_MAC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_MODULE_BUS_WIDTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_RTT_NOM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_RTT_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_RTT_PARK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC00>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC01>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC02>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC06>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC07>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC08>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC09>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC0A>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC0B>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC0C>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC0D>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC0E>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_BC0F>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F0BC1x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BC2x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BC3x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BC4x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BC5x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F0BC6x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F70BC7x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BC8x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BC9x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BCAx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F30BCBx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F0BCCx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F0BCDx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F0BCEx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F0BCFx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F1BCCx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F1BCDx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F1BCEx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F1BCFx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F2BCEx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F4BC0x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F4BC1x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F4BC2x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F4BC3x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F4BC4x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F4BC5x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F4BC6x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F5BC0x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F5BC1x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F5BC2x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F5BC3x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F5BC5x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F5BC6x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F6BC0x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F6BC1x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F6BC2x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F6BC3x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F6BC4x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F6BC5x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BC8x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BC9x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BCAx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BCBx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BCCx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BCDx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BCEx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DIMM_DDR4_F74BCFx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_RON>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TDQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_ODIC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_LRDIMM_TRAINING_PATTERN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EFF_ODT_RD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EFF_ODT_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EFF_DPHY_WLO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EFF_DPHY_RLO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TREFI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRTP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRFC_DLR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TFAW_DLR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TRRD_DLR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EFF_DRAM_TXS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_IGNORE_PLUG_RULES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MVPD_FWMS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 8>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_REORDER_QUEUE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EFF_RANK_GROUP_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_RTT_NOM_OVERRIDE_DISABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EFF_WR_CRC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_WR_VREF_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_NVDIMM_IN_PORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT8_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT8_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT16_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT16_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT32_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT32_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT64_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT64_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT8_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT8_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][3][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 3>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT16_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT16_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][3][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint16_t, 2>, 3>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT32_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT32_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][3];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 3>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT64_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_UINT64_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint64_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT8_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT8_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT16_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int16_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT16_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int16_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT32_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int32_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT32_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int32_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT64_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int64_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT64_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int64_t Type;
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT8_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<int8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT8_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type[2][3][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<int8_t, 4>, 3>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT16_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int16_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<int16_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT16_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int16_t Type[2][3][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<int16_t, 2>, 3>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT32_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int32_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<int32_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT32_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int32_t Type[2][3];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<int32_t, 3>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT64_ARRAY_1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int64_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<int64_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SCRATCH_INT64_ARRAY_2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int64_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<int64_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_I2C_DEV_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_I2C_SUB_POS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FAPI_POS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_IS_SIMULATION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXECUTION_PLATFORM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_FLAGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_RECONFIGURE_LOOP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MINI_EC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_DMI_CEN_TX_MARGIN_RATIO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IO_DMI_CEN_TX_FFE_POSTCURSOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_DMI_REFCLOCK_RCVR_TERM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_DDR_REFCLOCK_RCVR_TERM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_MODES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_RANK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_START_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_END_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ERROR_CAPTURE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_MAX_TIMEOUT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_PRINT_PORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_STOP_ON_ERROR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_DATA_SEED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_INTER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_NUM_ROWS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_NUM_COLS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_RANK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_BANK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_SLAVE_RANK_ON>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_STR_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_ADDR_RAND>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_DRAM_DEVICE_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_CUSTOM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_DENSITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_BANKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_ROWS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_COLUMNS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_NOMINAL_VOLTAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_NUM_RANKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_DRAM_WIDTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_MEMORY_BUS_WIDTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TCKMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_CAS_LATENCIES_SUPPORTED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TAAMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRCDMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRPMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRASMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRCMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TFAWMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_OPTIONAL_FEATURES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_THERMAL_AND_REFRESH_OPTIONS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ROW_REPAIR_SPPR_SUPPORTED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_THERMAL_SENSOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_DEVICE_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_SIGNAL_LOADING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_DIE_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TCKMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TAAMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TRCDMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TRPMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TRCMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURERS_JEDEC_ID_CODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_LOCATION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_DATE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_ID_MODULE_SERIAL_NUMBER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_PART_NUMBER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[20];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 20> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_REVISION_CODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_DRAM_MANUFACTURER_JEDEC_ID_CODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_BAD_DQ_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[80];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 80> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FTB_DIVIDEND>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FTB_DIVISOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MTB_DIVIDEND>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MTB_DIVISOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TWRMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRRDMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRFCMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TWTRMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRTPMIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_SDRAM_BANKGROUPS_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TIMEBASE_MTB_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TIMEBASE_FTB_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TCKMAX_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRFC1MIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRFC2MIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRFC4MIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRRDSMIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TRRDLMIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_TCCDLMIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TCCDLMIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TRRDLMIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TRRDSMIN_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_FINE_OFFSET_TCKMAX_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_CRC_BASE_CONFIG_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_DRAM_STEPPING_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_CRC_MNFG_SEC_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_ADDR_MAP_REG_TO_DRAM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SPD_MODULE_REVISION_CODE_DDR4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_ISDIMM_MBVPD_INDEX>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_DQ_TO_DIMM_CONN_DQ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[80];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 80> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MBA_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MBA_DIMM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MBVPD_VERSION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DATABUS_UTIL_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_ADDRESS_MIRRORING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_FREQ_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_FREQ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_FREQ_BIAS_PERCENTAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VREF_CAL_CNTL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VREF_CAL_DELTA_FROM_NOMINAL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_RANKS_CONFIGED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_NUM_RANKS_PER_DIMM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CUSTOM_DIMM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_WIDTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_GEN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_PRIMARY_RANK_GROUP0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_PRIMARY_RANK_GROUP1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_PRIMARY_RANK_GROUP2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_PRIMARY_RANK_GROUP3>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SECONDARY_RANK_GROUP0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SECONDARY_RANK_GROUP1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SECONDARY_RANK_GROUP2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SECONDARY_RANK_GROUP3>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_TERTIARY_RANK_GROUP0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_TERTIARY_RANK_GROUP1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_TERTIARY_RANK_GROUP2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_TERTIARY_RANK_GROUP3>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_QUATERNARY_RANK_GROUP0>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_QUATERNARY_RANK_GROUP1>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_QUATERNARY_RANK_GROUP2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_QUATERNARY_RANK_GROUP3>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_WR_VREF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_WR_VREF_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_RD_VREF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CEN_RD_VREF_SCHMOO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_BANKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_ROWS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_COLS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_DENSITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRCD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRRD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRRD_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRAS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRFI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRFC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TWTR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TWTR_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TRTP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TFAW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_BL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_CL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_AL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_CWL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_RBT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_DLL_RESET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_DLL_PPD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_DLL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TDQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_WR_LVL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_OUTPUT_BUFFER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_PASR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_ASR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_SRT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_MPR_LOC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_MPR_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_RCD_CNTL_WORD_0_15>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint64_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC00>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC01>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC02>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC06_07>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC08>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC09>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC10>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC11>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC12>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC13>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC14>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC15>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_1x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_2x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_3x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_4x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_5x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_6x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_7x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_8x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_9x>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_Ax>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DIMM_DDR4_RC_Bx>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_ADDR_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_TEST_VALID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_PARAM_VALID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_VREF_STEP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_WR_EYE_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_RD_EYE_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_DQS_CLK_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_RD_GATE_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_MEMCAL_INTERVAL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_ZQCAL_INTERVAL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_IBM_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_NUM_DROPS_PER_PORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_STACK_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_NUM_MASTER_RANKS_PER_DIMM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_NUM_PACKAGES_PER_RANK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_NUM_DIES_PER_PACKAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_MEM_THROTTLE_DENOMINATOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_CHIP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_MEM_WATT_TARGET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_POWER_SLOPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_POWER_SLOPE2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_POWER_INT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_POWER_INT2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_TOTAL_POWER_SLOPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_TOTAL_POWER_SLOPE2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_TOTAL_POWER_INT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_TOTAL_POWER_INT2>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint32_t, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_CHANNEL_PAIR_MAXPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_DENOMINATOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_CHIP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_MBA>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DERIVED_MBA_ADDR_INTERLEAVE_BIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DERIVED_MBA_CACHELINE_INTERLEAVE_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_MBA_CACHELINE_INTERLEAVE_MODE_CONTROL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_CACHE_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_PREFETCH_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_CLEANER_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_EFF_DIMM_FUNCTIONAL_VECTOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_LPASR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_MPR_PAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_GEARDOWN_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_PER_DRAM_ACCESS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_TEMP_READOUT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_FINE_REFRESH_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CRC_WR_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_MPR_RD_FORMAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_MAX_POWERDOWN_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_TEMP_REF_RANGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_TEMP_REF_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_INT_VREF_MON>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CS_CMD_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_SELF_REF_ABORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_RD_PREAMBLE_TRAIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_RD_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_WR_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CA_PARITY_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CRC_ERROR_CLEAR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CA_PARITY_ERROR_STATUS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_ODT_INPUT_BUFF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_CA_PARITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DATA_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_WRITE_DBI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_READ_DBI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_VREF_DQ_TRAIN_VALUE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_VREF_DQ_TRAIN_RANGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_VREF_DQ_TRAIN_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_EFF_DRAM_TCCD_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_TCCD_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_EFF_WRITE_CRC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_CAL_STEP_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DRAMINIT_RESET_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_SLEW_RATE_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_SLEW_RATE_ADR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_CEN_MSS_ALLOW_SINGLE_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DQS_SWIZZLE_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_PSRO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_NWELL_MISPLACEMENT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_BLUEWATERFALL_BROKEN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_PATTERN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_TEST_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_PRINTING_DISABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_DATA_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_USER_RANK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_USER_BANK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_SCHMOO_MULTIPLE_SETUP_CALL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_LRDIMM_RANK_MULT_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_THROTTLE_CONTROL_RAS_WEIGHT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_THROTTLE_CONTROL_CAS_WEIGHT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_RANDOM_SEED_VALUE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MCBIST_RANDOM_SEED_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_INIT_STATE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_EFF_VPD_VERSION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_NEST_CAPABLE_FREQUENCIES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_STRICT_MBA_PLUG_RULE_CHECKING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_CDIMM_MASTER_I2C_TEMP_SENSOR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_CDIMM_SPARE_I2C_TEMP_SENSOR_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_THERMAL_MEMORY_POWER_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_POWER_CONTROL_REQUESTED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DDR3_VDDR_SLOPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DDR3_VDDR_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DDR4_VDDR_SLOPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_DDR4_VDDR_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_COMPLIANT_DIMMS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_MCS_PREFETCH_RETRY_THRESHOLD>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_DDR3_VDDR_MAX_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_DDR4_VDDR_MAX_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_MEM_SENSOR_CACHE_ADDR_MAP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MRW_WR_VREF_CHECK_VREF_STEP_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_DRAM_CWL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_RDIMM_BUFFER_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_REORDER_QUEUE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_2N_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_VPD_DQ_MAP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[72];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 72> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC0F>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_CS_CMD_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_CA_PARITY_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC02>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC0B>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC1X>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F0RC7X>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F1RC00>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F1RC02>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F1RC03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F1RC04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_DDR4_F1RC05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_POS_METADATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_DRAM_GEN_METADATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_DIMM_TYPE_METADATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OMI_EDPL_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_IGNORE_PLUG_RULES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OMI_DL_PREIPL_PRBS_TIME>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_SI_SIGNATURE_HASH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_SI_DIMM_RCD_IBT_CA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DIMM_RCD_IBT_CKE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DIMM_RCD_IBT_CS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DIMM_RCD_IBT_ODT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DRAM_DRV_IMP_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DRAM_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DRAM_RTT_NOM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DRAM_RTT_PARK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_DRAM_RTT_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_GEARDOWN_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_RCV_EQ_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_EQ_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_IMP_CLK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_IMP_CMD_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_IMP_CNTL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_IMP_CSCID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_SLEW_RATE_CLK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_SLEW_RATE_CNTL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_SLEW_RATE_CSCID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_RCV_IMP_ALERT_N>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_MC_RCV_IMP_DQ_DQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_ODT_RD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_ODT_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_SI_VREF_DRAM_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_SI_VREF_MC_RD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_SI_WINDAGE_RD_CTR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int16_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_SUPPORTED_FREQ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_MRW_IS_PLANAR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_THERMAL_MEMORY_POWER_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[10];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 10> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MRW_PWR_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[100];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 100> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MRW_PWR_SLOPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[100];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 100> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MRW_REFRESH_RATE_REQUEST>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_IDLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_MEM_M_DRAM_CLOCKS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_MAX_DRAM_DATABUS_UTIL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_POWER_CONTROL_REQUESTED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_IDLE_POWER_CONTROL_REQUESTED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_VMEM_REGULATOR_POWER_LIMIT_PER_DIMM_ADJ_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR3>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR4>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_MAX_NUMBER_DIMMS_POSSIBLE_PER_VMEM_REGULATOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_AVDD_OFFSET_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_VDD_OFFSET_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_VCS_OFFSET_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_VPP_OFFSET_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_VDDR_OFFSET_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_FINE_REFRESH_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_TEMP_REFRESH_RANGE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_RESET_DELAY_BEFORE_CAL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_PERIODIC_MEMCAL_MODE_OPTIONS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_PERIODIC_ZQCAL_MODE_OPTIONS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_DRAM_2N_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_DRAM_WRITE_CRC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_TEMP_REFRESH_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_FORCE_BCMODE_OFF>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_NVDIMM_PLUG_RULES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_ALLOW_UNSUPPORTED_RCW>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_SUPPORTED_DRAM_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[25];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 25> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MRW_OCMB_PWR_SLOPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[50];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 50> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MRW_OCMB_PWR_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[50];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 50> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[25];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 25> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_MRW_SAFEMODE_DRAM_DATABUS_UTIL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_PORT_POS_OF_FAIL_THROTTLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_GEN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DIMM_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_HYBRID_MEMORY_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_HYBRID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_HOST_TO_DDR_SPEED_RATIO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_DENSITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_BANK_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_BANK_GROUP_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_COLUMN_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_ROW_BITS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PRIM_DIE_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PRIM_STACK_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PRIM_BUS_WIDTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_PPR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_SOFT_PPR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRCD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRAS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRFC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TFAW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRRD_S>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRRD_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRRD_DLR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TCCD_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TWR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TWTR_S>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TWTR_L>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TMAW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_WIDTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_LOGICAL_RANKS_PER_DIMM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_3DS_HEIGHT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_REGISTER_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_MODULE_MFG_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_MFG_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_MODULE_HEIGHT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_RCD_MFG_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_REGISTER_REV>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PACKAGE_RANK_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][20];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 20>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_NIBBLE_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][20];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 20>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DIMM_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DIMM_SPARE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_CL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_MDS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_NUM_MASTER_RANKS_PER_DIMM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DIMM_RANKS_CONFIGED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TREFI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRTP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DRAM_TRFC_DLR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_FREQ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_VOLT_VDDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_VOLT_VPP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_SPD_REVISION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_ENABLE_US_TMPL_1>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_ENABLE_US_TMPL_5>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_ENABLE_US_TMPL_9>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_ENABLE_US_TMPL_A>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_ENABLE_US_TMPL_B>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_TMPL_0_PACING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_TMPL_1_PACING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_TMPL_5_PACING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_TMPL_9_PACING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_TMPL_B_PACING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_SHRT_BACKOFF_TIMER>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_METADATA_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_PASID_BASE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_ACTAG_BASE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_AFU_ACTAG_LEN>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXPLR_PASID_LEN>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PSTATES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_BYTE_ENABLES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_NIBBLE_ENABLES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_FOUR_RANK_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_MRAM_SUPPORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_DDP_COMPATIBILITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_TSV_8H_SUPPORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_MODULE_THERMAL_SENSORS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_0_AVAIL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_0_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_0_USAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_0_I2C_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_1_AVAIL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_1_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_1_USAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_1_I2C_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_DIFF_AVAIL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_DIFF_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_DIFF_USAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_THERM_SENSOR_DIFF_I2C_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_SUPPORTED_RCD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_ENTERPRISE_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_ENTERPRISE_POLICY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_NONENTERPRISE_MODE_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_HALF_DIMM_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_HALF_DIMM_MODE_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MEM_MVPD_FWMS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EXP_INIT_VREF_DQ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EXP_INIT_PHY_VREF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EXP_RCD_DIC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EXP_RCD_VOLTAGE_CTRL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EXP_DRAM_ADDRESS_MIRRORING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EXP_RCD_SLEW_RATE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EXP_SPD_CL_SUPPORTED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EXP_SPD_TAA_MIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_REORDER_QUEUE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EXP_FIRMWARE_EMULATION_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_STRUCT_MMIO_ENDIAN_CTRL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_STRUCT_ENDIAN>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_STRUCT_MMIO_WORD_SWAP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_OMI_CFG_ENDIAN_CTRL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_ECID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[14];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 14> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EXP_DFIMRL_CLK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_ATXDLY_A>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEM_EFF_ATXDLY_B>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_PHY_INIT_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_CHECKSTOP_OBJ_HANDLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_RECOV_OBJ_HANDLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_SPECATTN_OBJ_HANDLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_APPINTR_OBJ_HANDLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CHECK_FOR_READY_TIMEOUT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_FAIL_BEHAVIOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_DISABLE_THERM_INIT_READ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC00>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC01>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC0B>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC0E>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC0F>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC1X>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F0RC7X>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F1RC00>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F1RC02>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F1RC03>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F1RC04>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DDR4_F1RC05>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_RBT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_TM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_DLL_RESET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_BURST_LENGTH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_DLL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_ODIC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_AL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_WR_LVL_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_RTT_NOM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_RTT_PARK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_TDQS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_OUTPUT_BUFFER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_LPASR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DRAM_RTT_WR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_MPR_PAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_MPR_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_GEARDOWN_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_PER_DRAM_ACCESS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_TEMP_READOUT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_CRC_WR_LATENCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_MPR_RD_FORMAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_MAX_POWERDOWN_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_INTERNAL_VREF_MONITOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_SELF_REF_ABORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_RD_PREAMBLE_TRAIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_RD_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_WR_PREAMBLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_CRC_ERROR_CLEAR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_CA_PARITY_ERROR_STATUS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_ODT_INPUT_BUFF>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_CA_PARITY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_DATA_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_WRITE_DBI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_READ_DBI>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4][20];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 20>, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4][20];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 20>, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2][4][20];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 20>, 4>, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_PORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_SLOT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_RUNTIME_MEM_M_DRAM_CLOCKS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_MEM_PORT_POS_OF_FAIL_THROTTLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_MEM_WATT_TARGET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EXP_TOTAL_PWR_SLOPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EXP_TOTAL_PWR_INTERCEPT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_EXP_DATABUS_UTIL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_PORT_MAXPOWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_EXP_DIMM_THERMAL_LIMIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_FW_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_OPENCAPI_LOOPBACK_TEST>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_TRANSPORT_LAYER>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DL_LAYER_BOOT_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DFE_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_LANE_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_SERDES_FREQUENCY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_OCMB_EXP_BOOT_CONFIG_ADAPTATION_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_MFG_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_MFG_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_SETTING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_RANGE_SELECT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_DELAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_ORDER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_REDUNDANCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_PHASE_COMB>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_REDUNDANCY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_REVISION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_REVISION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_PHASE_COMB>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC0_SEQUENCE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_PMIC1_SEQUENCE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC0_SWA_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC0_SWB_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC0_SWC_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC0_SWD_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC1_SWA_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC1_SWB_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC1_SWC_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_EFF_EFD_PMIC1_SWD_VOLTAGE_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_AFFINITY_PATH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EntityPath Type;
};

template<>
class AttributeTraits<ATTR_ALL_MCS_IN_INTERLEAVING_GROUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ALTFSI_MASTER_CHIP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EntityPath Type;
};

template<>
class AttributeTraits<ATTR_ALTFSI_MASTER_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ASYNC_NEST_FREQ_MHZ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_ATTN_AREA_1_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ATTN_AREA_2_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_AVDD_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_BLOCK_SPEC_DECONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_BMC_FRU_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_BPM_UPDATE_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_BRAZOS_RX_FIFO_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CDM_DOMAIN>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef CDM_DOMAIN Type;
};

template<>
class AttributeTraits<ATTR_CDM_POLICIES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CEC_IPL_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef CecIplType Type;
};

template<>
class AttributeTraits<ATTR_CENTAUR_ECID_FRU_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_AVDD_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VCS_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VCS_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VDDR_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VDD_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VDD_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VPP_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CEN_MSS_VOLT_VPP_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CHIPLET_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CHIP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CHIP_UNIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CLASS>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef CLASS Type;
};

template<>
class AttributeTraits<ATTR_CPU_ATTR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CVPD_MAX_SECTIONS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_CVPD_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_DATA_CACHE_LINE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DATA_CACHE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DCACHE_ASSOC_SETS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DCACHE_LINE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DECONFIG_GARDABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DISABLE_I2C_ENGINE2_PORT0_DIAG_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DO_ABUS_DECONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DPO_MIN_FREQ_PERCENT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int32_t Type;
};

template<>
class AttributeTraits<ATTR_DUMMY_HEAP_ZERO_DEFAULT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DUMMY_RO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DUMMY_RW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[1][3][5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<std::array<uint8_t, 5>, 3>, 1> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_DUMMY_WO>
{
    public:
        enum { writeable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EEPROM_CONTENT_TYPE>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef EEPROM_CONTENT_TYPE Type;
};

template<>
class AttributeTraits<ATTR_EEPROM_NV_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EepromNvInfo Type;
};

template<>
class AttributeTraits<ATTR_EEPROM_SBE_BACKUP_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EepromSbeBackupInfo Type;
};

template<>
class AttributeTraits<ATTR_EEPROM_SBE_PRIMARY_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EepromSbePrimaryInfo Type;
};

template<>
class AttributeTraits<ATTR_EEPROM_VPD_BACKUP_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EepromVpdBackupInfo Type;
};

template<>
class AttributeTraits<ATTR_EEPROM_VPD_PRIMARY_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EepromVpdPrimaryInfo Type;
};

template<>
class AttributeTraits<ATTR_EFFECTIVE_EC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ENABLED_THREADS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ENGINE_TYPE>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef ENGINE_TYPE Type;
};

template<>
class AttributeTraits<ATTR_FABRIC_CHIP_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FABRIC_GROUP_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FABRIC_TO_PHYSICAL_NODE_MAP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_FAPI_I2C_CONTROL_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef FapiI2cControlInfo Type;
};

template<>
class AttributeTraits<ATTR_FAPI_NAME>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[64];
};

template<>
class AttributeTraits<ATTR_FIELD_CORE_OVERRIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FIELD_TH_P8EX_L2_COL_REPAIRS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FIELD_TH_P8EX_L2_LINE_DELETES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FIELD_TH_P8EX_L3_COL_REPAIRS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FIELD_TH_P8EX_L3_LINE_DELETES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FORCE_NVDIMM_RESET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FORCE_SRAM_MMIO_OVER_I2C>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_CORE_CEILING_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_CORE_MAX>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FREQ_PROC_REFCLOCK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FRU_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FSI_MASTER_CHIP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EntityPath Type;
};

template<>
class AttributeTraits<ATTR_FSI_MASTER_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FSI_MASTER_TYPE>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef FSI_MASTER_TYPE Type;
};

template<>
class AttributeTraits<ATTR_FSI_OPTION_FLAGS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef FsiOptionFlags Type;
};

template<>
class AttributeTraits<ATTR_FSI_SLAVE_CASCADE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FSP_BAR_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_FSP_BASE_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_FUSED_CORE_MODE_HB>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FUSED_CORE_OPTION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_HRMOR_NODAL_BASE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_HB_RSV_MEM_SIZE_MB>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HB_SETTINGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef HbSettings Type;
};

template<>
class AttributeTraits<ATTR_HDAT_EC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HDAT_HBRT_NUM_SECTIONS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HDAT_HBRT_SECTION_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[9];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 9> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_RSV_MEM_NUM_SECTIONS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HDDW_ORDER>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HOMER_PHYS_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_HOSTSVC_PLID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HOT_PLUG_POWER_CONTROLLER_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[8][8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 8>, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HUID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HWAS_STATE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef HwasState Type;
};

template<>
class AttributeTraits<ATTR_HWAS_STATE_CHANGED_FLAG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_HWAS_STATE_CHANGED_SUBSCRIPTION_MASK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_I2C_BUS_SPEED_ARRAY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[4][13];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 13>, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_I2C_CONTROL_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef I2cControlInfo Type;
};

template<>
class AttributeTraits<ATTR_I2C_SWITCHES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef I2cSwitches Type;
};

template<>
class AttributeTraits<ATTR_IBSCOM_MCS_BASE_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_IBSCOM_PROC_BASE_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ICACHE_ASSOC_SETS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_ICACHE_BLOCK_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_ICACHE_LINE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_ICACHE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_IMT_BAR_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_IMT_BASE_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_INIT_TO_AVAILABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_INTP_BASE_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ISTEP_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ISTEP_PAUSE_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ISTEP_PAUSE_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IS_INTER_ENCLOSURE_BUS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_KEY_CLEAR_REQUEST>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef KEY_CLEAR_REQUEST Type;
};

template<>
class AttributeTraits<ATTR_KEY_TRANSITION_STATE>
{
    public:
        enum { writeable, readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef KEY_TRANSITION_STATE Type;
};

template<>
class AttributeTraits<ATTR_L2_CACHE_ASSOC_SETS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_L2_CACHE_LINE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_L2_CACHE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_L3_CACHE_LINE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_L3_CACHE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_LOCATION_CODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[64];
};

template<>
class AttributeTraits<ATTR_LPC_BUS_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_CHIPLETS_PER_PROC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_COMPUTE_NODES_PER_SYSTEM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_DIMMS_PER_MBA_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_DMI_PER_PROC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_EXS_PER_PROC_CHIP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_MBAS_PER_MEMBUF_CHIP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_MBA_PORTS_PER_MBA>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_MCS_PER_SYSTEM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_POWER>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_PROC_CHIPS_PER_NODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_SBE_SEEPROM_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MBA_DIMM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MBA_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MC_PLL_BUCKET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_AVDD_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_BASE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_VCS_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_VDDR_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_VDD_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MEM_VPP_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MFG_TRACE_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef MFG_WRAP_TEST_ABUS_LINKS_SET Type;
};

template<>
class AttributeTraits<ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MGC_LOAD_SOURCE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MIN_FREQ_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MIRROR_BASE_ADDRESS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_ABUS_MIN_EYE_HEIGHT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_ABUS_MIN_EYE_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_DMI_MIN_EYE_HEIGHT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_DMI_MIN_EYE_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_CEN_L4_CACHE_CES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_CEN_MBA_IPL_SOFT_CE_TH_ALGO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_CEN_MBA_RT_RCE_PER_RANK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_CEN_MBA_RT_SOFT_CE_TH_ALGO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_MEMORY_IMPES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_MEMORY_IUES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L2_CACHE_CES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L2_COL_REPAIRS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L2_DIR_CES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L2_LINE_DELETES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L3_CACHE_CES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L3_COL_REPAIRS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L3_DIR_CES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_P8EX_L3_LINE_DELETES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_TH_RCD_PARITY_ERRORS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MNFG_XBUS_MIN_EYE_WIDTH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MODEL>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef MODEL Type;
};

template<>
class AttributeTraits<ATTR_MRU_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_DDR3_VDDR_MAX_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_DDR3_VDDR_MAX_LIMIT_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_DDR4_VDDR_MAX_LIMIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_DDR4_VDDR_MAX_LIMIT_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_DEFAULT_RISK_LEVEL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_MEM_THROTTLE_DENOMINATOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_AVDD_PROGRAM>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_AVDD_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_AVDD_SLOPE_ACTIVE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_AVDD_SLOPE_INACTIVE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_VCS_INTERCEPT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_VCS_SLOPE_ACTIVE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_VCS_SLOPE_INACTIVE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_VDD_INTERCEPT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_VDD_SLOPE_ACTIVE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_CENT_VDD_SLOPE_INACTIVE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_DATABUS_UTIL_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MBA_ADDR_INTERLEAVE_BIT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_MBA_CACHELINE_INTERLEAVE_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_PHY_SEQ_REFRESH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MSS_VCS_PROGRAM>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VDDR_PROGRAM>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VDD_PROGRAM>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_AVDD_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR3_VDDR_SLOPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR4_VDDR_SLOPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VCS_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VCS_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VDDR_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VDD_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VDD_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_INTERCEPT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_INTERCEPT_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_OFFSET_MILLIVOLTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_SLOPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_SLOPE_POST_DRAM_INIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VPP_PROGRAM>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_PLL_FREQ_BUCKETS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_PLL_FREQ_I2CDIV_LIST>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_NEST_PLL_FREQ_LIST>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_NEST_VCS_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_VDDR_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_VDD_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_VDN_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_NEST_VIO_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_NOMINAL_FREQ_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_NPU_MMIO_BAR_BASE_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[4][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint64_t, 2>, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_NPU_MMIO_BAR_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[4][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint64_t, 2>, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_NUMERIC_POD_TYPE_TEST>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef NumericPodTypeTest Type;
};

template<>
class AttributeTraits<ATTR_NVDIMM_ARMED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef NvdimmArmed Type;
};

template<>
class AttributeTraits<ATTR_NVDIMM_AUTO_ARM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NVDIMM_ENCRYPTION_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NVDIMM_ENCRYPTION_KEYS_ANCHOR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[96];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 96> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_NVDIMM_ENCRYPTION_KEYS_FW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[96];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 96> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_NVDIMM_FORCE_FW_UPDATE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NVDIMM_UPDATE_I2C_BLOCK_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NV_OPS_TIMEOUT_MSEC>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[6];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 6> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_NV_STATUS_FLAG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OBUS_BRICK_LANE_MASK>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_OBUS_SLOT_INDEX>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OCC_MASTER_CAPABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OCMB_FW_UPDATE_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ORDINAL_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PARENT_DECONFIG_RULES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef ParentDeconfigRules Type;
};

template<>
class AttributeTraits<ATTR_PAYLOAD_BASE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PAYLOAD_ENTRY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PAYLOAD_IN_MIRROR_MEM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PAYLOAD_KIND>
{
    public:
        enum { writeable, readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef PAYLOAD_KIND Type;
};

template<>
class AttributeTraits<ATTR_PCIE_32BIT_DMA_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PCIE_32BIT_MMIO_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PCIE_64BIT_DMA_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PCIE_64BIT_MMIO_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PCIE_CAPABILITES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PCI_BASE_ADDRS_32>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PCI_BASE_ADDRS_64>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PCI_REFCLOCK_RCVR_TERM>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PEC_IS_BIFURCATABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_HX_KEYWORD_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[9];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 9> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_IOP_REVERSAL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_IOP_REVERSAL_BIFURCATED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_IOP_SWAP_BIFURCATED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_IOP_SWAP_NON_BIFURCATED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_LANE_MASK_BIFURCATED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PEC_PCIE_LANE_MASK_NON_BIFURCATED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PEER_PATH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EntityPath Type;
};

template<>
class AttributeTraits<ATTR_PEER_TARGET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef TARGETING::Target* Type;
};

template<>
class AttributeTraits<ATTR_PG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PHB_BASE_ADDRS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PHYS_PATH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EntityPath Type;
};

template<>
class AttributeTraits<ATTR_PIB_I2C_NEST_PLL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PIB_I2C_REFCLOCK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PLCK_IPL_ATTR_OVERRIDES_EXIST>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PNOR_FLASH_WORKAROUNDS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_POSITION>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_POS_ON_MEM_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_POWER_PATH>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef EntityPath Type;
};

template<>
class AttributeTraits<ATTR_PRD_HWP_PLID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PREVIOUS_WAKEUP_FAIL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PRIMARY_CAPABILITIES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef PrimaryCapabilities Type;
};

template<>
class AttributeTraits<ATTR_PROC_COMPATIBILITY_REQ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef PROC_COMPATIBILITY_REQ Type;
};

template<>
class AttributeTraits<ATTR_PROC_DCM_INSTALLED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_MASTER_TYPE>
{
    public:
        enum { writeable, readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_IOP_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_IOVALID_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_IS_SLOT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN3>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN4>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_LANE_MASK>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_NUM_PEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_LOFF_CONTROL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_PHASE_ROTATOR_CNTL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_SIGDET_CNTL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_SYSTEM_CNTL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PCS_TX_FIFO_CONFIG_OFFSET>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_PHB_ACTIVE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_PCIE_REFCLOCK_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_REFCLOCK_RCVR_TERM>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_SELECT_BOOT_SEEPROM_IMAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_SELECT_SEEPROM_IMAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PSI_BRIDGE_BASE_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PSI_HB_ESB_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_RAW_MTM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[64];
};

template<>
class AttributeTraits<ATTR_RCD_PARITY_RECONFIG_LOOPS_ALLOWED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_RCD_PARITY_RECONFIG_LOOP_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_RECONFIG_LOOP_TESTS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[5];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 5> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_RECONFIG_LOOP_TESTS_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_REDUNDANT_CLOCKS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_REL_POS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_RESOURCE_IS_CRITICAL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_RISK_LEVEL_ORIGIN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef RISK_LEVEL_ORIGIN Type;
};

template<>
class AttributeTraits<ATTR_RNG_BAR_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_RNG_BASE_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_RUN_MAX_MEM_PATTERNS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SAVED_STATE_UINT8>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_COMMIT_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_IS_STARTED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_RELEASE_TAG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[21];
};

template<>
class AttributeTraits<ATTR_SBE_UPDATE_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_VERSION_INFO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SCOM_SWITCHES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef ScomSwitches Type;
};

template<>
class AttributeTraits<ATTR_SECUREBOOT_PROTECT_DECONFIGURED_TPM>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SLOT_INDEX>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SLOT_NAME>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SOCKET_POWER_NOMINAL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SOCKET_POWER_TURBO>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SP_FUNCTIONS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef SpFunctions Type;
};

template<>
class AttributeTraits<ATTR_START_MEM_ADDRESS_FOR_PAYLOAD_TCE_TOKEN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_STOP5_DISABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SUPPORTED_STOP_STATES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SUPPORTS_DYNAMIC_MEM_VOLT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYNC_BETWEEN_STEPS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_FAMILY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[64];
};

template<>
class AttributeTraits<ATTR_SYSTEM_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[64];
};

template<>
class AttributeTraits<ATTR_TCE_START_TOKEN_FOR_HDAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TCE_START_TOKEN_FOR_PAYLOAD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TEMP_SENSOR_I2C_CONFIG>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef TempSensorI2cConfig Type;
};

template<>
class AttributeTraits<ATTR_TEST_MAX_STRING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[4];
};

template<>
class AttributeTraits<ATTR_TEST_MIN_STRING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[10];
};

template<>
class AttributeTraits<ATTR_TEST_NEGATIVE_FCN>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef int8_t Type;
};

template<>
class AttributeTraits<ATTR_TEST_NO_DEFAULT_STRING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[10];
};

template<>
class AttributeTraits<ATTR_TEST_NULL_STRING>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[10];
};

template<>
class AttributeTraits<ATTR_THREAD_COUNT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TIME_BASE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TLB_DATA_ASSOC_SETS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TLB_DATA_ENTRIES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TLB_INSTR_ASSOC_SETS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TLB_INSTR_ENTRIES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TLB_RESERVE_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TOD_CPU_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[44];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 44> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_TOD_ROLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TPM_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef TpmInfo Type;
};

template<>
class AttributeTraits<ATTR_TPM_POISONED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TPM_REQUIRED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TPM_UNUSABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TYPE>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef TYPE Type;
};

template<>
class AttributeTraits<ATTR_UCD_MFR_REVISION_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_ULTRA_TURBO_FREQ_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_USE_TCES_FOR_DMAS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VCS_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_VDDR_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_VDD_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_VENDOR_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VPD_REC_NUM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_VPP_ID>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_WOF_POWER_LIMIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_WOF_TABLE_LID_NUMBER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_XIVE_CONTROLLER_BAR_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_XIVE_THREAD_MGMT1_BAR_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_XSCOM_BASE_ADDRESS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ALLOW_ATTR_OVERRIDES_IN_SECURE_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ASSUME_SBE_QUIESCED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ATTN_CHK_ALL_PROCS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CENTAUR_REGISTER_CACHE_PTR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_CHKSTOP_ANALYSIS_ON_STARTUP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_CLEAR_DIMM_SPD_ENABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_DIMM_SPD_BYTE_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DRTM_PAYLOAD_ADDR_MB_HB>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_DYNAMIC_I2C_DEVICE_ADDRESS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_EARLY_TESTCASES_ISTEP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_EEPROM_PAGE_ARRAY>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[4][4];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint8_t, 4>, 4> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_FORCE_PRE_PAYLOAD_DRTM>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_FORCE_SBE_UPDATE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef FORCE_SBE_UPDATE Type;
};

template<>
class AttributeTraits<ATTR_FREQ_MCA_MHZ>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_FSI_MASTER_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_FSI_SCOM_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_GPIO_INFO_PHYS_PRES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef GpioInfoPhysPres Type;
};

template<>
class AttributeTraits<ATTR_HBRT_HYP_ID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_HB_EXISTING_IMAGE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_INITIATED_PM_RESET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_MUTEX_SERIALIZE_TEST_LOCK>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_HB_MUTEX_TEST_LOCK>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_HB_NODE_COMM_ABUS_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_HB_NODE_COMM_XBUS_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_HB_RECURSIVE_MUTEX_TEST_LOCK>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_HB_RSV_MEM_NEXT_SECTION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_SBE_SEEPROM_VERSION_MISMATCH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_TARGET_SCOMABLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_TPM_INIT_ATTEMPTED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HB_TPM_LOG_MGR_PTR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_HB_TPM_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_HIDDEN_ERRLOGS_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HOMER_HCODE_LOADED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HOMER_VIRT_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_I2C_ENGINE_MUTEX_0>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_I2C_ENGINE_MUTEX_1>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_I2C_ENGINE_MUTEX_2>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_I2C_ENGINE_MUTEX_3>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_I2C_PAGE_MUTEX_0>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_I2C_PAGE_MUTEX_1>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_I2C_PAGE_MUTEX_2>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_I2C_PAGE_MUTEX_3>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_IBSCOM_ENABLE_OVERRIDE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IBSCOM_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_IBSCOM_VIRTUAL_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_IPC_NODE_BUFFER_GLOBAL_ADDRESS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint64_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_IS_DRTM_MPIPL_HB>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IS_MPIPL_HB>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IS_SLAVE_DRAWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MASTER_MBOX_SCRATCH>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[8];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 8> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_MEMD_OFFSET>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MMIO_PHYS_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MMIO_VM_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_MPIPL_HB_MDRT_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_DDR3_VDDR_MAX_LIMIT_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_DDR4_VDDR_MAX_LIMIT_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_INTERCEPT_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_VOLT_VPP_SLOPE_EFF_CONFIG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_NVDIMM_READING_PAGE4>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_NVDIMM_READING_VENDOR_LOG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OCC_COMMON_AREA_PHYS_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OCMB_COUNTER_HB>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_OVERRIDES_ATTEMPTED_FLAG>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PART_NUMBER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[48];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 48> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PDA_THREAD_REG_ENTRY_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PDA_THREAD_REG_STATE_ENTRY_FORMAT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PHYS_PRES_ASSERTED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PHYS_PRES_FAKE_ASSERT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PHYS_PRES_REIPL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PHYS_PRES_REQUEST_OPEN_WINDOW>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_ARCH_DUMP_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_COMM_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SBE_FFDC_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_SCAN_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_SCOM_IND_MUTEX>
{
    public:
        enum { writeable, readable, hbMutex, notFspMutex, fspAccessible };
        typedef mutex_t* Type;
};

template<>
class AttributeTraits<ATTR_SERIAL_NUMBER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[18];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 18> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_SPCWKUP_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_TPM_MODEL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_TPM_ROLE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VPD_SWITCHES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef VpdSwitches Type;
};

template<>
class AttributeTraits<ATTR_WOF_FREQUENCY_UPLIFT_SELECTED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[22][13];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 13>, 22> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_XSCOM_VIRTUAL_ADDR>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_ASCII_VPD_LX_KEYWORD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_CHIP_VER>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef CHIP_VER Type;
};

template<>
class AttributeTraits<ATTR_CREATE_DEF_PARTITION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ENLARGED_IO_SLOT_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HUGE_PAGE_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_HUGE_PAGE_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HW_VER>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef HW_VER Type;
};

template<>
class AttributeTraits<ATTR_HYPERVISOR_IPL_DESTINATION>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HYPERVISOR_IPL_SIDE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HYP_DISPATCH_WHEEL>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IPL_ATTRIBUTES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef IplAttributes Type;
};

template<>
class AttributeTraits<ATTR_IPS_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IPS_ENTER_TIME_SECONDS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_IPS_ENTER_UTILIZATION_PERCENT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_IPS_EXIT_TIME_SECONDS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_IPS_EXIT_UTILIZATION_PERCENT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_LMB_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_HSL_OPTICONNECT_CONNECTIONS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_PM_MODE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_PM_MODE_FREQ_PERCENT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_OS_IPL_MODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PHYP_SYSTEM_TYPE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_PRESERVE_MDC_PARTITION_VPD>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_HW_TOPOLOGY>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_ROLE>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef ROLE Type;
};

template<>
class AttributeTraits<ATTR_SLCA_INDEX>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SLCA_RID>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_SW_VER>
{
    public:
        enum { readable, hasStringConversion, notHbMutex, notFspMutex, fspAccessible };
        typedef SW_VER Type;
};

template<>
class AttributeTraits<ATTR_SYSTEM_BRAND_NAME>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[64];
};

template<>
class AttributeTraits<ATTR_VLAN_SWITCHES>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VTPM_ENABLED>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_ADC_CHANNEL_FUNC_IDS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_ADC_CHANNEL_GAINS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_ADC_CHANNEL_GNDS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_ADC_CHANNEL_OFFSETS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_ADC_CHANNEL_SENSOR_NUMBERS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_ALLOW_CALLHOME_ESELS_TO_BMC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_APSS_GPIO_PORT_MODES>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_APSS_GPIO_PORT_PINS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[16];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_BMC_HW_CHIP_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[32];
};

template<>
class AttributeTraits<ATTR_BMC_MANUFACTURER>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[32];
};

template<>
class AttributeTraits<ATTR_BMC_MAX_ERROR_LOG_SIZE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_BMC_SW_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[32];
};

template<>
class AttributeTraits<ATTR_CALCULATED_MAX_SYS_POWER_EXCLUDING_GPUS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_CALCULATED_PROC_MEMORY_POWER_DROP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_CUMULATIVE_PMCOMPLEX_RESET_COUNT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_GPIO_INFO>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef GpioInfo Type;
};

template<>
class AttributeTraits<ATTR_GPU_SENSORS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[3][7];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 7>, 3> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_ADDR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_BUS_FREQ>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_DEVICE_LABEL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[32][64];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<char, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_DEVICE_PURPOSE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_DEVICE_TYPE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_ELEMENTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_ENGINE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_MASTER_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HDAT_I2C_SLAVE_PORT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[32];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 32> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_HTMGT_INTERNAL_FLAGS>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_HTMGT_PMCOMPLEX_RESET_COUNT_TIMER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_HTMGT_SAFEMODE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_HTMGT_SAVED_POWER_LIMIT>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_IPMI_INSTANCE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_IPMI_MAX_BUFFER_SIZE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_IPMI_SENSORS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[16][2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<std::array<uint16_t, 2>, 16> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_IS_MPIPL_SUPPORTED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_MAX_VDD_CURRENT_READING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MISC_SYSTEM_COMPONENTS_MAX_POWER_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_MSS_UTIL_N_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type;
};

template<>
class AttributeTraits<ATTR_NO_APSS_PROC_POWER_VCS_VIO_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_N_PLUS_ONE_MEM_POWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_N_PLUS_ONE_N_PER_CHIP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_N_PLUS_ONE_N_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_OCC_CONTROL_DATA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type[256];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint8_t, 256> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_OPAL_MODEL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef char Type[32];
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_DIMM_ERROR_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_DIMM_READ_TIMEOUT_SEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_DIMM_THROTTLE_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_GPU_ERROR_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_GPU_MEM_ERROR_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_GPU_MEM_READ_TIMEOUT_SEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_GPU_READ_TIMEOUT_SEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_MEMCTRL_ERROR_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_MEMCTRL_READ_TIMEOUT_SEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_MEMCTRL_THROTTLE_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_POWER_CAP>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_THROTTLING>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_MIN_POWER_CAP_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_N_BULK_POWER_LIMIT_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_N_MAX_MEM_POWER_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_N_PLUS_ONE_BULK_POWER_LIMIT_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_N_PLUS_ONE_HPC_BULK_POWER_LIMIT_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_N_PLUS_ONE_MAX_MEM_POWER_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_PROC_DVFS_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_PROC_ERROR_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_PROC_READ_TIMEOUT_SEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_PROC_WEIGHT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_QUAD_WEIGHT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_REGULATOR_EFFICIENCY_FACTOR>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_SOFT_MIN_PCAP_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint64_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_TURBO_MODE_SUPPORTED>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_VRM_READ_TIMEOUT_SEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_VRM_VDD_DVFS_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_VRM_VDD_ERROR_TEMP_DEG_C>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OPEN_POWER_VRM_VDD_READ_TIMEOUT_SEC>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OP_TRACE_LITE>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_OT_MEM_POWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_OT_MIN_N_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_POWERCAP_MEM_POWER>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint32_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint32_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_POWERCAP_N_PER_CHIP>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_POWERCAP_N_PER_MBA>
{
    public:
        enum { writeable, readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type[2];
#ifdef __HOSTBOOT_MODULE
        typedef std::array<uint16_t, 2> TypeStdArr;
#endif
};

template<>
class AttributeTraits<ATTR_PROC_MHZ_PER_WATT>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_PROC_SOCKET_POWER_WATTS>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

template<>
class AttributeTraits<ATTR_REPORT_THROTTLE_BELOW_NOMINAL>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint8_t Type;
};

template<>
class AttributeTraits<ATTR_VDD_CURRENT_OVERFLOW_WORKAROUND_ENABLE>
{
    public:
        enum { readable, notHbMutex, notFspMutex, fspAccessible };
        typedef uint16_t Type;
};

/**
 *  @brief Mapping of alias type name to underlying type
 */
// Type aliases and/or sizes for ATTR_SBE_IMAGE_MINIMUM_VALID_ECS attribute
typedef uint8_t SBE_IMAGE_MINIMUM_VALID_ECS_ATTR;
typedef uint8_t ATTR_SBE_IMAGE_MINIMUM_VALID_ECS_type;

// Type aliases and/or sizes for ATTR_EXTERNAL_VRM_STEPSIZE attribute
typedef uint32_t EXTERNAL_VRM_STEPSIZE_ATTR;
typedef uint32_t ATTR_EXTERNAL_VRM_STEPSIZE_type;

// Type aliases and/or sizes for ATTR_EXTERNAL_VRM_TRANSITION_START_NS attribute
typedef uint32_t EXTERNAL_VRM_TRANSITION_START_NS_ATTR;
typedef uint32_t ATTR_EXTERNAL_VRM_TRANSITION_START_NS_type;

// Type aliases and/or sizes for ATTR_EXTERNAL_VRM_TRANSITION_RATE_INC_UV_PER_US attribute
typedef uint32_t EXTERNAL_VRM_TRANSITION_RATE_INC_UV_PER_US_ATTR;
typedef uint32_t ATTR_EXTERNAL_VRM_TRANSITION_RATE_INC_UV_PER_US_type;

// Type aliases and/or sizes for ATTR_EXTERNAL_VRM_TRANSITION_RATE_DEC_UV_PER_US attribute
typedef uint32_t EXTERNAL_VRM_TRANSITION_RATE_DEC_UV_PER_US_ATTR;
typedef uint32_t ATTR_EXTERNAL_VRM_TRANSITION_RATE_DEC_UV_PER_US_type;

// Type aliases and/or sizes for ATTR_EXTERNAL_VRM_TRANSITION_STABILIZATION_TIME_NS attribute
typedef uint32_t EXTERNAL_VRM_TRANSITION_STABILIZATION_TIME_NS_ATTR;
typedef uint32_t ATTR_EXTERNAL_VRM_TRANSITION_STABILIZATION_TIME_NS_type;

// Type aliases and/or sizes for ATTR_SAFE_MODE_NOVDM_UPLIFT_MV attribute
typedef uint32_t SAFE_MODE_NOVDM_UPLIFT_MV_ATTR;
typedef uint32_t ATTR_SAFE_MODE_NOVDM_UPLIFT_MV_type;

// Type aliases and/or sizes for ATTR_SAFE_MODE_VOLTAGE_MV attribute
typedef uint32_t SAFE_MODE_VOLTAGE_MV_ATTR;
typedef uint32_t ATTR_SAFE_MODE_VOLTAGE_MV_type;

// Type aliases and/or sizes for ATTR_SAFE_MODE_FREQUENCY_MHZ attribute
typedef uint32_t SAFE_MODE_FREQUENCY_MHZ_ATTR;
typedef uint32_t ATTR_SAFE_MODE_FREQUENCY_MHZ_type;

// Type aliases and/or sizes for ATTR_AVSBUS_FREQUENCY attribute
typedef uint32_t AVSBUS_FREQUENCY_ATTR;
typedef uint32_t ATTR_AVSBUS_FREQUENCY_type;

// Type aliases and/or sizes for ATTR_VDD_AVSBUS_BUSNUM attribute
typedef uint8_t VDD_AVSBUS_BUSNUM_ATTR;
typedef uint8_t ATTR_VDD_AVSBUS_BUSNUM_type;

// Type aliases and/or sizes for ATTR_VDN_AVSBUS_BUSNUM attribute
typedef uint8_t VDN_AVSBUS_BUSNUM_ATTR;
typedef uint8_t ATTR_VDN_AVSBUS_BUSNUM_type;

// Type aliases and/or sizes for ATTR_VCS_AVSBUS_BUSNUM attribute
typedef uint8_t VCS_AVSBUS_BUSNUM_ATTR;
typedef uint8_t ATTR_VCS_AVSBUS_BUSNUM_type;

// Type aliases and/or sizes for ATTR_VDD_AVSBUS_RAIL attribute
typedef uint8_t VDD_AVSBUS_RAIL_ATTR;
typedef uint8_t ATTR_VDD_AVSBUS_RAIL_type;

// Type aliases and/or sizes for ATTR_VDN_AVSBUS_RAIL attribute
typedef uint8_t VDN_AVSBUS_RAIL_ATTR;
typedef uint8_t ATTR_VDN_AVSBUS_RAIL_type;

// Type aliases and/or sizes for ATTR_VCS_AVSBUS_RAIL attribute
typedef uint8_t VCS_AVSBUS_RAIL_ATTR;
typedef uint8_t ATTR_VCS_AVSBUS_RAIL_type;

// Type aliases and/or sizes for ATTR_VCS_I2C_BUSNUM attribute
typedef uint8_t VCS_I2C_BUSNUM_ATTR;
typedef uint8_t ATTR_VCS_I2C_BUSNUM_type;

// Type aliases and/or sizes for ATTR_VDD_BOOT_VOLTAGE attribute
typedef uint32_t VDD_BOOT_VOLTAGE_ATTR;
typedef uint32_t ATTR_VDD_BOOT_VOLTAGE_type;

// Type aliases and/or sizes for ATTR_VDN_BOOT_VOLTAGE attribute
typedef uint32_t VDN_BOOT_VOLTAGE_ATTR;
typedef uint32_t ATTR_VDN_BOOT_VOLTAGE_type;

// Type aliases and/or sizes for ATTR_VCS_BOOT_VOLTAGE attribute
typedef uint32_t VCS_BOOT_VOLTAGE_ATTR;
typedef uint32_t ATTR_VCS_BOOT_VOLTAGE_type;

// Type aliases and/or sizes for ATTR_SPIPSS_FREQUENCY attribute
typedef uint32_t SPIPSS_FREQUENCY_ATTR;
typedef uint32_t ATTR_SPIPSS_FREQUENCY_type;

// Type aliases and/or sizes for ATTR_PM_APSS_CHIP_SELECT attribute
typedef uint8_t PM_APSS_CHIP_SELECT_ATTR;
typedef uint8_t ATTR_PM_APSS_CHIP_SELECT_type;

// Type aliases and/or sizes for ATTR_PROC_R_LOADLINE_VDD_UOHM attribute
typedef uint32_t PROC_R_LOADLINE_VDD_UOHM_ATTR;
typedef uint32_t ATTR_PROC_R_LOADLINE_VDD_UOHM_type;

// Type aliases and/or sizes for ATTR_PROC_R_DISTLOSS_VDD_UOHM attribute
typedef uint32_t PROC_R_DISTLOSS_VDD_UOHM_ATTR;
typedef uint32_t ATTR_PROC_R_DISTLOSS_VDD_UOHM_type;

// Type aliases and/or sizes for ATTR_PROC_VRM_VOFFSET_VDD_UV attribute
typedef uint32_t PROC_VRM_VOFFSET_VDD_UV_ATTR;
typedef uint32_t ATTR_PROC_VRM_VOFFSET_VDD_UV_type;

// Type aliases and/or sizes for ATTR_PROC_R_LOADLINE_VDN_UOHM attribute
typedef uint32_t PROC_R_LOADLINE_VDN_UOHM_ATTR;
typedef uint32_t ATTR_PROC_R_LOADLINE_VDN_UOHM_type;

// Type aliases and/or sizes for ATTR_PROC_R_DISTLOSS_VDN_UOHM attribute
typedef uint32_t PROC_R_DISTLOSS_VDN_UOHM_ATTR;
typedef uint32_t ATTR_PROC_R_DISTLOSS_VDN_UOHM_type;

// Type aliases and/or sizes for ATTR_PROC_VRM_VOFFSET_VDN_UV attribute
typedef uint32_t PROC_VRM_VOFFSET_VDN_UV_ATTR;
typedef uint32_t ATTR_PROC_VRM_VOFFSET_VDN_UV_type;

// Type aliases and/or sizes for ATTR_PROC_R_LOADLINE_VCS_UOHM attribute
typedef uint32_t PROC_R_LOADLINE_VCS_UOHM_ATTR;
typedef uint32_t ATTR_PROC_R_LOADLINE_VCS_UOHM_type;

// Type aliases and/or sizes for ATTR_PROC_R_DISTLOSS_VCS_UOHM attribute
typedef uint32_t PROC_R_DISTLOSS_VCS_UOHM_ATTR;
typedef uint32_t ATTR_PROC_R_DISTLOSS_VCS_UOHM_type;

// Type aliases and/or sizes for ATTR_PROC_VRM_VOFFSET_VCS_UV attribute
typedef uint32_t PROC_VRM_VOFFSET_VCS_UV_ATTR;
typedef uint32_t ATTR_PROC_VRM_VOFFSET_VCS_UV_type;

// Type aliases and/or sizes for ATTR_FREQ_BIAS_ULTRATURBO attribute
typedef int8_t FREQ_BIAS_ULTRATURBO_ATTR;
typedef int8_t ATTR_FREQ_BIAS_ULTRATURBO_type;

// Type aliases and/or sizes for ATTR_FREQ_BIAS_TURBO attribute
typedef int8_t FREQ_BIAS_TURBO_ATTR;
typedef int8_t ATTR_FREQ_BIAS_TURBO_type;

// Type aliases and/or sizes for ATTR_FREQ_BIAS_NOMINAL attribute
typedef int8_t FREQ_BIAS_NOMINAL_ATTR;
typedef int8_t ATTR_FREQ_BIAS_NOMINAL_type;

// Type aliases and/or sizes for ATTR_FREQ_BIAS_POWERSAVE attribute
typedef int8_t FREQ_BIAS_POWERSAVE_ATTR;
typedef int8_t ATTR_FREQ_BIAS_POWERSAVE_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_EXT_VDD_BIAS_ULTRATURBO attribute
typedef int8_t VOLTAGE_EXT_VDD_BIAS_ULTRATURBO_ATTR;
typedef int8_t ATTR_VOLTAGE_EXT_VDD_BIAS_ULTRATURBO_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_EXT_VDD_BIAS_TURBO attribute
typedef int8_t VOLTAGE_EXT_VDD_BIAS_TURBO_ATTR;
typedef int8_t ATTR_VOLTAGE_EXT_VDD_BIAS_TURBO_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_EXT_VDD_BIAS_NOMINAL attribute
typedef int8_t VOLTAGE_EXT_VDD_BIAS_NOMINAL_ATTR;
typedef int8_t ATTR_VOLTAGE_EXT_VDD_BIAS_NOMINAL_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_EXT_VDD_BIAS_POWERSAVE attribute
typedef int8_t VOLTAGE_EXT_VDD_BIAS_POWERSAVE_ATTR;
typedef int8_t ATTR_VOLTAGE_EXT_VDD_BIAS_POWERSAVE_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_EXT_VCS_BIAS attribute
typedef int8_t VOLTAGE_EXT_VCS_BIAS_ATTR;
typedef int8_t ATTR_VOLTAGE_EXT_VCS_BIAS_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_EXT_VDN_BIAS attribute
typedef int8_t VOLTAGE_EXT_VDN_BIAS_ATTR;
typedef int8_t ATTR_VOLTAGE_EXT_VDN_BIAS_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_INT_VDD_BIAS_ULTRATURBO attribute
typedef int8_t VOLTAGE_INT_VDD_BIAS_ULTRATURBO_ATTR;
typedef int8_t ATTR_VOLTAGE_INT_VDD_BIAS_ULTRATURBO_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_INT_VDD_BIAS_TURBO attribute
typedef int8_t VOLTAGE_INT_VDD_BIAS_TURBO_ATTR;
typedef int8_t ATTR_VOLTAGE_INT_VDD_BIAS_TURBO_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_INT_VDD_BIAS_NOMINAL attribute
typedef int8_t VOLTAGE_INT_VDD_BIAS_NOMINAL_ATTR;
typedef int8_t ATTR_VOLTAGE_INT_VDD_BIAS_NOMINAL_type;

// Type aliases and/or sizes for ATTR_VOLTAGE_INT_VDD_BIAS_POWERSAVE attribute
typedef int8_t VOLTAGE_INT_VDD_BIAS_POWERSAVE_ATTR;
typedef int8_t ATTR_VOLTAGE_INT_VDD_BIAS_POWERSAVE_type;

// Type aliases and/or sizes for ATTR_STOP4_DISABLE attribute
typedef uint8_t STOP4_DISABLE_ATTR;
typedef uint8_t ATTR_STOP4_DISABLE_type;

// Type aliases and/or sizes for ATTR_STOP8_DISABLE attribute
typedef uint8_t STOP8_DISABLE_ATTR;
typedef uint8_t ATTR_STOP8_DISABLE_type;

// Type aliases and/or sizes for ATTR_STOP11_DISABLE attribute
typedef uint8_t STOP11_DISABLE_ATTR;
typedef uint8_t ATTR_STOP11_DISABLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_CORE_PERIODIC_QUIESCE_DISABLE attribute
typedef uint8_t SYSTEM_CORE_PERIODIC_QUIESCE_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_CORE_PERIODIC_QUIESCE_DISABLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_POUNDV_VALIDITY_HALT_DISABLE attribute
typedef uint8_t SYSTEM_POUNDV_VALIDITY_HALT_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_POUNDV_VALIDITY_HALT_DISABLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_WOF_DISABLE attribute
typedef uint8_t SYSTEM_WOF_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_WOF_DISABLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_WOV_UNDERV_DISABLE attribute
typedef uint8_t SYSTEM_WOV_UNDERV_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_WOV_UNDERV_DISABLE_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_FORCE attribute
typedef uint8_t WOV_UNDERV_FORCE_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_FORCE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_WOV_OVERV_DISABLE attribute
typedef uint8_t SYSTEM_WOV_OVERV_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_WOV_OVERV_DISABLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_IVRM_DISABLE attribute
typedef uint8_t SYSTEM_IVRM_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_IVRM_DISABLE_type;

// Type aliases and/or sizes for ATTR_WOF_ENABLE_FRATIO attribute
typedef uint8_t WOF_ENABLE_FRATIO_ATTR;
typedef uint8_t ATTR_WOF_ENABLE_FRATIO_type;

// Type aliases and/or sizes for ATTR_WOF_ENABLE_VRATIO attribute
typedef uint8_t WOF_ENABLE_VRATIO_ATTR;
typedef uint8_t ATTR_WOF_ENABLE_VRATIO_type;

// Type aliases and/or sizes for ATTR_WOF_VRATIO_SELECT attribute
typedef uint8_t WOF_VRATIO_SELECT_ATTR;
typedef uint8_t ATTR_WOF_VRATIO_SELECT_type;

// Type aliases and/or sizes for ATTR_WOV_SAMPLE_125US attribute
typedef uint32_t WOV_SAMPLE_125US_ATTR;
typedef uint32_t ATTR_WOV_SAMPLE_125US_type;

// Type aliases and/or sizes for ATTR_WOV_MAX_DROOP_10THPCT attribute
typedef uint32_t WOV_MAX_DROOP_10THPCT_ATTR;
typedef uint32_t ATTR_WOV_MAX_DROOP_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_VMIN_MV attribute
typedef uint16_t WOV_UNDERV_VMIN_MV_ATTR;
typedef uint16_t ATTR_WOV_UNDERV_VMIN_MV_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_LIMIT_ALGORITHM attribute
typedef uint8_t WOV_UNDERV_LIMIT_ALGORITHM_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_LIMIT_ALGORITHM_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_LIMIT_MV attribute
typedef uint16_t WOV_UNDERV_LIMIT_MV_ATTR;
typedef uint16_t ATTR_WOV_UNDERV_LIMIT_MV_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_LIMIT_10THPCT attribute
typedef uint8_t WOV_UNDERV_LIMIT_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_LIMIT_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_OVERV_VMAX_SETPOINT_MV attribute
typedef uint16_t WOV_OVERV_VMAX_SETPOINT_MV_ATTR;
typedef uint16_t ATTR_WOV_OVERV_VMAX_SETPOINT_MV_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_STEP_INCR_10THPCT attribute
typedef uint8_t WOV_UNDERV_STEP_INCR_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_STEP_INCR_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_STEP_DECR_10THPCT attribute
typedef uint8_t WOV_UNDERV_STEP_DECR_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_STEP_DECR_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_MAX_10THPCT attribute
typedef uint8_t WOV_UNDERV_MAX_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_MAX_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_PERF_LOSS_THRESH_10THPCT attribute
typedef uint8_t WOV_UNDERV_PERF_LOSS_THRESH_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_PERF_LOSS_THRESH_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_OVERV_STEP_INCR_10THPCT attribute
typedef uint8_t WOV_OVERV_STEP_INCR_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_OVERV_STEP_INCR_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_OVERV_STEP_DECR_10THPCT attribute
typedef uint8_t WOV_OVERV_STEP_DECR_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_OVERV_STEP_DECR_10THPCT_type;

// Type aliases and/or sizes for ATTR_WOV_OVERV_MAX_10THPCT attribute
typedef uint8_t WOV_OVERV_MAX_10THPCT_ATTR;
typedef uint8_t ATTR_WOV_OVERV_MAX_10THPCT_type;

// Type aliases and/or sizes for ATTR_PBAX_GROUPID attribute
typedef uint8_t PBAX_GROUPID_ATTR;
typedef uint8_t ATTR_PBAX_GROUPID_type;

// Type aliases and/or sizes for ATTR_PBAX_CHIPID attribute
typedef uint8_t PBAX_CHIPID_ATTR;
typedef uint8_t ATTR_PBAX_CHIPID_type;

// Type aliases and/or sizes for ATTR_PBAX_BRDCST_ID_VECTOR attribute
typedef uint8_t PBAX_BRDCST_ID_VECTOR_ATTR;
typedef uint8_t ATTR_PBAX_BRDCST_ID_VECTOR_type;

// Type aliases and/or sizes for ATTR_POUNDV_BUCKET_NUM_OVERRIDE attribute
typedef uint8_t POUNDV_BUCKET_NUM_OVERRIDE_ATTR;
typedef uint8_t ATTR_POUNDV_BUCKET_NUM_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_POUNDV_BUCKET_NUM attribute
typedef uint8_t POUNDV_BUCKET_NUM_ATTR;
typedef uint8_t ATTR_POUNDV_BUCKET_NUM_type;

// Type aliases and/or sizes for ATTR_DUMP_STOP_INFO_SUPPRESS_ERROR_TRACE attribute
typedef uint8_t DUMP_STOP_INFO_SUPPRESS_ERROR_TRACE_ATTR;
typedef uint8_t ATTR_DUMP_STOP_INFO_SUPPRESS_ERROR_TRACE_type;

// Type aliases and/or sizes for ATTR_DUMP_STOP_INFO_ENABLE_ERRORLOG attribute
typedef uint8_t DUMP_STOP_INFO_ENABLE_ERRORLOG_ATTR;
typedef uint8_t ATTR_DUMP_STOP_INFO_ENABLE_ERRORLOG_type;

// Type aliases and/or sizes for ATTR_SYSTEM_VDM_DISABLE attribute
typedef uint8_t SYSTEM_VDM_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_VDM_DISABLE_type;

// Type aliases and/or sizes for ATTR_VDM_DROOP_SMALL_OVERRIDE attribute
typedef uint8_t VDM_DROOP_SMALL_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_DROOP_SMALL_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_DROOP_SMALL_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_DROOP_LARGE_OVERRIDE attribute
typedef uint8_t VDM_DROOP_LARGE_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_DROOP_LARGE_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_DROOP_LARGE_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_DROOP_EXTREME_OVERRIDE attribute
typedef uint8_t VDM_DROOP_EXTREME_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_DROOP_EXTREME_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_DROOP_EXTREME_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_OVERVOLT_OVERRIDE attribute
typedef uint8_t VDM_OVERVOLT_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_OVERVOLT_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_OVERVOLT_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_SMALL_FREQ_DROP_N_S_OVERRIDE attribute
typedef uint8_t VDM_SMALL_FREQ_DROP_N_S_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_SMALL_FREQ_DROP_N_S_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_SMALL_FREQ_DROP_N_S_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_LARGE_FREQ_DROP_N_L_OVERRIDE attribute
typedef uint8_t VDM_LARGE_FREQ_DROP_N_L_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_LARGE_FREQ_DROP_N_L_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_LARGE_FREQ_DROP_N_L_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_FREQ_RETURN_L_S_OVERRIDE attribute
typedef uint8_t VDM_FREQ_RETURN_L_S_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_FREQ_RETURN_L_S_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_FREQ_RETURN_L_S_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_FREQ_RETURN_S_N_OVERRIDE attribute
typedef uint8_t VDM_FREQ_RETURN_S_N_OVERRIDE_ATTR[5];
typedef uint8_t ATTR_VDM_FREQ_RETURN_S_N_OVERRIDE_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_FREQ_RETURN_S_N_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_EXTREME_THOTTLE_ENABLE attribute
typedef uint8_t VDM_EXTREME_THOTTLE_ENABLE_ATTR;
typedef uint8_t ATTR_VDM_EXTREME_THOTTLE_ENABLE_type;

// Type aliases and/or sizes for ATTR_VDM_FMAX_OVERRIDE_KHZ attribute
typedef uint16_t VDM_FMAX_OVERRIDE_KHZ_ATTR[5];
typedef uint16_t ATTR_VDM_FMAX_OVERRIDE_KHZ_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 5> ATTR_VDM_FMAX_OVERRIDE_KHZ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_FMIN_OVERRIDE_KHZ attribute
typedef uint16_t VDM_FMIN_OVERRIDE_KHZ_ATTR[5];
typedef uint16_t ATTR_VDM_FMIN_OVERRIDE_KHZ_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 5> ATTR_VDM_FMIN_OVERRIDE_KHZ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_VID_COMPARE_OVERRIDE_MV attribute
typedef uint8_t VDM_VID_COMPARE_OVERRIDE_MV_ATTR[5];
typedef uint8_t ATTR_VDM_VID_COMPARE_OVERRIDE_MV_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_VDM_VID_COMPARE_OVERRIDE_MV_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VDM_VID_COMPARE_BIAS_0P5PCT attribute
typedef int8_t VDM_VID_COMPARE_BIAS_0P5PCT_ATTR[4];
typedef int8_t ATTR_VDM_VID_COMPARE_BIAS_0P5PCT_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<int8_t, 4> ATTR_VDM_VID_COMPARE_BIAS_0P5PCT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_DPLL_VDM_RESPONSE attribute
typedef uint8_t DPLL_VDM_RESPONSE_ATTR;
typedef uint8_t ATTR_DPLL_VDM_RESPONSE_type;

// Type aliases and/or sizes for ATTR_IVRM_DEADZONE_MV attribute
typedef uint8_t IVRM_DEADZONE_MV_ATTR;
typedef uint8_t ATTR_IVRM_DEADZONE_MV_type;

// Type aliases and/or sizes for ATTR_IVRM_STRENGTH_LOOKUP attribute
typedef uint8_t IVRM_STRENGTH_LOOKUP_ATTR[64];
typedef uint8_t ATTR_IVRM_STRENGTH_LOOKUP_type[64];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 64> ATTR_IVRM_STRENGTH_LOOKUP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_IVRM_VIN_MULTIPLIER attribute
typedef uint8_t IVRM_VIN_MULTIPLIER_ATTR[64];
typedef uint8_t ATTR_IVRM_VIN_MULTIPLIER_type[64];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 64> ATTR_IVRM_VIN_MULTIPLIER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_IVRM_VIN_MAX_MV attribute
typedef uint16_t IVRM_VIN_MAX_MV_ATTR;
typedef uint16_t ATTR_IVRM_VIN_MAX_MV_type;

// Type aliases and/or sizes for ATTR_IVRM_STEP_DELAY_NS attribute
typedef uint16_t IVRM_STEP_DELAY_NS_ATTR;
typedef uint16_t ATTR_IVRM_STEP_DELAY_NS_type;

// Type aliases and/or sizes for ATTR_IVRM_STABILIZATION_DELAY_NS attribute
typedef uint16_t IVRM_STABILIZATION_DELAY_NS_ATTR;
typedef uint16_t ATTR_IVRM_STABILIZATION_DELAY_NS_type;

// Type aliases and/or sizes for ATTR_TDP_RDP_CURRENT_FACTOR attribute
typedef uint32_t TDP_RDP_CURRENT_FACTOR_ATTR;
typedef uint32_t ATTR_TDP_RDP_CURRENT_FACTOR_type;

// Type aliases and/or sizes for ATTR_SYSTEM_RESCLK_DISABLE attribute
typedef uint8_t SYSTEM_RESCLK_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_RESCLK_DISABLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_RESCLK_STEP_DELAY attribute
typedef uint16_t SYSTEM_RESCLK_STEP_DELAY_ATTR;
typedef uint16_t ATTR_SYSTEM_RESCLK_STEP_DELAY_type;

// Type aliases and/or sizes for ATTR_SYSTEM_RESCLK_FREQ_REGIONS attribute
typedef uint16_t SYSTEM_RESCLK_FREQ_REGIONS_ATTR[8];
typedef uint16_t ATTR_SYSTEM_RESCLK_FREQ_REGIONS_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 8> ATTR_SYSTEM_RESCLK_FREQ_REGIONS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SYSTEM_RESCLK_FREQ_REGION_INDEX attribute
typedef uint8_t SYSTEM_RESCLK_FREQ_REGION_INDEX_ATTR[8];
typedef uint8_t ATTR_SYSTEM_RESCLK_FREQ_REGION_INDEX_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 8> ATTR_SYSTEM_RESCLK_FREQ_REGION_INDEX_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SYSTEM_RESCLK_VALUE attribute
typedef uint16_t SYSTEM_RESCLK_VALUE_ATTR[64];
typedef uint16_t ATTR_SYSTEM_RESCLK_VALUE_type[64];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 64> ATTR_SYSTEM_RESCLK_VALUE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SYSTEM_RESCLK_L3_VALUE attribute
typedef uint8_t SYSTEM_RESCLK_L3_VALUE_ATTR[4];
typedef uint8_t ATTR_SYSTEM_RESCLK_L3_VALUE_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_SYSTEM_RESCLK_L3_VALUE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SYSTEM_RESCLK_L3_VOLTAGE_THRESHOLD_MV attribute
typedef uint16_t SYSTEM_RESCLK_L3_VOLTAGE_THRESHOLD_MV_ATTR;
typedef uint16_t ATTR_SYSTEM_RESCLK_L3_VOLTAGE_THRESHOLD_MV_type;

// Type aliases and/or sizes for ATTR_SYSTEM_RING_DBG_MODE attribute
typedef uint8_t SYSTEM_RING_DBG_MODE_ATTR;
typedef uint8_t ATTR_SYSTEM_RING_DBG_MODE_type;

// Type aliases and/or sizes for ATTR_PERF_24x7_INVOCATION_TIME_MS attribute
typedef uint8_t PERF_24x7_INVOCATION_TIME_MS_ATTR;
typedef uint8_t ATTR_PERF_24x7_INVOCATION_TIME_MS_type;

// Type aliases and/or sizes for ATTR_CME_INSTRUCTION_TRACE_ENABLE attribute
typedef uint8_t CME_INSTRUCTION_TRACE_ENABLE_ATTR;
typedef uint8_t ATTR_CME_INSTRUCTION_TRACE_ENABLE_type;

// Type aliases and/or sizes for ATTR_CME_CHTM_TRACE_ENABLE attribute
typedef uint8_t CME_CHTM_TRACE_ENABLE_ATTR;
typedef uint8_t ATTR_CME_CHTM_TRACE_ENABLE_type;

// Type aliases and/or sizes for ATTR_CME_CHTM_TRACE_MEMORY_CONFIG attribute
typedef uint64_t CME_CHTM_TRACE_MEMORY_CONFIG_ATTR;
typedef uint64_t ATTR_CME_CHTM_TRACE_MEMORY_CONFIG_type;

// Type aliases and/or sizes for ATTR_PGPE_HCODE_FUNCTION_ENABLE attribute
typedef uint8_t PGPE_HCODE_FUNCTION_ENABLE_ATTR;
typedef uint8_t ATTR_PGPE_HCODE_FUNCTION_ENABLE_type;

// Type aliases and/or sizes for ATTR_POUND_W_STATIC_DATA_ENABLE attribute
typedef uint8_t POUND_W_STATIC_DATA_ENABLE_ATTR;
typedef uint8_t ATTR_POUND_W_STATIC_DATA_ENABLE_type;

// Type aliases and/or sizes for ATTR_AUX_FUNC_INVOCATION_TIME_MS attribute
typedef uint8_t AUX_FUNC_INVOCATION_TIME_MS_ATTR;
typedef uint8_t ATTR_AUX_FUNC_INVOCATION_TIME_MS_type;

// Type aliases and/or sizes for ATTR_SYS_VFRT_STATIC_DATA_ENABLE attribute
typedef uint8_t SYS_VFRT_STATIC_DATA_ENABLE_ATTR;
typedef uint8_t ATTR_SYS_VFRT_STATIC_DATA_ENABLE_type;

// Type aliases and/or sizes for ATTR_NEST_LEAKAGE_PERCENT attribute
typedef uint8_t NEST_LEAKAGE_PERCENT_ATTR;
typedef uint8_t ATTR_NEST_LEAKAGE_PERCENT_type;

// Type aliases and/or sizes for ATTR_SYSTEM_PSTATES_MODE attribute
typedef uint8_t SYSTEM_PSTATES_MODE_ATTR;
typedef uint8_t ATTR_SYSTEM_PSTATES_MODE_type;

// Type aliases and/or sizes for ATTR_CORE_THROTTLE_ASSERT_COUNT attribute
typedef uint32_t CORE_THROTTLE_ASSERT_COUNT_ATTR;
typedef uint32_t ATTR_CORE_THROTTLE_ASSERT_COUNT_type;

// Type aliases and/or sizes for ATTR_CORE_THROTTLE_DEASSERT_COUNT attribute
typedef uint32_t CORE_THROTTLE_DEASSERT_COUNT_ATTR;
typedef uint32_t ATTR_CORE_THROTTLE_DEASSERT_COUNT_type;

// Type aliases and/or sizes for ATTR_PM_GLOBAL_FIR_TRACE_EN attribute
typedef uint8_t PM_GLOBAL_FIR_TRACE_EN_ATTR;
typedef uint8_t ATTR_PM_GLOBAL_FIR_TRACE_EN_type;

// Type aliases and/or sizes for ATTR_FREQ_PB_MHZ_POUNDV_FALLBACK attribute
typedef uint32_t FREQ_PB_MHZ_POUNDV_FALLBACK_ATTR;
typedef uint32_t ATTR_FREQ_PB_MHZ_POUNDV_FALLBACK_type;

// Type aliases and/or sizes for ATTR_SKIP_WAKEUP attribute
typedef uint8_t SKIP_WAKEUP_ATTR;
typedef uint8_t ATTR_SKIP_WAKEUP_type;

// Type aliases and/or sizes for ATTR_PM_MALF_ALERT_ENABLE attribute
typedef uint8_t PM_MALF_ALERT_ENABLE_ATTR;
typedef uint8_t ATTR_PM_MALF_ALERT_ENABLE_type;

// Type aliases and/or sizes for ATTR_PM_RESET_FFDC_ENABLE attribute
typedef uint8_t PM_RESET_FFDC_ENABLE_ATTR;
typedef uint8_t ATTR_PM_RESET_FFDC_ENABLE_type;

// Type aliases and/or sizes for ATTR_NEST_24x7_PERF_ACCUM_ENABLE attribute
typedef uint8_t NEST_24x7_PERF_ACCUM_ENABLE_ATTR;
typedef uint8_t ATTR_NEST_24x7_PERF_ACCUM_ENABLE_type;

// Type aliases and/or sizes for ATTR_UNSECURE_HOMER_ADDRESS attribute
typedef uint64_t UNSECURE_HOMER_ADDRESS_ATTR;
typedef uint64_t ATTR_UNSECURE_HOMER_ADDRESS_type;

// Type aliases and/or sizes for ATTR_UNSECURE_HOMER_SIZE attribute
typedef uint32_t UNSECURE_HOMER_SIZE_ATTR;
typedef uint32_t ATTR_UNSECURE_HOMER_SIZE_type;

// Type aliases and/or sizes for ATTR_PMCR_MOST_RECENT_MODE attribute
typedef uint8_t PMCR_MOST_RECENT_MODE_ATTR;
typedef uint8_t ATTR_PMCR_MOST_RECENT_MODE_type;

// Type aliases and/or sizes for ATTR_ENABLED_OBUS_BRICKS attribute
typedef uint64_t ENABLED_OBUS_BRICKS_ATTR;
typedef uint64_t ATTR_ENABLED_OBUS_BRICKS_type;

// Type aliases and/or sizes for ATTR_IO_XBUS_DCCAL_FLAGS attribute
typedef uint8_t IO_XBUS_DCCAL_FLAGS_ATTR;
typedef uint8_t ATTR_IO_XBUS_DCCAL_FLAGS_type;

// Type aliases and/or sizes for ATTR_IO_X_DEBUG attribute
typedef uint8_t IO_X_DEBUG_ATTR;
typedef uint8_t ATTR_IO_X_DEBUG_type;

// Type aliases and/or sizes for ATTR_IO_X_MFG_CHK attribute
typedef uint8_t IO_X_MFG_CHK_ATTR;
typedef uint8_t ATTR_IO_X_MFG_CHK_type;

// Type aliases and/or sizes for ATTR_IO_X_MFG_MIN_EYE_WIDTH attribute
typedef uint8_t IO_X_MFG_MIN_EYE_WIDTH_ATTR;
typedef uint8_t ATTR_IO_X_MFG_MIN_EYE_WIDTH_type;

// Type aliases and/or sizes for ATTR_IO_XBUS_MASTER_MODE attribute
typedef uint8_t IO_XBUS_MASTER_MODE_ATTR;
typedef uint8_t ATTR_IO_XBUS_MASTER_MODE_type;

// Type aliases and/or sizes for ATTR_IO_XBUS_TX_MARGIN_RATIO attribute
typedef uint8_t IO_XBUS_TX_MARGIN_RATIO_ATTR;
typedef uint8_t ATTR_IO_XBUS_TX_MARGIN_RATIO_type;

// Type aliases and/or sizes for ATTR_IO_XBUS_TX_FFE_PRECURSOR attribute
typedef uint8_t IO_XBUS_TX_FFE_PRECURSOR_ATTR;
typedef uint8_t ATTR_IO_XBUS_TX_FFE_PRECURSOR_type;

// Type aliases and/or sizes for ATTR_IO_XBUS_CHAN_EQ attribute
typedef uint8_t IO_XBUS_CHAN_EQ_ATTR;
typedef uint8_t ATTR_IO_XBUS_CHAN_EQ_type;

// Type aliases and/or sizes for ATTR_IO_XBUS_GRP0_PRE_BAD_LANE_DATA attribute
typedef uint32_t IO_XBUS_GRP0_PRE_BAD_LANE_DATA_ATTR;
typedef uint32_t ATTR_IO_XBUS_GRP0_PRE_BAD_LANE_DATA_type;

// Type aliases and/or sizes for ATTR_IO_XBUS_GRP1_PRE_BAD_LANE_DATA attribute
typedef uint32_t IO_XBUS_GRP1_PRE_BAD_LANE_DATA_ATTR;
typedef uint32_t ATTR_IO_XBUS_GRP1_PRE_BAD_LANE_DATA_type;

// Type aliases and/or sizes for ATTR_DO_MSS_WR_VREF attribute
typedef uint8_t DO_MSS_WR_VREF_ATTR;
typedef uint8_t ATTR_DO_MSS_WR_VREF_type;

// Type aliases and/or sizes for ATTR_SKIP_HW_VREF_CAL attribute
typedef uint8_t SKIP_HW_VREF_CAL_ATTR;
typedef uint8_t ATTR_SKIP_HW_VREF_CAL_type;

// Type aliases and/or sizes for ATTR_SKIP_RD_VREF_VREFSENSE_OVERRIDE attribute
typedef uint8_t SKIP_RD_VREF_VREFSENSE_OVERRIDE_ATTR;
typedef uint8_t ATTR_SKIP_RD_VREF_VREFSENSE_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_DO_MSS_VREF_DAC attribute
typedef uint8_t DO_MSS_VREF_DAC_ATTR;
typedef uint8_t ATTR_DO_MSS_VREF_DAC_type;

// Type aliases and/or sizes for ATTR_DO_MSS_TRAINING_BAD_BITS attribute
typedef uint8_t DO_MSS_TRAINING_BAD_BITS_ATTR;
typedef uint8_t ATTR_DO_MSS_TRAINING_BAD_BITS_type;

// Type aliases and/or sizes for ATTR_MSS_RUN_DCD_CALIBRATION attribute
typedef uint8_t MSS_RUN_DCD_CALIBRATION_ATTR;
typedef uint8_t ATTR_MSS_RUN_DCD_CALIBRATION_type;

// Type aliases and/or sizes for ATTR_OMI_INBAND_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t OMI_INBAND_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_OMI_INBAND_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_NVDIMM_SLOT_POSITION attribute
typedef uint8_t MSS_MRW_NVDIMM_SLOT_POSITION_ATTR;
typedef uint8_t ATTR_MSS_MRW_NVDIMM_SLOT_POSITION_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_UNSUPPORTED_RANK_CONFIG attribute
typedef uint64_t MSS_MRW_UNSUPPORTED_RANK_CONFIG_ATTR[2];
typedef uint64_t ATTR_MSS_MRW_UNSUPPORTED_RANK_CONFIG_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 2> ATTR_MSS_MRW_UNSUPPORTED_RANK_CONFIG_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_MT attribute
typedef uint8_t VPD_OVERRIDE_MT_ATTR[255];
typedef uint8_t ATTR_VPD_OVERRIDE_MT_type[255];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 255> ATTR_VPD_OVERRIDE_MT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_MT_ENABLE attribute
typedef uint8_t VPD_OVERRIDE_MT_ENABLE_ATTR;
typedef uint8_t ATTR_VPD_OVERRIDE_MT_ENABLE_type;

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_MR attribute
typedef uint8_t VPD_OVERRIDE_MR_ATTR[255];
typedef uint8_t ATTR_VPD_OVERRIDE_MR_type[255];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 255> ATTR_VPD_OVERRIDE_MR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_MR_ENABLE attribute
typedef uint8_t VPD_OVERRIDE_MR_ENABLE_ATTR;
typedef uint8_t ATTR_VPD_OVERRIDE_MR_ENABLE_type;

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_DQ attribute
typedef uint8_t VPD_OVERRIDE_DQ_ATTR[160];
typedef uint8_t ATTR_VPD_OVERRIDE_DQ_type[160];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 160> ATTR_VPD_OVERRIDE_DQ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_DQ_ENABLE attribute
typedef uint8_t VPD_OVERRIDE_DQ_ENABLE_ATTR;
typedef uint8_t ATTR_VPD_OVERRIDE_DQ_ENABLE_type;

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_CK attribute
typedef uint8_t VPD_OVERRIDE_CK_ATTR[16];
typedef uint8_t ATTR_VPD_OVERRIDE_CK_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 16> ATTR_VPD_OVERRIDE_CK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_CK_ENABLE attribute
typedef uint8_t VPD_OVERRIDE_CK_ENABLE_ATTR;
typedef uint8_t ATTR_VPD_OVERRIDE_CK_ENABLE_type;

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_MW attribute
typedef uint8_t VPD_OVERRIDE_MW_ATTR[255];
typedef uint8_t ATTR_VPD_OVERRIDE_MW_type[255];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 255> ATTR_VPD_OVERRIDE_MW_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_VPD_OVERRIDE_MW_ENABLE attribute
typedef uint8_t VPD_OVERRIDE_MW_ENABLE_ATTR;
typedef uint8_t ATTR_VPD_OVERRIDE_MW_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_OVERRIDE_MEMDIAGS_BCMODE attribute
typedef uint8_t MSS_OVERRIDE_MEMDIAGS_BCMODE_ATTR;
typedef uint8_t ATTR_MSS_OVERRIDE_MEMDIAGS_BCMODE_type;

// Type aliases and/or sizes for ATTR_EFF_DRAM_GEN attribute
typedef uint8_t EFF_DRAM_GEN_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_GEN_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_GEN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_TYPE attribute
typedef uint8_t EFF_DIMM_TYPE_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_TYPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_HYBRID_MEMORY_TYPE attribute
typedef uint8_t EFF_HYBRID_MEMORY_TYPE_ATTR[2][2];
typedef uint8_t ATTR_EFF_HYBRID_MEMORY_TYPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_HYBRID_MEMORY_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_HYBRID attribute
typedef uint8_t EFF_HYBRID_ATTR[2][2];
typedef uint8_t ATTR_EFF_HYBRID_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_HYBRID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_DENSITY attribute
typedef uint8_t EFF_DRAM_DENSITY_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_DENSITY_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_DENSITY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_BANK_BITS attribute
typedef uint8_t EFF_DRAM_BANK_BITS_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_BANK_BITS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_BANK_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_BANK_GROUP_BITS attribute
typedef uint8_t EFF_DRAM_BANK_GROUP_BITS_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_BANK_GROUP_BITS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_BANK_GROUP_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_COLUMN_BITS attribute
typedef uint8_t EFF_DRAM_COLUMN_BITS_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_COLUMN_BITS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_COLUMN_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_ROW_BITS attribute
typedef uint8_t EFF_DRAM_ROW_BITS_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_ROW_BITS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_ROW_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_PRIM_STACK_TYPE attribute
typedef uint8_t EFF_PRIM_STACK_TYPE_ATTR[2][2];
typedef uint8_t ATTR_EFF_PRIM_STACK_TYPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_PRIM_STACK_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_PPR attribute
typedef uint8_t EFF_DRAM_PPR_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_PPR_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_PPR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_SOFT_PPR attribute
typedef uint8_t EFF_DRAM_SOFT_PPR_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_SOFT_PPR_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_SOFT_PPR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRCD attribute
typedef uint8_t EFF_DRAM_TRCD_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRCD_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRCD_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRP attribute
typedef uint8_t EFF_DRAM_TRP_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRAS attribute
typedef uint8_t EFF_DRAM_TRAS_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRAS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRAS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRC attribute
typedef uint8_t EFF_DRAM_TRC_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRC_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRFC attribute
typedef uint16_t EFF_DRAM_TRFC_ATTR[2];
typedef uint16_t ATTR_EFF_DRAM_TRFC_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_EFF_DRAM_TRFC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TFAW attribute
typedef uint8_t EFF_DRAM_TFAW_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TFAW_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TFAW_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRRD_S attribute
typedef uint8_t EFF_DRAM_TRRD_S_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRRD_S_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRRD_S_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRRD_L attribute
typedef uint8_t EFF_DRAM_TRRD_L_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRRD_L_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRRD_L_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TCCD_L attribute
typedef uint8_t EFF_DRAM_TCCD_L_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TCCD_L_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TCCD_L_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TWR attribute
typedef uint8_t EFF_DRAM_TWR_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TWR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TWR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TWTR_S attribute
typedef uint8_t EFF_DRAM_TWTR_S_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TWTR_S_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TWTR_S_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TWTR_L attribute
typedef uint8_t EFF_DRAM_TWTR_L_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TWTR_L_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TWTR_L_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TMAW attribute
typedef uint16_t EFF_DRAM_TMAW_ATTR[2];
typedef uint16_t ATTR_EFF_DRAM_TMAW_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_EFF_DRAM_TMAW_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_WIDTH attribute
typedef uint8_t EFF_DRAM_WIDTH_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_WIDTH_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_WIDTH_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_RANK_MIX attribute
typedef uint8_t EFF_DRAM_RANK_MIX_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_RANK_MIX_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_RANK_MIX_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_NUM_RANKS_PER_DIMM attribute
typedef uint8_t EFF_NUM_RANKS_PER_DIMM_ATTR[2][2];
typedef uint8_t ATTR_EFF_NUM_RANKS_PER_DIMM_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_NUM_RANKS_PER_DIMM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_REGISTER_TYPE attribute
typedef uint8_t EFF_REGISTER_TYPE_ATTR[2][2];
typedef uint8_t ATTR_EFF_REGISTER_TYPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_REGISTER_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_MFG_ID attribute
typedef uint16_t EFF_DRAM_MFG_ID_ATTR[2][2];
typedef uint16_t ATTR_EFF_DRAM_MFG_ID_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 2>, 2> ATTR_EFF_DRAM_MFG_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_RCD_MFG_ID attribute
typedef uint16_t EFF_RCD_MFG_ID_ATTR[2][2];
typedef uint16_t ATTR_EFF_RCD_MFG_ID_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 2>, 2> ATTR_EFF_RCD_MFG_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_REGISTER_REV attribute
typedef uint8_t EFF_REGISTER_REV_ATTR[2][2];
typedef uint8_t ATTR_EFF_REGISTER_REV_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_REGISTER_REV_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_PACKAGE_RANK_MAP attribute
typedef uint8_t EFF_PACKAGE_RANK_MAP_ATTR[2][2][18];
typedef uint8_t ATTR_EFF_PACKAGE_RANK_MAP_type[2][2][18];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 18>, 2>, 2> ATTR_EFF_PACKAGE_RANK_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_NIBBLE_MAP attribute
typedef uint8_t EFF_NIBBLE_MAP_ATTR[2][2][18];
typedef uint8_t ATTR_EFF_NIBBLE_MAP_type[2][2][18];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 18>, 2>, 2> ATTR_EFF_NIBBLE_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_ROW_REPAIR_SUPPORTED_MRW attribute
typedef uint8_t ROW_REPAIR_SUPPORTED_MRW_ATTR;
typedef uint8_t ATTR_ROW_REPAIR_SUPPORTED_MRW_type;

// Type aliases and/or sizes for ATTR_EI_BUS_TX_MSBSWAP attribute
typedef uint8_t EI_BUS_TX_MSBSWAP_ATTR;
typedef uint8_t ATTR_EI_BUS_TX_MSBSWAP_type;

// Type aliases and/or sizes for ATTR_PROC_NPU_REGION_ENABLED attribute
typedef uint8_t PROC_NPU_REGION_ENABLED_ATTR;
typedef uint8_t ATTR_PROC_NPU_REGION_ENABLED_type;

// Type aliases and/or sizes for ATTR_CLOCK_PLL_MUX_TOD attribute
typedef uint8_t CLOCK_PLL_MUX_TOD_ATTR;
typedef uint8_t ATTR_CLOCK_PLL_MUX_TOD_type;

// Type aliases and/or sizes for ATTR_CLOCK_PLL_MUX attribute
typedef uint32_t CLOCK_PLL_MUX_ATTR;
typedef uint32_t ATTR_CLOCK_PLL_MUX_type;

// Type aliases and/or sizes for ATTR_CLOCK_PLL_MUX0 attribute
typedef uint8_t CLOCK_PLL_MUX0_ATTR;
typedef uint8_t ATTR_CLOCK_PLL_MUX0_type;

// Type aliases and/or sizes for ATTR_I2C_BUS_DIV_REF attribute
typedef uint16_t I2C_BUS_DIV_REF_ATTR;
typedef uint16_t ATTR_I2C_BUS_DIV_REF_type;

// Type aliases and/or sizes for ATTR_EQ_GARD attribute
typedef uint8_t EQ_GARD_ATTR;
typedef uint8_t ATTR_EQ_GARD_type;

// Type aliases and/or sizes for ATTR_EC_GARD attribute
typedef uint32_t EC_GARD_ATTR;
typedef uint32_t ATTR_EC_GARD_type;

// Type aliases and/or sizes for ATTR_SBE_RUNTIME_MODE attribute
typedef uint8_t SBE_RUNTIME_MODE_ATTR;
typedef uint8_t ATTR_SBE_RUNTIME_MODE_type;

// Type aliases and/or sizes for ATTR_IS_SP_MODE attribute
typedef uint8_t IS_SP_MODE_ATTR;
typedef uint8_t ATTR_IS_SP_MODE_type;

// Type aliases and/or sizes for ATTR_SBE_FFDC_ENABLE attribute
typedef uint8_t SBE_FFDC_ENABLE_ATTR;
typedef uint8_t ATTR_SBE_FFDC_ENABLE_type;

// Type aliases and/or sizes for ATTR_SBE_INTERNAL_FFDC_ENABLE attribute
typedef uint8_t SBE_INTERNAL_FFDC_ENABLE_ATTR;
typedef uint8_t ATTR_SBE_INTERNAL_FFDC_ENABLE_type;

// Type aliases and/or sizes for ATTR_NEST_PLL_BUCKET attribute
typedef uint8_t NEST_PLL_BUCKET_ATTR;
typedef uint8_t ATTR_NEST_PLL_BUCKET_type;

// Type aliases and/or sizes for ATTR_FILTER_PLL_BUCKET attribute
typedef uint8_t FILTER_PLL_BUCKET_ATTR;
typedef uint8_t ATTR_FILTER_PLL_BUCKET_type;

// Type aliases and/or sizes for ATTR_MRW_FILTER_PLL_BUCKET attribute
typedef uint8_t MRW_FILTER_PLL_BUCKET_ATTR;
typedef uint8_t ATTR_MRW_FILTER_PLL_BUCKET_type;

// Type aliases and/or sizes for ATTR_FORCE_SYNC_SS_PLL_SPREAD attribute
typedef uint8_t FORCE_SYNC_SS_PLL_SPREAD_ATTR;
typedef uint8_t ATTR_FORCE_SYNC_SS_PLL_SPREAD_type;

// Type aliases and/or sizes for ATTR_BOOT_FREQ_MULT attribute
typedef uint16_t BOOT_FREQ_MULT_ATTR;
typedef uint16_t ATTR_BOOT_FREQ_MULT_type;

// Type aliases and/or sizes for ATTR_RISK_LEVEL attribute
typedef uint8_t RISK_LEVEL_ATTR;
typedef uint8_t ATTR_RISK_LEVEL_type;

// Type aliases and/or sizes for ATTR_DISABLE_HBBL_VECTORS attribute
typedef uint8_t DISABLE_HBBL_VECTORS_ATTR;
typedef uint8_t ATTR_DISABLE_HBBL_VECTORS_type;

// Type aliases and/or sizes for ATTR_BACKUP_SEEPROM_SELECT attribute
typedef uint8_t BACKUP_SEEPROM_SELECT_ATTR;
typedef uint8_t ATTR_BACKUP_SEEPROM_SELECT_type;

// Type aliases and/or sizes for ATTR_BOOT_FLAGS attribute
typedef uint32_t BOOT_FLAGS_ATTR;
typedef uint32_t ATTR_BOOT_FLAGS_type;

// Type aliases and/or sizes for ATTR_BOOT_FREQ_MHZ attribute
typedef uint32_t BOOT_FREQ_MHZ_ATTR;
typedef uint32_t ATTR_BOOT_FREQ_MHZ_type;

// Type aliases and/or sizes for ATTR_BRANCH_PIBMEM_ADDR attribute
typedef uint8_t BRANCH_PIBMEM_ADDR_ATTR;
typedef uint8_t ATTR_BRANCH_PIBMEM_ADDR_type;

// Type aliases and/or sizes for ATTR_CHIP_REGIONS_TO_ENABLE attribute
typedef uint32_t CHIP_REGIONS_TO_ENABLE_ATTR;
typedef uint32_t ATTR_CHIP_REGIONS_TO_ENABLE_type;

// Type aliases and/or sizes for ATTR_DEVICE_ID attribute
typedef uint8_t DEVICE_ID_ATTR;
typedef uint8_t ATTR_DEVICE_ID_type;

// Type aliases and/or sizes for ATTR_ECID attribute
typedef uint64_t ECID_ATTR[2];
typedef uint64_t ATTR_ECID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 2> ATTR_ECID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_I2C_BUS_DIV_NEST attribute
typedef uint8_t I2C_BUS_DIV_NEST_ATTR;
typedef uint8_t ATTR_I2C_BUS_DIV_NEST_type;

// Type aliases and/or sizes for ATTR_LEN_OF_SEEPROM_DATA attribute
typedef uint8_t LEN_OF_SEEPROM_DATA_ATTR;
typedef uint8_t ATTR_LEN_OF_SEEPROM_DATA_type;

// Type aliases and/or sizes for ATTR_MB_BIT_RATE_DIVISOR_PLL attribute
typedef uint8_t MB_BIT_RATE_DIVISOR_PLL_ATTR;
typedef uint8_t ATTR_MB_BIT_RATE_DIVISOR_PLL_type;

// Type aliases and/or sizes for ATTR_MB_BIT_RATE_DIVISOR_REFCLK attribute
typedef uint8_t MB_BIT_RATE_DIVISOR_REFCLK_ATTR;
typedef uint8_t ATTR_MB_BIT_RATE_DIVISOR_REFCLK_type;

// Type aliases and/or sizes for ATTR_MC_SYNC_MODE attribute
typedef uint8_t MC_SYNC_MODE_ATTR;
typedef uint8_t ATTR_MC_SYNC_MODE_type;

// Type aliases and/or sizes for ATTR_PROC_PB_BNDY_DMIPLL_DATA attribute
typedef uint8_t PROC_PB_BNDY_DMIPLL_DATA_ATTR;
typedef uint8_t ATTR_PROC_PB_BNDY_DMIPLL_DATA_type;

// Type aliases and/or sizes for ATTR_PROC_PB_BNDY_DMIPLL_FOR_DCCAL_DATA attribute
typedef uint8_t PROC_PB_BNDY_DMIPLL_FOR_DCCAL_DATA_ATTR;
typedef uint8_t ATTR_PROC_PB_BNDY_DMIPLL_FOR_DCCAL_DATA_type;

// Type aliases and/or sizes for ATTR_PROC_PERV_BNDY_PLL_DATA attribute
typedef uint8_t PROC_PERV_BNDY_PLL_DATA_ATTR;
typedef uint8_t ATTR_PROC_PERV_BNDY_PLL_DATA_type;

// Type aliases and/or sizes for ATTR_PROC_SBE_MASTER_CHIP attribute
typedef uint8_t PROC_SBE_MASTER_CHIP_ATTR;
typedef uint8_t ATTR_PROC_SBE_MASTER_CHIP_type;

// Type aliases and/or sizes for ATTR_SBE_SEEPROM_I2C_DEVICE_ADDRESS attribute
typedef uint64_t SBE_SEEPROM_I2C_DEVICE_ADDRESS_ATTR;
typedef uint64_t ATTR_SBE_SEEPROM_I2C_DEVICE_ADDRESS_type;

// Type aliases and/or sizes for ATTR_SBE_SEEPROM_I2C_PORT attribute
typedef uint64_t SBE_SEEPROM_I2C_PORT_ATTR;
typedef uint64_t ATTR_SBE_SEEPROM_I2C_PORT_type;

// Type aliases and/or sizes for ATTR_START_PIBMEM_ADDR attribute
typedef uint8_t START_PIBMEM_ADDR_ATTR;
typedef uint8_t ATTR_START_PIBMEM_ADDR_type;

// Type aliases and/or sizes for ATTR_START_SEEPROM_ADDR attribute
typedef uint8_t START_SEEPROM_ADDR_ATTR;
typedef uint8_t ATTR_START_SEEPROM_ADDR_type;

// Type aliases and/or sizes for ATTR_WAIT_N0 attribute
typedef uint8_t WAIT_N0_ATTR;
typedef uint8_t ATTR_WAIT_N0_type;

// Type aliases and/or sizes for ATTR_WAIT_N1 attribute
typedef uint8_t WAIT_N1_ATTR;
typedef uint8_t ATTR_WAIT_N1_type;

// Type aliases and/or sizes for ATTR_WAIT_N2 attribute
typedef uint8_t WAIT_N2_ATTR;
typedef uint8_t ATTR_WAIT_N2_type;

// Type aliases and/or sizes for ATTR_WAIT_N3 attribute
typedef uint8_t WAIT_N3_ATTR;
typedef uint8_t ATTR_WAIT_N3_type;

// Type aliases and/or sizes for ATTR_SYS_FORCE_ALL_CORES attribute
typedef uint8_t SYS_FORCE_ALL_CORES_ATTR;
typedef uint8_t ATTR_SYS_FORCE_ALL_CORES_type;

// Type aliases and/or sizes for ATTR_MASTER_CORE attribute
typedef uint8_t MASTER_CORE_ATTR;
typedef uint8_t ATTR_MASTER_CORE_type;

// Type aliases and/or sizes for ATTR_MASTER_EX attribute
typedef uint8_t MASTER_EX_ATTR;
typedef uint8_t ATTR_MASTER_EX_type;

// Type aliases and/or sizes for ATTR_SECURITY_ENABLE attribute
typedef uint8_t SECURITY_ENABLE_ATTR;
typedef uint8_t ATTR_SECURITY_ENABLE_type;

// Type aliases and/or sizes for ATTR_PFET_OFF_CONTROLS attribute
typedef uint32_t PFET_OFF_CONTROLS_ATTR;
typedef uint32_t ATTR_PFET_OFF_CONTROLS_type;

// Type aliases and/or sizes for ATTR_OBUS_RATIO_VALUE attribute
typedef uint8_t OBUS_RATIO_VALUE_ATTR;
typedef uint8_t ATTR_OBUS_RATIO_VALUE_type;

// Type aliases and/or sizes for ATTR_PIBMEM_REPAIR0 attribute
typedef uint64_t PIBMEM_REPAIR0_ATTR;
typedef uint64_t ATTR_PIBMEM_REPAIR0_type;

// Type aliases and/or sizes for ATTR_PIBMEM_REPAIR1 attribute
typedef uint64_t PIBMEM_REPAIR1_ATTR;
typedef uint64_t ATTR_PIBMEM_REPAIR1_type;

// Type aliases and/or sizes for ATTR_PIBMEM_REPAIR2 attribute
typedef uint64_t PIBMEM_REPAIR2_ATTR;
typedef uint64_t ATTR_PIBMEM_REPAIR2_type;

// Type aliases and/or sizes for ATTR_SENSEADJ_STEP attribute
typedef uint8_t SENSEADJ_STEP_ATTR;
typedef uint8_t ATTR_SENSEADJ_STEP_type;

// Type aliases and/or sizes for ATTR_CP_FILTER_BYPASS attribute
typedef uint8_t CP_FILTER_BYPASS_ATTR;
typedef uint8_t ATTR_CP_FILTER_BYPASS_type;

// Type aliases and/or sizes for ATTR_SS_FILTER_BYPASS attribute
typedef uint8_t SS_FILTER_BYPASS_ATTR;
typedef uint8_t ATTR_SS_FILTER_BYPASS_type;

// Type aliases and/or sizes for ATTR_IO_FILTER_BYPASS attribute
typedef uint8_t IO_FILTER_BYPASS_ATTR;
typedef uint8_t ATTR_IO_FILTER_BYPASS_type;

// Type aliases and/or sizes for ATTR_DPLL_BYPASS attribute
typedef uint8_t DPLL_BYPASS_ATTR;
typedef uint8_t ATTR_DPLL_BYPASS_type;

// Type aliases and/or sizes for ATTR_NEST_MEM_X_O_PCI_BYPASS attribute
typedef uint8_t NEST_MEM_X_O_PCI_BYPASS_ATTR;
typedef uint8_t ATTR_NEST_MEM_X_O_PCI_BYPASS_type;

// Type aliases and/or sizes for ATTR_TARGET_HAS_POWER attribute
typedef uint8_t TARGET_HAS_POWER_ATTR;
typedef uint8_t ATTR_TARGET_HAS_POWER_type;

// Type aliases and/or sizes for ATTR_TARGET_HAS_CLOCK attribute
typedef uint8_t TARGET_HAS_CLOCK_ATTR;
typedef uint8_t ATTR_TARGET_HAS_CLOCK_type;

// Type aliases and/or sizes for ATTR_TARGET_IS_SCOMMABLE attribute
typedef uint8_t TARGET_IS_SCOMMABLE_ATTR;
typedef uint8_t ATTR_TARGET_IS_SCOMMABLE_type;

// Type aliases and/or sizes for ATTR_SBE_SYS_CONFIG attribute
typedef uint64_t SBE_SYS_CONFIG_ATTR;
typedef uint64_t ATTR_SBE_SYS_CONFIG_type;

// Type aliases and/or sizes for ATTR_CP_REFCLOCK_RCVR_TERM attribute
typedef uint8_t CP_REFCLOCK_RCVR_TERM_ATTR;
typedef uint8_t ATTR_CP_REFCLOCK_RCVR_TERM_type;

// Type aliases and/or sizes for ATTR_IO_REFCLOCK_RCVR_TERM attribute
typedef uint8_t IO_REFCLOCK_RCVR_TERM_ATTR;
typedef uint8_t ATTR_IO_REFCLOCK_RCVR_TERM_type;

// Type aliases and/or sizes for ATTR_SECTOR_BUFFER_STRENGTH attribute
typedef uint8_t SECTOR_BUFFER_STRENGTH_ATTR;
typedef uint8_t ATTR_SECTOR_BUFFER_STRENGTH_type;

// Type aliases and/or sizes for ATTR_PULSE_MODE_ENABLE attribute
typedef uint8_t PULSE_MODE_ENABLE_ATTR;
typedef uint8_t ATTR_PULSE_MODE_ENABLE_type;

// Type aliases and/or sizes for ATTR_PULSE_MODE_VALUE attribute
typedef uint8_t PULSE_MODE_VALUE_ATTR;
typedef uint8_t ATTR_PULSE_MODE_VALUE_type;

// Type aliases and/or sizes for ATTR_NDL_MESHCTRL_SETUP attribute
typedef uint8_t NDL_MESHCTRL_SETUP_ATTR;
typedef uint8_t ATTR_NDL_MESHCTRL_SETUP_type;

// Type aliases and/or sizes for ATTR_START_CBS_FIFO_RESET_SKIP attribute
typedef uint8_t START_CBS_FIFO_RESET_SKIP_ATTR;
typedef uint8_t ATTR_START_CBS_FIFO_RESET_SKIP_type;

// Type aliases and/or sizes for ATTR_SYSTEM_CORECACHE_SKEWADJ_DISABLE attribute
typedef uint8_t SYSTEM_CORECACHE_SKEWADJ_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_CORECACHE_SKEWADJ_DISABLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_CORECACHE_DCADJ_DISABLE attribute
typedef uint8_t SYSTEM_CORECACHE_DCADJ_DISABLE_ATTR;
typedef uint8_t ATTR_SYSTEM_CORECACHE_DCADJ_DISABLE_type;

// Type aliases and/or sizes for ATTR_PROC_MEM_TO_USE attribute
typedef uint8_t PROC_MEM_TO_USE_ATTR;
typedef uint8_t ATTR_PROC_MEM_TO_USE_type;

// Type aliases and/or sizes for ATTR_ORIG_FIR_SETTINGS_ACTION0 attribute
typedef uint64_t ORIG_FIR_SETTINGS_ACTION0_ATTR;
typedef uint64_t ATTR_ORIG_FIR_SETTINGS_ACTION0_type;

// Type aliases and/or sizes for ATTR_ORIG_FIR_SETTINGS_ACTION1 attribute
typedef uint64_t ORIG_FIR_SETTINGS_ACTION1_ATTR;
typedef uint64_t ATTR_ORIG_FIR_SETTINGS_ACTION1_type;

// Type aliases and/or sizes for ATTR_LPC_CONSOLE_CNFG attribute
typedef uint8_t LPC_CONSOLE_CNFG_ATTR;
typedef uint8_t ATTR_LPC_CONSOLE_CNFG_type;

// Type aliases and/or sizes for ATTR_X_EREPAIR_THRESHOLD_FIELD attribute
typedef uint8_t X_EREPAIR_THRESHOLD_FIELD_ATTR;
typedef uint8_t ATTR_X_EREPAIR_THRESHOLD_FIELD_type;

// Type aliases and/or sizes for ATTR_X_EREPAIR_THRESHOLD_MNFG attribute
typedef uint8_t X_EREPAIR_THRESHOLD_MNFG_ATTR;
typedef uint8_t ATTR_X_EREPAIR_THRESHOLD_MNFG_type;

// Type aliases and/or sizes for ATTR_O_EREPAIR_THRESHOLD_FIELD attribute
typedef uint8_t O_EREPAIR_THRESHOLD_FIELD_ATTR;
typedef uint8_t ATTR_O_EREPAIR_THRESHOLD_FIELD_type;

// Type aliases and/or sizes for ATTR_O_EREPAIR_THRESHOLD_MNFG attribute
typedef uint8_t O_EREPAIR_THRESHOLD_MNFG_ATTR;
typedef uint8_t ATTR_O_EREPAIR_THRESHOLD_MNFG_type;

// Type aliases and/or sizes for ATTR_DMI_EREPAIR_THRESHOLD_FIELD attribute
typedef uint8_t DMI_EREPAIR_THRESHOLD_FIELD_ATTR;
typedef uint8_t ATTR_DMI_EREPAIR_THRESHOLD_FIELD_type;

// Type aliases and/or sizes for ATTR_DMI_EREPAIR_THRESHOLD_MNFG attribute
typedef uint8_t DMI_EREPAIR_THRESHOLD_MNFG_ATTR;
typedef uint8_t ATTR_DMI_EREPAIR_THRESHOLD_MNFG_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_CKE_MAP attribute
typedef uint8_t MSS_VPD_CKE_MAP_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_VPD_CKE_MAP_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_VPD_CKE_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_IO_OBUS_DCCAL_FLAGS attribute
typedef uint8_t IO_OBUS_DCCAL_FLAGS_ATTR;
typedef uint8_t ATTR_IO_OBUS_DCCAL_FLAGS_type;

// Type aliases and/or sizes for ATTR_IO_OBUS_PAT_A_DETECT_RUN attribute
typedef uint8_t IO_OBUS_PAT_A_DETECT_RUN_ATTR;
typedef uint8_t ATTR_IO_OBUS_PAT_A_DETECT_RUN_type;

// Type aliases and/or sizes for ATTR_IO_OBUS_PAT_A_DETECT_RX_AMP_VALUE attribute
typedef uint8_t IO_OBUS_PAT_A_DETECT_RX_AMP_VALUE_ATTR;
typedef uint8_t ATTR_IO_OBUS_PAT_A_DETECT_RX_AMP_VALUE_type;

// Type aliases and/or sizes for ATTR_IO_OBUS_PAT_A_CAPTURE attribute
typedef uint32_t IO_OBUS_PAT_A_CAPTURE_ATTR[24];
typedef uint32_t ATTR_IO_OBUS_PAT_A_CAPTURE_type[24];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 24> ATTR_IO_OBUS_PAT_A_CAPTURE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_IO_OBUS_TRAIN_FOR_RECOVERY attribute
typedef uint8_t IO_OBUS_TRAIN_FOR_RECOVERY_ATTR;
typedef uint8_t ATTR_IO_OBUS_TRAIN_FOR_RECOVERY_type;

// Type aliases and/or sizes for ATTR_IO_OBUS_LINK_SPARE_MARK attribute
typedef uint8_t IO_OBUS_LINK_SPARE_MARK_ATTR[2];
typedef uint8_t ATTR_IO_OBUS_LINK_SPARE_MARK_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_IO_OBUS_LINK_SPARE_MARK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_IO_OBUS_LANE_PDWN attribute
typedef uint32_t IO_OBUS_LANE_PDWN_ATTR;
typedef uint32_t ATTR_IO_OBUS_LANE_PDWN_type;

// Type aliases and/or sizes for ATTR_IO_O_DEBUG attribute
typedef uint8_t IO_O_DEBUG_ATTR;
typedef uint8_t ATTR_IO_O_DEBUG_type;

// Type aliases and/or sizes for ATTR_IO_O_MFG_CHK attribute
typedef uint8_t IO_O_MFG_CHK_ATTR;
typedef uint8_t ATTR_IO_O_MFG_CHK_type;

// Type aliases and/or sizes for ATTR_IO_O_MFG_MIN_EYE_WIDTH attribute
typedef uint8_t IO_O_MFG_MIN_EYE_WIDTH_ATTR;
typedef uint8_t ATTR_IO_O_MFG_MIN_EYE_WIDTH_type;

// Type aliases and/or sizes for ATTR_IO_OBUS_TX_MARGIN_RATIO attribute
typedef uint8_t IO_OBUS_TX_MARGIN_RATIO_ATTR;
typedef uint8_t ATTR_IO_OBUS_TX_MARGIN_RATIO_type;

// Type aliases and/or sizes for ATTR_IO_OBUS_TX_FFE_PRECURSOR attribute
typedef uint8_t IO_OBUS_TX_FFE_PRECURSOR_ATTR;
typedef uint8_t ATTR_IO_OBUS_TX_FFE_PRECURSOR_type;

// Type aliases and/or sizes for ATTR_IO_OBUS_TX_FFE_POSTCURSOR attribute
typedef uint8_t IO_OBUS_TX_FFE_POSTCURSOR_ATTR;
typedef uint8_t ATTR_IO_OBUS_TX_FFE_POSTCURSOR_type;

// Type aliases and/or sizes for ATTR_IO_OLLFIR_MASK attribute
typedef uint64_t IO_OLLFIR_MASK_ATTR;
typedef uint64_t ATTR_IO_OLLFIR_MASK_type;

// Type aliases and/or sizes for ATTR_IO_PB_IOOFIR_MASK attribute
typedef uint64_t IO_PB_IOOFIR_MASK_ATTR;
typedef uint64_t ATTR_IO_PB_IOOFIR_MASK_type;

// Type aliases and/or sizes for ATTR_IO_O_CHANNEL_TYPE attribute
typedef uint8_t IO_O_CHANNEL_TYPE_ATTR;
typedef uint8_t ATTR_IO_O_CHANNEL_TYPE_type;

// Type aliases and/or sizes for ATTR_IO_O_MFG_STRESS_PR_OFFSET_EVEN attribute
typedef uint8_t IO_O_MFG_STRESS_PR_OFFSET_EVEN_ATTR;
typedef uint8_t ATTR_IO_O_MFG_STRESS_PR_OFFSET_EVEN_type;

// Type aliases and/or sizes for ATTR_IO_O_MFG_STRESS_PR_OFFSET_ODD attribute
typedef uint8_t IO_O_MFG_STRESS_PR_OFFSET_ODD_ATTR;
typedef uint8_t ATTR_IO_O_MFG_STRESS_PR_OFFSET_ODD_type;

// Type aliases and/or sizes for ATTR_IO_O_MNFG_ERROR_THRESHOLD attribute
typedef uint8_t IO_O_MNFG_ERROR_THRESHOLD_ATTR;
typedef uint8_t ATTR_IO_O_MNFG_ERROR_THRESHOLD_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_0_VERSION_LAYOUT attribute
typedef uint8_t MSS_VPD_MT_0_VERSION_LAYOUT_ATTR;
typedef uint8_t ATTR_MSS_VPD_MT_0_VERSION_LAYOUT_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_1_VERSION_DATA attribute
typedef uint8_t MSS_VPD_MT_1_VERSION_DATA_ATTR;
typedef uint8_t ATTR_MSS_VPD_MT_1_VERSION_DATA_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_2_SIGNATURE_HASH attribute
typedef uint32_t MSS_VPD_MT_2_SIGNATURE_HASH_ATTR;
typedef uint32_t ATTR_MSS_VPD_MT_2_SIGNATURE_HASH_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CA attribute
typedef uint8_t MSS_VPD_MT_DIMM_RCD_IBT_CA_ATTR[2][2];
typedef uint8_t ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CA_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CKE attribute
typedef uint8_t MSS_VPD_MT_DIMM_RCD_IBT_CKE_ATTR[2][2];
typedef uint8_t ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CKE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CKE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CS attribute
typedef uint8_t MSS_VPD_MT_DIMM_RCD_IBT_CS_ATTR[2][2];
typedef uint8_t ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_ODT attribute
typedef uint8_t MSS_VPD_MT_DIMM_RCD_IBT_ODT_ATTR[2][2];
typedef uint8_t ATTR_MSS_VPD_MT_DIMM_RCD_IBT_ODT_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_MSS_VPD_MT_DIMM_RCD_IBT_ODT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS attribute
typedef uint8_t MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DRAM_RTT_NOM attribute
typedef uint8_t MSS_VPD_MT_DRAM_RTT_NOM_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_VPD_MT_DRAM_RTT_NOM_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_VPD_MT_DRAM_RTT_NOM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DRAM_RTT_PARK attribute
typedef uint8_t MSS_VPD_MT_DRAM_RTT_PARK_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_VPD_MT_DRAM_RTT_PARK_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_VPD_MT_DRAM_RTT_PARK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_DRAM_RTT_WR attribute
typedef uint8_t MSS_VPD_MT_DRAM_RTT_WR_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_VPD_MT_DRAM_RTT_WR_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_VPD_MT_DRAM_RTT_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_BIAS_TRIM attribute
typedef uint8_t MSS_VPD_MT_MC_BIAS_TRIM_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MT_MC_BIAS_TRIM_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MT_MC_BIAS_TRIM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP attribute
typedef uint32_t MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP_ATTR[2];
typedef uint32_t ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN attribute
typedef uint32_t MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN_ATTR[2];
typedef uint32_t ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP attribute
typedef uint32_t MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP_ATTR[2];
typedef uint32_t ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DQ_CTLE_CAP attribute
typedef uint64_t MSS_VPD_MT_MC_DQ_CTLE_CAP_ATTR[2];
typedef uint64_t ATTR_MSS_VPD_MT_MC_DQ_CTLE_CAP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 2> ATTR_MSS_VPD_MT_MC_DQ_CTLE_CAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DQ_CTLE_RES attribute
typedef uint64_t MSS_VPD_MT_MC_DQ_CTLE_RES_ATTR[2];
typedef uint64_t ATTR_MSS_VPD_MT_MC_DQ_CTLE_RES_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 2> ATTR_MSS_VPD_MT_MC_DQ_CTLE_RES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DRV_IMP_CLK attribute
typedef uint8_t MSS_VPD_MT_MC_DRV_IMP_CLK_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MT_MC_DRV_IMP_CLK_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MT_MC_DRV_IMP_CLK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR attribute
typedef uint8_t MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DRV_IMP_CNTL attribute
typedef uint8_t MSS_VPD_MT_MC_DRV_IMP_CNTL_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MT_MC_DRV_IMP_CNTL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MT_MC_DRV_IMP_CNTL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DRV_IMP_CSCID attribute
typedef uint8_t MSS_VPD_MT_MC_DRV_IMP_CSCID_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MT_MC_DRV_IMP_CSCID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MT_MC_DRV_IMP_CSCID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_DRV_IMP_DQ_DQS attribute
typedef uint8_t MSS_VPD_MT_MC_DRV_IMP_DQ_DQS_ATTR[2][5];
typedef uint8_t ATTR_MSS_VPD_MT_MC_DRV_IMP_DQ_DQS_type[2][5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 5>, 2> ATTR_MSS_VPD_MT_MC_DRV_IMP_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_MC_RCV_IMP_DQ_DQS attribute
typedef uint8_t MSS_VPD_MT_MC_RCV_IMP_DQ_DQS_ATTR[2][5];
typedef uint8_t ATTR_MSS_VPD_MT_MC_RCV_IMP_DQ_DQS_type[2][5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 5>, 2> ATTR_MSS_VPD_MT_MC_RCV_IMP_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_ODT_RD attribute
typedef uint8_t MSS_VPD_MT_ODT_RD_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_VPD_MT_ODT_RD_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_VPD_MT_ODT_RD_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_ODT_WR attribute
typedef uint8_t MSS_VPD_MT_ODT_WR_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_VPD_MT_ODT_WR_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_VPD_MT_ODT_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_PREAMBLE attribute
typedef uint8_t MSS_VPD_MT_PREAMBLE_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MT_PREAMBLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MT_PREAMBLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_VREF_DRAM_WR attribute
typedef uint8_t MSS_VPD_MT_VREF_DRAM_WR_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MT_VREF_DRAM_WR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MT_VREF_DRAM_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_VREF_MC_RD attribute
typedef uint32_t MSS_VPD_MT_VREF_MC_RD_ATTR[2];
typedef uint32_t ATTR_MSS_VPD_MT_VREF_MC_RD_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_VPD_MT_VREF_MC_RD_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MT_WINDAGE_RD_CTR attribute
typedef int16_t MSS_VPD_MT_WINDAGE_RD_CTR_ATTR[2];
typedef int16_t ATTR_MSS_VPD_MT_WINDAGE_RD_CTR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<int16_t, 2> ATTR_MSS_VPD_MT_WINDAGE_RD_CTR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_DSTLCFG_MMIO_ADDRBIT_POS attribute
typedef uint8_t PROC_DSTLCFG_MMIO_ADDRBIT_POS_ATTR;
typedef uint8_t ATTR_PROC_DSTLCFG_MMIO_ADDRBIT_POS_type;

// Type aliases and/or sizes for ATTR_PROC_ENABLE_DL_TMPL_1 attribute
typedef uint8_t PROC_ENABLE_DL_TMPL_1_ATTR;
typedef uint8_t ATTR_PROC_ENABLE_DL_TMPL_1_type;

// Type aliases and/or sizes for ATTR_PROC_ENABLE_DL_TMPL_4 attribute
typedef uint8_t PROC_ENABLE_DL_TMPL_4_ATTR;
typedef uint8_t ATTR_PROC_ENABLE_DL_TMPL_4_type;

// Type aliases and/or sizes for ATTR_PROC_ENABLE_DL_TMPL_7 attribute
typedef uint8_t PROC_ENABLE_DL_TMPL_7_ATTR;
typedef uint8_t ATTR_PROC_ENABLE_DL_TMPL_7_type;

// Type aliases and/or sizes for ATTR_PROC_ENABLE_DL_TMPL_A attribute
typedef uint8_t PROC_ENABLE_DL_TMPL_A_ATTR;
typedef uint8_t ATTR_PROC_ENABLE_DL_TMPL_A_type;

// Type aliases and/or sizes for ATTR_PROC_TMPL_0_PACING attribute
typedef uint8_t PROC_TMPL_0_PACING_ATTR;
typedef uint8_t ATTR_PROC_TMPL_0_PACING_type;

// Type aliases and/or sizes for ATTR_PROC_TMPL_1_PACING attribute
typedef uint8_t PROC_TMPL_1_PACING_ATTR;
typedef uint8_t ATTR_PROC_TMPL_1_PACING_type;

// Type aliases and/or sizes for ATTR_PROC_TMPL_4_PACING attribute
typedef uint8_t PROC_TMPL_4_PACING_ATTR;
typedef uint8_t ATTR_PROC_TMPL_4_PACING_type;

// Type aliases and/or sizes for ATTR_PROC_TMPL_7_PACING attribute
typedef uint8_t PROC_TMPL_7_PACING_ATTR;
typedef uint8_t ATTR_PROC_TMPL_7_PACING_type;

// Type aliases and/or sizes for ATTR_PROC_TMPL_A_PACING attribute
typedef uint8_t PROC_TMPL_A_PACING_ATTR;
typedef uint8_t ATTR_PROC_TMPL_A_PACING_type;

// Type aliases and/or sizes for ATTR_PROC_OMI_OC_MAJOR_VER attribute
typedef uint8_t PROC_OMI_OC_MAJOR_VER_ATTR;
typedef uint8_t ATTR_PROC_OMI_OC_MAJOR_VER_type;

// Type aliases and/or sizes for ATTR_PROC_OMI_OC_MINOR_VER attribute
typedef uint8_t PROC_OMI_OC_MINOR_VER_ATTR;
typedef uint8_t ATTR_PROC_OMI_OC_MINOR_VER_type;

// Type aliases and/or sizes for ATTR_HB_HRMOR_BYTES attribute
typedef uint64_t HB_HRMOR_BYTES_ATTR;
typedef uint64_t ATTR_HB_HRMOR_BYTES_type;

// Type aliases and/or sizes for ATTR_FREQ_PROC_REFCLOCK_KHZ attribute
typedef uint32_t FREQ_PROC_REFCLOCK_KHZ_ATTR;
typedef uint32_t ATTR_FREQ_PROC_REFCLOCK_KHZ_type;

// Type aliases and/or sizes for ATTR_FREQ_DPLL_REFCLOCK_KHZ attribute
typedef uint32_t FREQ_DPLL_REFCLOCK_KHZ_ATTR;
typedef uint32_t ATTR_FREQ_DPLL_REFCLOCK_KHZ_type;

// Type aliases and/or sizes for ATTR_FREQ_MEM_REFCLOCK attribute
typedef uint32_t FREQ_MEM_REFCLOCK_ATTR;
typedef uint32_t ATTR_FREQ_MEM_REFCLOCK_type;

// Type aliases and/or sizes for ATTR_MAX_ALLOWED_DIMM_FREQ attribute
typedef uint32_t MAX_ALLOWED_DIMM_FREQ_ATTR[5];
typedef uint32_t ATTR_MAX_ALLOWED_DIMM_FREQ_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 5> ATTR_MAX_ALLOWED_DIMM_FREQ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEMVPD_POS attribute
typedef uint8_t MEMVPD_POS_ATTR;
typedef uint8_t ATTR_MEMVPD_POS_type;

// Type aliases and/or sizes for ATTR_REQUIRED_SYNCH_MODE attribute
typedef uint8_t REQUIRED_SYNCH_MODE_ATTR;
typedef uint8_t ATTR_REQUIRED_SYNCH_MODE_type;

// Type aliases and/or sizes for ATTR_NHTM_TRACE_TYPE attribute
typedef uint8_t NHTM_TRACE_TYPE_ATTR;
typedef uint8_t ATTR_NHTM_TRACE_TYPE_type;

// Type aliases and/or sizes for ATTR_CHTM_TRACE_TYPE attribute
typedef uint8_t CHTM_TRACE_TYPE_ATTR[24];
typedef uint8_t ATTR_CHTM_TRACE_TYPE_type[24];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 24> ATTR_CHTM_TRACE_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HTMSC_TTYPEFILT_PAT attribute
typedef uint8_t HTMSC_TTYPEFILT_PAT_ATTR;
typedef uint8_t ATTR_HTMSC_TTYPEFILT_PAT_type;

// Type aliases and/or sizes for ATTR_HTMSC_TSIZEFILT_PAT attribute
typedef uint8_t HTMSC_TSIZEFILT_PAT_ATTR;
typedef uint8_t ATTR_HTMSC_TSIZEFILT_PAT_type;

// Type aliases and/or sizes for ATTR_HTMSC_TTYPEFILT_MASK attribute
typedef uint8_t HTMSC_TTYPEFILT_MASK_ATTR;
typedef uint8_t ATTR_HTMSC_TTYPEFILT_MASK_type;

// Type aliases and/or sizes for ATTR_HTMSC_TSIZEFILT_MASK attribute
typedef uint8_t HTMSC_TSIZEFILT_MASK_ATTR;
typedef uint8_t ATTR_HTMSC_TSIZEFILT_MASK_type;

// Type aliases and/or sizes for ATTR_HTMSC_TTYPEFILT_INVERT attribute
typedef uint8_t HTMSC_TTYPEFILT_INVERT_ATTR;
typedef uint8_t ATTR_HTMSC_TTYPEFILT_INVERT_type;

// Type aliases and/or sizes for ATTR_HTMSC_CRESPFILT_INVERT attribute
typedef uint8_t HTMSC_CRESPFILT_INVERT_ATTR;
typedef uint8_t ATTR_HTMSC_CRESPFILT_INVERT_type;

// Type aliases and/or sizes for ATTR_HTMSC_FILT_PAT attribute
typedef uint32_t HTMSC_FILT_PAT_ATTR;
typedef uint32_t ATTR_HTMSC_FILT_PAT_type;

// Type aliases and/or sizes for ATTR_HTMSC_FILT_CRESP_PAT attribute
typedef uint8_t HTMSC_FILT_CRESP_PAT_ATTR;
typedef uint8_t ATTR_HTMSC_FILT_CRESP_PAT_type;

// Type aliases and/or sizes for ATTR_HTMSC_FILT_MASK attribute
typedef uint32_t HTMSC_FILT_MASK_ATTR;
typedef uint32_t ATTR_HTMSC_FILT_MASK_type;

// Type aliases and/or sizes for ATTR_HTMSC_FILT_CRESP_MASK attribute
typedef uint8_t HTMSC_FILT_CRESP_MASK_ATTR;
typedef uint8_t ATTR_HTMSC_FILT_CRESP_MASK_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_CONTENT_SEL attribute
typedef uint8_t NHTM_HTMSC_MODE_CONTENT_SEL_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_CONTENT_SEL_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_CAPTURE_GENERATED_WRITES attribute
typedef uint8_t NHTM_HTMSC_MODE_CAPTURE_GENERATED_WRITES_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_CAPTURE_GENERATED_WRITES_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_CAPTURE_ENABLE_FILTER_ALL attribute
typedef uint8_t NHTM_HTMSC_MODE_CAPTURE_ENABLE_FILTER_ALL_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_CAPTURE_ENABLE_FILTER_ALL_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_CAPTURE_PRECISE_CRESP_MODE attribute
typedef uint8_t NHTM_HTMSC_MODE_CAPTURE_PRECISE_CRESP_MODE_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_CAPTURE_PRECISE_CRESP_MODE_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_CAPTURE_LIMIT_MEM_ALLOCATION attribute
typedef uint8_t NHTM_HTMSC_MODE_CAPTURE_LIMIT_MEM_ALLOCATION_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_CAPTURE_LIMIT_MEM_ALLOCATION_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_CAPTURE_PMISC_ONLY_CMD attribute
typedef uint8_t NHTM_HTMSC_MODE_CAPTURE_PMISC_ONLY_CMD_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_CAPTURE_PMISC_ONLY_CMD_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_SYNC_STAMP_FORCE attribute
typedef uint8_t NHTM_HTMSC_MODE_SYNC_STAMP_FORCE_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_SYNC_STAMP_FORCE_type;

// Type aliases and/or sizes for ATTR_NHTM_HTMSC_MODE_WRITETOIO attribute
typedef uint8_t NHTM_HTMSC_MODE_WRITETOIO_ATTR;
typedef uint8_t ATTR_NHTM_HTMSC_MODE_WRITETOIO_type;

// Type aliases and/or sizes for ATTR_CHTM_HTMSC_MODE_CONTENT_SEL attribute
typedef uint8_t CHTM_HTMSC_MODE_CONTENT_SEL_ATTR;
typedef uint8_t ATTR_CHTM_HTMSC_MODE_CONTENT_SEL_type;

// Type aliases and/or sizes for ATTR_CHTM_HTMSC_MODE_CAPTURE attribute
typedef uint32_t CHTM_HTMSC_MODE_CAPTURE_ATTR;
typedef uint32_t ATTR_CHTM_HTMSC_MODE_CAPTURE_type;

// Type aliases and/or sizes for ATTR_CHTM_HTMSC_MODE_CORE_INSTR_STALL attribute
typedef uint8_t CHTM_HTMSC_MODE_CORE_INSTR_STALL_ATTR;
typedef uint8_t ATTR_CHTM_HTMSC_MODE_CORE_INSTR_STALL_type;

// Type aliases and/or sizes for ATTR_HTMSC_MODE_WRAP attribute
typedef uint8_t HTMSC_MODE_WRAP_ATTR;
typedef uint8_t ATTR_HTMSC_MODE_WRAP_type;

// Type aliases and/or sizes for ATTR_HTMSC_MODE_DIS_TSTAMP attribute
typedef uint8_t HTMSC_MODE_DIS_TSTAMP_ATTR;
typedef uint8_t ATTR_HTMSC_MODE_DIS_TSTAMP_type;

// Type aliases and/or sizes for ATTR_HTMSC_MODE_SINGLE_TSTAMP attribute
typedef uint8_t HTMSC_MODE_SINGLE_TSTAMP_ATTR;
typedef uint8_t ATTR_HTMSC_MODE_SINGLE_TSTAMP_type;

// Type aliases and/or sizes for ATTR_HTMSC_MODE_MARKERS_ONLY attribute
typedef uint8_t HTMSC_MODE_MARKERS_ONLY_ATTR;
typedef uint8_t ATTR_HTMSC_MODE_MARKERS_ONLY_type;

// Type aliases and/or sizes for ATTR_HTMSC_MODE_DIS_FORCE_GROUP_SCOPE attribute
typedef uint8_t HTMSC_MODE_DIS_FORCE_GROUP_SCOPE_ATTR;
typedef uint8_t ATTR_HTMSC_MODE_DIS_FORCE_GROUP_SCOPE_type;

// Type aliases and/or sizes for ATTR_HTMSC_MODE_VGTARGET attribute
typedef uint32_t HTMSC_MODE_VGTARGET_ATTR;
typedef uint32_t ATTR_HTMSC_MODE_VGTARGET_type;

// Type aliases and/or sizes for ATTR_HTMSC_MEM_SCOPE attribute
typedef uint8_t HTMSC_MEM_SCOPE_ATTR;
typedef uint8_t ATTR_HTMSC_MEM_SCOPE_type;

// Type aliases and/or sizes for ATTR_HTMSC_MEM_PRIORITY attribute
typedef uint8_t HTMSC_MEM_PRIORITY_ATTR;
typedef uint8_t ATTR_HTMSC_MEM_PRIORITY_type;

// Type aliases and/or sizes for ATTR_NHTM_CTRL_TRIG attribute
typedef uint8_t NHTM_CTRL_TRIG_ATTR;
typedef uint8_t ATTR_NHTM_CTRL_TRIG_type;

// Type aliases and/or sizes for ATTR_NHTM_CTRL_MARK attribute
typedef uint8_t NHTM_CTRL_MARK_ATTR;
typedef uint8_t ATTR_NHTM_CTRL_MARK_type;

// Type aliases and/or sizes for ATTR_CHTM_CTRL_TRIG attribute
typedef uint8_t CHTM_CTRL_TRIG_ATTR;
typedef uint8_t ATTR_CHTM_CTRL_TRIG_type;

// Type aliases and/or sizes for ATTR_CHTM_CTRL_MARK attribute
typedef uint8_t CHTM_CTRL_MARK_ATTR;
typedef uint8_t ATTR_CHTM_CTRL_MARK_type;

// Type aliases and/or sizes for ATTR_HTMSC_CTRL_DBG0_STOP attribute
typedef uint8_t HTMSC_CTRL_DBG0_STOP_ATTR;
typedef uint8_t ATTR_HTMSC_CTRL_DBG0_STOP_type;

// Type aliases and/or sizes for ATTR_HTMSC_CTRL_DBG1_STOP attribute
typedef uint8_t HTMSC_CTRL_DBG1_STOP_ATTR;
typedef uint8_t ATTR_HTMSC_CTRL_DBG1_STOP_type;

// Type aliases and/or sizes for ATTR_HTMSC_CTRL_RUN_STOP attribute
typedef uint8_t HTMSC_CTRL_RUN_STOP_ATTR;
typedef uint8_t ATTR_HTMSC_CTRL_RUN_STOP_type;

// Type aliases and/or sizes for ATTR_HTMSC_CTRL_OTHER_DBG0_STOP attribute
typedef uint8_t HTMSC_CTRL_OTHER_DBG0_STOP_ATTR;
typedef uint8_t ATTR_HTMSC_CTRL_OTHER_DBG0_STOP_type;

// Type aliases and/or sizes for ATTR_HTMSC_CTRL_XSTOP_STOP attribute
typedef uint8_t HTMSC_CTRL_XSTOP_STOP_ATTR;
typedef uint8_t ATTR_HTMSC_CTRL_XSTOP_STOP_type;

// Type aliases and/or sizes for ATTR_HTMSC_CTRL_CHIP0_STOP attribute
typedef uint8_t HTMSC_CTRL_CHIP0_STOP_ATTR;
typedef uint8_t ATTR_HTMSC_CTRL_CHIP0_STOP_type;

// Type aliases and/or sizes for ATTR_HTMSC_CTRL_CHIP1_STOP attribute
typedef uint8_t HTMSC_CTRL_CHIP1_STOP_ATTR;
typedef uint8_t ATTR_HTMSC_CTRL_CHIP1_STOP_type;

// Type aliases and/or sizes for ATTR_HTMSC_IMA_PDBAR_SPLIT_CORE_MODE attribute
typedef uint8_t HTMSC_IMA_PDBAR_SPLIT_CORE_MODE_ATTR;
typedef uint8_t ATTR_HTMSC_IMA_PDBAR_SPLIT_CORE_MODE_type;

// Type aliases and/or sizes for ATTR_HTMSC_IMA_PDBAR_SCOPE attribute
typedef uint8_t HTMSC_IMA_PDBAR_SCOPE_ATTR;
typedef uint8_t ATTR_HTMSC_IMA_PDBAR_SCOPE_type;

// Type aliases and/or sizes for ATTR_HTMSC_IMA_PDBAR_ADDR attribute
typedef uint64_t HTMSC_IMA_PDBAR_ADDR_ATTR;
typedef uint64_t ATTR_HTMSC_IMA_PDBAR_ADDR_type;

// Type aliases and/or sizes for ATTR_FREQ_PB_MHZ attribute
typedef uint32_t FREQ_PB_MHZ_ATTR;
typedef uint32_t ATTR_FREQ_PB_MHZ_type;

// Type aliases and/or sizes for ATTR_FREQ_OMI_MHZ attribute
typedef uint32_t FREQ_OMI_MHZ_ATTR;
typedef uint32_t ATTR_FREQ_OMI_MHZ_type;

// Type aliases and/or sizes for ATTR_OMI_PLL_VCO attribute
typedef uint8_t OMI_PLL_VCO_ATTR;
typedef uint8_t ATTR_OMI_PLL_VCO_type;

// Type aliases and/or sizes for ATTR_FREQ_O_MHZ attribute
typedef uint32_t FREQ_O_MHZ_ATTR[4];
typedef uint32_t ATTR_FREQ_O_MHZ_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 4> ATTR_FREQ_O_MHZ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_FREQ_A_MHZ attribute
typedef uint32_t FREQ_A_MHZ_ATTR;
typedef uint32_t ATTR_FREQ_A_MHZ_type;

// Type aliases and/or sizes for ATTR_FREQ_X_MHZ attribute
typedef uint32_t FREQ_X_MHZ_ATTR;
typedef uint32_t ATTR_FREQ_X_MHZ_type;

// Type aliases and/or sizes for ATTR_PM_SAFE_FREQUENCY_MHZ attribute
typedef uint32_t PM_SAFE_FREQUENCY_MHZ_ATTR;
typedef uint32_t ATTR_PM_SAFE_FREQUENCY_MHZ_type;

// Type aliases and/or sizes for ATTR_PM_SAFE_VOLTAGE_MV attribute
typedef uint32_t PM_SAFE_VOLTAGE_MV_ATTR;
typedef uint32_t ATTR_PM_SAFE_VOLTAGE_MV_type;

// Type aliases and/or sizes for ATTR_FREQ_PCIE_MHZ attribute
typedef uint32_t FREQ_PCIE_MHZ_ATTR;
typedef uint32_t ATTR_FREQ_PCIE_MHZ_type;

// Type aliases and/or sizes for ATTR_DD1_SLOW_PCI_REF_CLOCK attribute
typedef uint8_t DD1_SLOW_PCI_REF_CLOCK_ATTR;
typedef uint8_t ATTR_DD1_SLOW_PCI_REF_CLOCK_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_ASYNC_SAFE_MODE attribute
typedef uint8_t PROC_FABRIC_ASYNC_SAFE_MODE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_ASYNC_SAFE_MODE_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_BUS_WIDTH attribute
typedef uint8_t PROC_FABRIC_A_BUS_WIDTH_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_A_BUS_WIDTH_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_BUS_WIDTH attribute
typedef uint8_t PROC_FABRIC_X_BUS_WIDTH_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_X_BUS_WIDTH_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_CORE_FLOOR_RATIO attribute
typedef uint8_t PROC_FABRIC_CORE_FLOOR_RATIO_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_CORE_FLOOR_RATIO_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_CORE_CEILING_RATIO attribute
typedef uint8_t PROC_FABRIC_CORE_CEILING_RATIO_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_CORE_CEILING_RATIO_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_PUMP_MODE attribute
typedef uint8_t PROC_FABRIC_PUMP_MODE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_PUMP_MODE_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_CCSM_MODE attribute
typedef uint8_t PROC_FABRIC_CCSM_MODE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_CCSM_MODE_type;

// Type aliases and/or sizes for ATTR_OPTICS_CONFIG_MODE attribute
typedef uint8_t OPTICS_CONFIG_MODE_ATTR;
typedef uint8_t ATTR_OPTICS_CONFIG_MODE_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_SMP_OPTICS_MODE attribute
typedef uint8_t PROC_FABRIC_SMP_OPTICS_MODE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_SMP_OPTICS_MODE_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_OPTICS_CONFIG_MODE attribute
typedef uint8_t PROC_FABRIC_OPTICS_CONFIG_MODE_ATTR[4];
typedef uint8_t ATTR_PROC_FABRIC_OPTICS_CONFIG_MODE_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_FABRIC_OPTICS_CONFIG_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_CAPI_MODE attribute
typedef uint8_t PROC_FABRIC_CAPI_MODE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_CAPI_MODE_type;

// Type aliases and/or sizes for ATTR_FABRIC_PRESENT_GROUPS attribute
typedef uint8_t FABRIC_PRESENT_GROUPS_ATTR;
typedef uint8_t ATTR_FABRIC_PRESENT_GROUPS_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_SYSTEM_ID attribute
typedef uint32_t PROC_FABRIC_SYSTEM_ID_ATTR;
typedef uint32_t ATTR_PROC_FABRIC_SYSTEM_ID_type;

// Type aliases and/or sizes for ATTR_FABRIC_ADDR_EXTENSION_GROUP_ID attribute
typedef uint8_t FABRIC_ADDR_EXTENSION_GROUP_ID_ATTR;
typedef uint8_t ATTR_FABRIC_ADDR_EXTENSION_GROUP_ID_type;

// Type aliases and/or sizes for ATTR_FABRIC_ADDR_EXTENSION_CHIP_ID attribute
typedef uint8_t FABRIC_ADDR_EXTENSION_CHIP_ID_ATTR;
typedef uint8_t ATTR_FABRIC_ADDR_EXTENSION_CHIP_ID_type;

// Type aliases and/or sizes for ATTR_PROC_EFF_FABRIC_GROUP_ID attribute
typedef uint8_t PROC_EFF_FABRIC_GROUP_ID_ATTR;
typedef uint8_t ATTR_PROC_EFF_FABRIC_GROUP_ID_type;

// Type aliases and/or sizes for ATTR_PROC_EFF_FABRIC_CHIP_ID attribute
typedef uint8_t PROC_EFF_FABRIC_CHIP_ID_ATTR;
typedef uint8_t ATTR_PROC_EFF_FABRIC_CHIP_ID_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_SYSTEM_MASTER_CHIP attribute
typedef uint8_t PROC_FABRIC_SYSTEM_MASTER_CHIP_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_SYSTEM_MASTER_CHIP_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_GROUP_MASTER_CHIP attribute
typedef uint8_t PROC_FABRIC_GROUP_MASTER_CHIP_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_GROUP_MASTER_CHIP_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_LINK_ACTIVE attribute
typedef uint8_t PROC_FABRIC_LINK_ACTIVE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_LINK_ACTIVE_type;

// Type aliases and/or sizes for ATTR_LINK_TRAIN attribute
typedef uint8_t LINK_TRAIN_ATTR;
typedef uint8_t ATTR_LINK_TRAIN_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_ATTACHED_CHIP_CNFG attribute
typedef uint8_t PROC_FABRIC_X_ATTACHED_CHIP_CNFG_ATTR[7];
typedef uint8_t ATTR_PROC_FABRIC_X_ATTACHED_CHIP_CNFG_type[7];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 7> ATTR_PROC_FABRIC_X_ATTACHED_CHIP_CNFG_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_LINKS_CNFG attribute
typedef uint8_t PROC_FABRIC_X_LINKS_CNFG_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_X_LINKS_CNFG_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_ATTACHED_CHIP_CNFG attribute
typedef uint8_t PROC_FABRIC_A_ATTACHED_CHIP_CNFG_ATTR[4];
typedef uint8_t ATTR_PROC_FABRIC_A_ATTACHED_CHIP_CNFG_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_FABRIC_A_ATTACHED_CHIP_CNFG_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_LINKS_CNFG attribute
typedef uint8_t PROC_FABRIC_A_LINKS_CNFG_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_A_LINKS_CNFG_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_ATTACHED_CHIP_ID attribute
typedef uint8_t PROC_FABRIC_X_ATTACHED_CHIP_ID_ATTR[7];
typedef uint8_t ATTR_PROC_FABRIC_X_ATTACHED_CHIP_ID_type[7];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 7> ATTR_PROC_FABRIC_X_ATTACHED_CHIP_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_ATTACHED_LINK_ID attribute
typedef uint8_t PROC_FABRIC_X_ATTACHED_LINK_ID_ATTR[7];
typedef uint8_t ATTR_PROC_FABRIC_X_ATTACHED_LINK_ID_type[7];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 7> ATTR_PROC_FABRIC_X_ATTACHED_LINK_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_ATTACHED_CHIP_ID attribute
typedef uint8_t PROC_FABRIC_A_ATTACHED_CHIP_ID_ATTR[4];
typedef uint8_t ATTR_PROC_FABRIC_A_ATTACHED_CHIP_ID_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_FABRIC_A_ATTACHED_CHIP_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_ATTACHED_LINK_ID attribute
typedef uint8_t PROC_FABRIC_A_ATTACHED_LINK_ID_ATTR[4];
typedef uint8_t ATTR_PROC_FABRIC_A_ATTACHED_LINK_ID_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_FABRIC_A_ATTACHED_LINK_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_AGGREGATE attribute
typedef uint8_t PROC_FABRIC_X_AGGREGATE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_X_AGGREGATE_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_ADDR_DIS attribute
typedef uint8_t PROC_FABRIC_X_ADDR_DIS_ATTR[7];
typedef uint8_t ATTR_PROC_FABRIC_X_ADDR_DIS_type[7];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 7> ATTR_PROC_FABRIC_X_ADDR_DIS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_X_LINK_DELAY attribute
typedef uint32_t PROC_FABRIC_X_LINK_DELAY_ATTR[7];
typedef uint32_t ATTR_PROC_FABRIC_X_LINK_DELAY_type[7];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 7> ATTR_PROC_FABRIC_X_LINK_DELAY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_AGGREGATE attribute
typedef uint8_t PROC_FABRIC_A_AGGREGATE_ATTR;
typedef uint8_t ATTR_PROC_FABRIC_A_AGGREGATE_type;

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_ADDR_DIS attribute
typedef uint8_t PROC_FABRIC_A_ADDR_DIS_ATTR[4];
typedef uint8_t ATTR_PROC_FABRIC_A_ADDR_DIS_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_FABRIC_A_ADDR_DIS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_FABRIC_A_LINK_DELAY attribute
typedef uint32_t PROC_FABRIC_A_LINK_DELAY_ATTR[4];
typedef uint32_t ATTR_PROC_FABRIC_A_LINK_DELAY_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 4> ATTR_PROC_FABRIC_A_LINK_DELAY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_EPS_GB_PERCENTAGE attribute
typedef int8_t PROC_EPS_GB_PERCENTAGE_ATTR;
typedef int8_t ATTR_PROC_EPS_GB_PERCENTAGE_type;

// Type aliases and/or sizes for ATTR_PROC_EPS_TABLE_TYPE attribute
typedef uint8_t PROC_EPS_TABLE_TYPE_ATTR;
typedef uint8_t ATTR_PROC_EPS_TABLE_TYPE_type;

// Type aliases and/or sizes for ATTR_PROC_EPS_READ_CYCLES_T0 attribute
typedef uint32_t PROC_EPS_READ_CYCLES_T0_ATTR;
typedef uint32_t ATTR_PROC_EPS_READ_CYCLES_T0_type;

// Type aliases and/or sizes for ATTR_PROC_EPS_READ_CYCLES_T1 attribute
typedef uint32_t PROC_EPS_READ_CYCLES_T1_ATTR;
typedef uint32_t ATTR_PROC_EPS_READ_CYCLES_T1_type;

// Type aliases and/or sizes for ATTR_PROC_EPS_READ_CYCLES_T2 attribute
typedef uint32_t PROC_EPS_READ_CYCLES_T2_ATTR;
typedef uint32_t ATTR_PROC_EPS_READ_CYCLES_T2_type;

// Type aliases and/or sizes for ATTR_PROC_EPS_WRITE_CYCLES_T1 attribute
typedef uint32_t PROC_EPS_WRITE_CYCLES_T1_ATTR;
typedef uint32_t ATTR_PROC_EPS_WRITE_CYCLES_T1_type;

// Type aliases and/or sizes for ATTR_PROC_EPS_WRITE_CYCLES_T2 attribute
typedef uint32_t PROC_EPS_WRITE_CYCLES_T2_ATTR;
typedef uint32_t ATTR_PROC_EPS_WRITE_CYCLES_T2_type;

// Type aliases and/or sizes for ATTR_DMI_REFCLOCK_SWIZZLE attribute
typedef uint8_t DMI_REFCLOCK_SWIZZLE_ATTR;
typedef uint8_t ATTR_DMI_REFCLOCK_SWIZZLE_type;

// Type aliases and/or sizes for ATTR_OMI_REFCLOCK_SWIZZLE attribute
typedef uint8_t OMI_REFCLOCK_SWIZZLE_ATTR;
typedef uint8_t ATTR_OMI_REFCLOCK_SWIZZLE_type;

// Type aliases and/or sizes for ATTR_SYSTEM_IPL_PHASE attribute
typedef uint8_t SYSTEM_IPL_PHASE_ATTR;
typedef uint8_t ATTR_SYSTEM_IPL_PHASE_type;

// Type aliases and/or sizes for ATTR_PROC_XSCOM_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_XSCOM_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_XSCOM_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_LPC_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_LPC_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_LPC_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_MAX_INTERLEAVE_GROUP_SIZE attribute
typedef uint64_t MAX_INTERLEAVE_GROUP_SIZE_ATTR;
typedef uint64_t ATTR_MAX_INTERLEAVE_GROUP_SIZE_type;

// Type aliases and/or sizes for ATTR_MEM_MIRROR_PLACEMENT_POLICY attribute
typedef uint8_t MEM_MIRROR_PLACEMENT_POLICY_ATTR;
typedef uint8_t ATTR_MEM_MIRROR_PLACEMENT_POLICY_type;

// Type aliases and/or sizes for ATTR_PROC_MEM_BASES attribute
typedef uint64_t PROC_MEM_BASES_ATTR[8];
typedef uint64_t ATTR_PROC_MEM_BASES_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MEM_BASES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_MEM_SIZES attribute
typedef uint64_t PROC_MEM_SIZES_ATTR[8];
typedef uint64_t ATTR_PROC_MEM_SIZES_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MEM_SIZES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_MIRROR_BASES attribute
typedef uint64_t PROC_MIRROR_BASES_ATTR[8];
typedef uint64_t ATTR_PROC_MIRROR_BASES_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MIRROR_BASES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_MIRROR_SIZES attribute
typedef uint64_t PROC_MIRROR_SIZES_ATTR[8];
typedef uint64_t ATTR_PROC_MIRROR_SIZES_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MIRROR_SIZES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_INTERLEAVE_ENABLE attribute
typedef uint8_t MSS_INTERLEAVE_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_INTERLEAVE_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_INTERLEAVE_GRANULARITY attribute
typedef uint8_t MSS_INTERLEAVE_GRANULARITY_ATTR;
typedef uint8_t ATTR_MSS_INTERLEAVE_GRANULARITY_type;

// Type aliases and/or sizes for ATTR_MSS_MEM_MC_IN_GROUP attribute
typedef uint8_t MSS_MEM_MC_IN_GROUP_ATTR[8];
typedef uint8_t ATTR_MSS_MEM_MC_IN_GROUP_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 8> ATTR_MSS_MEM_MC_IN_GROUP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MCS_GROUP_32 attribute
typedef uint32_t MSS_MCS_GROUP_32_ATTR[16][21];
typedef uint32_t ATTR_MSS_MCS_GROUP_32_type[16][21];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 21>, 16> ATTR_MSS_MCS_GROUP_32_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MEM_IPL_COMPLETE attribute
typedef uint8_t MSS_MEM_IPL_COMPLETE_ATTR;
typedef uint8_t ATTR_MSS_MEM_IPL_COMPLETE_type;

// Type aliases and/or sizes for ATTR_MRW_HW_MIRRORING_ENABLE attribute
typedef uint8_t MRW_HW_MIRRORING_ENABLE_ATTR;
typedef uint8_t ATTR_MRW_HW_MIRRORING_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NHTM_BAR_BASE_ADDR attribute
typedef uint64_t PROC_NHTM_BAR_BASE_ADDR_ATTR;
typedef uint64_t ATTR_PROC_NHTM_BAR_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_PROC_NHTM_BAR_SIZE attribute
typedef uint64_t PROC_NHTM_BAR_SIZE_ATTR;
typedef uint64_t ATTR_PROC_NHTM_BAR_SIZE_type;

// Type aliases and/or sizes for ATTR_PROC_CHTM_BAR_BASE_ADDR attribute
typedef uint64_t PROC_CHTM_BAR_BASE_ADDR_ATTR[24];
typedef uint64_t ATTR_PROC_CHTM_BAR_BASE_ADDR_type[24];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 24> ATTR_PROC_CHTM_BAR_BASE_ADDR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_CHTM_BAR_SIZES attribute
typedef uint64_t PROC_CHTM_BAR_SIZES_ATTR[24];
typedef uint64_t ATTR_PROC_CHTM_BAR_SIZES_type[24];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 24> ATTR_PROC_CHTM_BAR_SIZES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_SMF_BAR_BASE_ADDR attribute
typedef uint64_t PROC_SMF_BAR_BASE_ADDR_ATTR;
typedef uint64_t ATTR_PROC_SMF_BAR_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_PROC_SMF_BAR_SIZE attribute
typedef uint64_t PROC_SMF_BAR_SIZE_ATTR;
typedef uint64_t ATTR_PROC_SMF_BAR_SIZE_type;

// Type aliases and/or sizes for ATTR_SMF_CONFIG attribute
typedef uint8_t SMF_CONFIG_ATTR;
typedef uint8_t ATTR_SMF_CONFIG_type;

// Type aliases and/or sizes for ATTR_SMF_ENABLED attribute
typedef uint8_t SMF_ENABLED_ATTR;
typedef uint8_t ATTR_SMF_ENABLED_type;

// Type aliases and/or sizes for ATTR_PROC_OCC_SANDBOX_BASE_ADDR attribute
typedef uint64_t PROC_OCC_SANDBOX_BASE_ADDR_ATTR;
typedef uint64_t ATTR_PROC_OCC_SANDBOX_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_PROC_OCC_SANDBOX_SIZE attribute
typedef uint64_t PROC_OCC_SANDBOX_SIZE_ATTR;
typedef uint64_t ATTR_PROC_OCC_SANDBOX_SIZE_type;

// Type aliases and/or sizes for ATTR_PROC_MEM_BASES_ACK attribute
typedef uint64_t PROC_MEM_BASES_ACK_ATTR[8];
typedef uint64_t ATTR_PROC_MEM_BASES_ACK_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MEM_BASES_ACK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_MEM_SIZES_ACK attribute
typedef uint64_t PROC_MEM_SIZES_ACK_ATTR[8];
typedef uint64_t ATTR_PROC_MEM_SIZES_ACK_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MEM_SIZES_ACK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_MIRROR_BASES_ACK attribute
typedef uint64_t PROC_MIRROR_BASES_ACK_ATTR[8];
typedef uint64_t ATTR_PROC_MIRROR_BASES_ACK_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MIRROR_BASES_ACK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_MIRROR_SIZES_ACK attribute
typedef uint64_t PROC_MIRROR_SIZES_ACK_ATTR[8];
typedef uint64_t ATTR_PROC_MIRROR_SIZES_ACK_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_PROC_MIRROR_SIZES_ACK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HTM_QUEUES attribute
typedef uint8_t HTM_QUEUES_ATTR[8];
typedef uint8_t ATTR_HTM_QUEUES_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 8> ATTR_HTM_QUEUES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_ENABLE_MEM_EARLY_DATA_SCOM attribute
typedef uint8_t ENABLE_MEM_EARLY_DATA_SCOM_ATTR;
typedef uint8_t ATTR_ENABLE_MEM_EARLY_DATA_SCOM_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_DQ_MAP attribute
typedef uint8_t MSS_VPD_DQ_MAP_ATTR[2][72];
typedef uint8_t ATTR_MSS_VPD_DQ_MAP_type[2][72];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 72>, 2> ATTR_MSS_VPD_DQ_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_0_VERSION_LAYOUT attribute
typedef uint8_t MSS_VPD_MR_0_VERSION_LAYOUT_ATTR;
typedef uint8_t ATTR_MSS_VPD_MR_0_VERSION_LAYOUT_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_1_VERSION_DATA attribute
typedef uint8_t MSS_VPD_MR_1_VERSION_DATA_ATTR;
typedef uint8_t ATTR_MSS_VPD_MR_1_VERSION_DATA_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_2_SIGNATURE_HASH attribute
typedef uint32_t MSS_VPD_MR_2_SIGNATURE_HASH_ATTR;
typedef uint32_t ATTR_MSS_VPD_MR_2_SIGNATURE_HASH_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_DPHY_GPO attribute
typedef uint8_t MSS_VPD_MR_DPHY_GPO_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_DPHY_GPO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_DPHY_GPO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_DPHY_RLO attribute
typedef uint8_t MSS_VPD_MR_DPHY_RLO_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_DPHY_RLO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_DPHY_RLO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_DPHY_WLO attribute
typedef uint8_t MSS_VPD_MR_DPHY_WLO_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_DPHY_WLO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_DPHY_WLO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_2N_MODE_AUTOSET attribute
typedef uint8_t MSS_VPD_MR_MC_2N_MODE_AUTOSET_ATTR;
typedef uint8_t ATTR_MSS_VPD_MR_MC_2N_MODE_AUTOSET_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1 attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP attribute
typedef uint8_t MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP_ATTR[2];
typedef uint8_t ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_TSYS_ADR attribute
typedef uint8_t MSS_VPD_MR_TSYS_ADR_ATTR;
typedef uint8_t ATTR_MSS_VPD_MR_TSYS_ADR_type;

// Type aliases and/or sizes for ATTR_MSS_VPD_MR_TSYS_DATA attribute
typedef uint8_t MSS_VPD_MR_TSYS_DATA_ATTR;
typedef uint8_t ATTR_MSS_VPD_MR_TSYS_DATA_type;

// Type aliases and/or sizes for ATTR_PROC_FSP_BAR_ENABLE attribute
typedef uint8_t PROC_FSP_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_FSP_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_FSP_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_FSP_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_FSP_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_FSP_BAR_SIZE attribute
typedef uint64_t PROC_FSP_BAR_SIZE_ATTR;
typedef uint64_t ATTR_PROC_FSP_BAR_SIZE_type;

// Type aliases and/or sizes for ATTR_PROC_FSP_MMIO_MASK_SIZE attribute
typedef uint64_t PROC_FSP_MMIO_MASK_SIZE_ATTR;
typedef uint64_t ATTR_PROC_FSP_MMIO_MASK_SIZE_type;

// Type aliases and/or sizes for ATTR_PROC_NPU_PHY0_BAR_ENABLE attribute
typedef uint8_t PROC_NPU_PHY0_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NPU_PHY0_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NPU_PHY0_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_NPU_PHY0_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_NPU_PHY0_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NPU_PHY1_BAR_ENABLE attribute
typedef uint8_t PROC_NPU_PHY1_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NPU_PHY1_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NPU_PHY1_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_NPU_PHY1_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_NPU_PHY1_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NPU_MMIO_BAR_ENABLE attribute
typedef uint8_t PROC_NPU_MMIO_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NPU_MMIO_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NPU_MMIO_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_NPU_MMIO_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_NPU_MMIO_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NPU0_MMIO_BAR_ENABLE attribute
typedef uint8_t PROC_NPU0_MMIO_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NPU0_MMIO_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NPU0_MMIO_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_NPU0_MMIO_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_NPU0_MMIO_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NPU1_MMIO_BAR_ENABLE attribute
typedef uint8_t PROC_NPU1_MMIO_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NPU1_MMIO_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NPU1_MMIO_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_NPU1_MMIO_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_NPU1_MMIO_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NPU2_MMIO_BAR_ENABLE attribute
typedef uint8_t PROC_NPU2_MMIO_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NPU2_MMIO_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NPU2_MMIO_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_NPU2_MMIO_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_NPU2_MMIO_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NPU0_PRI_CONFIG attribute
typedef uint8_t PROC_NPU0_PRI_CONFIG_ATTR[4];
typedef uint8_t ATTR_PROC_NPU0_PRI_CONFIG_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_NPU0_PRI_CONFIG_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_NPU1_PRI_CONFIG attribute
typedef uint8_t PROC_NPU1_PRI_CONFIG_ATTR[4];
typedef uint8_t ATTR_PROC_NPU1_PRI_CONFIG_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_NPU1_PRI_CONFIG_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_NPU2_PRI_CONFIG attribute
typedef uint8_t PROC_NPU2_PRI_CONFIG_ATTR[4];
typedef uint8_t ATTR_PROC_NPU2_PRI_CONFIG_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_NPU2_PRI_CONFIG_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PSI_BRIDGE_BAR_ENABLE attribute
typedef uint8_t PROC_PSI_BRIDGE_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_PSI_BRIDGE_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_PSI_BRIDGE_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_PSI_BRIDGE_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_PSI_BRIDGE_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NX_RNG_BAR_ENABLE attribute
typedef uint8_t PROC_NX_RNG_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NX_RNG_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NX_RNG_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_NX_RNG_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_NX_RNG_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_NX_RNG_FAILED_INT_ENABLE attribute
typedef uint8_t PROC_NX_RNG_FAILED_INT_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_NX_RNG_FAILED_INT_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_NX_RNG_FAILED_INT_ADDR attribute
typedef uint64_t PROC_NX_RNG_FAILED_INT_ADDR_ATTR;
typedef uint64_t ATTR_PROC_NX_RNG_FAILED_INT_ADDR_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_PC_BAR_ENABLE attribute
typedef uint8_t PROC_INT_CQ_PC_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_INT_CQ_PC_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_MASK attribute
typedef uint64_t PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_MASK_ATTR;
typedef uint64_t ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_MASK_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_VC_BAR_ENABLE attribute
typedef uint8_t PROC_INT_CQ_VC_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_INT_CQ_VC_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_MASK attribute
typedef uint64_t PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_MASK_ATTR;
typedef uint64_t ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_MASK_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_TM1_BAR_ENABLE attribute
typedef uint8_t PROC_INT_CQ_TM1_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_INT_CQ_TM1_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_TM1_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_INT_CQ_TM1_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_INT_CQ_TM1_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_TM1_BAR_PAGE_SIZE attribute
typedef uint8_t PROC_INT_CQ_TM1_BAR_PAGE_SIZE_ATTR;
typedef uint8_t ATTR_PROC_INT_CQ_TM1_BAR_PAGE_SIZE_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_IC_BAR_ENABLE attribute
typedef uint8_t PROC_INT_CQ_IC_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_INT_CQ_IC_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_IC_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_INT_CQ_IC_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_PROC_INT_CQ_IC_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_INT_CQ_IC_BAR_PAGE_SIZE attribute
typedef uint8_t PROC_INT_CQ_IC_BAR_PAGE_SIZE_ATTR;
typedef uint8_t ATTR_PROC_INT_CQ_IC_BAR_PAGE_SIZE_type;

// Type aliases and/or sizes for ATTR_IO_OMI_DCCAL_FLAGS attribute
typedef uint8_t IO_OMI_DCCAL_FLAGS_ATTR;
typedef uint8_t ATTR_IO_OMI_DCCAL_FLAGS_type;

// Type aliases and/or sizes for ATTR_IO_OMI_DEBUG attribute
typedef uint8_t IO_OMI_DEBUG_ATTR;
typedef uint8_t ATTR_IO_OMI_DEBUG_type;

// Type aliases and/or sizes for ATTR_IO_OMI_MFG_CHK attribute
typedef uint8_t IO_OMI_MFG_CHK_ATTR;
typedef uint8_t ATTR_IO_OMI_MFG_CHK_type;

// Type aliases and/or sizes for ATTR_IO_OMI_MFG_MIN_EYE_WIDTH attribute
typedef uint8_t IO_OMI_MFG_MIN_EYE_WIDTH_ATTR;
typedef uint8_t ATTR_IO_OMI_MFG_MIN_EYE_WIDTH_type;

// Type aliases and/or sizes for ATTR_IO_OMI_TX_MARGIN_RATIO attribute
typedef uint8_t IO_OMI_TX_MARGIN_RATIO_ATTR;
typedef uint8_t ATTR_IO_OMI_TX_MARGIN_RATIO_type;

// Type aliases and/or sizes for ATTR_IO_OMI_TX_FFE_PRECURSOR attribute
typedef uint8_t IO_OMI_TX_FFE_PRECURSOR_ATTR;
typedef uint8_t ATTR_IO_OMI_TX_FFE_PRECURSOR_type;

// Type aliases and/or sizes for ATTR_IO_OMI_TX_FFE_POSTCURSOR attribute
typedef uint8_t IO_OMI_TX_FFE_POSTCURSOR_ATTR;
typedef uint8_t ATTR_IO_OMI_TX_FFE_POSTCURSOR_type;

// Type aliases and/or sizes for ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_EVEN attribute
typedef uint8_t IO_OMI_MFG_STRESS_PR_OFFSET_EVEN_ATTR;
typedef uint8_t ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_EVEN_type;

// Type aliases and/or sizes for ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_ODD attribute
typedef uint8_t IO_OMI_MFG_STRESS_PR_OFFSET_ODD_ATTR;
typedef uint8_t ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_ODD_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_IOP_SWAP attribute
typedef uint8_t PROC_PCIE_IOP_SWAP_ATTR;
typedef uint8_t ATTR_PROC_PCIE_IOP_SWAP_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_BAR_ENABLE attribute
typedef uint8_t PROC_PCIE_BAR_ENABLE_ATTR[3];
typedef uint8_t ATTR_PROC_PCIE_BAR_ENABLE_type[3];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 3> ATTR_PROC_PCIE_BAR_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET_ATTR[6];
typedef uint64_t ATTR_PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET_type[6];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 6> ATTR_PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET_ATTR[6];
typedef uint64_t ATTR_PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET_type[6];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 6> ATTR_PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET_ATTR[6];
typedef uint64_t ATTR_PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET_type[6];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 6> ATTR_PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_BAR_SIZE attribute
typedef uint64_t PROC_PCIE_BAR_SIZE_ATTR[3];
typedef uint64_t ATTR_PROC_PCIE_BAR_SIZE_type[3];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 3> ATTR_PROC_PCIE_BAR_SIZE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_CDR_GAIN attribute
typedef uint8_t PROC_PCIE_PCS_RX_CDR_GAIN_ATTR[4];
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_CDR_GAIN_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_PCIE_PCS_RX_CDR_GAIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_PK_INIT attribute
typedef uint8_t PROC_PCIE_PCS_RX_PK_INIT_ATTR[4][16];
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_PK_INIT_type[4][16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 16>, 4> ATTR_PROC_PCIE_PCS_RX_PK_INIT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_INIT_GAIN attribute
typedef uint8_t PROC_PCIE_PCS_RX_INIT_GAIN_ATTR[4][16];
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_INIT_GAIN_type[4][16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 16>, 4> ATTR_PROC_PCIE_PCS_RX_INIT_GAIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_SIGDET_LVL attribute
typedef uint8_t PROC_PCIE_PCS_RX_SIGDET_LVL_ATTR[4];
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_SIGDET_LVL_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_PCIE_PCS_RX_SIGDET_LVL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_ROT_CDR_LOOKAHEAD attribute
typedef uint8_t PROC_PCIE_PCS_RX_ROT_CDR_LOOKAHEAD_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_ROT_CDR_LOOKAHEAD_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_ROT_CDR_SSC attribute
typedef uint8_t PROC_PCIE_PCS_RX_ROT_CDR_SSC_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_ROT_CDR_SSC_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_ROT_EXTEL attribute
typedef uint8_t PROC_PCIE_PCS_RX_ROT_EXTEL_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_ROT_EXTEL_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_ROT_RST_FW attribute
typedef uint8_t PROC_PCIE_PCS_RX_ROT_RST_FW_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_ROT_RST_FW_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLA attribute
typedef uint8_t PROC_PCIE_PCS_PCLCK_CNTL_PLLA_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLA_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLB attribute
typedef uint8_t PROC_PCIE_PCS_PCLCK_CNTL_PLLB_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLB_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_TX_DCLCK_ROT attribute
typedef uint16_t PROC_PCIE_PCS_TX_DCLCK_ROT_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_TX_DCLCK_ROT_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG1 attribute
typedef uint16_t PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG1_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG1_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG2 attribute
typedef uint16_t PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG2_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG2_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_TX_POWER_SEQ_ENABLE attribute
typedef uint8_t PROC_PCIE_PCS_TX_POWER_SEQ_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_TX_POWER_SEQ_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG1 attribute
typedef uint16_t PROC_PCIE_PCS_RX_VGA_CNTL_REG1_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG1_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG2 attribute
typedef uint16_t PROC_PCIE_PCS_RX_VGA_CNTL_REG2_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG2_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_DFE_FDDC attribute
typedef uint8_t PROC_PCIE_PCS_RX_DFE_FDDC_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_RX_DFE_FDDC_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_M_CNTL attribute
typedef uint16_t PROC_PCIE_PCS_M_CNTL_ATTR[4];
typedef uint16_t ATTR_PROC_PCIE_PCS_M_CNTL_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 4> ATTR_PROC_PCIE_PCS_M_CNTL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_CACHE_INJ_MODE attribute
typedef uint8_t PROC_PCIE_CACHE_INJ_MODE_ATTR;
typedef uint8_t ATTR_PROC_PCIE_CACHE_INJ_MODE_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_CACHE_INJ_THROTTLE attribute
typedef uint8_t PROC_PCIE_CACHE_INJ_THROTTLE_ATTR;
typedef uint8_t ATTR_PROC_PCIE_CACHE_INJ_THROTTLE_type;

// Type aliases and/or sizes for ATTR_OMI_DL_LN_REV_ENABLE attribute
typedef uint8_t OMI_DL_LN_REV_ENABLE_ATTR;
typedef uint8_t ATTR_OMI_DL_LN_REV_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_DPLL_DIVIDER attribute
typedef uint32_t PROC_DPLL_DIVIDER_ATTR;
typedef uint32_t ATTR_PROC_DPLL_DIVIDER_type;

// Type aliases and/or sizes for ATTR_STOPGPE_BOOT_COPIER_IVPR_OFFSET attribute
typedef uint32_t STOPGPE_BOOT_COPIER_IVPR_OFFSET_ATTR;
typedef uint32_t ATTR_STOPGPE_BOOT_COPIER_IVPR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PSTATEGPE_BOOT_COPIER_IVPR_OFFSET attribute
typedef uint32_t PSTATEGPE_BOOT_COPIER_IVPR_OFFSET_ATTR;
typedef uint32_t ATTR_PSTATEGPE_BOOT_COPIER_IVPR_OFFSET_type;

// Type aliases and/or sizes for ATTR_PM_SPWUP_IGNORE_XSTOP_FLAG attribute
typedef uint8_t PM_SPWUP_IGNORE_XSTOP_FLAG_ATTR;
typedef uint8_t ATTR_PM_SPWUP_IGNORE_XSTOP_FLAG_type;

// Type aliases and/or sizes for ATTR_OCC_LFIR attribute
typedef uint64_t OCC_LFIR_ATTR;
typedef uint64_t ATTR_OCC_LFIR_type;

// Type aliases and/or sizes for ATTR_PBA_LFIR attribute
typedef uint64_t PBA_LFIR_ATTR;
typedef uint64_t ATTR_PBA_LFIR_type;

// Type aliases and/or sizes for ATTR_PM_FIRINIT_DONE_ONCE_FLAG attribute
typedef uint8_t PM_FIRINIT_DONE_ONCE_FLAG_ATTR;
typedef uint8_t ATTR_PM_FIRINIT_DONE_ONCE_FLAG_type;

// Type aliases and/or sizes for ATTR_QUAD_PPM_ERRMASK attribute
typedef uint32_t QUAD_PPM_ERRMASK_ATTR;
typedef uint32_t ATTR_QUAD_PPM_ERRMASK_type;

// Type aliases and/or sizes for ATTR_CORE_PPM_ERRMASK attribute
typedef uint32_t CORE_PPM_ERRMASK_ATTR;
typedef uint32_t ATTR_CORE_PPM_ERRMASK_type;

// Type aliases and/or sizes for ATTR_CME_LOCAL_FIRMASK attribute
typedef uint32_t CME_LOCAL_FIRMASK_ATTR;
typedef uint32_t ATTR_CME_LOCAL_FIRMASK_type;

// Type aliases and/or sizes for ATTR_L2_HASCLOCKS attribute
typedef uint8_t L2_HASCLOCKS_ATTR;
typedef uint8_t ATTR_L2_HASCLOCKS_type;

// Type aliases and/or sizes for ATTR_L3_HASCLOCKS attribute
typedef uint8_t L3_HASCLOCKS_ATTR;
typedef uint8_t ATTR_L3_HASCLOCKS_type;

// Type aliases and/or sizes for ATTR_C0_EXEC_HASCLOCKS attribute
typedef uint8_t C0_EXEC_HASCLOCKS_ATTR;
typedef uint8_t ATTR_C0_EXEC_HASCLOCKS_type;

// Type aliases and/or sizes for ATTR_C1_EXEC_HASCLOCKS attribute
typedef uint8_t C1_EXEC_HASCLOCKS_ATTR;
typedef uint8_t ATTR_C1_EXEC_HASCLOCKS_type;

// Type aliases and/or sizes for ATTR_C0_PC_HASCLOCKS attribute
typedef uint8_t C0_PC_HASCLOCKS_ATTR;
typedef uint8_t ATTR_C0_PC_HASCLOCKS_type;

// Type aliases and/or sizes for ATTR_C1_PC_HASCLOCKS attribute
typedef uint8_t C1_PC_HASCLOCKS_ATTR;
typedef uint8_t ATTR_C1_PC_HASCLOCKS_type;

// Type aliases and/or sizes for ATTR_L2_HASPOWER attribute
typedef uint8_t L2_HASPOWER_ATTR;
typedef uint8_t ATTR_L2_HASPOWER_type;

// Type aliases and/or sizes for ATTR_L3_HASPOWER attribute
typedef uint8_t L3_HASPOWER_ATTR;
typedef uint8_t ATTR_L3_HASPOWER_type;

// Type aliases and/or sizes for ATTR_C0_HASPOWER attribute
typedef uint8_t C0_HASPOWER_ATTR;
typedef uint8_t ATTR_C0_HASPOWER_type;

// Type aliases and/or sizes for ATTR_C1_HASPOWER attribute
typedef uint8_t C1_HASPOWER_ATTR;
typedef uint8_t ATTR_C1_HASPOWER_type;

// Type aliases and/or sizes for ATTR_PM_SPIPSS_FRAME_SIZE attribute
typedef uint8_t PM_SPIPSS_FRAME_SIZE_ATTR;
typedef uint8_t ATTR_PM_SPIPSS_FRAME_SIZE_type;

// Type aliases and/or sizes for ATTR_PM_SPIPSS_IN_DELAY attribute
typedef uint8_t PM_SPIPSS_IN_DELAY_ATTR;
typedef uint8_t ATTR_PM_SPIPSS_IN_DELAY_type;

// Type aliases and/or sizes for ATTR_PM_SPIPSS_CLOCK_POLARITY attribute
typedef uint8_t PM_SPIPSS_CLOCK_POLARITY_ATTR;
typedef uint8_t ATTR_PM_SPIPSS_CLOCK_POLARITY_type;

// Type aliases and/or sizes for ATTR_PM_SPIPSS_CLOCK_PHASE attribute
typedef uint8_t PM_SPIPSS_CLOCK_PHASE_ATTR;
typedef uint8_t ATTR_PM_SPIPSS_CLOCK_PHASE_type;

// Type aliases and/or sizes for ATTR_PM_SPIPSS_CLOCK_DIVIDER attribute
typedef uint16_t PM_SPIPSS_CLOCK_DIVIDER_ATTR;
typedef uint16_t ATTR_PM_SPIPSS_CLOCK_DIVIDER_type;

// Type aliases and/or sizes for ATTR_PM_SPIPSS_INTER_FRAME_DELAY_SETTING attribute
typedef uint32_t PM_SPIPSS_INTER_FRAME_DELAY_SETTING_ATTR;
typedef uint32_t ATTR_PM_SPIPSS_INTER_FRAME_DELAY_SETTING_type;

// Type aliases and/or sizes for ATTR_PSTATES_ENABLED attribute
typedef uint8_t PSTATES_ENABLED_ATTR;
typedef uint8_t ATTR_PSTATES_ENABLED_type;

// Type aliases and/or sizes for ATTR_RESCLK_ENABLED attribute
typedef uint8_t RESCLK_ENABLED_ATTR;
typedef uint8_t ATTR_RESCLK_ENABLED_type;

// Type aliases and/or sizes for ATTR_VDM_ENABLED attribute
typedef uint8_t VDM_ENABLED_ATTR;
typedef uint8_t ATTR_VDM_ENABLED_type;

// Type aliases and/or sizes for ATTR_IVRM_ENABLED attribute
typedef uint8_t IVRM_ENABLED_ATTR;
typedef uint8_t ATTR_IVRM_ENABLED_type;

// Type aliases and/or sizes for ATTR_WOF_ENABLED attribute
typedef uint8_t WOF_ENABLED_ATTR;
typedef uint8_t ATTR_WOF_ENABLED_type;

// Type aliases and/or sizes for ATTR_WOV_OVERV_ENABLED attribute
typedef uint8_t WOV_OVERV_ENABLED_ATTR;
typedef uint8_t ATTR_WOV_OVERV_ENABLED_type;

// Type aliases and/or sizes for ATTR_WOV_UNDERV_ENABLED attribute
typedef uint8_t WOV_UNDERV_ENABLED_ATTR;
typedef uint8_t ATTR_WOV_UNDERV_ENABLED_type;

// Type aliases and/or sizes for ATTR_SGPE_PHANTOM_HALT_ENABLE attribute
typedef uint8_t SGPE_PHANTOM_HALT_ENABLE_ATTR;
typedef uint8_t ATTR_SGPE_PHANTOM_HALT_ENABLE_type;

// Type aliases and/or sizes for ATTR_PGPE_PHANTOM_HALT_ENABLE attribute
typedef uint8_t PGPE_PHANTOM_HALT_ENABLE_ATTR;
typedef uint8_t ATTR_PGPE_PHANTOM_HALT_ENABLE_type;

// Type aliases and/or sizes for ATTR_CME_STOP_PHANTOM_HALT_ENABLE attribute
typedef uint8_t CME_STOP_PHANTOM_HALT_ENABLE_ATTR;
typedef uint8_t ATTR_CME_STOP_PHANTOM_HALT_ENABLE_type;

// Type aliases and/or sizes for ATTR_CME_PSTATE_PHANTOM_HALT_ENABLE attribute
typedef uint8_t CME_PSTATE_PHANTOM_HALT_ENABLE_ATTR;
typedef uint8_t ATTR_CME_PSTATE_PHANTOM_HALT_ENABLE_type;

// Type aliases and/or sizes for ATTR_CORE_INSIDE_SPECIAL_WAKEUP attribute
typedef uint8_t CORE_INSIDE_SPECIAL_WAKEUP_ATTR;
typedef uint8_t ATTR_CORE_INSIDE_SPECIAL_WAKEUP_type;

// Type aliases and/or sizes for ATTR_EX_INSIDE_SPECIAL_WAKEUP attribute
typedef uint8_t EX_INSIDE_SPECIAL_WAKEUP_ATTR;
typedef uint8_t ATTR_EX_INSIDE_SPECIAL_WAKEUP_type;

// Type aliases and/or sizes for ATTR_EQ_INSIDE_SPECIAL_WAKEUP attribute
typedef uint8_t EQ_INSIDE_SPECIAL_WAKEUP_ATTR;
typedef uint8_t ATTR_EQ_INSIDE_SPECIAL_WAKEUP_type;

// Type aliases and/or sizes for ATTR_INITIATED_PM_RESET attribute
typedef uint8_t INITIATED_PM_RESET_ATTR;
typedef uint8_t ATTR_INITIATED_PM_RESET_type;

// Type aliases and/or sizes for ATTR_PM_RESET_PHASE attribute
typedef uint8_t PM_RESET_PHASE_ATTR;
typedef uint8_t ATTR_PM_RESET_PHASE_type;

// Type aliases and/or sizes for ATTR_PM_MALF_CYCLE attribute
typedef uint8_t PM_MALF_CYCLE_ATTR;
typedef uint8_t ATTR_PM_MALF_CYCLE_type;

// Type aliases and/or sizes for ATTR_OMI_DL_GROUP_POS attribute
typedef uint8_t OMI_DL_GROUP_POS_ATTR;
typedef uint8_t ATTR_OMI_DL_GROUP_POS_type;

// Type aliases and/or sizes for ATTR_OMI_DL_NUM attribute
typedef uint8_t OMI_DL_NUM_ATTR;
typedef uint8_t ATTR_OMI_DL_NUM_type;

// Type aliases and/or sizes for ATTR_IO_DMI_PROC_DCCAL_FLAGS attribute
typedef uint8_t IO_DMI_PROC_DCCAL_FLAGS_ATTR;
typedef uint8_t ATTR_IO_DMI_PROC_DCCAL_FLAGS_type;

// Type aliases and/or sizes for ATTR_IO_DMI_PROC_DEBUG attribute
typedef uint8_t IO_DMI_PROC_DEBUG_ATTR;
typedef uint8_t ATTR_IO_DMI_PROC_DEBUG_type;

// Type aliases and/or sizes for ATTR_IO_DMI_PROC_TX_MARGIN_RATIO attribute
typedef uint8_t IO_DMI_PROC_TX_MARGIN_RATIO_ATTR;
typedef uint8_t ATTR_IO_DMI_PROC_TX_MARGIN_RATIO_type;

// Type aliases and/or sizes for ATTR_IO_DMI_PROC_TX_FFE_PRECURSOR attribute
typedef uint8_t IO_DMI_PROC_TX_FFE_PRECURSOR_ATTR;
typedef uint8_t ATTR_IO_DMI_PROC_TX_FFE_PRECURSOR_type;

// Type aliases and/or sizes for ATTR_DMI_INBAND_BAR_ENABLE attribute
typedef uint8_t DMI_INBAND_BAR_ENABLE_ATTR;
typedef uint8_t ATTR_DMI_INBAND_BAR_ENABLE_type;

// Type aliases and/or sizes for ATTR_DMI_INBAND_BAR_BASE_ADDR_OFFSET attribute
typedef uint64_t DMI_INBAND_BAR_BASE_ADDR_OFFSET_ATTR;
typedef uint64_t ATTR_DMI_INBAND_BAR_BASE_ADDR_OFFSET_type;

// Type aliases and/or sizes for ATTR_MSS_FREQ_OVERRIDE attribute
typedef uint64_t MSS_FREQ_OVERRIDE_ATTR;
typedef uint64_t ATTR_MSS_FREQ_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_MSS_FREQ attribute
typedef uint64_t MSS_FREQ_ATTR;
typedef uint64_t ATTR_MSS_FREQ_type;

// Type aliases and/or sizes for ATTR_MSS_FREQ_BIAS_PERCENTAGE attribute
typedef uint32_t MSS_FREQ_BIAS_PERCENTAGE_ATTR;
typedef uint32_t ATTR_MSS_FREQ_BIAS_PERCENTAGE_type;

// Type aliases and/or sizes for ATTR_EFF_DIMM_SPARE attribute
typedef uint8_t EFF_DIMM_SPARE_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_DIMM_SPARE_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_DIMM_SPARE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_WR_VREF attribute
typedef uint32_t EFF_DRAM_WR_VREF_ATTR[2];
typedef uint32_t ATTR_EFF_DRAM_WR_VREF_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_EFF_DRAM_WR_VREF_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_WR_VREF_SCHMOO attribute
typedef uint32_t EFF_DRAM_WR_VREF_SCHMOO_ATTR[2];
typedef uint32_t ATTR_EFF_DRAM_WR_VREF_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_EFF_DRAM_WR_VREF_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_WRDDR4_VREF_SCHMOO attribute
typedef uint32_t EFF_DRAM_WRDDR4_VREF_SCHMOO_ATTR[2];
typedef uint32_t ATTR_EFF_DRAM_WRDDR4_VREF_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_EFF_DRAM_WRDDR4_VREF_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_SIZE attribute
typedef uint32_t EFF_DIMM_SIZE_ATTR[2][2];
typedef uint32_t ATTR_EFF_DIMM_SIZE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_EFF_DIMM_SIZE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_CL attribute
typedef uint8_t EFF_DRAM_CL_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_CL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_CL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_AL attribute
typedef uint8_t EFF_DRAM_AL_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_AL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_AL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_CWL attribute
typedef uint8_t EFF_DRAM_CWL_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_CWL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_CWL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_RBT attribute
typedef uint8_t EFF_DRAM_RBT_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_RBT_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_RBT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TM attribute
typedef uint8_t EFF_DRAM_TM_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_TM_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_TM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_DLL_RESET attribute
typedef uint8_t EFF_DRAM_DLL_RESET_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_DLL_RESET_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_DLL_RESET_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_DLL_PPD attribute
typedef uint8_t EFF_DRAM_DLL_PPD_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_DLL_PPD_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_DLL_PPD_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_DLL_ENABLE attribute
typedef uint8_t EFF_DRAM_DLL_ENABLE_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_DLL_ENABLE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_DLL_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_WR_LVL_ENABLE attribute
typedef uint8_t EFF_DRAM_WR_LVL_ENABLE_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_WR_LVL_ENABLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_WR_LVL_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_OUTPUT_BUFFER attribute
typedef uint8_t EFF_DRAM_OUTPUT_BUFFER_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_OUTPUT_BUFFER_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_OUTPUT_BUFFER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_PASR attribute
typedef uint8_t EFF_DRAM_PASR_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_PASR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_PASR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_ASR attribute
typedef uint8_t EFF_DRAM_ASR_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_ASR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_ASR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_SRT attribute
typedef uint8_t EFF_DRAM_SRT_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_SRT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_SRT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_MPR_LOC attribute
typedef uint8_t EFF_MPR_LOC_ATTR[2];
typedef uint8_t ATTR_EFF_MPR_LOC_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_MPR_LOC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_MPR_MODE attribute
typedef uint8_t EFF_MPR_MODE_ATTR[2];
typedef uint8_t ATTR_EFF_MPR_MODE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_MPR_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC00 attribute
typedef uint8_t EFF_DIMM_DDR4_RC00_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC00_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC00_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC01 attribute
typedef uint8_t EFF_DIMM_DDR4_RC01_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC01_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC01_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC02 attribute
typedef uint8_t EFF_DIMM_DDR4_RC02_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC02_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC02_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC03 attribute
typedef uint8_t EFF_DIMM_DDR4_RC03_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC03_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC04 attribute
typedef uint8_t EFF_DIMM_DDR4_RC04_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC04_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC05 attribute
typedef uint8_t EFF_DIMM_DDR4_RC05_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC05_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC06_07 attribute
typedef uint8_t EFF_DIMM_DDR4_RC06_07_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC06_07_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC06_07_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC08 attribute
typedef uint8_t EFF_DIMM_DDR4_RC08_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC08_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC08_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC09 attribute
typedef uint8_t EFF_DIMM_DDR4_RC09_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC09_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC09_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC0A attribute
typedef uint8_t EFF_DIMM_DDR4_RC0A_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC0A_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC0A_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC0B attribute
typedef uint8_t EFF_DIMM_DDR4_RC0B_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC0B_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC0B_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC0C attribute
typedef uint8_t EFF_DIMM_DDR4_RC0C_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC0C_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC0C_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC0D attribute
typedef uint8_t EFF_DIMM_DDR4_RC0D_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC0D_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC0D_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC0E attribute
typedef uint8_t EFF_DIMM_DDR4_RC0E_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC0E_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC0E_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC0F attribute
typedef uint8_t EFF_DIMM_DDR4_RC0F_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC0F_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC0F_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_1x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_1x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_1x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_1x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_2x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_2x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_2x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_2x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_3x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_3x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_3x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_3x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_4x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_4x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_4x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_4x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_5x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_5x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_5x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_5x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_6x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_6x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_6x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_6x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_7x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_7x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_7x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_7x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_8x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_8x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_8x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_8x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_9x attribute
typedef uint8_t EFF_DIMM_DDR4_RC_9x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_9x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_9x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_Ax attribute
typedef uint8_t EFF_DIMM_DDR4_RC_Ax_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_Ax_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_Ax_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_RC_Bx attribute
typedef uint8_t EFF_DIMM_DDR4_RC_Bx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_RC_Bx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_RC_Bx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_RCD_MIRROR_MODE attribute
typedef uint8_t EFF_DIMM_RCD_MIRROR_MODE_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_RCD_MIRROR_MODE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_RCD_MIRROR_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_MODE attribute
typedef uint8_t EFF_SCHMOO_MODE_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_MODE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_ADDR_MODE attribute
typedef uint8_t EFF_SCHMOO_ADDR_MODE_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_ADDR_MODE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_ADDR_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_TEST_VALID attribute
typedef uint8_t EFF_SCHMOO_TEST_VALID_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_TEST_VALID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_TEST_VALID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_PARAM_VALID attribute
typedef uint8_t EFF_SCHMOO_PARAM_VALID_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_PARAM_VALID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_PARAM_VALID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_WR_EYE_MIN_MARGIN attribute
typedef uint8_t EFF_SCHMOO_WR_EYE_MIN_MARGIN_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_WR_EYE_MIN_MARGIN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_WR_EYE_MIN_MARGIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_RD_EYE_MIN_MARGIN attribute
typedef uint8_t EFF_SCHMOO_RD_EYE_MIN_MARGIN_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_RD_EYE_MIN_MARGIN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_RD_EYE_MIN_MARGIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_DQS_CLK_MIN_MARGIN attribute
typedef uint8_t EFF_SCHMOO_DQS_CLK_MIN_MARGIN_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_DQS_CLK_MIN_MARGIN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_DQS_CLK_MIN_MARGIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_RD_GATE_MIN_MARGIN attribute
typedef uint8_t EFF_SCHMOO_RD_GATE_MIN_MARGIN_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_RD_GATE_MIN_MARGIN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_RD_GATE_MIN_MARGIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN attribute
typedef uint8_t EFF_SCHMOO_ADDR_CMD_MIN_MARGIN_ATTR[2];
typedef uint8_t ATTR_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_MEMCAL_INTERVAL attribute
typedef uint32_t EFF_MEMCAL_INTERVAL_ATTR[2];
typedef uint32_t ATTR_EFF_MEMCAL_INTERVAL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_EFF_MEMCAL_INTERVAL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_ZQCAL_INTERVAL attribute
typedef uint32_t EFF_ZQCAL_INTERVAL_ATTR[2];
typedef uint32_t ATTR_EFF_ZQCAL_INTERVAL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_EFF_ZQCAL_INTERVAL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_IBM_TYPE attribute
typedef uint8_t EFF_IBM_TYPE_ATTR[2][2];
typedef uint8_t ATTR_EFF_IBM_TYPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_IBM_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_NUM_MASTER_RANKS_PER_DIMM attribute
typedef uint8_t EFF_NUM_MASTER_RANKS_PER_DIMM_ATTR[2][2];
typedef uint8_t ATTR_EFF_NUM_MASTER_RANKS_PER_DIMM_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_NUM_MASTER_RANKS_PER_DIMM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_RANKS_CONFIGED attribute
typedef uint8_t EFF_DIMM_RANKS_CONFIGED_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_RANKS_CONFIGED_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_RANKS_CONFIGED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_PRIM_DIE_COUNT attribute
typedef uint8_t EFF_PRIM_DIE_COUNT_ATTR[2][2];
typedef uint8_t ATTR_EFF_PRIM_DIE_COUNT_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_PRIM_DIE_COUNT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
typedef uint16_t MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT_ATTR[2];
typedef uint16_t ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MEM_M_DRAM_CLOCKS attribute
typedef uint32_t MSS_MEM_M_DRAM_CLOCKS_ATTR[2];
typedef uint32_t ATTR_MSS_MEM_M_DRAM_CLOCKS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_MEM_M_DRAM_CLOCKS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
typedef uint16_t MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT_ATTR[2];
typedef uint16_t ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MEM_WATT_TARGET attribute
typedef uint32_t MSS_MEM_WATT_TARGET_ATTR[2][2];
typedef uint32_t ATTR_MSS_MEM_WATT_TARGET_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_MSS_MEM_WATT_TARGET_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_TOTAL_PWR_SLOPE attribute
typedef uint16_t MSS_TOTAL_PWR_SLOPE_ATTR[2][2];
typedef uint16_t ATTR_MSS_TOTAL_PWR_SLOPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 2>, 2> ATTR_MSS_TOTAL_PWR_SLOPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_TOTAL_PWR_INTERCEPT attribute
typedef uint16_t MSS_TOTAL_PWR_INTERCEPT_ATTR[2][2];
typedef uint16_t ATTR_MSS_TOTAL_PWR_INTERCEPT_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 2>, 2> ATTR_MSS_TOTAL_PWR_INTERCEPT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_DIMM_MAXBANDWIDTH_GBS attribute
typedef uint32_t MSS_DIMM_MAXBANDWIDTH_GBS_ATTR[2][2];
typedef uint32_t ATTR_MSS_DIMM_MAXBANDWIDTH_GBS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_MSS_DIMM_MAXBANDWIDTH_GBS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_DIMM_MAXBANDWIDTH_MRS attribute
typedef uint32_t MSS_DIMM_MAXBANDWIDTH_MRS_ATTR[2][2];
typedef uint32_t ATTR_MSS_DIMM_MAXBANDWIDTH_MRS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_MSS_DIMM_MAXBANDWIDTH_MRS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS attribute
typedef uint32_t MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS_ATTR[2];
typedef uint32_t ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS attribute
typedef uint32_t MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS_ATTR[2];
typedef uint32_t ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_DIMM_MAXPOWER attribute
typedef uint32_t MSS_DIMM_MAXPOWER_ATTR[2][2];
typedef uint32_t ATTR_MSS_DIMM_MAXPOWER_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_MSS_DIMM_MAXPOWER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_DATABUS_UTIL attribute
typedef uint32_t MSS_DATABUS_UTIL_ATTR[2];
typedef uint32_t ATTR_MSS_DATABUS_UTIL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_DATABUS_UTIL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_PORT_MAXPOWER attribute
typedef uint32_t MSS_PORT_MAXPOWER_ATTR[2];
typedef uint32_t ATTR_MSS_PORT_MAXPOWER_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_PORT_MAXPOWER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_DIMM_THERMAL_LIMIT attribute
typedef uint32_t MSS_DIMM_THERMAL_LIMIT_ATTR[2][2];
typedef uint32_t ATTR_MSS_DIMM_THERMAL_LIMIT_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_MSS_DIMM_THERMAL_LIMIT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
typedef uint16_t MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT_ATTR[2];
typedef uint16_t ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_RUNTIME_MEM_M_DRAM_CLOCKS attribute
typedef uint32_t MSS_RUNTIME_MEM_M_DRAM_CLOCKS_ATTR[2];
typedef uint32_t ATTR_MSS_RUNTIME_MEM_M_DRAM_CLOCKS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_RUNTIME_MEM_M_DRAM_CLOCKS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
typedef uint16_t MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT_ATTR[2];
typedef uint16_t ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EFF_DIMM_FUNCTIONAL_VECTOR attribute
typedef uint8_t MSS_EFF_DIMM_FUNCTIONAL_VECTOR_ATTR[2];
typedef uint8_t ATTR_MSS_EFF_DIMM_FUNCTIONAL_VECTOR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EFF_DIMM_FUNCTIONAL_VECTOR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_LPASR attribute
typedef uint8_t EFF_DRAM_LPASR_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_LPASR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_LPASR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_MPR_PAGE attribute
typedef uint8_t EFF_MPR_PAGE_ATTR[2];
typedef uint8_t ATTR_EFF_MPR_PAGE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_MPR_PAGE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_GEARDOWN_MODE attribute
typedef uint8_t EFF_GEARDOWN_MODE_ATTR[2];
typedef uint8_t ATTR_EFF_GEARDOWN_MODE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_GEARDOWN_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_PER_DRAM_ACCESS attribute
typedef uint8_t EFF_PER_DRAM_ACCESS_ATTR[2];
typedef uint8_t ATTR_EFF_PER_DRAM_ACCESS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_PER_DRAM_ACCESS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_TEMP_READOUT attribute
typedef uint8_t EFF_TEMP_READOUT_ATTR[2];
typedef uint8_t ATTR_EFF_TEMP_READOUT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_TEMP_READOUT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_CRC_WR_LATENCY attribute
typedef uint8_t EFF_CRC_WR_LATENCY_ATTR[2];
typedef uint8_t ATTR_EFF_CRC_WR_LATENCY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_CRC_WR_LATENCY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_MPR_RD_FORMAT attribute
typedef uint8_t EFF_MPR_RD_FORMAT_ATTR[2];
typedef uint8_t ATTR_EFF_MPR_RD_FORMAT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_MPR_RD_FORMAT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_MAX_POWERDOWN_MODE attribute
typedef uint8_t EFF_MAX_POWERDOWN_MODE_ATTR[2];
typedef uint8_t ATTR_EFF_MAX_POWERDOWN_MODE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_MAX_POWERDOWN_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_INTERNAL_VREF_MONITOR attribute
typedef uint8_t EFF_INTERNAL_VREF_MONITOR_ATTR[2];
typedef uint8_t ATTR_EFF_INTERNAL_VREF_MONITOR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_INTERNAL_VREF_MONITOR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_CS_CMD_LATENCY attribute
typedef uint8_t EFF_CS_CMD_LATENCY_ATTR[2];
typedef uint8_t ATTR_EFF_CS_CMD_LATENCY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_CS_CMD_LATENCY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_SELF_REF_ABORT attribute
typedef uint8_t EFF_SELF_REF_ABORT_ATTR[2];
typedef uint8_t ATTR_EFF_SELF_REF_ABORT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_SELF_REF_ABORT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_RD_PREAMBLE_TRAIN attribute
typedef uint8_t EFF_RD_PREAMBLE_TRAIN_ATTR[2];
typedef uint8_t ATTR_EFF_RD_PREAMBLE_TRAIN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_RD_PREAMBLE_TRAIN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_RD_PREAMBLE attribute
typedef uint8_t EFF_RD_PREAMBLE_ATTR[2];
typedef uint8_t ATTR_EFF_RD_PREAMBLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_RD_PREAMBLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_WR_PREAMBLE attribute
typedef uint8_t EFF_WR_PREAMBLE_ATTR[2];
typedef uint8_t ATTR_EFF_WR_PREAMBLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_WR_PREAMBLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_CA_PARITY_LATENCY attribute
typedef uint8_t EFF_CA_PARITY_LATENCY_ATTR[2];
typedef uint8_t ATTR_EFF_CA_PARITY_LATENCY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_CA_PARITY_LATENCY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_CRC_ERROR_CLEAR attribute
typedef uint8_t EFF_CRC_ERROR_CLEAR_ATTR[2];
typedef uint8_t ATTR_EFF_CRC_ERROR_CLEAR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_CRC_ERROR_CLEAR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_CA_PARITY_ERROR_STATUS attribute
typedef uint8_t EFF_CA_PARITY_ERROR_STATUS_ATTR[2];
typedef uint8_t ATTR_EFF_CA_PARITY_ERROR_STATUS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_CA_PARITY_ERROR_STATUS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_ODT_INPUT_BUFF attribute
typedef uint8_t EFF_ODT_INPUT_BUFF_ATTR[2];
typedef uint8_t ATTR_EFF_ODT_INPUT_BUFF_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_ODT_INPUT_BUFF_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_CA_PARITY attribute
typedef uint8_t EFF_CA_PARITY_ATTR[2];
typedef uint8_t ATTR_EFF_CA_PARITY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_CA_PARITY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DATA_MASK attribute
typedef uint8_t EFF_DATA_MASK_ATTR[2];
typedef uint8_t ATTR_EFF_DATA_MASK_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DATA_MASK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_WRITE_DBI attribute
typedef uint8_t EFF_WRITE_DBI_ATTR[2];
typedef uint8_t ATTR_EFF_WRITE_DBI_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_WRITE_DBI_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_READ_DBI attribute
typedef uint8_t EFF_READ_DBI_ATTR[2];
typedef uint8_t ATTR_EFF_READ_DBI_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_READ_DBI_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_VREF_DQ_TRAIN_VALUE attribute
typedef uint8_t EFF_VREF_DQ_TRAIN_VALUE_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_VREF_DQ_TRAIN_VALUE_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_VREF_DQ_TRAIN_VALUE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_VREF_DQ_TRAIN_RANGE attribute
typedef uint8_t EFF_VREF_DQ_TRAIN_RANGE_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_VREF_DQ_TRAIN_RANGE_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_VREF_DQ_TRAIN_RANGE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_VREF_DQ_TRAIN_ENABLE attribute
typedef uint8_t EFF_VREF_DQ_TRAIN_ENABLE_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_VREF_DQ_TRAIN_ENABLE_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_VREF_DQ_TRAIN_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_WRITE_CRC attribute
typedef uint8_t EFF_WRITE_CRC_ATTR[2];
typedef uint8_t ATTR_EFF_WRITE_CRC_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_WRITE_CRC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CAL_STEP_ENABLE attribute
typedef uint32_t MSS_CAL_STEP_ENABLE_ATTR[2];
typedef uint32_t ATTR_MSS_CAL_STEP_ENABLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_CAL_STEP_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CUSTOM_TRAINING_ADV_PATTERNS attribute
typedef uint32_t MSS_CUSTOM_TRAINING_ADV_PATTERNS_ATTR[2];
typedef uint32_t ATTR_MSS_CUSTOM_TRAINING_ADV_PATTERNS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_CUSTOM_TRAINING_ADV_PATTERNS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS attribute
typedef uint32_t MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS_ATTR[2];
typedef uint32_t ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2 attribute
typedef uint32_t MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2_ATTR[2];
typedef uint32_t ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CUSTOM_TRAINING_ADV_WR_PATTERN attribute
typedef uint8_t MSS_CUSTOM_TRAINING_ADV_WR_PATTERN_ATTR[2];
typedef uint8_t ATTR_MSS_CUSTOM_TRAINING_ADV_WR_PATTERN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_CUSTOM_TRAINING_ADV_WR_PATTERN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VREF_CAL_ENABLE attribute
typedef uint16_t MSS_VREF_CAL_ENABLE_ATTR[2];
typedef uint16_t ATTR_MSS_VREF_CAL_ENABLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MSS_VREF_CAL_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_RDVREF_CAL_ENABLE attribute
typedef uint16_t MSS_RDVREF_CAL_ENABLE_ATTR[2];
typedef uint16_t ATTR_MSS_RDVREF_CAL_ENABLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MSS_RDVREF_CAL_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_CAL_ABORT_ON_ERROR attribute
typedef uint8_t MSS_CAL_ABORT_ON_ERROR_ATTR;
typedef uint8_t ATTR_MSS_CAL_ABORT_ON_ERROR_type;

// Type aliases and/or sizes for ATTR_MSS_SLEW_RATE_DATA attribute
typedef uint8_t MSS_SLEW_RATE_DATA_ATTR[2][4][4];
typedef uint8_t ATTR_MSS_SLEW_RATE_DATA_type[2][4][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> ATTR_MSS_SLEW_RATE_DATA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_SLEW_RATE_ADR attribute
typedef uint8_t MSS_SLEW_RATE_ADR_ATTR[2][4][4];
typedef uint8_t ATTR_MSS_SLEW_RATE_ADR_type[2][4][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> ATTR_MSS_SLEW_RATE_ADR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCHMOO_MULTIPLE_SETUP_CALL attribute
typedef uint8_t SCHMOO_MULTIPLE_SETUP_CALL_ATTR[2];
typedef uint8_t ATTR_SCHMOO_MULTIPLE_SETUP_CALL_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_SCHMOO_MULTIPLE_SETUP_CALL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_BUFFER_LATENCY attribute
typedef uint8_t EFF_BUFFER_LATENCY_ATTR[2];
typedef uint8_t ATTR_EFF_BUFFER_LATENCY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_BUFFER_LATENCY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_LRDIMM_WORD_X attribute
typedef uint64_t EFF_LRDIMM_WORD_X_ATTR[2][2];
typedef uint64_t ATTR_EFF_LRDIMM_WORD_X_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint64_t, 2>, 2> ATTR_EFF_LRDIMM_WORD_X_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_LRDIMM_MR12_REG attribute
typedef uint8_t LRDIMM_MR12_REG_ATTR[2][2];
typedef uint8_t ATTR_LRDIMM_MR12_REG_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_LRDIMM_MR12_REG_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_LRDIMM_ADDITIONAL_CNTL_WORDS attribute
typedef uint64_t LRDIMM_ADDITIONAL_CNTL_WORDS_ATTR[2][2];
typedef uint64_t ATTR_LRDIMM_ADDITIONAL_CNTL_WORDS_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint64_t, 2>, 2> ATTR_LRDIMM_ADDITIONAL_CNTL_WORDS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_LRDIMM_RANK_MULT_MODE attribute
typedef uint8_t LRDIMM_RANK_MULT_MODE_ATTR[2];
typedef uint8_t ATTR_LRDIMM_RANK_MULT_MODE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_LRDIMM_RANK_MULT_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EFF_VPD_VERSION attribute
typedef uint32_t MSS_EFF_VPD_VERSION_ATTR[2];
typedef uint32_t ATTR_MSS_EFF_VPD_VERSION_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MSS_EFF_VPD_VERSION_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VOLT_OVERRIDE attribute
typedef uint8_t MSS_VOLT_OVERRIDE_ATTR;
typedef uint8_t ATTR_MSS_VOLT_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_MSS_VDDR_OVERIDE_SPD attribute
typedef uint8_t MSS_VDDR_OVERIDE_SPD_ATTR;
typedef uint8_t ATTR_MSS_VDDR_OVERIDE_SPD_type;

// Type aliases and/or sizes for ATTR_MSS_VMEM_REGULATOR_MAX_DIMM_COUNT attribute
typedef uint8_t MSS_VMEM_REGULATOR_MAX_DIMM_COUNT_ATTR;
typedef uint8_t ATTR_MSS_VMEM_REGULATOR_MAX_DIMM_COUNT_type;

// Type aliases and/or sizes for ATTR_MSS_VREF_DAC_NIBBLE attribute
typedef uint8_t MSS_VREF_DAC_NIBBLE_ATTR[2];
typedef uint8_t ATTR_MSS_VREF_DAC_NIBBLE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_VREF_DAC_NIBBLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VCCD_OVERRIDE attribute
typedef uint8_t MSS_VCCD_OVERRIDE_ATTR;
typedef uint8_t ATTR_MSS_VCCD_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_EFF_DRAM_MAC attribute
typedef uint16_t EFF_DRAM_MAC_ATTR[2][2];
typedef uint16_t ATTR_EFF_DRAM_MAC_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 2>, 2> ATTR_EFF_DRAM_MAC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_MODULE_BUS_WIDTH attribute
typedef uint8_t EFF_DRAM_MODULE_BUS_WIDTH_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_MODULE_BUS_WIDTH_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_MODULE_BUS_WIDTH_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_RTT_NOM attribute
typedef uint8_t EFF_DRAM_RTT_NOM_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_DRAM_RTT_NOM_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_DRAM_RTT_NOM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_RTT_WR attribute
typedef uint8_t EFF_DRAM_RTT_WR_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_DRAM_RTT_WR_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_DRAM_RTT_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_RTT_PARK attribute
typedef uint8_t EFF_DRAM_RTT_PARK_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_DRAM_RTT_PARK_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_DRAM_RTT_PARK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC00 attribute
typedef uint8_t EFF_DIMM_DDR4_BC00_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC00_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC00_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC01 attribute
typedef uint8_t EFF_DIMM_DDR4_BC01_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC01_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC01_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC02 attribute
typedef uint8_t EFF_DIMM_DDR4_BC02_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC02_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC02_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC03 attribute
typedef uint8_t EFF_DIMM_DDR4_BC03_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC03_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC04 attribute
typedef uint8_t EFF_DIMM_DDR4_BC04_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC04_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC05 attribute
typedef uint8_t EFF_DIMM_DDR4_BC05_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC05_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC06 attribute
typedef uint8_t EFF_DIMM_DDR4_BC06_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC06_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC06_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC07 attribute
typedef uint8_t EFF_DIMM_DDR4_BC07_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC07_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC07_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC08 attribute
typedef uint8_t EFF_DIMM_DDR4_BC08_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC08_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC08_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC09 attribute
typedef uint8_t EFF_DIMM_DDR4_BC09_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC09_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC09_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC0A attribute
typedef uint8_t EFF_DIMM_DDR4_BC0A_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC0A_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC0A_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC0B attribute
typedef uint8_t EFF_DIMM_DDR4_BC0B_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC0B_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC0B_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC0C attribute
typedef uint8_t EFF_DIMM_DDR4_BC0C_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC0C_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC0C_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC0D attribute
typedef uint8_t EFF_DIMM_DDR4_BC0D_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC0D_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC0D_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC0E attribute
typedef uint8_t EFF_DIMM_DDR4_BC0E_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC0E_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC0E_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_BC0F attribute
typedef uint8_t EFF_DIMM_DDR4_BC0F_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_BC0F_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_BC0F_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F0BC1x attribute
typedef uint8_t EFF_DIMM_DDR4_F0BC1x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F0BC1x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F0BC1x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BC2x attribute
typedef uint8_t EFF_DIMM_DDR4_F30BC2x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BC2x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BC2x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BC3x attribute
typedef uint8_t EFF_DIMM_DDR4_F30BC3x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BC3x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BC3x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BC4x attribute
typedef uint8_t EFF_DIMM_DDR4_F30BC4x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BC4x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BC4x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BC5x attribute
typedef uint8_t EFF_DIMM_DDR4_F30BC5x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BC5x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BC5x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F0BC6x attribute
typedef uint8_t EFF_DIMM_DDR4_F0BC6x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F0BC6x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F0BC6x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F70BC7x attribute
typedef uint8_t EFF_DIMM_DDR4_F70BC7x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F70BC7x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F70BC7x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BC8x attribute
typedef uint8_t EFF_DIMM_DDR4_F30BC8x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BC8x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BC8x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BC9x attribute
typedef uint8_t EFF_DIMM_DDR4_F30BC9x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BC9x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BC9x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BCAx attribute
typedef uint8_t EFF_DIMM_DDR4_F30BCAx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BCAx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BCAx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F30BCBx attribute
typedef uint8_t EFF_DIMM_DDR4_F30BCBx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F30BCBx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F30BCBx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F0BCCx attribute
typedef uint8_t EFF_DIMM_DDR4_F0BCCx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F0BCCx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F0BCCx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F0BCDx attribute
typedef uint8_t EFF_DIMM_DDR4_F0BCDx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F0BCDx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F0BCDx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F0BCEx attribute
typedef uint8_t EFF_DIMM_DDR4_F0BCEx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F0BCEx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F0BCEx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F0BCFx attribute
typedef uint8_t EFF_DIMM_DDR4_F0BCFx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F0BCFx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F0BCFx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F1BCCx attribute
typedef uint8_t EFF_DIMM_DDR4_F1BCCx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F1BCCx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F1BCCx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F1BCDx attribute
typedef uint8_t EFF_DIMM_DDR4_F1BCDx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F1BCDx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F1BCDx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F1BCEx attribute
typedef uint8_t EFF_DIMM_DDR4_F1BCEx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F1BCEx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F1BCEx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F1BCFx attribute
typedef uint8_t EFF_DIMM_DDR4_F1BCFx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F1BCFx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F1BCFx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F2BCEx attribute
typedef uint8_t EFF_DIMM_DDR4_F2BCEx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F2BCEx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F2BCEx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F4BC0x attribute
typedef uint8_t EFF_DIMM_DDR4_F4BC0x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F4BC0x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F4BC0x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F4BC1x attribute
typedef uint8_t EFF_DIMM_DDR4_F4BC1x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F4BC1x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F4BC1x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F4BC2x attribute
typedef uint8_t EFF_DIMM_DDR4_F4BC2x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F4BC2x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F4BC2x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F4BC3x attribute
typedef uint8_t EFF_DIMM_DDR4_F4BC3x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F4BC3x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F4BC3x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F4BC4x attribute
typedef uint8_t EFF_DIMM_DDR4_F4BC4x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F4BC4x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F4BC4x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F4BC5x attribute
typedef uint8_t EFF_DIMM_DDR4_F4BC5x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F4BC5x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F4BC5x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F4BC6x attribute
typedef uint8_t EFF_DIMM_DDR4_F4BC6x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F4BC6x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F4BC6x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F5BC0x attribute
typedef uint8_t EFF_DIMM_DDR4_F5BC0x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F5BC0x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F5BC0x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F5BC1x attribute
typedef uint8_t EFF_DIMM_DDR4_F5BC1x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F5BC1x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F5BC1x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F5BC2x attribute
typedef uint8_t EFF_DIMM_DDR4_F5BC2x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F5BC2x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F5BC2x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F5BC3x attribute
typedef uint8_t EFF_DIMM_DDR4_F5BC3x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F5BC3x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F5BC3x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F5BC5x attribute
typedef uint8_t EFF_DIMM_DDR4_F5BC5x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F5BC5x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F5BC5x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F5BC6x attribute
typedef uint8_t EFF_DIMM_DDR4_F5BC6x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F5BC6x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F5BC6x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F6BC0x attribute
typedef uint8_t EFF_DIMM_DDR4_F6BC0x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F6BC0x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F6BC0x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F6BC1x attribute
typedef uint8_t EFF_DIMM_DDR4_F6BC1x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F6BC1x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F6BC1x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F6BC2x attribute
typedef uint8_t EFF_DIMM_DDR4_F6BC2x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F6BC2x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F6BC2x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F6BC3x attribute
typedef uint8_t EFF_DIMM_DDR4_F6BC3x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F6BC3x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F6BC3x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F6BC4x attribute
typedef uint8_t EFF_DIMM_DDR4_F6BC4x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F6BC4x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F6BC4x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F6BC5x attribute
typedef uint8_t EFF_DIMM_DDR4_F6BC5x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F6BC5x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F6BC5x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BC8x attribute
typedef uint8_t EFF_DIMM_DDR4_F74BC8x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BC8x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BC8x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BC9x attribute
typedef uint8_t EFF_DIMM_DDR4_F74BC9x_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BC9x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BC9x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BCAx attribute
typedef uint8_t EFF_DIMM_DDR4_F74BCAx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BCAx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BCAx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BCBx attribute
typedef uint8_t EFF_DIMM_DDR4_F74BCBx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BCBx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BCBx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BCCx attribute
typedef uint8_t EFF_DIMM_DDR4_F74BCCx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BCCx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BCCx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BCDx attribute
typedef uint8_t EFF_DIMM_DDR4_F74BCDx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BCDx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BCDx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BCEx attribute
typedef uint8_t EFF_DIMM_DDR4_F74BCEx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BCEx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BCEx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DIMM_DDR4_F74BCFx attribute
typedef uint8_t EFF_DIMM_DDR4_F74BCFx_ATTR[2][2];
typedef uint8_t ATTR_EFF_DIMM_DDR4_F74BCFx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DIMM_DDR4_F74BCFx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_RON attribute
typedef uint8_t EFF_DRAM_RON_ATTR[2][2];
typedef uint8_t ATTR_EFF_DRAM_RON_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_EFF_DRAM_RON_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TDQS attribute
typedef uint8_t EFF_DRAM_TDQS_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TDQS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TDQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_ODIC attribute
typedef uint8_t EFF_DRAM_ODIC_ATTR[2][2][4];
typedef uint8_t ATTR_EFF_DRAM_ODIC_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_EFF_DRAM_ODIC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_LRDIMM_TRAINING_PATTERN attribute
typedef uint8_t MSS_LRDIMM_TRAINING_PATTERN_ATTR[5];
typedef uint8_t ATTR_MSS_LRDIMM_TRAINING_PATTERN_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 5> ATTR_MSS_LRDIMM_TRAINING_PATTERN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EFF_ODT_RD attribute
typedef uint8_t MSS_EFF_ODT_RD_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_EFF_ODT_RD_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_EFF_ODT_RD_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EFF_ODT_WR attribute
typedef uint8_t MSS_EFF_ODT_WR_ATTR[2][2][4];
typedef uint8_t ATTR_MSS_EFF_ODT_WR_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_MSS_EFF_ODT_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EFF_DPHY_WLO attribute
typedef uint8_t MSS_EFF_DPHY_WLO_ATTR[2];
typedef uint8_t ATTR_MSS_EFF_DPHY_WLO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EFF_DPHY_WLO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EFF_DPHY_RLO attribute
typedef uint8_t MSS_EFF_DPHY_RLO_ATTR[2];
typedef uint8_t ATTR_MSS_EFF_DPHY_RLO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EFF_DPHY_RLO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TREFI attribute
typedef uint16_t EFF_DRAM_TREFI_ATTR[2];
typedef uint16_t ATTR_EFF_DRAM_TREFI_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_EFF_DRAM_TREFI_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRTP attribute
typedef uint8_t EFF_DRAM_TRTP_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRTP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRTP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRFC_DLR attribute
typedef uint8_t EFF_DRAM_TRFC_DLR_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRFC_DLR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRFC_DLR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TFAW_DLR attribute
typedef uint8_t EFF_DRAM_TFAW_DLR_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TFAW_DLR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TFAW_DLR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TRRD_DLR attribute
typedef uint8_t EFF_DRAM_TRRD_DLR_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TRRD_DLR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TRRD_DLR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EFF_DRAM_TXS attribute
typedef uint8_t EFF_DRAM_TXS_ATTR[2];
typedef uint8_t ATTR_EFF_DRAM_TXS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_EFF_DRAM_TXS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_IGNORE_PLUG_RULES attribute
typedef uint8_t MSS_IGNORE_PLUG_RULES_ATTR;
typedef uint8_t ATTR_MSS_IGNORE_PLUG_RULES_type;

// Type aliases and/or sizes for ATTR_MSS_MVPD_FWMS attribute
typedef uint32_t MSS_MVPD_FWMS_ATTR[2][8];
typedef uint32_t ATTR_MSS_MVPD_FWMS_type[2][8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 8>, 2> ATTR_MSS_MVPD_FWMS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_REORDER_QUEUE_SETTING attribute
typedef uint8_t MSS_REORDER_QUEUE_SETTING_ATTR;
typedef uint8_t ATTR_MSS_REORDER_QUEUE_SETTING_type;

// Type aliases and/or sizes for ATTR_EFF_RANK_GROUP_OVERRIDE attribute
typedef uint16_t EFF_RANK_GROUP_OVERRIDE_ATTR[2][2];
typedef uint16_t ATTR_EFF_RANK_GROUP_OVERRIDE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 2>, 2> ATTR_EFF_RANK_GROUP_OVERRIDE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_RTT_NOM_OVERRIDE_DISABLE attribute
typedef uint8_t MSS_RTT_NOM_OVERRIDE_DISABLE_ATTR[2][2];
typedef uint8_t ATTR_MSS_RTT_NOM_OVERRIDE_DISABLE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_MSS_RTT_NOM_OVERRIDE_DISABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EFF_WR_CRC attribute
typedef uint8_t MSS_EFF_WR_CRC_ATTR[2];
typedef uint8_t ATTR_MSS_EFF_WR_CRC_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EFF_WR_CRC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_WR_VREF_OFFSET attribute
typedef int8_t MSS_WR_VREF_OFFSET_ATTR;
typedef int8_t ATTR_MSS_WR_VREF_OFFSET_type;

// Type aliases and/or sizes for ATTR_SBE_NVDIMM_IN_PORT attribute
typedef uint8_t SBE_NVDIMM_IN_PORT_ATTR;
typedef uint8_t ATTR_SBE_NVDIMM_IN_PORT_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT8_1 attribute
typedef uint8_t SCRATCH_UINT8_1_ATTR;
typedef uint8_t ATTR_SCRATCH_UINT8_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT8_2 attribute
typedef uint8_t SCRATCH_UINT8_2_ATTR;
typedef uint8_t ATTR_SCRATCH_UINT8_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT16_1 attribute
typedef uint16_t SCRATCH_UINT16_1_ATTR;
typedef uint16_t ATTR_SCRATCH_UINT16_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT16_2 attribute
typedef uint16_t SCRATCH_UINT16_2_ATTR;
typedef uint16_t ATTR_SCRATCH_UINT16_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT32_1 attribute
typedef uint32_t SCRATCH_UINT32_1_ATTR;
typedef uint32_t ATTR_SCRATCH_UINT32_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT32_2 attribute
typedef uint32_t SCRATCH_UINT32_2_ATTR;
typedef uint32_t ATTR_SCRATCH_UINT32_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT64_1 attribute
typedef uint64_t SCRATCH_UINT64_1_ATTR;
typedef uint64_t ATTR_SCRATCH_UINT64_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT64_2 attribute
typedef uint64_t SCRATCH_UINT64_2_ATTR;
typedef uint64_t ATTR_SCRATCH_UINT64_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_UINT8_ARRAY_1 attribute
typedef uint8_t SCRATCH_UINT8_ARRAY_1_ATTR[32];
typedef uint8_t ATTR_SCRATCH_UINT8_ARRAY_1_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 32> ATTR_SCRATCH_UINT8_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_UINT8_ARRAY_2 attribute
typedef uint8_t SCRATCH_UINT8_ARRAY_2_ATTR[2][3][4];
typedef uint8_t ATTR_SCRATCH_UINT8_ARRAY_2_type[2][3][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 3>, 2> ATTR_SCRATCH_UINT8_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_UINT16_ARRAY_1 attribute
typedef uint16_t SCRATCH_UINT16_ARRAY_1_ATTR[16];
typedef uint16_t ATTR_SCRATCH_UINT16_ARRAY_1_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 16> ATTR_SCRATCH_UINT16_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_UINT16_ARRAY_2 attribute
typedef uint16_t SCRATCH_UINT16_ARRAY_2_ATTR[2][3][2];
typedef uint16_t ATTR_SCRATCH_UINT16_ARRAY_2_type[2][3][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint16_t, 2>, 3>, 2> ATTR_SCRATCH_UINT16_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_UINT32_ARRAY_1 attribute
typedef uint32_t SCRATCH_UINT32_ARRAY_1_ATTR[8];
typedef uint32_t ATTR_SCRATCH_UINT32_ARRAY_1_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 8> ATTR_SCRATCH_UINT32_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_UINT32_ARRAY_2 attribute
typedef uint32_t SCRATCH_UINT32_ARRAY_2_ATTR[2][3];
typedef uint32_t ATTR_SCRATCH_UINT32_ARRAY_2_type[2][3];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 3>, 2> ATTR_SCRATCH_UINT32_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_UINT64_ARRAY_1 attribute
typedef uint64_t SCRATCH_UINT64_ARRAY_1_ATTR[4];
typedef uint64_t ATTR_SCRATCH_UINT64_ARRAY_1_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 4> ATTR_SCRATCH_UINT64_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_UINT64_ARRAY_2 attribute
typedef uint64_t SCRATCH_UINT64_ARRAY_2_ATTR[2][2];
typedef uint64_t ATTR_SCRATCH_UINT64_ARRAY_2_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint64_t, 2>, 2> ATTR_SCRATCH_UINT64_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT8_1 attribute
typedef int8_t SCRATCH_INT8_1_ATTR;
typedef int8_t ATTR_SCRATCH_INT8_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT8_2 attribute
typedef int8_t SCRATCH_INT8_2_ATTR;
typedef int8_t ATTR_SCRATCH_INT8_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT16_1 attribute
typedef int16_t SCRATCH_INT16_1_ATTR;
typedef int16_t ATTR_SCRATCH_INT16_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT16_2 attribute
typedef int16_t SCRATCH_INT16_2_ATTR;
typedef int16_t ATTR_SCRATCH_INT16_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT32_1 attribute
typedef int32_t SCRATCH_INT32_1_ATTR;
typedef int32_t ATTR_SCRATCH_INT32_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT32_2 attribute
typedef int32_t SCRATCH_INT32_2_ATTR;
typedef int32_t ATTR_SCRATCH_INT32_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT64_1 attribute
typedef int64_t SCRATCH_INT64_1_ATTR;
typedef int64_t ATTR_SCRATCH_INT64_1_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT64_2 attribute
typedef int64_t SCRATCH_INT64_2_ATTR;
typedef int64_t ATTR_SCRATCH_INT64_2_type;

// Type aliases and/or sizes for ATTR_SCRATCH_INT8_ARRAY_1 attribute
typedef int8_t SCRATCH_INT8_ARRAY_1_ATTR[32];
typedef int8_t ATTR_SCRATCH_INT8_ARRAY_1_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<int8_t, 32> ATTR_SCRATCH_INT8_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT8_ARRAY_2 attribute
typedef int8_t SCRATCH_INT8_ARRAY_2_ATTR[2][3][4];
typedef int8_t ATTR_SCRATCH_INT8_ARRAY_2_type[2][3][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<int8_t, 4>, 3>, 2> ATTR_SCRATCH_INT8_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT16_ARRAY_1 attribute
typedef int16_t SCRATCH_INT16_ARRAY_1_ATTR[16];
typedef int16_t ATTR_SCRATCH_INT16_ARRAY_1_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<int16_t, 16> ATTR_SCRATCH_INT16_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT16_ARRAY_2 attribute
typedef int16_t SCRATCH_INT16_ARRAY_2_ATTR[2][3][2];
typedef int16_t ATTR_SCRATCH_INT16_ARRAY_2_type[2][3][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<int16_t, 2>, 3>, 2> ATTR_SCRATCH_INT16_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT32_ARRAY_1 attribute
typedef int32_t SCRATCH_INT32_ARRAY_1_ATTR[8];
typedef int32_t ATTR_SCRATCH_INT32_ARRAY_1_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<int32_t, 8> ATTR_SCRATCH_INT32_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT32_ARRAY_2 attribute
typedef int32_t SCRATCH_INT32_ARRAY_2_ATTR[2][3];
typedef int32_t ATTR_SCRATCH_INT32_ARRAY_2_type[2][3];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<int32_t, 3>, 2> ATTR_SCRATCH_INT32_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT64_ARRAY_1 attribute
typedef int64_t SCRATCH_INT64_ARRAY_1_ATTR[4];
typedef int64_t ATTR_SCRATCH_INT64_ARRAY_1_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<int64_t, 4> ATTR_SCRATCH_INT64_ARRAY_1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SCRATCH_INT64_ARRAY_2 attribute
typedef int64_t SCRATCH_INT64_ARRAY_2_ATTR[2][2];
typedef int64_t ATTR_SCRATCH_INT64_ARRAY_2_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<int64_t, 2>, 2> ATTR_SCRATCH_INT64_ARRAY_2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_I2C_DEV_TYPE attribute
typedef uint8_t I2C_DEV_TYPE_ATTR;
typedef uint8_t ATTR_I2C_DEV_TYPE_type;

// Type aliases and/or sizes for ATTR_I2C_SUB_POS attribute
typedef uint8_t I2C_SUB_POS_ATTR;
typedef uint8_t ATTR_I2C_SUB_POS_type;

// Type aliases and/or sizes for ATTR_FAPI_POS attribute
typedef uint32_t FAPI_POS_ATTR;
typedef uint32_t ATTR_FAPI_POS_type;

// Type aliases and/or sizes for ATTR_IS_SIMULATION attribute
typedef uint8_t IS_SIMULATION_ATTR;
typedef uint8_t ATTR_IS_SIMULATION_type;

// Type aliases and/or sizes for ATTR_EXECUTION_PLATFORM attribute
typedef uint8_t EXECUTION_PLATFORM_ATTR;
typedef uint8_t ATTR_EXECUTION_PLATFORM_type;

// Type aliases and/or sizes for ATTR_MNFG_FLAGS attribute
typedef uint64_t MNFG_FLAGS_ATTR;
typedef uint64_t ATTR_MNFG_FLAGS_type;

// Type aliases and/or sizes for ATTR_RECONFIGURE_LOOP attribute
typedef uint8_t RECONFIGURE_LOOP_ATTR;
typedef uint8_t ATTR_RECONFIGURE_LOOP_type;

// Type aliases and/or sizes for ATTR_MINI_EC attribute
typedef uint8_t MINI_EC_ATTR;
typedef uint8_t ATTR_MINI_EC_type;

// Type aliases and/or sizes for ATTR_IO_DMI_CEN_TX_MARGIN_RATIO attribute
typedef uint8_t IO_DMI_CEN_TX_MARGIN_RATIO_ATTR;
typedef uint8_t ATTR_IO_DMI_CEN_TX_MARGIN_RATIO_type;

// Type aliases and/or sizes for ATTR_IO_DMI_CEN_TX_FFE_POSTCURSOR attribute
typedef uint8_t IO_DMI_CEN_TX_FFE_POSTCURSOR_ATTR;
typedef uint8_t ATTR_IO_DMI_CEN_TX_FFE_POSTCURSOR_type;

// Type aliases and/or sizes for ATTR_CEN_DMI_REFCLOCK_RCVR_TERM attribute
typedef uint8_t CEN_DMI_REFCLOCK_RCVR_TERM_ATTR;
typedef uint8_t ATTR_CEN_DMI_REFCLOCK_RCVR_TERM_type;

// Type aliases and/or sizes for ATTR_CEN_DDR_REFCLOCK_RCVR_TERM attribute
typedef uint8_t CEN_DDR_REFCLOCK_RCVR_TERM_ATTR;
typedef uint8_t ATTR_CEN_DDR_REFCLOCK_RCVR_TERM_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_MODES attribute
typedef uint8_t CEN_MCBIST_ADDR_MODES_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_MODES_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_RANK attribute
typedef uint8_t CEN_MCBIST_RANK_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_RANK_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_START_ADDR attribute
typedef uint64_t CEN_MCBIST_START_ADDR_ATTR;
typedef uint64_t ATTR_CEN_MCBIST_START_ADDR_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_END_ADDR attribute
typedef uint64_t CEN_MCBIST_END_ADDR_ATTR;
typedef uint64_t ATTR_CEN_MCBIST_END_ADDR_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ERROR_CAPTURE attribute
typedef uint8_t CEN_MCBIST_ERROR_CAPTURE_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ERROR_CAPTURE_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_MAX_TIMEOUT attribute
typedef uint64_t CEN_MCBIST_MAX_TIMEOUT_ATTR;
typedef uint64_t ATTR_CEN_MCBIST_MAX_TIMEOUT_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_PRINT_PORT attribute
typedef uint8_t CEN_MCBIST_PRINT_PORT_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_PRINT_PORT_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_STOP_ON_ERROR attribute
typedef uint8_t CEN_MCBIST_STOP_ON_ERROR_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_STOP_ON_ERROR_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_DATA_SEED attribute
typedef uint32_t CEN_MCBIST_DATA_SEED_ATTR;
typedef uint32_t ATTR_CEN_MCBIST_DATA_SEED_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_INTER attribute
typedef uint8_t CEN_MCBIST_ADDR_INTER_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_INTER_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_NUM_ROWS attribute
typedef uint8_t CEN_MCBIST_ADDR_NUM_ROWS_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_NUM_ROWS_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_NUM_COLS attribute
typedef uint8_t CEN_MCBIST_ADDR_NUM_COLS_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_NUM_COLS_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_RANK attribute
typedef uint8_t CEN_MCBIST_ADDR_RANK_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_RANK_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_BANK attribute
typedef uint8_t CEN_MCBIST_ADDR_BANK_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_BANK_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_SLAVE_RANK_ON attribute
typedef uint8_t CEN_MCBIST_ADDR_SLAVE_RANK_ON_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_SLAVE_RANK_ON_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_STR_MAP attribute
typedef uint64_t CEN_MCBIST_ADDR_STR_MAP_ATTR;
typedef uint64_t ATTR_CEN_MCBIST_ADDR_STR_MAP_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_ADDR_RAND attribute
typedef uint8_t CEN_MCBIST_ADDR_RAND_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_ADDR_RAND_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_DRAM_DEVICE_TYPE attribute
typedef uint8_t CEN_SPD_DRAM_DEVICE_TYPE_ATTR;
typedef uint8_t ATTR_CEN_SPD_DRAM_DEVICE_TYPE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_TYPE attribute
typedef uint8_t CEN_SPD_MODULE_TYPE_ATTR;
typedef uint8_t ATTR_CEN_SPD_MODULE_TYPE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_CUSTOM attribute
typedef uint8_t CEN_SPD_CUSTOM_ATTR;
typedef uint8_t ATTR_CEN_SPD_CUSTOM_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_DENSITY attribute
typedef uint8_t CEN_SPD_SDRAM_DENSITY_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_DENSITY_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_BANKS attribute
typedef uint8_t CEN_SPD_SDRAM_BANKS_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_BANKS_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_ROWS attribute
typedef uint8_t CEN_SPD_SDRAM_ROWS_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_ROWS_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_COLUMNS attribute
typedef uint8_t CEN_SPD_SDRAM_COLUMNS_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_COLUMNS_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_NOMINAL_VOLTAGE attribute
typedef uint8_t CEN_SPD_MODULE_NOMINAL_VOLTAGE_ATTR;
typedef uint8_t ATTR_CEN_SPD_MODULE_NOMINAL_VOLTAGE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_NUM_RANKS attribute
typedef uint8_t CEN_SPD_NUM_RANKS_ATTR;
typedef uint8_t ATTR_CEN_SPD_NUM_RANKS_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_DRAM_WIDTH attribute
typedef uint8_t CEN_SPD_DRAM_WIDTH_ATTR;
typedef uint8_t ATTR_CEN_SPD_DRAM_WIDTH_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_MEMORY_BUS_WIDTH attribute
typedef uint8_t CEN_SPD_MODULE_MEMORY_BUS_WIDTH_ATTR;
typedef uint8_t ATTR_CEN_SPD_MODULE_MEMORY_BUS_WIDTH_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TCKMIN attribute
typedef uint8_t CEN_SPD_TCKMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TCKMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_CAS_LATENCIES_SUPPORTED attribute
typedef uint32_t CEN_SPD_CAS_LATENCIES_SUPPORTED_ATTR;
typedef uint32_t ATTR_CEN_SPD_CAS_LATENCIES_SUPPORTED_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TAAMIN attribute
typedef uint8_t CEN_SPD_TAAMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TAAMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRCDMIN attribute
typedef uint8_t CEN_SPD_TRCDMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TRCDMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRPMIN attribute
typedef uint8_t CEN_SPD_TRPMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TRPMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRASMIN attribute
typedef uint32_t CEN_SPD_TRASMIN_ATTR;
typedef uint32_t ATTR_CEN_SPD_TRASMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRCMIN attribute
typedef uint32_t CEN_SPD_TRCMIN_ATTR;
typedef uint32_t ATTR_CEN_SPD_TRCMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TFAWMIN attribute
typedef uint32_t CEN_SPD_TFAWMIN_ATTR;
typedef uint32_t ATTR_CEN_SPD_TFAWMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_OPTIONAL_FEATURES attribute
typedef uint8_t CEN_SPD_SDRAM_OPTIONAL_FEATURES_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_OPTIONAL_FEATURES_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_THERMAL_AND_REFRESH_OPTIONS attribute
typedef uint8_t CEN_SPD_SDRAM_THERMAL_AND_REFRESH_OPTIONS_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_THERMAL_AND_REFRESH_OPTIONS_type;

// Type aliases and/or sizes for ATTR_ROW_REPAIR_SPPR_SUPPORTED attribute
typedef uint8_t ROW_REPAIR_SPPR_SUPPORTED_ATTR;
typedef uint8_t ATTR_ROW_REPAIR_SPPR_SUPPORTED_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_THERMAL_SENSOR attribute
typedef uint8_t CEN_SPD_MODULE_THERMAL_SENSOR_ATTR;
typedef uint8_t ATTR_CEN_SPD_MODULE_THERMAL_SENSOR_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_DEVICE_TYPE attribute
typedef uint8_t CEN_SPD_SDRAM_DEVICE_TYPE_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_SIGNAL_LOADING attribute
typedef uint8_t CEN_SPD_SDRAM_DEVICE_TYPE_SIGNAL_LOADING_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_SIGNAL_LOADING_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_DIE_COUNT attribute
typedef uint8_t CEN_SPD_SDRAM_DIE_COUNT_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_DIE_COUNT_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TCKMIN attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TCKMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TCKMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TAAMIN attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TAAMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TAAMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TRCDMIN attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TRCDMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TRCDMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TRPMIN attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TRPMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TRPMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TRCMIN attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TRCMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TRCMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURERS_JEDEC_ID_CODE attribute
typedef uint32_t CEN_SPD_MODULE_ID_MODULE_MANUFACTURERS_JEDEC_ID_CODE_ATTR;
typedef uint32_t ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURERS_JEDEC_ID_CODE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_LOCATION attribute
typedef uint8_t CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_LOCATION_ATTR;
typedef uint8_t ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_LOCATION_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_DATE attribute
typedef uint32_t CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_DATE_ATTR;
typedef uint32_t ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_DATE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_ID_MODULE_SERIAL_NUMBER attribute
typedef uint32_t CEN_SPD_MODULE_ID_MODULE_SERIAL_NUMBER_ATTR;
typedef uint32_t ATTR_CEN_SPD_MODULE_ID_MODULE_SERIAL_NUMBER_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_PART_NUMBER attribute
typedef uint8_t CEN_SPD_MODULE_PART_NUMBER_ATTR[20];
typedef uint8_t ATTR_CEN_SPD_MODULE_PART_NUMBER_type[20];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 20> ATTR_CEN_SPD_MODULE_PART_NUMBER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_REVISION_CODE attribute
typedef uint32_t CEN_SPD_MODULE_REVISION_CODE_ATTR;
typedef uint32_t ATTR_CEN_SPD_MODULE_REVISION_CODE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_DRAM_MANUFACTURER_JEDEC_ID_CODE attribute
typedef uint32_t CEN_SPD_DRAM_MANUFACTURER_JEDEC_ID_CODE_ATTR;
typedef uint32_t ATTR_CEN_SPD_DRAM_MANUFACTURER_JEDEC_ID_CODE_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_BAD_DQ_DATA attribute
typedef uint8_t CEN_SPD_BAD_DQ_DATA_ATTR[80];
typedef uint8_t ATTR_CEN_SPD_BAD_DQ_DATA_type[80];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 80> ATTR_CEN_SPD_BAD_DQ_DATA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_SPD_FTB_DIVIDEND attribute
typedef uint8_t CEN_SPD_FTB_DIVIDEND_ATTR;
typedef uint8_t ATTR_CEN_SPD_FTB_DIVIDEND_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FTB_DIVISOR attribute
typedef uint8_t CEN_SPD_FTB_DIVISOR_ATTR;
typedef uint8_t ATTR_CEN_SPD_FTB_DIVISOR_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MTB_DIVIDEND attribute
typedef uint8_t CEN_SPD_MTB_DIVIDEND_ATTR;
typedef uint8_t ATTR_CEN_SPD_MTB_DIVIDEND_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MTB_DIVISOR attribute
typedef uint8_t CEN_SPD_MTB_DIVISOR_ATTR;
typedef uint8_t ATTR_CEN_SPD_MTB_DIVISOR_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TWRMIN attribute
typedef uint8_t CEN_SPD_TWRMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TWRMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRRDMIN attribute
typedef uint8_t CEN_SPD_TRRDMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TRRDMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRFCMIN attribute
typedef uint32_t CEN_SPD_TRFCMIN_ATTR;
typedef uint32_t ATTR_CEN_SPD_TRFCMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TWTRMIN attribute
typedef uint8_t CEN_SPD_TWTRMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TWTRMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRTPMIN attribute
typedef uint8_t CEN_SPD_TRTPMIN_ATTR;
typedef uint8_t ATTR_CEN_SPD_TRTPMIN_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_SDRAM_BANKGROUPS_DDR4 attribute
typedef uint8_t CEN_SPD_SDRAM_BANKGROUPS_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_SDRAM_BANKGROUPS_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TIMEBASE_MTB_DDR4 attribute
typedef uint8_t CEN_SPD_TIMEBASE_MTB_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_TIMEBASE_MTB_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TIMEBASE_FTB_DDR4 attribute
typedef uint8_t CEN_SPD_TIMEBASE_FTB_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_TIMEBASE_FTB_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TCKMAX_DDR4 attribute
typedef uint8_t CEN_SPD_TCKMAX_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_TCKMAX_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRFC1MIN_DDR4 attribute
typedef uint32_t CEN_SPD_TRFC1MIN_DDR4_ATTR;
typedef uint32_t ATTR_CEN_SPD_TRFC1MIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRFC2MIN_DDR4 attribute
typedef uint32_t CEN_SPD_TRFC2MIN_DDR4_ATTR;
typedef uint32_t ATTR_CEN_SPD_TRFC2MIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRFC4MIN_DDR4 attribute
typedef uint32_t CEN_SPD_TRFC4MIN_DDR4_ATTR;
typedef uint32_t ATTR_CEN_SPD_TRFC4MIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRRDSMIN_DDR4 attribute
typedef uint8_t CEN_SPD_TRRDSMIN_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_TRRDSMIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TRRDLMIN_DDR4 attribute
typedef uint8_t CEN_SPD_TRRDLMIN_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_TRRDLMIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_TCCDLMIN_DDR4 attribute
typedef uint8_t CEN_SPD_TCCDLMIN_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_TCCDLMIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TCCDLMIN_DDR4 attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TCCDLMIN_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TCCDLMIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TRRDLMIN_DDR4 attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TRRDLMIN_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TRRDLMIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TRRDSMIN_DDR4 attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TRRDSMIN_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TRRDSMIN_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_FINE_OFFSET_TCKMAX_DDR4 attribute
typedef uint8_t CEN_SPD_FINE_OFFSET_TCKMAX_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_FINE_OFFSET_TCKMAX_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_CRC_BASE_CONFIG_DDR4 attribute
typedef uint32_t CEN_SPD_CRC_BASE_CONFIG_DDR4_ATTR;
typedef uint32_t ATTR_CEN_SPD_CRC_BASE_CONFIG_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_DRAM_STEPPING_DDR4 attribute
typedef uint8_t CEN_SPD_DRAM_STEPPING_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_DRAM_STEPPING_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_CRC_MNFG_SEC_DDR4 attribute
typedef uint32_t CEN_SPD_CRC_MNFG_SEC_DDR4_ATTR;
typedef uint32_t ATTR_CEN_SPD_CRC_MNFG_SEC_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_ADDR_MAP_REG_TO_DRAM attribute
typedef uint8_t CEN_SPD_ADDR_MAP_REG_TO_DRAM_ATTR;
typedef uint8_t ATTR_CEN_SPD_ADDR_MAP_REG_TO_DRAM_type;

// Type aliases and/or sizes for ATTR_CEN_SPD_MODULE_REVISION_CODE_DDR4 attribute
typedef uint8_t CEN_SPD_MODULE_REVISION_CODE_DDR4_ATTR;
typedef uint8_t ATTR_CEN_SPD_MODULE_REVISION_CODE_DDR4_type;

// Type aliases and/or sizes for ATTR_CEN_ISDIMM_MBVPD_INDEX attribute
typedef uint8_t CEN_ISDIMM_MBVPD_INDEX_ATTR;
typedef uint8_t ATTR_CEN_ISDIMM_MBVPD_INDEX_type;

// Type aliases and/or sizes for ATTR_CEN_DQ_TO_DIMM_CONN_DQ attribute
typedef uint8_t CEN_DQ_TO_DIMM_CONN_DQ_ATTR[80];
typedef uint8_t ATTR_CEN_DQ_TO_DIMM_CONN_DQ_type[80];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 80> ATTR_CEN_DQ_TO_DIMM_CONN_DQ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MBA_PORT attribute
typedef uint8_t CEN_MBA_PORT_ATTR;
typedef uint8_t ATTR_CEN_MBA_PORT_type;

// Type aliases and/or sizes for ATTR_CEN_MBA_DIMM attribute
typedef uint8_t CEN_MBA_DIMM_ATTR;
typedef uint8_t ATTR_CEN_MBA_DIMM_type;

// Type aliases and/or sizes for ATTR_CEN_MBVPD_VERSION attribute
typedef uint32_t CEN_MBVPD_VERSION_ATTR;
typedef uint32_t ATTR_CEN_MBVPD_VERSION_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DATABUS_UTIL_PER_MBA attribute
typedef uint8_t CEN_MSS_DATABUS_UTIL_PER_MBA_ATTR;
typedef uint8_t ATTR_CEN_MSS_DATABUS_UTIL_PER_MBA_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_ADDRESS_MIRRORING attribute
typedef uint8_t CEN_EFF_DRAM_ADDRESS_MIRRORING_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DRAM_ADDRESS_MIRRORING_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DRAM_ADDRESS_MIRRORING_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_FREQ_OVERRIDE attribute
typedef uint32_t CEN_MSS_FREQ_OVERRIDE_ATTR;
typedef uint32_t ATTR_CEN_MSS_FREQ_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_FREQ attribute
typedef uint32_t CEN_MSS_FREQ_ATTR;
typedef uint32_t ATTR_CEN_MSS_FREQ_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_FREQ_BIAS_PERCENTAGE attribute
typedef uint32_t CEN_MSS_FREQ_BIAS_PERCENTAGE_ATTR;
typedef uint32_t ATTR_CEN_MSS_FREQ_BIAS_PERCENTAGE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VREF_CAL_CNTL attribute
typedef uint8_t CEN_MSS_VREF_CAL_CNTL_ATTR;
typedef uint8_t ATTR_CEN_MSS_VREF_CAL_CNTL_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VREF_CAL_DELTA_FROM_NOMINAL attribute
typedef uint8_t CEN_MSS_VREF_CAL_DELTA_FROM_NOMINAL_ATTR;
typedef uint8_t ATTR_CEN_MSS_VREF_CAL_DELTA_FROM_NOMINAL_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_RANKS_CONFIGED attribute
typedef uint8_t CEN_EFF_DIMM_RANKS_CONFIGED_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_RANKS_CONFIGED_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_RANKS_CONFIGED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_NUM_RANKS_PER_DIMM attribute
typedef uint8_t CEN_EFF_NUM_RANKS_PER_DIMM_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_NUM_RANKS_PER_DIMM_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_NUM_RANKS_PER_DIMM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_TYPE attribute
typedef uint8_t CEN_EFF_DIMM_TYPE_ATTR;
typedef uint8_t ATTR_CEN_EFF_DIMM_TYPE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_CUSTOM_DIMM attribute
typedef uint8_t CEN_EFF_CUSTOM_DIMM_ATTR;
typedef uint8_t ATTR_CEN_EFF_CUSTOM_DIMM_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_WIDTH attribute
typedef uint8_t CEN_EFF_DRAM_WIDTH_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_WIDTH_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_GEN attribute
typedef uint8_t CEN_EFF_DRAM_GEN_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_GEN_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_PRIMARY_RANK_GROUP0 attribute
typedef uint8_t CEN_EFF_PRIMARY_RANK_GROUP0_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_PRIMARY_RANK_GROUP0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_PRIMARY_RANK_GROUP0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_PRIMARY_RANK_GROUP1 attribute
typedef uint8_t CEN_EFF_PRIMARY_RANK_GROUP1_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_PRIMARY_RANK_GROUP1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_PRIMARY_RANK_GROUP1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_PRIMARY_RANK_GROUP2 attribute
typedef uint8_t CEN_EFF_PRIMARY_RANK_GROUP2_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_PRIMARY_RANK_GROUP2_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_PRIMARY_RANK_GROUP2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_PRIMARY_RANK_GROUP3 attribute
typedef uint8_t CEN_EFF_PRIMARY_RANK_GROUP3_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_PRIMARY_RANK_GROUP3_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_PRIMARY_RANK_GROUP3_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_SECONDARY_RANK_GROUP0 attribute
typedef uint8_t CEN_EFF_SECONDARY_RANK_GROUP0_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_SECONDARY_RANK_GROUP0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_SECONDARY_RANK_GROUP0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_SECONDARY_RANK_GROUP1 attribute
typedef uint8_t CEN_EFF_SECONDARY_RANK_GROUP1_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_SECONDARY_RANK_GROUP1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_SECONDARY_RANK_GROUP1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_SECONDARY_RANK_GROUP2 attribute
typedef uint8_t CEN_EFF_SECONDARY_RANK_GROUP2_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_SECONDARY_RANK_GROUP2_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_SECONDARY_RANK_GROUP2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_SECONDARY_RANK_GROUP3 attribute
typedef uint8_t CEN_EFF_SECONDARY_RANK_GROUP3_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_SECONDARY_RANK_GROUP3_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_SECONDARY_RANK_GROUP3_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_TERTIARY_RANK_GROUP0 attribute
typedef uint8_t CEN_EFF_TERTIARY_RANK_GROUP0_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_TERTIARY_RANK_GROUP0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_TERTIARY_RANK_GROUP0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_TERTIARY_RANK_GROUP1 attribute
typedef uint8_t CEN_EFF_TERTIARY_RANK_GROUP1_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_TERTIARY_RANK_GROUP1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_TERTIARY_RANK_GROUP1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_TERTIARY_RANK_GROUP2 attribute
typedef uint8_t CEN_EFF_TERTIARY_RANK_GROUP2_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_TERTIARY_RANK_GROUP2_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_TERTIARY_RANK_GROUP2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_TERTIARY_RANK_GROUP3 attribute
typedef uint8_t CEN_EFF_TERTIARY_RANK_GROUP3_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_TERTIARY_RANK_GROUP3_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_TERTIARY_RANK_GROUP3_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP0 attribute
typedef uint8_t CEN_EFF_QUATERNARY_RANK_GROUP0_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_QUATERNARY_RANK_GROUP0_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_QUATERNARY_RANK_GROUP0_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP1 attribute
typedef uint8_t CEN_EFF_QUATERNARY_RANK_GROUP1_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_QUATERNARY_RANK_GROUP1_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_QUATERNARY_RANK_GROUP1_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP2 attribute
typedef uint8_t CEN_EFF_QUATERNARY_RANK_GROUP2_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_QUATERNARY_RANK_GROUP2_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_QUATERNARY_RANK_GROUP2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP3 attribute
typedef uint8_t CEN_EFF_QUATERNARY_RANK_GROUP3_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_QUATERNARY_RANK_GROUP3_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_QUATERNARY_RANK_GROUP3_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_WR_VREF attribute
typedef uint32_t CEN_EFF_DRAM_WR_VREF_ATTR[2];
typedef uint32_t ATTR_CEN_EFF_DRAM_WR_VREF_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_CEN_EFF_DRAM_WR_VREF_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_WR_VREF_SCHMOO attribute
typedef uint32_t CEN_EFF_DRAM_WR_VREF_SCHMOO_ATTR[2];
typedef uint32_t ATTR_CEN_EFF_DRAM_WR_VREF_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_CEN_EFF_DRAM_WR_VREF_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS attribute
typedef uint8_t CEN_EFF_CEN_DRV_IMP_DQ_DQS_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO attribute
typedef uint32_t CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO_ATTR[2];
typedef uint32_t ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS attribute
typedef uint8_t CEN_EFF_CEN_SLEW_RATE_DQ_DQS_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS attribute
typedef uint8_t CEN_EFF_CEN_RCV_IMP_DQ_DQS_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO attribute
typedef uint32_t CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO_ATTR[2];
typedef uint32_t ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO attribute
typedef uint8_t CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO_ATTR[2];
typedef uint8_t ATTR_CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_RD_VREF attribute
typedef uint32_t CEN_EFF_RD_VREF_ATTR[2];
typedef uint32_t ATTR_CEN_EFF_RD_VREF_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_CEN_EFF_RD_VREF_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_CEN_RD_VREF_SCHMOO attribute
typedef uint32_t CEN_EFF_CEN_RD_VREF_SCHMOO_ATTR[2];
typedef uint32_t ATTR_CEN_EFF_CEN_RD_VREF_SCHMOO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_CEN_EFF_CEN_RD_VREF_SCHMOO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_SIZE attribute
typedef uint8_t CEN_EFF_DIMM_SIZE_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_SIZE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_SIZE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_BANKS attribute
typedef uint8_t CEN_EFF_DRAM_BANKS_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_BANKS_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_ROWS attribute
typedef uint8_t CEN_EFF_DRAM_ROWS_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_ROWS_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_COLS attribute
typedef uint8_t CEN_EFF_DRAM_COLS_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_COLS_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_DENSITY attribute
typedef uint8_t CEN_EFF_DRAM_DENSITY_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_DENSITY_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRCD attribute
typedef uint8_t CEN_EFF_DRAM_TRCD_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TRCD_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRRD attribute
typedef uint8_t CEN_EFF_DRAM_TRRD_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TRRD_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRRD_L attribute
typedef uint8_t CEN_EFF_DRAM_TRRD_L_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TRRD_L_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRP attribute
typedef uint8_t CEN_EFF_DRAM_TRP_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TRP_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRAS attribute
typedef uint8_t CEN_EFF_DRAM_TRAS_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TRAS_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRC attribute
typedef uint8_t CEN_EFF_DRAM_TRC_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TRC_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRFI attribute
typedef uint32_t CEN_EFF_DRAM_TRFI_ATTR;
typedef uint32_t ATTR_CEN_EFF_DRAM_TRFI_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRFC attribute
typedef uint32_t CEN_EFF_DRAM_TRFC_ATTR;
typedef uint32_t ATTR_CEN_EFF_DRAM_TRFC_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TWTR attribute
typedef uint8_t CEN_EFF_DRAM_TWTR_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TWTR_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TWTR_L attribute
typedef uint8_t CEN_EFF_DRAM_TWTR_L_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TWTR_L_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TRTP attribute
typedef uint8_t CEN_EFF_DRAM_TRTP_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TRTP_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TFAW attribute
typedef uint8_t CEN_EFF_DRAM_TFAW_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TFAW_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_BL attribute
typedef uint8_t CEN_EFF_DRAM_BL_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_BL_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_CL attribute
typedef uint8_t CEN_EFF_DRAM_CL_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_CL_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_AL attribute
typedef uint8_t CEN_EFF_DRAM_AL_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_AL_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_CWL attribute
typedef uint8_t CEN_EFF_DRAM_CWL_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_CWL_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_RBT attribute
typedef uint8_t CEN_EFF_DRAM_RBT_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_RBT_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TM attribute
typedef uint8_t CEN_EFF_DRAM_TM_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TM_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_DLL_RESET attribute
typedef uint8_t CEN_EFF_DRAM_DLL_RESET_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_DLL_RESET_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_WR attribute
typedef uint8_t CEN_EFF_DRAM_WR_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_WR_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_DLL_PPD attribute
typedef uint8_t CEN_EFF_DRAM_DLL_PPD_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_DLL_PPD_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_DLL_ENABLE attribute
typedef uint8_t CEN_EFF_DRAM_DLL_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_DLL_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TDQS attribute
typedef uint8_t CEN_EFF_DRAM_TDQS_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TDQS_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_WR_LVL_ENABLE attribute
typedef uint8_t CEN_EFF_DRAM_WR_LVL_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_WR_LVL_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_OUTPUT_BUFFER attribute
typedef uint8_t CEN_EFF_DRAM_OUTPUT_BUFFER_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_OUTPUT_BUFFER_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_PASR attribute
typedef uint8_t CEN_EFF_DRAM_PASR_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_PASR_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_ASR attribute
typedef uint8_t CEN_EFF_DRAM_ASR_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_ASR_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_SRT attribute
typedef uint8_t CEN_EFF_DRAM_SRT_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_SRT_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_MPR_LOC attribute
typedef uint8_t CEN_EFF_MPR_LOC_ATTR;
typedef uint8_t ATTR_CEN_EFF_MPR_LOC_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_MPR_MODE attribute
typedef uint8_t CEN_EFF_MPR_MODE_ATTR;
typedef uint8_t ATTR_CEN_EFF_MPR_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_RCD_CNTL_WORD_0_15 attribute
typedef uint64_t CEN_EFF_DIMM_RCD_CNTL_WORD_0_15_ATTR[2][2];
typedef uint64_t ATTR_CEN_EFF_DIMM_RCD_CNTL_WORD_0_15_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint64_t, 2>, 2> ATTR_CEN_EFF_DIMM_RCD_CNTL_WORD_0_15_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC00 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC00_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC00_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC00_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC01 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC01_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC01_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC01_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC02 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC02_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC02_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC02_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC03 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC03_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC03_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC04 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC04_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC04_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC05 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC05_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC05_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC06_07 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC06_07_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC06_07_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC06_07_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC08 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC08_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC08_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC08_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC09 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC09_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC09_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC09_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC10 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC10_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC10_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC10_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC11 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC11_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC11_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC11_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC12 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC12_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC12_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC12_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC13 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC13_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC13_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC13_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC14 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC14_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC14_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC14_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC15 attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC15_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC15_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC15_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_1x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_1x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_1x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_1x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_2x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_2x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_2x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_2x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_3x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_3x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_3x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_3x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_4x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_4x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_4x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_4x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_5x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_5x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_5x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_5x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_6x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_6x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_6x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_6x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_7x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_7x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_7x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_7x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_8x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_8x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_8x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_8x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_9x attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_9x_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_9x_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_9x_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_Ax attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_Ax_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_Ax_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_Ax_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DIMM_DDR4_RC_Bx attribute
typedef uint8_t CEN_EFF_DIMM_DDR4_RC_Bx_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_DIMM_DDR4_RC_Bx_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_DIMM_DDR4_RC_Bx_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_MODE attribute
typedef uint8_t CEN_EFF_SCHMOO_MODE_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_ADDR_MODE attribute
typedef uint8_t CEN_EFF_SCHMOO_ADDR_MODE_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_ADDR_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_TEST_VALID attribute
typedef uint8_t CEN_EFF_SCHMOO_TEST_VALID_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_TEST_VALID_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_PARAM_VALID attribute
typedef uint8_t CEN_EFF_SCHMOO_PARAM_VALID_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_PARAM_VALID_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_VREF_STEP attribute
typedef uint8_t CEN_EFF_SCHMOO_VREF_STEP_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_VREF_STEP_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_WR_EYE_MIN_MARGIN attribute
typedef uint8_t CEN_EFF_SCHMOO_WR_EYE_MIN_MARGIN_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_WR_EYE_MIN_MARGIN_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_RD_EYE_MIN_MARGIN attribute
typedef uint8_t CEN_EFF_SCHMOO_RD_EYE_MIN_MARGIN_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_RD_EYE_MIN_MARGIN_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_DQS_CLK_MIN_MARGIN attribute
typedef uint8_t CEN_EFF_SCHMOO_DQS_CLK_MIN_MARGIN_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_DQS_CLK_MIN_MARGIN_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_RD_GATE_MIN_MARGIN attribute
typedef uint8_t CEN_EFF_SCHMOO_RD_GATE_MIN_MARGIN_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_RD_GATE_MIN_MARGIN_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN attribute
typedef uint8_t CEN_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN_ATTR;
typedef uint8_t ATTR_CEN_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_MEMCAL_INTERVAL attribute
typedef uint32_t CEN_EFF_MEMCAL_INTERVAL_ATTR;
typedef uint32_t ATTR_CEN_EFF_MEMCAL_INTERVAL_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_ZQCAL_INTERVAL attribute
typedef uint32_t CEN_EFF_ZQCAL_INTERVAL_ATTR;
typedef uint32_t ATTR_CEN_EFF_ZQCAL_INTERVAL_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_IBM_TYPE attribute
typedef uint8_t CEN_EFF_IBM_TYPE_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_IBM_TYPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_IBM_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_NUM_DROPS_PER_PORT attribute
typedef uint8_t CEN_EFF_NUM_DROPS_PER_PORT_ATTR;
typedef uint8_t ATTR_CEN_EFF_NUM_DROPS_PER_PORT_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_STACK_TYPE attribute
typedef uint8_t CEN_EFF_STACK_TYPE_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_STACK_TYPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_STACK_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_NUM_MASTER_RANKS_PER_DIMM attribute
typedef uint8_t CEN_EFF_NUM_MASTER_RANKS_PER_DIMM_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_NUM_MASTER_RANKS_PER_DIMM_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_NUM_MASTER_RANKS_PER_DIMM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_NUM_PACKAGES_PER_RANK attribute
typedef uint8_t CEN_EFF_NUM_PACKAGES_PER_RANK_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_NUM_PACKAGES_PER_RANK_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_NUM_PACKAGES_PER_RANK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_NUM_DIES_PER_PACKAGE attribute
typedef uint8_t CEN_EFF_NUM_DIES_PER_PACKAGE_ATTR[2][2];
typedef uint8_t ATTR_CEN_EFF_NUM_DIES_PER_PACKAGE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 2>, 2> ATTR_CEN_EFF_NUM_DIES_PER_PACKAGE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_MBA attribute
typedef uint32_t CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_MBA_ATTR;
typedef uint32_t ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_MBA_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_MEM_THROTTLE_DENOMINATOR attribute
typedef uint32_t CEN_MSS_MEM_THROTTLE_DENOMINATOR_ATTR;
typedef uint32_t ATTR_CEN_MSS_MEM_THROTTLE_DENOMINATOR_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
typedef uint32_t CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_CHIP_ATTR;
typedef uint32_t ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_CHIP_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_MEM_WATT_TARGET attribute
typedef uint32_t CEN_MSS_MEM_WATT_TARGET_ATTR;
typedef uint32_t ATTR_CEN_MSS_MEM_WATT_TARGET_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_POWER_SLOPE attribute
typedef uint32_t CEN_MSS_POWER_SLOPE_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_POWER_SLOPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_POWER_SLOPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_POWER_SLOPE2 attribute
typedef uint32_t CEN_MSS_POWER_SLOPE2_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_POWER_SLOPE2_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_POWER_SLOPE2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_POWER_INT attribute
typedef uint32_t CEN_MSS_POWER_INT_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_POWER_INT_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_POWER_INT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_POWER_INT2 attribute
typedef uint32_t CEN_MSS_POWER_INT2_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_POWER_INT2_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_POWER_INT2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_TOTAL_POWER_SLOPE attribute
typedef uint32_t CEN_MSS_TOTAL_POWER_SLOPE_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_TOTAL_POWER_SLOPE_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_TOTAL_POWER_SLOPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_TOTAL_POWER_SLOPE2 attribute
typedef uint32_t CEN_MSS_TOTAL_POWER_SLOPE2_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_TOTAL_POWER_SLOPE2_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_TOTAL_POWER_SLOPE2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_TOTAL_POWER_INT attribute
typedef uint32_t CEN_MSS_TOTAL_POWER_INT_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_TOTAL_POWER_INT_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_TOTAL_POWER_INT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_TOTAL_POWER_INT2 attribute
typedef uint32_t CEN_MSS_TOTAL_POWER_INT2_ATTR[2][2];
typedef uint32_t ATTR_CEN_MSS_TOTAL_POWER_INT2_type[2][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint32_t, 2>, 2> ATTR_CEN_MSS_TOTAL_POWER_INT2_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_CHANNEL_PAIR_MAXPOWER attribute
typedef uint32_t CEN_MSS_CHANNEL_PAIR_MAXPOWER_ATTR;
typedef uint32_t ATTR_CEN_MSS_CHANNEL_PAIR_MAXPOWER_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_MBA attribute
typedef uint32_t CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_MBA_ATTR;
typedef uint32_t ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_MBA_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_DENOMINATOR attribute
typedef uint32_t CEN_MSS_RUNTIME_MEM_THROTTLE_DENOMINATOR_ATTR;
typedef uint32_t ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_DENOMINATOR_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
typedef uint32_t CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_CHIP_ATTR;
typedef uint32_t ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_CHIP_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_MBA attribute
typedef uint32_t CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_MBA_ATTR;
typedef uint32_t ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_MBA_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
typedef uint32_t CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP_ATTR;
typedef uint32_t ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DERIVED_MBA_ADDR_INTERLEAVE_BIT attribute
typedef uint8_t CEN_MSS_DERIVED_MBA_ADDR_INTERLEAVE_BIT_ATTR;
typedef uint8_t ATTR_CEN_MSS_DERIVED_MBA_ADDR_INTERLEAVE_BIT_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DERIVED_MBA_CACHELINE_INTERLEAVE_MODE attribute
typedef uint8_t CEN_MSS_DERIVED_MBA_CACHELINE_INTERLEAVE_MODE_ATTR;
typedef uint8_t ATTR_CEN_MSS_DERIVED_MBA_CACHELINE_INTERLEAVE_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_MBA_CACHELINE_INTERLEAVE_MODE_CONTROL attribute
typedef uint8_t CEN_MRW_MBA_CACHELINE_INTERLEAVE_MODE_CONTROL_ATTR;
typedef uint8_t ATTR_CEN_MRW_MBA_CACHELINE_INTERLEAVE_MODE_CONTROL_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_CACHE_ENABLE attribute
typedef uint8_t CEN_MSS_CACHE_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_MSS_CACHE_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_PREFETCH_ENABLE attribute
typedef uint8_t CEN_MSS_PREFETCH_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_MSS_PREFETCH_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_CLEANER_ENABLE attribute
typedef uint8_t CEN_MSS_CLEANER_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_MSS_CLEANER_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_EFF_DIMM_FUNCTIONAL_VECTOR attribute
typedef uint8_t CEN_MSS_EFF_DIMM_FUNCTIONAL_VECTOR_ATTR;
typedef uint8_t ATTR_CEN_MSS_EFF_DIMM_FUNCTIONAL_VECTOR_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_LPASR attribute
typedef uint8_t CEN_EFF_DRAM_LPASR_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_LPASR_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_MPR_PAGE attribute
typedef uint8_t CEN_EFF_MPR_PAGE_ATTR;
typedef uint8_t ATTR_CEN_EFF_MPR_PAGE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_GEARDOWN_MODE attribute
typedef uint8_t CEN_EFF_GEARDOWN_MODE_ATTR;
typedef uint8_t ATTR_CEN_EFF_GEARDOWN_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_PER_DRAM_ACCESS attribute
typedef uint8_t CEN_EFF_PER_DRAM_ACCESS_ATTR;
typedef uint8_t ATTR_CEN_EFF_PER_DRAM_ACCESS_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_TEMP_READOUT attribute
typedef uint8_t CEN_EFF_TEMP_READOUT_ATTR;
typedef uint8_t ATTR_CEN_EFF_TEMP_READOUT_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_FINE_REFRESH_MODE attribute
typedef uint8_t CEN_EFF_FINE_REFRESH_MODE_ATTR;
typedef uint8_t ATTR_CEN_EFF_FINE_REFRESH_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_CRC_WR_LATENCY attribute
typedef uint8_t CEN_EFF_CRC_WR_LATENCY_ATTR;
typedef uint8_t ATTR_CEN_EFF_CRC_WR_LATENCY_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_MPR_RD_FORMAT attribute
typedef uint8_t CEN_EFF_MPR_RD_FORMAT_ATTR;
typedef uint8_t ATTR_CEN_EFF_MPR_RD_FORMAT_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_MAX_POWERDOWN_MODE attribute
typedef uint8_t CEN_EFF_MAX_POWERDOWN_MODE_ATTR;
typedef uint8_t ATTR_CEN_EFF_MAX_POWERDOWN_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_TEMP_REF_RANGE attribute
typedef uint8_t CEN_EFF_TEMP_REF_RANGE_ATTR;
typedef uint8_t ATTR_CEN_EFF_TEMP_REF_RANGE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_TEMP_REF_MODE attribute
typedef uint8_t CEN_EFF_TEMP_REF_MODE_ATTR;
typedef uint8_t ATTR_CEN_EFF_TEMP_REF_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_INT_VREF_MON attribute
typedef uint8_t CEN_EFF_INT_VREF_MON_ATTR;
typedef uint8_t ATTR_CEN_EFF_INT_VREF_MON_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_CS_CMD_LATENCY attribute
typedef uint8_t CEN_EFF_CS_CMD_LATENCY_ATTR;
typedef uint8_t ATTR_CEN_EFF_CS_CMD_LATENCY_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_SELF_REF_ABORT attribute
typedef uint8_t CEN_EFF_SELF_REF_ABORT_ATTR;
typedef uint8_t ATTR_CEN_EFF_SELF_REF_ABORT_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_RD_PREAMBLE_TRAIN attribute
typedef uint8_t CEN_EFF_RD_PREAMBLE_TRAIN_ATTR;
typedef uint8_t ATTR_CEN_EFF_RD_PREAMBLE_TRAIN_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_RD_PREAMBLE attribute
typedef uint8_t CEN_EFF_RD_PREAMBLE_ATTR;
typedef uint8_t ATTR_CEN_EFF_RD_PREAMBLE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_WR_PREAMBLE attribute
typedef uint8_t CEN_EFF_WR_PREAMBLE_ATTR;
typedef uint8_t ATTR_CEN_EFF_WR_PREAMBLE_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_CA_PARITY_LATENCY attribute
typedef uint8_t CEN_EFF_CA_PARITY_LATENCY_ATTR;
typedef uint8_t ATTR_CEN_EFF_CA_PARITY_LATENCY_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_CRC_ERROR_CLEAR attribute
typedef uint8_t CEN_EFF_CRC_ERROR_CLEAR_ATTR;
typedef uint8_t ATTR_CEN_EFF_CRC_ERROR_CLEAR_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_CA_PARITY_ERROR_STATUS attribute
typedef uint8_t CEN_EFF_CA_PARITY_ERROR_STATUS_ATTR;
typedef uint8_t ATTR_CEN_EFF_CA_PARITY_ERROR_STATUS_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_ODT_INPUT_BUFF attribute
typedef uint8_t CEN_EFF_ODT_INPUT_BUFF_ATTR;
typedef uint8_t ATTR_CEN_EFF_ODT_INPUT_BUFF_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_CA_PARITY attribute
typedef uint8_t CEN_EFF_CA_PARITY_ATTR;
typedef uint8_t ATTR_CEN_EFF_CA_PARITY_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_DATA_MASK attribute
typedef uint8_t CEN_EFF_DATA_MASK_ATTR;
typedef uint8_t ATTR_CEN_EFF_DATA_MASK_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_WRITE_DBI attribute
typedef uint8_t CEN_EFF_WRITE_DBI_ATTR;
typedef uint8_t ATTR_CEN_EFF_WRITE_DBI_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_READ_DBI attribute
typedef uint8_t CEN_EFF_READ_DBI_ATTR;
typedef uint8_t ATTR_CEN_EFF_READ_DBI_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_VREF_DQ_TRAIN_VALUE attribute
typedef uint8_t CEN_EFF_VREF_DQ_TRAIN_VALUE_ATTR[2][2][4];
typedef uint8_t ATTR_CEN_EFF_VREF_DQ_TRAIN_VALUE_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_CEN_EFF_VREF_DQ_TRAIN_VALUE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_VREF_DQ_TRAIN_RANGE attribute
typedef uint8_t CEN_EFF_VREF_DQ_TRAIN_RANGE_ATTR[2][2][4];
typedef uint8_t ATTR_CEN_EFF_VREF_DQ_TRAIN_RANGE_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_CEN_EFF_VREF_DQ_TRAIN_RANGE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_VREF_DQ_TRAIN_ENABLE attribute
typedef uint8_t CEN_EFF_VREF_DQ_TRAIN_ENABLE_ATTR[2][2][4];
typedef uint8_t ATTR_CEN_EFF_VREF_DQ_TRAIN_ENABLE_type[2][2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 2>, 2> ATTR_CEN_EFF_VREF_DQ_TRAIN_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_EFF_DRAM_TCCD_L attribute
typedef uint8_t CEN_EFF_DRAM_TCCD_L_ATTR;
typedef uint8_t ATTR_CEN_EFF_DRAM_TCCD_L_type;

// Type aliases and/or sizes for ATTR_CEN_TCCD_L attribute
typedef uint8_t CEN_TCCD_L_ATTR;
typedef uint8_t ATTR_CEN_TCCD_L_type;

// Type aliases and/or sizes for ATTR_CEN_EFF_WRITE_CRC attribute
typedef uint8_t CEN_EFF_WRITE_CRC_ATTR;
typedef uint8_t ATTR_CEN_EFF_WRITE_CRC_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_CAL_STEP_ENABLE attribute
typedef uint8_t CEN_MSS_CAL_STEP_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_MSS_CAL_STEP_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DRAMINIT_RESET_DISABLE attribute
typedef uint8_t CEN_MSS_DRAMINIT_RESET_DISABLE_ATTR;
typedef uint8_t ATTR_CEN_MSS_DRAMINIT_RESET_DISABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_SLEW_RATE_DATA attribute
typedef uint8_t CEN_MSS_SLEW_RATE_DATA_ATTR[2][4][4];
typedef uint8_t ATTR_CEN_MSS_SLEW_RATE_DATA_type[2][4][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> ATTR_CEN_MSS_SLEW_RATE_DATA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_SLEW_RATE_ADR attribute
typedef uint8_t CEN_MSS_SLEW_RATE_ADR_ATTR[2][4][4];
typedef uint8_t ATTR_CEN_MSS_SLEW_RATE_ADR_type[2][4][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 4>, 4>, 2> ATTR_CEN_MSS_SLEW_RATE_ADR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_CEN_MSS_ALLOW_SINGLE_PORT attribute
typedef uint8_t CEN_MSS_ALLOW_SINGLE_PORT_ATTR;
typedef uint8_t ATTR_CEN_MSS_ALLOW_SINGLE_PORT_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DQS_SWIZZLE_TYPE attribute
typedef uint8_t CEN_MSS_DQS_SWIZZLE_TYPE_ATTR;
typedef uint8_t ATTR_CEN_MSS_DQS_SWIZZLE_TYPE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_PSRO attribute
typedef uint8_t CEN_MSS_PSRO_ATTR;
typedef uint8_t ATTR_CEN_MSS_PSRO_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_NWELL_MISPLACEMENT attribute
typedef uint8_t CEN_MSS_NWELL_MISPLACEMENT_ATTR;
typedef uint8_t ATTR_CEN_MSS_NWELL_MISPLACEMENT_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_BLUEWATERFALL_BROKEN attribute
typedef uint8_t CEN_MSS_BLUEWATERFALL_BROKEN_ATTR;
typedef uint8_t ATTR_CEN_MSS_BLUEWATERFALL_BROKEN_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_PATTERN attribute
typedef uint32_t CEN_MCBIST_PATTERN_ATTR;
typedef uint32_t ATTR_CEN_MCBIST_PATTERN_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_TEST_TYPE attribute
typedef uint32_t CEN_MCBIST_TEST_TYPE_ATTR;
typedef uint32_t ATTR_CEN_MCBIST_TEST_TYPE_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_PRINTING_DISABLE attribute
typedef uint8_t CEN_MCBIST_PRINTING_DISABLE_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_PRINTING_DISABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_DATA_ENABLE attribute
typedef uint8_t CEN_MCBIST_DATA_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_DATA_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_USER_RANK attribute
typedef uint8_t CEN_MCBIST_USER_RANK_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_USER_RANK_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_USER_BANK attribute
typedef uint8_t CEN_MCBIST_USER_BANK_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_USER_BANK_type;

// Type aliases and/or sizes for ATTR_CEN_SCHMOO_MULTIPLE_SETUP_CALL attribute
typedef uint8_t CEN_SCHMOO_MULTIPLE_SETUP_CALL_ATTR;
typedef uint8_t ATTR_CEN_SCHMOO_MULTIPLE_SETUP_CALL_type;

// Type aliases and/or sizes for ATTR_CEN_LRDIMM_RANK_MULT_MODE attribute
typedef uint8_t CEN_LRDIMM_RANK_MULT_MODE_ATTR;
typedef uint8_t ATTR_CEN_LRDIMM_RANK_MULT_MODE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_THROTTLE_CONTROL_RAS_WEIGHT attribute
typedef uint8_t CEN_MSS_THROTTLE_CONTROL_RAS_WEIGHT_ATTR;
typedef uint8_t ATTR_CEN_MSS_THROTTLE_CONTROL_RAS_WEIGHT_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_THROTTLE_CONTROL_CAS_WEIGHT attribute
typedef uint8_t CEN_MSS_THROTTLE_CONTROL_CAS_WEIGHT_ATTR;
typedef uint8_t ATTR_CEN_MSS_THROTTLE_CONTROL_CAS_WEIGHT_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_RANDOM_SEED_VALUE attribute
typedef uint32_t CEN_MCBIST_RANDOM_SEED_VALUE_ATTR;
typedef uint32_t ATTR_CEN_MCBIST_RANDOM_SEED_VALUE_type;

// Type aliases and/or sizes for ATTR_CEN_MCBIST_RANDOM_SEED_TYPE attribute
typedef uint8_t CEN_MCBIST_RANDOM_SEED_TYPE_ATTR;
typedef uint8_t ATTR_CEN_MCBIST_RANDOM_SEED_TYPE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_INIT_STATE attribute
typedef uint8_t CEN_MSS_INIT_STATE_ATTR;
typedef uint8_t ATTR_CEN_MSS_INIT_STATE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_EFF_VPD_VERSION attribute
typedef uint32_t CEN_MSS_EFF_VPD_VERSION_ATTR;
typedef uint32_t ATTR_CEN_MSS_EFF_VPD_VERSION_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_NEST_CAPABLE_FREQUENCIES attribute
typedef uint8_t CEN_MSS_NEST_CAPABLE_FREQUENCIES_ATTR;
typedef uint8_t ATTR_CEN_MSS_NEST_CAPABLE_FREQUENCIES_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_STRICT_MBA_PLUG_RULE_CHECKING attribute
typedef uint8_t CEN_MRW_STRICT_MBA_PLUG_RULE_CHECKING_ATTR;
typedef uint8_t ATTR_CEN_MRW_STRICT_MBA_PLUG_RULE_CHECKING_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_CDIMM_MASTER_I2C_TEMP_SENSOR_ENABLE attribute
typedef uint8_t CEN_MRW_CDIMM_MASTER_I2C_TEMP_SENSOR_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_MRW_CDIMM_MASTER_I2C_TEMP_SENSOR_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_CDIMM_SPARE_I2C_TEMP_SENSOR_ENABLE attribute
typedef uint8_t CEN_MRW_CDIMM_SPARE_I2C_TEMP_SENSOR_ENABLE_ATTR;
typedef uint8_t ATTR_CEN_MRW_CDIMM_SPARE_I2C_TEMP_SENSOR_ENABLE_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_THERMAL_MEMORY_POWER_LIMIT attribute
typedef uint32_t CEN_MRW_THERMAL_MEMORY_POWER_LIMIT_ATTR;
typedef uint32_t ATTR_CEN_MRW_THERMAL_MEMORY_POWER_LIMIT_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_POWER_CONTROL_REQUESTED attribute
typedef uint8_t CEN_MRW_POWER_CONTROL_REQUESTED_ATTR;
typedef uint8_t ATTR_CEN_MRW_POWER_CONTROL_REQUESTED_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DDR3_VDDR_SLOPE attribute
typedef uint32_t CEN_MSS_DDR3_VDDR_SLOPE_ATTR;
typedef uint32_t ATTR_CEN_MSS_DDR3_VDDR_SLOPE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DDR3_VDDR_INTERCEPT attribute
typedef uint32_t CEN_MSS_DDR3_VDDR_INTERCEPT_ATTR;
typedef uint32_t ATTR_CEN_MSS_DDR3_VDDR_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DDR4_VDDR_SLOPE attribute
typedef uint32_t CEN_MSS_DDR4_VDDR_SLOPE_ATTR;
typedef uint32_t ATTR_CEN_MSS_DDR4_VDDR_SLOPE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_DDR4_VDDR_INTERCEPT attribute
typedef uint32_t CEN_MSS_DDR4_VDDR_INTERCEPT_ATTR;
typedef uint32_t ATTR_CEN_MSS_DDR4_VDDR_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_OVERRIDE attribute
typedef uint8_t CEN_MSS_VOLT_OVERRIDE_ATTR;
typedef uint8_t ATTR_CEN_MSS_VOLT_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_COMPLIANT_DIMMS attribute
typedef uint8_t CEN_MSS_VOLT_COMPLIANT_DIMMS_ATTR;
typedef uint8_t ATTR_CEN_MSS_VOLT_COMPLIANT_DIMMS_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_MCS_PREFETCH_RETRY_THRESHOLD attribute
typedef uint8_t CEN_MRW_MCS_PREFETCH_RETRY_THRESHOLD_ATTR;
typedef uint8_t ATTR_CEN_MRW_MCS_PREFETCH_RETRY_THRESHOLD_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_DDR3_VDDR_MAX_LIMIT attribute
typedef uint32_t CEN_MRW_DDR3_VDDR_MAX_LIMIT_ATTR;
typedef uint32_t ATTR_CEN_MRW_DDR3_VDDR_MAX_LIMIT_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_DDR4_VDDR_MAX_LIMIT attribute
typedef uint32_t CEN_MRW_DDR4_VDDR_MAX_LIMIT_ATTR;
typedef uint32_t ATTR_CEN_MRW_DDR4_VDDR_MAX_LIMIT_type;

// Type aliases and/or sizes for ATTR_MRW_MEM_SENSOR_CACHE_ADDR_MAP attribute
typedef uint32_t MRW_MEM_SENSOR_CACHE_ADDR_MAP_ATTR;
typedef uint32_t ATTR_MRW_MEM_SENSOR_CACHE_ADDR_MAP_type;

// Type aliases and/or sizes for ATTR_CEN_MRW_WR_VREF_CHECK_VREF_STEP_SIZE attribute
typedef uint8_t CEN_MRW_WR_VREF_CHECK_VREF_STEP_SIZE_ATTR;
typedef uint8_t ATTR_CEN_MRW_WR_VREF_CHECK_VREF_STEP_SIZE_type;

// Type aliases and/or sizes for ATTR_MEM_DRAM_CWL attribute
typedef uint8_t MEM_DRAM_CWL_ATTR;
typedef uint8_t ATTR_MEM_DRAM_CWL_type;

// Type aliases and/or sizes for ATTR_MEM_RDIMM_BUFFER_DELAY attribute
typedef uint8_t MEM_RDIMM_BUFFER_DELAY_ATTR;
typedef uint8_t ATTR_MEM_RDIMM_BUFFER_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_REORDER_QUEUE_SETTING attribute
typedef uint8_t MEM_REORDER_QUEUE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_REORDER_QUEUE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_2N_MODE attribute
typedef uint8_t MEM_2N_MODE_ATTR;
typedef uint8_t ATTR_MEM_2N_MODE_type;

// Type aliases and/or sizes for ATTR_MEM_VPD_DQ_MAP attribute
typedef uint8_t MEM_VPD_DQ_MAP_ATTR[72];
typedef uint8_t ATTR_MEM_VPD_DQ_MAP_type[72];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 72> ATTR_MEM_VPD_DQ_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC0F attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC0F_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC0F_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC0F_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_CS_CMD_LATENCY attribute
typedef uint8_t MEM_CS_CMD_LATENCY_ATTR[2];
typedef uint8_t ATTR_MEM_CS_CMD_LATENCY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_CS_CMD_LATENCY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_CA_PARITY_LATENCY attribute
typedef uint8_t MEM_CA_PARITY_LATENCY_ATTR[2];
typedef uint8_t ATTR_MEM_CA_PARITY_LATENCY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_CA_PARITY_LATENCY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC02 attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC02_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC02_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC02_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC03 attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC03_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC03_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC04 attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC04_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC04_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC05 attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC05_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC05_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC0B attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC0B_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC0B_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC0B_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC1X attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC1X_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC1X_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC1X_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F0RC7X attribute
typedef uint8_t MEM_DIMM_DDR4_F0RC7X_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F0RC7X_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F0RC7X_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F1RC00 attribute
typedef uint8_t MEM_DIMM_DDR4_F1RC00_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F1RC00_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F1RC00_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F1RC02 attribute
typedef uint8_t MEM_DIMM_DDR4_F1RC02_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F1RC02_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F1RC02_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F1RC03 attribute
typedef uint8_t MEM_DIMM_DDR4_F1RC03_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F1RC03_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F1RC03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F1RC04 attribute
typedef uint8_t MEM_DIMM_DDR4_F1RC04_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F1RC04_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F1RC04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_DDR4_F1RC05 attribute
typedef uint8_t MEM_DIMM_DDR4_F1RC05_ATTR[2];
typedef uint8_t ATTR_MEM_DIMM_DDR4_F1RC05_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_DIMM_DDR4_F1RC05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_DIMM_POS_METADATA attribute
typedef uint32_t MEM_DIMM_POS_METADATA_ATTR;
typedef uint32_t ATTR_MEM_DIMM_POS_METADATA_type;

// Type aliases and/or sizes for ATTR_MEM_DRAM_GEN_METADATA attribute
typedef uint8_t MEM_DRAM_GEN_METADATA_ATTR;
typedef uint8_t ATTR_MEM_DRAM_GEN_METADATA_type;

// Type aliases and/or sizes for ATTR_MEM_DIMM_TYPE_METADATA attribute
typedef uint8_t MEM_DIMM_TYPE_METADATA_ATTR;
typedef uint8_t ATTR_MEM_DIMM_TYPE_METADATA_type;

// Type aliases and/or sizes for ATTR_MSS_OMI_EDPL_DISABLE attribute
typedef uint8_t MSS_OMI_EDPL_DISABLE_ATTR;
typedef uint8_t ATTR_MSS_OMI_EDPL_DISABLE_type;

// Type aliases and/or sizes for ATTR_MEM_IGNORE_PLUG_RULES attribute
typedef uint8_t MEM_IGNORE_PLUG_RULES_ATTR;
typedef uint8_t ATTR_MEM_IGNORE_PLUG_RULES_type;

// Type aliases and/or sizes for ATTR_OMI_DL_PREIPL_PRBS_TIME attribute
typedef uint32_t OMI_DL_PREIPL_PRBS_TIME_ATTR;
typedef uint32_t ATTR_OMI_DL_PREIPL_PRBS_TIME_type;

// Type aliases and/or sizes for ATTR_MEM_SI_SIGNATURE_HASH attribute
typedef uint32_t MEM_SI_SIGNATURE_HASH_ATTR;
typedef uint32_t ATTR_MEM_SI_SIGNATURE_HASH_type;

// Type aliases and/or sizes for ATTR_MEM_SI_DIMM_RCD_IBT_CA attribute
typedef uint8_t MEM_SI_DIMM_RCD_IBT_CA_ATTR[2];
typedef uint8_t ATTR_MEM_SI_DIMM_RCD_IBT_CA_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_SI_DIMM_RCD_IBT_CA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DIMM_RCD_IBT_CKE attribute
typedef uint8_t MEM_SI_DIMM_RCD_IBT_CKE_ATTR[2];
typedef uint8_t ATTR_MEM_SI_DIMM_RCD_IBT_CKE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_SI_DIMM_RCD_IBT_CKE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DIMM_RCD_IBT_CS attribute
typedef uint8_t MEM_SI_DIMM_RCD_IBT_CS_ATTR[2];
typedef uint8_t ATTR_MEM_SI_DIMM_RCD_IBT_CS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_SI_DIMM_RCD_IBT_CS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DIMM_RCD_IBT_ODT attribute
typedef uint8_t MEM_SI_DIMM_RCD_IBT_ODT_ATTR[2];
typedef uint8_t ATTR_MEM_SI_DIMM_RCD_IBT_ODT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_SI_DIMM_RCD_IBT_ODT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DRAM_DRV_IMP_DQ_DQS attribute
typedef uint8_t MEM_SI_DRAM_DRV_IMP_DQ_DQS_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_DRAM_DRV_IMP_DQ_DQS_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_DRAM_DRV_IMP_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DRAM_PREAMBLE attribute
typedef uint8_t MEM_SI_DRAM_PREAMBLE_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_DRAM_PREAMBLE_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_DRAM_PREAMBLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DRAM_RTT_NOM attribute
typedef uint8_t MEM_SI_DRAM_RTT_NOM_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_DRAM_RTT_NOM_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_DRAM_RTT_NOM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DRAM_RTT_PARK attribute
typedef uint8_t MEM_SI_DRAM_RTT_PARK_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_DRAM_RTT_PARK_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_DRAM_RTT_PARK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_DRAM_RTT_WR attribute
typedef uint8_t MEM_SI_DRAM_RTT_WR_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_DRAM_RTT_WR_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_DRAM_RTT_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_GEARDOWN_MODE attribute
typedef uint8_t MEM_SI_GEARDOWN_MODE_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_GEARDOWN_MODE_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_GEARDOWN_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_DQ_DQS attribute
typedef uint8_t MEM_SI_MC_DRV_DQ_DQS_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_DQ_DQS_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_RCV_EQ_DQ_DQS attribute
typedef uint8_t MEM_SI_MC_RCV_EQ_DQ_DQS_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_RCV_EQ_DQ_DQS_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_RCV_EQ_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_EQ_DQ_DQS attribute
typedef uint8_t MEM_SI_MC_DRV_EQ_DQ_DQS_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_EQ_DQ_DQS_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_EQ_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_IMP_CLK attribute
typedef uint8_t MEM_SI_MC_DRV_IMP_CLK_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_IMP_CLK_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_IMP_CLK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_IMP_CMD_ADDR attribute
typedef uint8_t MEM_SI_MC_DRV_IMP_CMD_ADDR_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_IMP_CMD_ADDR_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_IMP_CMD_ADDR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_IMP_CNTL attribute
typedef uint8_t MEM_SI_MC_DRV_IMP_CNTL_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_IMP_CNTL_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_IMP_CNTL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_IMP_CSCID attribute
typedef uint8_t MEM_SI_MC_DRV_IMP_CSCID_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_IMP_CSCID_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_IMP_CSCID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN attribute
typedef uint16_t MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN_ATTR[2][4];
typedef uint16_t ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 4>, 2> ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP attribute
typedef uint16_t MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP_ATTR[2][4];
typedef uint16_t ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 4>, 2> ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CLK attribute
typedef uint8_t MEM_SI_MC_DRV_SLEW_RATE_CLK_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_SLEW_RATE_CLK_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_SLEW_RATE_CLK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR attribute
typedef uint8_t MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CNTL attribute
typedef uint8_t MEM_SI_MC_DRV_SLEW_RATE_CNTL_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_SLEW_RATE_CNTL_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_SLEW_RATE_CNTL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CSCID attribute
typedef uint8_t MEM_SI_MC_DRV_SLEW_RATE_CSCID_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_SLEW_RATE_CSCID_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_SLEW_RATE_CSCID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS attribute
typedef uint8_t MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_RCV_IMP_ALERT_N attribute
typedef uint8_t MEM_SI_MC_RCV_IMP_ALERT_N_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_RCV_IMP_ALERT_N_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_RCV_IMP_ALERT_N_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_MC_RCV_IMP_DQ_DQS attribute
typedef uint8_t MEM_SI_MC_RCV_IMP_DQ_DQS_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_MC_RCV_IMP_DQ_DQS_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_MC_RCV_IMP_DQ_DQS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_ODT_RD attribute
typedef uint8_t MEM_SI_ODT_RD_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_ODT_RD_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_ODT_RD_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_ODT_WR attribute
typedef uint8_t MEM_SI_ODT_WR_ATTR[2][4];
typedef uint8_t ATTR_MEM_SI_ODT_WR_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_SI_ODT_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_SI_VREF_DRAM_WR attribute
typedef uint8_t MEM_SI_VREF_DRAM_WR_ATTR;
typedef uint8_t ATTR_MEM_SI_VREF_DRAM_WR_type;

// Type aliases and/or sizes for ATTR_MEM_SI_VREF_MC_RD attribute
typedef uint32_t MEM_SI_VREF_MC_RD_ATTR;
typedef uint32_t ATTR_MEM_SI_VREF_MC_RD_type;

// Type aliases and/or sizes for ATTR_MEM_SI_WINDAGE_RD_CTR attribute
typedef int16_t MEM_SI_WINDAGE_RD_CTR_ATTR;
typedef int16_t ATTR_MEM_SI_WINDAGE_RD_CTR_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_SUPPORTED_FREQ attribute
typedef uint32_t MSS_MRW_SUPPORTED_FREQ_ATTR[4];
typedef uint32_t ATTR_MSS_MRW_SUPPORTED_FREQ_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 4> ATTR_MSS_MRW_SUPPORTED_FREQ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_MRW_IS_PLANAR attribute
typedef uint8_t MEM_MRW_IS_PLANAR_ATTR;
typedef uint8_t ATTR_MEM_MRW_IS_PLANAR_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
typedef uint16_t MSS_MRW_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT_ATTR;
typedef uint16_t ATTR_MSS_MRW_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_THERMAL_MEMORY_POWER_LIMIT attribute
typedef uint64_t MSS_MRW_THERMAL_MEMORY_POWER_LIMIT_ATTR[10];
typedef uint64_t ATTR_MSS_MRW_THERMAL_MEMORY_POWER_LIMIT_type[10];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 10> ATTR_MSS_MRW_THERMAL_MEMORY_POWER_LIMIT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MRW_PWR_INTERCEPT attribute
typedef uint64_t MSS_MRW_PWR_INTERCEPT_ATTR[100];
typedef uint64_t ATTR_MSS_MRW_PWR_INTERCEPT_type[100];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 100> ATTR_MSS_MRW_PWR_INTERCEPT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MRW_PWR_SLOPE attribute
typedef uint64_t MSS_MRW_PWR_SLOPE_ATTR[100];
typedef uint64_t ATTR_MSS_MRW_PWR_SLOPE_type[100];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 100> ATTR_MSS_MRW_PWR_SLOPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MRW_REFRESH_RATE_REQUEST attribute
typedef uint8_t MSS_MRW_REFRESH_RATE_REQUEST_ATTR;
typedef uint8_t ATTR_MSS_MRW_REFRESH_RATE_REQUEST_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT attribute
typedef uint8_t MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_ATTR;
typedef uint8_t ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_IDLE attribute
typedef uint8_t MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_IDLE_ATTR;
typedef uint8_t ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_IDLE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_MEM_M_DRAM_CLOCKS attribute
typedef uint32_t MSS_MRW_MEM_M_DRAM_CLOCKS_ATTR;
typedef uint32_t ATTR_MSS_MRW_MEM_M_DRAM_CLOCKS_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_MAX_DRAM_DATABUS_UTIL attribute
typedef uint32_t MSS_MRW_MAX_DRAM_DATABUS_UTIL_ATTR;
typedef uint32_t ATTR_MSS_MRW_MAX_DRAM_DATABUS_UTIL_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_POWER_CONTROL_REQUESTED attribute
typedef uint8_t MSS_MRW_POWER_CONTROL_REQUESTED_ATTR;
typedef uint8_t ATTR_MSS_MRW_POWER_CONTROL_REQUESTED_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_IDLE_POWER_CONTROL_REQUESTED attribute
typedef uint8_t MSS_MRW_IDLE_POWER_CONTROL_REQUESTED_ATTR;
typedef uint8_t ATTR_MSS_MRW_IDLE_POWER_CONTROL_REQUESTED_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_VMEM_REGULATOR_POWER_LIMIT_PER_DIMM_ADJ_ENABLE attribute
typedef uint8_t MSS_MRW_VMEM_REGULATOR_POWER_LIMIT_PER_DIMM_ADJ_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_MRW_VMEM_REGULATOR_POWER_LIMIT_PER_DIMM_ADJ_ENABLE_type;

// Type aliases and/or sizes for ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR3 attribute
typedef uint32_t MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR3_ATTR;
typedef uint32_t ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR3_type;

// Type aliases and/or sizes for ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR4 attribute
typedef uint32_t MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR4_ATTR;
typedef uint32_t ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR4_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_MAX_NUMBER_DIMMS_POSSIBLE_PER_VMEM_REGULATOR attribute
typedef uint8_t MSS_MRW_MAX_NUMBER_DIMMS_POSSIBLE_PER_VMEM_REGULATOR_ATTR;
typedef uint8_t ATTR_MSS_MRW_MAX_NUMBER_DIMMS_POSSIBLE_PER_VMEM_REGULATOR_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_AVDD_OFFSET_ENABLE attribute
typedef uint8_t MSS_MRW_AVDD_OFFSET_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_MRW_AVDD_OFFSET_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_VDD_OFFSET_ENABLE attribute
typedef uint8_t MSS_MRW_VDD_OFFSET_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_MRW_VDD_OFFSET_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_VCS_OFFSET_ENABLE attribute
typedef uint8_t MSS_MRW_VCS_OFFSET_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_MRW_VCS_OFFSET_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_VPP_OFFSET_ENABLE attribute
typedef uint8_t MSS_MRW_VPP_OFFSET_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_MRW_VPP_OFFSET_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_VDDR_OFFSET_ENABLE attribute
typedef uint8_t MSS_MRW_VDDR_OFFSET_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_MRW_VDDR_OFFSET_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_FINE_REFRESH_MODE attribute
typedef uint8_t MSS_MRW_FINE_REFRESH_MODE_ATTR;
typedef uint8_t ATTR_MSS_MRW_FINE_REFRESH_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_TEMP_REFRESH_RANGE attribute
typedef uint8_t MSS_MRW_TEMP_REFRESH_RANGE_ATTR;
typedef uint8_t ATTR_MSS_MRW_TEMP_REFRESH_RANGE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_RESET_DELAY_BEFORE_CAL attribute
typedef uint8_t MSS_MRW_RESET_DELAY_BEFORE_CAL_ATTR;
typedef uint8_t ATTR_MSS_MRW_RESET_DELAY_BEFORE_CAL_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_PERIODIC_MEMCAL_MODE_OPTIONS attribute
typedef uint16_t MSS_MRW_PERIODIC_MEMCAL_MODE_OPTIONS_ATTR;
typedef uint16_t ATTR_MSS_MRW_PERIODIC_MEMCAL_MODE_OPTIONS_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_PERIODIC_ZQCAL_MODE_OPTIONS attribute
typedef uint16_t MSS_MRW_PERIODIC_ZQCAL_MODE_OPTIONS_ATTR;
typedef uint16_t ATTR_MSS_MRW_PERIODIC_ZQCAL_MODE_OPTIONS_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_DRAM_2N_MODE attribute
typedef uint8_t MSS_MRW_DRAM_2N_MODE_ATTR;
typedef uint8_t ATTR_MSS_MRW_DRAM_2N_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_DRAM_WRITE_CRC attribute
typedef uint8_t MSS_MRW_DRAM_WRITE_CRC_ATTR;
typedef uint8_t ATTR_MSS_MRW_DRAM_WRITE_CRC_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_TEMP_REFRESH_MODE attribute
typedef uint8_t MSS_MRW_TEMP_REFRESH_MODE_ATTR;
typedef uint8_t ATTR_MSS_MRW_TEMP_REFRESH_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_FORCE_BCMODE_OFF attribute
typedef uint8_t MSS_MRW_FORCE_BCMODE_OFF_ATTR;
typedef uint8_t ATTR_MSS_MRW_FORCE_BCMODE_OFF_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_NVDIMM_PLUG_RULES attribute
typedef uint64_t MSS_MRW_NVDIMM_PLUG_RULES_ATTR;
typedef uint64_t ATTR_MSS_MRW_NVDIMM_PLUG_RULES_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_ALLOW_UNSUPPORTED_RCW attribute
typedef uint8_t MSS_MRW_ALLOW_UNSUPPORTED_RCW_ATTR;
typedef uint8_t ATTR_MSS_MRW_ALLOW_UNSUPPORTED_RCW_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_SUPPORTED_DRAM_WIDTH attribute
typedef uint8_t MSS_MRW_SUPPORTED_DRAM_WIDTH_ATTR;
typedef uint8_t ATTR_MSS_MRW_SUPPORTED_DRAM_WIDTH_type;

// Type aliases and/or sizes for ATTR_MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT attribute
typedef uint64_t MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT_ATTR[25];
typedef uint64_t ATTR_MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT_type[25];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 25> ATTR_MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MRW_OCMB_PWR_SLOPE attribute
typedef uint64_t MSS_MRW_OCMB_PWR_SLOPE_ATTR[50];
typedef uint64_t ATTR_MSS_MRW_OCMB_PWR_SLOPE_type[50];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 50> ATTR_MSS_MRW_OCMB_PWR_SLOPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MRW_OCMB_PWR_INTERCEPT attribute
typedef uint64_t MSS_MRW_OCMB_PWR_INTERCEPT_ATTR[50];
typedef uint64_t ATTR_MSS_MRW_OCMB_PWR_INTERCEPT_type[50];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 50> ATTR_MSS_MRW_OCMB_PWR_INTERCEPT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT attribute
typedef uint64_t MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT_ATTR[25];
typedef uint64_t ATTR_MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT_type[25];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 25> ATTR_MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_MRW_SAFEMODE_DRAM_DATABUS_UTIL attribute
typedef uint32_t MSS_MRW_SAFEMODE_DRAM_DATABUS_UTIL_ATTR;
typedef uint32_t ATTR_MSS_MRW_SAFEMODE_DRAM_DATABUS_UTIL_type;

// Type aliases and/or sizes for ATTR_MEM_PORT_POS_OF_FAIL_THROTTLE attribute
typedef uint64_t MEM_PORT_POS_OF_FAIL_THROTTLE_ATTR;
typedef uint64_t ATTR_MEM_PORT_POS_OF_FAIL_THROTTLE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_GEN attribute
typedef uint8_t MEM_EFF_DRAM_GEN_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_GEN_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_GEN_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DIMM_TYPE attribute
typedef uint8_t MEM_EFF_DIMM_TYPE_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DIMM_TYPE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DIMM_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_HYBRID_MEMORY_TYPE attribute
typedef uint8_t MEM_EFF_HYBRID_MEMORY_TYPE_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_HYBRID_MEMORY_TYPE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_HYBRID_MEMORY_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_HYBRID attribute
typedef uint8_t MEM_EFF_HYBRID_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_HYBRID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_HYBRID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_HOST_TO_DDR_SPEED_RATIO attribute
typedef uint8_t MEM_EFF_HOST_TO_DDR_SPEED_RATIO_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_HOST_TO_DDR_SPEED_RATIO_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_HOST_TO_DDR_SPEED_RATIO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_DENSITY attribute
typedef uint8_t MEM_EFF_DRAM_DENSITY_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_DENSITY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_DENSITY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_BANK_BITS attribute
typedef uint8_t MEM_EFF_DRAM_BANK_BITS_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_BANK_BITS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_BANK_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_BANK_GROUP_BITS attribute
typedef uint8_t MEM_EFF_DRAM_BANK_GROUP_BITS_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_BANK_GROUP_BITS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_BANK_GROUP_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_COLUMN_BITS attribute
typedef uint8_t MEM_EFF_DRAM_COLUMN_BITS_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_COLUMN_BITS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_COLUMN_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_ROW_BITS attribute
typedef uint8_t MEM_EFF_DRAM_ROW_BITS_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_ROW_BITS_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_ROW_BITS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_PRIM_DIE_COUNT attribute
typedef uint8_t MEM_EFF_PRIM_DIE_COUNT_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_PRIM_DIE_COUNT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_PRIM_DIE_COUNT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_PRIM_STACK_TYPE attribute
typedef uint8_t MEM_EFF_PRIM_STACK_TYPE_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_PRIM_STACK_TYPE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_PRIM_STACK_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_PRIM_BUS_WIDTH attribute
typedef uint8_t MEM_EFF_PRIM_BUS_WIDTH_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_PRIM_BUS_WIDTH_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_PRIM_BUS_WIDTH_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_PPR attribute
typedef uint8_t MEM_EFF_DRAM_PPR_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_PPR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_PPR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_SOFT_PPR attribute
typedef uint8_t MEM_EFF_DRAM_SOFT_PPR_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_SOFT_PPR_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_SOFT_PPR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRCD attribute
typedef uint8_t MEM_EFF_DRAM_TRCD_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRCD_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRP attribute
typedef uint8_t MEM_EFF_DRAM_TRP_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRP_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRAS attribute
typedef uint8_t MEM_EFF_DRAM_TRAS_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRAS_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRC attribute
typedef uint8_t MEM_EFF_DRAM_TRC_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRC_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRFC attribute
typedef uint16_t MEM_EFF_DRAM_TRFC_ATTR;
typedef uint16_t ATTR_MEM_EFF_DRAM_TRFC_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TFAW attribute
typedef uint8_t MEM_EFF_DRAM_TFAW_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TFAW_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRRD_S attribute
typedef uint8_t MEM_EFF_DRAM_TRRD_S_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRRD_S_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRRD_L attribute
typedef uint8_t MEM_EFF_DRAM_TRRD_L_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRRD_L_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRRD_DLR attribute
typedef uint8_t MEM_EFF_DRAM_TRRD_DLR_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRRD_DLR_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TCCD_L attribute
typedef uint8_t MEM_EFF_DRAM_TCCD_L_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TCCD_L_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TWR attribute
typedef uint8_t MEM_EFF_DRAM_TWR_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TWR_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TWTR_S attribute
typedef uint8_t MEM_EFF_DRAM_TWTR_S_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TWTR_S_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TWTR_L attribute
typedef uint8_t MEM_EFF_DRAM_TWTR_L_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TWTR_L_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TMAW attribute
typedef uint16_t MEM_EFF_DRAM_TMAW_ATTR;
typedef uint16_t ATTR_MEM_EFF_DRAM_TMAW_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_WIDTH attribute
typedef uint8_t MEM_EFF_DRAM_WIDTH_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DRAM_WIDTH_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DRAM_WIDTH_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_LOGICAL_RANKS_PER_DIMM attribute
typedef uint8_t MEM_EFF_LOGICAL_RANKS_PER_DIMM_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_LOGICAL_RANKS_PER_DIMM_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_LOGICAL_RANKS_PER_DIMM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_3DS_HEIGHT attribute
typedef uint16_t MEM_3DS_HEIGHT_ATTR[2];
typedef uint16_t ATTR_MEM_3DS_HEIGHT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MEM_3DS_HEIGHT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_REGISTER_TYPE attribute
typedef uint8_t MEM_EFF_REGISTER_TYPE_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_REGISTER_TYPE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_REGISTER_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_MODULE_MFG_ID attribute
typedef uint16_t MEM_EFF_MODULE_MFG_ID_ATTR[2];
typedef uint16_t ATTR_MEM_EFF_MODULE_MFG_ID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MEM_EFF_MODULE_MFG_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_MFG_ID attribute
typedef uint16_t MEM_EFF_DRAM_MFG_ID_ATTR[2];
typedef uint16_t ATTR_MEM_EFF_DRAM_MFG_ID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MEM_EFF_DRAM_MFG_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_MODULE_HEIGHT attribute
typedef uint8_t MEM_EFF_DRAM_MODULE_HEIGHT_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_MODULE_HEIGHT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_RCD_MFG_ID attribute
typedef uint16_t MEM_EFF_RCD_MFG_ID_ATTR[2];
typedef uint16_t ATTR_MEM_EFF_RCD_MFG_ID_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MEM_EFF_RCD_MFG_ID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_REGISTER_REV attribute
typedef uint8_t MEM_EFF_REGISTER_REV_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_REGISTER_REV_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_REGISTER_REV_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_PACKAGE_RANK_MAP attribute
typedef uint8_t MEM_EFF_PACKAGE_RANK_MAP_ATTR[2][20];
typedef uint8_t ATTR_MEM_EFF_PACKAGE_RANK_MAP_type[2][20];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 20>, 2> ATTR_MEM_EFF_PACKAGE_RANK_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_NIBBLE_MAP attribute
typedef uint8_t MEM_EFF_NIBBLE_MAP_ATTR[2][20];
typedef uint8_t ATTR_MEM_EFF_NIBBLE_MAP_type[2][20];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 20>, 2> ATTR_MEM_EFF_NIBBLE_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DIMM_SIZE attribute
typedef uint32_t MEM_EFF_DIMM_SIZE_ATTR[2];
typedef uint32_t ATTR_MEM_EFF_DIMM_SIZE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MEM_EFF_DIMM_SIZE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DIMM_SPARE attribute
typedef uint8_t MEM_EFF_DIMM_SPARE_ATTR[2][4];
typedef uint8_t ATTR_MEM_EFF_DIMM_SPARE_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_EFF_DIMM_SPARE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_CL attribute
typedef uint8_t MEM_EFF_DRAM_CL_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_CL_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_MDS attribute
typedef uint8_t MEM_EFF_DRAM_MDS_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_MDS_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_NUM_MASTER_RANKS_PER_DIMM attribute
typedef uint8_t MEM_EFF_NUM_MASTER_RANKS_PER_DIMM_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_NUM_MASTER_RANKS_PER_DIMM_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_NUM_MASTER_RANKS_PER_DIMM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DIMM_RANKS_CONFIGED attribute
typedef uint8_t MEM_EFF_DIMM_RANKS_CONFIGED_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DIMM_RANKS_CONFIGED_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DIMM_RANKS_CONFIGED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TREFI attribute
typedef uint16_t MEM_EFF_DRAM_TREFI_ATTR;
typedef uint16_t ATTR_MEM_EFF_DRAM_TREFI_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRTP attribute
typedef uint8_t MEM_EFF_DRAM_TRTP_ATTR;
typedef uint8_t ATTR_MEM_EFF_DRAM_TRTP_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_DRAM_TRFC_DLR attribute
typedef uint16_t MEM_EFF_DRAM_TRFC_DLR_ATTR;
typedef uint16_t ATTR_MEM_EFF_DRAM_TRFC_DLR_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_FREQ attribute
typedef uint64_t MEM_EFF_FREQ_ATTR;
typedef uint64_t ATTR_MEM_EFF_FREQ_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_VOLT_VDDR attribute
typedef uint32_t MEM_EFF_VOLT_VDDR_ATTR;
typedef uint32_t ATTR_MEM_EFF_VOLT_VDDR_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_VOLT_VPP attribute
typedef uint32_t MEM_EFF_VOLT_VPP_ATTR;
typedef uint32_t ATTR_MEM_EFF_VOLT_VPP_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_SPD_REVISION attribute
typedef uint8_t MEM_EFF_SPD_REVISION_ATTR;
typedef uint8_t ATTR_MEM_EFF_SPD_REVISION_type;

// Type aliases and/or sizes for ATTR_EXPLR_ENABLE_US_TMPL_1 attribute
typedef uint8_t EXPLR_ENABLE_US_TMPL_1_ATTR;
typedef uint8_t ATTR_EXPLR_ENABLE_US_TMPL_1_type;

// Type aliases and/or sizes for ATTR_EXPLR_ENABLE_US_TMPL_5 attribute
typedef uint8_t EXPLR_ENABLE_US_TMPL_5_ATTR;
typedef uint8_t ATTR_EXPLR_ENABLE_US_TMPL_5_type;

// Type aliases and/or sizes for ATTR_EXPLR_ENABLE_US_TMPL_9 attribute
typedef uint8_t EXPLR_ENABLE_US_TMPL_9_ATTR;
typedef uint8_t ATTR_EXPLR_ENABLE_US_TMPL_9_type;

// Type aliases and/or sizes for ATTR_EXPLR_ENABLE_US_TMPL_A attribute
typedef uint8_t EXPLR_ENABLE_US_TMPL_A_ATTR;
typedef uint8_t ATTR_EXPLR_ENABLE_US_TMPL_A_type;

// Type aliases and/or sizes for ATTR_EXPLR_ENABLE_US_TMPL_B attribute
typedef uint8_t EXPLR_ENABLE_US_TMPL_B_ATTR;
typedef uint8_t ATTR_EXPLR_ENABLE_US_TMPL_B_type;

// Type aliases and/or sizes for ATTR_EXPLR_TMPL_0_PACING attribute
typedef uint8_t EXPLR_TMPL_0_PACING_ATTR;
typedef uint8_t ATTR_EXPLR_TMPL_0_PACING_type;

// Type aliases and/or sizes for ATTR_EXPLR_TMPL_1_PACING attribute
typedef uint8_t EXPLR_TMPL_1_PACING_ATTR;
typedef uint8_t ATTR_EXPLR_TMPL_1_PACING_type;

// Type aliases and/or sizes for ATTR_EXPLR_TMPL_5_PACING attribute
typedef uint8_t EXPLR_TMPL_5_PACING_ATTR;
typedef uint8_t ATTR_EXPLR_TMPL_5_PACING_type;

// Type aliases and/or sizes for ATTR_EXPLR_TMPL_9_PACING attribute
typedef uint8_t EXPLR_TMPL_9_PACING_ATTR;
typedef uint8_t ATTR_EXPLR_TMPL_9_PACING_type;

// Type aliases and/or sizes for ATTR_EXPLR_TMPL_B_PACING attribute
typedef uint8_t EXPLR_TMPL_B_PACING_ATTR;
typedef uint8_t ATTR_EXPLR_TMPL_B_PACING_type;

// Type aliases and/or sizes for ATTR_EXPLR_SHRT_BACKOFF_TIMER attribute
typedef uint8_t EXPLR_SHRT_BACKOFF_TIMER_ATTR;
typedef uint8_t ATTR_EXPLR_SHRT_BACKOFF_TIMER_type;

// Type aliases and/or sizes for ATTR_EXPLR_METADATA_ENABLE attribute
typedef uint8_t EXPLR_METADATA_ENABLE_ATTR;
typedef uint8_t ATTR_EXPLR_METADATA_ENABLE_type;

// Type aliases and/or sizes for ATTR_EXPLR_PASID_BASE attribute
typedef uint32_t EXPLR_PASID_BASE_ATTR;
typedef uint32_t ATTR_EXPLR_PASID_BASE_type;

// Type aliases and/or sizes for ATTR_EXPLR_ACTAG_BASE attribute
typedef uint32_t EXPLR_ACTAG_BASE_ATTR;
typedef uint32_t ATTR_EXPLR_ACTAG_BASE_type;

// Type aliases and/or sizes for ATTR_EXPLR_AFU_ACTAG_LEN attribute
typedef uint32_t EXPLR_AFU_ACTAG_LEN_ATTR;
typedef uint32_t ATTR_EXPLR_AFU_ACTAG_LEN_type;

// Type aliases and/or sizes for ATTR_EXPLR_PASID_LEN attribute
typedef uint8_t EXPLR_PASID_LEN_ATTR;
typedef uint8_t ATTR_EXPLR_PASID_LEN_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PSTATES attribute
typedef uint8_t MEM_EFF_PSTATES_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_PSTATES_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_PSTATES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_BYTE_ENABLES attribute
typedef uint16_t MEM_EFF_BYTE_ENABLES_ATTR[2];
typedef uint16_t ATTR_MEM_EFF_BYTE_ENABLES_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_MEM_EFF_BYTE_ENABLES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_NIBBLE_ENABLES attribute
typedef uint32_t MEM_EFF_NIBBLE_ENABLES_ATTR[2];
typedef uint32_t ATTR_MEM_EFF_NIBBLE_ENABLES_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_MEM_EFF_NIBBLE_ENABLES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_FOUR_RANK_MODE attribute
typedef uint8_t MEM_EFF_FOUR_RANK_MODE_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_FOUR_RANK_MODE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_FOUR_RANK_MODE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_MRAM_SUPPORT attribute
typedef uint8_t MEM_EFF_MRAM_SUPPORT_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_MRAM_SUPPORT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_MRAM_SUPPORT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_DDP_COMPATIBILITY attribute
typedef uint8_t MEM_EFF_DDP_COMPATIBILITY_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_DDP_COMPATIBILITY_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_DDP_COMPATIBILITY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_TSV_8H_SUPPORT attribute
typedef uint8_t MEM_EFF_TSV_8H_SUPPORT_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_TSV_8H_SUPPORT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_TSV_8H_SUPPORT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_MODULE_THERMAL_SENSORS attribute
typedef uint8_t MEM_EFF_MODULE_THERMAL_SENSORS_ATTR;
typedef uint8_t ATTR_MEM_EFF_MODULE_THERMAL_SENSORS_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_0_AVAIL attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_0_AVAIL_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_0_AVAIL_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_0_TYPE attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_0_TYPE_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_0_TYPE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_0_USAGE attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_0_USAGE_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_0_USAGE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_0_I2C_ADDR attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_0_I2C_ADDR_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_0_I2C_ADDR_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_1_AVAIL attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_1_AVAIL_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_1_AVAIL_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_1_TYPE attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_1_TYPE_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_1_TYPE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_1_USAGE attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_1_USAGE_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_1_USAGE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_1_I2C_ADDR attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_1_I2C_ADDR_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_1_I2C_ADDR_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_DIFF_AVAIL attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_DIFF_AVAIL_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_DIFF_AVAIL_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_DIFF_TYPE attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_DIFF_TYPE_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_DIFF_TYPE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_DIFF_USAGE attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_DIFF_USAGE_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_DIFF_USAGE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_THERM_SENSOR_DIFF_I2C_ADDR attribute
typedef uint8_t MEM_EFF_THERM_SENSOR_DIFF_I2C_ADDR_ATTR;
typedef uint8_t ATTR_MEM_EFF_THERM_SENSOR_DIFF_I2C_ADDR_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_SUPPORTED_RCD attribute
typedef uint8_t MEM_EFF_SUPPORTED_RCD_ATTR[2];
typedef uint8_t ATTR_MEM_EFF_SUPPORTED_RCD_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EFF_SUPPORTED_RCD_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_OCMB_ENTERPRISE_MODE attribute
typedef uint8_t MSS_OCMB_ENTERPRISE_MODE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_ENTERPRISE_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_ENTERPRISE_POLICY attribute
typedef uint8_t MSS_OCMB_ENTERPRISE_POLICY_ATTR;
typedef uint8_t ATTR_MSS_OCMB_ENTERPRISE_POLICY_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_NONENTERPRISE_MODE_OVERRIDE attribute
typedef uint8_t MSS_OCMB_NONENTERPRISE_MODE_OVERRIDE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_NONENTERPRISE_MODE_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_HALF_DIMM_MODE attribute
typedef uint8_t MSS_OCMB_HALF_DIMM_MODE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_HALF_DIMM_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_HALF_DIMM_MODE_OVERRIDE attribute
typedef uint8_t MSS_OCMB_HALF_DIMM_MODE_OVERRIDE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_HALF_DIMM_MODE_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_MSS_MEM_MVPD_FWMS attribute
typedef uint32_t MSS_MEM_MVPD_FWMS_ATTR[8];
typedef uint32_t ATTR_MSS_MEM_MVPD_FWMS_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 8> ATTR_MSS_MEM_MVPD_FWMS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EXP_INIT_VREF_DQ attribute
typedef uint8_t MEM_EXP_INIT_VREF_DQ_ATTR[2][4];
typedef uint8_t ATTR_MEM_EXP_INIT_VREF_DQ_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_EXP_INIT_VREF_DQ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EXP_INIT_PHY_VREF attribute
typedef uint8_t MEM_EXP_INIT_PHY_VREF_ATTR[2][4];
typedef uint8_t ATTR_MEM_EXP_INIT_PHY_VREF_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MEM_EXP_INIT_PHY_VREF_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EXP_RCD_DIC attribute
typedef uint16_t MEM_EXP_RCD_DIC_ATTR;
typedef uint16_t ATTR_MEM_EXP_RCD_DIC_type;

// Type aliases and/or sizes for ATTR_MEM_EXP_RCD_VOLTAGE_CTRL attribute
typedef uint16_t MEM_EXP_RCD_VOLTAGE_CTRL_ATTR;
typedef uint16_t ATTR_MEM_EXP_RCD_VOLTAGE_CTRL_type;

// Type aliases and/or sizes for ATTR_MEM_EXP_DRAM_ADDRESS_MIRRORING attribute
typedef uint8_t MEM_EXP_DRAM_ADDRESS_MIRRORING_ATTR[2];
typedef uint8_t ATTR_MEM_EXP_DRAM_ADDRESS_MIRRORING_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MEM_EXP_DRAM_ADDRESS_MIRRORING_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EXP_RCD_SLEW_RATE attribute
typedef uint16_t MEM_EXP_RCD_SLEW_RATE_ATTR;
typedef uint16_t ATTR_MEM_EXP_RCD_SLEW_RATE_type;

// Type aliases and/or sizes for ATTR_MEM_EXP_SPD_CL_SUPPORTED attribute
typedef uint32_t MEM_EXP_SPD_CL_SUPPORTED_ATTR;
typedef uint32_t ATTR_MEM_EXP_SPD_CL_SUPPORTED_type;

// Type aliases and/or sizes for ATTR_MEM_EXP_SPD_TAA_MIN attribute
typedef uint16_t MEM_EXP_SPD_TAA_MIN_ATTR;
typedef uint16_t ATTR_MEM_EXP_SPD_TAA_MIN_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_REORDER_QUEUE_SETTING attribute
typedef uint8_t MSS_EXP_REORDER_QUEUE_SETTING_ATTR;
typedef uint8_t ATTR_MSS_EXP_REORDER_QUEUE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EXP_FIRMWARE_EMULATION_MODE attribute
typedef uint8_t MEM_EXP_FIRMWARE_EMULATION_MODE_ATTR;
typedef uint8_t ATTR_MEM_EXP_FIRMWARE_EMULATION_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_STRUCT_MMIO_ENDIAN_CTRL attribute
typedef uint8_t MSS_OCMB_EXP_STRUCT_MMIO_ENDIAN_CTRL_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_STRUCT_MMIO_ENDIAN_CTRL_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_STRUCT_ENDIAN attribute
typedef uint8_t MSS_OCMB_EXP_STRUCT_ENDIAN_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_STRUCT_ENDIAN_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_STRUCT_MMIO_WORD_SWAP attribute
typedef uint8_t MSS_OCMB_EXP_STRUCT_MMIO_WORD_SWAP_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_STRUCT_MMIO_WORD_SWAP_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_OMI_CFG_ENDIAN_CTRL attribute
typedef uint8_t MSS_OCMB_EXP_OMI_CFG_ENDIAN_CTRL_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_OMI_CFG_ENDIAN_CTRL_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_ECID attribute
typedef uint16_t MSS_OCMB_ECID_ATTR[14];
typedef uint16_t ATTR_MSS_OCMB_ECID_type[14];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 14> ATTR_MSS_OCMB_ECID_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EXP_DFIMRL_CLK attribute
typedef uint8_t MEM_EXP_DFIMRL_CLK_ATTR;
typedef uint8_t ATTR_MEM_EXP_DFIMRL_CLK_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_ATXDLY_A attribute
typedef uint8_t MEM_EFF_ATXDLY_A_ATTR[8];
typedef uint8_t ATTR_MEM_EFF_ATXDLY_A_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 8> ATTR_MEM_EFF_ATXDLY_A_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEM_EFF_ATXDLY_B attribute
typedef uint8_t MEM_EFF_ATXDLY_B_ATTR[8];
typedef uint8_t ATTR_MEM_EFF_ATXDLY_B_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 8> ATTR_MEM_EFF_ATXDLY_B_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_OCMB_PHY_INIT_MODE attribute
typedef uint8_t MSS_OCMB_PHY_INIT_MODE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_PHY_INIT_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_CHECKSTOP_OBJ_HANDLE attribute
typedef uint64_t MSS_OCMB_CHECKSTOP_OBJ_HANDLE_ATTR;
typedef uint64_t ATTR_MSS_OCMB_CHECKSTOP_OBJ_HANDLE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_RECOV_OBJ_HANDLE attribute
typedef uint64_t MSS_OCMB_RECOV_OBJ_HANDLE_ATTR;
typedef uint64_t ATTR_MSS_OCMB_RECOV_OBJ_HANDLE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_SPECATTN_OBJ_HANDLE attribute
typedef uint64_t MSS_OCMB_SPECATTN_OBJ_HANDLE_ATTR;
typedef uint64_t ATTR_MSS_OCMB_SPECATTN_OBJ_HANDLE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_APPINTR_OBJ_HANDLE attribute
typedef uint64_t MSS_OCMB_APPINTR_OBJ_HANDLE_ATTR;
typedef uint64_t ATTR_MSS_OCMB_APPINTR_OBJ_HANDLE_type;

// Type aliases and/or sizes for ATTR_MSS_CHECK_FOR_READY_TIMEOUT attribute
typedef uint16_t MSS_CHECK_FOR_READY_TIMEOUT_ATTR;
typedef uint16_t ATTR_MSS_CHECK_FOR_READY_TIMEOUT_type;

// Type aliases and/or sizes for ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST attribute
typedef uint8_t MSS_POST_MEMDIAGS_READ_SUBTEST_ATTR;
typedef uint8_t ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_type;

// Type aliases and/or sizes for ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_FAIL_BEHAVIOR attribute
typedef uint8_t MSS_POST_MEMDIAGS_READ_SUBTEST_FAIL_BEHAVIOR_ATTR;
typedef uint8_t ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_FAIL_BEHAVIOR_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_DISABLE_THERM_INIT_READ attribute
typedef uint8_t MSS_OCMB_DISABLE_THERM_INIT_READ_ATTR;
typedef uint8_t ATTR_MSS_OCMB_DISABLE_THERM_INIT_READ_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC00 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC00_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC00_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC00_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC01 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC01_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC01_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC01_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC03 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC03_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC03_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC04 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC04_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC04_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC05 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC05_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC05_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC0B attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC0B_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC0B_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC0B_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC0E attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC0E_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC0E_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC0E_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC0F attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC0F_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC0F_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC0F_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC1X attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC1X_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC1X_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC1X_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F0RC7X attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F0RC7X_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F0RC7X_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F0RC7X_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F1RC00 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F1RC00_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F1RC00_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F1RC00_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F1RC02 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F1RC02_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F1RC02_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F1RC02_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F1RC03 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F1RC03_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F1RC03_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F1RC03_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F1RC04 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F1RC04_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F1RC04_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F1RC04_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DDR4_F1RC05 attribute
typedef uint8_t MSS_EXP_RESP_DDR4_F1RC05_ATTR[2];
typedef uint8_t ATTR_MSS_EXP_RESP_DDR4_F1RC05_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_EXP_RESP_DDR4_F1RC05_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_RBT attribute
typedef uint8_t MSS_EXP_RESP_DRAM_RBT_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_RBT_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_TM attribute
typedef uint8_t MSS_EXP_RESP_DRAM_TM_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_TM_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_DLL_RESET attribute
typedef uint8_t MSS_EXP_RESP_DRAM_DLL_RESET_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_DLL_RESET_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_BURST_LENGTH attribute
typedef uint8_t MSS_EXP_RESP_DRAM_BURST_LENGTH_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_BURST_LENGTH_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_DLL_ENABLE attribute
typedef uint8_t MSS_EXP_RESP_DRAM_DLL_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_DLL_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_ODIC attribute
typedef uint8_t MSS_EXP_RESP_DRAM_ODIC_ATTR[2][4];
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_ODIC_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MSS_EXP_RESP_DRAM_ODIC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_AL attribute
typedef uint8_t MSS_EXP_RESP_DRAM_AL_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_AL_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_WR_LVL_ENABLE attribute
typedef uint8_t MSS_EXP_RESP_DRAM_WR_LVL_ENABLE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_WR_LVL_ENABLE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_RTT_NOM attribute
typedef uint8_t MSS_EXP_RESP_DRAM_RTT_NOM_ATTR[2][4];
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_RTT_NOM_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MSS_EXP_RESP_DRAM_RTT_NOM_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_RTT_PARK attribute
typedef uint8_t MSS_EXP_RESP_DRAM_RTT_PARK_ATTR[2][4];
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_RTT_PARK_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MSS_EXP_RESP_DRAM_RTT_PARK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_TDQS attribute
typedef uint8_t MSS_EXP_RESP_DRAM_TDQS_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_TDQS_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_OUTPUT_BUFFER attribute
typedef uint8_t MSS_EXP_RESP_DRAM_OUTPUT_BUFFER_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_OUTPUT_BUFFER_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_LPASR attribute
typedef uint8_t MSS_EXP_RESP_DRAM_LPASR_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_LPASR_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DRAM_RTT_WR attribute
typedef uint8_t MSS_EXP_RESP_DRAM_RTT_WR_ATTR[2][4];
typedef uint8_t ATTR_MSS_EXP_RESP_DRAM_RTT_WR_type[2][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 2> ATTR_MSS_EXP_RESP_DRAM_RTT_WR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_MPR_PAGE attribute
typedef uint8_t MSS_EXP_RESP_MPR_PAGE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_MPR_PAGE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_MPR_MODE attribute
typedef uint8_t MSS_EXP_RESP_MPR_MODE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_MPR_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_GEARDOWN_MODE attribute
typedef uint8_t MSS_EXP_RESP_GEARDOWN_MODE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_GEARDOWN_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_PER_DRAM_ACCESS attribute
typedef uint8_t MSS_EXP_RESP_PER_DRAM_ACCESS_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_PER_DRAM_ACCESS_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_TEMP_READOUT attribute
typedef uint8_t MSS_EXP_RESP_TEMP_READOUT_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_TEMP_READOUT_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_CRC_WR_LATENCY attribute
typedef uint8_t MSS_EXP_RESP_CRC_WR_LATENCY_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_CRC_WR_LATENCY_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_MPR_RD_FORMAT attribute
typedef uint8_t MSS_EXP_RESP_MPR_RD_FORMAT_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_MPR_RD_FORMAT_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_MAX_POWERDOWN_MODE attribute
typedef uint8_t MSS_EXP_RESP_MAX_POWERDOWN_MODE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_MAX_POWERDOWN_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_INTERNAL_VREF_MONITOR attribute
typedef uint8_t MSS_EXP_RESP_INTERNAL_VREF_MONITOR_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_INTERNAL_VREF_MONITOR_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_SELF_REF_ABORT attribute
typedef uint8_t MSS_EXP_RESP_SELF_REF_ABORT_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_SELF_REF_ABORT_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_RD_PREAMBLE_TRAIN attribute
typedef uint8_t MSS_EXP_RESP_RD_PREAMBLE_TRAIN_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_RD_PREAMBLE_TRAIN_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_RD_PREAMBLE attribute
typedef uint8_t MSS_EXP_RESP_RD_PREAMBLE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_RD_PREAMBLE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_WR_PREAMBLE attribute
typedef uint8_t MSS_EXP_RESP_WR_PREAMBLE_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_WR_PREAMBLE_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_CRC_ERROR_CLEAR attribute
typedef uint8_t MSS_EXP_RESP_CRC_ERROR_CLEAR_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_CRC_ERROR_CLEAR_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_CA_PARITY_ERROR_STATUS attribute
typedef uint8_t MSS_EXP_RESP_CA_PARITY_ERROR_STATUS_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_CA_PARITY_ERROR_STATUS_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_ODT_INPUT_BUFF attribute
typedef uint8_t MSS_EXP_RESP_ODT_INPUT_BUFF_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_ODT_INPUT_BUFF_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_CA_PARITY attribute
typedef uint8_t MSS_EXP_RESP_CA_PARITY_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_CA_PARITY_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_DATA_MASK attribute
typedef uint8_t MSS_EXP_RESP_DATA_MASK_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_DATA_MASK_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_WRITE_DBI attribute
typedef uint8_t MSS_EXP_RESP_WRITE_DBI_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_WRITE_DBI_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_READ_DBI attribute
typedef uint8_t MSS_EXP_RESP_READ_DBI_ATTR;
typedef uint8_t ATTR_MSS_EXP_RESP_READ_DBI_type;

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE attribute
typedef uint8_t MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE_ATTR[2][4][20];
typedef uint8_t ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE_type[2][4][20];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 20>, 4>, 2> ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE attribute
typedef uint8_t MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE_ATTR[2][4][20];
typedef uint8_t ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE_type[2][4][20];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 20>, 4>, 2> ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE attribute
typedef uint8_t MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE_ATTR[2][4][20];
typedef uint8_t ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE_type[2][4][20];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 20>, 4>, 2> ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
typedef uint16_t EXP_MEM_THROTTLED_N_COMMANDS_PER_PORT_ATTR;
typedef uint16_t ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_PORT_type;

// Type aliases and/or sizes for ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
typedef uint16_t EXP_MEM_THROTTLED_N_COMMANDS_PER_SLOT_ATTR;
typedef uint16_t ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type;

// Type aliases and/or sizes for ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
typedef uint16_t EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT_ATTR;
typedef uint16_t ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT_type;

// Type aliases and/or sizes for ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
typedef uint16_t EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT_ATTR;
typedef uint16_t ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type;

// Type aliases and/or sizes for ATTR_EXP_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
typedef uint16_t EXP_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT_ATTR;
typedef uint16_t ATTR_EXP_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT_type;

// Type aliases and/or sizes for ATTR_EXP_RUNTIME_MEM_M_DRAM_CLOCKS attribute
typedef uint32_t EXP_RUNTIME_MEM_M_DRAM_CLOCKS_ATTR;
typedef uint32_t ATTR_EXP_RUNTIME_MEM_M_DRAM_CLOCKS_type;

// Type aliases and/or sizes for ATTR_EXP_MEM_PORT_POS_OF_FAIL_THROTTLE attribute
typedef uint64_t EXP_MEM_PORT_POS_OF_FAIL_THROTTLE_ATTR;
typedef uint64_t ATTR_EXP_MEM_PORT_POS_OF_FAIL_THROTTLE_type;

// Type aliases and/or sizes for ATTR_EXP_MEM_WATT_TARGET attribute
typedef uint32_t EXP_MEM_WATT_TARGET_ATTR[2];
typedef uint32_t ATTR_EXP_MEM_WATT_TARGET_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_EXP_MEM_WATT_TARGET_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EXP_TOTAL_PWR_SLOPE attribute
typedef uint16_t EXP_TOTAL_PWR_SLOPE_ATTR[2];
typedef uint16_t ATTR_EXP_TOTAL_PWR_SLOPE_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_EXP_TOTAL_PWR_SLOPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EXP_TOTAL_PWR_INTERCEPT attribute
typedef uint16_t EXP_TOTAL_PWR_INTERCEPT_ATTR[2];
typedef uint16_t ATTR_EXP_TOTAL_PWR_INTERCEPT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_EXP_TOTAL_PWR_INTERCEPT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_EXP_DATABUS_UTIL attribute
typedef uint32_t EXP_DATABUS_UTIL_ATTR;
typedef uint32_t ATTR_EXP_DATABUS_UTIL_type;

// Type aliases and/or sizes for ATTR_EXP_PORT_MAXPOWER attribute
typedef uint32_t EXP_PORT_MAXPOWER_ATTR;
typedef uint32_t ATTR_EXP_PORT_MAXPOWER_type;

// Type aliases and/or sizes for ATTR_EXP_DIMM_THERMAL_LIMIT attribute
typedef uint32_t EXP_DIMM_THERMAL_LIMIT_ATTR[2];
typedef uint32_t ATTR_EXP_DIMM_THERMAL_LIMIT_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_EXP_DIMM_THERMAL_LIMIT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_FW_MODE attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_FW_MODE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_FW_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_OPENCAPI_LOOPBACK_TEST attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_OPENCAPI_LOOPBACK_TEST_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_OPENCAPI_LOOPBACK_TEST_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_TRANSPORT_LAYER attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_TRANSPORT_LAYER_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_TRANSPORT_LAYER_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DL_LAYER_BOOT_MODE attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_DL_LAYER_BOOT_MODE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DL_LAYER_BOOT_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DFE_DISABLE attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_DFE_DISABLE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DFE_DISABLE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_LANE_MODE attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_LANE_MODE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_LANE_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_SERDES_FREQUENCY attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_SERDES_FREQUENCY_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_SERDES_FREQUENCY_type;

// Type aliases and/or sizes for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_ADAPTATION_MODE attribute
typedef uint8_t MSS_OCMB_EXP_BOOT_CONFIG_ADAPTATION_MODE_ATTR;
typedef uint8_t ATTR_MSS_OCMB_EXP_BOOT_CONFIG_ADAPTATION_MODE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_MFG_ID attribute
typedef uint16_t MEM_EFF_PMIC0_MFG_ID_ATTR;
typedef uint16_t ATTR_MEM_EFF_PMIC0_MFG_ID_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC0_SWA_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC0_SWA_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC0_SWA_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC0_SWA_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC0_SWA_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC0_SWB_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC0_SWB_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC0_SWB_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC0_SWB_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC0_SWB_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC0_SWC_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC0_SWC_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC0_SWC_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC0_SWC_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC0_SWC_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC0_SWD_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC0_SWD_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC0_SWD_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC0_SWD_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC0_SWD_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_MFG_ID attribute
typedef uint16_t MEM_EFF_PMIC1_MFG_ID_ATTR;
typedef uint16_t ATTR_MEM_EFF_PMIC1_MFG_ID_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC1_SWA_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC1_SWA_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC1_SWA_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC1_SWA_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC1_SWA_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC1_SWB_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC1_SWB_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC1_SWB_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC1_SWB_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC1_SWB_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC1_SWC_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC1_SWC_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC1_SWC_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC1_SWC_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC1_SWC_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_SETTING attribute
typedef uint8_t MEM_EFF_PMIC1_SWD_VOLTAGE_SETTING_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_SETTING_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_RANGE_SELECT attribute
typedef uint8_t MEM_EFF_PMIC1_SWD_VOLTAGE_RANGE_SELECT_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_RANGE_SELECT_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_PMIC1_SWD_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_DELAY attribute
typedef uint8_t MEM_EFF_PMIC1_SWD_SEQUENCE_DELAY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_DELAY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_ORDER attribute
typedef uint8_t MEM_EFF_PMIC1_SWD_SEQUENCE_ORDER_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_ORDER_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_REDUNDANCY attribute
typedef uint8_t MEM_EFF_PMIC0_REDUNDANCY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_REDUNDANCY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_PHASE_COMB attribute
typedef uint8_t MEM_EFF_PMIC0_PHASE_COMB_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_PHASE_COMB_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_REDUNDANCY attribute
typedef uint8_t MEM_EFF_PMIC1_REDUNDANCY_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_REDUNDANCY_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_REVISION attribute
typedef uint8_t MEM_EFF_PMIC0_REVISION_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_REVISION_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_REVISION attribute
typedef uint8_t MEM_EFF_PMIC1_REVISION_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_REVISION_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_PHASE_COMB attribute
typedef uint8_t MEM_EFF_PMIC1_PHASE_COMB_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_PHASE_COMB_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC0_SEQUENCE attribute
typedef uint8_t MEM_EFF_PMIC0_SEQUENCE_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC0_SEQUENCE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_PMIC1_SEQUENCE attribute
typedef uint8_t MEM_EFF_PMIC1_SEQUENCE_ATTR;
typedef uint8_t ATTR_MEM_EFF_PMIC1_SEQUENCE_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC0_SWA_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC0_SWA_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC0_SWA_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC0_SWB_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC0_SWB_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC0_SWB_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC0_SWC_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC0_SWC_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC0_SWC_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC0_SWD_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC0_SWD_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC0_SWD_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC1_SWA_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC1_SWA_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC1_SWA_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC1_SWB_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC1_SWB_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC1_SWB_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC1_SWC_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC1_SWC_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC1_SWC_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_MEM_EFF_EFD_PMIC1_SWD_VOLTAGE_OFFSET attribute
typedef int8_t MEM_EFF_EFD_PMIC1_SWD_VOLTAGE_OFFSET_ATTR;
typedef int8_t ATTR_MEM_EFF_EFD_PMIC1_SWD_VOLTAGE_OFFSET_type;

// Type aliases and/or sizes for ATTR_AFFINITY_PATH attribute
typedef EntityPath AFFINITY_PATH_ATTR;
typedef EntityPath ATTR_AFFINITY_PATH_type;

// Type aliases and/or sizes for ATTR_ALL_MCS_IN_INTERLEAVING_GROUP attribute
typedef uint8_t ALL_MCS_IN_INTERLEAVING_GROUP_ATTR;
typedef uint8_t ATTR_ALL_MCS_IN_INTERLEAVING_GROUP_type;

// Type aliases and/or sizes for ATTR_ALTFSI_MASTER_CHIP attribute
typedef EntityPath ALTFSI_MASTER_CHIP_ATTR;
typedef EntityPath ATTR_ALTFSI_MASTER_CHIP_type;

// Type aliases and/or sizes for ATTR_ALTFSI_MASTER_PORT attribute
typedef uint8_t ALTFSI_MASTER_PORT_ATTR;
typedef uint8_t ATTR_ALTFSI_MASTER_PORT_type;

// Type aliases and/or sizes for ATTR_ASYNC_NEST_FREQ_MHZ attribute
typedef uint32_t ASYNC_NEST_FREQ_MHZ_ATTR;
typedef uint32_t ATTR_ASYNC_NEST_FREQ_MHZ_type;

// Type aliases and/or sizes for ATTR_ATTN_AREA_1_ADDR attribute
typedef uint64_t ATTN_AREA_1_ADDR_ATTR;
typedef uint64_t ATTR_ATTN_AREA_1_ADDR_type;

// Type aliases and/or sizes for ATTR_ATTN_AREA_2_ADDR attribute
typedef uint64_t ATTN_AREA_2_ADDR_ATTR;
typedef uint64_t ATTR_ATTN_AREA_2_ADDR_type;

// Type aliases and/or sizes for ATTR_AVDD_ID attribute
typedef uint16_t AVDD_ID_ATTR;
typedef uint16_t ATTR_AVDD_ID_type;

// Type aliases and/or sizes for ATTR_BLOCK_SPEC_DECONFIG attribute
typedef uint8_t BLOCK_SPEC_DECONFIG_ATTR;
typedef uint8_t ATTR_BLOCK_SPEC_DECONFIG_type;

// Type aliases and/or sizes for ATTR_BMC_FRU_ID attribute
typedef uint32_t BMC_FRU_ID_ATTR;
typedef uint32_t ATTR_BMC_FRU_ID_type;

// Type aliases and/or sizes for ATTR_BPM_UPDATE_OVERRIDE attribute
typedef uint16_t BPM_UPDATE_OVERRIDE_ATTR;
typedef uint16_t ATTR_BPM_UPDATE_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_BRAZOS_RX_FIFO_OVERRIDE attribute
typedef uint8_t BRAZOS_RX_FIFO_OVERRIDE_ATTR;
typedef uint8_t ATTR_BRAZOS_RX_FIFO_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_CDM_DOMAIN attribute
typedef CDM_DOMAIN CDM_DOMAIN_ATTR;
typedef CDM_DOMAIN ATTR_CDM_DOMAIN_type;

// Type aliases and/or sizes for ATTR_CDM_POLICIES attribute
typedef uint8_t CDM_POLICIES_ATTR;
typedef uint8_t ATTR_CDM_POLICIES_type;

// Type aliases and/or sizes for ATTR_CEC_IPL_TYPE attribute
typedef CecIplType CEC_IPL_TYPE_ATTR;
typedef CecIplType ATTR_CEC_IPL_TYPE_type;

// Type aliases and/or sizes for ATTR_CENTAUR_ECID_FRU_ID attribute
typedef uint32_t CENTAUR_ECID_FRU_ID_ATTR;
typedef uint32_t ATTR_CENTAUR_ECID_FRU_ID_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_AVDD_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_AVDD_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_AVDD_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VCS_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VCS_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VCS_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VCS_OFFSET_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VCS_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VCS_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VDDR_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VDDR_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VDDR_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VDD_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VDD_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VDD_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VDD_OFFSET_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VDD_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VDD_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VPP_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VPP_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VPP_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CEN_MSS_VOLT_VPP_OFFSET_MILLIVOLTS attribute
typedef uint32_t CEN_MSS_VOLT_VPP_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_CEN_MSS_VOLT_VPP_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_CHIPLET_ID attribute
typedef uint8_t CHIPLET_ID_ATTR;
typedef uint8_t ATTR_CHIPLET_ID_type;

// Type aliases and/or sizes for ATTR_CHIP_ID attribute
typedef uint32_t CHIP_ID_ATTR;
typedef uint32_t ATTR_CHIP_ID_type;

// Type aliases and/or sizes for ATTR_CHIP_UNIT attribute
typedef uint8_t CHIP_UNIT_ATTR;
typedef uint8_t ATTR_CHIP_UNIT_type;

// Type aliases and/or sizes for ATTR_CLASS attribute
typedef CLASS CLASS_ATTR;
typedef CLASS ATTR_CLASS_type;

// Type aliases and/or sizes for ATTR_CPU_ATTR attribute
typedef uint32_t CPU_ATTR_ATTR;
typedef uint32_t ATTR_CPU_ATTR_type;

// Type aliases and/or sizes for ATTR_CVPD_MAX_SECTIONS attribute
typedef uint32_t CVPD_MAX_SECTIONS_ATTR;
typedef uint32_t ATTR_CVPD_MAX_SECTIONS_type;

// Type aliases and/or sizes for ATTR_CVPD_SIZE attribute
typedef uint64_t CVPD_SIZE_ATTR;
typedef uint64_t ATTR_CVPD_SIZE_type;

// Type aliases and/or sizes for ATTR_DATA_CACHE_LINE_SIZE attribute
typedef uint32_t DATA_CACHE_LINE_SIZE_ATTR;
typedef uint32_t ATTR_DATA_CACHE_LINE_SIZE_type;

// Type aliases and/or sizes for ATTR_DATA_CACHE_SIZE attribute
typedef uint32_t DATA_CACHE_SIZE_ATTR;
typedef uint32_t ATTR_DATA_CACHE_SIZE_type;

// Type aliases and/or sizes for ATTR_DCACHE_ASSOC_SETS attribute
typedef uint32_t DCACHE_ASSOC_SETS_ATTR;
typedef uint32_t ATTR_DCACHE_ASSOC_SETS_type;

// Type aliases and/or sizes for ATTR_DCACHE_LINE_SIZE attribute
typedef uint32_t DCACHE_LINE_SIZE_ATTR;
typedef uint32_t ATTR_DCACHE_LINE_SIZE_type;

// Type aliases and/or sizes for ATTR_DECONFIG_GARDABLE attribute
typedef uint8_t DECONFIG_GARDABLE_ATTR;
typedef uint8_t ATTR_DECONFIG_GARDABLE_type;

// Type aliases and/or sizes for ATTR_DISABLE_I2C_ENGINE2_PORT0_DIAG_MODE attribute
typedef uint8_t DISABLE_I2C_ENGINE2_PORT0_DIAG_MODE_ATTR;
typedef uint8_t ATTR_DISABLE_I2C_ENGINE2_PORT0_DIAG_MODE_type;

// Type aliases and/or sizes for ATTR_DO_ABUS_DECONFIG attribute
typedef uint8_t DO_ABUS_DECONFIG_ATTR;
typedef uint8_t ATTR_DO_ABUS_DECONFIG_type;

// Type aliases and/or sizes for ATTR_DPO_MIN_FREQ_PERCENT attribute
typedef int32_t DPO_MIN_FREQ_PERCENT_ATTR;
typedef int32_t ATTR_DPO_MIN_FREQ_PERCENT_type;

// Type aliases and/or sizes for ATTR_DUMMY_HEAP_ZERO_DEFAULT attribute
typedef uint8_t DUMMY_HEAP_ZERO_DEFAULT_ATTR;
typedef uint8_t ATTR_DUMMY_HEAP_ZERO_DEFAULT_type;

// Type aliases and/or sizes for ATTR_DUMMY_RO attribute
typedef uint8_t DUMMY_RO_ATTR;
typedef uint8_t ATTR_DUMMY_RO_type;

// Type aliases and/or sizes for ATTR_DUMMY_RW attribute
typedef uint8_t DUMMY_RW_ATTR[1][3][5];
typedef uint8_t ATTR_DUMMY_RW_type[1][3][5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<std::array<uint8_t, 5>, 3>, 1> ATTR_DUMMY_RW_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_DUMMY_WO attribute
typedef uint8_t DUMMY_WO_ATTR;
typedef uint8_t ATTR_DUMMY_WO_type;

// Type aliases and/or sizes for ATTR_EC attribute
typedef uint8_t EC_ATTR;
typedef uint8_t ATTR_EC_type;

// Type aliases and/or sizes for ATTR_EEPROM_CONTENT_TYPE attribute
typedef EEPROM_CONTENT_TYPE EEPROM_CONTENT_TYPE_ATTR;
typedef EEPROM_CONTENT_TYPE ATTR_EEPROM_CONTENT_TYPE_type;

// Type aliases and/or sizes for ATTR_EEPROM_NV_INFO attribute
typedef EepromNvInfo EEPROM_NV_INFO_ATTR;
typedef EepromNvInfo ATTR_EEPROM_NV_INFO_type;

// Type aliases and/or sizes for ATTR_EEPROM_SBE_BACKUP_INFO attribute
typedef EepromSbeBackupInfo EEPROM_SBE_BACKUP_INFO_ATTR;
typedef EepromSbeBackupInfo ATTR_EEPROM_SBE_BACKUP_INFO_type;

// Type aliases and/or sizes for ATTR_EEPROM_SBE_PRIMARY_INFO attribute
typedef EepromSbePrimaryInfo EEPROM_SBE_PRIMARY_INFO_ATTR;
typedef EepromSbePrimaryInfo ATTR_EEPROM_SBE_PRIMARY_INFO_type;

// Type aliases and/or sizes for ATTR_EEPROM_VPD_BACKUP_INFO attribute
typedef EepromVpdBackupInfo EEPROM_VPD_BACKUP_INFO_ATTR;
typedef EepromVpdBackupInfo ATTR_EEPROM_VPD_BACKUP_INFO_type;

// Type aliases and/or sizes for ATTR_EEPROM_VPD_PRIMARY_INFO attribute
typedef EepromVpdPrimaryInfo EEPROM_VPD_PRIMARY_INFO_ATTR;
typedef EepromVpdPrimaryInfo ATTR_EEPROM_VPD_PRIMARY_INFO_type;

// Type aliases and/or sizes for ATTR_EFFECTIVE_EC attribute
typedef uint8_t EFFECTIVE_EC_ATTR;
typedef uint8_t ATTR_EFFECTIVE_EC_type;

// Type aliases and/or sizes for ATTR_ENABLED_THREADS attribute
typedef uint64_t ENABLED_THREADS_ATTR;
typedef uint64_t ATTR_ENABLED_THREADS_type;

// Type aliases and/or sizes for ATTR_ENGINE_TYPE attribute
typedef ENGINE_TYPE ENGINE_TYPE_ATTR;
typedef ENGINE_TYPE ATTR_ENGINE_TYPE_type;

// Type aliases and/or sizes for ATTR_FABRIC_CHIP_ID attribute
typedef uint8_t FABRIC_CHIP_ID_ATTR;
typedef uint8_t ATTR_FABRIC_CHIP_ID_type;

// Type aliases and/or sizes for ATTR_FABRIC_GROUP_ID attribute
typedef uint8_t FABRIC_GROUP_ID_ATTR;
typedef uint8_t ATTR_FABRIC_GROUP_ID_type;

// Type aliases and/or sizes for ATTR_FABRIC_TO_PHYSICAL_NODE_MAP attribute
typedef uint8_t FABRIC_TO_PHYSICAL_NODE_MAP_ATTR[8];
typedef uint8_t ATTR_FABRIC_TO_PHYSICAL_NODE_MAP_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 8> ATTR_FABRIC_TO_PHYSICAL_NODE_MAP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_FAPI_I2C_CONTROL_INFO attribute
typedef FapiI2cControlInfo FAPI_I2C_CONTROL_INFO_ATTR;
typedef FapiI2cControlInfo ATTR_FAPI_I2C_CONTROL_INFO_type;

// Type aliases and/or sizes for ATTR_FAPI_NAME attribute
typedef char FAPI_NAME_ATTR[64];
typedef char ATTR_FAPI_NAME_type[64];
const size_t ATTR_FAPI_NAME_max_chars = 63;

// Type aliases and/or sizes for ATTR_FIELD_CORE_OVERRIDE attribute
typedef uint32_t FIELD_CORE_OVERRIDE_ATTR;
typedef uint32_t ATTR_FIELD_CORE_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_FIELD_TH_P8EX_L2_COL_REPAIRS attribute
typedef uint8_t FIELD_TH_P8EX_L2_COL_REPAIRS_ATTR;
typedef uint8_t ATTR_FIELD_TH_P8EX_L2_COL_REPAIRS_type;

// Type aliases and/or sizes for ATTR_FIELD_TH_P8EX_L2_LINE_DELETES attribute
typedef uint8_t FIELD_TH_P8EX_L2_LINE_DELETES_ATTR;
typedef uint8_t ATTR_FIELD_TH_P8EX_L2_LINE_DELETES_type;

// Type aliases and/or sizes for ATTR_FIELD_TH_P8EX_L3_COL_REPAIRS attribute
typedef uint8_t FIELD_TH_P8EX_L3_COL_REPAIRS_ATTR;
typedef uint8_t ATTR_FIELD_TH_P8EX_L3_COL_REPAIRS_type;

// Type aliases and/or sizes for ATTR_FIELD_TH_P8EX_L3_LINE_DELETES attribute
typedef uint8_t FIELD_TH_P8EX_L3_LINE_DELETES_ATTR;
typedef uint8_t ATTR_FIELD_TH_P8EX_L3_LINE_DELETES_type;

// Type aliases and/or sizes for ATTR_FORCE_NVDIMM_RESET attribute
typedef uint8_t FORCE_NVDIMM_RESET_ATTR;
typedef uint8_t ATTR_FORCE_NVDIMM_RESET_type;

// Type aliases and/or sizes for ATTR_FORCE_SRAM_MMIO_OVER_I2C attribute
typedef uint8_t FORCE_SRAM_MMIO_OVER_I2C_ATTR;
typedef uint8_t ATTR_FORCE_SRAM_MMIO_OVER_I2C_type;

// Type aliases and/or sizes for ATTR_FREQ_CORE_CEILING_MHZ attribute
typedef uint32_t FREQ_CORE_CEILING_MHZ_ATTR;
typedef uint32_t ATTR_FREQ_CORE_CEILING_MHZ_type;

// Type aliases and/or sizes for ATTR_FREQ_CORE_MAX attribute
typedef uint32_t FREQ_CORE_MAX_ATTR;
typedef uint32_t ATTR_FREQ_CORE_MAX_type;

// Type aliases and/or sizes for ATTR_FREQ_PROC_REFCLOCK attribute
typedef uint32_t FREQ_PROC_REFCLOCK_ATTR;
typedef uint32_t ATTR_FREQ_PROC_REFCLOCK_type;

// Type aliases and/or sizes for ATTR_FRU_ID attribute
typedef uint32_t FRU_ID_ATTR;
typedef uint32_t ATTR_FRU_ID_type;

// Type aliases and/or sizes for ATTR_FSI_MASTER_CHIP attribute
typedef EntityPath FSI_MASTER_CHIP_ATTR;
typedef EntityPath ATTR_FSI_MASTER_CHIP_type;

// Type aliases and/or sizes for ATTR_FSI_MASTER_PORT attribute
typedef uint8_t FSI_MASTER_PORT_ATTR;
typedef uint8_t ATTR_FSI_MASTER_PORT_type;

// Type aliases and/or sizes for ATTR_FSI_MASTER_TYPE attribute
typedef FSI_MASTER_TYPE FSI_MASTER_TYPE_ATTR;
typedef FSI_MASTER_TYPE ATTR_FSI_MASTER_TYPE_type;

// Type aliases and/or sizes for ATTR_FSI_OPTION_FLAGS attribute
typedef FsiOptionFlags FSI_OPTION_FLAGS_ATTR;
typedef FsiOptionFlags ATTR_FSI_OPTION_FLAGS_type;

// Type aliases and/or sizes for ATTR_FSI_SLAVE_CASCADE attribute
typedef uint8_t FSI_SLAVE_CASCADE_ATTR;
typedef uint8_t ATTR_FSI_SLAVE_CASCADE_type;

// Type aliases and/or sizes for ATTR_FSP_BAR_SIZE attribute
typedef uint64_t FSP_BAR_SIZE_ATTR;
typedef uint64_t ATTR_FSP_BAR_SIZE_type;

// Type aliases and/or sizes for ATTR_FSP_BASE_ADDR attribute
typedef uint64_t FSP_BASE_ADDR_ATTR;
typedef uint64_t ATTR_FSP_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_FUSED_CORE_MODE_HB attribute
typedef uint8_t FUSED_CORE_MODE_HB_ATTR;
typedef uint8_t ATTR_FUSED_CORE_MODE_HB_type;

// Type aliases and/or sizes for ATTR_FUSED_CORE_OPTION attribute
typedef uint8_t FUSED_CORE_OPTION_ATTR;
typedef uint8_t ATTR_FUSED_CORE_OPTION_type;

// Type aliases and/or sizes for ATTR_HB_HRMOR_NODAL_BASE attribute
typedef uint64_t HB_HRMOR_NODAL_BASE_ATTR;
typedef uint64_t ATTR_HB_HRMOR_NODAL_BASE_type;

// Type aliases and/or sizes for ATTR_HB_RSV_MEM_SIZE_MB attribute
typedef uint32_t HB_RSV_MEM_SIZE_MB_ATTR;
typedef uint32_t ATTR_HB_RSV_MEM_SIZE_MB_type;

// Type aliases and/or sizes for ATTR_HB_SETTINGS attribute
typedef HbSettings HB_SETTINGS_ATTR;
typedef HbSettings ATTR_HB_SETTINGS_type;

// Type aliases and/or sizes for ATTR_HDAT_EC attribute
typedef uint8_t HDAT_EC_ATTR;
typedef uint8_t ATTR_HDAT_EC_type;

// Type aliases and/or sizes for ATTR_HDAT_HBRT_NUM_SECTIONS attribute
typedef uint32_t HDAT_HBRT_NUM_SECTIONS_ATTR;
typedef uint32_t ATTR_HDAT_HBRT_NUM_SECTIONS_type;

// Type aliases and/or sizes for ATTR_HDAT_HBRT_SECTION_SIZE attribute
typedef uint64_t HDAT_HBRT_SECTION_SIZE_ATTR[9];
typedef uint64_t ATTR_HDAT_HBRT_SECTION_SIZE_type[9];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 9> ATTR_HDAT_HBRT_SECTION_SIZE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_RSV_MEM_NUM_SECTIONS attribute
typedef uint32_t HDAT_RSV_MEM_NUM_SECTIONS_ATTR;
typedef uint32_t ATTR_HDAT_RSV_MEM_NUM_SECTIONS_type;

// Type aliases and/or sizes for ATTR_HDDW_ORDER attribute
typedef uint8_t HDDW_ORDER_ATTR;
typedef uint8_t ATTR_HDDW_ORDER_type;

// Type aliases and/or sizes for ATTR_HOMER_PHYS_ADDR attribute
typedef uint64_t HOMER_PHYS_ADDR_ATTR;
typedef uint64_t ATTR_HOMER_PHYS_ADDR_type;

// Type aliases and/or sizes for ATTR_HOSTSVC_PLID attribute
typedef uint32_t HOSTSVC_PLID_ATTR;
typedef uint32_t ATTR_HOSTSVC_PLID_type;

// Type aliases and/or sizes for ATTR_HOT_PLUG_POWER_CONTROLLER_INFO attribute
typedef uint8_t HOT_PLUG_POWER_CONTROLLER_INFO_ATTR[8][8];
typedef uint8_t ATTR_HOT_PLUG_POWER_CONTROLLER_INFO_type[8][8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 8>, 8> ATTR_HOT_PLUG_POWER_CONTROLLER_INFO_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HUID attribute
typedef uint32_t HUID_ATTR;
typedef uint32_t ATTR_HUID_type;

// Type aliases and/or sizes for ATTR_HWAS_STATE attribute
typedef HwasState HWAS_STATE_ATTR;
typedef HwasState ATTR_HWAS_STATE_type;

// Type aliases and/or sizes for ATTR_HWAS_STATE_CHANGED_FLAG attribute
typedef uint64_t HWAS_STATE_CHANGED_FLAG_ATTR;
typedef uint64_t ATTR_HWAS_STATE_CHANGED_FLAG_type;

// Type aliases and/or sizes for ATTR_HWAS_STATE_CHANGED_SUBSCRIPTION_MASK attribute
typedef uint64_t HWAS_STATE_CHANGED_SUBSCRIPTION_MASK_ATTR;
typedef uint64_t ATTR_HWAS_STATE_CHANGED_SUBSCRIPTION_MASK_type;

// Type aliases and/or sizes for ATTR_I2C_BUS_SPEED_ARRAY attribute
typedef uint16_t I2C_BUS_SPEED_ARRAY_ATTR[4][13];
typedef uint16_t ATTR_I2C_BUS_SPEED_ARRAY_type[4][13];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 13>, 4> ATTR_I2C_BUS_SPEED_ARRAY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_I2C_CONTROL_INFO attribute
typedef I2cControlInfo I2C_CONTROL_INFO_ATTR;
typedef I2cControlInfo ATTR_I2C_CONTROL_INFO_type;

// Type aliases and/or sizes for ATTR_I2C_SWITCHES attribute
typedef I2cSwitches I2C_SWITCHES_ATTR;
typedef I2cSwitches ATTR_I2C_SWITCHES_type;

// Type aliases and/or sizes for ATTR_IBSCOM_MCS_BASE_ADDR attribute
typedef uint64_t IBSCOM_MCS_BASE_ADDR_ATTR;
typedef uint64_t ATTR_IBSCOM_MCS_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_IBSCOM_PROC_BASE_ADDR attribute
typedef uint64_t IBSCOM_PROC_BASE_ADDR_ATTR;
typedef uint64_t ATTR_IBSCOM_PROC_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_ICACHE_ASSOC_SETS attribute
typedef uint32_t ICACHE_ASSOC_SETS_ATTR;
typedef uint32_t ATTR_ICACHE_ASSOC_SETS_type;

// Type aliases and/or sizes for ATTR_ICACHE_BLOCK_SIZE attribute
typedef uint32_t ICACHE_BLOCK_SIZE_ATTR;
typedef uint32_t ATTR_ICACHE_BLOCK_SIZE_type;

// Type aliases and/or sizes for ATTR_ICACHE_LINE_SIZE attribute
typedef uint32_t ICACHE_LINE_SIZE_ATTR;
typedef uint32_t ATTR_ICACHE_LINE_SIZE_type;

// Type aliases and/or sizes for ATTR_ICACHE_SIZE attribute
typedef uint32_t ICACHE_SIZE_ATTR;
typedef uint32_t ATTR_ICACHE_SIZE_type;

// Type aliases and/or sizes for ATTR_IMT_BAR_SIZE attribute
typedef uint64_t IMT_BAR_SIZE_ATTR;
typedef uint64_t ATTR_IMT_BAR_SIZE_type;

// Type aliases and/or sizes for ATTR_IMT_BASE_ADDR attribute
typedef uint64_t IMT_BASE_ADDR_ATTR;
typedef uint64_t ATTR_IMT_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_INIT_TO_AVAILABLE attribute
typedef uint8_t INIT_TO_AVAILABLE_ATTR;
typedef uint8_t ATTR_INIT_TO_AVAILABLE_type;

// Type aliases and/or sizes for ATTR_INTP_BASE_ADDR attribute
typedef uint64_t INTP_BASE_ADDR_ATTR;
typedef uint64_t ATTR_INTP_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_ISTEP_MODE attribute
typedef uint8_t ISTEP_MODE_ATTR;
typedef uint8_t ATTR_ISTEP_MODE_type;

// Type aliases and/or sizes for ATTR_ISTEP_PAUSE_CONFIG attribute
typedef uint64_t ISTEP_PAUSE_CONFIG_ATTR;
typedef uint64_t ATTR_ISTEP_PAUSE_CONFIG_type;

// Type aliases and/or sizes for ATTR_ISTEP_PAUSE_ENABLE attribute
typedef uint8_t ISTEP_PAUSE_ENABLE_ATTR;
typedef uint8_t ATTR_ISTEP_PAUSE_ENABLE_type;

// Type aliases and/or sizes for ATTR_IS_INTER_ENCLOSURE_BUS attribute
typedef uint8_t IS_INTER_ENCLOSURE_BUS_ATTR;
typedef uint8_t ATTR_IS_INTER_ENCLOSURE_BUS_type;

// Type aliases and/or sizes for ATTR_KEY_CLEAR_REQUEST attribute
typedef KEY_CLEAR_REQUEST KEY_CLEAR_REQUEST_ATTR;
typedef KEY_CLEAR_REQUEST ATTR_KEY_CLEAR_REQUEST_type;

// Type aliases and/or sizes for ATTR_KEY_TRANSITION_STATE attribute
typedef KEY_TRANSITION_STATE KEY_TRANSITION_STATE_ATTR;
typedef KEY_TRANSITION_STATE ATTR_KEY_TRANSITION_STATE_type;

// Type aliases and/or sizes for ATTR_L2_CACHE_ASSOC_SETS attribute
typedef uint32_t L2_CACHE_ASSOC_SETS_ATTR;
typedef uint32_t ATTR_L2_CACHE_ASSOC_SETS_type;

// Type aliases and/or sizes for ATTR_L2_CACHE_LINE_SIZE attribute
typedef uint32_t L2_CACHE_LINE_SIZE_ATTR;
typedef uint32_t ATTR_L2_CACHE_LINE_SIZE_type;

// Type aliases and/or sizes for ATTR_L2_CACHE_SIZE attribute
typedef uint32_t L2_CACHE_SIZE_ATTR;
typedef uint32_t ATTR_L2_CACHE_SIZE_type;

// Type aliases and/or sizes for ATTR_L3_CACHE_LINE_SIZE attribute
typedef uint32_t L3_CACHE_LINE_SIZE_ATTR;
typedef uint32_t ATTR_L3_CACHE_LINE_SIZE_type;

// Type aliases and/or sizes for ATTR_L3_CACHE_SIZE attribute
typedef uint32_t L3_CACHE_SIZE_ATTR;
typedef uint32_t ATTR_L3_CACHE_SIZE_type;

// Type aliases and/or sizes for ATTR_LOCATION_CODE attribute
typedef char LOCATION_CODE_ATTR[64];
typedef char ATTR_LOCATION_CODE_type[64];
const size_t ATTR_LOCATION_CODE_max_chars = 63;

// Type aliases and/or sizes for ATTR_LPC_BUS_ADDR attribute
typedef uint64_t LPC_BUS_ADDR_ATTR;
typedef uint64_t ATTR_LPC_BUS_ADDR_type;

// Type aliases and/or sizes for ATTR_MAX_CHIPLETS_PER_PROC attribute
typedef uint8_t MAX_CHIPLETS_PER_PROC_ATTR;
typedef uint8_t ATTR_MAX_CHIPLETS_PER_PROC_type;

// Type aliases and/or sizes for ATTR_MAX_COMPUTE_NODES_PER_SYSTEM attribute
typedef uint8_t MAX_COMPUTE_NODES_PER_SYSTEM_ATTR;
typedef uint8_t ATTR_MAX_COMPUTE_NODES_PER_SYSTEM_type;

// Type aliases and/or sizes for ATTR_MAX_DIMMS_PER_MBA_PORT attribute
typedef uint8_t MAX_DIMMS_PER_MBA_PORT_ATTR;
typedef uint8_t ATTR_MAX_DIMMS_PER_MBA_PORT_type;

// Type aliases and/or sizes for ATTR_MAX_DMI_PER_PROC attribute
typedef uint8_t MAX_DMI_PER_PROC_ATTR;
typedef uint8_t ATTR_MAX_DMI_PER_PROC_type;

// Type aliases and/or sizes for ATTR_MAX_EXS_PER_PROC_CHIP attribute
typedef uint8_t MAX_EXS_PER_PROC_CHIP_ATTR;
typedef uint8_t ATTR_MAX_EXS_PER_PROC_CHIP_type;

// Type aliases and/or sizes for ATTR_MAX_MBAS_PER_MEMBUF_CHIP attribute
typedef uint8_t MAX_MBAS_PER_MEMBUF_CHIP_ATTR;
typedef uint8_t ATTR_MAX_MBAS_PER_MEMBUF_CHIP_type;

// Type aliases and/or sizes for ATTR_MAX_MBA_PORTS_PER_MBA attribute
typedef uint8_t MAX_MBA_PORTS_PER_MBA_ATTR;
typedef uint8_t ATTR_MAX_MBA_PORTS_PER_MBA_type;

// Type aliases and/or sizes for ATTR_MAX_MCS_PER_SYSTEM attribute
typedef uint8_t MAX_MCS_PER_SYSTEM_ATTR;
typedef uint8_t ATTR_MAX_MCS_PER_SYSTEM_type;

// Type aliases and/or sizes for ATTR_MAX_POWER attribute
typedef uint8_t MAX_POWER_ATTR;
typedef uint8_t ATTR_MAX_POWER_type;

// Type aliases and/or sizes for ATTR_MAX_PROC_CHIPS_PER_NODE attribute
typedef uint8_t MAX_PROC_CHIPS_PER_NODE_ATTR;
typedef uint8_t ATTR_MAX_PROC_CHIPS_PER_NODE_type;

// Type aliases and/or sizes for ATTR_MAX_SBE_SEEPROM_SIZE attribute
typedef uint32_t MAX_SBE_SEEPROM_SIZE_ATTR;
typedef uint32_t ATTR_MAX_SBE_SEEPROM_SIZE_type;

// Type aliases and/or sizes for ATTR_MBA_DIMM attribute
typedef uint8_t MBA_DIMM_ATTR;
typedef uint8_t ATTR_MBA_DIMM_type;

// Type aliases and/or sizes for ATTR_MBA_PORT attribute
typedef uint8_t MBA_PORT_ATTR;
typedef uint8_t ATTR_MBA_PORT_type;

// Type aliases and/or sizes for ATTR_MC_PLL_BUCKET attribute
typedef uint8_t MC_PLL_BUCKET_ATTR;
typedef uint8_t ATTR_MC_PLL_BUCKET_type;

// Type aliases and/or sizes for ATTR_MEM_AVDD_OFFSET_MILLIVOLTS attribute
typedef uint32_t MEM_AVDD_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MEM_AVDD_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MEM_BASE attribute
typedef uint64_t MEM_BASE_ATTR;
typedef uint64_t ATTR_MEM_BASE_type;

// Type aliases and/or sizes for ATTR_MEM_PORT attribute
typedef uint8_t MEM_PORT_ATTR;
typedef uint8_t ATTR_MEM_PORT_type;

// Type aliases and/or sizes for ATTR_MEM_VCS_OFFSET_MILLIVOLTS attribute
typedef uint32_t MEM_VCS_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MEM_VCS_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MEM_VDDR_OFFSET_MILLIVOLTS attribute
typedef uint32_t MEM_VDDR_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MEM_VDDR_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MEM_VDD_OFFSET_MILLIVOLTS attribute
typedef uint32_t MEM_VDD_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MEM_VDD_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MEM_VPP_OFFSET_MILLIVOLTS attribute
typedef uint32_t MEM_VPP_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MEM_VPP_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MFG_TRACE_ENABLE attribute
typedef uint8_t MFG_TRACE_ENABLE_ATTR;
typedef uint8_t ATTR_MFG_TRACE_ENABLE_type;

// Type aliases and/or sizes for ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET attribute
typedef MFG_WRAP_TEST_ABUS_LINKS_SET MFG_WRAP_TEST_ABUS_LINKS_SET_ATTR;
typedef MFG_WRAP_TEST_ABUS_LINKS_SET ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_type;

// Type aliases and/or sizes for ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_ENABLE attribute
typedef uint8_t MFG_WRAP_TEST_ABUS_LINKS_SET_ENABLE_ATTR;
typedef uint8_t ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_ENABLE_type;

// Type aliases and/or sizes for ATTR_MGC_LOAD_SOURCE attribute
typedef uint8_t MGC_LOAD_SOURCE_ATTR;
typedef uint8_t ATTR_MGC_LOAD_SOURCE_type;

// Type aliases and/or sizes for ATTR_MIN_FREQ_MHZ attribute
typedef uint32_t MIN_FREQ_MHZ_ATTR;
typedef uint32_t ATTR_MIN_FREQ_MHZ_type;

// Type aliases and/or sizes for ATTR_MIRROR_BASE_ADDRESS attribute
typedef uint64_t MIRROR_BASE_ADDRESS_ATTR;
typedef uint64_t ATTR_MIRROR_BASE_ADDRESS_type;

// Type aliases and/or sizes for ATTR_MNFG_ABUS_MIN_EYE_HEIGHT attribute
typedef uint8_t MNFG_ABUS_MIN_EYE_HEIGHT_ATTR;
typedef uint8_t ATTR_MNFG_ABUS_MIN_EYE_HEIGHT_type;

// Type aliases and/or sizes for ATTR_MNFG_ABUS_MIN_EYE_WIDTH attribute
typedef uint8_t MNFG_ABUS_MIN_EYE_WIDTH_ATTR;
typedef uint8_t ATTR_MNFG_ABUS_MIN_EYE_WIDTH_type;

// Type aliases and/or sizes for ATTR_MNFG_DMI_MIN_EYE_HEIGHT attribute
typedef uint8_t MNFG_DMI_MIN_EYE_HEIGHT_ATTR;
typedef uint8_t ATTR_MNFG_DMI_MIN_EYE_HEIGHT_type;

// Type aliases and/or sizes for ATTR_MNFG_DMI_MIN_EYE_WIDTH attribute
typedef uint8_t MNFG_DMI_MIN_EYE_WIDTH_ATTR;
typedef uint8_t ATTR_MNFG_DMI_MIN_EYE_WIDTH_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_CEN_L4_CACHE_CES attribute
typedef uint8_t MNFG_TH_CEN_L4_CACHE_CES_ATTR;
typedef uint8_t ATTR_MNFG_TH_CEN_L4_CACHE_CES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_CEN_MBA_IPL_SOFT_CE_TH_ALGO attribute
typedef uint8_t MNFG_TH_CEN_MBA_IPL_SOFT_CE_TH_ALGO_ATTR;
typedef uint8_t ATTR_MNFG_TH_CEN_MBA_IPL_SOFT_CE_TH_ALGO_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_CEN_MBA_RT_RCE_PER_RANK attribute
typedef uint8_t MNFG_TH_CEN_MBA_RT_RCE_PER_RANK_ATTR;
typedef uint8_t ATTR_MNFG_TH_CEN_MBA_RT_RCE_PER_RANK_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_CEN_MBA_RT_SOFT_CE_TH_ALGO attribute
typedef uint8_t MNFG_TH_CEN_MBA_RT_SOFT_CE_TH_ALGO_ATTR;
typedef uint8_t ATTR_MNFG_TH_CEN_MBA_RT_SOFT_CE_TH_ALGO_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_MEMORY_IMPES attribute
typedef uint8_t MNFG_TH_MEMORY_IMPES_ATTR;
typedef uint8_t ATTR_MNFG_TH_MEMORY_IMPES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_MEMORY_IUES attribute
typedef uint8_t MNFG_TH_MEMORY_IUES_ATTR;
typedef uint8_t ATTR_MNFG_TH_MEMORY_IUES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L2_CACHE_CES attribute
typedef uint8_t MNFG_TH_P8EX_L2_CACHE_CES_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L2_CACHE_CES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L2_COL_REPAIRS attribute
typedef uint8_t MNFG_TH_P8EX_L2_COL_REPAIRS_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L2_COL_REPAIRS_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L2_DIR_CES attribute
typedef uint8_t MNFG_TH_P8EX_L2_DIR_CES_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L2_DIR_CES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L2_LINE_DELETES attribute
typedef uint8_t MNFG_TH_P8EX_L2_LINE_DELETES_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L2_LINE_DELETES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L3_CACHE_CES attribute
typedef uint8_t MNFG_TH_P8EX_L3_CACHE_CES_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L3_CACHE_CES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L3_COL_REPAIRS attribute
typedef uint8_t MNFG_TH_P8EX_L3_COL_REPAIRS_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L3_COL_REPAIRS_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L3_DIR_CES attribute
typedef uint8_t MNFG_TH_P8EX_L3_DIR_CES_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L3_DIR_CES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_P8EX_L3_LINE_DELETES attribute
typedef uint8_t MNFG_TH_P8EX_L3_LINE_DELETES_ATTR;
typedef uint8_t ATTR_MNFG_TH_P8EX_L3_LINE_DELETES_type;

// Type aliases and/or sizes for ATTR_MNFG_TH_RCD_PARITY_ERRORS attribute
typedef uint8_t MNFG_TH_RCD_PARITY_ERRORS_ATTR;
typedef uint8_t ATTR_MNFG_TH_RCD_PARITY_ERRORS_type;

// Type aliases and/or sizes for ATTR_MNFG_XBUS_MIN_EYE_WIDTH attribute
typedef uint8_t MNFG_XBUS_MIN_EYE_WIDTH_ATTR;
typedef uint8_t ATTR_MNFG_XBUS_MIN_EYE_WIDTH_type;

// Type aliases and/or sizes for ATTR_MODEL attribute
typedef MODEL MODEL_ATTR;
typedef MODEL ATTR_MODEL_type;

// Type aliases and/or sizes for ATTR_MRU_ID attribute
typedef uint32_t MRU_ID_ATTR;
typedef uint32_t ATTR_MRU_ID_type;

// Type aliases and/or sizes for ATTR_MRW_DDR3_VDDR_MAX_LIMIT attribute
typedef uint32_t MRW_DDR3_VDDR_MAX_LIMIT_ATTR;
typedef uint32_t ATTR_MRW_DDR3_VDDR_MAX_LIMIT_type;

// Type aliases and/or sizes for ATTR_MRW_DDR3_VDDR_MAX_LIMIT_POST_DRAM_INIT attribute
typedef uint32_t MRW_DDR3_VDDR_MAX_LIMIT_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MRW_DDR3_VDDR_MAX_LIMIT_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MRW_DDR4_VDDR_MAX_LIMIT attribute
typedef uint32_t MRW_DDR4_VDDR_MAX_LIMIT_ATTR;
typedef uint32_t ATTR_MRW_DDR4_VDDR_MAX_LIMIT_type;

// Type aliases and/or sizes for ATTR_MRW_DDR4_VDDR_MAX_LIMIT_POST_DRAM_INIT attribute
typedef uint32_t MRW_DDR4_VDDR_MAX_LIMIT_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MRW_DDR4_VDDR_MAX_LIMIT_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MRW_DEFAULT_RISK_LEVEL attribute
typedef uint32_t MRW_DEFAULT_RISK_LEVEL_ATTR;
typedef uint32_t ATTR_MRW_DEFAULT_RISK_LEVEL_type;

// Type aliases and/or sizes for ATTR_MRW_MEM_THROTTLE_DENOMINATOR attribute
typedef uint32_t MRW_MEM_THROTTLE_DENOMINATOR_ATTR;
typedef uint32_t ATTR_MRW_MEM_THROTTLE_DENOMINATOR_type;

// Type aliases and/or sizes for ATTR_MSS_AVDD_PROGRAM attribute
typedef uint8_t MSS_AVDD_PROGRAM_ATTR;
typedef uint8_t ATTR_MSS_AVDD_PROGRAM_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_AVDD_INTERCEPT attribute
typedef uint32_t MSS_CENT_AVDD_INTERCEPT_ATTR;
typedef uint32_t ATTR_MSS_CENT_AVDD_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_AVDD_SLOPE_ACTIVE attribute
typedef uint32_t MSS_CENT_AVDD_SLOPE_ACTIVE_ATTR;
typedef uint32_t ATTR_MSS_CENT_AVDD_SLOPE_ACTIVE_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_AVDD_SLOPE_INACTIVE attribute
typedef uint32_t MSS_CENT_AVDD_SLOPE_INACTIVE_ATTR;
typedef uint32_t ATTR_MSS_CENT_AVDD_SLOPE_INACTIVE_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_VCS_INTERCEPT attribute
typedef uint32_t MSS_CENT_VCS_INTERCEPT_ATTR;
typedef uint32_t ATTR_MSS_CENT_VCS_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_VCS_SLOPE_ACTIVE attribute
typedef uint32_t MSS_CENT_VCS_SLOPE_ACTIVE_ATTR;
typedef uint32_t ATTR_MSS_CENT_VCS_SLOPE_ACTIVE_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_VCS_SLOPE_INACTIVE attribute
typedef uint32_t MSS_CENT_VCS_SLOPE_INACTIVE_ATTR;
typedef uint32_t ATTR_MSS_CENT_VCS_SLOPE_INACTIVE_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_VDD_INTERCEPT attribute
typedef uint32_t MSS_CENT_VDD_INTERCEPT_ATTR;
typedef uint32_t ATTR_MSS_CENT_VDD_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_VDD_SLOPE_ACTIVE attribute
typedef uint32_t MSS_CENT_VDD_SLOPE_ACTIVE_ATTR;
typedef uint32_t ATTR_MSS_CENT_VDD_SLOPE_ACTIVE_type;

// Type aliases and/or sizes for ATTR_MSS_CENT_VDD_SLOPE_INACTIVE attribute
typedef uint32_t MSS_CENT_VDD_SLOPE_INACTIVE_ATTR;
typedef uint32_t ATTR_MSS_CENT_VDD_SLOPE_INACTIVE_type;

// Type aliases and/or sizes for ATTR_MSS_DATABUS_UTIL_PER_MBA attribute
typedef uint8_t MSS_DATABUS_UTIL_PER_MBA_ATTR;
typedef uint8_t ATTR_MSS_DATABUS_UTIL_PER_MBA_type;

// Type aliases and/or sizes for ATTR_MSS_MBA_ADDR_INTERLEAVE_BIT attribute
typedef uint8_t MSS_MBA_ADDR_INTERLEAVE_BIT_ATTR;
typedef uint8_t ATTR_MSS_MBA_ADDR_INTERLEAVE_BIT_type;

// Type aliases and/or sizes for ATTR_MSS_MBA_CACHELINE_INTERLEAVE_MODE attribute
typedef uint8_t MSS_MBA_CACHELINE_INTERLEAVE_MODE_ATTR;
typedef uint8_t ATTR_MSS_MBA_CACHELINE_INTERLEAVE_MODE_type;

// Type aliases and/or sizes for ATTR_MSS_PHY_SEQ_REFRESH attribute
typedef uint8_t MSS_PHY_SEQ_REFRESH_ATTR[2];
typedef uint8_t ATTR_MSS_PHY_SEQ_REFRESH_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_MSS_PHY_SEQ_REFRESH_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MSS_VCS_PROGRAM attribute
typedef uint8_t MSS_VCS_PROGRAM_ATTR;
typedef uint8_t ATTR_MSS_VCS_PROGRAM_type;

// Type aliases and/or sizes for ATTR_MSS_VDDR_PROGRAM attribute
typedef uint8_t MSS_VDDR_PROGRAM_ATTR;
typedef uint8_t ATTR_MSS_VDDR_PROGRAM_type;

// Type aliases and/or sizes for ATTR_MSS_VDD_PROGRAM attribute
typedef uint8_t MSS_VDD_PROGRAM_ATTR;
typedef uint8_t ATTR_MSS_VDD_PROGRAM_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_AVDD_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_AVDD_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_AVDD_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_AVDD_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT attribute
typedef uint32_t MSS_VOLT_DDR3_VDDR_INTERCEPT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_POST_DRAM_INIT attribute
typedef uint32_t MSS_VOLT_DDR3_VDDR_INTERCEPT_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR3_VDDR_SLOPE attribute
typedef uint32_t MSS_VOLT_DDR3_VDDR_SLOPE_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_POST_DRAM_INIT attribute
typedef uint32_t MSS_VOLT_DDR3_VDDR_SLOPE_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT attribute
typedef uint32_t MSS_VOLT_DDR4_VDDR_INTERCEPT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_POST_DRAM_INIT attribute
typedef uint32_t MSS_VOLT_DDR4_VDDR_INTERCEPT_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR4_VDDR_SLOPE attribute
typedef uint32_t MSS_VOLT_DDR4_VDDR_SLOPE_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_POST_DRAM_INIT attribute
typedef uint32_t MSS_VOLT_DDR4_VDDR_SLOPE_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VCS_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VCS_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VCS_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VCS_OFFSET_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VCS_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VCS_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VDDR_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VDDR_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VDDR_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VDDR_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VDD_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VDD_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VDD_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VDD_OFFSET_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VDD_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VDD_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_INTERCEPT attribute
typedef uint32_t MSS_VOLT_VPP_INTERCEPT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_INTERCEPT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_INTERCEPT_POST_DRAM_INIT attribute
typedef uint32_t MSS_VOLT_VPP_INTERCEPT_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_INTERCEPT_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VPP_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_OFFSET_MILLIVOLTS attribute
typedef uint32_t MSS_VOLT_VPP_OFFSET_MILLIVOLTS_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_OFFSET_MILLIVOLTS_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_SLOPE attribute
typedef uint32_t MSS_VOLT_VPP_SLOPE_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_SLOPE_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_SLOPE_POST_DRAM_INIT attribute
typedef uint32_t MSS_VOLT_VPP_SLOPE_POST_DRAM_INIT_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_SLOPE_POST_DRAM_INIT_type;

// Type aliases and/or sizes for ATTR_MSS_VPP_PROGRAM attribute
typedef uint8_t MSS_VPP_PROGRAM_ATTR;
typedef uint8_t ATTR_MSS_VPP_PROGRAM_type;

// Type aliases and/or sizes for ATTR_NEST_PLL_FREQ_BUCKETS attribute
typedef uint8_t NEST_PLL_FREQ_BUCKETS_ATTR;
typedef uint8_t ATTR_NEST_PLL_FREQ_BUCKETS_type;

// Type aliases and/or sizes for ATTR_NEST_PLL_FREQ_I2CDIV_LIST attribute
typedef uint32_t NEST_PLL_FREQ_I2CDIV_LIST_ATTR[5];
typedef uint32_t ATTR_NEST_PLL_FREQ_I2CDIV_LIST_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 5> ATTR_NEST_PLL_FREQ_I2CDIV_LIST_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_NEST_PLL_FREQ_LIST attribute
typedef uint32_t NEST_PLL_FREQ_LIST_ATTR[5];
typedef uint32_t ATTR_NEST_PLL_FREQ_LIST_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 5> ATTR_NEST_PLL_FREQ_LIST_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_NEST_VCS_ID attribute
typedef uint16_t NEST_VCS_ID_ATTR;
typedef uint16_t ATTR_NEST_VCS_ID_type;

// Type aliases and/or sizes for ATTR_NEST_VDDR_ID attribute
typedef uint16_t NEST_VDDR_ID_ATTR;
typedef uint16_t ATTR_NEST_VDDR_ID_type;

// Type aliases and/or sizes for ATTR_NEST_VDD_ID attribute
typedef uint16_t NEST_VDD_ID_ATTR;
typedef uint16_t ATTR_NEST_VDD_ID_type;

// Type aliases and/or sizes for ATTR_NEST_VDN_ID attribute
typedef uint16_t NEST_VDN_ID_ATTR;
typedef uint16_t ATTR_NEST_VDN_ID_type;

// Type aliases and/or sizes for ATTR_NEST_VIO_ID attribute
typedef uint16_t NEST_VIO_ID_ATTR;
typedef uint16_t ATTR_NEST_VIO_ID_type;

// Type aliases and/or sizes for ATTR_NOMINAL_FREQ_MHZ attribute
typedef uint32_t NOMINAL_FREQ_MHZ_ATTR;
typedef uint32_t ATTR_NOMINAL_FREQ_MHZ_type;

// Type aliases and/or sizes for ATTR_NPU_MMIO_BAR_BASE_ADDR attribute
typedef uint64_t NPU_MMIO_BAR_BASE_ADDR_ATTR[4][2];
typedef uint64_t ATTR_NPU_MMIO_BAR_BASE_ADDR_type[4][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint64_t, 2>, 4> ATTR_NPU_MMIO_BAR_BASE_ADDR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_NPU_MMIO_BAR_SIZE attribute
typedef uint64_t NPU_MMIO_BAR_SIZE_ATTR[4][2];
typedef uint64_t ATTR_NPU_MMIO_BAR_SIZE_type[4][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint64_t, 2>, 4> ATTR_NPU_MMIO_BAR_SIZE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_NUMERIC_POD_TYPE_TEST attribute
typedef NumericPodTypeTest NUMERIC_POD_TYPE_TEST_ATTR;
typedef NumericPodTypeTest ATTR_NUMERIC_POD_TYPE_TEST_type;

// Type aliases and/or sizes for ATTR_NVDIMM_ARMED attribute
typedef NvdimmArmed NVDIMM_ARMED_ATTR;
typedef NvdimmArmed ATTR_NVDIMM_ARMED_type;

// Type aliases and/or sizes for ATTR_NVDIMM_AUTO_ARM attribute
typedef uint8_t NVDIMM_AUTO_ARM_ATTR;
typedef uint8_t ATTR_NVDIMM_AUTO_ARM_type;

// Type aliases and/or sizes for ATTR_NVDIMM_ENCRYPTION_ENABLE attribute
typedef uint8_t NVDIMM_ENCRYPTION_ENABLE_ATTR;
typedef uint8_t ATTR_NVDIMM_ENCRYPTION_ENABLE_type;

// Type aliases and/or sizes for ATTR_NVDIMM_ENCRYPTION_KEYS_ANCHOR attribute
typedef uint8_t NVDIMM_ENCRYPTION_KEYS_ANCHOR_ATTR[96];
typedef uint8_t ATTR_NVDIMM_ENCRYPTION_KEYS_ANCHOR_type[96];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 96> ATTR_NVDIMM_ENCRYPTION_KEYS_ANCHOR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_NVDIMM_ENCRYPTION_KEYS_FW attribute
typedef uint8_t NVDIMM_ENCRYPTION_KEYS_FW_ATTR[96];
typedef uint8_t ATTR_NVDIMM_ENCRYPTION_KEYS_FW_type[96];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 96> ATTR_NVDIMM_ENCRYPTION_KEYS_FW_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_NVDIMM_FORCE_FW_UPDATE attribute
typedef uint8_t NVDIMM_FORCE_FW_UPDATE_ATTR;
typedef uint8_t ATTR_NVDIMM_FORCE_FW_UPDATE_type;

// Type aliases and/or sizes for ATTR_NVDIMM_UPDATE_I2C_BLOCK_SIZE attribute
typedef uint8_t NVDIMM_UPDATE_I2C_BLOCK_SIZE_ATTR;
typedef uint8_t ATTR_NVDIMM_UPDATE_I2C_BLOCK_SIZE_type;

// Type aliases and/or sizes for ATTR_NV_OPS_TIMEOUT_MSEC attribute
typedef uint32_t NV_OPS_TIMEOUT_MSEC_ATTR[6];
typedef uint32_t ATTR_NV_OPS_TIMEOUT_MSEC_type[6];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 6> ATTR_NV_OPS_TIMEOUT_MSEC_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_NV_STATUS_FLAG attribute
typedef uint8_t NV_STATUS_FLAG_ATTR;
typedef uint8_t ATTR_NV_STATUS_FLAG_type;

// Type aliases and/or sizes for ATTR_OBUS_BRICK_LANE_MASK attribute
typedef uint32_t OBUS_BRICK_LANE_MASK_ATTR;
typedef uint32_t ATTR_OBUS_BRICK_LANE_MASK_type;

// Type aliases and/or sizes for ATTR_OBUS_SLOT_INDEX attribute
typedef uint8_t OBUS_SLOT_INDEX_ATTR;
typedef uint8_t ATTR_OBUS_SLOT_INDEX_type;

// Type aliases and/or sizes for ATTR_OCC_MASTER_CAPABLE attribute
typedef uint8_t OCC_MASTER_CAPABLE_ATTR;
typedef uint8_t ATTR_OCC_MASTER_CAPABLE_type;

// Type aliases and/or sizes for ATTR_OCMB_FW_UPDATE_OVERRIDE attribute
typedef uint8_t OCMB_FW_UPDATE_OVERRIDE_ATTR;
typedef uint8_t ATTR_OCMB_FW_UPDATE_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_ORDINAL_ID attribute
typedef uint32_t ORDINAL_ID_ATTR;
typedef uint32_t ATTR_ORDINAL_ID_type;

// Type aliases and/or sizes for ATTR_PARENT_DECONFIG_RULES attribute
typedef ParentDeconfigRules PARENT_DECONFIG_RULES_ATTR;
typedef ParentDeconfigRules ATTR_PARENT_DECONFIG_RULES_type;

// Type aliases and/or sizes for ATTR_PAYLOAD_BASE attribute
typedef uint64_t PAYLOAD_BASE_ATTR;
typedef uint64_t ATTR_PAYLOAD_BASE_type;

// Type aliases and/or sizes for ATTR_PAYLOAD_ENTRY attribute
typedef uint64_t PAYLOAD_ENTRY_ATTR;
typedef uint64_t ATTR_PAYLOAD_ENTRY_type;

// Type aliases and/or sizes for ATTR_PAYLOAD_IN_MIRROR_MEM attribute
typedef uint8_t PAYLOAD_IN_MIRROR_MEM_ATTR;
typedef uint8_t ATTR_PAYLOAD_IN_MIRROR_MEM_type;

// Type aliases and/or sizes for ATTR_PAYLOAD_KIND attribute
typedef PAYLOAD_KIND PAYLOAD_KIND_ATTR;
typedef PAYLOAD_KIND ATTR_PAYLOAD_KIND_type;

// Type aliases and/or sizes for ATTR_PCIE_32BIT_DMA_SIZE attribute
typedef uint8_t PCIE_32BIT_DMA_SIZE_ATTR;
typedef uint8_t ATTR_PCIE_32BIT_DMA_SIZE_type;

// Type aliases and/or sizes for ATTR_PCIE_32BIT_MMIO_SIZE attribute
typedef uint8_t PCIE_32BIT_MMIO_SIZE_ATTR;
typedef uint8_t ATTR_PCIE_32BIT_MMIO_SIZE_type;

// Type aliases and/or sizes for ATTR_PCIE_64BIT_DMA_SIZE attribute
typedef uint8_t PCIE_64BIT_DMA_SIZE_ATTR;
typedef uint8_t ATTR_PCIE_64BIT_DMA_SIZE_type;

// Type aliases and/or sizes for ATTR_PCIE_64BIT_MMIO_SIZE attribute
typedef uint8_t PCIE_64BIT_MMIO_SIZE_ATTR;
typedef uint8_t ATTR_PCIE_64BIT_MMIO_SIZE_type;

// Type aliases and/or sizes for ATTR_PCIE_CAPABILITES attribute
typedef uint8_t PCIE_CAPABILITES_ATTR;
typedef uint8_t ATTR_PCIE_CAPABILITES_type;

// Type aliases and/or sizes for ATTR_PCI_BASE_ADDRS_32 attribute
typedef uint64_t PCI_BASE_ADDRS_32_ATTR[4];
typedef uint64_t ATTR_PCI_BASE_ADDRS_32_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 4> ATTR_PCI_BASE_ADDRS_32_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PCI_BASE_ADDRS_64 attribute
typedef uint64_t PCI_BASE_ADDRS_64_ATTR[4];
typedef uint64_t ATTR_PCI_BASE_ADDRS_64_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 4> ATTR_PCI_BASE_ADDRS_64_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PCI_REFCLOCK_RCVR_TERM attribute
typedef uint8_t PCI_REFCLOCK_RCVR_TERM_ATTR;
typedef uint8_t ATTR_PCI_REFCLOCK_RCVR_TERM_type;

// Type aliases and/or sizes for ATTR_PEC_IS_BIFURCATABLE attribute
typedef uint8_t PEC_IS_BIFURCATABLE_ATTR;
typedef uint8_t ATTR_PEC_IS_BIFURCATABLE_type;

// Type aliases and/or sizes for ATTR_PEC_PCIE_HX_KEYWORD_DATA attribute
typedef uint8_t PEC_PCIE_HX_KEYWORD_DATA_ATTR[9];
typedef uint8_t ATTR_PEC_PCIE_HX_KEYWORD_DATA_type[9];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 9> ATTR_PEC_PCIE_HX_KEYWORD_DATA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PEC_PCIE_IOP_REVERSAL attribute
typedef uint8_t PEC_PCIE_IOP_REVERSAL_ATTR[4];
typedef uint8_t ATTR_PEC_PCIE_IOP_REVERSAL_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PEC_PCIE_IOP_REVERSAL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PEC_PCIE_IOP_REVERSAL_BIFURCATED attribute
typedef uint8_t PEC_PCIE_IOP_REVERSAL_BIFURCATED_ATTR[4];
typedef uint8_t ATTR_PEC_PCIE_IOP_REVERSAL_BIFURCATED_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PEC_PCIE_IOP_REVERSAL_BIFURCATED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED attribute
typedef uint8_t PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED_ATTR[4];
typedef uint8_t ATTR_PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PEC_PCIE_IOP_SWAP_BIFURCATED attribute
typedef uint8_t PEC_PCIE_IOP_SWAP_BIFURCATED_ATTR;
typedef uint8_t ATTR_PEC_PCIE_IOP_SWAP_BIFURCATED_type;

// Type aliases and/or sizes for ATTR_PEC_PCIE_IOP_SWAP_NON_BIFURCATED attribute
typedef uint8_t PEC_PCIE_IOP_SWAP_NON_BIFURCATED_ATTR;
typedef uint8_t ATTR_PEC_PCIE_IOP_SWAP_NON_BIFURCATED_type;

// Type aliases and/or sizes for ATTR_PEC_PCIE_LANE_MASK_BIFURCATED attribute
typedef uint16_t PEC_PCIE_LANE_MASK_BIFURCATED_ATTR[4];
typedef uint16_t ATTR_PEC_PCIE_LANE_MASK_BIFURCATED_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 4> ATTR_PEC_PCIE_LANE_MASK_BIFURCATED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PEC_PCIE_LANE_MASK_NON_BIFURCATED attribute
typedef uint16_t PEC_PCIE_LANE_MASK_NON_BIFURCATED_ATTR[4];
typedef uint16_t ATTR_PEC_PCIE_LANE_MASK_NON_BIFURCATED_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 4> ATTR_PEC_PCIE_LANE_MASK_NON_BIFURCATED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PEER_PATH attribute
typedef EntityPath PEER_PATH_ATTR;
typedef EntityPath ATTR_PEER_PATH_type;

// Type aliases and/or sizes for ATTR_PEER_TARGET attribute
typedef TARGETING::Target* PEER_TARGET_ATTR;
typedef TARGETING::Target* ATTR_PEER_TARGET_type;

// Type aliases and/or sizes for ATTR_PG attribute
typedef uint16_t PG_ATTR;
typedef uint16_t ATTR_PG_type;

// Type aliases and/or sizes for ATTR_PHB_BASE_ADDRS attribute
typedef uint64_t PHB_BASE_ADDRS_ATTR[4];
typedef uint64_t ATTR_PHB_BASE_ADDRS_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 4> ATTR_PHB_BASE_ADDRS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PHYS_PATH attribute
typedef EntityPath PHYS_PATH_ATTR;
typedef EntityPath ATTR_PHYS_PATH_type;

// Type aliases and/or sizes for ATTR_PIB_I2C_NEST_PLL attribute
typedef uint32_t PIB_I2C_NEST_PLL_ATTR;
typedef uint32_t ATTR_PIB_I2C_NEST_PLL_type;

// Type aliases and/or sizes for ATTR_PIB_I2C_REFCLOCK attribute
typedef uint32_t PIB_I2C_REFCLOCK_ATTR;
typedef uint32_t ATTR_PIB_I2C_REFCLOCK_type;

// Type aliases and/or sizes for ATTR_PLCK_IPL_ATTR_OVERRIDES_EXIST attribute
typedef uint8_t PLCK_IPL_ATTR_OVERRIDES_EXIST_ATTR;
typedef uint8_t ATTR_PLCK_IPL_ATTR_OVERRIDES_EXIST_type;

// Type aliases and/or sizes for ATTR_PNOR_FLASH_WORKAROUNDS attribute
typedef uint32_t PNOR_FLASH_WORKAROUNDS_ATTR;
typedef uint32_t ATTR_PNOR_FLASH_WORKAROUNDS_type;

// Type aliases and/or sizes for ATTR_POSITION attribute
typedef uint16_t POSITION_ATTR;
typedef uint16_t ATTR_POSITION_type;

// Type aliases and/or sizes for ATTR_POS_ON_MEM_PORT attribute
typedef uint8_t POS_ON_MEM_PORT_ATTR;
typedef uint8_t ATTR_POS_ON_MEM_PORT_type;

// Type aliases and/or sizes for ATTR_POWER_PATH attribute
typedef EntityPath POWER_PATH_ATTR;
typedef EntityPath ATTR_POWER_PATH_type;

// Type aliases and/or sizes for ATTR_PRD_HWP_PLID attribute
typedef uint32_t PRD_HWP_PLID_ATTR;
typedef uint32_t ATTR_PRD_HWP_PLID_type;

// Type aliases and/or sizes for ATTR_PREVIOUS_WAKEUP_FAIL attribute
typedef uint8_t PREVIOUS_WAKEUP_FAIL_ATTR;
typedef uint8_t ATTR_PREVIOUS_WAKEUP_FAIL_type;

// Type aliases and/or sizes for ATTR_PRIMARY_CAPABILITIES attribute
typedef PrimaryCapabilities PRIMARY_CAPABILITIES_ATTR;
typedef PrimaryCapabilities ATTR_PRIMARY_CAPABILITIES_type;

// Type aliases and/or sizes for ATTR_PROC_COMPATIBILITY_REQ attribute
typedef PROC_COMPATIBILITY_REQ PROC_COMPATIBILITY_REQ_ATTR;
typedef PROC_COMPATIBILITY_REQ ATTR_PROC_COMPATIBILITY_REQ_type;

// Type aliases and/or sizes for ATTR_PROC_DCM_INSTALLED attribute
typedef uint8_t PROC_DCM_INSTALLED_ATTR;
typedef uint8_t ATTR_PROC_DCM_INSTALLED_type;

// Type aliases and/or sizes for ATTR_PROC_MASTER_TYPE attribute
typedef uint8_t PROC_MASTER_TYPE_ATTR;
typedef uint8_t ATTR_PROC_MASTER_TYPE_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_IOP_CONFIG attribute
typedef uint8_t PROC_PCIE_IOP_CONFIG_ATTR;
typedef uint8_t ATTR_PROC_PCIE_IOP_CONFIG_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_IOVALID_ENABLE attribute
typedef uint8_t PROC_PCIE_IOVALID_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_PCIE_IOVALID_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_IS_SLOT attribute
typedef uint8_t PROC_PCIE_IS_SLOT_ATTR[4];
typedef uint8_t ATTR_PROC_PCIE_IS_SLOT_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 4> ATTR_PROC_PCIE_IS_SLOT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN3 attribute
typedef uint16_t PROC_PCIE_LANE_EQUALIZATION_GEN3_ATTR[16];
typedef uint16_t ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN3_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 16> ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN3_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN4 attribute
typedef uint16_t PROC_PCIE_LANE_EQUALIZATION_GEN4_ATTR[16];
typedef uint16_t ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN4_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 16> ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN4_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_LANE_MASK attribute
typedef uint16_t PROC_PCIE_LANE_MASK_ATTR[4];
typedef uint16_t ATTR_PROC_PCIE_LANE_MASK_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 4> ATTR_PROC_PCIE_LANE_MASK_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_NUM_PEC attribute
typedef uint8_t PROC_PCIE_NUM_PEC_ATTR;
typedef uint8_t ATTR_PROC_PCIE_NUM_PEC_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_LOFF_CONTROL attribute
typedef uint16_t PROC_PCIE_PCS_RX_LOFF_CONTROL_ATTR[4];
typedef uint16_t ATTR_PROC_PCIE_PCS_RX_LOFF_CONTROL_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 4> ATTR_PROC_PCIE_PCS_RX_LOFF_CONTROL_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_PHASE_ROTATOR_CNTL attribute
typedef uint16_t PROC_PCIE_PCS_RX_PHASE_ROTATOR_CNTL_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_RX_PHASE_ROTATOR_CNTL_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_SIGDET_CNTL attribute
typedef uint16_t PROC_PCIE_PCS_RX_SIGDET_CNTL_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_RX_SIGDET_CNTL_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3 attribute
typedef uint16_t PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3_ATTR[4];
typedef uint16_t ATTR_PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3_type[4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 4> ATTR_PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_SYSTEM_CNTL attribute
typedef uint16_t PROC_PCIE_PCS_SYSTEM_CNTL_ATTR;
typedef uint16_t ATTR_PROC_PCIE_PCS_SYSTEM_CNTL_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PCS_TX_FIFO_CONFIG_OFFSET attribute
typedef uint8_t PROC_PCIE_PCS_TX_FIFO_CONFIG_OFFSET_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PCS_TX_FIFO_CONFIG_OFFSET_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_PHB_ACTIVE attribute
typedef uint8_t PROC_PCIE_PHB_ACTIVE_ATTR;
typedef uint8_t ATTR_PROC_PCIE_PHB_ACTIVE_type;

// Type aliases and/or sizes for ATTR_PROC_PCIE_REFCLOCK_ENABLE attribute
typedef uint8_t PROC_PCIE_REFCLOCK_ENABLE_ATTR;
typedef uint8_t ATTR_PROC_PCIE_REFCLOCK_ENABLE_type;

// Type aliases and/or sizes for ATTR_PROC_REFCLOCK_RCVR_TERM attribute
typedef uint8_t PROC_REFCLOCK_RCVR_TERM_ATTR;
typedef uint8_t ATTR_PROC_REFCLOCK_RCVR_TERM_type;

// Type aliases and/or sizes for ATTR_PROC_SELECT_BOOT_SEEPROM_IMAGE attribute
typedef uint8_t PROC_SELECT_BOOT_SEEPROM_IMAGE_ATTR;
typedef uint8_t ATTR_PROC_SELECT_BOOT_SEEPROM_IMAGE_type;

// Type aliases and/or sizes for ATTR_PROC_SELECT_SEEPROM_IMAGE attribute
typedef uint8_t PROC_SELECT_SEEPROM_IMAGE_ATTR;
typedef uint8_t ATTR_PROC_SELECT_SEEPROM_IMAGE_type;

// Type aliases and/or sizes for ATTR_PSI_BRIDGE_BASE_ADDR attribute
typedef uint64_t PSI_BRIDGE_BASE_ADDR_ATTR;
typedef uint64_t ATTR_PSI_BRIDGE_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_PSI_HB_ESB_ADDR attribute
typedef uint64_t PSI_HB_ESB_ADDR_ATTR;
typedef uint64_t ATTR_PSI_HB_ESB_ADDR_type;

// Type aliases and/or sizes for ATTR_RAW_MTM attribute
typedef char RAW_MTM_ATTR[64];
typedef char ATTR_RAW_MTM_type[64];
const size_t ATTR_RAW_MTM_max_chars = 63;

// Type aliases and/or sizes for ATTR_RCD_PARITY_RECONFIG_LOOPS_ALLOWED attribute
typedef uint8_t RCD_PARITY_RECONFIG_LOOPS_ALLOWED_ATTR;
typedef uint8_t ATTR_RCD_PARITY_RECONFIG_LOOPS_ALLOWED_type;

// Type aliases and/or sizes for ATTR_RCD_PARITY_RECONFIG_LOOP_COUNT attribute
typedef uint8_t RCD_PARITY_RECONFIG_LOOP_COUNT_ATTR;
typedef uint8_t ATTR_RCD_PARITY_RECONFIG_LOOP_COUNT_type;

// Type aliases and/or sizes for ATTR_RECONFIG_LOOP_TESTS attribute
typedef uint64_t RECONFIG_LOOP_TESTS_ATTR[5];
typedef uint64_t ATTR_RECONFIG_LOOP_TESTS_type[5];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 5> ATTR_RECONFIG_LOOP_TESTS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_RECONFIG_LOOP_TESTS_ENABLE attribute
typedef uint8_t RECONFIG_LOOP_TESTS_ENABLE_ATTR;
typedef uint8_t ATTR_RECONFIG_LOOP_TESTS_ENABLE_type;

// Type aliases and/or sizes for ATTR_REDUNDANT_CLOCKS attribute
typedef uint8_t REDUNDANT_CLOCKS_ATTR;
typedef uint8_t ATTR_REDUNDANT_CLOCKS_type;

// Type aliases and/or sizes for ATTR_REL_POS attribute
typedef uint8_t REL_POS_ATTR;
typedef uint8_t ATTR_REL_POS_type;

// Type aliases and/or sizes for ATTR_RESOURCE_IS_CRITICAL attribute
typedef uint8_t RESOURCE_IS_CRITICAL_ATTR;
typedef uint8_t ATTR_RESOURCE_IS_CRITICAL_type;

// Type aliases and/or sizes for ATTR_RISK_LEVEL_ORIGIN attribute
typedef RISK_LEVEL_ORIGIN RISK_LEVEL_ORIGIN_ATTR;
typedef RISK_LEVEL_ORIGIN ATTR_RISK_LEVEL_ORIGIN_type;

// Type aliases and/or sizes for ATTR_RNG_BAR_SIZE attribute
typedef uint64_t RNG_BAR_SIZE_ATTR;
typedef uint64_t ATTR_RNG_BAR_SIZE_type;

// Type aliases and/or sizes for ATTR_RNG_BASE_ADDR attribute
typedef uint64_t RNG_BASE_ADDR_ATTR;
typedef uint64_t ATTR_RNG_BASE_ADDR_type;

// Type aliases and/or sizes for ATTR_RUN_MAX_MEM_PATTERNS attribute
typedef uint8_t RUN_MAX_MEM_PATTERNS_ATTR;
typedef uint8_t ATTR_RUN_MAX_MEM_PATTERNS_type;

// Type aliases and/or sizes for ATTR_SAVED_STATE_UINT8 attribute
typedef uint8_t SAVED_STATE_UINT8_ATTR;
typedef uint8_t ATTR_SAVED_STATE_UINT8_type;

// Type aliases and/or sizes for ATTR_SBE_COMMIT_ID attribute
typedef uint32_t SBE_COMMIT_ID_ATTR;
typedef uint32_t ATTR_SBE_COMMIT_ID_type;

// Type aliases and/or sizes for ATTR_SBE_IS_STARTED attribute
typedef uint8_t SBE_IS_STARTED_ATTR;
typedef uint8_t ATTR_SBE_IS_STARTED_type;

// Type aliases and/or sizes for ATTR_SBE_RELEASE_TAG attribute
typedef char SBE_RELEASE_TAG_ATTR[21];
typedef char ATTR_SBE_RELEASE_TAG_type[21];
const size_t ATTR_SBE_RELEASE_TAG_max_chars = 20;

// Type aliases and/or sizes for ATTR_SBE_UPDATE_DISABLE attribute
typedef uint8_t SBE_UPDATE_DISABLE_ATTR;
typedef uint8_t ATTR_SBE_UPDATE_DISABLE_type;

// Type aliases and/or sizes for ATTR_SBE_VERSION_INFO attribute
typedef uint32_t SBE_VERSION_INFO_ATTR;
typedef uint32_t ATTR_SBE_VERSION_INFO_type;

// Type aliases and/or sizes for ATTR_SCOM_SWITCHES attribute
typedef ScomSwitches SCOM_SWITCHES_ATTR;
typedef ScomSwitches ATTR_SCOM_SWITCHES_type;

// Type aliases and/or sizes for ATTR_SECUREBOOT_PROTECT_DECONFIGURED_TPM attribute
typedef uint8_t SECUREBOOT_PROTECT_DECONFIGURED_TPM_ATTR;
typedef uint8_t ATTR_SECUREBOOT_PROTECT_DECONFIGURED_TPM_type;

// Type aliases and/or sizes for ATTR_SLOT_INDEX attribute
typedef uint8_t SLOT_INDEX_ATTR;
typedef uint8_t ATTR_SLOT_INDEX_type;

// Type aliases and/or sizes for ATTR_SLOT_NAME attribute
typedef uint8_t SLOT_NAME_ATTR;
typedef uint8_t ATTR_SLOT_NAME_type;

// Type aliases and/or sizes for ATTR_SOCKET_POWER_NOMINAL attribute
typedef uint32_t SOCKET_POWER_NOMINAL_ATTR;
typedef uint32_t ATTR_SOCKET_POWER_NOMINAL_type;

// Type aliases and/or sizes for ATTR_SOCKET_POWER_TURBO attribute
typedef uint32_t SOCKET_POWER_TURBO_ATTR;
typedef uint32_t ATTR_SOCKET_POWER_TURBO_type;

// Type aliases and/or sizes for ATTR_SP_FUNCTIONS attribute
typedef SpFunctions SP_FUNCTIONS_ATTR;
typedef SpFunctions ATTR_SP_FUNCTIONS_type;

// Type aliases and/or sizes for ATTR_START_MEM_ADDRESS_FOR_PAYLOAD_TCE_TOKEN attribute
typedef uint64_t START_MEM_ADDRESS_FOR_PAYLOAD_TCE_TOKEN_ATTR;
typedef uint64_t ATTR_START_MEM_ADDRESS_FOR_PAYLOAD_TCE_TOKEN_type;

// Type aliases and/or sizes for ATTR_STOP5_DISABLE attribute
typedef uint8_t STOP5_DISABLE_ATTR;
typedef uint8_t ATTR_STOP5_DISABLE_type;

// Type aliases and/or sizes for ATTR_SUPPORTED_STOP_STATES attribute
typedef uint32_t SUPPORTED_STOP_STATES_ATTR;
typedef uint32_t ATTR_SUPPORTED_STOP_STATES_type;

// Type aliases and/or sizes for ATTR_SUPPORTS_DYNAMIC_MEM_VOLT attribute
typedef uint8_t SUPPORTS_DYNAMIC_MEM_VOLT_ATTR;
typedef uint8_t ATTR_SUPPORTS_DYNAMIC_MEM_VOLT_type;

// Type aliases and/or sizes for ATTR_SYNC_BETWEEN_STEPS attribute
typedef uint8_t SYNC_BETWEEN_STEPS_ATTR;
typedef uint8_t ATTR_SYNC_BETWEEN_STEPS_type;

// Type aliases and/or sizes for ATTR_SYSTEM_FAMILY attribute
typedef char SYSTEM_FAMILY_ATTR[64];
typedef char ATTR_SYSTEM_FAMILY_type[64];
const size_t ATTR_SYSTEM_FAMILY_max_chars = 63;

// Type aliases and/or sizes for ATTR_SYSTEM_TYPE attribute
typedef char SYSTEM_TYPE_ATTR[64];
typedef char ATTR_SYSTEM_TYPE_type[64];
const size_t ATTR_SYSTEM_TYPE_max_chars = 63;

// Type aliases and/or sizes for ATTR_TCE_START_TOKEN_FOR_HDAT attribute
typedef uint32_t TCE_START_TOKEN_FOR_HDAT_ATTR;
typedef uint32_t ATTR_TCE_START_TOKEN_FOR_HDAT_type;

// Type aliases and/or sizes for ATTR_TCE_START_TOKEN_FOR_PAYLOAD attribute
typedef uint32_t TCE_START_TOKEN_FOR_PAYLOAD_ATTR;
typedef uint32_t ATTR_TCE_START_TOKEN_FOR_PAYLOAD_type;

// Type aliases and/or sizes for ATTR_TEMP_SENSOR_I2C_CONFIG attribute
typedef TempSensorI2cConfig TEMP_SENSOR_I2C_CONFIG_ATTR;
typedef TempSensorI2cConfig ATTR_TEMP_SENSOR_I2C_CONFIG_type;

// Type aliases and/or sizes for ATTR_TEST_MAX_STRING attribute
typedef char TEST_MAX_STRING_ATTR[4];
typedef char ATTR_TEST_MAX_STRING_type[4];
const size_t ATTR_TEST_MAX_STRING_max_chars = 3;

// Type aliases and/or sizes for ATTR_TEST_MIN_STRING attribute
typedef char TEST_MIN_STRING_ATTR[10];
typedef char ATTR_TEST_MIN_STRING_type[10];
const size_t ATTR_TEST_MIN_STRING_max_chars = 9;

// Type aliases and/or sizes for ATTR_TEST_NEGATIVE_FCN attribute
typedef int8_t TEST_NEGATIVE_FCN_ATTR;
typedef int8_t ATTR_TEST_NEGATIVE_FCN_type;

// Type aliases and/or sizes for ATTR_TEST_NO_DEFAULT_STRING attribute
typedef char TEST_NO_DEFAULT_STRING_ATTR[10];
typedef char ATTR_TEST_NO_DEFAULT_STRING_type[10];
const size_t ATTR_TEST_NO_DEFAULT_STRING_max_chars = 9;

// Type aliases and/or sizes for ATTR_TEST_NULL_STRING attribute
typedef char TEST_NULL_STRING_ATTR[10];
typedef char ATTR_TEST_NULL_STRING_type[10];
const size_t ATTR_TEST_NULL_STRING_max_chars = 9;

// Type aliases and/or sizes for ATTR_THREAD_COUNT attribute
typedef uint32_t THREAD_COUNT_ATTR;
typedef uint32_t ATTR_THREAD_COUNT_type;

// Type aliases and/or sizes for ATTR_TIME_BASE attribute
typedef uint32_t TIME_BASE_ATTR;
typedef uint32_t ATTR_TIME_BASE_type;

// Type aliases and/or sizes for ATTR_TLB_DATA_ASSOC_SETS attribute
typedef uint32_t TLB_DATA_ASSOC_SETS_ATTR;
typedef uint32_t ATTR_TLB_DATA_ASSOC_SETS_type;

// Type aliases and/or sizes for ATTR_TLB_DATA_ENTRIES attribute
typedef uint32_t TLB_DATA_ENTRIES_ATTR;
typedef uint32_t ATTR_TLB_DATA_ENTRIES_type;

// Type aliases and/or sizes for ATTR_TLB_INSTR_ASSOC_SETS attribute
typedef uint32_t TLB_INSTR_ASSOC_SETS_ATTR;
typedef uint32_t ATTR_TLB_INSTR_ASSOC_SETS_type;

// Type aliases and/or sizes for ATTR_TLB_INSTR_ENTRIES attribute
typedef uint32_t TLB_INSTR_ENTRIES_ATTR;
typedef uint32_t ATTR_TLB_INSTR_ENTRIES_type;

// Type aliases and/or sizes for ATTR_TLB_RESERVE_SIZE attribute
typedef uint32_t TLB_RESERVE_SIZE_ATTR;
typedef uint32_t ATTR_TLB_RESERVE_SIZE_type;

// Type aliases and/or sizes for ATTR_TOD_CPU_DATA attribute
typedef uint8_t TOD_CPU_DATA_ATTR[44];
typedef uint8_t ATTR_TOD_CPU_DATA_type[44];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 44> ATTR_TOD_CPU_DATA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_TOD_ROLE attribute
typedef uint8_t TOD_ROLE_ATTR;
typedef uint8_t ATTR_TOD_ROLE_type;

// Type aliases and/or sizes for ATTR_TPM_INFO attribute
typedef TpmInfo TPM_INFO_ATTR;
typedef TpmInfo ATTR_TPM_INFO_type;

// Type aliases and/or sizes for ATTR_TPM_POISONED attribute
typedef uint8_t TPM_POISONED_ATTR;
typedef uint8_t ATTR_TPM_POISONED_type;

// Type aliases and/or sizes for ATTR_TPM_REQUIRED attribute
typedef uint8_t TPM_REQUIRED_ATTR;
typedef uint8_t ATTR_TPM_REQUIRED_type;

// Type aliases and/or sizes for ATTR_TPM_UNUSABLE attribute
typedef uint8_t TPM_UNUSABLE_ATTR;
typedef uint8_t ATTR_TPM_UNUSABLE_type;

// Type aliases and/or sizes for ATTR_TYPE attribute
typedef TYPE TYPE_ATTR;
typedef TYPE ATTR_TYPE_type;

// Type aliases and/or sizes for ATTR_UCD_MFR_REVISION_OVERRIDE attribute
typedef uint16_t UCD_MFR_REVISION_OVERRIDE_ATTR;
typedef uint16_t ATTR_UCD_MFR_REVISION_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_ULTRA_TURBO_FREQ_MHZ attribute
typedef uint32_t ULTRA_TURBO_FREQ_MHZ_ATTR;
typedef uint32_t ATTR_ULTRA_TURBO_FREQ_MHZ_type;

// Type aliases and/or sizes for ATTR_USE_TCES_FOR_DMAS attribute
typedef uint8_t USE_TCES_FOR_DMAS_ATTR;
typedef uint8_t ATTR_USE_TCES_FOR_DMAS_type;

// Type aliases and/or sizes for ATTR_VCS_ID attribute
typedef uint16_t VCS_ID_ATTR;
typedef uint16_t ATTR_VCS_ID_type;

// Type aliases and/or sizes for ATTR_VDDR_ID attribute
typedef uint16_t VDDR_ID_ATTR;
typedef uint16_t ATTR_VDDR_ID_type;

// Type aliases and/or sizes for ATTR_VDD_ID attribute
typedef uint16_t VDD_ID_ATTR;
typedef uint16_t ATTR_VDD_ID_type;

// Type aliases and/or sizes for ATTR_VENDOR_ID attribute
typedef uint8_t VENDOR_ID_ATTR;
typedef uint8_t ATTR_VENDOR_ID_type;

// Type aliases and/or sizes for ATTR_VPD_REC_NUM attribute
typedef uint16_t VPD_REC_NUM_ATTR;
typedef uint16_t ATTR_VPD_REC_NUM_type;

// Type aliases and/or sizes for ATTR_VPP_ID attribute
typedef uint16_t VPP_ID_ATTR;
typedef uint16_t ATTR_VPP_ID_type;

// Type aliases and/or sizes for ATTR_WOF_POWER_LIMIT attribute
typedef uint8_t WOF_POWER_LIMIT_ATTR;
typedef uint8_t ATTR_WOF_POWER_LIMIT_type;

// Type aliases and/or sizes for ATTR_WOF_TABLE_LID_NUMBER attribute
typedef uint32_t WOF_TABLE_LID_NUMBER_ATTR;
typedef uint32_t ATTR_WOF_TABLE_LID_NUMBER_type;

// Type aliases and/or sizes for ATTR_XIVE_CONTROLLER_BAR_ADDR attribute
typedef uint64_t XIVE_CONTROLLER_BAR_ADDR_ATTR;
typedef uint64_t ATTR_XIVE_CONTROLLER_BAR_ADDR_type;

// Type aliases and/or sizes for ATTR_XIVE_THREAD_MGMT1_BAR_ADDR attribute
typedef uint64_t XIVE_THREAD_MGMT1_BAR_ADDR_ATTR;
typedef uint64_t ATTR_XIVE_THREAD_MGMT1_BAR_ADDR_type;

// Type aliases and/or sizes for ATTR_XSCOM_BASE_ADDRESS attribute
typedef uint64_t XSCOM_BASE_ADDRESS_ATTR;
typedef uint64_t ATTR_XSCOM_BASE_ADDRESS_type;

// Type aliases and/or sizes for ATTR_ALLOW_ATTR_OVERRIDES_IN_SECURE_MODE attribute
typedef uint8_t ALLOW_ATTR_OVERRIDES_IN_SECURE_MODE_ATTR;
typedef uint8_t ATTR_ALLOW_ATTR_OVERRIDES_IN_SECURE_MODE_type;

// Type aliases and/or sizes for ATTR_ASSUME_SBE_QUIESCED attribute
typedef uint8_t ASSUME_SBE_QUIESCED_ATTR;
typedef uint8_t ATTR_ASSUME_SBE_QUIESCED_type;

// Type aliases and/or sizes for ATTR_ATTN_CHK_ALL_PROCS attribute
typedef uint8_t ATTN_CHK_ALL_PROCS_ATTR;
typedef uint8_t ATTR_ATTN_CHK_ALL_PROCS_type;

// Type aliases and/or sizes for ATTR_CENTAUR_REGISTER_CACHE_PTR attribute
typedef uint64_t CENTAUR_REGISTER_CACHE_PTR_ATTR;
typedef uint64_t ATTR_CENTAUR_REGISTER_CACHE_PTR_type;

// Type aliases and/or sizes for ATTR_CHKSTOP_ANALYSIS_ON_STARTUP attribute
typedef uint8_t CHKSTOP_ANALYSIS_ON_STARTUP_ATTR;
typedef uint8_t ATTR_CHKSTOP_ANALYSIS_ON_STARTUP_type;

// Type aliases and/or sizes for ATTR_CLEAR_DIMM_SPD_ENABLE attribute
typedef uint8_t CLEAR_DIMM_SPD_ENABLE_ATTR;
typedef uint8_t ATTR_CLEAR_DIMM_SPD_ENABLE_type;

// Type aliases and/or sizes for ATTR_DIMM_SPD_BYTE_SIZE attribute
typedef uint32_t DIMM_SPD_BYTE_SIZE_ATTR;
typedef uint32_t ATTR_DIMM_SPD_BYTE_SIZE_type;

// Type aliases and/or sizes for ATTR_DRTM_PAYLOAD_ADDR_MB_HB attribute
typedef uint32_t DRTM_PAYLOAD_ADDR_MB_HB_ATTR;
typedef uint32_t ATTR_DRTM_PAYLOAD_ADDR_MB_HB_type;

// Type aliases and/or sizes for ATTR_DYNAMIC_I2C_DEVICE_ADDRESS attribute
typedef uint8_t DYNAMIC_I2C_DEVICE_ADDRESS_ATTR;
typedef uint8_t ATTR_DYNAMIC_I2C_DEVICE_ADDRESS_type;

// Type aliases and/or sizes for ATTR_EARLY_TESTCASES_ISTEP attribute
typedef uint16_t EARLY_TESTCASES_ISTEP_ATTR;
typedef uint16_t ATTR_EARLY_TESTCASES_ISTEP_type;

// Type aliases and/or sizes for ATTR_EEPROM_PAGE_ARRAY attribute
typedef uint8_t EEPROM_PAGE_ARRAY_ATTR[4][4];
typedef uint8_t ATTR_EEPROM_PAGE_ARRAY_type[4][4];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint8_t, 4>, 4> ATTR_EEPROM_PAGE_ARRAY_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_FORCE_PRE_PAYLOAD_DRTM attribute
typedef uint8_t FORCE_PRE_PAYLOAD_DRTM_ATTR;
typedef uint8_t ATTR_FORCE_PRE_PAYLOAD_DRTM_type;

// Type aliases and/or sizes for ATTR_FORCE_SBE_UPDATE attribute
typedef FORCE_SBE_UPDATE FORCE_SBE_UPDATE_ATTR;
typedef FORCE_SBE_UPDATE ATTR_FORCE_SBE_UPDATE_type;

// Type aliases and/or sizes for ATTR_FREQ_MCA_MHZ attribute
typedef uint32_t FREQ_MCA_MHZ_ATTR;
typedef uint32_t ATTR_FREQ_MCA_MHZ_type;

// Type aliases and/or sizes for ATTR_FSI_MASTER_MUTEX attribute
typedef mutex_t* FSI_MASTER_MUTEX_ATTR;
typedef mutex_t* ATTR_FSI_MASTER_MUTEX_type;

// Type aliases and/or sizes for ATTR_FSI_SCOM_MUTEX attribute
typedef mutex_t* FSI_SCOM_MUTEX_ATTR;
typedef mutex_t* ATTR_FSI_SCOM_MUTEX_type;

// Type aliases and/or sizes for ATTR_GPIO_INFO_PHYS_PRES attribute
typedef GpioInfoPhysPres GPIO_INFO_PHYS_PRES_ATTR;
typedef GpioInfoPhysPres ATTR_GPIO_INFO_PHYS_PRES_type;

// Type aliases and/or sizes for ATTR_HBRT_HYP_ID attribute
typedef uint64_t HBRT_HYP_ID_ATTR;
typedef uint64_t ATTR_HBRT_HYP_ID_type;

// Type aliases and/or sizes for ATTR_HB_EXISTING_IMAGE attribute
typedef uint8_t HB_EXISTING_IMAGE_ATTR;
typedef uint8_t ATTR_HB_EXISTING_IMAGE_type;

// Type aliases and/or sizes for ATTR_HB_INITIATED_PM_RESET attribute
typedef uint8_t HB_INITIATED_PM_RESET_ATTR;
typedef uint8_t ATTR_HB_INITIATED_PM_RESET_type;

// Type aliases and/or sizes for ATTR_HB_MUTEX_SERIALIZE_TEST_LOCK attribute
typedef mutex_t* HB_MUTEX_SERIALIZE_TEST_LOCK_ATTR;
typedef mutex_t* ATTR_HB_MUTEX_SERIALIZE_TEST_LOCK_type;

// Type aliases and/or sizes for ATTR_HB_MUTEX_TEST_LOCK attribute
typedef mutex_t* HB_MUTEX_TEST_LOCK_ATTR;
typedef mutex_t* ATTR_HB_MUTEX_TEST_LOCK_type;

// Type aliases and/or sizes for ATTR_HB_NODE_COMM_ABUS_MUTEX attribute
typedef mutex_t* HB_NODE_COMM_ABUS_MUTEX_ATTR;
typedef mutex_t* ATTR_HB_NODE_COMM_ABUS_MUTEX_type;

// Type aliases and/or sizes for ATTR_HB_NODE_COMM_XBUS_MUTEX attribute
typedef mutex_t* HB_NODE_COMM_XBUS_MUTEX_ATTR;
typedef mutex_t* ATTR_HB_NODE_COMM_XBUS_MUTEX_type;

// Type aliases and/or sizes for ATTR_HB_RECURSIVE_MUTEX_TEST_LOCK attribute
typedef mutex_t* HB_RECURSIVE_MUTEX_TEST_LOCK_ATTR;
typedef mutex_t* ATTR_HB_RECURSIVE_MUTEX_TEST_LOCK_type;

// Type aliases and/or sizes for ATTR_HB_RSV_MEM_NEXT_SECTION attribute
typedef uint8_t HB_RSV_MEM_NEXT_SECTION_ATTR;
typedef uint8_t ATTR_HB_RSV_MEM_NEXT_SECTION_type;

// Type aliases and/or sizes for ATTR_HB_SBE_SEEPROM_VERSION_MISMATCH attribute
typedef uint8_t HB_SBE_SEEPROM_VERSION_MISMATCH_ATTR;
typedef uint8_t ATTR_HB_SBE_SEEPROM_VERSION_MISMATCH_type;

// Type aliases and/or sizes for ATTR_HB_TARGET_SCOMABLE attribute
typedef uint8_t HB_TARGET_SCOMABLE_ATTR;
typedef uint8_t ATTR_HB_TARGET_SCOMABLE_type;

// Type aliases and/or sizes for ATTR_HB_TPM_INIT_ATTEMPTED attribute
typedef uint8_t HB_TPM_INIT_ATTEMPTED_ATTR;
typedef uint8_t ATTR_HB_TPM_INIT_ATTEMPTED_type;

// Type aliases and/or sizes for ATTR_HB_TPM_LOG_MGR_PTR attribute
typedef uint64_t HB_TPM_LOG_MGR_PTR_ATTR;
typedef uint64_t ATTR_HB_TPM_LOG_MGR_PTR_type;

// Type aliases and/or sizes for ATTR_HB_TPM_MUTEX attribute
typedef mutex_t* HB_TPM_MUTEX_ATTR;
typedef mutex_t* ATTR_HB_TPM_MUTEX_type;

// Type aliases and/or sizes for ATTR_HIDDEN_ERRLOGS_ENABLE attribute
typedef uint8_t HIDDEN_ERRLOGS_ENABLE_ATTR;
typedef uint8_t ATTR_HIDDEN_ERRLOGS_ENABLE_type;

// Type aliases and/or sizes for ATTR_HOMER_HCODE_LOADED attribute
typedef uint8_t HOMER_HCODE_LOADED_ATTR;
typedef uint8_t ATTR_HOMER_HCODE_LOADED_type;

// Type aliases and/or sizes for ATTR_HOMER_VIRT_ADDR attribute
typedef uint64_t HOMER_VIRT_ADDR_ATTR;
typedef uint64_t ATTR_HOMER_VIRT_ADDR_type;

// Type aliases and/or sizes for ATTR_I2C_ENGINE_MUTEX_0 attribute
typedef mutex_t* I2C_ENGINE_MUTEX_0_ATTR;
typedef mutex_t* ATTR_I2C_ENGINE_MUTEX_0_type;

// Type aliases and/or sizes for ATTR_I2C_ENGINE_MUTEX_1 attribute
typedef mutex_t* I2C_ENGINE_MUTEX_1_ATTR;
typedef mutex_t* ATTR_I2C_ENGINE_MUTEX_1_type;

// Type aliases and/or sizes for ATTR_I2C_ENGINE_MUTEX_2 attribute
typedef mutex_t* I2C_ENGINE_MUTEX_2_ATTR;
typedef mutex_t* ATTR_I2C_ENGINE_MUTEX_2_type;

// Type aliases and/or sizes for ATTR_I2C_ENGINE_MUTEX_3 attribute
typedef mutex_t* I2C_ENGINE_MUTEX_3_ATTR;
typedef mutex_t* ATTR_I2C_ENGINE_MUTEX_3_type;

// Type aliases and/or sizes for ATTR_I2C_PAGE_MUTEX_0 attribute
typedef mutex_t* I2C_PAGE_MUTEX_0_ATTR;
typedef mutex_t* ATTR_I2C_PAGE_MUTEX_0_type;

// Type aliases and/or sizes for ATTR_I2C_PAGE_MUTEX_1 attribute
typedef mutex_t* I2C_PAGE_MUTEX_1_ATTR;
typedef mutex_t* ATTR_I2C_PAGE_MUTEX_1_type;

// Type aliases and/or sizes for ATTR_I2C_PAGE_MUTEX_2 attribute
typedef mutex_t* I2C_PAGE_MUTEX_2_ATTR;
typedef mutex_t* ATTR_I2C_PAGE_MUTEX_2_type;

// Type aliases and/or sizes for ATTR_I2C_PAGE_MUTEX_3 attribute
typedef mutex_t* I2C_PAGE_MUTEX_3_ATTR;
typedef mutex_t* ATTR_I2C_PAGE_MUTEX_3_type;

// Type aliases and/or sizes for ATTR_IBSCOM_ENABLE_OVERRIDE attribute
typedef uint8_t IBSCOM_ENABLE_OVERRIDE_ATTR;
typedef uint8_t ATTR_IBSCOM_ENABLE_OVERRIDE_type;

// Type aliases and/or sizes for ATTR_IBSCOM_MUTEX attribute
typedef mutex_t* IBSCOM_MUTEX_ATTR;
typedef mutex_t* ATTR_IBSCOM_MUTEX_type;

// Type aliases and/or sizes for ATTR_IBSCOM_VIRTUAL_ADDR attribute
typedef uint64_t IBSCOM_VIRTUAL_ADDR_ATTR;
typedef uint64_t ATTR_IBSCOM_VIRTUAL_ADDR_type;

// Type aliases and/or sizes for ATTR_IPC_NODE_BUFFER_GLOBAL_ADDRESS attribute
typedef uint64_t IPC_NODE_BUFFER_GLOBAL_ADDRESS_ATTR[8];
typedef uint64_t ATTR_IPC_NODE_BUFFER_GLOBAL_ADDRESS_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint64_t, 8> ATTR_IPC_NODE_BUFFER_GLOBAL_ADDRESS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_IS_DRTM_MPIPL_HB attribute
typedef uint8_t IS_DRTM_MPIPL_HB_ATTR;
typedef uint8_t ATTR_IS_DRTM_MPIPL_HB_type;

// Type aliases and/or sizes for ATTR_IS_MPIPL_HB attribute
typedef uint8_t IS_MPIPL_HB_ATTR;
typedef uint8_t ATTR_IS_MPIPL_HB_type;

// Type aliases and/or sizes for ATTR_IS_SLAVE_DRAWER attribute
typedef uint8_t IS_SLAVE_DRAWER_ATTR;
typedef uint8_t ATTR_IS_SLAVE_DRAWER_type;

// Type aliases and/or sizes for ATTR_MASTER_MBOX_SCRATCH attribute
typedef uint32_t MASTER_MBOX_SCRATCH_ATTR[8];
typedef uint32_t ATTR_MASTER_MBOX_SCRATCH_type[8];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 8> ATTR_MASTER_MBOX_SCRATCH_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_MEMD_OFFSET attribute
typedef uint64_t MEMD_OFFSET_ATTR;
typedef uint64_t ATTR_MEMD_OFFSET_type;

// Type aliases and/or sizes for ATTR_MMIO_PHYS_ADDR attribute
typedef uint64_t MMIO_PHYS_ADDR_ATTR;
typedef uint64_t ATTR_MMIO_PHYS_ADDR_type;

// Type aliases and/or sizes for ATTR_MMIO_VM_ADDR attribute
typedef uint64_t MMIO_VM_ADDR_ATTR;
typedef uint64_t ATTR_MMIO_VM_ADDR_type;

// Type aliases and/or sizes for ATTR_MPIPL_HB_MDRT_COUNT attribute
typedef uint32_t MPIPL_HB_MDRT_COUNT_ATTR;
typedef uint32_t ATTR_MPIPL_HB_MDRT_COUNT_type;

// Type aliases and/or sizes for ATTR_MRW_DDR3_VDDR_MAX_LIMIT_EFF_CONFIG attribute
typedef uint32_t MRW_DDR3_VDDR_MAX_LIMIT_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MRW_DDR3_VDDR_MAX_LIMIT_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_MRW_DDR4_VDDR_MAX_LIMIT_EFF_CONFIG attribute
typedef uint32_t MRW_DDR4_VDDR_MAX_LIMIT_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MRW_DDR4_VDDR_MAX_LIMIT_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_EFF_CONFIG attribute
typedef uint32_t MSS_VOLT_DDR3_VDDR_INTERCEPT_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_EFF_CONFIG attribute
typedef uint32_t MSS_VOLT_DDR3_VDDR_SLOPE_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_EFF_CONFIG attribute
typedef uint32_t MSS_VOLT_DDR4_VDDR_INTERCEPT_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_EFF_CONFIG attribute
typedef uint32_t MSS_VOLT_DDR4_VDDR_SLOPE_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_INTERCEPT_EFF_CONFIG attribute
typedef uint32_t MSS_VOLT_VPP_INTERCEPT_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_INTERCEPT_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_MSS_VOLT_VPP_SLOPE_EFF_CONFIG attribute
typedef uint32_t MSS_VOLT_VPP_SLOPE_EFF_CONFIG_ATTR;
typedef uint32_t ATTR_MSS_VOLT_VPP_SLOPE_EFF_CONFIG_type;

// Type aliases and/or sizes for ATTR_NVDIMM_READING_PAGE4 attribute
typedef uint8_t NVDIMM_READING_PAGE4_ATTR;
typedef uint8_t ATTR_NVDIMM_READING_PAGE4_type;

// Type aliases and/or sizes for ATTR_NVDIMM_READING_VENDOR_LOG attribute
typedef uint8_t NVDIMM_READING_VENDOR_LOG_ATTR;
typedef uint8_t ATTR_NVDIMM_READING_VENDOR_LOG_type;

// Type aliases and/or sizes for ATTR_OCC_COMMON_AREA_PHYS_ADDR attribute
typedef uint64_t OCC_COMMON_AREA_PHYS_ADDR_ATTR;
typedef uint64_t ATTR_OCC_COMMON_AREA_PHYS_ADDR_type;

// Type aliases and/or sizes for ATTR_OCMB_COUNTER_HB attribute
typedef uint32_t OCMB_COUNTER_HB_ATTR;
typedef uint32_t ATTR_OCMB_COUNTER_HB_type;

// Type aliases and/or sizes for ATTR_OVERRIDES_ATTEMPTED_FLAG attribute
typedef uint8_t OVERRIDES_ATTEMPTED_FLAG_ATTR;
typedef uint8_t ATTR_OVERRIDES_ATTEMPTED_FLAG_type;

// Type aliases and/or sizes for ATTR_PART_NUMBER attribute
typedef uint8_t PART_NUMBER_ATTR[48];
typedef uint8_t ATTR_PART_NUMBER_type[48];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 48> ATTR_PART_NUMBER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_ADDR attribute
typedef uint64_t PDA_CAPTURED_THREAD_REG_ARRAY_ADDR_ATTR;
typedef uint64_t ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_ADDR_type;

// Type aliases and/or sizes for ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_SIZE attribute
typedef uint32_t PDA_CAPTURED_THREAD_REG_ARRAY_SIZE_ATTR;
typedef uint32_t ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_SIZE_type;

// Type aliases and/or sizes for ATTR_PDA_THREAD_REG_ENTRY_SIZE attribute
typedef uint32_t PDA_THREAD_REG_ENTRY_SIZE_ATTR;
typedef uint32_t ATTR_PDA_THREAD_REG_ENTRY_SIZE_type;

// Type aliases and/or sizes for ATTR_PDA_THREAD_REG_STATE_ENTRY_FORMAT attribute
typedef uint8_t PDA_THREAD_REG_STATE_ENTRY_FORMAT_ATTR;
typedef uint8_t ATTR_PDA_THREAD_REG_STATE_ENTRY_FORMAT_type;

// Type aliases and/or sizes for ATTR_PHYS_PRES_ASSERTED attribute
typedef uint8_t PHYS_PRES_ASSERTED_ATTR;
typedef uint8_t ATTR_PHYS_PRES_ASSERTED_type;

// Type aliases and/or sizes for ATTR_PHYS_PRES_FAKE_ASSERT attribute
typedef uint8_t PHYS_PRES_FAKE_ASSERT_ATTR;
typedef uint8_t ATTR_PHYS_PRES_FAKE_ASSERT_type;

// Type aliases and/or sizes for ATTR_PHYS_PRES_REIPL attribute
typedef uint8_t PHYS_PRES_REIPL_ATTR;
typedef uint8_t ATTR_PHYS_PRES_REIPL_type;

// Type aliases and/or sizes for ATTR_PHYS_PRES_REQUEST_OPEN_WINDOW attribute
typedef uint8_t PHYS_PRES_REQUEST_OPEN_WINDOW_ATTR;
typedef uint8_t ATTR_PHYS_PRES_REQUEST_OPEN_WINDOW_type;

// Type aliases and/or sizes for ATTR_SBE_ARCH_DUMP_ADDR attribute
typedef uint64_t SBE_ARCH_DUMP_ADDR_ATTR;
typedef uint64_t ATTR_SBE_ARCH_DUMP_ADDR_type;

// Type aliases and/or sizes for ATTR_SBE_COMM_ADDR attribute
typedef uint64_t SBE_COMM_ADDR_ATTR;
typedef uint64_t ATTR_SBE_COMM_ADDR_type;

// Type aliases and/or sizes for ATTR_SBE_FFDC_ADDR attribute
typedef uint64_t SBE_FFDC_ADDR_ATTR;
typedef uint64_t ATTR_SBE_FFDC_ADDR_type;

// Type aliases and/or sizes for ATTR_SCAN_MUTEX attribute
typedef mutex_t* SCAN_MUTEX_ATTR;
typedef mutex_t* ATTR_SCAN_MUTEX_type;

// Type aliases and/or sizes for ATTR_SCOM_IND_MUTEX attribute
typedef mutex_t* SCOM_IND_MUTEX_ATTR;
typedef mutex_t* ATTR_SCOM_IND_MUTEX_type;

// Type aliases and/or sizes for ATTR_SERIAL_NUMBER attribute
typedef uint8_t SERIAL_NUMBER_ATTR[18];
typedef uint8_t ATTR_SERIAL_NUMBER_type[18];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 18> ATTR_SERIAL_NUMBER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_SPCWKUP_COUNT attribute
typedef uint32_t SPCWKUP_COUNT_ATTR;
typedef uint32_t ATTR_SPCWKUP_COUNT_type;

// Type aliases and/or sizes for ATTR_TPM_MODEL attribute
typedef uint8_t TPM_MODEL_ATTR;
typedef uint8_t ATTR_TPM_MODEL_type;

// Type aliases and/or sizes for ATTR_TPM_ROLE attribute
typedef uint8_t TPM_ROLE_ATTR;
typedef uint8_t ATTR_TPM_ROLE_type;

// Type aliases and/or sizes for ATTR_VPD_SWITCHES attribute
typedef VpdSwitches VPD_SWITCHES_ATTR;
typedef VpdSwitches ATTR_VPD_SWITCHES_type;

// Type aliases and/or sizes for ATTR_WOF_FREQUENCY_UPLIFT_SELECTED attribute
typedef uint16_t WOF_FREQUENCY_UPLIFT_SELECTED_ATTR[22][13];
typedef uint16_t ATTR_WOF_FREQUENCY_UPLIFT_SELECTED_type[22][13];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 13>, 22> ATTR_WOF_FREQUENCY_UPLIFT_SELECTED_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_XSCOM_VIRTUAL_ADDR attribute
typedef uint64_t XSCOM_VIRTUAL_ADDR_ATTR;
typedef uint64_t ATTR_XSCOM_VIRTUAL_ADDR_type;

// Type aliases and/or sizes for ATTR_ASCII_VPD_LX_KEYWORD attribute
typedef uint64_t ASCII_VPD_LX_KEYWORD_ATTR;
typedef uint64_t ATTR_ASCII_VPD_LX_KEYWORD_type;

// Type aliases and/or sizes for ATTR_CHIP_VER attribute
typedef CHIP_VER CHIP_VER_ATTR;
typedef CHIP_VER ATTR_CHIP_VER_type;

// Type aliases and/or sizes for ATTR_CREATE_DEF_PARTITION attribute
typedef uint8_t CREATE_DEF_PARTITION_ATTR;
typedef uint8_t ATTR_CREATE_DEF_PARTITION_type;

// Type aliases and/or sizes for ATTR_ENLARGED_IO_SLOT_COUNT attribute
typedef uint8_t ENLARGED_IO_SLOT_COUNT_ATTR;
typedef uint8_t ATTR_ENLARGED_IO_SLOT_COUNT_type;

// Type aliases and/or sizes for ATTR_HUGE_PAGE_COUNT attribute
typedef uint16_t HUGE_PAGE_COUNT_ATTR;
typedef uint16_t ATTR_HUGE_PAGE_COUNT_type;

// Type aliases and/or sizes for ATTR_HUGE_PAGE_SIZE attribute
typedef uint8_t HUGE_PAGE_SIZE_ATTR;
typedef uint8_t ATTR_HUGE_PAGE_SIZE_type;

// Type aliases and/or sizes for ATTR_HW_VER attribute
typedef HW_VER HW_VER_ATTR;
typedef HW_VER ATTR_HW_VER_type;

// Type aliases and/or sizes for ATTR_HYPERVISOR_IPL_DESTINATION attribute
typedef uint8_t HYPERVISOR_IPL_DESTINATION_ATTR;
typedef uint8_t ATTR_HYPERVISOR_IPL_DESTINATION_type;

// Type aliases and/or sizes for ATTR_HYPERVISOR_IPL_SIDE attribute
typedef uint8_t HYPERVISOR_IPL_SIDE_ATTR;
typedef uint8_t ATTR_HYPERVISOR_IPL_SIDE_type;

// Type aliases and/or sizes for ATTR_HYP_DISPATCH_WHEEL attribute
typedef uint8_t HYP_DISPATCH_WHEEL_ATTR;
typedef uint8_t ATTR_HYP_DISPATCH_WHEEL_type;

// Type aliases and/or sizes for ATTR_IPL_ATTRIBUTES attribute
typedef IplAttributes IPL_ATTRIBUTES_ATTR;
typedef IplAttributes ATTR_IPL_ATTRIBUTES_type;

// Type aliases and/or sizes for ATTR_IPS_ENABLE attribute
typedef uint8_t IPS_ENABLE_ATTR;
typedef uint8_t ATTR_IPS_ENABLE_type;

// Type aliases and/or sizes for ATTR_IPS_ENTER_TIME_SECONDS attribute
typedef uint16_t IPS_ENTER_TIME_SECONDS_ATTR;
typedef uint16_t ATTR_IPS_ENTER_TIME_SECONDS_type;

// Type aliases and/or sizes for ATTR_IPS_ENTER_UTILIZATION_PERCENT attribute
typedef uint8_t IPS_ENTER_UTILIZATION_PERCENT_ATTR;
typedef uint8_t ATTR_IPS_ENTER_UTILIZATION_PERCENT_type;

// Type aliases and/or sizes for ATTR_IPS_EXIT_TIME_SECONDS attribute
typedef uint16_t IPS_EXIT_TIME_SECONDS_ATTR;
typedef uint16_t ATTR_IPS_EXIT_TIME_SECONDS_type;

// Type aliases and/or sizes for ATTR_IPS_EXIT_UTILIZATION_PERCENT attribute
typedef uint8_t IPS_EXIT_UTILIZATION_PERCENT_ATTR;
typedef uint8_t ATTR_IPS_EXIT_UTILIZATION_PERCENT_type;

// Type aliases and/or sizes for ATTR_LMB_SIZE attribute
typedef uint8_t LMB_SIZE_ATTR;
typedef uint8_t ATTR_LMB_SIZE_type;

// Type aliases and/or sizes for ATTR_MAX_HSL_OPTICONNECT_CONNECTIONS attribute
typedef uint8_t MAX_HSL_OPTICONNECT_CONNECTIONS_ATTR;
typedef uint8_t ATTR_MAX_HSL_OPTICONNECT_CONNECTIONS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_PM_MODE attribute
typedef uint8_t OPEN_POWER_PM_MODE_ATTR;
typedef uint8_t ATTR_OPEN_POWER_PM_MODE_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_PM_MODE_FREQ_PERCENT attribute
typedef uint16_t OPEN_POWER_PM_MODE_FREQ_PERCENT_ATTR;
typedef uint16_t ATTR_OPEN_POWER_PM_MODE_FREQ_PERCENT_type;

// Type aliases and/or sizes for ATTR_OS_IPL_MODE attribute
typedef uint8_t OS_IPL_MODE_ATTR;
typedef uint8_t ATTR_OS_IPL_MODE_type;

// Type aliases and/or sizes for ATTR_PHYP_SYSTEM_TYPE attribute
typedef uint32_t PHYP_SYSTEM_TYPE_ATTR;
typedef uint32_t ATTR_PHYP_SYSTEM_TYPE_type;

// Type aliases and/or sizes for ATTR_PRESERVE_MDC_PARTITION_VPD attribute
typedef uint8_t PRESERVE_MDC_PARTITION_VPD_ATTR;
typedef uint8_t ATTR_PRESERVE_MDC_PARTITION_VPD_type;

// Type aliases and/or sizes for ATTR_PROC_HW_TOPOLOGY attribute
typedef uint16_t PROC_HW_TOPOLOGY_ATTR;
typedef uint16_t ATTR_PROC_HW_TOPOLOGY_type;

// Type aliases and/or sizes for ATTR_ROLE attribute
typedef ROLE ROLE_ATTR;
typedef ROLE ATTR_ROLE_type;

// Type aliases and/or sizes for ATTR_SLCA_INDEX attribute
typedef uint32_t SLCA_INDEX_ATTR;
typedef uint32_t ATTR_SLCA_INDEX_type;

// Type aliases and/or sizes for ATTR_SLCA_RID attribute
typedef uint32_t SLCA_RID_ATTR;
typedef uint32_t ATTR_SLCA_RID_type;

// Type aliases and/or sizes for ATTR_SW_VER attribute
typedef SW_VER SW_VER_ATTR;
typedef SW_VER ATTR_SW_VER_type;

// Type aliases and/or sizes for ATTR_SYSTEM_BRAND_NAME attribute
typedef char SYSTEM_BRAND_NAME_ATTR[64];
typedef char ATTR_SYSTEM_BRAND_NAME_type[64];
const size_t ATTR_SYSTEM_BRAND_NAME_max_chars = 63;

// Type aliases and/or sizes for ATTR_VLAN_SWITCHES attribute
typedef uint8_t VLAN_SWITCHES_ATTR;
typedef uint8_t ATTR_VLAN_SWITCHES_type;

// Type aliases and/or sizes for ATTR_VTPM_ENABLED attribute
typedef uint8_t VTPM_ENABLED_ATTR;
typedef uint8_t ATTR_VTPM_ENABLED_type;

// Type aliases and/or sizes for ATTR_ADC_CHANNEL_FUNC_IDS attribute
typedef uint8_t ADC_CHANNEL_FUNC_IDS_ATTR[16];
typedef uint8_t ATTR_ADC_CHANNEL_FUNC_IDS_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 16> ATTR_ADC_CHANNEL_FUNC_IDS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_ADC_CHANNEL_GAINS attribute
typedef uint32_t ADC_CHANNEL_GAINS_ATTR[16];
typedef uint32_t ATTR_ADC_CHANNEL_GAINS_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 16> ATTR_ADC_CHANNEL_GAINS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_ADC_CHANNEL_GNDS attribute
typedef uint8_t ADC_CHANNEL_GNDS_ATTR[16];
typedef uint8_t ATTR_ADC_CHANNEL_GNDS_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 16> ATTR_ADC_CHANNEL_GNDS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_ADC_CHANNEL_OFFSETS attribute
typedef uint32_t ADC_CHANNEL_OFFSETS_ATTR[16];
typedef uint32_t ATTR_ADC_CHANNEL_OFFSETS_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 16> ATTR_ADC_CHANNEL_OFFSETS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_ADC_CHANNEL_SENSOR_NUMBERS attribute
typedef uint32_t ADC_CHANNEL_SENSOR_NUMBERS_ATTR[16];
typedef uint32_t ATTR_ADC_CHANNEL_SENSOR_NUMBERS_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 16> ATTR_ADC_CHANNEL_SENSOR_NUMBERS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_ALLOW_CALLHOME_ESELS_TO_BMC attribute
typedef uint8_t ALLOW_CALLHOME_ESELS_TO_BMC_ATTR;
typedef uint8_t ATTR_ALLOW_CALLHOME_ESELS_TO_BMC_type;

// Type aliases and/or sizes for ATTR_APSS_GPIO_PORT_MODES attribute
typedef uint8_t APSS_GPIO_PORT_MODES_ATTR[2];
typedef uint8_t ATTR_APSS_GPIO_PORT_MODES_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 2> ATTR_APSS_GPIO_PORT_MODES_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_APSS_GPIO_PORT_PINS attribute
typedef uint8_t APSS_GPIO_PORT_PINS_ATTR[16];
typedef uint8_t ATTR_APSS_GPIO_PORT_PINS_type[16];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 16> ATTR_APSS_GPIO_PORT_PINS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_BMC_HW_CHIP_TYPE attribute
typedef char BMC_HW_CHIP_TYPE_ATTR[32];
typedef char ATTR_BMC_HW_CHIP_TYPE_type[32];
const size_t ATTR_BMC_HW_CHIP_TYPE_max_chars = 31;

// Type aliases and/or sizes for ATTR_BMC_MANUFACTURER attribute
typedef char BMC_MANUFACTURER_ATTR[32];
typedef char ATTR_BMC_MANUFACTURER_type[32];
const size_t ATTR_BMC_MANUFACTURER_max_chars = 31;

// Type aliases and/or sizes for ATTR_BMC_MAX_ERROR_LOG_SIZE attribute
typedef uint32_t BMC_MAX_ERROR_LOG_SIZE_ATTR;
typedef uint32_t ATTR_BMC_MAX_ERROR_LOG_SIZE_type;

// Type aliases and/or sizes for ATTR_BMC_SW_TYPE attribute
typedef char BMC_SW_TYPE_ATTR[32];
typedef char ATTR_BMC_SW_TYPE_type[32];
const size_t ATTR_BMC_SW_TYPE_max_chars = 31;

// Type aliases and/or sizes for ATTR_CALCULATED_MAX_SYS_POWER_EXCLUDING_GPUS attribute
typedef uint16_t CALCULATED_MAX_SYS_POWER_EXCLUDING_GPUS_ATTR;
typedef uint16_t ATTR_CALCULATED_MAX_SYS_POWER_EXCLUDING_GPUS_type;

// Type aliases and/or sizes for ATTR_CALCULATED_PROC_MEMORY_POWER_DROP attribute
typedef uint16_t CALCULATED_PROC_MEMORY_POWER_DROP_ATTR;
typedef uint16_t ATTR_CALCULATED_PROC_MEMORY_POWER_DROP_type;

// Type aliases and/or sizes for ATTR_CUMULATIVE_PMCOMPLEX_RESET_COUNT attribute
typedef uint8_t CUMULATIVE_PMCOMPLEX_RESET_COUNT_ATTR;
typedef uint8_t ATTR_CUMULATIVE_PMCOMPLEX_RESET_COUNT_type;

// Type aliases and/or sizes for ATTR_GPIO_INFO attribute
typedef GpioInfo GPIO_INFO_ATTR;
typedef GpioInfo ATTR_GPIO_INFO_type;

// Type aliases and/or sizes for ATTR_GPU_SENSORS attribute
typedef uint16_t GPU_SENSORS_ATTR[3][7];
typedef uint16_t ATTR_GPU_SENSORS_type[3][7];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 7>, 3> ATTR_GPU_SENSORS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_I2C_ADDR attribute
typedef uint8_t HDAT_I2C_ADDR_ATTR[32];
typedef uint8_t ATTR_HDAT_I2C_ADDR_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 32> ATTR_HDAT_I2C_ADDR_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_I2C_BUS_FREQ attribute
typedef uint32_t HDAT_I2C_BUS_FREQ_ATTR[32];
typedef uint32_t ATTR_HDAT_I2C_BUS_FREQ_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 32> ATTR_HDAT_I2C_BUS_FREQ_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_I2C_DEVICE_LABEL attribute
typedef char HDAT_I2C_DEVICE_LABEL_ATTR[32][64];
typedef char ATTR_HDAT_I2C_DEVICE_LABEL_type[32][64];
#ifdef __HOSTBOOT_MODULE
typedef std::array<char, 32> ATTR_HDAT_I2C_DEVICE_LABEL_typeStdArr;
#endif
const size_t ATTR_HDAT_I2C_DEVICE_LABEL_max_chars = 63;

// Type aliases and/or sizes for ATTR_HDAT_I2C_DEVICE_PURPOSE attribute
typedef uint8_t HDAT_I2C_DEVICE_PURPOSE_ATTR[32];
typedef uint8_t ATTR_HDAT_I2C_DEVICE_PURPOSE_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 32> ATTR_HDAT_I2C_DEVICE_PURPOSE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_I2C_DEVICE_TYPE attribute
typedef uint8_t HDAT_I2C_DEVICE_TYPE_ATTR[32];
typedef uint8_t ATTR_HDAT_I2C_DEVICE_TYPE_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 32> ATTR_HDAT_I2C_DEVICE_TYPE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_I2C_ELEMENTS attribute
typedef uint8_t HDAT_I2C_ELEMENTS_ATTR;
typedef uint8_t ATTR_HDAT_I2C_ELEMENTS_type;

// Type aliases and/or sizes for ATTR_HDAT_I2C_ENGINE attribute
typedef uint8_t HDAT_I2C_ENGINE_ATTR[32];
typedef uint8_t ATTR_HDAT_I2C_ENGINE_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 32> ATTR_HDAT_I2C_ENGINE_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_I2C_MASTER_PORT attribute
typedef uint8_t HDAT_I2C_MASTER_PORT_ATTR[32];
typedef uint8_t ATTR_HDAT_I2C_MASTER_PORT_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 32> ATTR_HDAT_I2C_MASTER_PORT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HDAT_I2C_SLAVE_PORT attribute
typedef uint8_t HDAT_I2C_SLAVE_PORT_ATTR[32];
typedef uint8_t ATTR_HDAT_I2C_SLAVE_PORT_type[32];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 32> ATTR_HDAT_I2C_SLAVE_PORT_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_HTMGT_INTERNAL_FLAGS attribute
typedef uint32_t HTMGT_INTERNAL_FLAGS_ATTR;
typedef uint32_t ATTR_HTMGT_INTERNAL_FLAGS_type;

// Type aliases and/or sizes for ATTR_HTMGT_PMCOMPLEX_RESET_COUNT_TIMER attribute
typedef uint64_t HTMGT_PMCOMPLEX_RESET_COUNT_TIMER_ATTR;
typedef uint64_t ATTR_HTMGT_PMCOMPLEX_RESET_COUNT_TIMER_type;

// Type aliases and/or sizes for ATTR_HTMGT_SAFEMODE attribute
typedef uint8_t HTMGT_SAFEMODE_ATTR;
typedef uint8_t ATTR_HTMGT_SAFEMODE_type;

// Type aliases and/or sizes for ATTR_HTMGT_SAVED_POWER_LIMIT attribute
typedef uint32_t HTMGT_SAVED_POWER_LIMIT_ATTR;
typedef uint32_t ATTR_HTMGT_SAVED_POWER_LIMIT_type;

// Type aliases and/or sizes for ATTR_IPMI_INSTANCE attribute
typedef uint32_t IPMI_INSTANCE_ATTR;
typedef uint32_t ATTR_IPMI_INSTANCE_type;

// Type aliases and/or sizes for ATTR_IPMI_MAX_BUFFER_SIZE attribute
typedef uint32_t IPMI_MAX_BUFFER_SIZE_ATTR;
typedef uint32_t ATTR_IPMI_MAX_BUFFER_SIZE_type;

// Type aliases and/or sizes for ATTR_IPMI_SENSORS attribute
typedef uint16_t IPMI_SENSORS_ATTR[16][2];
typedef uint16_t ATTR_IPMI_SENSORS_type[16][2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<std::array<uint16_t, 2>, 16> ATTR_IPMI_SENSORS_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_IS_MPIPL_SUPPORTED attribute
typedef uint8_t IS_MPIPL_SUPPORTED_ATTR;
typedef uint8_t ATTR_IS_MPIPL_SUPPORTED_type;

// Type aliases and/or sizes for ATTR_MAX_VDD_CURRENT_READING attribute
typedef uint16_t MAX_VDD_CURRENT_READING_ATTR;
typedef uint16_t ATTR_MAX_VDD_CURRENT_READING_type;

// Type aliases and/or sizes for ATTR_MISC_SYSTEM_COMPONENTS_MAX_POWER_WATTS attribute
typedef uint16_t MISC_SYSTEM_COMPONENTS_MAX_POWER_WATTS_ATTR;
typedef uint16_t ATTR_MISC_SYSTEM_COMPONENTS_MAX_POWER_WATTS_type;

// Type aliases and/or sizes for ATTR_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
typedef uint32_t MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP_ATTR;
typedef uint32_t ATTR_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP_type;

// Type aliases and/or sizes for ATTR_MSS_UTIL_N_PER_MBA attribute
typedef uint32_t MSS_UTIL_N_PER_MBA_ATTR;
typedef uint32_t ATTR_MSS_UTIL_N_PER_MBA_type;

// Type aliases and/or sizes for ATTR_NO_APSS_PROC_POWER_VCS_VIO_WATTS attribute
typedef uint16_t NO_APSS_PROC_POWER_VCS_VIO_WATTS_ATTR;
typedef uint16_t ATTR_NO_APSS_PROC_POWER_VCS_VIO_WATTS_type;

// Type aliases and/or sizes for ATTR_N_PLUS_ONE_MEM_POWER attribute
typedef uint32_t N_PLUS_ONE_MEM_POWER_ATTR[2];
typedef uint32_t ATTR_N_PLUS_ONE_MEM_POWER_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_N_PLUS_ONE_MEM_POWER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_N_PLUS_ONE_N_PER_CHIP attribute
typedef uint16_t N_PLUS_ONE_N_PER_CHIP_ATTR[2];
typedef uint16_t ATTR_N_PLUS_ONE_N_PER_CHIP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_N_PLUS_ONE_N_PER_CHIP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_N_PLUS_ONE_N_PER_MBA attribute
typedef uint16_t N_PLUS_ONE_N_PER_MBA_ATTR[2];
typedef uint16_t ATTR_N_PLUS_ONE_N_PER_MBA_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_N_PLUS_ONE_N_PER_MBA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_OCC_CONTROL_DATA attribute
typedef uint8_t OCC_CONTROL_DATA_ATTR[256];
typedef uint8_t ATTR_OCC_CONTROL_DATA_type[256];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint8_t, 256> ATTR_OCC_CONTROL_DATA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_OPAL_MODEL attribute
typedef char OPAL_MODEL_ATTR[32];
typedef char ATTR_OPAL_MODEL_type[32];
const size_t ATTR_OPAL_MODEL_max_chars = 31;

// Type aliases and/or sizes for ATTR_OPEN_POWER_DIMM_ERROR_TEMP_DEG_C attribute
typedef uint64_t OPEN_POWER_DIMM_ERROR_TEMP_DEG_C_ATTR;
typedef uint64_t ATTR_OPEN_POWER_DIMM_ERROR_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_DIMM_READ_TIMEOUT_SEC attribute
typedef uint64_t OPEN_POWER_DIMM_READ_TIMEOUT_SEC_ATTR;
typedef uint64_t ATTR_OPEN_POWER_DIMM_READ_TIMEOUT_SEC_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_DIMM_THROTTLE_TEMP_DEG_C attribute
typedef uint64_t OPEN_POWER_DIMM_THROTTLE_TEMP_DEG_C_ATTR;
typedef uint64_t ATTR_OPEN_POWER_DIMM_THROTTLE_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_GPU_ERROR_TEMP_DEG_C attribute
typedef uint8_t OPEN_POWER_GPU_ERROR_TEMP_DEG_C_ATTR;
typedef uint8_t ATTR_OPEN_POWER_GPU_ERROR_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_GPU_MEM_ERROR_TEMP_DEG_C attribute
typedef uint8_t OPEN_POWER_GPU_MEM_ERROR_TEMP_DEG_C_ATTR;
typedef uint8_t ATTR_OPEN_POWER_GPU_MEM_ERROR_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_GPU_MEM_READ_TIMEOUT_SEC attribute
typedef uint8_t OPEN_POWER_GPU_MEM_READ_TIMEOUT_SEC_ATTR;
typedef uint8_t ATTR_OPEN_POWER_GPU_MEM_READ_TIMEOUT_SEC_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_GPU_READ_TIMEOUT_SEC attribute
typedef uint8_t OPEN_POWER_GPU_READ_TIMEOUT_SEC_ATTR;
typedef uint8_t ATTR_OPEN_POWER_GPU_READ_TIMEOUT_SEC_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_MEMCTRL_ERROR_TEMP_DEG_C attribute
typedef uint64_t OPEN_POWER_MEMCTRL_ERROR_TEMP_DEG_C_ATTR;
typedef uint64_t ATTR_OPEN_POWER_MEMCTRL_ERROR_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_MEMCTRL_READ_TIMEOUT_SEC attribute
typedef uint64_t OPEN_POWER_MEMCTRL_READ_TIMEOUT_SEC_ATTR;
typedef uint64_t ATTR_OPEN_POWER_MEMCTRL_READ_TIMEOUT_SEC_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_MEMCTRL_THROTTLE_TEMP_DEG_C attribute
typedef uint64_t OPEN_POWER_MEMCTRL_THROTTLE_TEMP_DEG_C_ATTR;
typedef uint64_t ATTR_OPEN_POWER_MEMCTRL_THROTTLE_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_POWER_CAP attribute
typedef uint8_t OPEN_POWER_MIN_MEM_UTILIZATION_POWER_CAP_ATTR;
typedef uint8_t ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_POWER_CAP_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_THROTTLING attribute
typedef uint8_t OPEN_POWER_MIN_MEM_UTILIZATION_THROTTLING_ATTR;
typedef uint8_t ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_THROTTLING_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_MIN_POWER_CAP_WATTS attribute
typedef uint64_t OPEN_POWER_MIN_POWER_CAP_WATTS_ATTR;
typedef uint64_t ATTR_OPEN_POWER_MIN_POWER_CAP_WATTS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_N_BULK_POWER_LIMIT_WATTS attribute
typedef uint64_t OPEN_POWER_N_BULK_POWER_LIMIT_WATTS_ATTR;
typedef uint64_t ATTR_OPEN_POWER_N_BULK_POWER_LIMIT_WATTS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_N_MAX_MEM_POWER_WATTS attribute
typedef uint64_t OPEN_POWER_N_MAX_MEM_POWER_WATTS_ATTR;
typedef uint64_t ATTR_OPEN_POWER_N_MAX_MEM_POWER_WATTS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_N_PLUS_ONE_BULK_POWER_LIMIT_WATTS attribute
typedef uint64_t OPEN_POWER_N_PLUS_ONE_BULK_POWER_LIMIT_WATTS_ATTR;
typedef uint64_t ATTR_OPEN_POWER_N_PLUS_ONE_BULK_POWER_LIMIT_WATTS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_N_PLUS_ONE_HPC_BULK_POWER_LIMIT_WATTS attribute
typedef uint64_t OPEN_POWER_N_PLUS_ONE_HPC_BULK_POWER_LIMIT_WATTS_ATTR;
typedef uint64_t ATTR_OPEN_POWER_N_PLUS_ONE_HPC_BULK_POWER_LIMIT_WATTS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_N_PLUS_ONE_MAX_MEM_POWER_WATTS attribute
typedef uint64_t OPEN_POWER_N_PLUS_ONE_MAX_MEM_POWER_WATTS_ATTR;
typedef uint64_t ATTR_OPEN_POWER_N_PLUS_ONE_MAX_MEM_POWER_WATTS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_PROC_DVFS_TEMP_DEG_C attribute
typedef uint64_t OPEN_POWER_PROC_DVFS_TEMP_DEG_C_ATTR;
typedef uint64_t ATTR_OPEN_POWER_PROC_DVFS_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_PROC_ERROR_TEMP_DEG_C attribute
typedef uint64_t OPEN_POWER_PROC_ERROR_TEMP_DEG_C_ATTR;
typedef uint64_t ATTR_OPEN_POWER_PROC_ERROR_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_PROC_READ_TIMEOUT_SEC attribute
typedef uint64_t OPEN_POWER_PROC_READ_TIMEOUT_SEC_ATTR;
typedef uint64_t ATTR_OPEN_POWER_PROC_READ_TIMEOUT_SEC_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_PROC_WEIGHT attribute
typedef uint64_t OPEN_POWER_PROC_WEIGHT_ATTR;
typedef uint64_t ATTR_OPEN_POWER_PROC_WEIGHT_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_QUAD_WEIGHT attribute
typedef uint64_t OPEN_POWER_QUAD_WEIGHT_ATTR;
typedef uint64_t ATTR_OPEN_POWER_QUAD_WEIGHT_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_REGULATOR_EFFICIENCY_FACTOR attribute
typedef uint64_t OPEN_POWER_REGULATOR_EFFICIENCY_FACTOR_ATTR;
typedef uint64_t ATTR_OPEN_POWER_REGULATOR_EFFICIENCY_FACTOR_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_SOFT_MIN_PCAP_WATTS attribute
typedef uint64_t OPEN_POWER_SOFT_MIN_PCAP_WATTS_ATTR;
typedef uint64_t ATTR_OPEN_POWER_SOFT_MIN_PCAP_WATTS_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_TURBO_MODE_SUPPORTED attribute
typedef uint8_t OPEN_POWER_TURBO_MODE_SUPPORTED_ATTR;
typedef uint8_t ATTR_OPEN_POWER_TURBO_MODE_SUPPORTED_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_VRM_READ_TIMEOUT_SEC attribute
typedef uint8_t OPEN_POWER_VRM_READ_TIMEOUT_SEC_ATTR;
typedef uint8_t ATTR_OPEN_POWER_VRM_READ_TIMEOUT_SEC_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_VRM_VDD_DVFS_TEMP_DEG_C attribute
typedef uint8_t OPEN_POWER_VRM_VDD_DVFS_TEMP_DEG_C_ATTR;
typedef uint8_t ATTR_OPEN_POWER_VRM_VDD_DVFS_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_VRM_VDD_ERROR_TEMP_DEG_C attribute
typedef uint8_t OPEN_POWER_VRM_VDD_ERROR_TEMP_DEG_C_ATTR;
typedef uint8_t ATTR_OPEN_POWER_VRM_VDD_ERROR_TEMP_DEG_C_type;

// Type aliases and/or sizes for ATTR_OPEN_POWER_VRM_VDD_READ_TIMEOUT_SEC attribute
typedef uint8_t OPEN_POWER_VRM_VDD_READ_TIMEOUT_SEC_ATTR;
typedef uint8_t ATTR_OPEN_POWER_VRM_VDD_READ_TIMEOUT_SEC_type;

// Type aliases and/or sizes for ATTR_OP_TRACE_LITE attribute
typedef uint8_t OP_TRACE_LITE_ATTR;
typedef uint8_t ATTR_OP_TRACE_LITE_type;

// Type aliases and/or sizes for ATTR_OT_MEM_POWER attribute
typedef uint32_t OT_MEM_POWER_ATTR[2];
typedef uint32_t ATTR_OT_MEM_POWER_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_OT_MEM_POWER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_OT_MIN_N_PER_MBA attribute
typedef uint16_t OT_MIN_N_PER_MBA_ATTR[2];
typedef uint16_t ATTR_OT_MIN_N_PER_MBA_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_OT_MIN_N_PER_MBA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_POWERCAP_MEM_POWER attribute
typedef uint32_t POWERCAP_MEM_POWER_ATTR[2];
typedef uint32_t ATTR_POWERCAP_MEM_POWER_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint32_t, 2> ATTR_POWERCAP_MEM_POWER_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_POWERCAP_N_PER_CHIP attribute
typedef uint16_t POWERCAP_N_PER_CHIP_ATTR[2];
typedef uint16_t ATTR_POWERCAP_N_PER_CHIP_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_POWERCAP_N_PER_CHIP_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_POWERCAP_N_PER_MBA attribute
typedef uint16_t POWERCAP_N_PER_MBA_ATTR[2];
typedef uint16_t ATTR_POWERCAP_N_PER_MBA_type[2];
#ifdef __HOSTBOOT_MODULE
typedef std::array<uint16_t, 2> ATTR_POWERCAP_N_PER_MBA_typeStdArr;
#endif

// Type aliases and/or sizes for ATTR_PROC_MHZ_PER_WATT attribute
typedef uint16_t PROC_MHZ_PER_WATT_ATTR;
typedef uint16_t ATTR_PROC_MHZ_PER_WATT_type;

// Type aliases and/or sizes for ATTR_PROC_SOCKET_POWER_WATTS attribute
typedef uint16_t PROC_SOCKET_POWER_WATTS_ATTR;
typedef uint16_t ATTR_PROC_SOCKET_POWER_WATTS_type;

// Type aliases and/or sizes for ATTR_REPORT_THROTTLE_BELOW_NOMINAL attribute
typedef uint8_t REPORT_THROTTLE_BELOW_NOMINAL_ATTR;
typedef uint8_t ATTR_REPORT_THROTTLE_BELOW_NOMINAL_type;

// Type aliases and/or sizes for ATTR_VDD_CURRENT_OVERFLOW_WORKAROUND_ENABLE attribute
typedef uint16_t VDD_CURRENT_OVERFLOW_WORKAROUND_ENABLE_ATTR;
typedef uint16_t ATTR_VDD_CURRENT_OVERFLOW_WORKAROUND_ENABLE_type;


/**
 *  @brief Function to return size of specified attribute
 *
 *  @param[in] i_attrId Attribute ID for attribute to look up size
 *  @return uint32_t Size of the attribute
 *
 *  @retval Size of the attribute if succeeded
 *  @retval 0 if failed
 *
 */
inline uint32_t attrSizeLookup(ATTRIBUTE_ID i_attrId)
{
    uint32_t l_attrSize = 0;

    switch(i_attrId) {
    // Get size for ATTR_SBE_IMAGE_MINIMUM_VALID_ECS attribute
    case ATTR_SBE_IMAGE_MINIMUM_VALID_ECS:
        l_attrSize = sizeof(ATTR_SBE_IMAGE_MINIMUM_VALID_ECS_type);
        break;

    // Get size for ATTR_EXTERNAL_VRM_STEPSIZE attribute
    case ATTR_EXTERNAL_VRM_STEPSIZE:
        l_attrSize = sizeof(ATTR_EXTERNAL_VRM_STEPSIZE_type);
        break;

    // Get size for ATTR_EXTERNAL_VRM_TRANSITION_START_NS attribute
    case ATTR_EXTERNAL_VRM_TRANSITION_START_NS:
        l_attrSize = sizeof(ATTR_EXTERNAL_VRM_TRANSITION_START_NS_type);
        break;

    // Get size for ATTR_EXTERNAL_VRM_TRANSITION_RATE_INC_UV_PER_US attribute
    case ATTR_EXTERNAL_VRM_TRANSITION_RATE_INC_UV_PER_US:
        l_attrSize = sizeof(ATTR_EXTERNAL_VRM_TRANSITION_RATE_INC_UV_PER_US_type);
        break;

    // Get size for ATTR_EXTERNAL_VRM_TRANSITION_RATE_DEC_UV_PER_US attribute
    case ATTR_EXTERNAL_VRM_TRANSITION_RATE_DEC_UV_PER_US:
        l_attrSize = sizeof(ATTR_EXTERNAL_VRM_TRANSITION_RATE_DEC_UV_PER_US_type);
        break;

    // Get size for ATTR_EXTERNAL_VRM_TRANSITION_STABILIZATION_TIME_NS attribute
    case ATTR_EXTERNAL_VRM_TRANSITION_STABILIZATION_TIME_NS:
        l_attrSize = sizeof(ATTR_EXTERNAL_VRM_TRANSITION_STABILIZATION_TIME_NS_type);
        break;

    // Get size for ATTR_SAFE_MODE_NOVDM_UPLIFT_MV attribute
    case ATTR_SAFE_MODE_NOVDM_UPLIFT_MV:
        l_attrSize = sizeof(ATTR_SAFE_MODE_NOVDM_UPLIFT_MV_type);
        break;

    // Get size for ATTR_SAFE_MODE_VOLTAGE_MV attribute
    case ATTR_SAFE_MODE_VOLTAGE_MV:
        l_attrSize = sizeof(ATTR_SAFE_MODE_VOLTAGE_MV_type);
        break;

    // Get size for ATTR_SAFE_MODE_FREQUENCY_MHZ attribute
    case ATTR_SAFE_MODE_FREQUENCY_MHZ:
        l_attrSize = sizeof(ATTR_SAFE_MODE_FREQUENCY_MHZ_type);
        break;

    // Get size for ATTR_AVSBUS_FREQUENCY attribute
    case ATTR_AVSBUS_FREQUENCY:
        l_attrSize = sizeof(ATTR_AVSBUS_FREQUENCY_type);
        break;

    // Get size for ATTR_VDD_AVSBUS_BUSNUM attribute
    case ATTR_VDD_AVSBUS_BUSNUM:
        l_attrSize = sizeof(ATTR_VDD_AVSBUS_BUSNUM_type);
        break;

    // Get size for ATTR_VDN_AVSBUS_BUSNUM attribute
    case ATTR_VDN_AVSBUS_BUSNUM:
        l_attrSize = sizeof(ATTR_VDN_AVSBUS_BUSNUM_type);
        break;

    // Get size for ATTR_VCS_AVSBUS_BUSNUM attribute
    case ATTR_VCS_AVSBUS_BUSNUM:
        l_attrSize = sizeof(ATTR_VCS_AVSBUS_BUSNUM_type);
        break;

    // Get size for ATTR_VDD_AVSBUS_RAIL attribute
    case ATTR_VDD_AVSBUS_RAIL:
        l_attrSize = sizeof(ATTR_VDD_AVSBUS_RAIL_type);
        break;

    // Get size for ATTR_VDN_AVSBUS_RAIL attribute
    case ATTR_VDN_AVSBUS_RAIL:
        l_attrSize = sizeof(ATTR_VDN_AVSBUS_RAIL_type);
        break;

    // Get size for ATTR_VCS_AVSBUS_RAIL attribute
    case ATTR_VCS_AVSBUS_RAIL:
        l_attrSize = sizeof(ATTR_VCS_AVSBUS_RAIL_type);
        break;

    // Get size for ATTR_VCS_I2C_BUSNUM attribute
    case ATTR_VCS_I2C_BUSNUM:
        l_attrSize = sizeof(ATTR_VCS_I2C_BUSNUM_type);
        break;

    // Get size for ATTR_VDD_BOOT_VOLTAGE attribute
    case ATTR_VDD_BOOT_VOLTAGE:
        l_attrSize = sizeof(ATTR_VDD_BOOT_VOLTAGE_type);
        break;

    // Get size for ATTR_VDN_BOOT_VOLTAGE attribute
    case ATTR_VDN_BOOT_VOLTAGE:
        l_attrSize = sizeof(ATTR_VDN_BOOT_VOLTAGE_type);
        break;

    // Get size for ATTR_VCS_BOOT_VOLTAGE attribute
    case ATTR_VCS_BOOT_VOLTAGE:
        l_attrSize = sizeof(ATTR_VCS_BOOT_VOLTAGE_type);
        break;

    // Get size for ATTR_SPIPSS_FREQUENCY attribute
    case ATTR_SPIPSS_FREQUENCY:
        l_attrSize = sizeof(ATTR_SPIPSS_FREQUENCY_type);
        break;

    // Get size for ATTR_PM_APSS_CHIP_SELECT attribute
    case ATTR_PM_APSS_CHIP_SELECT:
        l_attrSize = sizeof(ATTR_PM_APSS_CHIP_SELECT_type);
        break;

    // Get size for ATTR_PROC_R_LOADLINE_VDD_UOHM attribute
    case ATTR_PROC_R_LOADLINE_VDD_UOHM:
        l_attrSize = sizeof(ATTR_PROC_R_LOADLINE_VDD_UOHM_type);
        break;

    // Get size for ATTR_PROC_R_DISTLOSS_VDD_UOHM attribute
    case ATTR_PROC_R_DISTLOSS_VDD_UOHM:
        l_attrSize = sizeof(ATTR_PROC_R_DISTLOSS_VDD_UOHM_type);
        break;

    // Get size for ATTR_PROC_VRM_VOFFSET_VDD_UV attribute
    case ATTR_PROC_VRM_VOFFSET_VDD_UV:
        l_attrSize = sizeof(ATTR_PROC_VRM_VOFFSET_VDD_UV_type);
        break;

    // Get size for ATTR_PROC_R_LOADLINE_VDN_UOHM attribute
    case ATTR_PROC_R_LOADLINE_VDN_UOHM:
        l_attrSize = sizeof(ATTR_PROC_R_LOADLINE_VDN_UOHM_type);
        break;

    // Get size for ATTR_PROC_R_DISTLOSS_VDN_UOHM attribute
    case ATTR_PROC_R_DISTLOSS_VDN_UOHM:
        l_attrSize = sizeof(ATTR_PROC_R_DISTLOSS_VDN_UOHM_type);
        break;

    // Get size for ATTR_PROC_VRM_VOFFSET_VDN_UV attribute
    case ATTR_PROC_VRM_VOFFSET_VDN_UV:
        l_attrSize = sizeof(ATTR_PROC_VRM_VOFFSET_VDN_UV_type);
        break;

    // Get size for ATTR_PROC_R_LOADLINE_VCS_UOHM attribute
    case ATTR_PROC_R_LOADLINE_VCS_UOHM:
        l_attrSize = sizeof(ATTR_PROC_R_LOADLINE_VCS_UOHM_type);
        break;

    // Get size for ATTR_PROC_R_DISTLOSS_VCS_UOHM attribute
    case ATTR_PROC_R_DISTLOSS_VCS_UOHM:
        l_attrSize = sizeof(ATTR_PROC_R_DISTLOSS_VCS_UOHM_type);
        break;

    // Get size for ATTR_PROC_VRM_VOFFSET_VCS_UV attribute
    case ATTR_PROC_VRM_VOFFSET_VCS_UV:
        l_attrSize = sizeof(ATTR_PROC_VRM_VOFFSET_VCS_UV_type);
        break;

    // Get size for ATTR_FREQ_BIAS_ULTRATURBO attribute
    case ATTR_FREQ_BIAS_ULTRATURBO:
        l_attrSize = sizeof(ATTR_FREQ_BIAS_ULTRATURBO_type);
        break;

    // Get size for ATTR_FREQ_BIAS_TURBO attribute
    case ATTR_FREQ_BIAS_TURBO:
        l_attrSize = sizeof(ATTR_FREQ_BIAS_TURBO_type);
        break;

    // Get size for ATTR_FREQ_BIAS_NOMINAL attribute
    case ATTR_FREQ_BIAS_NOMINAL:
        l_attrSize = sizeof(ATTR_FREQ_BIAS_NOMINAL_type);
        break;

    // Get size for ATTR_FREQ_BIAS_POWERSAVE attribute
    case ATTR_FREQ_BIAS_POWERSAVE:
        l_attrSize = sizeof(ATTR_FREQ_BIAS_POWERSAVE_type);
        break;

    // Get size for ATTR_VOLTAGE_EXT_VDD_BIAS_ULTRATURBO attribute
    case ATTR_VOLTAGE_EXT_VDD_BIAS_ULTRATURBO:
        l_attrSize = sizeof(ATTR_VOLTAGE_EXT_VDD_BIAS_ULTRATURBO_type);
        break;

    // Get size for ATTR_VOLTAGE_EXT_VDD_BIAS_TURBO attribute
    case ATTR_VOLTAGE_EXT_VDD_BIAS_TURBO:
        l_attrSize = sizeof(ATTR_VOLTAGE_EXT_VDD_BIAS_TURBO_type);
        break;

    // Get size for ATTR_VOLTAGE_EXT_VDD_BIAS_NOMINAL attribute
    case ATTR_VOLTAGE_EXT_VDD_BIAS_NOMINAL:
        l_attrSize = sizeof(ATTR_VOLTAGE_EXT_VDD_BIAS_NOMINAL_type);
        break;

    // Get size for ATTR_VOLTAGE_EXT_VDD_BIAS_POWERSAVE attribute
    case ATTR_VOLTAGE_EXT_VDD_BIAS_POWERSAVE:
        l_attrSize = sizeof(ATTR_VOLTAGE_EXT_VDD_BIAS_POWERSAVE_type);
        break;

    // Get size for ATTR_VOLTAGE_EXT_VCS_BIAS attribute
    case ATTR_VOLTAGE_EXT_VCS_BIAS:
        l_attrSize = sizeof(ATTR_VOLTAGE_EXT_VCS_BIAS_type);
        break;

    // Get size for ATTR_VOLTAGE_EXT_VDN_BIAS attribute
    case ATTR_VOLTAGE_EXT_VDN_BIAS:
        l_attrSize = sizeof(ATTR_VOLTAGE_EXT_VDN_BIAS_type);
        break;

    // Get size for ATTR_VOLTAGE_INT_VDD_BIAS_ULTRATURBO attribute
    case ATTR_VOLTAGE_INT_VDD_BIAS_ULTRATURBO:
        l_attrSize = sizeof(ATTR_VOLTAGE_INT_VDD_BIAS_ULTRATURBO_type);
        break;

    // Get size for ATTR_VOLTAGE_INT_VDD_BIAS_TURBO attribute
    case ATTR_VOLTAGE_INT_VDD_BIAS_TURBO:
        l_attrSize = sizeof(ATTR_VOLTAGE_INT_VDD_BIAS_TURBO_type);
        break;

    // Get size for ATTR_VOLTAGE_INT_VDD_BIAS_NOMINAL attribute
    case ATTR_VOLTAGE_INT_VDD_BIAS_NOMINAL:
        l_attrSize = sizeof(ATTR_VOLTAGE_INT_VDD_BIAS_NOMINAL_type);
        break;

    // Get size for ATTR_VOLTAGE_INT_VDD_BIAS_POWERSAVE attribute
    case ATTR_VOLTAGE_INT_VDD_BIAS_POWERSAVE:
        l_attrSize = sizeof(ATTR_VOLTAGE_INT_VDD_BIAS_POWERSAVE_type);
        break;

    // Get size for ATTR_STOP4_DISABLE attribute
    case ATTR_STOP4_DISABLE:
        l_attrSize = sizeof(ATTR_STOP4_DISABLE_type);
        break;

    // Get size for ATTR_STOP8_DISABLE attribute
    case ATTR_STOP8_DISABLE:
        l_attrSize = sizeof(ATTR_STOP8_DISABLE_type);
        break;

    // Get size for ATTR_STOP11_DISABLE attribute
    case ATTR_STOP11_DISABLE:
        l_attrSize = sizeof(ATTR_STOP11_DISABLE_type);
        break;

    // Get size for ATTR_SYSTEM_CORE_PERIODIC_QUIESCE_DISABLE attribute
    case ATTR_SYSTEM_CORE_PERIODIC_QUIESCE_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_CORE_PERIODIC_QUIESCE_DISABLE_type);
        break;

    // Get size for ATTR_SYSTEM_POUNDV_VALIDITY_HALT_DISABLE attribute
    case ATTR_SYSTEM_POUNDV_VALIDITY_HALT_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_POUNDV_VALIDITY_HALT_DISABLE_type);
        break;

    // Get size for ATTR_SYSTEM_WOF_DISABLE attribute
    case ATTR_SYSTEM_WOF_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_WOF_DISABLE_type);
        break;

    // Get size for ATTR_SYSTEM_WOV_UNDERV_DISABLE attribute
    case ATTR_SYSTEM_WOV_UNDERV_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_WOV_UNDERV_DISABLE_type);
        break;

    // Get size for ATTR_WOV_UNDERV_FORCE attribute
    case ATTR_WOV_UNDERV_FORCE:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_FORCE_type);
        break;

    // Get size for ATTR_SYSTEM_WOV_OVERV_DISABLE attribute
    case ATTR_SYSTEM_WOV_OVERV_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_WOV_OVERV_DISABLE_type);
        break;

    // Get size for ATTR_SYSTEM_IVRM_DISABLE attribute
    case ATTR_SYSTEM_IVRM_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_IVRM_DISABLE_type);
        break;

    // Get size for ATTR_WOF_ENABLE_FRATIO attribute
    case ATTR_WOF_ENABLE_FRATIO:
        l_attrSize = sizeof(ATTR_WOF_ENABLE_FRATIO_type);
        break;

    // Get size for ATTR_WOF_ENABLE_VRATIO attribute
    case ATTR_WOF_ENABLE_VRATIO:
        l_attrSize = sizeof(ATTR_WOF_ENABLE_VRATIO_type);
        break;

    // Get size for ATTR_WOF_VRATIO_SELECT attribute
    case ATTR_WOF_VRATIO_SELECT:
        l_attrSize = sizeof(ATTR_WOF_VRATIO_SELECT_type);
        break;

    // Get size for ATTR_WOV_SAMPLE_125US attribute
    case ATTR_WOV_SAMPLE_125US:
        l_attrSize = sizeof(ATTR_WOV_SAMPLE_125US_type);
        break;

    // Get size for ATTR_WOV_MAX_DROOP_10THPCT attribute
    case ATTR_WOV_MAX_DROOP_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_MAX_DROOP_10THPCT_type);
        break;

    // Get size for ATTR_WOV_UNDERV_VMIN_MV attribute
    case ATTR_WOV_UNDERV_VMIN_MV:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_VMIN_MV_type);
        break;

    // Get size for ATTR_WOV_UNDERV_LIMIT_ALGORITHM attribute
    case ATTR_WOV_UNDERV_LIMIT_ALGORITHM:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_LIMIT_ALGORITHM_type);
        break;

    // Get size for ATTR_WOV_UNDERV_LIMIT_MV attribute
    case ATTR_WOV_UNDERV_LIMIT_MV:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_LIMIT_MV_type);
        break;

    // Get size for ATTR_WOV_UNDERV_LIMIT_10THPCT attribute
    case ATTR_WOV_UNDERV_LIMIT_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_LIMIT_10THPCT_type);
        break;

    // Get size for ATTR_WOV_OVERV_VMAX_SETPOINT_MV attribute
    case ATTR_WOV_OVERV_VMAX_SETPOINT_MV:
        l_attrSize = sizeof(ATTR_WOV_OVERV_VMAX_SETPOINT_MV_type);
        break;

    // Get size for ATTR_WOV_UNDERV_STEP_INCR_10THPCT attribute
    case ATTR_WOV_UNDERV_STEP_INCR_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_STEP_INCR_10THPCT_type);
        break;

    // Get size for ATTR_WOV_UNDERV_STEP_DECR_10THPCT attribute
    case ATTR_WOV_UNDERV_STEP_DECR_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_STEP_DECR_10THPCT_type);
        break;

    // Get size for ATTR_WOV_UNDERV_MAX_10THPCT attribute
    case ATTR_WOV_UNDERV_MAX_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_MAX_10THPCT_type);
        break;

    // Get size for ATTR_WOV_UNDERV_PERF_LOSS_THRESH_10THPCT attribute
    case ATTR_WOV_UNDERV_PERF_LOSS_THRESH_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_PERF_LOSS_THRESH_10THPCT_type);
        break;

    // Get size for ATTR_WOV_OVERV_STEP_INCR_10THPCT attribute
    case ATTR_WOV_OVERV_STEP_INCR_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_OVERV_STEP_INCR_10THPCT_type);
        break;

    // Get size for ATTR_WOV_OVERV_STEP_DECR_10THPCT attribute
    case ATTR_WOV_OVERV_STEP_DECR_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_OVERV_STEP_DECR_10THPCT_type);
        break;

    // Get size for ATTR_WOV_OVERV_MAX_10THPCT attribute
    case ATTR_WOV_OVERV_MAX_10THPCT:
        l_attrSize = sizeof(ATTR_WOV_OVERV_MAX_10THPCT_type);
        break;

    // Get size for ATTR_PBAX_GROUPID attribute
    case ATTR_PBAX_GROUPID:
        l_attrSize = sizeof(ATTR_PBAX_GROUPID_type);
        break;

    // Get size for ATTR_PBAX_CHIPID attribute
    case ATTR_PBAX_CHIPID:
        l_attrSize = sizeof(ATTR_PBAX_CHIPID_type);
        break;

    // Get size for ATTR_PBAX_BRDCST_ID_VECTOR attribute
    case ATTR_PBAX_BRDCST_ID_VECTOR:
        l_attrSize = sizeof(ATTR_PBAX_BRDCST_ID_VECTOR_type);
        break;

    // Get size for ATTR_POUNDV_BUCKET_NUM_OVERRIDE attribute
    case ATTR_POUNDV_BUCKET_NUM_OVERRIDE:
        l_attrSize = sizeof(ATTR_POUNDV_BUCKET_NUM_OVERRIDE_type);
        break;

    // Get size for ATTR_POUNDV_BUCKET_NUM attribute
    case ATTR_POUNDV_BUCKET_NUM:
        l_attrSize = sizeof(ATTR_POUNDV_BUCKET_NUM_type);
        break;

    // Get size for ATTR_DUMP_STOP_INFO_SUPPRESS_ERROR_TRACE attribute
    case ATTR_DUMP_STOP_INFO_SUPPRESS_ERROR_TRACE:
        l_attrSize = sizeof(ATTR_DUMP_STOP_INFO_SUPPRESS_ERROR_TRACE_type);
        break;

    // Get size for ATTR_DUMP_STOP_INFO_ENABLE_ERRORLOG attribute
    case ATTR_DUMP_STOP_INFO_ENABLE_ERRORLOG:
        l_attrSize = sizeof(ATTR_DUMP_STOP_INFO_ENABLE_ERRORLOG_type);
        break;

    // Get size for ATTR_SYSTEM_VDM_DISABLE attribute
    case ATTR_SYSTEM_VDM_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_VDM_DISABLE_type);
        break;

    // Get size for ATTR_VDM_DROOP_SMALL_OVERRIDE attribute
    case ATTR_VDM_DROOP_SMALL_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_DROOP_SMALL_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_DROOP_LARGE_OVERRIDE attribute
    case ATTR_VDM_DROOP_LARGE_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_DROOP_LARGE_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_DROOP_EXTREME_OVERRIDE attribute
    case ATTR_VDM_DROOP_EXTREME_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_DROOP_EXTREME_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_OVERVOLT_OVERRIDE attribute
    case ATTR_VDM_OVERVOLT_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_OVERVOLT_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_SMALL_FREQ_DROP_N_S_OVERRIDE attribute
    case ATTR_VDM_SMALL_FREQ_DROP_N_S_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_SMALL_FREQ_DROP_N_S_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_LARGE_FREQ_DROP_N_L_OVERRIDE attribute
    case ATTR_VDM_LARGE_FREQ_DROP_N_L_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_LARGE_FREQ_DROP_N_L_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_FREQ_RETURN_L_S_OVERRIDE attribute
    case ATTR_VDM_FREQ_RETURN_L_S_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_FREQ_RETURN_L_S_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_FREQ_RETURN_S_N_OVERRIDE attribute
    case ATTR_VDM_FREQ_RETURN_S_N_OVERRIDE:
        l_attrSize = sizeof(ATTR_VDM_FREQ_RETURN_S_N_OVERRIDE_type);
        break;

    // Get size for ATTR_VDM_EXTREME_THOTTLE_ENABLE attribute
    case ATTR_VDM_EXTREME_THOTTLE_ENABLE:
        l_attrSize = sizeof(ATTR_VDM_EXTREME_THOTTLE_ENABLE_type);
        break;

    // Get size for ATTR_VDM_FMAX_OVERRIDE_KHZ attribute
    case ATTR_VDM_FMAX_OVERRIDE_KHZ:
        l_attrSize = sizeof(ATTR_VDM_FMAX_OVERRIDE_KHZ_type);
        break;

    // Get size for ATTR_VDM_FMIN_OVERRIDE_KHZ attribute
    case ATTR_VDM_FMIN_OVERRIDE_KHZ:
        l_attrSize = sizeof(ATTR_VDM_FMIN_OVERRIDE_KHZ_type);
        break;

    // Get size for ATTR_VDM_VID_COMPARE_OVERRIDE_MV attribute
    case ATTR_VDM_VID_COMPARE_OVERRIDE_MV:
        l_attrSize = sizeof(ATTR_VDM_VID_COMPARE_OVERRIDE_MV_type);
        break;

    // Get size for ATTR_VDM_VID_COMPARE_BIAS_0P5PCT attribute
    case ATTR_VDM_VID_COMPARE_BIAS_0P5PCT:
        l_attrSize = sizeof(ATTR_VDM_VID_COMPARE_BIAS_0P5PCT_type);
        break;

    // Get size for ATTR_DPLL_VDM_RESPONSE attribute
    case ATTR_DPLL_VDM_RESPONSE:
        l_attrSize = sizeof(ATTR_DPLL_VDM_RESPONSE_type);
        break;

    // Get size for ATTR_IVRM_DEADZONE_MV attribute
    case ATTR_IVRM_DEADZONE_MV:
        l_attrSize = sizeof(ATTR_IVRM_DEADZONE_MV_type);
        break;

    // Get size for ATTR_IVRM_STRENGTH_LOOKUP attribute
    case ATTR_IVRM_STRENGTH_LOOKUP:
        l_attrSize = sizeof(ATTR_IVRM_STRENGTH_LOOKUP_type);
        break;

    // Get size for ATTR_IVRM_VIN_MULTIPLIER attribute
    case ATTR_IVRM_VIN_MULTIPLIER:
        l_attrSize = sizeof(ATTR_IVRM_VIN_MULTIPLIER_type);
        break;

    // Get size for ATTR_IVRM_VIN_MAX_MV attribute
    case ATTR_IVRM_VIN_MAX_MV:
        l_attrSize = sizeof(ATTR_IVRM_VIN_MAX_MV_type);
        break;

    // Get size for ATTR_IVRM_STEP_DELAY_NS attribute
    case ATTR_IVRM_STEP_DELAY_NS:
        l_attrSize = sizeof(ATTR_IVRM_STEP_DELAY_NS_type);
        break;

    // Get size for ATTR_IVRM_STABILIZATION_DELAY_NS attribute
    case ATTR_IVRM_STABILIZATION_DELAY_NS:
        l_attrSize = sizeof(ATTR_IVRM_STABILIZATION_DELAY_NS_type);
        break;

    // Get size for ATTR_TDP_RDP_CURRENT_FACTOR attribute
    case ATTR_TDP_RDP_CURRENT_FACTOR:
        l_attrSize = sizeof(ATTR_TDP_RDP_CURRENT_FACTOR_type);
        break;

    // Get size for ATTR_SYSTEM_RESCLK_DISABLE attribute
    case ATTR_SYSTEM_RESCLK_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_RESCLK_DISABLE_type);
        break;

    // Get size for ATTR_SYSTEM_RESCLK_STEP_DELAY attribute
    case ATTR_SYSTEM_RESCLK_STEP_DELAY:
        l_attrSize = sizeof(ATTR_SYSTEM_RESCLK_STEP_DELAY_type);
        break;

    // Get size for ATTR_SYSTEM_RESCLK_FREQ_REGIONS attribute
    case ATTR_SYSTEM_RESCLK_FREQ_REGIONS:
        l_attrSize = sizeof(ATTR_SYSTEM_RESCLK_FREQ_REGIONS_type);
        break;

    // Get size for ATTR_SYSTEM_RESCLK_FREQ_REGION_INDEX attribute
    case ATTR_SYSTEM_RESCLK_FREQ_REGION_INDEX:
        l_attrSize = sizeof(ATTR_SYSTEM_RESCLK_FREQ_REGION_INDEX_type);
        break;

    // Get size for ATTR_SYSTEM_RESCLK_VALUE attribute
    case ATTR_SYSTEM_RESCLK_VALUE:
        l_attrSize = sizeof(ATTR_SYSTEM_RESCLK_VALUE_type);
        break;

    // Get size for ATTR_SYSTEM_RESCLK_L3_VALUE attribute
    case ATTR_SYSTEM_RESCLK_L3_VALUE:
        l_attrSize = sizeof(ATTR_SYSTEM_RESCLK_L3_VALUE_type);
        break;

    // Get size for ATTR_SYSTEM_RESCLK_L3_VOLTAGE_THRESHOLD_MV attribute
    case ATTR_SYSTEM_RESCLK_L3_VOLTAGE_THRESHOLD_MV:
        l_attrSize = sizeof(ATTR_SYSTEM_RESCLK_L3_VOLTAGE_THRESHOLD_MV_type);
        break;

    // Get size for ATTR_SYSTEM_RING_DBG_MODE attribute
    case ATTR_SYSTEM_RING_DBG_MODE:
        l_attrSize = sizeof(ATTR_SYSTEM_RING_DBG_MODE_type);
        break;

    // Get size for ATTR_PERF_24x7_INVOCATION_TIME_MS attribute
    case ATTR_PERF_24x7_INVOCATION_TIME_MS:
        l_attrSize = sizeof(ATTR_PERF_24x7_INVOCATION_TIME_MS_type);
        break;

    // Get size for ATTR_CME_INSTRUCTION_TRACE_ENABLE attribute
    case ATTR_CME_INSTRUCTION_TRACE_ENABLE:
        l_attrSize = sizeof(ATTR_CME_INSTRUCTION_TRACE_ENABLE_type);
        break;

    // Get size for ATTR_CME_CHTM_TRACE_ENABLE attribute
    case ATTR_CME_CHTM_TRACE_ENABLE:
        l_attrSize = sizeof(ATTR_CME_CHTM_TRACE_ENABLE_type);
        break;

    // Get size for ATTR_CME_CHTM_TRACE_MEMORY_CONFIG attribute
    case ATTR_CME_CHTM_TRACE_MEMORY_CONFIG:
        l_attrSize = sizeof(ATTR_CME_CHTM_TRACE_MEMORY_CONFIG_type);
        break;

    // Get size for ATTR_PGPE_HCODE_FUNCTION_ENABLE attribute
    case ATTR_PGPE_HCODE_FUNCTION_ENABLE:
        l_attrSize = sizeof(ATTR_PGPE_HCODE_FUNCTION_ENABLE_type);
        break;

    // Get size for ATTR_POUND_W_STATIC_DATA_ENABLE attribute
    case ATTR_POUND_W_STATIC_DATA_ENABLE:
        l_attrSize = sizeof(ATTR_POUND_W_STATIC_DATA_ENABLE_type);
        break;

    // Get size for ATTR_AUX_FUNC_INVOCATION_TIME_MS attribute
    case ATTR_AUX_FUNC_INVOCATION_TIME_MS:
        l_attrSize = sizeof(ATTR_AUX_FUNC_INVOCATION_TIME_MS_type);
        break;

    // Get size for ATTR_SYS_VFRT_STATIC_DATA_ENABLE attribute
    case ATTR_SYS_VFRT_STATIC_DATA_ENABLE:
        l_attrSize = sizeof(ATTR_SYS_VFRT_STATIC_DATA_ENABLE_type);
        break;

    // Get size for ATTR_NEST_LEAKAGE_PERCENT attribute
    case ATTR_NEST_LEAKAGE_PERCENT:
        l_attrSize = sizeof(ATTR_NEST_LEAKAGE_PERCENT_type);
        break;

    // Get size for ATTR_SYSTEM_PSTATES_MODE attribute
    case ATTR_SYSTEM_PSTATES_MODE:
        l_attrSize = sizeof(ATTR_SYSTEM_PSTATES_MODE_type);
        break;

    // Get size for ATTR_CORE_THROTTLE_ASSERT_COUNT attribute
    case ATTR_CORE_THROTTLE_ASSERT_COUNT:
        l_attrSize = sizeof(ATTR_CORE_THROTTLE_ASSERT_COUNT_type);
        break;

    // Get size for ATTR_CORE_THROTTLE_DEASSERT_COUNT attribute
    case ATTR_CORE_THROTTLE_DEASSERT_COUNT:
        l_attrSize = sizeof(ATTR_CORE_THROTTLE_DEASSERT_COUNT_type);
        break;

    // Get size for ATTR_PM_GLOBAL_FIR_TRACE_EN attribute
    case ATTR_PM_GLOBAL_FIR_TRACE_EN:
        l_attrSize = sizeof(ATTR_PM_GLOBAL_FIR_TRACE_EN_type);
        break;

    // Get size for ATTR_FREQ_PB_MHZ_POUNDV_FALLBACK attribute
    case ATTR_FREQ_PB_MHZ_POUNDV_FALLBACK:
        l_attrSize = sizeof(ATTR_FREQ_PB_MHZ_POUNDV_FALLBACK_type);
        break;

    // Get size for ATTR_SKIP_WAKEUP attribute
    case ATTR_SKIP_WAKEUP:
        l_attrSize = sizeof(ATTR_SKIP_WAKEUP_type);
        break;

    // Get size for ATTR_PM_MALF_ALERT_ENABLE attribute
    case ATTR_PM_MALF_ALERT_ENABLE:
        l_attrSize = sizeof(ATTR_PM_MALF_ALERT_ENABLE_type);
        break;

    // Get size for ATTR_PM_RESET_FFDC_ENABLE attribute
    case ATTR_PM_RESET_FFDC_ENABLE:
        l_attrSize = sizeof(ATTR_PM_RESET_FFDC_ENABLE_type);
        break;

    // Get size for ATTR_NEST_24x7_PERF_ACCUM_ENABLE attribute
    case ATTR_NEST_24x7_PERF_ACCUM_ENABLE:
        l_attrSize = sizeof(ATTR_NEST_24x7_PERF_ACCUM_ENABLE_type);
        break;

    // Get size for ATTR_UNSECURE_HOMER_ADDRESS attribute
    case ATTR_UNSECURE_HOMER_ADDRESS:
        l_attrSize = sizeof(ATTR_UNSECURE_HOMER_ADDRESS_type);
        break;

    // Get size for ATTR_UNSECURE_HOMER_SIZE attribute
    case ATTR_UNSECURE_HOMER_SIZE:
        l_attrSize = sizeof(ATTR_UNSECURE_HOMER_SIZE_type);
        break;

    // Get size for ATTR_PMCR_MOST_RECENT_MODE attribute
    case ATTR_PMCR_MOST_RECENT_MODE:
        l_attrSize = sizeof(ATTR_PMCR_MOST_RECENT_MODE_type);
        break;

    // Get size for ATTR_ENABLED_OBUS_BRICKS attribute
    case ATTR_ENABLED_OBUS_BRICKS:
        l_attrSize = sizeof(ATTR_ENABLED_OBUS_BRICKS_type);
        break;

    // Get size for ATTR_IO_XBUS_DCCAL_FLAGS attribute
    case ATTR_IO_XBUS_DCCAL_FLAGS:
        l_attrSize = sizeof(ATTR_IO_XBUS_DCCAL_FLAGS_type);
        break;

    // Get size for ATTR_IO_X_DEBUG attribute
    case ATTR_IO_X_DEBUG:
        l_attrSize = sizeof(ATTR_IO_X_DEBUG_type);
        break;

    // Get size for ATTR_IO_X_MFG_CHK attribute
    case ATTR_IO_X_MFG_CHK:
        l_attrSize = sizeof(ATTR_IO_X_MFG_CHK_type);
        break;

    // Get size for ATTR_IO_X_MFG_MIN_EYE_WIDTH attribute
    case ATTR_IO_X_MFG_MIN_EYE_WIDTH:
        l_attrSize = sizeof(ATTR_IO_X_MFG_MIN_EYE_WIDTH_type);
        break;

    // Get size for ATTR_IO_XBUS_MASTER_MODE attribute
    case ATTR_IO_XBUS_MASTER_MODE:
        l_attrSize = sizeof(ATTR_IO_XBUS_MASTER_MODE_type);
        break;

    // Get size for ATTR_IO_XBUS_TX_MARGIN_RATIO attribute
    case ATTR_IO_XBUS_TX_MARGIN_RATIO:
        l_attrSize = sizeof(ATTR_IO_XBUS_TX_MARGIN_RATIO_type);
        break;

    // Get size for ATTR_IO_XBUS_TX_FFE_PRECURSOR attribute
    case ATTR_IO_XBUS_TX_FFE_PRECURSOR:
        l_attrSize = sizeof(ATTR_IO_XBUS_TX_FFE_PRECURSOR_type);
        break;

    // Get size for ATTR_IO_XBUS_CHAN_EQ attribute
    case ATTR_IO_XBUS_CHAN_EQ:
        l_attrSize = sizeof(ATTR_IO_XBUS_CHAN_EQ_type);
        break;

    // Get size for ATTR_IO_XBUS_GRP0_PRE_BAD_LANE_DATA attribute
    case ATTR_IO_XBUS_GRP0_PRE_BAD_LANE_DATA:
        l_attrSize = sizeof(ATTR_IO_XBUS_GRP0_PRE_BAD_LANE_DATA_type);
        break;

    // Get size for ATTR_IO_XBUS_GRP1_PRE_BAD_LANE_DATA attribute
    case ATTR_IO_XBUS_GRP1_PRE_BAD_LANE_DATA:
        l_attrSize = sizeof(ATTR_IO_XBUS_GRP1_PRE_BAD_LANE_DATA_type);
        break;

    // Get size for ATTR_DO_MSS_WR_VREF attribute
    case ATTR_DO_MSS_WR_VREF:
        l_attrSize = sizeof(ATTR_DO_MSS_WR_VREF_type);
        break;

    // Get size for ATTR_SKIP_HW_VREF_CAL attribute
    case ATTR_SKIP_HW_VREF_CAL:
        l_attrSize = sizeof(ATTR_SKIP_HW_VREF_CAL_type);
        break;

    // Get size for ATTR_SKIP_RD_VREF_VREFSENSE_OVERRIDE attribute
    case ATTR_SKIP_RD_VREF_VREFSENSE_OVERRIDE:
        l_attrSize = sizeof(ATTR_SKIP_RD_VREF_VREFSENSE_OVERRIDE_type);
        break;

    // Get size for ATTR_DO_MSS_VREF_DAC attribute
    case ATTR_DO_MSS_VREF_DAC:
        l_attrSize = sizeof(ATTR_DO_MSS_VREF_DAC_type);
        break;

    // Get size for ATTR_DO_MSS_TRAINING_BAD_BITS attribute
    case ATTR_DO_MSS_TRAINING_BAD_BITS:
        l_attrSize = sizeof(ATTR_DO_MSS_TRAINING_BAD_BITS_type);
        break;

    // Get size for ATTR_MSS_RUN_DCD_CALIBRATION attribute
    case ATTR_MSS_RUN_DCD_CALIBRATION:
        l_attrSize = sizeof(ATTR_MSS_RUN_DCD_CALIBRATION_type);
        break;

    // Get size for ATTR_OMI_INBAND_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_OMI_INBAND_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_OMI_INBAND_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_MSS_MRW_NVDIMM_SLOT_POSITION attribute
    case ATTR_MSS_MRW_NVDIMM_SLOT_POSITION:
        l_attrSize = sizeof(ATTR_MSS_MRW_NVDIMM_SLOT_POSITION_type);
        break;

    // Get size for ATTR_MSS_MRW_UNSUPPORTED_RANK_CONFIG attribute
    case ATTR_MSS_MRW_UNSUPPORTED_RANK_CONFIG:
        l_attrSize = sizeof(ATTR_MSS_MRW_UNSUPPORTED_RANK_CONFIG_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_MT attribute
    case ATTR_VPD_OVERRIDE_MT:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_MT_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_MT_ENABLE attribute
    case ATTR_VPD_OVERRIDE_MT_ENABLE:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_MT_ENABLE_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_MR attribute
    case ATTR_VPD_OVERRIDE_MR:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_MR_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_MR_ENABLE attribute
    case ATTR_VPD_OVERRIDE_MR_ENABLE:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_MR_ENABLE_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_DQ attribute
    case ATTR_VPD_OVERRIDE_DQ:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_DQ_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_DQ_ENABLE attribute
    case ATTR_VPD_OVERRIDE_DQ_ENABLE:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_DQ_ENABLE_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_CK attribute
    case ATTR_VPD_OVERRIDE_CK:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_CK_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_CK_ENABLE attribute
    case ATTR_VPD_OVERRIDE_CK_ENABLE:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_CK_ENABLE_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_MW attribute
    case ATTR_VPD_OVERRIDE_MW:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_MW_type);
        break;

    // Get size for ATTR_VPD_OVERRIDE_MW_ENABLE attribute
    case ATTR_VPD_OVERRIDE_MW_ENABLE:
        l_attrSize = sizeof(ATTR_VPD_OVERRIDE_MW_ENABLE_type);
        break;

    // Get size for ATTR_MSS_OVERRIDE_MEMDIAGS_BCMODE attribute
    case ATTR_MSS_OVERRIDE_MEMDIAGS_BCMODE:
        l_attrSize = sizeof(ATTR_MSS_OVERRIDE_MEMDIAGS_BCMODE_type);
        break;

    // Get size for ATTR_EFF_DRAM_GEN attribute
    case ATTR_EFF_DRAM_GEN:
        l_attrSize = sizeof(ATTR_EFF_DRAM_GEN_type);
        break;

    // Get size for ATTR_EFF_DIMM_TYPE attribute
    case ATTR_EFF_DIMM_TYPE:
        l_attrSize = sizeof(ATTR_EFF_DIMM_TYPE_type);
        break;

    // Get size for ATTR_EFF_HYBRID_MEMORY_TYPE attribute
    case ATTR_EFF_HYBRID_MEMORY_TYPE:
        l_attrSize = sizeof(ATTR_EFF_HYBRID_MEMORY_TYPE_type);
        break;

    // Get size for ATTR_EFF_HYBRID attribute
    case ATTR_EFF_HYBRID:
        l_attrSize = sizeof(ATTR_EFF_HYBRID_type);
        break;

    // Get size for ATTR_EFF_DRAM_DENSITY attribute
    case ATTR_EFF_DRAM_DENSITY:
        l_attrSize = sizeof(ATTR_EFF_DRAM_DENSITY_type);
        break;

    // Get size for ATTR_EFF_DRAM_BANK_BITS attribute
    case ATTR_EFF_DRAM_BANK_BITS:
        l_attrSize = sizeof(ATTR_EFF_DRAM_BANK_BITS_type);
        break;

    // Get size for ATTR_EFF_DRAM_BANK_GROUP_BITS attribute
    case ATTR_EFF_DRAM_BANK_GROUP_BITS:
        l_attrSize = sizeof(ATTR_EFF_DRAM_BANK_GROUP_BITS_type);
        break;

    // Get size for ATTR_EFF_DRAM_COLUMN_BITS attribute
    case ATTR_EFF_DRAM_COLUMN_BITS:
        l_attrSize = sizeof(ATTR_EFF_DRAM_COLUMN_BITS_type);
        break;

    // Get size for ATTR_EFF_DRAM_ROW_BITS attribute
    case ATTR_EFF_DRAM_ROW_BITS:
        l_attrSize = sizeof(ATTR_EFF_DRAM_ROW_BITS_type);
        break;

    // Get size for ATTR_EFF_PRIM_STACK_TYPE attribute
    case ATTR_EFF_PRIM_STACK_TYPE:
        l_attrSize = sizeof(ATTR_EFF_PRIM_STACK_TYPE_type);
        break;

    // Get size for ATTR_EFF_DRAM_PPR attribute
    case ATTR_EFF_DRAM_PPR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_PPR_type);
        break;

    // Get size for ATTR_EFF_DRAM_SOFT_PPR attribute
    case ATTR_EFF_DRAM_SOFT_PPR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_SOFT_PPR_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRCD attribute
    case ATTR_EFF_DRAM_TRCD:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRCD_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRP attribute
    case ATTR_EFF_DRAM_TRP:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRP_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRAS attribute
    case ATTR_EFF_DRAM_TRAS:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRAS_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRC attribute
    case ATTR_EFF_DRAM_TRC:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRC_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRFC attribute
    case ATTR_EFF_DRAM_TRFC:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRFC_type);
        break;

    // Get size for ATTR_EFF_DRAM_TFAW attribute
    case ATTR_EFF_DRAM_TFAW:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TFAW_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRRD_S attribute
    case ATTR_EFF_DRAM_TRRD_S:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRRD_S_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRRD_L attribute
    case ATTR_EFF_DRAM_TRRD_L:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRRD_L_type);
        break;

    // Get size for ATTR_EFF_DRAM_TCCD_L attribute
    case ATTR_EFF_DRAM_TCCD_L:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TCCD_L_type);
        break;

    // Get size for ATTR_EFF_DRAM_TWR attribute
    case ATTR_EFF_DRAM_TWR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TWR_type);
        break;

    // Get size for ATTR_EFF_DRAM_TWTR_S attribute
    case ATTR_EFF_DRAM_TWTR_S:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TWTR_S_type);
        break;

    // Get size for ATTR_EFF_DRAM_TWTR_L attribute
    case ATTR_EFF_DRAM_TWTR_L:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TWTR_L_type);
        break;

    // Get size for ATTR_EFF_DRAM_TMAW attribute
    case ATTR_EFF_DRAM_TMAW:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TMAW_type);
        break;

    // Get size for ATTR_EFF_DRAM_WIDTH attribute
    case ATTR_EFF_DRAM_WIDTH:
        l_attrSize = sizeof(ATTR_EFF_DRAM_WIDTH_type);
        break;

    // Get size for ATTR_EFF_DRAM_RANK_MIX attribute
    case ATTR_EFF_DRAM_RANK_MIX:
        l_attrSize = sizeof(ATTR_EFF_DRAM_RANK_MIX_type);
        break;

    // Get size for ATTR_EFF_NUM_RANKS_PER_DIMM attribute
    case ATTR_EFF_NUM_RANKS_PER_DIMM:
        l_attrSize = sizeof(ATTR_EFF_NUM_RANKS_PER_DIMM_type);
        break;

    // Get size for ATTR_EFF_REGISTER_TYPE attribute
    case ATTR_EFF_REGISTER_TYPE:
        l_attrSize = sizeof(ATTR_EFF_REGISTER_TYPE_type);
        break;

    // Get size for ATTR_EFF_DRAM_MFG_ID attribute
    case ATTR_EFF_DRAM_MFG_ID:
        l_attrSize = sizeof(ATTR_EFF_DRAM_MFG_ID_type);
        break;

    // Get size for ATTR_EFF_RCD_MFG_ID attribute
    case ATTR_EFF_RCD_MFG_ID:
        l_attrSize = sizeof(ATTR_EFF_RCD_MFG_ID_type);
        break;

    // Get size for ATTR_EFF_REGISTER_REV attribute
    case ATTR_EFF_REGISTER_REV:
        l_attrSize = sizeof(ATTR_EFF_REGISTER_REV_type);
        break;

    // Get size for ATTR_EFF_PACKAGE_RANK_MAP attribute
    case ATTR_EFF_PACKAGE_RANK_MAP:
        l_attrSize = sizeof(ATTR_EFF_PACKAGE_RANK_MAP_type);
        break;

    // Get size for ATTR_EFF_NIBBLE_MAP attribute
    case ATTR_EFF_NIBBLE_MAP:
        l_attrSize = sizeof(ATTR_EFF_NIBBLE_MAP_type);
        break;

    // Get size for ATTR_ROW_REPAIR_SUPPORTED_MRW attribute
    case ATTR_ROW_REPAIR_SUPPORTED_MRW:
        l_attrSize = sizeof(ATTR_ROW_REPAIR_SUPPORTED_MRW_type);
        break;

    // Get size for ATTR_EI_BUS_TX_MSBSWAP attribute
    case ATTR_EI_BUS_TX_MSBSWAP:
        l_attrSize = sizeof(ATTR_EI_BUS_TX_MSBSWAP_type);
        break;

    // Get size for ATTR_PROC_NPU_REGION_ENABLED attribute
    case ATTR_PROC_NPU_REGION_ENABLED:
        l_attrSize = sizeof(ATTR_PROC_NPU_REGION_ENABLED_type);
        break;

    // Get size for ATTR_CLOCK_PLL_MUX_TOD attribute
    case ATTR_CLOCK_PLL_MUX_TOD:
        l_attrSize = sizeof(ATTR_CLOCK_PLL_MUX_TOD_type);
        break;

    // Get size for ATTR_CLOCK_PLL_MUX attribute
    case ATTR_CLOCK_PLL_MUX:
        l_attrSize = sizeof(ATTR_CLOCK_PLL_MUX_type);
        break;

    // Get size for ATTR_CLOCK_PLL_MUX0 attribute
    case ATTR_CLOCK_PLL_MUX0:
        l_attrSize = sizeof(ATTR_CLOCK_PLL_MUX0_type);
        break;

    // Get size for ATTR_I2C_BUS_DIV_REF attribute
    case ATTR_I2C_BUS_DIV_REF:
        l_attrSize = sizeof(ATTR_I2C_BUS_DIV_REF_type);
        break;

    // Get size for ATTR_EQ_GARD attribute
    case ATTR_EQ_GARD:
        l_attrSize = sizeof(ATTR_EQ_GARD_type);
        break;

    // Get size for ATTR_EC_GARD attribute
    case ATTR_EC_GARD:
        l_attrSize = sizeof(ATTR_EC_GARD_type);
        break;

    // Get size for ATTR_SBE_RUNTIME_MODE attribute
    case ATTR_SBE_RUNTIME_MODE:
        l_attrSize = sizeof(ATTR_SBE_RUNTIME_MODE_type);
        break;

    // Get size for ATTR_IS_SP_MODE attribute
    case ATTR_IS_SP_MODE:
        l_attrSize = sizeof(ATTR_IS_SP_MODE_type);
        break;

    // Get size for ATTR_SBE_FFDC_ENABLE attribute
    case ATTR_SBE_FFDC_ENABLE:
        l_attrSize = sizeof(ATTR_SBE_FFDC_ENABLE_type);
        break;

    // Get size for ATTR_SBE_INTERNAL_FFDC_ENABLE attribute
    case ATTR_SBE_INTERNAL_FFDC_ENABLE:
        l_attrSize = sizeof(ATTR_SBE_INTERNAL_FFDC_ENABLE_type);
        break;

    // Get size for ATTR_NEST_PLL_BUCKET attribute
    case ATTR_NEST_PLL_BUCKET:
        l_attrSize = sizeof(ATTR_NEST_PLL_BUCKET_type);
        break;

    // Get size for ATTR_FILTER_PLL_BUCKET attribute
    case ATTR_FILTER_PLL_BUCKET:
        l_attrSize = sizeof(ATTR_FILTER_PLL_BUCKET_type);
        break;

    // Get size for ATTR_MRW_FILTER_PLL_BUCKET attribute
    case ATTR_MRW_FILTER_PLL_BUCKET:
        l_attrSize = sizeof(ATTR_MRW_FILTER_PLL_BUCKET_type);
        break;

    // Get size for ATTR_FORCE_SYNC_SS_PLL_SPREAD attribute
    case ATTR_FORCE_SYNC_SS_PLL_SPREAD:
        l_attrSize = sizeof(ATTR_FORCE_SYNC_SS_PLL_SPREAD_type);
        break;

    // Get size for ATTR_BOOT_FREQ_MULT attribute
    case ATTR_BOOT_FREQ_MULT:
        l_attrSize = sizeof(ATTR_BOOT_FREQ_MULT_type);
        break;

    // Get size for ATTR_RISK_LEVEL attribute
    case ATTR_RISK_LEVEL:
        l_attrSize = sizeof(ATTR_RISK_LEVEL_type);
        break;

    // Get size for ATTR_DISABLE_HBBL_VECTORS attribute
    case ATTR_DISABLE_HBBL_VECTORS:
        l_attrSize = sizeof(ATTR_DISABLE_HBBL_VECTORS_type);
        break;

    // Get size for ATTR_BACKUP_SEEPROM_SELECT attribute
    case ATTR_BACKUP_SEEPROM_SELECT:
        l_attrSize = sizeof(ATTR_BACKUP_SEEPROM_SELECT_type);
        break;

    // Get size for ATTR_BOOT_FLAGS attribute
    case ATTR_BOOT_FLAGS:
        l_attrSize = sizeof(ATTR_BOOT_FLAGS_type);
        break;

    // Get size for ATTR_BOOT_FREQ_MHZ attribute
    case ATTR_BOOT_FREQ_MHZ:
        l_attrSize = sizeof(ATTR_BOOT_FREQ_MHZ_type);
        break;

    // Get size for ATTR_BRANCH_PIBMEM_ADDR attribute
    case ATTR_BRANCH_PIBMEM_ADDR:
        l_attrSize = sizeof(ATTR_BRANCH_PIBMEM_ADDR_type);
        break;

    // Get size for ATTR_CHIP_REGIONS_TO_ENABLE attribute
    case ATTR_CHIP_REGIONS_TO_ENABLE:
        l_attrSize = sizeof(ATTR_CHIP_REGIONS_TO_ENABLE_type);
        break;

    // Get size for ATTR_DEVICE_ID attribute
    case ATTR_DEVICE_ID:
        l_attrSize = sizeof(ATTR_DEVICE_ID_type);
        break;

    // Get size for ATTR_ECID attribute
    case ATTR_ECID:
        l_attrSize = sizeof(ATTR_ECID_type);
        break;

    // Get size for ATTR_I2C_BUS_DIV_NEST attribute
    case ATTR_I2C_BUS_DIV_NEST:
        l_attrSize = sizeof(ATTR_I2C_BUS_DIV_NEST_type);
        break;

    // Get size for ATTR_LEN_OF_SEEPROM_DATA attribute
    case ATTR_LEN_OF_SEEPROM_DATA:
        l_attrSize = sizeof(ATTR_LEN_OF_SEEPROM_DATA_type);
        break;

    // Get size for ATTR_MB_BIT_RATE_DIVISOR_PLL attribute
    case ATTR_MB_BIT_RATE_DIVISOR_PLL:
        l_attrSize = sizeof(ATTR_MB_BIT_RATE_DIVISOR_PLL_type);
        break;

    // Get size for ATTR_MB_BIT_RATE_DIVISOR_REFCLK attribute
    case ATTR_MB_BIT_RATE_DIVISOR_REFCLK:
        l_attrSize = sizeof(ATTR_MB_BIT_RATE_DIVISOR_REFCLK_type);
        break;

    // Get size for ATTR_MC_SYNC_MODE attribute
    case ATTR_MC_SYNC_MODE:
        l_attrSize = sizeof(ATTR_MC_SYNC_MODE_type);
        break;

    // Get size for ATTR_PROC_PB_BNDY_DMIPLL_DATA attribute
    case ATTR_PROC_PB_BNDY_DMIPLL_DATA:
        l_attrSize = sizeof(ATTR_PROC_PB_BNDY_DMIPLL_DATA_type);
        break;

    // Get size for ATTR_PROC_PB_BNDY_DMIPLL_FOR_DCCAL_DATA attribute
    case ATTR_PROC_PB_BNDY_DMIPLL_FOR_DCCAL_DATA:
        l_attrSize = sizeof(ATTR_PROC_PB_BNDY_DMIPLL_FOR_DCCAL_DATA_type);
        break;

    // Get size for ATTR_PROC_PERV_BNDY_PLL_DATA attribute
    case ATTR_PROC_PERV_BNDY_PLL_DATA:
        l_attrSize = sizeof(ATTR_PROC_PERV_BNDY_PLL_DATA_type);
        break;

    // Get size for ATTR_PROC_SBE_MASTER_CHIP attribute
    case ATTR_PROC_SBE_MASTER_CHIP:
        l_attrSize = sizeof(ATTR_PROC_SBE_MASTER_CHIP_type);
        break;

    // Get size for ATTR_SBE_SEEPROM_I2C_DEVICE_ADDRESS attribute
    case ATTR_SBE_SEEPROM_I2C_DEVICE_ADDRESS:
        l_attrSize = sizeof(ATTR_SBE_SEEPROM_I2C_DEVICE_ADDRESS_type);
        break;

    // Get size for ATTR_SBE_SEEPROM_I2C_PORT attribute
    case ATTR_SBE_SEEPROM_I2C_PORT:
        l_attrSize = sizeof(ATTR_SBE_SEEPROM_I2C_PORT_type);
        break;

    // Get size for ATTR_START_PIBMEM_ADDR attribute
    case ATTR_START_PIBMEM_ADDR:
        l_attrSize = sizeof(ATTR_START_PIBMEM_ADDR_type);
        break;

    // Get size for ATTR_START_SEEPROM_ADDR attribute
    case ATTR_START_SEEPROM_ADDR:
        l_attrSize = sizeof(ATTR_START_SEEPROM_ADDR_type);
        break;

    // Get size for ATTR_WAIT_N0 attribute
    case ATTR_WAIT_N0:
        l_attrSize = sizeof(ATTR_WAIT_N0_type);
        break;

    // Get size for ATTR_WAIT_N1 attribute
    case ATTR_WAIT_N1:
        l_attrSize = sizeof(ATTR_WAIT_N1_type);
        break;

    // Get size for ATTR_WAIT_N2 attribute
    case ATTR_WAIT_N2:
        l_attrSize = sizeof(ATTR_WAIT_N2_type);
        break;

    // Get size for ATTR_WAIT_N3 attribute
    case ATTR_WAIT_N3:
        l_attrSize = sizeof(ATTR_WAIT_N3_type);
        break;

    // Get size for ATTR_SYS_FORCE_ALL_CORES attribute
    case ATTR_SYS_FORCE_ALL_CORES:
        l_attrSize = sizeof(ATTR_SYS_FORCE_ALL_CORES_type);
        break;

    // Get size for ATTR_MASTER_CORE attribute
    case ATTR_MASTER_CORE:
        l_attrSize = sizeof(ATTR_MASTER_CORE_type);
        break;

    // Get size for ATTR_MASTER_EX attribute
    case ATTR_MASTER_EX:
        l_attrSize = sizeof(ATTR_MASTER_EX_type);
        break;

    // Get size for ATTR_SECURITY_ENABLE attribute
    case ATTR_SECURITY_ENABLE:
        l_attrSize = sizeof(ATTR_SECURITY_ENABLE_type);
        break;

    // Get size for ATTR_PFET_OFF_CONTROLS attribute
    case ATTR_PFET_OFF_CONTROLS:
        l_attrSize = sizeof(ATTR_PFET_OFF_CONTROLS_type);
        break;

    // Get size for ATTR_OBUS_RATIO_VALUE attribute
    case ATTR_OBUS_RATIO_VALUE:
        l_attrSize = sizeof(ATTR_OBUS_RATIO_VALUE_type);
        break;

    // Get size for ATTR_PIBMEM_REPAIR0 attribute
    case ATTR_PIBMEM_REPAIR0:
        l_attrSize = sizeof(ATTR_PIBMEM_REPAIR0_type);
        break;

    // Get size for ATTR_PIBMEM_REPAIR1 attribute
    case ATTR_PIBMEM_REPAIR1:
        l_attrSize = sizeof(ATTR_PIBMEM_REPAIR1_type);
        break;

    // Get size for ATTR_PIBMEM_REPAIR2 attribute
    case ATTR_PIBMEM_REPAIR2:
        l_attrSize = sizeof(ATTR_PIBMEM_REPAIR2_type);
        break;

    // Get size for ATTR_SENSEADJ_STEP attribute
    case ATTR_SENSEADJ_STEP:
        l_attrSize = sizeof(ATTR_SENSEADJ_STEP_type);
        break;

    // Get size for ATTR_CP_FILTER_BYPASS attribute
    case ATTR_CP_FILTER_BYPASS:
        l_attrSize = sizeof(ATTR_CP_FILTER_BYPASS_type);
        break;

    // Get size for ATTR_SS_FILTER_BYPASS attribute
    case ATTR_SS_FILTER_BYPASS:
        l_attrSize = sizeof(ATTR_SS_FILTER_BYPASS_type);
        break;

    // Get size for ATTR_IO_FILTER_BYPASS attribute
    case ATTR_IO_FILTER_BYPASS:
        l_attrSize = sizeof(ATTR_IO_FILTER_BYPASS_type);
        break;

    // Get size for ATTR_DPLL_BYPASS attribute
    case ATTR_DPLL_BYPASS:
        l_attrSize = sizeof(ATTR_DPLL_BYPASS_type);
        break;

    // Get size for ATTR_NEST_MEM_X_O_PCI_BYPASS attribute
    case ATTR_NEST_MEM_X_O_PCI_BYPASS:
        l_attrSize = sizeof(ATTR_NEST_MEM_X_O_PCI_BYPASS_type);
        break;

    // Get size for ATTR_TARGET_HAS_POWER attribute
    case ATTR_TARGET_HAS_POWER:
        l_attrSize = sizeof(ATTR_TARGET_HAS_POWER_type);
        break;

    // Get size for ATTR_TARGET_HAS_CLOCK attribute
    case ATTR_TARGET_HAS_CLOCK:
        l_attrSize = sizeof(ATTR_TARGET_HAS_CLOCK_type);
        break;

    // Get size for ATTR_TARGET_IS_SCOMMABLE attribute
    case ATTR_TARGET_IS_SCOMMABLE:
        l_attrSize = sizeof(ATTR_TARGET_IS_SCOMMABLE_type);
        break;

    // Get size for ATTR_SBE_SYS_CONFIG attribute
    case ATTR_SBE_SYS_CONFIG:
        l_attrSize = sizeof(ATTR_SBE_SYS_CONFIG_type);
        break;

    // Get size for ATTR_CP_REFCLOCK_RCVR_TERM attribute
    case ATTR_CP_REFCLOCK_RCVR_TERM:
        l_attrSize = sizeof(ATTR_CP_REFCLOCK_RCVR_TERM_type);
        break;

    // Get size for ATTR_IO_REFCLOCK_RCVR_TERM attribute
    case ATTR_IO_REFCLOCK_RCVR_TERM:
        l_attrSize = sizeof(ATTR_IO_REFCLOCK_RCVR_TERM_type);
        break;

    // Get size for ATTR_SECTOR_BUFFER_STRENGTH attribute
    case ATTR_SECTOR_BUFFER_STRENGTH:
        l_attrSize = sizeof(ATTR_SECTOR_BUFFER_STRENGTH_type);
        break;

    // Get size for ATTR_PULSE_MODE_ENABLE attribute
    case ATTR_PULSE_MODE_ENABLE:
        l_attrSize = sizeof(ATTR_PULSE_MODE_ENABLE_type);
        break;

    // Get size for ATTR_PULSE_MODE_VALUE attribute
    case ATTR_PULSE_MODE_VALUE:
        l_attrSize = sizeof(ATTR_PULSE_MODE_VALUE_type);
        break;

    // Get size for ATTR_NDL_MESHCTRL_SETUP attribute
    case ATTR_NDL_MESHCTRL_SETUP:
        l_attrSize = sizeof(ATTR_NDL_MESHCTRL_SETUP_type);
        break;

    // Get size for ATTR_START_CBS_FIFO_RESET_SKIP attribute
    case ATTR_START_CBS_FIFO_RESET_SKIP:
        l_attrSize = sizeof(ATTR_START_CBS_FIFO_RESET_SKIP_type);
        break;

    // Get size for ATTR_SYSTEM_CORECACHE_SKEWADJ_DISABLE attribute
    case ATTR_SYSTEM_CORECACHE_SKEWADJ_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_CORECACHE_SKEWADJ_DISABLE_type);
        break;

    // Get size for ATTR_SYSTEM_CORECACHE_DCADJ_DISABLE attribute
    case ATTR_SYSTEM_CORECACHE_DCADJ_DISABLE:
        l_attrSize = sizeof(ATTR_SYSTEM_CORECACHE_DCADJ_DISABLE_type);
        break;

    // Get size for ATTR_PROC_MEM_TO_USE attribute
    case ATTR_PROC_MEM_TO_USE:
        l_attrSize = sizeof(ATTR_PROC_MEM_TO_USE_type);
        break;

    // Get size for ATTR_ORIG_FIR_SETTINGS_ACTION0 attribute
    case ATTR_ORIG_FIR_SETTINGS_ACTION0:
        l_attrSize = sizeof(ATTR_ORIG_FIR_SETTINGS_ACTION0_type);
        break;

    // Get size for ATTR_ORIG_FIR_SETTINGS_ACTION1 attribute
    case ATTR_ORIG_FIR_SETTINGS_ACTION1:
        l_attrSize = sizeof(ATTR_ORIG_FIR_SETTINGS_ACTION1_type);
        break;

    // Get size for ATTR_LPC_CONSOLE_CNFG attribute
    case ATTR_LPC_CONSOLE_CNFG:
        l_attrSize = sizeof(ATTR_LPC_CONSOLE_CNFG_type);
        break;

    // Get size for ATTR_X_EREPAIR_THRESHOLD_FIELD attribute
    case ATTR_X_EREPAIR_THRESHOLD_FIELD:
        l_attrSize = sizeof(ATTR_X_EREPAIR_THRESHOLD_FIELD_type);
        break;

    // Get size for ATTR_X_EREPAIR_THRESHOLD_MNFG attribute
    case ATTR_X_EREPAIR_THRESHOLD_MNFG:
        l_attrSize = sizeof(ATTR_X_EREPAIR_THRESHOLD_MNFG_type);
        break;

    // Get size for ATTR_O_EREPAIR_THRESHOLD_FIELD attribute
    case ATTR_O_EREPAIR_THRESHOLD_FIELD:
        l_attrSize = sizeof(ATTR_O_EREPAIR_THRESHOLD_FIELD_type);
        break;

    // Get size for ATTR_O_EREPAIR_THRESHOLD_MNFG attribute
    case ATTR_O_EREPAIR_THRESHOLD_MNFG:
        l_attrSize = sizeof(ATTR_O_EREPAIR_THRESHOLD_MNFG_type);
        break;

    // Get size for ATTR_DMI_EREPAIR_THRESHOLD_FIELD attribute
    case ATTR_DMI_EREPAIR_THRESHOLD_FIELD:
        l_attrSize = sizeof(ATTR_DMI_EREPAIR_THRESHOLD_FIELD_type);
        break;

    // Get size for ATTR_DMI_EREPAIR_THRESHOLD_MNFG attribute
    case ATTR_DMI_EREPAIR_THRESHOLD_MNFG:
        l_attrSize = sizeof(ATTR_DMI_EREPAIR_THRESHOLD_MNFG_type);
        break;

    // Get size for ATTR_MSS_VPD_CKE_MAP attribute
    case ATTR_MSS_VPD_CKE_MAP:
        l_attrSize = sizeof(ATTR_MSS_VPD_CKE_MAP_type);
        break;

    // Get size for ATTR_IO_OBUS_DCCAL_FLAGS attribute
    case ATTR_IO_OBUS_DCCAL_FLAGS:
        l_attrSize = sizeof(ATTR_IO_OBUS_DCCAL_FLAGS_type);
        break;

    // Get size for ATTR_IO_OBUS_PAT_A_DETECT_RUN attribute
    case ATTR_IO_OBUS_PAT_A_DETECT_RUN:
        l_attrSize = sizeof(ATTR_IO_OBUS_PAT_A_DETECT_RUN_type);
        break;

    // Get size for ATTR_IO_OBUS_PAT_A_DETECT_RX_AMP_VALUE attribute
    case ATTR_IO_OBUS_PAT_A_DETECT_RX_AMP_VALUE:
        l_attrSize = sizeof(ATTR_IO_OBUS_PAT_A_DETECT_RX_AMP_VALUE_type);
        break;

    // Get size for ATTR_IO_OBUS_PAT_A_CAPTURE attribute
    case ATTR_IO_OBUS_PAT_A_CAPTURE:
        l_attrSize = sizeof(ATTR_IO_OBUS_PAT_A_CAPTURE_type);
        break;

    // Get size for ATTR_IO_OBUS_TRAIN_FOR_RECOVERY attribute
    case ATTR_IO_OBUS_TRAIN_FOR_RECOVERY:
        l_attrSize = sizeof(ATTR_IO_OBUS_TRAIN_FOR_RECOVERY_type);
        break;

    // Get size for ATTR_IO_OBUS_LINK_SPARE_MARK attribute
    case ATTR_IO_OBUS_LINK_SPARE_MARK:
        l_attrSize = sizeof(ATTR_IO_OBUS_LINK_SPARE_MARK_type);
        break;

    // Get size for ATTR_IO_OBUS_LANE_PDWN attribute
    case ATTR_IO_OBUS_LANE_PDWN:
        l_attrSize = sizeof(ATTR_IO_OBUS_LANE_PDWN_type);
        break;

    // Get size for ATTR_IO_O_DEBUG attribute
    case ATTR_IO_O_DEBUG:
        l_attrSize = sizeof(ATTR_IO_O_DEBUG_type);
        break;

    // Get size for ATTR_IO_O_MFG_CHK attribute
    case ATTR_IO_O_MFG_CHK:
        l_attrSize = sizeof(ATTR_IO_O_MFG_CHK_type);
        break;

    // Get size for ATTR_IO_O_MFG_MIN_EYE_WIDTH attribute
    case ATTR_IO_O_MFG_MIN_EYE_WIDTH:
        l_attrSize = sizeof(ATTR_IO_O_MFG_MIN_EYE_WIDTH_type);
        break;

    // Get size for ATTR_IO_OBUS_TX_MARGIN_RATIO attribute
    case ATTR_IO_OBUS_TX_MARGIN_RATIO:
        l_attrSize = sizeof(ATTR_IO_OBUS_TX_MARGIN_RATIO_type);
        break;

    // Get size for ATTR_IO_OBUS_TX_FFE_PRECURSOR attribute
    case ATTR_IO_OBUS_TX_FFE_PRECURSOR:
        l_attrSize = sizeof(ATTR_IO_OBUS_TX_FFE_PRECURSOR_type);
        break;

    // Get size for ATTR_IO_OBUS_TX_FFE_POSTCURSOR attribute
    case ATTR_IO_OBUS_TX_FFE_POSTCURSOR:
        l_attrSize = sizeof(ATTR_IO_OBUS_TX_FFE_POSTCURSOR_type);
        break;

    // Get size for ATTR_IO_OLLFIR_MASK attribute
    case ATTR_IO_OLLFIR_MASK:
        l_attrSize = sizeof(ATTR_IO_OLLFIR_MASK_type);
        break;

    // Get size for ATTR_IO_PB_IOOFIR_MASK attribute
    case ATTR_IO_PB_IOOFIR_MASK:
        l_attrSize = sizeof(ATTR_IO_PB_IOOFIR_MASK_type);
        break;

    // Get size for ATTR_IO_O_CHANNEL_TYPE attribute
    case ATTR_IO_O_CHANNEL_TYPE:
        l_attrSize = sizeof(ATTR_IO_O_CHANNEL_TYPE_type);
        break;

    // Get size for ATTR_IO_O_MFG_STRESS_PR_OFFSET_EVEN attribute
    case ATTR_IO_O_MFG_STRESS_PR_OFFSET_EVEN:
        l_attrSize = sizeof(ATTR_IO_O_MFG_STRESS_PR_OFFSET_EVEN_type);
        break;

    // Get size for ATTR_IO_O_MFG_STRESS_PR_OFFSET_ODD attribute
    case ATTR_IO_O_MFG_STRESS_PR_OFFSET_ODD:
        l_attrSize = sizeof(ATTR_IO_O_MFG_STRESS_PR_OFFSET_ODD_type);
        break;

    // Get size for ATTR_IO_O_MNFG_ERROR_THRESHOLD attribute
    case ATTR_IO_O_MNFG_ERROR_THRESHOLD:
        l_attrSize = sizeof(ATTR_IO_O_MNFG_ERROR_THRESHOLD_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_0_VERSION_LAYOUT attribute
    case ATTR_MSS_VPD_MT_0_VERSION_LAYOUT:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_0_VERSION_LAYOUT_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_1_VERSION_DATA attribute
    case ATTR_MSS_VPD_MT_1_VERSION_DATA:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_1_VERSION_DATA_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_2_SIGNATURE_HASH attribute
    case ATTR_MSS_VPD_MT_2_SIGNATURE_HASH:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_2_SIGNATURE_HASH_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CA attribute
    case ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CA:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CA_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CKE attribute
    case ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CKE:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CKE_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CS attribute
    case ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CS:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DIMM_RCD_IBT_CS_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DIMM_RCD_IBT_ODT attribute
    case ATTR_MSS_VPD_MT_DIMM_RCD_IBT_ODT:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DIMM_RCD_IBT_ODT_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS attribute
    case ATTR_MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DRAM_DRV_IMP_DQ_DQS_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DRAM_RTT_NOM attribute
    case ATTR_MSS_VPD_MT_DRAM_RTT_NOM:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DRAM_RTT_NOM_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DRAM_RTT_PARK attribute
    case ATTR_MSS_VPD_MT_DRAM_RTT_PARK:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DRAM_RTT_PARK_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_DRAM_RTT_WR attribute
    case ATTR_MSS_VPD_MT_DRAM_RTT_WR:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_DRAM_RTT_WR_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_BIAS_TRIM attribute
    case ATTR_MSS_VPD_MT_MC_BIAS_TRIM:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_BIAS_TRIM_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP attribute
    case ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_RD_UP_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN attribute
    case ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_DOWN_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP attribute
    case ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DQ_ACBOOST_WR_UP_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DQ_CTLE_CAP attribute
    case ATTR_MSS_VPD_MT_MC_DQ_CTLE_CAP:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DQ_CTLE_CAP_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DQ_CTLE_RES attribute
    case ATTR_MSS_VPD_MT_MC_DQ_CTLE_RES:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DQ_CTLE_RES_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DRV_IMP_CLK attribute
    case ATTR_MSS_VPD_MT_MC_DRV_IMP_CLK:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DRV_IMP_CLK_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR attribute
    case ATTR_MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DRV_IMP_CMD_ADDR_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DRV_IMP_CNTL attribute
    case ATTR_MSS_VPD_MT_MC_DRV_IMP_CNTL:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DRV_IMP_CNTL_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DRV_IMP_CSCID attribute
    case ATTR_MSS_VPD_MT_MC_DRV_IMP_CSCID:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DRV_IMP_CSCID_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_DRV_IMP_DQ_DQS attribute
    case ATTR_MSS_VPD_MT_MC_DRV_IMP_DQ_DQS:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_DRV_IMP_DQ_DQS_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_MC_RCV_IMP_DQ_DQS attribute
    case ATTR_MSS_VPD_MT_MC_RCV_IMP_DQ_DQS:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_MC_RCV_IMP_DQ_DQS_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_ODT_RD attribute
    case ATTR_MSS_VPD_MT_ODT_RD:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_ODT_RD_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_ODT_WR attribute
    case ATTR_MSS_VPD_MT_ODT_WR:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_ODT_WR_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_PREAMBLE attribute
    case ATTR_MSS_VPD_MT_PREAMBLE:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_PREAMBLE_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_VREF_DRAM_WR attribute
    case ATTR_MSS_VPD_MT_VREF_DRAM_WR:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_VREF_DRAM_WR_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_VREF_MC_RD attribute
    case ATTR_MSS_VPD_MT_VREF_MC_RD:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_VREF_MC_RD_type);
        break;

    // Get size for ATTR_MSS_VPD_MT_WINDAGE_RD_CTR attribute
    case ATTR_MSS_VPD_MT_WINDAGE_RD_CTR:
        l_attrSize = sizeof(ATTR_MSS_VPD_MT_WINDAGE_RD_CTR_type);
        break;

    // Get size for ATTR_PROC_DSTLCFG_MMIO_ADDRBIT_POS attribute
    case ATTR_PROC_DSTLCFG_MMIO_ADDRBIT_POS:
        l_attrSize = sizeof(ATTR_PROC_DSTLCFG_MMIO_ADDRBIT_POS_type);
        break;

    // Get size for ATTR_PROC_ENABLE_DL_TMPL_1 attribute
    case ATTR_PROC_ENABLE_DL_TMPL_1:
        l_attrSize = sizeof(ATTR_PROC_ENABLE_DL_TMPL_1_type);
        break;

    // Get size for ATTR_PROC_ENABLE_DL_TMPL_4 attribute
    case ATTR_PROC_ENABLE_DL_TMPL_4:
        l_attrSize = sizeof(ATTR_PROC_ENABLE_DL_TMPL_4_type);
        break;

    // Get size for ATTR_PROC_ENABLE_DL_TMPL_7 attribute
    case ATTR_PROC_ENABLE_DL_TMPL_7:
        l_attrSize = sizeof(ATTR_PROC_ENABLE_DL_TMPL_7_type);
        break;

    // Get size for ATTR_PROC_ENABLE_DL_TMPL_A attribute
    case ATTR_PROC_ENABLE_DL_TMPL_A:
        l_attrSize = sizeof(ATTR_PROC_ENABLE_DL_TMPL_A_type);
        break;

    // Get size for ATTR_PROC_TMPL_0_PACING attribute
    case ATTR_PROC_TMPL_0_PACING:
        l_attrSize = sizeof(ATTR_PROC_TMPL_0_PACING_type);
        break;

    // Get size for ATTR_PROC_TMPL_1_PACING attribute
    case ATTR_PROC_TMPL_1_PACING:
        l_attrSize = sizeof(ATTR_PROC_TMPL_1_PACING_type);
        break;

    // Get size for ATTR_PROC_TMPL_4_PACING attribute
    case ATTR_PROC_TMPL_4_PACING:
        l_attrSize = sizeof(ATTR_PROC_TMPL_4_PACING_type);
        break;

    // Get size for ATTR_PROC_TMPL_7_PACING attribute
    case ATTR_PROC_TMPL_7_PACING:
        l_attrSize = sizeof(ATTR_PROC_TMPL_7_PACING_type);
        break;

    // Get size for ATTR_PROC_TMPL_A_PACING attribute
    case ATTR_PROC_TMPL_A_PACING:
        l_attrSize = sizeof(ATTR_PROC_TMPL_A_PACING_type);
        break;

    // Get size for ATTR_PROC_OMI_OC_MAJOR_VER attribute
    case ATTR_PROC_OMI_OC_MAJOR_VER:
        l_attrSize = sizeof(ATTR_PROC_OMI_OC_MAJOR_VER_type);
        break;

    // Get size for ATTR_PROC_OMI_OC_MINOR_VER attribute
    case ATTR_PROC_OMI_OC_MINOR_VER:
        l_attrSize = sizeof(ATTR_PROC_OMI_OC_MINOR_VER_type);
        break;

    // Get size for ATTR_HB_HRMOR_BYTES attribute
    case ATTR_HB_HRMOR_BYTES:
        l_attrSize = sizeof(ATTR_HB_HRMOR_BYTES_type);
        break;

    // Get size for ATTR_FREQ_PROC_REFCLOCK_KHZ attribute
    case ATTR_FREQ_PROC_REFCLOCK_KHZ:
        l_attrSize = sizeof(ATTR_FREQ_PROC_REFCLOCK_KHZ_type);
        break;

    // Get size for ATTR_FREQ_DPLL_REFCLOCK_KHZ attribute
    case ATTR_FREQ_DPLL_REFCLOCK_KHZ:
        l_attrSize = sizeof(ATTR_FREQ_DPLL_REFCLOCK_KHZ_type);
        break;

    // Get size for ATTR_FREQ_MEM_REFCLOCK attribute
    case ATTR_FREQ_MEM_REFCLOCK:
        l_attrSize = sizeof(ATTR_FREQ_MEM_REFCLOCK_type);
        break;

    // Get size for ATTR_MAX_ALLOWED_DIMM_FREQ attribute
    case ATTR_MAX_ALLOWED_DIMM_FREQ:
        l_attrSize = sizeof(ATTR_MAX_ALLOWED_DIMM_FREQ_type);
        break;

    // Get size for ATTR_MEMVPD_POS attribute
    case ATTR_MEMVPD_POS:
        l_attrSize = sizeof(ATTR_MEMVPD_POS_type);
        break;

    // Get size for ATTR_REQUIRED_SYNCH_MODE attribute
    case ATTR_REQUIRED_SYNCH_MODE:
        l_attrSize = sizeof(ATTR_REQUIRED_SYNCH_MODE_type);
        break;

    // Get size for ATTR_NHTM_TRACE_TYPE attribute
    case ATTR_NHTM_TRACE_TYPE:
        l_attrSize = sizeof(ATTR_NHTM_TRACE_TYPE_type);
        break;

    // Get size for ATTR_CHTM_TRACE_TYPE attribute
    case ATTR_CHTM_TRACE_TYPE:
        l_attrSize = sizeof(ATTR_CHTM_TRACE_TYPE_type);
        break;

    // Get size for ATTR_HTMSC_TTYPEFILT_PAT attribute
    case ATTR_HTMSC_TTYPEFILT_PAT:
        l_attrSize = sizeof(ATTR_HTMSC_TTYPEFILT_PAT_type);
        break;

    // Get size for ATTR_HTMSC_TSIZEFILT_PAT attribute
    case ATTR_HTMSC_TSIZEFILT_PAT:
        l_attrSize = sizeof(ATTR_HTMSC_TSIZEFILT_PAT_type);
        break;

    // Get size for ATTR_HTMSC_TTYPEFILT_MASK attribute
    case ATTR_HTMSC_TTYPEFILT_MASK:
        l_attrSize = sizeof(ATTR_HTMSC_TTYPEFILT_MASK_type);
        break;

    // Get size for ATTR_HTMSC_TSIZEFILT_MASK attribute
    case ATTR_HTMSC_TSIZEFILT_MASK:
        l_attrSize = sizeof(ATTR_HTMSC_TSIZEFILT_MASK_type);
        break;

    // Get size for ATTR_HTMSC_TTYPEFILT_INVERT attribute
    case ATTR_HTMSC_TTYPEFILT_INVERT:
        l_attrSize = sizeof(ATTR_HTMSC_TTYPEFILT_INVERT_type);
        break;

    // Get size for ATTR_HTMSC_CRESPFILT_INVERT attribute
    case ATTR_HTMSC_CRESPFILT_INVERT:
        l_attrSize = sizeof(ATTR_HTMSC_CRESPFILT_INVERT_type);
        break;

    // Get size for ATTR_HTMSC_FILT_PAT attribute
    case ATTR_HTMSC_FILT_PAT:
        l_attrSize = sizeof(ATTR_HTMSC_FILT_PAT_type);
        break;

    // Get size for ATTR_HTMSC_FILT_CRESP_PAT attribute
    case ATTR_HTMSC_FILT_CRESP_PAT:
        l_attrSize = sizeof(ATTR_HTMSC_FILT_CRESP_PAT_type);
        break;

    // Get size for ATTR_HTMSC_FILT_MASK attribute
    case ATTR_HTMSC_FILT_MASK:
        l_attrSize = sizeof(ATTR_HTMSC_FILT_MASK_type);
        break;

    // Get size for ATTR_HTMSC_FILT_CRESP_MASK attribute
    case ATTR_HTMSC_FILT_CRESP_MASK:
        l_attrSize = sizeof(ATTR_HTMSC_FILT_CRESP_MASK_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_CONTENT_SEL attribute
    case ATTR_NHTM_HTMSC_MODE_CONTENT_SEL:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_CONTENT_SEL_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_CAPTURE_GENERATED_WRITES attribute
    case ATTR_NHTM_HTMSC_MODE_CAPTURE_GENERATED_WRITES:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_CAPTURE_GENERATED_WRITES_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_CAPTURE_ENABLE_FILTER_ALL attribute
    case ATTR_NHTM_HTMSC_MODE_CAPTURE_ENABLE_FILTER_ALL:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_CAPTURE_ENABLE_FILTER_ALL_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_CAPTURE_PRECISE_CRESP_MODE attribute
    case ATTR_NHTM_HTMSC_MODE_CAPTURE_PRECISE_CRESP_MODE:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_CAPTURE_PRECISE_CRESP_MODE_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_CAPTURE_LIMIT_MEM_ALLOCATION attribute
    case ATTR_NHTM_HTMSC_MODE_CAPTURE_LIMIT_MEM_ALLOCATION:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_CAPTURE_LIMIT_MEM_ALLOCATION_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_CAPTURE_PMISC_ONLY_CMD attribute
    case ATTR_NHTM_HTMSC_MODE_CAPTURE_PMISC_ONLY_CMD:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_CAPTURE_PMISC_ONLY_CMD_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_SYNC_STAMP_FORCE attribute
    case ATTR_NHTM_HTMSC_MODE_SYNC_STAMP_FORCE:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_SYNC_STAMP_FORCE_type);
        break;

    // Get size for ATTR_NHTM_HTMSC_MODE_WRITETOIO attribute
    case ATTR_NHTM_HTMSC_MODE_WRITETOIO:
        l_attrSize = sizeof(ATTR_NHTM_HTMSC_MODE_WRITETOIO_type);
        break;

    // Get size for ATTR_CHTM_HTMSC_MODE_CONTENT_SEL attribute
    case ATTR_CHTM_HTMSC_MODE_CONTENT_SEL:
        l_attrSize = sizeof(ATTR_CHTM_HTMSC_MODE_CONTENT_SEL_type);
        break;

    // Get size for ATTR_CHTM_HTMSC_MODE_CAPTURE attribute
    case ATTR_CHTM_HTMSC_MODE_CAPTURE:
        l_attrSize = sizeof(ATTR_CHTM_HTMSC_MODE_CAPTURE_type);
        break;

    // Get size for ATTR_CHTM_HTMSC_MODE_CORE_INSTR_STALL attribute
    case ATTR_CHTM_HTMSC_MODE_CORE_INSTR_STALL:
        l_attrSize = sizeof(ATTR_CHTM_HTMSC_MODE_CORE_INSTR_STALL_type);
        break;

    // Get size for ATTR_HTMSC_MODE_WRAP attribute
    case ATTR_HTMSC_MODE_WRAP:
        l_attrSize = sizeof(ATTR_HTMSC_MODE_WRAP_type);
        break;

    // Get size for ATTR_HTMSC_MODE_DIS_TSTAMP attribute
    case ATTR_HTMSC_MODE_DIS_TSTAMP:
        l_attrSize = sizeof(ATTR_HTMSC_MODE_DIS_TSTAMP_type);
        break;

    // Get size for ATTR_HTMSC_MODE_SINGLE_TSTAMP attribute
    case ATTR_HTMSC_MODE_SINGLE_TSTAMP:
        l_attrSize = sizeof(ATTR_HTMSC_MODE_SINGLE_TSTAMP_type);
        break;

    // Get size for ATTR_HTMSC_MODE_MARKERS_ONLY attribute
    case ATTR_HTMSC_MODE_MARKERS_ONLY:
        l_attrSize = sizeof(ATTR_HTMSC_MODE_MARKERS_ONLY_type);
        break;

    // Get size for ATTR_HTMSC_MODE_DIS_FORCE_GROUP_SCOPE attribute
    case ATTR_HTMSC_MODE_DIS_FORCE_GROUP_SCOPE:
        l_attrSize = sizeof(ATTR_HTMSC_MODE_DIS_FORCE_GROUP_SCOPE_type);
        break;

    // Get size for ATTR_HTMSC_MODE_VGTARGET attribute
    case ATTR_HTMSC_MODE_VGTARGET:
        l_attrSize = sizeof(ATTR_HTMSC_MODE_VGTARGET_type);
        break;

    // Get size for ATTR_HTMSC_MEM_SCOPE attribute
    case ATTR_HTMSC_MEM_SCOPE:
        l_attrSize = sizeof(ATTR_HTMSC_MEM_SCOPE_type);
        break;

    // Get size for ATTR_HTMSC_MEM_PRIORITY attribute
    case ATTR_HTMSC_MEM_PRIORITY:
        l_attrSize = sizeof(ATTR_HTMSC_MEM_PRIORITY_type);
        break;

    // Get size for ATTR_NHTM_CTRL_TRIG attribute
    case ATTR_NHTM_CTRL_TRIG:
        l_attrSize = sizeof(ATTR_NHTM_CTRL_TRIG_type);
        break;

    // Get size for ATTR_NHTM_CTRL_MARK attribute
    case ATTR_NHTM_CTRL_MARK:
        l_attrSize = sizeof(ATTR_NHTM_CTRL_MARK_type);
        break;

    // Get size for ATTR_CHTM_CTRL_TRIG attribute
    case ATTR_CHTM_CTRL_TRIG:
        l_attrSize = sizeof(ATTR_CHTM_CTRL_TRIG_type);
        break;

    // Get size for ATTR_CHTM_CTRL_MARK attribute
    case ATTR_CHTM_CTRL_MARK:
        l_attrSize = sizeof(ATTR_CHTM_CTRL_MARK_type);
        break;

    // Get size for ATTR_HTMSC_CTRL_DBG0_STOP attribute
    case ATTR_HTMSC_CTRL_DBG0_STOP:
        l_attrSize = sizeof(ATTR_HTMSC_CTRL_DBG0_STOP_type);
        break;

    // Get size for ATTR_HTMSC_CTRL_DBG1_STOP attribute
    case ATTR_HTMSC_CTRL_DBG1_STOP:
        l_attrSize = sizeof(ATTR_HTMSC_CTRL_DBG1_STOP_type);
        break;

    // Get size for ATTR_HTMSC_CTRL_RUN_STOP attribute
    case ATTR_HTMSC_CTRL_RUN_STOP:
        l_attrSize = sizeof(ATTR_HTMSC_CTRL_RUN_STOP_type);
        break;

    // Get size for ATTR_HTMSC_CTRL_OTHER_DBG0_STOP attribute
    case ATTR_HTMSC_CTRL_OTHER_DBG0_STOP:
        l_attrSize = sizeof(ATTR_HTMSC_CTRL_OTHER_DBG0_STOP_type);
        break;

    // Get size for ATTR_HTMSC_CTRL_XSTOP_STOP attribute
    case ATTR_HTMSC_CTRL_XSTOP_STOP:
        l_attrSize = sizeof(ATTR_HTMSC_CTRL_XSTOP_STOP_type);
        break;

    // Get size for ATTR_HTMSC_CTRL_CHIP0_STOP attribute
    case ATTR_HTMSC_CTRL_CHIP0_STOP:
        l_attrSize = sizeof(ATTR_HTMSC_CTRL_CHIP0_STOP_type);
        break;

    // Get size for ATTR_HTMSC_CTRL_CHIP1_STOP attribute
    case ATTR_HTMSC_CTRL_CHIP1_STOP:
        l_attrSize = sizeof(ATTR_HTMSC_CTRL_CHIP1_STOP_type);
        break;

    // Get size for ATTR_HTMSC_IMA_PDBAR_SPLIT_CORE_MODE attribute
    case ATTR_HTMSC_IMA_PDBAR_SPLIT_CORE_MODE:
        l_attrSize = sizeof(ATTR_HTMSC_IMA_PDBAR_SPLIT_CORE_MODE_type);
        break;

    // Get size for ATTR_HTMSC_IMA_PDBAR_SCOPE attribute
    case ATTR_HTMSC_IMA_PDBAR_SCOPE:
        l_attrSize = sizeof(ATTR_HTMSC_IMA_PDBAR_SCOPE_type);
        break;

    // Get size for ATTR_HTMSC_IMA_PDBAR_ADDR attribute
    case ATTR_HTMSC_IMA_PDBAR_ADDR:
        l_attrSize = sizeof(ATTR_HTMSC_IMA_PDBAR_ADDR_type);
        break;

    // Get size for ATTR_FREQ_PB_MHZ attribute
    case ATTR_FREQ_PB_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_PB_MHZ_type);
        break;

    // Get size for ATTR_FREQ_OMI_MHZ attribute
    case ATTR_FREQ_OMI_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_OMI_MHZ_type);
        break;

    // Get size for ATTR_OMI_PLL_VCO attribute
    case ATTR_OMI_PLL_VCO:
        l_attrSize = sizeof(ATTR_OMI_PLL_VCO_type);
        break;

    // Get size for ATTR_FREQ_O_MHZ attribute
    case ATTR_FREQ_O_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_O_MHZ_type);
        break;

    // Get size for ATTR_FREQ_A_MHZ attribute
    case ATTR_FREQ_A_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_A_MHZ_type);
        break;

    // Get size for ATTR_FREQ_X_MHZ attribute
    case ATTR_FREQ_X_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_X_MHZ_type);
        break;

    // Get size for ATTR_PM_SAFE_FREQUENCY_MHZ attribute
    case ATTR_PM_SAFE_FREQUENCY_MHZ:
        l_attrSize = sizeof(ATTR_PM_SAFE_FREQUENCY_MHZ_type);
        break;

    // Get size for ATTR_PM_SAFE_VOLTAGE_MV attribute
    case ATTR_PM_SAFE_VOLTAGE_MV:
        l_attrSize = sizeof(ATTR_PM_SAFE_VOLTAGE_MV_type);
        break;

    // Get size for ATTR_FREQ_PCIE_MHZ attribute
    case ATTR_FREQ_PCIE_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_PCIE_MHZ_type);
        break;

    // Get size for ATTR_DD1_SLOW_PCI_REF_CLOCK attribute
    case ATTR_DD1_SLOW_PCI_REF_CLOCK:
        l_attrSize = sizeof(ATTR_DD1_SLOW_PCI_REF_CLOCK_type);
        break;

    // Get size for ATTR_PROC_FABRIC_ASYNC_SAFE_MODE attribute
    case ATTR_PROC_FABRIC_ASYNC_SAFE_MODE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_ASYNC_SAFE_MODE_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_BUS_WIDTH attribute
    case ATTR_PROC_FABRIC_A_BUS_WIDTH:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_BUS_WIDTH_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_BUS_WIDTH attribute
    case ATTR_PROC_FABRIC_X_BUS_WIDTH:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_BUS_WIDTH_type);
        break;

    // Get size for ATTR_PROC_FABRIC_CORE_FLOOR_RATIO attribute
    case ATTR_PROC_FABRIC_CORE_FLOOR_RATIO:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_CORE_FLOOR_RATIO_type);
        break;

    // Get size for ATTR_PROC_FABRIC_CORE_CEILING_RATIO attribute
    case ATTR_PROC_FABRIC_CORE_CEILING_RATIO:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_CORE_CEILING_RATIO_type);
        break;

    // Get size for ATTR_PROC_FABRIC_PUMP_MODE attribute
    case ATTR_PROC_FABRIC_PUMP_MODE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_PUMP_MODE_type);
        break;

    // Get size for ATTR_PROC_FABRIC_CCSM_MODE attribute
    case ATTR_PROC_FABRIC_CCSM_MODE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_CCSM_MODE_type);
        break;

    // Get size for ATTR_OPTICS_CONFIG_MODE attribute
    case ATTR_OPTICS_CONFIG_MODE:
        l_attrSize = sizeof(ATTR_OPTICS_CONFIG_MODE_type);
        break;

    // Get size for ATTR_PROC_FABRIC_SMP_OPTICS_MODE attribute
    case ATTR_PROC_FABRIC_SMP_OPTICS_MODE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_SMP_OPTICS_MODE_type);
        break;

    // Get size for ATTR_PROC_FABRIC_OPTICS_CONFIG_MODE attribute
    case ATTR_PROC_FABRIC_OPTICS_CONFIG_MODE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_OPTICS_CONFIG_MODE_type);
        break;

    // Get size for ATTR_PROC_FABRIC_CAPI_MODE attribute
    case ATTR_PROC_FABRIC_CAPI_MODE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_CAPI_MODE_type);
        break;

    // Get size for ATTR_FABRIC_PRESENT_GROUPS attribute
    case ATTR_FABRIC_PRESENT_GROUPS:
        l_attrSize = sizeof(ATTR_FABRIC_PRESENT_GROUPS_type);
        break;

    // Get size for ATTR_PROC_FABRIC_SYSTEM_ID attribute
    case ATTR_PROC_FABRIC_SYSTEM_ID:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_SYSTEM_ID_type);
        break;

    // Get size for ATTR_FABRIC_ADDR_EXTENSION_GROUP_ID attribute
    case ATTR_FABRIC_ADDR_EXTENSION_GROUP_ID:
        l_attrSize = sizeof(ATTR_FABRIC_ADDR_EXTENSION_GROUP_ID_type);
        break;

    // Get size for ATTR_FABRIC_ADDR_EXTENSION_CHIP_ID attribute
    case ATTR_FABRIC_ADDR_EXTENSION_CHIP_ID:
        l_attrSize = sizeof(ATTR_FABRIC_ADDR_EXTENSION_CHIP_ID_type);
        break;

    // Get size for ATTR_PROC_EFF_FABRIC_GROUP_ID attribute
    case ATTR_PROC_EFF_FABRIC_GROUP_ID:
        l_attrSize = sizeof(ATTR_PROC_EFF_FABRIC_GROUP_ID_type);
        break;

    // Get size for ATTR_PROC_EFF_FABRIC_CHIP_ID attribute
    case ATTR_PROC_EFF_FABRIC_CHIP_ID:
        l_attrSize = sizeof(ATTR_PROC_EFF_FABRIC_CHIP_ID_type);
        break;

    // Get size for ATTR_PROC_FABRIC_SYSTEM_MASTER_CHIP attribute
    case ATTR_PROC_FABRIC_SYSTEM_MASTER_CHIP:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_SYSTEM_MASTER_CHIP_type);
        break;

    // Get size for ATTR_PROC_FABRIC_GROUP_MASTER_CHIP attribute
    case ATTR_PROC_FABRIC_GROUP_MASTER_CHIP:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_GROUP_MASTER_CHIP_type);
        break;

    // Get size for ATTR_PROC_FABRIC_LINK_ACTIVE attribute
    case ATTR_PROC_FABRIC_LINK_ACTIVE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_LINK_ACTIVE_type);
        break;

    // Get size for ATTR_LINK_TRAIN attribute
    case ATTR_LINK_TRAIN:
        l_attrSize = sizeof(ATTR_LINK_TRAIN_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_ATTACHED_CHIP_CNFG attribute
    case ATTR_PROC_FABRIC_X_ATTACHED_CHIP_CNFG:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_ATTACHED_CHIP_CNFG_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_LINKS_CNFG attribute
    case ATTR_PROC_FABRIC_X_LINKS_CNFG:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_LINKS_CNFG_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_ATTACHED_CHIP_CNFG attribute
    case ATTR_PROC_FABRIC_A_ATTACHED_CHIP_CNFG:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_ATTACHED_CHIP_CNFG_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_LINKS_CNFG attribute
    case ATTR_PROC_FABRIC_A_LINKS_CNFG:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_LINKS_CNFG_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_ATTACHED_CHIP_ID attribute
    case ATTR_PROC_FABRIC_X_ATTACHED_CHIP_ID:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_ATTACHED_CHIP_ID_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_ATTACHED_LINK_ID attribute
    case ATTR_PROC_FABRIC_X_ATTACHED_LINK_ID:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_ATTACHED_LINK_ID_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_ATTACHED_CHIP_ID attribute
    case ATTR_PROC_FABRIC_A_ATTACHED_CHIP_ID:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_ATTACHED_CHIP_ID_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_ATTACHED_LINK_ID attribute
    case ATTR_PROC_FABRIC_A_ATTACHED_LINK_ID:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_ATTACHED_LINK_ID_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_AGGREGATE attribute
    case ATTR_PROC_FABRIC_X_AGGREGATE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_AGGREGATE_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_ADDR_DIS attribute
    case ATTR_PROC_FABRIC_X_ADDR_DIS:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_ADDR_DIS_type);
        break;

    // Get size for ATTR_PROC_FABRIC_X_LINK_DELAY attribute
    case ATTR_PROC_FABRIC_X_LINK_DELAY:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_X_LINK_DELAY_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_AGGREGATE attribute
    case ATTR_PROC_FABRIC_A_AGGREGATE:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_AGGREGATE_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_ADDR_DIS attribute
    case ATTR_PROC_FABRIC_A_ADDR_DIS:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_ADDR_DIS_type);
        break;

    // Get size for ATTR_PROC_FABRIC_A_LINK_DELAY attribute
    case ATTR_PROC_FABRIC_A_LINK_DELAY:
        l_attrSize = sizeof(ATTR_PROC_FABRIC_A_LINK_DELAY_type);
        break;

    // Get size for ATTR_PROC_EPS_GB_PERCENTAGE attribute
    case ATTR_PROC_EPS_GB_PERCENTAGE:
        l_attrSize = sizeof(ATTR_PROC_EPS_GB_PERCENTAGE_type);
        break;

    // Get size for ATTR_PROC_EPS_TABLE_TYPE attribute
    case ATTR_PROC_EPS_TABLE_TYPE:
        l_attrSize = sizeof(ATTR_PROC_EPS_TABLE_TYPE_type);
        break;

    // Get size for ATTR_PROC_EPS_READ_CYCLES_T0 attribute
    case ATTR_PROC_EPS_READ_CYCLES_T0:
        l_attrSize = sizeof(ATTR_PROC_EPS_READ_CYCLES_T0_type);
        break;

    // Get size for ATTR_PROC_EPS_READ_CYCLES_T1 attribute
    case ATTR_PROC_EPS_READ_CYCLES_T1:
        l_attrSize = sizeof(ATTR_PROC_EPS_READ_CYCLES_T1_type);
        break;

    // Get size for ATTR_PROC_EPS_READ_CYCLES_T2 attribute
    case ATTR_PROC_EPS_READ_CYCLES_T2:
        l_attrSize = sizeof(ATTR_PROC_EPS_READ_CYCLES_T2_type);
        break;

    // Get size for ATTR_PROC_EPS_WRITE_CYCLES_T1 attribute
    case ATTR_PROC_EPS_WRITE_CYCLES_T1:
        l_attrSize = sizeof(ATTR_PROC_EPS_WRITE_CYCLES_T1_type);
        break;

    // Get size for ATTR_PROC_EPS_WRITE_CYCLES_T2 attribute
    case ATTR_PROC_EPS_WRITE_CYCLES_T2:
        l_attrSize = sizeof(ATTR_PROC_EPS_WRITE_CYCLES_T2_type);
        break;

    // Get size for ATTR_DMI_REFCLOCK_SWIZZLE attribute
    case ATTR_DMI_REFCLOCK_SWIZZLE:
        l_attrSize = sizeof(ATTR_DMI_REFCLOCK_SWIZZLE_type);
        break;

    // Get size for ATTR_OMI_REFCLOCK_SWIZZLE attribute
    case ATTR_OMI_REFCLOCK_SWIZZLE:
        l_attrSize = sizeof(ATTR_OMI_REFCLOCK_SWIZZLE_type);
        break;

    // Get size for ATTR_SYSTEM_IPL_PHASE attribute
    case ATTR_SYSTEM_IPL_PHASE:
        l_attrSize = sizeof(ATTR_SYSTEM_IPL_PHASE_type);
        break;

    // Get size for ATTR_PROC_XSCOM_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_XSCOM_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_XSCOM_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_LPC_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_LPC_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_LPC_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_MAX_INTERLEAVE_GROUP_SIZE attribute
    case ATTR_MAX_INTERLEAVE_GROUP_SIZE:
        l_attrSize = sizeof(ATTR_MAX_INTERLEAVE_GROUP_SIZE_type);
        break;

    // Get size for ATTR_MEM_MIRROR_PLACEMENT_POLICY attribute
    case ATTR_MEM_MIRROR_PLACEMENT_POLICY:
        l_attrSize = sizeof(ATTR_MEM_MIRROR_PLACEMENT_POLICY_type);
        break;

    // Get size for ATTR_PROC_MEM_BASES attribute
    case ATTR_PROC_MEM_BASES:
        l_attrSize = sizeof(ATTR_PROC_MEM_BASES_type);
        break;

    // Get size for ATTR_PROC_MEM_SIZES attribute
    case ATTR_PROC_MEM_SIZES:
        l_attrSize = sizeof(ATTR_PROC_MEM_SIZES_type);
        break;

    // Get size for ATTR_PROC_MIRROR_BASES attribute
    case ATTR_PROC_MIRROR_BASES:
        l_attrSize = sizeof(ATTR_PROC_MIRROR_BASES_type);
        break;

    // Get size for ATTR_PROC_MIRROR_SIZES attribute
    case ATTR_PROC_MIRROR_SIZES:
        l_attrSize = sizeof(ATTR_PROC_MIRROR_SIZES_type);
        break;

    // Get size for ATTR_MSS_INTERLEAVE_ENABLE attribute
    case ATTR_MSS_INTERLEAVE_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_INTERLEAVE_ENABLE_type);
        break;

    // Get size for ATTR_MSS_INTERLEAVE_GRANULARITY attribute
    case ATTR_MSS_INTERLEAVE_GRANULARITY:
        l_attrSize = sizeof(ATTR_MSS_INTERLEAVE_GRANULARITY_type);
        break;

    // Get size for ATTR_MSS_MEM_MC_IN_GROUP attribute
    case ATTR_MSS_MEM_MC_IN_GROUP:
        l_attrSize = sizeof(ATTR_MSS_MEM_MC_IN_GROUP_type);
        break;

    // Get size for ATTR_MSS_MCS_GROUP_32 attribute
    case ATTR_MSS_MCS_GROUP_32:
        l_attrSize = sizeof(ATTR_MSS_MCS_GROUP_32_type);
        break;

    // Get size for ATTR_MSS_MEM_IPL_COMPLETE attribute
    case ATTR_MSS_MEM_IPL_COMPLETE:
        l_attrSize = sizeof(ATTR_MSS_MEM_IPL_COMPLETE_type);
        break;

    // Get size for ATTR_MRW_HW_MIRRORING_ENABLE attribute
    case ATTR_MRW_HW_MIRRORING_ENABLE:
        l_attrSize = sizeof(ATTR_MRW_HW_MIRRORING_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NHTM_BAR_BASE_ADDR attribute
    case ATTR_PROC_NHTM_BAR_BASE_ADDR:
        l_attrSize = sizeof(ATTR_PROC_NHTM_BAR_BASE_ADDR_type);
        break;

    // Get size for ATTR_PROC_NHTM_BAR_SIZE attribute
    case ATTR_PROC_NHTM_BAR_SIZE:
        l_attrSize = sizeof(ATTR_PROC_NHTM_BAR_SIZE_type);
        break;

    // Get size for ATTR_PROC_CHTM_BAR_BASE_ADDR attribute
    case ATTR_PROC_CHTM_BAR_BASE_ADDR:
        l_attrSize = sizeof(ATTR_PROC_CHTM_BAR_BASE_ADDR_type);
        break;

    // Get size for ATTR_PROC_CHTM_BAR_SIZES attribute
    case ATTR_PROC_CHTM_BAR_SIZES:
        l_attrSize = sizeof(ATTR_PROC_CHTM_BAR_SIZES_type);
        break;

    // Get size for ATTR_PROC_SMF_BAR_BASE_ADDR attribute
    case ATTR_PROC_SMF_BAR_BASE_ADDR:
        l_attrSize = sizeof(ATTR_PROC_SMF_BAR_BASE_ADDR_type);
        break;

    // Get size for ATTR_PROC_SMF_BAR_SIZE attribute
    case ATTR_PROC_SMF_BAR_SIZE:
        l_attrSize = sizeof(ATTR_PROC_SMF_BAR_SIZE_type);
        break;

    // Get size for ATTR_SMF_CONFIG attribute
    case ATTR_SMF_CONFIG:
        l_attrSize = sizeof(ATTR_SMF_CONFIG_type);
        break;

    // Get size for ATTR_SMF_ENABLED attribute
    case ATTR_SMF_ENABLED:
        l_attrSize = sizeof(ATTR_SMF_ENABLED_type);
        break;

    // Get size for ATTR_PROC_OCC_SANDBOX_BASE_ADDR attribute
    case ATTR_PROC_OCC_SANDBOX_BASE_ADDR:
        l_attrSize = sizeof(ATTR_PROC_OCC_SANDBOX_BASE_ADDR_type);
        break;

    // Get size for ATTR_PROC_OCC_SANDBOX_SIZE attribute
    case ATTR_PROC_OCC_SANDBOX_SIZE:
        l_attrSize = sizeof(ATTR_PROC_OCC_SANDBOX_SIZE_type);
        break;

    // Get size for ATTR_PROC_MEM_BASES_ACK attribute
    case ATTR_PROC_MEM_BASES_ACK:
        l_attrSize = sizeof(ATTR_PROC_MEM_BASES_ACK_type);
        break;

    // Get size for ATTR_PROC_MEM_SIZES_ACK attribute
    case ATTR_PROC_MEM_SIZES_ACK:
        l_attrSize = sizeof(ATTR_PROC_MEM_SIZES_ACK_type);
        break;

    // Get size for ATTR_PROC_MIRROR_BASES_ACK attribute
    case ATTR_PROC_MIRROR_BASES_ACK:
        l_attrSize = sizeof(ATTR_PROC_MIRROR_BASES_ACK_type);
        break;

    // Get size for ATTR_PROC_MIRROR_SIZES_ACK attribute
    case ATTR_PROC_MIRROR_SIZES_ACK:
        l_attrSize = sizeof(ATTR_PROC_MIRROR_SIZES_ACK_type);
        break;

    // Get size for ATTR_HTM_QUEUES attribute
    case ATTR_HTM_QUEUES:
        l_attrSize = sizeof(ATTR_HTM_QUEUES_type);
        break;

    // Get size for ATTR_ENABLE_MEM_EARLY_DATA_SCOM attribute
    case ATTR_ENABLE_MEM_EARLY_DATA_SCOM:
        l_attrSize = sizeof(ATTR_ENABLE_MEM_EARLY_DATA_SCOM_type);
        break;

    // Get size for ATTR_MSS_VPD_DQ_MAP attribute
    case ATTR_MSS_VPD_DQ_MAP:
        l_attrSize = sizeof(ATTR_MSS_VPD_DQ_MAP_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_0_VERSION_LAYOUT attribute
    case ATTR_MSS_VPD_MR_0_VERSION_LAYOUT:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_0_VERSION_LAYOUT_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_1_VERSION_DATA attribute
    case ATTR_MSS_VPD_MR_1_VERSION_DATA:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_1_VERSION_DATA_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_2_SIGNATURE_HASH attribute
    case ATTR_MSS_VPD_MR_2_SIGNATURE_HASH:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_2_SIGNATURE_HASH_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_DPHY_GPO attribute
    case ATTR_MSS_VPD_MR_DPHY_GPO:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_DPHY_GPO_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_DPHY_RLO attribute
    case ATTR_MSS_VPD_MR_DPHY_RLO:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_DPHY_RLO_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_DPHY_WLO attribute
    case ATTR_MSS_VPD_MR_DPHY_WLO:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_DPHY_WLO_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_2N_MODE_AUTOSET attribute
    case ATTR_MSS_VPD_MR_MC_2N_MODE_AUTOSET:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_2N_MODE_AUTOSET_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A00_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A01_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A02_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A03_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A04_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A05_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A06_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A07_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A08_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A09_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A10_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A11_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A12_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A13_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_A17_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BA1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_BG1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_ADDR_C2_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ACTN_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_CASN_A15_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_RASN_A16_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_ADDR_WEN_A14_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CMD_PAR_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CKE1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_CSN1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D0_ODT1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CKE1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_CSN1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT0_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1 attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_CNTL_D1_ODT1_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKN_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_D0_CLKP_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKN_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP attribute
    case ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_MC_PHASE_ROT_D1_CLKP_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_TSYS_ADR attribute
    case ATTR_MSS_VPD_MR_TSYS_ADR:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_TSYS_ADR_type);
        break;

    // Get size for ATTR_MSS_VPD_MR_TSYS_DATA attribute
    case ATTR_MSS_VPD_MR_TSYS_DATA:
        l_attrSize = sizeof(ATTR_MSS_VPD_MR_TSYS_DATA_type);
        break;

    // Get size for ATTR_PROC_FSP_BAR_ENABLE attribute
    case ATTR_PROC_FSP_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_FSP_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_FSP_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_FSP_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_FSP_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_FSP_BAR_SIZE attribute
    case ATTR_PROC_FSP_BAR_SIZE:
        l_attrSize = sizeof(ATTR_PROC_FSP_BAR_SIZE_type);
        break;

    // Get size for ATTR_PROC_FSP_MMIO_MASK_SIZE attribute
    case ATTR_PROC_FSP_MMIO_MASK_SIZE:
        l_attrSize = sizeof(ATTR_PROC_FSP_MMIO_MASK_SIZE_type);
        break;

    // Get size for ATTR_PROC_NPU_PHY0_BAR_ENABLE attribute
    case ATTR_PROC_NPU_PHY0_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NPU_PHY0_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NPU_PHY0_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_NPU_PHY0_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_NPU_PHY0_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NPU_PHY1_BAR_ENABLE attribute
    case ATTR_PROC_NPU_PHY1_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NPU_PHY1_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NPU_PHY1_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_NPU_PHY1_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_NPU_PHY1_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NPU_MMIO_BAR_ENABLE attribute
    case ATTR_PROC_NPU_MMIO_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NPU_MMIO_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NPU_MMIO_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_NPU_MMIO_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_NPU_MMIO_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NPU0_MMIO_BAR_ENABLE attribute
    case ATTR_PROC_NPU0_MMIO_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NPU0_MMIO_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NPU0_MMIO_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_NPU0_MMIO_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_NPU0_MMIO_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NPU1_MMIO_BAR_ENABLE attribute
    case ATTR_PROC_NPU1_MMIO_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NPU1_MMIO_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NPU1_MMIO_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_NPU1_MMIO_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_NPU1_MMIO_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NPU2_MMIO_BAR_ENABLE attribute
    case ATTR_PROC_NPU2_MMIO_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NPU2_MMIO_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NPU2_MMIO_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_NPU2_MMIO_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_NPU2_MMIO_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NPU0_PRI_CONFIG attribute
    case ATTR_PROC_NPU0_PRI_CONFIG:
        l_attrSize = sizeof(ATTR_PROC_NPU0_PRI_CONFIG_type);
        break;

    // Get size for ATTR_PROC_NPU1_PRI_CONFIG attribute
    case ATTR_PROC_NPU1_PRI_CONFIG:
        l_attrSize = sizeof(ATTR_PROC_NPU1_PRI_CONFIG_type);
        break;

    // Get size for ATTR_PROC_NPU2_PRI_CONFIG attribute
    case ATTR_PROC_NPU2_PRI_CONFIG:
        l_attrSize = sizeof(ATTR_PROC_NPU2_PRI_CONFIG_type);
        break;

    // Get size for ATTR_PROC_PSI_BRIDGE_BAR_ENABLE attribute
    case ATTR_PROC_PSI_BRIDGE_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_PSI_BRIDGE_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_PSI_BRIDGE_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_PSI_BRIDGE_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_PSI_BRIDGE_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NX_RNG_BAR_ENABLE attribute
    case ATTR_PROC_NX_RNG_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NX_RNG_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NX_RNG_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_NX_RNG_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_NX_RNG_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_NX_RNG_FAILED_INT_ENABLE attribute
    case ATTR_PROC_NX_RNG_FAILED_INT_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_NX_RNG_FAILED_INT_ENABLE_type);
        break;

    // Get size for ATTR_PROC_NX_RNG_FAILED_INT_ADDR attribute
    case ATTR_PROC_NX_RNG_FAILED_INT_ADDR:
        l_attrSize = sizeof(ATTR_PROC_NX_RNG_FAILED_INT_ADDR_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_PC_BAR_ENABLE attribute
    case ATTR_PROC_INT_CQ_PC_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_PC_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_MASK attribute
    case ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_MASK:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_PC_BAR_BASE_ADDR_OFFSET_MASK_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_VC_BAR_ENABLE attribute
    case ATTR_PROC_INT_CQ_VC_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_VC_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_MASK attribute
    case ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_MASK:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_VC_BAR_BASE_ADDR_OFFSET_MASK_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_TM1_BAR_ENABLE attribute
    case ATTR_PROC_INT_CQ_TM1_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_TM1_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_TM1_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_INT_CQ_TM1_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_TM1_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_TM1_BAR_PAGE_SIZE attribute
    case ATTR_PROC_INT_CQ_TM1_BAR_PAGE_SIZE:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_TM1_BAR_PAGE_SIZE_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_IC_BAR_ENABLE attribute
    case ATTR_PROC_INT_CQ_IC_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_IC_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_IC_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_INT_CQ_IC_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_IC_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_INT_CQ_IC_BAR_PAGE_SIZE attribute
    case ATTR_PROC_INT_CQ_IC_BAR_PAGE_SIZE:
        l_attrSize = sizeof(ATTR_PROC_INT_CQ_IC_BAR_PAGE_SIZE_type);
        break;

    // Get size for ATTR_IO_OMI_DCCAL_FLAGS attribute
    case ATTR_IO_OMI_DCCAL_FLAGS:
        l_attrSize = sizeof(ATTR_IO_OMI_DCCAL_FLAGS_type);
        break;

    // Get size for ATTR_IO_OMI_DEBUG attribute
    case ATTR_IO_OMI_DEBUG:
        l_attrSize = sizeof(ATTR_IO_OMI_DEBUG_type);
        break;

    // Get size for ATTR_IO_OMI_MFG_CHK attribute
    case ATTR_IO_OMI_MFG_CHK:
        l_attrSize = sizeof(ATTR_IO_OMI_MFG_CHK_type);
        break;

    // Get size for ATTR_IO_OMI_MFG_MIN_EYE_WIDTH attribute
    case ATTR_IO_OMI_MFG_MIN_EYE_WIDTH:
        l_attrSize = sizeof(ATTR_IO_OMI_MFG_MIN_EYE_WIDTH_type);
        break;

    // Get size for ATTR_IO_OMI_TX_MARGIN_RATIO attribute
    case ATTR_IO_OMI_TX_MARGIN_RATIO:
        l_attrSize = sizeof(ATTR_IO_OMI_TX_MARGIN_RATIO_type);
        break;

    // Get size for ATTR_IO_OMI_TX_FFE_PRECURSOR attribute
    case ATTR_IO_OMI_TX_FFE_PRECURSOR:
        l_attrSize = sizeof(ATTR_IO_OMI_TX_FFE_PRECURSOR_type);
        break;

    // Get size for ATTR_IO_OMI_TX_FFE_POSTCURSOR attribute
    case ATTR_IO_OMI_TX_FFE_POSTCURSOR:
        l_attrSize = sizeof(ATTR_IO_OMI_TX_FFE_POSTCURSOR_type);
        break;

    // Get size for ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_EVEN attribute
    case ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_EVEN:
        l_attrSize = sizeof(ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_EVEN_type);
        break;

    // Get size for ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_ODD attribute
    case ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_ODD:
        l_attrSize = sizeof(ATTR_IO_OMI_MFG_STRESS_PR_OFFSET_ODD_type);
        break;

    // Get size for ATTR_PROC_PCIE_IOP_SWAP attribute
    case ATTR_PROC_PCIE_IOP_SWAP:
        l_attrSize = sizeof(ATTR_PROC_PCIE_IOP_SWAP_type);
        break;

    // Get size for ATTR_PROC_PCIE_BAR_ENABLE attribute
    case ATTR_PROC_PCIE_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_BAR_ENABLE_type);
        break;

    // Get size for ATTR_PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_PCIE_MMIO_BAR0_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_PCIE_MMIO_BAR1_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_PCIE_REGISTER_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_PROC_PCIE_BAR_SIZE attribute
    case ATTR_PROC_PCIE_BAR_SIZE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_BAR_SIZE_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_CDR_GAIN attribute
    case ATTR_PROC_PCIE_PCS_RX_CDR_GAIN:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_CDR_GAIN_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_PK_INIT attribute
    case ATTR_PROC_PCIE_PCS_RX_PK_INIT:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_PK_INIT_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_INIT_GAIN attribute
    case ATTR_PROC_PCIE_PCS_RX_INIT_GAIN:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_INIT_GAIN_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_SIGDET_LVL attribute
    case ATTR_PROC_PCIE_PCS_RX_SIGDET_LVL:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_SIGDET_LVL_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_ROT_CDR_LOOKAHEAD attribute
    case ATTR_PROC_PCIE_PCS_RX_ROT_CDR_LOOKAHEAD:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_ROT_CDR_LOOKAHEAD_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_ROT_CDR_SSC attribute
    case ATTR_PROC_PCIE_PCS_RX_ROT_CDR_SSC:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_ROT_CDR_SSC_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_ROT_EXTEL attribute
    case ATTR_PROC_PCIE_PCS_RX_ROT_EXTEL:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_ROT_EXTEL_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_ROT_RST_FW attribute
    case ATTR_PROC_PCIE_PCS_RX_ROT_RST_FW:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_ROT_RST_FW_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLA attribute
    case ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLA:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLA_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLB attribute
    case ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLB:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_PCLCK_CNTL_PLLB_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_TX_DCLCK_ROT attribute
    case ATTR_PROC_PCIE_PCS_TX_DCLCK_ROT:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_TX_DCLCK_ROT_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG1 attribute
    case ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG1:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG1_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG2 attribute
    case ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG2:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_TX_PCIE_RECV_DETECT_CNTL_REG2_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_TX_POWER_SEQ_ENABLE attribute
    case ATTR_PROC_PCIE_PCS_TX_POWER_SEQ_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_TX_POWER_SEQ_ENABLE_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG1 attribute
    case ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG1:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG1_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG2 attribute
    case ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG2:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_VGA_CNTL_REG2_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_DFE_FDDC attribute
    case ATTR_PROC_PCIE_PCS_RX_DFE_FDDC:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_DFE_FDDC_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_M_CNTL attribute
    case ATTR_PROC_PCIE_PCS_M_CNTL:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_M_CNTL_type);
        break;

    // Get size for ATTR_PROC_PCIE_CACHE_INJ_MODE attribute
    case ATTR_PROC_PCIE_CACHE_INJ_MODE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_CACHE_INJ_MODE_type);
        break;

    // Get size for ATTR_PROC_PCIE_CACHE_INJ_THROTTLE attribute
    case ATTR_PROC_PCIE_CACHE_INJ_THROTTLE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_CACHE_INJ_THROTTLE_type);
        break;

    // Get size for ATTR_OMI_DL_LN_REV_ENABLE attribute
    case ATTR_OMI_DL_LN_REV_ENABLE:
        l_attrSize = sizeof(ATTR_OMI_DL_LN_REV_ENABLE_type);
        break;

    // Get size for ATTR_PROC_DPLL_DIVIDER attribute
    case ATTR_PROC_DPLL_DIVIDER:
        l_attrSize = sizeof(ATTR_PROC_DPLL_DIVIDER_type);
        break;

    // Get size for ATTR_STOPGPE_BOOT_COPIER_IVPR_OFFSET attribute
    case ATTR_STOPGPE_BOOT_COPIER_IVPR_OFFSET:
        l_attrSize = sizeof(ATTR_STOPGPE_BOOT_COPIER_IVPR_OFFSET_type);
        break;

    // Get size for ATTR_PSTATEGPE_BOOT_COPIER_IVPR_OFFSET attribute
    case ATTR_PSTATEGPE_BOOT_COPIER_IVPR_OFFSET:
        l_attrSize = sizeof(ATTR_PSTATEGPE_BOOT_COPIER_IVPR_OFFSET_type);
        break;

    // Get size for ATTR_PM_SPWUP_IGNORE_XSTOP_FLAG attribute
    case ATTR_PM_SPWUP_IGNORE_XSTOP_FLAG:
        l_attrSize = sizeof(ATTR_PM_SPWUP_IGNORE_XSTOP_FLAG_type);
        break;

    // Get size for ATTR_OCC_LFIR attribute
    case ATTR_OCC_LFIR:
        l_attrSize = sizeof(ATTR_OCC_LFIR_type);
        break;

    // Get size for ATTR_PBA_LFIR attribute
    case ATTR_PBA_LFIR:
        l_attrSize = sizeof(ATTR_PBA_LFIR_type);
        break;

    // Get size for ATTR_PM_FIRINIT_DONE_ONCE_FLAG attribute
    case ATTR_PM_FIRINIT_DONE_ONCE_FLAG:
        l_attrSize = sizeof(ATTR_PM_FIRINIT_DONE_ONCE_FLAG_type);
        break;

    // Get size for ATTR_QUAD_PPM_ERRMASK attribute
    case ATTR_QUAD_PPM_ERRMASK:
        l_attrSize = sizeof(ATTR_QUAD_PPM_ERRMASK_type);
        break;

    // Get size for ATTR_CORE_PPM_ERRMASK attribute
    case ATTR_CORE_PPM_ERRMASK:
        l_attrSize = sizeof(ATTR_CORE_PPM_ERRMASK_type);
        break;

    // Get size for ATTR_CME_LOCAL_FIRMASK attribute
    case ATTR_CME_LOCAL_FIRMASK:
        l_attrSize = sizeof(ATTR_CME_LOCAL_FIRMASK_type);
        break;

    // Get size for ATTR_L2_HASCLOCKS attribute
    case ATTR_L2_HASCLOCKS:
        l_attrSize = sizeof(ATTR_L2_HASCLOCKS_type);
        break;

    // Get size for ATTR_L3_HASCLOCKS attribute
    case ATTR_L3_HASCLOCKS:
        l_attrSize = sizeof(ATTR_L3_HASCLOCKS_type);
        break;

    // Get size for ATTR_C0_EXEC_HASCLOCKS attribute
    case ATTR_C0_EXEC_HASCLOCKS:
        l_attrSize = sizeof(ATTR_C0_EXEC_HASCLOCKS_type);
        break;

    // Get size for ATTR_C1_EXEC_HASCLOCKS attribute
    case ATTR_C1_EXEC_HASCLOCKS:
        l_attrSize = sizeof(ATTR_C1_EXEC_HASCLOCKS_type);
        break;

    // Get size for ATTR_C0_PC_HASCLOCKS attribute
    case ATTR_C0_PC_HASCLOCKS:
        l_attrSize = sizeof(ATTR_C0_PC_HASCLOCKS_type);
        break;

    // Get size for ATTR_C1_PC_HASCLOCKS attribute
    case ATTR_C1_PC_HASCLOCKS:
        l_attrSize = sizeof(ATTR_C1_PC_HASCLOCKS_type);
        break;

    // Get size for ATTR_L2_HASPOWER attribute
    case ATTR_L2_HASPOWER:
        l_attrSize = sizeof(ATTR_L2_HASPOWER_type);
        break;

    // Get size for ATTR_L3_HASPOWER attribute
    case ATTR_L3_HASPOWER:
        l_attrSize = sizeof(ATTR_L3_HASPOWER_type);
        break;

    // Get size for ATTR_C0_HASPOWER attribute
    case ATTR_C0_HASPOWER:
        l_attrSize = sizeof(ATTR_C0_HASPOWER_type);
        break;

    // Get size for ATTR_C1_HASPOWER attribute
    case ATTR_C1_HASPOWER:
        l_attrSize = sizeof(ATTR_C1_HASPOWER_type);
        break;

    // Get size for ATTR_PM_SPIPSS_FRAME_SIZE attribute
    case ATTR_PM_SPIPSS_FRAME_SIZE:
        l_attrSize = sizeof(ATTR_PM_SPIPSS_FRAME_SIZE_type);
        break;

    // Get size for ATTR_PM_SPIPSS_IN_DELAY attribute
    case ATTR_PM_SPIPSS_IN_DELAY:
        l_attrSize = sizeof(ATTR_PM_SPIPSS_IN_DELAY_type);
        break;

    // Get size for ATTR_PM_SPIPSS_CLOCK_POLARITY attribute
    case ATTR_PM_SPIPSS_CLOCK_POLARITY:
        l_attrSize = sizeof(ATTR_PM_SPIPSS_CLOCK_POLARITY_type);
        break;

    // Get size for ATTR_PM_SPIPSS_CLOCK_PHASE attribute
    case ATTR_PM_SPIPSS_CLOCK_PHASE:
        l_attrSize = sizeof(ATTR_PM_SPIPSS_CLOCK_PHASE_type);
        break;

    // Get size for ATTR_PM_SPIPSS_CLOCK_DIVIDER attribute
    case ATTR_PM_SPIPSS_CLOCK_DIVIDER:
        l_attrSize = sizeof(ATTR_PM_SPIPSS_CLOCK_DIVIDER_type);
        break;

    // Get size for ATTR_PM_SPIPSS_INTER_FRAME_DELAY_SETTING attribute
    case ATTR_PM_SPIPSS_INTER_FRAME_DELAY_SETTING:
        l_attrSize = sizeof(ATTR_PM_SPIPSS_INTER_FRAME_DELAY_SETTING_type);
        break;

    // Get size for ATTR_PSTATES_ENABLED attribute
    case ATTR_PSTATES_ENABLED:
        l_attrSize = sizeof(ATTR_PSTATES_ENABLED_type);
        break;

    // Get size for ATTR_RESCLK_ENABLED attribute
    case ATTR_RESCLK_ENABLED:
        l_attrSize = sizeof(ATTR_RESCLK_ENABLED_type);
        break;

    // Get size for ATTR_VDM_ENABLED attribute
    case ATTR_VDM_ENABLED:
        l_attrSize = sizeof(ATTR_VDM_ENABLED_type);
        break;

    // Get size for ATTR_IVRM_ENABLED attribute
    case ATTR_IVRM_ENABLED:
        l_attrSize = sizeof(ATTR_IVRM_ENABLED_type);
        break;

    // Get size for ATTR_WOF_ENABLED attribute
    case ATTR_WOF_ENABLED:
        l_attrSize = sizeof(ATTR_WOF_ENABLED_type);
        break;

    // Get size for ATTR_WOV_OVERV_ENABLED attribute
    case ATTR_WOV_OVERV_ENABLED:
        l_attrSize = sizeof(ATTR_WOV_OVERV_ENABLED_type);
        break;

    // Get size for ATTR_WOV_UNDERV_ENABLED attribute
    case ATTR_WOV_UNDERV_ENABLED:
        l_attrSize = sizeof(ATTR_WOV_UNDERV_ENABLED_type);
        break;

    // Get size for ATTR_SGPE_PHANTOM_HALT_ENABLE attribute
    case ATTR_SGPE_PHANTOM_HALT_ENABLE:
        l_attrSize = sizeof(ATTR_SGPE_PHANTOM_HALT_ENABLE_type);
        break;

    // Get size for ATTR_PGPE_PHANTOM_HALT_ENABLE attribute
    case ATTR_PGPE_PHANTOM_HALT_ENABLE:
        l_attrSize = sizeof(ATTR_PGPE_PHANTOM_HALT_ENABLE_type);
        break;

    // Get size for ATTR_CME_STOP_PHANTOM_HALT_ENABLE attribute
    case ATTR_CME_STOP_PHANTOM_HALT_ENABLE:
        l_attrSize = sizeof(ATTR_CME_STOP_PHANTOM_HALT_ENABLE_type);
        break;

    // Get size for ATTR_CME_PSTATE_PHANTOM_HALT_ENABLE attribute
    case ATTR_CME_PSTATE_PHANTOM_HALT_ENABLE:
        l_attrSize = sizeof(ATTR_CME_PSTATE_PHANTOM_HALT_ENABLE_type);
        break;

    // Get size for ATTR_CORE_INSIDE_SPECIAL_WAKEUP attribute
    case ATTR_CORE_INSIDE_SPECIAL_WAKEUP:
        l_attrSize = sizeof(ATTR_CORE_INSIDE_SPECIAL_WAKEUP_type);
        break;

    // Get size for ATTR_EX_INSIDE_SPECIAL_WAKEUP attribute
    case ATTR_EX_INSIDE_SPECIAL_WAKEUP:
        l_attrSize = sizeof(ATTR_EX_INSIDE_SPECIAL_WAKEUP_type);
        break;

    // Get size for ATTR_EQ_INSIDE_SPECIAL_WAKEUP attribute
    case ATTR_EQ_INSIDE_SPECIAL_WAKEUP:
        l_attrSize = sizeof(ATTR_EQ_INSIDE_SPECIAL_WAKEUP_type);
        break;

    // Get size for ATTR_INITIATED_PM_RESET attribute
    case ATTR_INITIATED_PM_RESET:
        l_attrSize = sizeof(ATTR_INITIATED_PM_RESET_type);
        break;

    // Get size for ATTR_PM_RESET_PHASE attribute
    case ATTR_PM_RESET_PHASE:
        l_attrSize = sizeof(ATTR_PM_RESET_PHASE_type);
        break;

    // Get size for ATTR_PM_MALF_CYCLE attribute
    case ATTR_PM_MALF_CYCLE:
        l_attrSize = sizeof(ATTR_PM_MALF_CYCLE_type);
        break;

    // Get size for ATTR_OMI_DL_GROUP_POS attribute
    case ATTR_OMI_DL_GROUP_POS:
        l_attrSize = sizeof(ATTR_OMI_DL_GROUP_POS_type);
        break;

    // Get size for ATTR_OMI_DL_NUM attribute
    case ATTR_OMI_DL_NUM:
        l_attrSize = sizeof(ATTR_OMI_DL_NUM_type);
        break;

    // Get size for ATTR_IO_DMI_PROC_DCCAL_FLAGS attribute
    case ATTR_IO_DMI_PROC_DCCAL_FLAGS:
        l_attrSize = sizeof(ATTR_IO_DMI_PROC_DCCAL_FLAGS_type);
        break;

    // Get size for ATTR_IO_DMI_PROC_DEBUG attribute
    case ATTR_IO_DMI_PROC_DEBUG:
        l_attrSize = sizeof(ATTR_IO_DMI_PROC_DEBUG_type);
        break;

    // Get size for ATTR_IO_DMI_PROC_TX_MARGIN_RATIO attribute
    case ATTR_IO_DMI_PROC_TX_MARGIN_RATIO:
        l_attrSize = sizeof(ATTR_IO_DMI_PROC_TX_MARGIN_RATIO_type);
        break;

    // Get size for ATTR_IO_DMI_PROC_TX_FFE_PRECURSOR attribute
    case ATTR_IO_DMI_PROC_TX_FFE_PRECURSOR:
        l_attrSize = sizeof(ATTR_IO_DMI_PROC_TX_FFE_PRECURSOR_type);
        break;

    // Get size for ATTR_DMI_INBAND_BAR_ENABLE attribute
    case ATTR_DMI_INBAND_BAR_ENABLE:
        l_attrSize = sizeof(ATTR_DMI_INBAND_BAR_ENABLE_type);
        break;

    // Get size for ATTR_DMI_INBAND_BAR_BASE_ADDR_OFFSET attribute
    case ATTR_DMI_INBAND_BAR_BASE_ADDR_OFFSET:
        l_attrSize = sizeof(ATTR_DMI_INBAND_BAR_BASE_ADDR_OFFSET_type);
        break;

    // Get size for ATTR_MSS_FREQ_OVERRIDE attribute
    case ATTR_MSS_FREQ_OVERRIDE:
        l_attrSize = sizeof(ATTR_MSS_FREQ_OVERRIDE_type);
        break;

    // Get size for ATTR_MSS_FREQ attribute
    case ATTR_MSS_FREQ:
        l_attrSize = sizeof(ATTR_MSS_FREQ_type);
        break;

    // Get size for ATTR_MSS_FREQ_BIAS_PERCENTAGE attribute
    case ATTR_MSS_FREQ_BIAS_PERCENTAGE:
        l_attrSize = sizeof(ATTR_MSS_FREQ_BIAS_PERCENTAGE_type);
        break;

    // Get size for ATTR_EFF_DIMM_SPARE attribute
    case ATTR_EFF_DIMM_SPARE:
        l_attrSize = sizeof(ATTR_EFF_DIMM_SPARE_type);
        break;

    // Get size for ATTR_EFF_DRAM_WR_VREF attribute
    case ATTR_EFF_DRAM_WR_VREF:
        l_attrSize = sizeof(ATTR_EFF_DRAM_WR_VREF_type);
        break;

    // Get size for ATTR_EFF_DRAM_WR_VREF_SCHMOO attribute
    case ATTR_EFF_DRAM_WR_VREF_SCHMOO:
        l_attrSize = sizeof(ATTR_EFF_DRAM_WR_VREF_SCHMOO_type);
        break;

    // Get size for ATTR_EFF_DRAM_WRDDR4_VREF_SCHMOO attribute
    case ATTR_EFF_DRAM_WRDDR4_VREF_SCHMOO:
        l_attrSize = sizeof(ATTR_EFF_DRAM_WRDDR4_VREF_SCHMOO_type);
        break;

    // Get size for ATTR_EFF_DIMM_SIZE attribute
    case ATTR_EFF_DIMM_SIZE:
        l_attrSize = sizeof(ATTR_EFF_DIMM_SIZE_type);
        break;

    // Get size for ATTR_EFF_DRAM_CL attribute
    case ATTR_EFF_DRAM_CL:
        l_attrSize = sizeof(ATTR_EFF_DRAM_CL_type);
        break;

    // Get size for ATTR_EFF_DRAM_AL attribute
    case ATTR_EFF_DRAM_AL:
        l_attrSize = sizeof(ATTR_EFF_DRAM_AL_type);
        break;

    // Get size for ATTR_EFF_DRAM_CWL attribute
    case ATTR_EFF_DRAM_CWL:
        l_attrSize = sizeof(ATTR_EFF_DRAM_CWL_type);
        break;

    // Get size for ATTR_EFF_DRAM_RBT attribute
    case ATTR_EFF_DRAM_RBT:
        l_attrSize = sizeof(ATTR_EFF_DRAM_RBT_type);
        break;

    // Get size for ATTR_EFF_DRAM_TM attribute
    case ATTR_EFF_DRAM_TM:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TM_type);
        break;

    // Get size for ATTR_EFF_DRAM_DLL_RESET attribute
    case ATTR_EFF_DRAM_DLL_RESET:
        l_attrSize = sizeof(ATTR_EFF_DRAM_DLL_RESET_type);
        break;

    // Get size for ATTR_EFF_DRAM_DLL_PPD attribute
    case ATTR_EFF_DRAM_DLL_PPD:
        l_attrSize = sizeof(ATTR_EFF_DRAM_DLL_PPD_type);
        break;

    // Get size for ATTR_EFF_DRAM_DLL_ENABLE attribute
    case ATTR_EFF_DRAM_DLL_ENABLE:
        l_attrSize = sizeof(ATTR_EFF_DRAM_DLL_ENABLE_type);
        break;

    // Get size for ATTR_EFF_DRAM_WR_LVL_ENABLE attribute
    case ATTR_EFF_DRAM_WR_LVL_ENABLE:
        l_attrSize = sizeof(ATTR_EFF_DRAM_WR_LVL_ENABLE_type);
        break;

    // Get size for ATTR_EFF_DRAM_OUTPUT_BUFFER attribute
    case ATTR_EFF_DRAM_OUTPUT_BUFFER:
        l_attrSize = sizeof(ATTR_EFF_DRAM_OUTPUT_BUFFER_type);
        break;

    // Get size for ATTR_EFF_DRAM_PASR attribute
    case ATTR_EFF_DRAM_PASR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_PASR_type);
        break;

    // Get size for ATTR_EFF_DRAM_ASR attribute
    case ATTR_EFF_DRAM_ASR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_ASR_type);
        break;

    // Get size for ATTR_EFF_DRAM_SRT attribute
    case ATTR_EFF_DRAM_SRT:
        l_attrSize = sizeof(ATTR_EFF_DRAM_SRT_type);
        break;

    // Get size for ATTR_EFF_MPR_LOC attribute
    case ATTR_EFF_MPR_LOC:
        l_attrSize = sizeof(ATTR_EFF_MPR_LOC_type);
        break;

    // Get size for ATTR_EFF_MPR_MODE attribute
    case ATTR_EFF_MPR_MODE:
        l_attrSize = sizeof(ATTR_EFF_MPR_MODE_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC00 attribute
    case ATTR_EFF_DIMM_DDR4_RC00:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC00_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC01 attribute
    case ATTR_EFF_DIMM_DDR4_RC01:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC01_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC02 attribute
    case ATTR_EFF_DIMM_DDR4_RC02:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC02_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC03 attribute
    case ATTR_EFF_DIMM_DDR4_RC03:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC03_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC04 attribute
    case ATTR_EFF_DIMM_DDR4_RC04:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC04_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC05 attribute
    case ATTR_EFF_DIMM_DDR4_RC05:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC05_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC06_07 attribute
    case ATTR_EFF_DIMM_DDR4_RC06_07:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC06_07_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC08 attribute
    case ATTR_EFF_DIMM_DDR4_RC08:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC08_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC09 attribute
    case ATTR_EFF_DIMM_DDR4_RC09:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC09_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC0A attribute
    case ATTR_EFF_DIMM_DDR4_RC0A:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC0A_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC0B attribute
    case ATTR_EFF_DIMM_DDR4_RC0B:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC0B_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC0C attribute
    case ATTR_EFF_DIMM_DDR4_RC0C:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC0C_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC0D attribute
    case ATTR_EFF_DIMM_DDR4_RC0D:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC0D_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC0E attribute
    case ATTR_EFF_DIMM_DDR4_RC0E:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC0E_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC0F attribute
    case ATTR_EFF_DIMM_DDR4_RC0F:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC0F_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_1x attribute
    case ATTR_EFF_DIMM_DDR4_RC_1x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_1x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_2x attribute
    case ATTR_EFF_DIMM_DDR4_RC_2x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_2x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_3x attribute
    case ATTR_EFF_DIMM_DDR4_RC_3x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_3x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_4x attribute
    case ATTR_EFF_DIMM_DDR4_RC_4x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_4x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_5x attribute
    case ATTR_EFF_DIMM_DDR4_RC_5x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_5x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_6x attribute
    case ATTR_EFF_DIMM_DDR4_RC_6x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_6x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_7x attribute
    case ATTR_EFF_DIMM_DDR4_RC_7x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_7x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_8x attribute
    case ATTR_EFF_DIMM_DDR4_RC_8x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_8x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_9x attribute
    case ATTR_EFF_DIMM_DDR4_RC_9x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_9x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_Ax attribute
    case ATTR_EFF_DIMM_DDR4_RC_Ax:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_Ax_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_RC_Bx attribute
    case ATTR_EFF_DIMM_DDR4_RC_Bx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_RC_Bx_type);
        break;

    // Get size for ATTR_EFF_DIMM_RCD_MIRROR_MODE attribute
    case ATTR_EFF_DIMM_RCD_MIRROR_MODE:
        l_attrSize = sizeof(ATTR_EFF_DIMM_RCD_MIRROR_MODE_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_MODE attribute
    case ATTR_EFF_SCHMOO_MODE:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_MODE_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_ADDR_MODE attribute
    case ATTR_EFF_SCHMOO_ADDR_MODE:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_ADDR_MODE_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_TEST_VALID attribute
    case ATTR_EFF_SCHMOO_TEST_VALID:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_TEST_VALID_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_PARAM_VALID attribute
    case ATTR_EFF_SCHMOO_PARAM_VALID:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_PARAM_VALID_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_WR_EYE_MIN_MARGIN attribute
    case ATTR_EFF_SCHMOO_WR_EYE_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_WR_EYE_MIN_MARGIN_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_RD_EYE_MIN_MARGIN attribute
    case ATTR_EFF_SCHMOO_RD_EYE_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_RD_EYE_MIN_MARGIN_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_DQS_CLK_MIN_MARGIN attribute
    case ATTR_EFF_SCHMOO_DQS_CLK_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_DQS_CLK_MIN_MARGIN_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_RD_GATE_MIN_MARGIN attribute
    case ATTR_EFF_SCHMOO_RD_GATE_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_RD_GATE_MIN_MARGIN_type);
        break;

    // Get size for ATTR_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN attribute
    case ATTR_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN_type);
        break;

    // Get size for ATTR_EFF_MEMCAL_INTERVAL attribute
    case ATTR_EFF_MEMCAL_INTERVAL:
        l_attrSize = sizeof(ATTR_EFF_MEMCAL_INTERVAL_type);
        break;

    // Get size for ATTR_EFF_ZQCAL_INTERVAL attribute
    case ATTR_EFF_ZQCAL_INTERVAL:
        l_attrSize = sizeof(ATTR_EFF_ZQCAL_INTERVAL_type);
        break;

    // Get size for ATTR_EFF_IBM_TYPE attribute
    case ATTR_EFF_IBM_TYPE:
        l_attrSize = sizeof(ATTR_EFF_IBM_TYPE_type);
        break;

    // Get size for ATTR_EFF_NUM_MASTER_RANKS_PER_DIMM attribute
    case ATTR_EFF_NUM_MASTER_RANKS_PER_DIMM:
        l_attrSize = sizeof(ATTR_EFF_NUM_MASTER_RANKS_PER_DIMM_type);
        break;

    // Get size for ATTR_EFF_DIMM_RANKS_CONFIGED attribute
    case ATTR_EFF_DIMM_RANKS_CONFIGED:
        l_attrSize = sizeof(ATTR_EFF_DIMM_RANKS_CONFIGED_type);
        break;

    // Get size for ATTR_EFF_PRIM_DIE_COUNT attribute
    case ATTR_EFF_PRIM_DIE_COUNT:
        l_attrSize = sizeof(ATTR_EFF_PRIM_DIE_COUNT_type);
        break;

    // Get size for ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
    case ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT:
        l_attrSize = sizeof(ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_PORT_type);
        break;

    // Get size for ATTR_MSS_MEM_M_DRAM_CLOCKS attribute
    case ATTR_MSS_MEM_M_DRAM_CLOCKS:
        l_attrSize = sizeof(ATTR_MSS_MEM_M_DRAM_CLOCKS_type);
        break;

    // Get size for ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
    case ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT:
        l_attrSize = sizeof(ATTR_MSS_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type);
        break;

    // Get size for ATTR_MSS_MEM_WATT_TARGET attribute
    case ATTR_MSS_MEM_WATT_TARGET:
        l_attrSize = sizeof(ATTR_MSS_MEM_WATT_TARGET_type);
        break;

    // Get size for ATTR_MSS_TOTAL_PWR_SLOPE attribute
    case ATTR_MSS_TOTAL_PWR_SLOPE:
        l_attrSize = sizeof(ATTR_MSS_TOTAL_PWR_SLOPE_type);
        break;

    // Get size for ATTR_MSS_TOTAL_PWR_INTERCEPT attribute
    case ATTR_MSS_TOTAL_PWR_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_TOTAL_PWR_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_DIMM_MAXBANDWIDTH_GBS attribute
    case ATTR_MSS_DIMM_MAXBANDWIDTH_GBS:
        l_attrSize = sizeof(ATTR_MSS_DIMM_MAXBANDWIDTH_GBS_type);
        break;

    // Get size for ATTR_MSS_DIMM_MAXBANDWIDTH_MRS attribute
    case ATTR_MSS_DIMM_MAXBANDWIDTH_MRS:
        l_attrSize = sizeof(ATTR_MSS_DIMM_MAXBANDWIDTH_MRS_type);
        break;

    // Get size for ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS attribute
    case ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS:
        l_attrSize = sizeof(ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_GBS_type);
        break;

    // Get size for ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS attribute
    case ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS:
        l_attrSize = sizeof(ATTR_MSS_CHANNEL_PAIR_MAXBANDWIDTH_MRS_type);
        break;

    // Get size for ATTR_MSS_DIMM_MAXPOWER attribute
    case ATTR_MSS_DIMM_MAXPOWER:
        l_attrSize = sizeof(ATTR_MSS_DIMM_MAXPOWER_type);
        break;

    // Get size for ATTR_MSS_DATABUS_UTIL attribute
    case ATTR_MSS_DATABUS_UTIL:
        l_attrSize = sizeof(ATTR_MSS_DATABUS_UTIL_type);
        break;

    // Get size for ATTR_MSS_PORT_MAXPOWER attribute
    case ATTR_MSS_PORT_MAXPOWER:
        l_attrSize = sizeof(ATTR_MSS_PORT_MAXPOWER_type);
        break;

    // Get size for ATTR_MSS_DIMM_THERMAL_LIMIT attribute
    case ATTR_MSS_DIMM_THERMAL_LIMIT:
        l_attrSize = sizeof(ATTR_MSS_DIMM_THERMAL_LIMIT_type);
        break;

    // Get size for ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
    case ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT:
        l_attrSize = sizeof(ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT_type);
        break;

    // Get size for ATTR_MSS_RUNTIME_MEM_M_DRAM_CLOCKS attribute
    case ATTR_MSS_RUNTIME_MEM_M_DRAM_CLOCKS:
        l_attrSize = sizeof(ATTR_MSS_RUNTIME_MEM_M_DRAM_CLOCKS_type);
        break;

    // Get size for ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
    case ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT:
        l_attrSize = sizeof(ATTR_MSS_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type);
        break;

    // Get size for ATTR_MSS_EFF_DIMM_FUNCTIONAL_VECTOR attribute
    case ATTR_MSS_EFF_DIMM_FUNCTIONAL_VECTOR:
        l_attrSize = sizeof(ATTR_MSS_EFF_DIMM_FUNCTIONAL_VECTOR_type);
        break;

    // Get size for ATTR_EFF_DRAM_LPASR attribute
    case ATTR_EFF_DRAM_LPASR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_LPASR_type);
        break;

    // Get size for ATTR_EFF_MPR_PAGE attribute
    case ATTR_EFF_MPR_PAGE:
        l_attrSize = sizeof(ATTR_EFF_MPR_PAGE_type);
        break;

    // Get size for ATTR_EFF_GEARDOWN_MODE attribute
    case ATTR_EFF_GEARDOWN_MODE:
        l_attrSize = sizeof(ATTR_EFF_GEARDOWN_MODE_type);
        break;

    // Get size for ATTR_EFF_PER_DRAM_ACCESS attribute
    case ATTR_EFF_PER_DRAM_ACCESS:
        l_attrSize = sizeof(ATTR_EFF_PER_DRAM_ACCESS_type);
        break;

    // Get size for ATTR_EFF_TEMP_READOUT attribute
    case ATTR_EFF_TEMP_READOUT:
        l_attrSize = sizeof(ATTR_EFF_TEMP_READOUT_type);
        break;

    // Get size for ATTR_EFF_CRC_WR_LATENCY attribute
    case ATTR_EFF_CRC_WR_LATENCY:
        l_attrSize = sizeof(ATTR_EFF_CRC_WR_LATENCY_type);
        break;

    // Get size for ATTR_EFF_MPR_RD_FORMAT attribute
    case ATTR_EFF_MPR_RD_FORMAT:
        l_attrSize = sizeof(ATTR_EFF_MPR_RD_FORMAT_type);
        break;

    // Get size for ATTR_EFF_MAX_POWERDOWN_MODE attribute
    case ATTR_EFF_MAX_POWERDOWN_MODE:
        l_attrSize = sizeof(ATTR_EFF_MAX_POWERDOWN_MODE_type);
        break;

    // Get size for ATTR_EFF_INTERNAL_VREF_MONITOR attribute
    case ATTR_EFF_INTERNAL_VREF_MONITOR:
        l_attrSize = sizeof(ATTR_EFF_INTERNAL_VREF_MONITOR_type);
        break;

    // Get size for ATTR_EFF_CS_CMD_LATENCY attribute
    case ATTR_EFF_CS_CMD_LATENCY:
        l_attrSize = sizeof(ATTR_EFF_CS_CMD_LATENCY_type);
        break;

    // Get size for ATTR_EFF_SELF_REF_ABORT attribute
    case ATTR_EFF_SELF_REF_ABORT:
        l_attrSize = sizeof(ATTR_EFF_SELF_REF_ABORT_type);
        break;

    // Get size for ATTR_EFF_RD_PREAMBLE_TRAIN attribute
    case ATTR_EFF_RD_PREAMBLE_TRAIN:
        l_attrSize = sizeof(ATTR_EFF_RD_PREAMBLE_TRAIN_type);
        break;

    // Get size for ATTR_EFF_RD_PREAMBLE attribute
    case ATTR_EFF_RD_PREAMBLE:
        l_attrSize = sizeof(ATTR_EFF_RD_PREAMBLE_type);
        break;

    // Get size for ATTR_EFF_WR_PREAMBLE attribute
    case ATTR_EFF_WR_PREAMBLE:
        l_attrSize = sizeof(ATTR_EFF_WR_PREAMBLE_type);
        break;

    // Get size for ATTR_EFF_CA_PARITY_LATENCY attribute
    case ATTR_EFF_CA_PARITY_LATENCY:
        l_attrSize = sizeof(ATTR_EFF_CA_PARITY_LATENCY_type);
        break;

    // Get size for ATTR_EFF_CRC_ERROR_CLEAR attribute
    case ATTR_EFF_CRC_ERROR_CLEAR:
        l_attrSize = sizeof(ATTR_EFF_CRC_ERROR_CLEAR_type);
        break;

    // Get size for ATTR_EFF_CA_PARITY_ERROR_STATUS attribute
    case ATTR_EFF_CA_PARITY_ERROR_STATUS:
        l_attrSize = sizeof(ATTR_EFF_CA_PARITY_ERROR_STATUS_type);
        break;

    // Get size for ATTR_EFF_ODT_INPUT_BUFF attribute
    case ATTR_EFF_ODT_INPUT_BUFF:
        l_attrSize = sizeof(ATTR_EFF_ODT_INPUT_BUFF_type);
        break;

    // Get size for ATTR_EFF_CA_PARITY attribute
    case ATTR_EFF_CA_PARITY:
        l_attrSize = sizeof(ATTR_EFF_CA_PARITY_type);
        break;

    // Get size for ATTR_EFF_DATA_MASK attribute
    case ATTR_EFF_DATA_MASK:
        l_attrSize = sizeof(ATTR_EFF_DATA_MASK_type);
        break;

    // Get size for ATTR_EFF_WRITE_DBI attribute
    case ATTR_EFF_WRITE_DBI:
        l_attrSize = sizeof(ATTR_EFF_WRITE_DBI_type);
        break;

    // Get size for ATTR_EFF_READ_DBI attribute
    case ATTR_EFF_READ_DBI:
        l_attrSize = sizeof(ATTR_EFF_READ_DBI_type);
        break;

    // Get size for ATTR_EFF_VREF_DQ_TRAIN_VALUE attribute
    case ATTR_EFF_VREF_DQ_TRAIN_VALUE:
        l_attrSize = sizeof(ATTR_EFF_VREF_DQ_TRAIN_VALUE_type);
        break;

    // Get size for ATTR_EFF_VREF_DQ_TRAIN_RANGE attribute
    case ATTR_EFF_VREF_DQ_TRAIN_RANGE:
        l_attrSize = sizeof(ATTR_EFF_VREF_DQ_TRAIN_RANGE_type);
        break;

    // Get size for ATTR_EFF_VREF_DQ_TRAIN_ENABLE attribute
    case ATTR_EFF_VREF_DQ_TRAIN_ENABLE:
        l_attrSize = sizeof(ATTR_EFF_VREF_DQ_TRAIN_ENABLE_type);
        break;

    // Get size for ATTR_EFF_WRITE_CRC attribute
    case ATTR_EFF_WRITE_CRC:
        l_attrSize = sizeof(ATTR_EFF_WRITE_CRC_type);
        break;

    // Get size for ATTR_MSS_CAL_STEP_ENABLE attribute
    case ATTR_MSS_CAL_STEP_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_CAL_STEP_ENABLE_type);
        break;

    // Get size for ATTR_MSS_CUSTOM_TRAINING_ADV_PATTERNS attribute
    case ATTR_MSS_CUSTOM_TRAINING_ADV_PATTERNS:
        l_attrSize = sizeof(ATTR_MSS_CUSTOM_TRAINING_ADV_PATTERNS_type);
        break;

    // Get size for ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS attribute
    case ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS:
        l_attrSize = sizeof(ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS_type);
        break;

    // Get size for ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2 attribute
    case ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2:
        l_attrSize = sizeof(ATTR_MSS_CUSTOM_TRAINING_ADV_BACKUP_PATTERNS2_type);
        break;

    // Get size for ATTR_MSS_CUSTOM_TRAINING_ADV_WR_PATTERN attribute
    case ATTR_MSS_CUSTOM_TRAINING_ADV_WR_PATTERN:
        l_attrSize = sizeof(ATTR_MSS_CUSTOM_TRAINING_ADV_WR_PATTERN_type);
        break;

    // Get size for ATTR_MSS_VREF_CAL_ENABLE attribute
    case ATTR_MSS_VREF_CAL_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_VREF_CAL_ENABLE_type);
        break;

    // Get size for ATTR_MSS_RDVREF_CAL_ENABLE attribute
    case ATTR_MSS_RDVREF_CAL_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_RDVREF_CAL_ENABLE_type);
        break;

    // Get size for ATTR_MSS_CAL_ABORT_ON_ERROR attribute
    case ATTR_MSS_CAL_ABORT_ON_ERROR:
        l_attrSize = sizeof(ATTR_MSS_CAL_ABORT_ON_ERROR_type);
        break;

    // Get size for ATTR_MSS_SLEW_RATE_DATA attribute
    case ATTR_MSS_SLEW_RATE_DATA:
        l_attrSize = sizeof(ATTR_MSS_SLEW_RATE_DATA_type);
        break;

    // Get size for ATTR_MSS_SLEW_RATE_ADR attribute
    case ATTR_MSS_SLEW_RATE_ADR:
        l_attrSize = sizeof(ATTR_MSS_SLEW_RATE_ADR_type);
        break;

    // Get size for ATTR_SCHMOO_MULTIPLE_SETUP_CALL attribute
    case ATTR_SCHMOO_MULTIPLE_SETUP_CALL:
        l_attrSize = sizeof(ATTR_SCHMOO_MULTIPLE_SETUP_CALL_type);
        break;

    // Get size for ATTR_EFF_BUFFER_LATENCY attribute
    case ATTR_EFF_BUFFER_LATENCY:
        l_attrSize = sizeof(ATTR_EFF_BUFFER_LATENCY_type);
        break;

    // Get size for ATTR_EFF_LRDIMM_WORD_X attribute
    case ATTR_EFF_LRDIMM_WORD_X:
        l_attrSize = sizeof(ATTR_EFF_LRDIMM_WORD_X_type);
        break;

    // Get size for ATTR_LRDIMM_MR12_REG attribute
    case ATTR_LRDIMM_MR12_REG:
        l_attrSize = sizeof(ATTR_LRDIMM_MR12_REG_type);
        break;

    // Get size for ATTR_LRDIMM_ADDITIONAL_CNTL_WORDS attribute
    case ATTR_LRDIMM_ADDITIONAL_CNTL_WORDS:
        l_attrSize = sizeof(ATTR_LRDIMM_ADDITIONAL_CNTL_WORDS_type);
        break;

    // Get size for ATTR_LRDIMM_RANK_MULT_MODE attribute
    case ATTR_LRDIMM_RANK_MULT_MODE:
        l_attrSize = sizeof(ATTR_LRDIMM_RANK_MULT_MODE_type);
        break;

    // Get size for ATTR_MSS_EFF_VPD_VERSION attribute
    case ATTR_MSS_EFF_VPD_VERSION:
        l_attrSize = sizeof(ATTR_MSS_EFF_VPD_VERSION_type);
        break;

    // Get size for ATTR_MSS_VOLT_OVERRIDE attribute
    case ATTR_MSS_VOLT_OVERRIDE:
        l_attrSize = sizeof(ATTR_MSS_VOLT_OVERRIDE_type);
        break;

    // Get size for ATTR_MSS_VDDR_OVERIDE_SPD attribute
    case ATTR_MSS_VDDR_OVERIDE_SPD:
        l_attrSize = sizeof(ATTR_MSS_VDDR_OVERIDE_SPD_type);
        break;

    // Get size for ATTR_MSS_VMEM_REGULATOR_MAX_DIMM_COUNT attribute
    case ATTR_MSS_VMEM_REGULATOR_MAX_DIMM_COUNT:
        l_attrSize = sizeof(ATTR_MSS_VMEM_REGULATOR_MAX_DIMM_COUNT_type);
        break;

    // Get size for ATTR_MSS_VREF_DAC_NIBBLE attribute
    case ATTR_MSS_VREF_DAC_NIBBLE:
        l_attrSize = sizeof(ATTR_MSS_VREF_DAC_NIBBLE_type);
        break;

    // Get size for ATTR_MSS_VCCD_OVERRIDE attribute
    case ATTR_MSS_VCCD_OVERRIDE:
        l_attrSize = sizeof(ATTR_MSS_VCCD_OVERRIDE_type);
        break;

    // Get size for ATTR_EFF_DRAM_MAC attribute
    case ATTR_EFF_DRAM_MAC:
        l_attrSize = sizeof(ATTR_EFF_DRAM_MAC_type);
        break;

    // Get size for ATTR_EFF_DRAM_MODULE_BUS_WIDTH attribute
    case ATTR_EFF_DRAM_MODULE_BUS_WIDTH:
        l_attrSize = sizeof(ATTR_EFF_DRAM_MODULE_BUS_WIDTH_type);
        break;

    // Get size for ATTR_EFF_DRAM_RTT_NOM attribute
    case ATTR_EFF_DRAM_RTT_NOM:
        l_attrSize = sizeof(ATTR_EFF_DRAM_RTT_NOM_type);
        break;

    // Get size for ATTR_EFF_DRAM_RTT_WR attribute
    case ATTR_EFF_DRAM_RTT_WR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_RTT_WR_type);
        break;

    // Get size for ATTR_EFF_DRAM_RTT_PARK attribute
    case ATTR_EFF_DRAM_RTT_PARK:
        l_attrSize = sizeof(ATTR_EFF_DRAM_RTT_PARK_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC00 attribute
    case ATTR_EFF_DIMM_DDR4_BC00:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC00_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC01 attribute
    case ATTR_EFF_DIMM_DDR4_BC01:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC01_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC02 attribute
    case ATTR_EFF_DIMM_DDR4_BC02:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC02_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC03 attribute
    case ATTR_EFF_DIMM_DDR4_BC03:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC03_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC04 attribute
    case ATTR_EFF_DIMM_DDR4_BC04:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC04_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC05 attribute
    case ATTR_EFF_DIMM_DDR4_BC05:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC05_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC06 attribute
    case ATTR_EFF_DIMM_DDR4_BC06:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC06_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC07 attribute
    case ATTR_EFF_DIMM_DDR4_BC07:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC07_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC08 attribute
    case ATTR_EFF_DIMM_DDR4_BC08:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC08_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC09 attribute
    case ATTR_EFF_DIMM_DDR4_BC09:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC09_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC0A attribute
    case ATTR_EFF_DIMM_DDR4_BC0A:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC0A_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC0B attribute
    case ATTR_EFF_DIMM_DDR4_BC0B:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC0B_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC0C attribute
    case ATTR_EFF_DIMM_DDR4_BC0C:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC0C_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC0D attribute
    case ATTR_EFF_DIMM_DDR4_BC0D:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC0D_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC0E attribute
    case ATTR_EFF_DIMM_DDR4_BC0E:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC0E_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_BC0F attribute
    case ATTR_EFF_DIMM_DDR4_BC0F:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_BC0F_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F0BC1x attribute
    case ATTR_EFF_DIMM_DDR4_F0BC1x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F0BC1x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BC2x attribute
    case ATTR_EFF_DIMM_DDR4_F30BC2x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BC2x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BC3x attribute
    case ATTR_EFF_DIMM_DDR4_F30BC3x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BC3x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BC4x attribute
    case ATTR_EFF_DIMM_DDR4_F30BC4x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BC4x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BC5x attribute
    case ATTR_EFF_DIMM_DDR4_F30BC5x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BC5x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F0BC6x attribute
    case ATTR_EFF_DIMM_DDR4_F0BC6x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F0BC6x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F70BC7x attribute
    case ATTR_EFF_DIMM_DDR4_F70BC7x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F70BC7x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BC8x attribute
    case ATTR_EFF_DIMM_DDR4_F30BC8x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BC8x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BC9x attribute
    case ATTR_EFF_DIMM_DDR4_F30BC9x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BC9x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BCAx attribute
    case ATTR_EFF_DIMM_DDR4_F30BCAx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BCAx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F30BCBx attribute
    case ATTR_EFF_DIMM_DDR4_F30BCBx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F30BCBx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F0BCCx attribute
    case ATTR_EFF_DIMM_DDR4_F0BCCx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F0BCCx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F0BCDx attribute
    case ATTR_EFF_DIMM_DDR4_F0BCDx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F0BCDx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F0BCEx attribute
    case ATTR_EFF_DIMM_DDR4_F0BCEx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F0BCEx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F0BCFx attribute
    case ATTR_EFF_DIMM_DDR4_F0BCFx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F0BCFx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F1BCCx attribute
    case ATTR_EFF_DIMM_DDR4_F1BCCx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F1BCCx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F1BCDx attribute
    case ATTR_EFF_DIMM_DDR4_F1BCDx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F1BCDx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F1BCEx attribute
    case ATTR_EFF_DIMM_DDR4_F1BCEx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F1BCEx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F1BCFx attribute
    case ATTR_EFF_DIMM_DDR4_F1BCFx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F1BCFx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F2BCEx attribute
    case ATTR_EFF_DIMM_DDR4_F2BCEx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F2BCEx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F4BC0x attribute
    case ATTR_EFF_DIMM_DDR4_F4BC0x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F4BC0x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F4BC1x attribute
    case ATTR_EFF_DIMM_DDR4_F4BC1x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F4BC1x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F4BC2x attribute
    case ATTR_EFF_DIMM_DDR4_F4BC2x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F4BC2x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F4BC3x attribute
    case ATTR_EFF_DIMM_DDR4_F4BC3x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F4BC3x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F4BC4x attribute
    case ATTR_EFF_DIMM_DDR4_F4BC4x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F4BC4x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F4BC5x attribute
    case ATTR_EFF_DIMM_DDR4_F4BC5x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F4BC5x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F4BC6x attribute
    case ATTR_EFF_DIMM_DDR4_F4BC6x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F4BC6x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F5BC0x attribute
    case ATTR_EFF_DIMM_DDR4_F5BC0x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F5BC0x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F5BC1x attribute
    case ATTR_EFF_DIMM_DDR4_F5BC1x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F5BC1x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F5BC2x attribute
    case ATTR_EFF_DIMM_DDR4_F5BC2x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F5BC2x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F5BC3x attribute
    case ATTR_EFF_DIMM_DDR4_F5BC3x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F5BC3x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F5BC5x attribute
    case ATTR_EFF_DIMM_DDR4_F5BC5x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F5BC5x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F5BC6x attribute
    case ATTR_EFF_DIMM_DDR4_F5BC6x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F5BC6x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F6BC0x attribute
    case ATTR_EFF_DIMM_DDR4_F6BC0x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F6BC0x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F6BC1x attribute
    case ATTR_EFF_DIMM_DDR4_F6BC1x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F6BC1x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F6BC2x attribute
    case ATTR_EFF_DIMM_DDR4_F6BC2x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F6BC2x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F6BC3x attribute
    case ATTR_EFF_DIMM_DDR4_F6BC3x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F6BC3x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F6BC4x attribute
    case ATTR_EFF_DIMM_DDR4_F6BC4x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F6BC4x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F6BC5x attribute
    case ATTR_EFF_DIMM_DDR4_F6BC5x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F6BC5x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BC8x attribute
    case ATTR_EFF_DIMM_DDR4_F74BC8x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BC8x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BC9x attribute
    case ATTR_EFF_DIMM_DDR4_F74BC9x:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BC9x_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BCAx attribute
    case ATTR_EFF_DIMM_DDR4_F74BCAx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BCAx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BCBx attribute
    case ATTR_EFF_DIMM_DDR4_F74BCBx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BCBx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BCCx attribute
    case ATTR_EFF_DIMM_DDR4_F74BCCx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BCCx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BCDx attribute
    case ATTR_EFF_DIMM_DDR4_F74BCDx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BCDx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BCEx attribute
    case ATTR_EFF_DIMM_DDR4_F74BCEx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BCEx_type);
        break;

    // Get size for ATTR_EFF_DIMM_DDR4_F74BCFx attribute
    case ATTR_EFF_DIMM_DDR4_F74BCFx:
        l_attrSize = sizeof(ATTR_EFF_DIMM_DDR4_F74BCFx_type);
        break;

    // Get size for ATTR_EFF_DRAM_RON attribute
    case ATTR_EFF_DRAM_RON:
        l_attrSize = sizeof(ATTR_EFF_DRAM_RON_type);
        break;

    // Get size for ATTR_EFF_DRAM_TDQS attribute
    case ATTR_EFF_DRAM_TDQS:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TDQS_type);
        break;

    // Get size for ATTR_EFF_DRAM_ODIC attribute
    case ATTR_EFF_DRAM_ODIC:
        l_attrSize = sizeof(ATTR_EFF_DRAM_ODIC_type);
        break;

    // Get size for ATTR_MSS_LRDIMM_TRAINING_PATTERN attribute
    case ATTR_MSS_LRDIMM_TRAINING_PATTERN:
        l_attrSize = sizeof(ATTR_MSS_LRDIMM_TRAINING_PATTERN_type);
        break;

    // Get size for ATTR_MSS_EFF_ODT_RD attribute
    case ATTR_MSS_EFF_ODT_RD:
        l_attrSize = sizeof(ATTR_MSS_EFF_ODT_RD_type);
        break;

    // Get size for ATTR_MSS_EFF_ODT_WR attribute
    case ATTR_MSS_EFF_ODT_WR:
        l_attrSize = sizeof(ATTR_MSS_EFF_ODT_WR_type);
        break;

    // Get size for ATTR_MSS_EFF_DPHY_WLO attribute
    case ATTR_MSS_EFF_DPHY_WLO:
        l_attrSize = sizeof(ATTR_MSS_EFF_DPHY_WLO_type);
        break;

    // Get size for ATTR_MSS_EFF_DPHY_RLO attribute
    case ATTR_MSS_EFF_DPHY_RLO:
        l_attrSize = sizeof(ATTR_MSS_EFF_DPHY_RLO_type);
        break;

    // Get size for ATTR_EFF_DRAM_TREFI attribute
    case ATTR_EFF_DRAM_TREFI:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TREFI_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRTP attribute
    case ATTR_EFF_DRAM_TRTP:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRTP_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRFC_DLR attribute
    case ATTR_EFF_DRAM_TRFC_DLR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRFC_DLR_type);
        break;

    // Get size for ATTR_EFF_DRAM_TFAW_DLR attribute
    case ATTR_EFF_DRAM_TFAW_DLR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TFAW_DLR_type);
        break;

    // Get size for ATTR_EFF_DRAM_TRRD_DLR attribute
    case ATTR_EFF_DRAM_TRRD_DLR:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TRRD_DLR_type);
        break;

    // Get size for ATTR_EFF_DRAM_TXS attribute
    case ATTR_EFF_DRAM_TXS:
        l_attrSize = sizeof(ATTR_EFF_DRAM_TXS_type);
        break;

    // Get size for ATTR_MSS_IGNORE_PLUG_RULES attribute
    case ATTR_MSS_IGNORE_PLUG_RULES:
        l_attrSize = sizeof(ATTR_MSS_IGNORE_PLUG_RULES_type);
        break;

    // Get size for ATTR_MSS_MVPD_FWMS attribute
    case ATTR_MSS_MVPD_FWMS:
        l_attrSize = sizeof(ATTR_MSS_MVPD_FWMS_type);
        break;

    // Get size for ATTR_MSS_REORDER_QUEUE_SETTING attribute
    case ATTR_MSS_REORDER_QUEUE_SETTING:
        l_attrSize = sizeof(ATTR_MSS_REORDER_QUEUE_SETTING_type);
        break;

    // Get size for ATTR_EFF_RANK_GROUP_OVERRIDE attribute
    case ATTR_EFF_RANK_GROUP_OVERRIDE:
        l_attrSize = sizeof(ATTR_EFF_RANK_GROUP_OVERRIDE_type);
        break;

    // Get size for ATTR_MSS_RTT_NOM_OVERRIDE_DISABLE attribute
    case ATTR_MSS_RTT_NOM_OVERRIDE_DISABLE:
        l_attrSize = sizeof(ATTR_MSS_RTT_NOM_OVERRIDE_DISABLE_type);
        break;

    // Get size for ATTR_MSS_EFF_WR_CRC attribute
    case ATTR_MSS_EFF_WR_CRC:
        l_attrSize = sizeof(ATTR_MSS_EFF_WR_CRC_type);
        break;

    // Get size for ATTR_MSS_WR_VREF_OFFSET attribute
    case ATTR_MSS_WR_VREF_OFFSET:
        l_attrSize = sizeof(ATTR_MSS_WR_VREF_OFFSET_type);
        break;

    // Get size for ATTR_SBE_NVDIMM_IN_PORT attribute
    case ATTR_SBE_NVDIMM_IN_PORT:
        l_attrSize = sizeof(ATTR_SBE_NVDIMM_IN_PORT_type);
        break;

    // Get size for ATTR_SCRATCH_UINT8_1 attribute
    case ATTR_SCRATCH_UINT8_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT8_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT8_2 attribute
    case ATTR_SCRATCH_UINT8_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT8_2_type);
        break;

    // Get size for ATTR_SCRATCH_UINT16_1 attribute
    case ATTR_SCRATCH_UINT16_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT16_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT16_2 attribute
    case ATTR_SCRATCH_UINT16_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT16_2_type);
        break;

    // Get size for ATTR_SCRATCH_UINT32_1 attribute
    case ATTR_SCRATCH_UINT32_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT32_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT32_2 attribute
    case ATTR_SCRATCH_UINT32_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT32_2_type);
        break;

    // Get size for ATTR_SCRATCH_UINT64_1 attribute
    case ATTR_SCRATCH_UINT64_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT64_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT64_2 attribute
    case ATTR_SCRATCH_UINT64_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT64_2_type);
        break;

    // Get size for ATTR_SCRATCH_UINT8_ARRAY_1 attribute
    case ATTR_SCRATCH_UINT8_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT8_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT8_ARRAY_2 attribute
    case ATTR_SCRATCH_UINT8_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT8_ARRAY_2_type);
        break;

    // Get size for ATTR_SCRATCH_UINT16_ARRAY_1 attribute
    case ATTR_SCRATCH_UINT16_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT16_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT16_ARRAY_2 attribute
    case ATTR_SCRATCH_UINT16_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT16_ARRAY_2_type);
        break;

    // Get size for ATTR_SCRATCH_UINT32_ARRAY_1 attribute
    case ATTR_SCRATCH_UINT32_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT32_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT32_ARRAY_2 attribute
    case ATTR_SCRATCH_UINT32_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT32_ARRAY_2_type);
        break;

    // Get size for ATTR_SCRATCH_UINT64_ARRAY_1 attribute
    case ATTR_SCRATCH_UINT64_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT64_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_UINT64_ARRAY_2 attribute
    case ATTR_SCRATCH_UINT64_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_UINT64_ARRAY_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT8_1 attribute
    case ATTR_SCRATCH_INT8_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT8_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT8_2 attribute
    case ATTR_SCRATCH_INT8_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT8_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT16_1 attribute
    case ATTR_SCRATCH_INT16_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT16_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT16_2 attribute
    case ATTR_SCRATCH_INT16_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT16_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT32_1 attribute
    case ATTR_SCRATCH_INT32_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT32_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT32_2 attribute
    case ATTR_SCRATCH_INT32_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT32_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT64_1 attribute
    case ATTR_SCRATCH_INT64_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT64_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT64_2 attribute
    case ATTR_SCRATCH_INT64_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT64_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT8_ARRAY_1 attribute
    case ATTR_SCRATCH_INT8_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT8_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT8_ARRAY_2 attribute
    case ATTR_SCRATCH_INT8_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT8_ARRAY_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT16_ARRAY_1 attribute
    case ATTR_SCRATCH_INT16_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT16_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT16_ARRAY_2 attribute
    case ATTR_SCRATCH_INT16_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT16_ARRAY_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT32_ARRAY_1 attribute
    case ATTR_SCRATCH_INT32_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT32_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT32_ARRAY_2 attribute
    case ATTR_SCRATCH_INT32_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT32_ARRAY_2_type);
        break;

    // Get size for ATTR_SCRATCH_INT64_ARRAY_1 attribute
    case ATTR_SCRATCH_INT64_ARRAY_1:
        l_attrSize = sizeof(ATTR_SCRATCH_INT64_ARRAY_1_type);
        break;

    // Get size for ATTR_SCRATCH_INT64_ARRAY_2 attribute
    case ATTR_SCRATCH_INT64_ARRAY_2:
        l_attrSize = sizeof(ATTR_SCRATCH_INT64_ARRAY_2_type);
        break;

    // Get size for ATTR_I2C_DEV_TYPE attribute
    case ATTR_I2C_DEV_TYPE:
        l_attrSize = sizeof(ATTR_I2C_DEV_TYPE_type);
        break;

    // Get size for ATTR_I2C_SUB_POS attribute
    case ATTR_I2C_SUB_POS:
        l_attrSize = sizeof(ATTR_I2C_SUB_POS_type);
        break;

    // Get size for ATTR_FAPI_POS attribute
    case ATTR_FAPI_POS:
        l_attrSize = sizeof(ATTR_FAPI_POS_type);
        break;

    // Get size for ATTR_IS_SIMULATION attribute
    case ATTR_IS_SIMULATION:
        l_attrSize = sizeof(ATTR_IS_SIMULATION_type);
        break;

    // Get size for ATTR_EXECUTION_PLATFORM attribute
    case ATTR_EXECUTION_PLATFORM:
        l_attrSize = sizeof(ATTR_EXECUTION_PLATFORM_type);
        break;

    // Get size for ATTR_MNFG_FLAGS attribute
    case ATTR_MNFG_FLAGS:
        l_attrSize = sizeof(ATTR_MNFG_FLAGS_type);
        break;

    // Get size for ATTR_RECONFIGURE_LOOP attribute
    case ATTR_RECONFIGURE_LOOP:
        l_attrSize = sizeof(ATTR_RECONFIGURE_LOOP_type);
        break;

    // Get size for ATTR_MINI_EC attribute
    case ATTR_MINI_EC:
        l_attrSize = sizeof(ATTR_MINI_EC_type);
        break;

    // Get size for ATTR_IO_DMI_CEN_TX_MARGIN_RATIO attribute
    case ATTR_IO_DMI_CEN_TX_MARGIN_RATIO:
        l_attrSize = sizeof(ATTR_IO_DMI_CEN_TX_MARGIN_RATIO_type);
        break;

    // Get size for ATTR_IO_DMI_CEN_TX_FFE_POSTCURSOR attribute
    case ATTR_IO_DMI_CEN_TX_FFE_POSTCURSOR:
        l_attrSize = sizeof(ATTR_IO_DMI_CEN_TX_FFE_POSTCURSOR_type);
        break;

    // Get size for ATTR_CEN_DMI_REFCLOCK_RCVR_TERM attribute
    case ATTR_CEN_DMI_REFCLOCK_RCVR_TERM:
        l_attrSize = sizeof(ATTR_CEN_DMI_REFCLOCK_RCVR_TERM_type);
        break;

    // Get size for ATTR_CEN_DDR_REFCLOCK_RCVR_TERM attribute
    case ATTR_CEN_DDR_REFCLOCK_RCVR_TERM:
        l_attrSize = sizeof(ATTR_CEN_DDR_REFCLOCK_RCVR_TERM_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_MODES attribute
    case ATTR_CEN_MCBIST_ADDR_MODES:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_MODES_type);
        break;

    // Get size for ATTR_CEN_MCBIST_RANK attribute
    case ATTR_CEN_MCBIST_RANK:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_RANK_type);
        break;

    // Get size for ATTR_CEN_MCBIST_START_ADDR attribute
    case ATTR_CEN_MCBIST_START_ADDR:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_START_ADDR_type);
        break;

    // Get size for ATTR_CEN_MCBIST_END_ADDR attribute
    case ATTR_CEN_MCBIST_END_ADDR:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_END_ADDR_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ERROR_CAPTURE attribute
    case ATTR_CEN_MCBIST_ERROR_CAPTURE:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ERROR_CAPTURE_type);
        break;

    // Get size for ATTR_CEN_MCBIST_MAX_TIMEOUT attribute
    case ATTR_CEN_MCBIST_MAX_TIMEOUT:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_MAX_TIMEOUT_type);
        break;

    // Get size for ATTR_CEN_MCBIST_PRINT_PORT attribute
    case ATTR_CEN_MCBIST_PRINT_PORT:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_PRINT_PORT_type);
        break;

    // Get size for ATTR_CEN_MCBIST_STOP_ON_ERROR attribute
    case ATTR_CEN_MCBIST_STOP_ON_ERROR:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_STOP_ON_ERROR_type);
        break;

    // Get size for ATTR_CEN_MCBIST_DATA_SEED attribute
    case ATTR_CEN_MCBIST_DATA_SEED:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_DATA_SEED_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_INTER attribute
    case ATTR_CEN_MCBIST_ADDR_INTER:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_INTER_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_NUM_ROWS attribute
    case ATTR_CEN_MCBIST_ADDR_NUM_ROWS:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_NUM_ROWS_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_NUM_COLS attribute
    case ATTR_CEN_MCBIST_ADDR_NUM_COLS:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_NUM_COLS_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_RANK attribute
    case ATTR_CEN_MCBIST_ADDR_RANK:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_RANK_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_BANK attribute
    case ATTR_CEN_MCBIST_ADDR_BANK:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_BANK_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_SLAVE_RANK_ON attribute
    case ATTR_CEN_MCBIST_ADDR_SLAVE_RANK_ON:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_SLAVE_RANK_ON_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_STR_MAP attribute
    case ATTR_CEN_MCBIST_ADDR_STR_MAP:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_STR_MAP_type);
        break;

    // Get size for ATTR_CEN_MCBIST_ADDR_RAND attribute
    case ATTR_CEN_MCBIST_ADDR_RAND:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_ADDR_RAND_type);
        break;

    // Get size for ATTR_CEN_SPD_DRAM_DEVICE_TYPE attribute
    case ATTR_CEN_SPD_DRAM_DEVICE_TYPE:
        l_attrSize = sizeof(ATTR_CEN_SPD_DRAM_DEVICE_TYPE_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_TYPE attribute
    case ATTR_CEN_SPD_MODULE_TYPE:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_TYPE_type);
        break;

    // Get size for ATTR_CEN_SPD_CUSTOM attribute
    case ATTR_CEN_SPD_CUSTOM:
        l_attrSize = sizeof(ATTR_CEN_SPD_CUSTOM_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_DENSITY attribute
    case ATTR_CEN_SPD_SDRAM_DENSITY:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_DENSITY_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_BANKS attribute
    case ATTR_CEN_SPD_SDRAM_BANKS:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_BANKS_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_ROWS attribute
    case ATTR_CEN_SPD_SDRAM_ROWS:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_ROWS_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_COLUMNS attribute
    case ATTR_CEN_SPD_SDRAM_COLUMNS:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_COLUMNS_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_NOMINAL_VOLTAGE attribute
    case ATTR_CEN_SPD_MODULE_NOMINAL_VOLTAGE:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_NOMINAL_VOLTAGE_type);
        break;

    // Get size for ATTR_CEN_SPD_NUM_RANKS attribute
    case ATTR_CEN_SPD_NUM_RANKS:
        l_attrSize = sizeof(ATTR_CEN_SPD_NUM_RANKS_type);
        break;

    // Get size for ATTR_CEN_SPD_DRAM_WIDTH attribute
    case ATTR_CEN_SPD_DRAM_WIDTH:
        l_attrSize = sizeof(ATTR_CEN_SPD_DRAM_WIDTH_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_MEMORY_BUS_WIDTH attribute
    case ATTR_CEN_SPD_MODULE_MEMORY_BUS_WIDTH:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_MEMORY_BUS_WIDTH_type);
        break;

    // Get size for ATTR_CEN_SPD_TCKMIN attribute
    case ATTR_CEN_SPD_TCKMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TCKMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_CAS_LATENCIES_SUPPORTED attribute
    case ATTR_CEN_SPD_CAS_LATENCIES_SUPPORTED:
        l_attrSize = sizeof(ATTR_CEN_SPD_CAS_LATENCIES_SUPPORTED_type);
        break;

    // Get size for ATTR_CEN_SPD_TAAMIN attribute
    case ATTR_CEN_SPD_TAAMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TAAMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TRCDMIN attribute
    case ATTR_CEN_SPD_TRCDMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRCDMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TRPMIN attribute
    case ATTR_CEN_SPD_TRPMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRPMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TRASMIN attribute
    case ATTR_CEN_SPD_TRASMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRASMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TRCMIN attribute
    case ATTR_CEN_SPD_TRCMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRCMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TFAWMIN attribute
    case ATTR_CEN_SPD_TFAWMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TFAWMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_OPTIONAL_FEATURES attribute
    case ATTR_CEN_SPD_SDRAM_OPTIONAL_FEATURES:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_OPTIONAL_FEATURES_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_THERMAL_AND_REFRESH_OPTIONS attribute
    case ATTR_CEN_SPD_SDRAM_THERMAL_AND_REFRESH_OPTIONS:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_THERMAL_AND_REFRESH_OPTIONS_type);
        break;

    // Get size for ATTR_ROW_REPAIR_SPPR_SUPPORTED attribute
    case ATTR_ROW_REPAIR_SPPR_SUPPORTED:
        l_attrSize = sizeof(ATTR_ROW_REPAIR_SPPR_SUPPORTED_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_THERMAL_SENSOR attribute
    case ATTR_CEN_SPD_MODULE_THERMAL_SENSOR:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_THERMAL_SENSOR_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_DEVICE_TYPE attribute
    case ATTR_CEN_SPD_SDRAM_DEVICE_TYPE:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_SIGNAL_LOADING attribute
    case ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_SIGNAL_LOADING:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_DEVICE_TYPE_SIGNAL_LOADING_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_DIE_COUNT attribute
    case ATTR_CEN_SPD_SDRAM_DIE_COUNT:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_DIE_COUNT_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TCKMIN attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TCKMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TCKMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TAAMIN attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TAAMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TAAMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TRCDMIN attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TRCDMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TRCDMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TRPMIN attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TRPMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TRPMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TRCMIN attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TRCMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TRCMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURERS_JEDEC_ID_CODE attribute
    case ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURERS_JEDEC_ID_CODE:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURERS_JEDEC_ID_CODE_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_LOCATION attribute
    case ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_LOCATION:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_LOCATION_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_DATE attribute
    case ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_DATE:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_ID_MODULE_MANUFACTURING_DATE_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_ID_MODULE_SERIAL_NUMBER attribute
    case ATTR_CEN_SPD_MODULE_ID_MODULE_SERIAL_NUMBER:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_ID_MODULE_SERIAL_NUMBER_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_PART_NUMBER attribute
    case ATTR_CEN_SPD_MODULE_PART_NUMBER:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_PART_NUMBER_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_REVISION_CODE attribute
    case ATTR_CEN_SPD_MODULE_REVISION_CODE:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_REVISION_CODE_type);
        break;

    // Get size for ATTR_CEN_SPD_DRAM_MANUFACTURER_JEDEC_ID_CODE attribute
    case ATTR_CEN_SPD_DRAM_MANUFACTURER_JEDEC_ID_CODE:
        l_attrSize = sizeof(ATTR_CEN_SPD_DRAM_MANUFACTURER_JEDEC_ID_CODE_type);
        break;

    // Get size for ATTR_CEN_SPD_BAD_DQ_DATA attribute
    case ATTR_CEN_SPD_BAD_DQ_DATA:
        l_attrSize = sizeof(ATTR_CEN_SPD_BAD_DQ_DATA_type);
        break;

    // Get size for ATTR_CEN_SPD_FTB_DIVIDEND attribute
    case ATTR_CEN_SPD_FTB_DIVIDEND:
        l_attrSize = sizeof(ATTR_CEN_SPD_FTB_DIVIDEND_type);
        break;

    // Get size for ATTR_CEN_SPD_FTB_DIVISOR attribute
    case ATTR_CEN_SPD_FTB_DIVISOR:
        l_attrSize = sizeof(ATTR_CEN_SPD_FTB_DIVISOR_type);
        break;

    // Get size for ATTR_CEN_SPD_MTB_DIVIDEND attribute
    case ATTR_CEN_SPD_MTB_DIVIDEND:
        l_attrSize = sizeof(ATTR_CEN_SPD_MTB_DIVIDEND_type);
        break;

    // Get size for ATTR_CEN_SPD_MTB_DIVISOR attribute
    case ATTR_CEN_SPD_MTB_DIVISOR:
        l_attrSize = sizeof(ATTR_CEN_SPD_MTB_DIVISOR_type);
        break;

    // Get size for ATTR_CEN_SPD_TWRMIN attribute
    case ATTR_CEN_SPD_TWRMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TWRMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TRRDMIN attribute
    case ATTR_CEN_SPD_TRRDMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRRDMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TRFCMIN attribute
    case ATTR_CEN_SPD_TRFCMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRFCMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TWTRMIN attribute
    case ATTR_CEN_SPD_TWTRMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TWTRMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_TRTPMIN attribute
    case ATTR_CEN_SPD_TRTPMIN:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRTPMIN_type);
        break;

    // Get size for ATTR_CEN_SPD_SDRAM_BANKGROUPS_DDR4 attribute
    case ATTR_CEN_SPD_SDRAM_BANKGROUPS_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_SDRAM_BANKGROUPS_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TIMEBASE_MTB_DDR4 attribute
    case ATTR_CEN_SPD_TIMEBASE_MTB_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TIMEBASE_MTB_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TIMEBASE_FTB_DDR4 attribute
    case ATTR_CEN_SPD_TIMEBASE_FTB_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TIMEBASE_FTB_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TCKMAX_DDR4 attribute
    case ATTR_CEN_SPD_TCKMAX_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TCKMAX_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TRFC1MIN_DDR4 attribute
    case ATTR_CEN_SPD_TRFC1MIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRFC1MIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TRFC2MIN_DDR4 attribute
    case ATTR_CEN_SPD_TRFC2MIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRFC2MIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TRFC4MIN_DDR4 attribute
    case ATTR_CEN_SPD_TRFC4MIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRFC4MIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TRRDSMIN_DDR4 attribute
    case ATTR_CEN_SPD_TRRDSMIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRRDSMIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TRRDLMIN_DDR4 attribute
    case ATTR_CEN_SPD_TRRDLMIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TRRDLMIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_TCCDLMIN_DDR4 attribute
    case ATTR_CEN_SPD_TCCDLMIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_TCCDLMIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TCCDLMIN_DDR4 attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TCCDLMIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TCCDLMIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TRRDLMIN_DDR4 attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TRRDLMIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TRRDLMIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TRRDSMIN_DDR4 attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TRRDSMIN_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TRRDSMIN_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_FINE_OFFSET_TCKMAX_DDR4 attribute
    case ATTR_CEN_SPD_FINE_OFFSET_TCKMAX_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_FINE_OFFSET_TCKMAX_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_CRC_BASE_CONFIG_DDR4 attribute
    case ATTR_CEN_SPD_CRC_BASE_CONFIG_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_CRC_BASE_CONFIG_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_DRAM_STEPPING_DDR4 attribute
    case ATTR_CEN_SPD_DRAM_STEPPING_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_DRAM_STEPPING_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_CRC_MNFG_SEC_DDR4 attribute
    case ATTR_CEN_SPD_CRC_MNFG_SEC_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_CRC_MNFG_SEC_DDR4_type);
        break;

    // Get size for ATTR_CEN_SPD_ADDR_MAP_REG_TO_DRAM attribute
    case ATTR_CEN_SPD_ADDR_MAP_REG_TO_DRAM:
        l_attrSize = sizeof(ATTR_CEN_SPD_ADDR_MAP_REG_TO_DRAM_type);
        break;

    // Get size for ATTR_CEN_SPD_MODULE_REVISION_CODE_DDR4 attribute
    case ATTR_CEN_SPD_MODULE_REVISION_CODE_DDR4:
        l_attrSize = sizeof(ATTR_CEN_SPD_MODULE_REVISION_CODE_DDR4_type);
        break;

    // Get size for ATTR_CEN_ISDIMM_MBVPD_INDEX attribute
    case ATTR_CEN_ISDIMM_MBVPD_INDEX:
        l_attrSize = sizeof(ATTR_CEN_ISDIMM_MBVPD_INDEX_type);
        break;

    // Get size for ATTR_CEN_DQ_TO_DIMM_CONN_DQ attribute
    case ATTR_CEN_DQ_TO_DIMM_CONN_DQ:
        l_attrSize = sizeof(ATTR_CEN_DQ_TO_DIMM_CONN_DQ_type);
        break;

    // Get size for ATTR_CEN_MBA_PORT attribute
    case ATTR_CEN_MBA_PORT:
        l_attrSize = sizeof(ATTR_CEN_MBA_PORT_type);
        break;

    // Get size for ATTR_CEN_MBA_DIMM attribute
    case ATTR_CEN_MBA_DIMM:
        l_attrSize = sizeof(ATTR_CEN_MBA_DIMM_type);
        break;

    // Get size for ATTR_CEN_MBVPD_VERSION attribute
    case ATTR_CEN_MBVPD_VERSION:
        l_attrSize = sizeof(ATTR_CEN_MBVPD_VERSION_type);
        break;

    // Get size for ATTR_CEN_MSS_DATABUS_UTIL_PER_MBA attribute
    case ATTR_CEN_MSS_DATABUS_UTIL_PER_MBA:
        l_attrSize = sizeof(ATTR_CEN_MSS_DATABUS_UTIL_PER_MBA_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_ADDRESS_MIRRORING attribute
    case ATTR_CEN_EFF_DRAM_ADDRESS_MIRRORING:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_ADDRESS_MIRRORING_type);
        break;

    // Get size for ATTR_CEN_MSS_FREQ_OVERRIDE attribute
    case ATTR_CEN_MSS_FREQ_OVERRIDE:
        l_attrSize = sizeof(ATTR_CEN_MSS_FREQ_OVERRIDE_type);
        break;

    // Get size for ATTR_CEN_MSS_FREQ attribute
    case ATTR_CEN_MSS_FREQ:
        l_attrSize = sizeof(ATTR_CEN_MSS_FREQ_type);
        break;

    // Get size for ATTR_CEN_MSS_FREQ_BIAS_PERCENTAGE attribute
    case ATTR_CEN_MSS_FREQ_BIAS_PERCENTAGE:
        l_attrSize = sizeof(ATTR_CEN_MSS_FREQ_BIAS_PERCENTAGE_type);
        break;

    // Get size for ATTR_CEN_MSS_VREF_CAL_CNTL attribute
    case ATTR_CEN_MSS_VREF_CAL_CNTL:
        l_attrSize = sizeof(ATTR_CEN_MSS_VREF_CAL_CNTL_type);
        break;

    // Get size for ATTR_CEN_MSS_VREF_CAL_DELTA_FROM_NOMINAL attribute
    case ATTR_CEN_MSS_VREF_CAL_DELTA_FROM_NOMINAL:
        l_attrSize = sizeof(ATTR_CEN_MSS_VREF_CAL_DELTA_FROM_NOMINAL_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_RANKS_CONFIGED attribute
    case ATTR_CEN_EFF_DIMM_RANKS_CONFIGED:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_RANKS_CONFIGED_type);
        break;

    // Get size for ATTR_CEN_EFF_NUM_RANKS_PER_DIMM attribute
    case ATTR_CEN_EFF_NUM_RANKS_PER_DIMM:
        l_attrSize = sizeof(ATTR_CEN_EFF_NUM_RANKS_PER_DIMM_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_TYPE attribute
    case ATTR_CEN_EFF_DIMM_TYPE:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_TYPE_type);
        break;

    // Get size for ATTR_CEN_EFF_CUSTOM_DIMM attribute
    case ATTR_CEN_EFF_CUSTOM_DIMM:
        l_attrSize = sizeof(ATTR_CEN_EFF_CUSTOM_DIMM_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_WIDTH attribute
    case ATTR_CEN_EFF_DRAM_WIDTH:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_WIDTH_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_GEN attribute
    case ATTR_CEN_EFF_DRAM_GEN:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_GEN_type);
        break;

    // Get size for ATTR_CEN_EFF_PRIMARY_RANK_GROUP0 attribute
    case ATTR_CEN_EFF_PRIMARY_RANK_GROUP0:
        l_attrSize = sizeof(ATTR_CEN_EFF_PRIMARY_RANK_GROUP0_type);
        break;

    // Get size for ATTR_CEN_EFF_PRIMARY_RANK_GROUP1 attribute
    case ATTR_CEN_EFF_PRIMARY_RANK_GROUP1:
        l_attrSize = sizeof(ATTR_CEN_EFF_PRIMARY_RANK_GROUP1_type);
        break;

    // Get size for ATTR_CEN_EFF_PRIMARY_RANK_GROUP2 attribute
    case ATTR_CEN_EFF_PRIMARY_RANK_GROUP2:
        l_attrSize = sizeof(ATTR_CEN_EFF_PRIMARY_RANK_GROUP2_type);
        break;

    // Get size for ATTR_CEN_EFF_PRIMARY_RANK_GROUP3 attribute
    case ATTR_CEN_EFF_PRIMARY_RANK_GROUP3:
        l_attrSize = sizeof(ATTR_CEN_EFF_PRIMARY_RANK_GROUP3_type);
        break;

    // Get size for ATTR_CEN_EFF_SECONDARY_RANK_GROUP0 attribute
    case ATTR_CEN_EFF_SECONDARY_RANK_GROUP0:
        l_attrSize = sizeof(ATTR_CEN_EFF_SECONDARY_RANK_GROUP0_type);
        break;

    // Get size for ATTR_CEN_EFF_SECONDARY_RANK_GROUP1 attribute
    case ATTR_CEN_EFF_SECONDARY_RANK_GROUP1:
        l_attrSize = sizeof(ATTR_CEN_EFF_SECONDARY_RANK_GROUP1_type);
        break;

    // Get size for ATTR_CEN_EFF_SECONDARY_RANK_GROUP2 attribute
    case ATTR_CEN_EFF_SECONDARY_RANK_GROUP2:
        l_attrSize = sizeof(ATTR_CEN_EFF_SECONDARY_RANK_GROUP2_type);
        break;

    // Get size for ATTR_CEN_EFF_SECONDARY_RANK_GROUP3 attribute
    case ATTR_CEN_EFF_SECONDARY_RANK_GROUP3:
        l_attrSize = sizeof(ATTR_CEN_EFF_SECONDARY_RANK_GROUP3_type);
        break;

    // Get size for ATTR_CEN_EFF_TERTIARY_RANK_GROUP0 attribute
    case ATTR_CEN_EFF_TERTIARY_RANK_GROUP0:
        l_attrSize = sizeof(ATTR_CEN_EFF_TERTIARY_RANK_GROUP0_type);
        break;

    // Get size for ATTR_CEN_EFF_TERTIARY_RANK_GROUP1 attribute
    case ATTR_CEN_EFF_TERTIARY_RANK_GROUP1:
        l_attrSize = sizeof(ATTR_CEN_EFF_TERTIARY_RANK_GROUP1_type);
        break;

    // Get size for ATTR_CEN_EFF_TERTIARY_RANK_GROUP2 attribute
    case ATTR_CEN_EFF_TERTIARY_RANK_GROUP2:
        l_attrSize = sizeof(ATTR_CEN_EFF_TERTIARY_RANK_GROUP2_type);
        break;

    // Get size for ATTR_CEN_EFF_TERTIARY_RANK_GROUP3 attribute
    case ATTR_CEN_EFF_TERTIARY_RANK_GROUP3:
        l_attrSize = sizeof(ATTR_CEN_EFF_TERTIARY_RANK_GROUP3_type);
        break;

    // Get size for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP0 attribute
    case ATTR_CEN_EFF_QUATERNARY_RANK_GROUP0:
        l_attrSize = sizeof(ATTR_CEN_EFF_QUATERNARY_RANK_GROUP0_type);
        break;

    // Get size for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP1 attribute
    case ATTR_CEN_EFF_QUATERNARY_RANK_GROUP1:
        l_attrSize = sizeof(ATTR_CEN_EFF_QUATERNARY_RANK_GROUP1_type);
        break;

    // Get size for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP2 attribute
    case ATTR_CEN_EFF_QUATERNARY_RANK_GROUP2:
        l_attrSize = sizeof(ATTR_CEN_EFF_QUATERNARY_RANK_GROUP2_type);
        break;

    // Get size for ATTR_CEN_EFF_QUATERNARY_RANK_GROUP3 attribute
    case ATTR_CEN_EFF_QUATERNARY_RANK_GROUP3:
        l_attrSize = sizeof(ATTR_CEN_EFF_QUATERNARY_RANK_GROUP3_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_WR_VREF attribute
    case ATTR_CEN_EFF_DRAM_WR_VREF:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_WR_VREF_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_WR_VREF_SCHMOO attribute
    case ATTR_CEN_EFF_DRAM_WR_VREF_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_WR_VREF_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS attribute
    case ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_DRV_IMP_DQ_DQS_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS attribute
    case ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_DRV_IMP_CLK_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_DRV_IMP_SPCKE_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_DRV_IMP_CNTL_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS attribute
    case ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_RCV_IMP_DQ_DQS_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_SLEW_RATE_DQ_DQS_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_SLEW_RATE_CLK_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_SLEW_RATE_SPCKE_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_SLEW_RATE_ADDR_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_SLEW_RATE_CNTL_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_RD_VREF attribute
    case ATTR_CEN_EFF_RD_VREF:
        l_attrSize = sizeof(ATTR_CEN_EFF_RD_VREF_type);
        break;

    // Get size for ATTR_CEN_EFF_CEN_RD_VREF_SCHMOO attribute
    case ATTR_CEN_EFF_CEN_RD_VREF_SCHMOO:
        l_attrSize = sizeof(ATTR_CEN_EFF_CEN_RD_VREF_SCHMOO_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_SIZE attribute
    case ATTR_CEN_EFF_DIMM_SIZE:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_SIZE_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_BANKS attribute
    case ATTR_CEN_EFF_DRAM_BANKS:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_BANKS_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_ROWS attribute
    case ATTR_CEN_EFF_DRAM_ROWS:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_ROWS_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_COLS attribute
    case ATTR_CEN_EFF_DRAM_COLS:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_COLS_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_DENSITY attribute
    case ATTR_CEN_EFF_DRAM_DENSITY:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_DENSITY_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRCD attribute
    case ATTR_CEN_EFF_DRAM_TRCD:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRCD_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRRD attribute
    case ATTR_CEN_EFF_DRAM_TRRD:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRRD_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRRD_L attribute
    case ATTR_CEN_EFF_DRAM_TRRD_L:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRRD_L_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRP attribute
    case ATTR_CEN_EFF_DRAM_TRP:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRP_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRAS attribute
    case ATTR_CEN_EFF_DRAM_TRAS:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRAS_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRC attribute
    case ATTR_CEN_EFF_DRAM_TRC:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRC_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRFI attribute
    case ATTR_CEN_EFF_DRAM_TRFI:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRFI_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRFC attribute
    case ATTR_CEN_EFF_DRAM_TRFC:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRFC_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TWTR attribute
    case ATTR_CEN_EFF_DRAM_TWTR:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TWTR_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TWTR_L attribute
    case ATTR_CEN_EFF_DRAM_TWTR_L:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TWTR_L_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TRTP attribute
    case ATTR_CEN_EFF_DRAM_TRTP:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TRTP_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TFAW attribute
    case ATTR_CEN_EFF_DRAM_TFAW:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TFAW_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_BL attribute
    case ATTR_CEN_EFF_DRAM_BL:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_BL_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_CL attribute
    case ATTR_CEN_EFF_DRAM_CL:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_CL_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_AL attribute
    case ATTR_CEN_EFF_DRAM_AL:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_AL_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_CWL attribute
    case ATTR_CEN_EFF_DRAM_CWL:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_CWL_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_RBT attribute
    case ATTR_CEN_EFF_DRAM_RBT:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_RBT_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TM attribute
    case ATTR_CEN_EFF_DRAM_TM:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TM_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_DLL_RESET attribute
    case ATTR_CEN_EFF_DRAM_DLL_RESET:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_DLL_RESET_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_WR attribute
    case ATTR_CEN_EFF_DRAM_WR:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_WR_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_DLL_PPD attribute
    case ATTR_CEN_EFF_DRAM_DLL_PPD:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_DLL_PPD_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_DLL_ENABLE attribute
    case ATTR_CEN_EFF_DRAM_DLL_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_DLL_ENABLE_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TDQS attribute
    case ATTR_CEN_EFF_DRAM_TDQS:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TDQS_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_WR_LVL_ENABLE attribute
    case ATTR_CEN_EFF_DRAM_WR_LVL_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_WR_LVL_ENABLE_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_OUTPUT_BUFFER attribute
    case ATTR_CEN_EFF_DRAM_OUTPUT_BUFFER:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_OUTPUT_BUFFER_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_PASR attribute
    case ATTR_CEN_EFF_DRAM_PASR:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_PASR_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_ASR attribute
    case ATTR_CEN_EFF_DRAM_ASR:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_ASR_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_SRT attribute
    case ATTR_CEN_EFF_DRAM_SRT:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_SRT_type);
        break;

    // Get size for ATTR_CEN_EFF_MPR_LOC attribute
    case ATTR_CEN_EFF_MPR_LOC:
        l_attrSize = sizeof(ATTR_CEN_EFF_MPR_LOC_type);
        break;

    // Get size for ATTR_CEN_EFF_MPR_MODE attribute
    case ATTR_CEN_EFF_MPR_MODE:
        l_attrSize = sizeof(ATTR_CEN_EFF_MPR_MODE_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_RCD_CNTL_WORD_0_15 attribute
    case ATTR_CEN_EFF_DIMM_RCD_CNTL_WORD_0_15:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_RCD_CNTL_WORD_0_15_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC00 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC00:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC00_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC01 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC01:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC01_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC02 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC02:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC02_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC03 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC03:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC03_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC04 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC04:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC04_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC05 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC05:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC05_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC06_07 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC06_07:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC06_07_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC08 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC08:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC08_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC09 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC09:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC09_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC10 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC10:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC10_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC11 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC11:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC11_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC12 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC12:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC12_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC13 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC13:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC13_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC14 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC14:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC14_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC15 attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC15:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC15_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_1x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_1x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_1x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_2x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_2x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_2x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_3x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_3x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_3x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_4x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_4x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_4x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_5x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_5x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_5x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_6x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_6x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_6x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_7x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_7x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_7x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_8x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_8x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_8x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_9x attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_9x:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_9x_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_Ax attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_Ax:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_Ax_type);
        break;

    // Get size for ATTR_CEN_EFF_DIMM_DDR4_RC_Bx attribute
    case ATTR_CEN_EFF_DIMM_DDR4_RC_Bx:
        l_attrSize = sizeof(ATTR_CEN_EFF_DIMM_DDR4_RC_Bx_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_MODE attribute
    case ATTR_CEN_EFF_SCHMOO_MODE:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_MODE_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_ADDR_MODE attribute
    case ATTR_CEN_EFF_SCHMOO_ADDR_MODE:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_ADDR_MODE_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_TEST_VALID attribute
    case ATTR_CEN_EFF_SCHMOO_TEST_VALID:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_TEST_VALID_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_PARAM_VALID attribute
    case ATTR_CEN_EFF_SCHMOO_PARAM_VALID:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_PARAM_VALID_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_VREF_STEP attribute
    case ATTR_CEN_EFF_SCHMOO_VREF_STEP:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_VREF_STEP_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_WR_EYE_MIN_MARGIN attribute
    case ATTR_CEN_EFF_SCHMOO_WR_EYE_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_WR_EYE_MIN_MARGIN_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_RD_EYE_MIN_MARGIN attribute
    case ATTR_CEN_EFF_SCHMOO_RD_EYE_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_RD_EYE_MIN_MARGIN_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_DQS_CLK_MIN_MARGIN attribute
    case ATTR_CEN_EFF_SCHMOO_DQS_CLK_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_DQS_CLK_MIN_MARGIN_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_RD_GATE_MIN_MARGIN attribute
    case ATTR_CEN_EFF_SCHMOO_RD_GATE_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_RD_GATE_MIN_MARGIN_type);
        break;

    // Get size for ATTR_CEN_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN attribute
    case ATTR_CEN_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN:
        l_attrSize = sizeof(ATTR_CEN_EFF_SCHMOO_ADDR_CMD_MIN_MARGIN_type);
        break;

    // Get size for ATTR_CEN_EFF_MEMCAL_INTERVAL attribute
    case ATTR_CEN_EFF_MEMCAL_INTERVAL:
        l_attrSize = sizeof(ATTR_CEN_EFF_MEMCAL_INTERVAL_type);
        break;

    // Get size for ATTR_CEN_EFF_ZQCAL_INTERVAL attribute
    case ATTR_CEN_EFF_ZQCAL_INTERVAL:
        l_attrSize = sizeof(ATTR_CEN_EFF_ZQCAL_INTERVAL_type);
        break;

    // Get size for ATTR_CEN_EFF_IBM_TYPE attribute
    case ATTR_CEN_EFF_IBM_TYPE:
        l_attrSize = sizeof(ATTR_CEN_EFF_IBM_TYPE_type);
        break;

    // Get size for ATTR_CEN_EFF_NUM_DROPS_PER_PORT attribute
    case ATTR_CEN_EFF_NUM_DROPS_PER_PORT:
        l_attrSize = sizeof(ATTR_CEN_EFF_NUM_DROPS_PER_PORT_type);
        break;

    // Get size for ATTR_CEN_EFF_STACK_TYPE attribute
    case ATTR_CEN_EFF_STACK_TYPE:
        l_attrSize = sizeof(ATTR_CEN_EFF_STACK_TYPE_type);
        break;

    // Get size for ATTR_CEN_EFF_NUM_MASTER_RANKS_PER_DIMM attribute
    case ATTR_CEN_EFF_NUM_MASTER_RANKS_PER_DIMM:
        l_attrSize = sizeof(ATTR_CEN_EFF_NUM_MASTER_RANKS_PER_DIMM_type);
        break;

    // Get size for ATTR_CEN_EFF_NUM_PACKAGES_PER_RANK attribute
    case ATTR_CEN_EFF_NUM_PACKAGES_PER_RANK:
        l_attrSize = sizeof(ATTR_CEN_EFF_NUM_PACKAGES_PER_RANK_type);
        break;

    // Get size for ATTR_CEN_EFF_NUM_DIES_PER_PACKAGE attribute
    case ATTR_CEN_EFF_NUM_DIES_PER_PACKAGE:
        l_attrSize = sizeof(ATTR_CEN_EFF_NUM_DIES_PER_PACKAGE_type);
        break;

    // Get size for ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_MBA attribute
    case ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_MBA:
        l_attrSize = sizeof(ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_MBA_type);
        break;

    // Get size for ATTR_CEN_MSS_MEM_THROTTLE_DENOMINATOR attribute
    case ATTR_CEN_MSS_MEM_THROTTLE_DENOMINATOR:
        l_attrSize = sizeof(ATTR_CEN_MSS_MEM_THROTTLE_DENOMINATOR_type);
        break;

    // Get size for ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
    case ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_CHIP:
        l_attrSize = sizeof(ATTR_CEN_MSS_MEM_THROTTLE_NUMERATOR_PER_CHIP_type);
        break;

    // Get size for ATTR_CEN_MSS_MEM_WATT_TARGET attribute
    case ATTR_CEN_MSS_MEM_WATT_TARGET:
        l_attrSize = sizeof(ATTR_CEN_MSS_MEM_WATT_TARGET_type);
        break;

    // Get size for ATTR_CEN_MSS_POWER_SLOPE attribute
    case ATTR_CEN_MSS_POWER_SLOPE:
        l_attrSize = sizeof(ATTR_CEN_MSS_POWER_SLOPE_type);
        break;

    // Get size for ATTR_CEN_MSS_POWER_SLOPE2 attribute
    case ATTR_CEN_MSS_POWER_SLOPE2:
        l_attrSize = sizeof(ATTR_CEN_MSS_POWER_SLOPE2_type);
        break;

    // Get size for ATTR_CEN_MSS_POWER_INT attribute
    case ATTR_CEN_MSS_POWER_INT:
        l_attrSize = sizeof(ATTR_CEN_MSS_POWER_INT_type);
        break;

    // Get size for ATTR_CEN_MSS_POWER_INT2 attribute
    case ATTR_CEN_MSS_POWER_INT2:
        l_attrSize = sizeof(ATTR_CEN_MSS_POWER_INT2_type);
        break;

    // Get size for ATTR_CEN_MSS_TOTAL_POWER_SLOPE attribute
    case ATTR_CEN_MSS_TOTAL_POWER_SLOPE:
        l_attrSize = sizeof(ATTR_CEN_MSS_TOTAL_POWER_SLOPE_type);
        break;

    // Get size for ATTR_CEN_MSS_TOTAL_POWER_SLOPE2 attribute
    case ATTR_CEN_MSS_TOTAL_POWER_SLOPE2:
        l_attrSize = sizeof(ATTR_CEN_MSS_TOTAL_POWER_SLOPE2_type);
        break;

    // Get size for ATTR_CEN_MSS_TOTAL_POWER_INT attribute
    case ATTR_CEN_MSS_TOTAL_POWER_INT:
        l_attrSize = sizeof(ATTR_CEN_MSS_TOTAL_POWER_INT_type);
        break;

    // Get size for ATTR_CEN_MSS_TOTAL_POWER_INT2 attribute
    case ATTR_CEN_MSS_TOTAL_POWER_INT2:
        l_attrSize = sizeof(ATTR_CEN_MSS_TOTAL_POWER_INT2_type);
        break;

    // Get size for ATTR_CEN_MSS_CHANNEL_PAIR_MAXPOWER attribute
    case ATTR_CEN_MSS_CHANNEL_PAIR_MAXPOWER:
        l_attrSize = sizeof(ATTR_CEN_MSS_CHANNEL_PAIR_MAXPOWER_type);
        break;

    // Get size for ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_MBA attribute
    case ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_MBA:
        l_attrSize = sizeof(ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_MBA_type);
        break;

    // Get size for ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_DENOMINATOR attribute
    case ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_DENOMINATOR:
        l_attrSize = sizeof(ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_DENOMINATOR_type);
        break;

    // Get size for ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
    case ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_CHIP:
        l_attrSize = sizeof(ATTR_CEN_MSS_RUNTIME_MEM_THROTTLE_NUMERATOR_PER_CHIP_type);
        break;

    // Get size for ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_MBA attribute
    case ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_MBA:
        l_attrSize = sizeof(ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_MBA_type);
        break;

    // Get size for ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
    case ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP:
        l_attrSize = sizeof(ATTR_CEN_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP_type);
        break;

    // Get size for ATTR_CEN_MSS_DERIVED_MBA_ADDR_INTERLEAVE_BIT attribute
    case ATTR_CEN_MSS_DERIVED_MBA_ADDR_INTERLEAVE_BIT:
        l_attrSize = sizeof(ATTR_CEN_MSS_DERIVED_MBA_ADDR_INTERLEAVE_BIT_type);
        break;

    // Get size for ATTR_CEN_MSS_DERIVED_MBA_CACHELINE_INTERLEAVE_MODE attribute
    case ATTR_CEN_MSS_DERIVED_MBA_CACHELINE_INTERLEAVE_MODE:
        l_attrSize = sizeof(ATTR_CEN_MSS_DERIVED_MBA_CACHELINE_INTERLEAVE_MODE_type);
        break;

    // Get size for ATTR_CEN_MRW_MBA_CACHELINE_INTERLEAVE_MODE_CONTROL attribute
    case ATTR_CEN_MRW_MBA_CACHELINE_INTERLEAVE_MODE_CONTROL:
        l_attrSize = sizeof(ATTR_CEN_MRW_MBA_CACHELINE_INTERLEAVE_MODE_CONTROL_type);
        break;

    // Get size for ATTR_CEN_MSS_CACHE_ENABLE attribute
    case ATTR_CEN_MSS_CACHE_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_MSS_CACHE_ENABLE_type);
        break;

    // Get size for ATTR_CEN_MSS_PREFETCH_ENABLE attribute
    case ATTR_CEN_MSS_PREFETCH_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_MSS_PREFETCH_ENABLE_type);
        break;

    // Get size for ATTR_CEN_MSS_CLEANER_ENABLE attribute
    case ATTR_CEN_MSS_CLEANER_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_MSS_CLEANER_ENABLE_type);
        break;

    // Get size for ATTR_CEN_MSS_EFF_DIMM_FUNCTIONAL_VECTOR attribute
    case ATTR_CEN_MSS_EFF_DIMM_FUNCTIONAL_VECTOR:
        l_attrSize = sizeof(ATTR_CEN_MSS_EFF_DIMM_FUNCTIONAL_VECTOR_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_LPASR attribute
    case ATTR_CEN_EFF_DRAM_LPASR:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_LPASR_type);
        break;

    // Get size for ATTR_CEN_EFF_MPR_PAGE attribute
    case ATTR_CEN_EFF_MPR_PAGE:
        l_attrSize = sizeof(ATTR_CEN_EFF_MPR_PAGE_type);
        break;

    // Get size for ATTR_CEN_EFF_GEARDOWN_MODE attribute
    case ATTR_CEN_EFF_GEARDOWN_MODE:
        l_attrSize = sizeof(ATTR_CEN_EFF_GEARDOWN_MODE_type);
        break;

    // Get size for ATTR_CEN_EFF_PER_DRAM_ACCESS attribute
    case ATTR_CEN_EFF_PER_DRAM_ACCESS:
        l_attrSize = sizeof(ATTR_CEN_EFF_PER_DRAM_ACCESS_type);
        break;

    // Get size for ATTR_CEN_EFF_TEMP_READOUT attribute
    case ATTR_CEN_EFF_TEMP_READOUT:
        l_attrSize = sizeof(ATTR_CEN_EFF_TEMP_READOUT_type);
        break;

    // Get size for ATTR_CEN_EFF_FINE_REFRESH_MODE attribute
    case ATTR_CEN_EFF_FINE_REFRESH_MODE:
        l_attrSize = sizeof(ATTR_CEN_EFF_FINE_REFRESH_MODE_type);
        break;

    // Get size for ATTR_CEN_EFF_CRC_WR_LATENCY attribute
    case ATTR_CEN_EFF_CRC_WR_LATENCY:
        l_attrSize = sizeof(ATTR_CEN_EFF_CRC_WR_LATENCY_type);
        break;

    // Get size for ATTR_CEN_EFF_MPR_RD_FORMAT attribute
    case ATTR_CEN_EFF_MPR_RD_FORMAT:
        l_attrSize = sizeof(ATTR_CEN_EFF_MPR_RD_FORMAT_type);
        break;

    // Get size for ATTR_CEN_EFF_MAX_POWERDOWN_MODE attribute
    case ATTR_CEN_EFF_MAX_POWERDOWN_MODE:
        l_attrSize = sizeof(ATTR_CEN_EFF_MAX_POWERDOWN_MODE_type);
        break;

    // Get size for ATTR_CEN_EFF_TEMP_REF_RANGE attribute
    case ATTR_CEN_EFF_TEMP_REF_RANGE:
        l_attrSize = sizeof(ATTR_CEN_EFF_TEMP_REF_RANGE_type);
        break;

    // Get size for ATTR_CEN_EFF_TEMP_REF_MODE attribute
    case ATTR_CEN_EFF_TEMP_REF_MODE:
        l_attrSize = sizeof(ATTR_CEN_EFF_TEMP_REF_MODE_type);
        break;

    // Get size for ATTR_CEN_EFF_INT_VREF_MON attribute
    case ATTR_CEN_EFF_INT_VREF_MON:
        l_attrSize = sizeof(ATTR_CEN_EFF_INT_VREF_MON_type);
        break;

    // Get size for ATTR_CEN_EFF_CS_CMD_LATENCY attribute
    case ATTR_CEN_EFF_CS_CMD_LATENCY:
        l_attrSize = sizeof(ATTR_CEN_EFF_CS_CMD_LATENCY_type);
        break;

    // Get size for ATTR_CEN_EFF_SELF_REF_ABORT attribute
    case ATTR_CEN_EFF_SELF_REF_ABORT:
        l_attrSize = sizeof(ATTR_CEN_EFF_SELF_REF_ABORT_type);
        break;

    // Get size for ATTR_CEN_EFF_RD_PREAMBLE_TRAIN attribute
    case ATTR_CEN_EFF_RD_PREAMBLE_TRAIN:
        l_attrSize = sizeof(ATTR_CEN_EFF_RD_PREAMBLE_TRAIN_type);
        break;

    // Get size for ATTR_CEN_EFF_RD_PREAMBLE attribute
    case ATTR_CEN_EFF_RD_PREAMBLE:
        l_attrSize = sizeof(ATTR_CEN_EFF_RD_PREAMBLE_type);
        break;

    // Get size for ATTR_CEN_EFF_WR_PREAMBLE attribute
    case ATTR_CEN_EFF_WR_PREAMBLE:
        l_attrSize = sizeof(ATTR_CEN_EFF_WR_PREAMBLE_type);
        break;

    // Get size for ATTR_CEN_EFF_CA_PARITY_LATENCY attribute
    case ATTR_CEN_EFF_CA_PARITY_LATENCY:
        l_attrSize = sizeof(ATTR_CEN_EFF_CA_PARITY_LATENCY_type);
        break;

    // Get size for ATTR_CEN_EFF_CRC_ERROR_CLEAR attribute
    case ATTR_CEN_EFF_CRC_ERROR_CLEAR:
        l_attrSize = sizeof(ATTR_CEN_EFF_CRC_ERROR_CLEAR_type);
        break;

    // Get size for ATTR_CEN_EFF_CA_PARITY_ERROR_STATUS attribute
    case ATTR_CEN_EFF_CA_PARITY_ERROR_STATUS:
        l_attrSize = sizeof(ATTR_CEN_EFF_CA_PARITY_ERROR_STATUS_type);
        break;

    // Get size for ATTR_CEN_EFF_ODT_INPUT_BUFF attribute
    case ATTR_CEN_EFF_ODT_INPUT_BUFF:
        l_attrSize = sizeof(ATTR_CEN_EFF_ODT_INPUT_BUFF_type);
        break;

    // Get size for ATTR_CEN_EFF_CA_PARITY attribute
    case ATTR_CEN_EFF_CA_PARITY:
        l_attrSize = sizeof(ATTR_CEN_EFF_CA_PARITY_type);
        break;

    // Get size for ATTR_CEN_EFF_DATA_MASK attribute
    case ATTR_CEN_EFF_DATA_MASK:
        l_attrSize = sizeof(ATTR_CEN_EFF_DATA_MASK_type);
        break;

    // Get size for ATTR_CEN_EFF_WRITE_DBI attribute
    case ATTR_CEN_EFF_WRITE_DBI:
        l_attrSize = sizeof(ATTR_CEN_EFF_WRITE_DBI_type);
        break;

    // Get size for ATTR_CEN_EFF_READ_DBI attribute
    case ATTR_CEN_EFF_READ_DBI:
        l_attrSize = sizeof(ATTR_CEN_EFF_READ_DBI_type);
        break;

    // Get size for ATTR_CEN_EFF_VREF_DQ_TRAIN_VALUE attribute
    case ATTR_CEN_EFF_VREF_DQ_TRAIN_VALUE:
        l_attrSize = sizeof(ATTR_CEN_EFF_VREF_DQ_TRAIN_VALUE_type);
        break;

    // Get size for ATTR_CEN_EFF_VREF_DQ_TRAIN_RANGE attribute
    case ATTR_CEN_EFF_VREF_DQ_TRAIN_RANGE:
        l_attrSize = sizeof(ATTR_CEN_EFF_VREF_DQ_TRAIN_RANGE_type);
        break;

    // Get size for ATTR_CEN_EFF_VREF_DQ_TRAIN_ENABLE attribute
    case ATTR_CEN_EFF_VREF_DQ_TRAIN_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_EFF_VREF_DQ_TRAIN_ENABLE_type);
        break;

    // Get size for ATTR_CEN_EFF_DRAM_TCCD_L attribute
    case ATTR_CEN_EFF_DRAM_TCCD_L:
        l_attrSize = sizeof(ATTR_CEN_EFF_DRAM_TCCD_L_type);
        break;

    // Get size for ATTR_CEN_TCCD_L attribute
    case ATTR_CEN_TCCD_L:
        l_attrSize = sizeof(ATTR_CEN_TCCD_L_type);
        break;

    // Get size for ATTR_CEN_EFF_WRITE_CRC attribute
    case ATTR_CEN_EFF_WRITE_CRC:
        l_attrSize = sizeof(ATTR_CEN_EFF_WRITE_CRC_type);
        break;

    // Get size for ATTR_CEN_MSS_CAL_STEP_ENABLE attribute
    case ATTR_CEN_MSS_CAL_STEP_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_MSS_CAL_STEP_ENABLE_type);
        break;

    // Get size for ATTR_CEN_MSS_DRAMINIT_RESET_DISABLE attribute
    case ATTR_CEN_MSS_DRAMINIT_RESET_DISABLE:
        l_attrSize = sizeof(ATTR_CEN_MSS_DRAMINIT_RESET_DISABLE_type);
        break;

    // Get size for ATTR_CEN_MSS_SLEW_RATE_DATA attribute
    case ATTR_CEN_MSS_SLEW_RATE_DATA:
        l_attrSize = sizeof(ATTR_CEN_MSS_SLEW_RATE_DATA_type);
        break;

    // Get size for ATTR_CEN_MSS_SLEW_RATE_ADR attribute
    case ATTR_CEN_MSS_SLEW_RATE_ADR:
        l_attrSize = sizeof(ATTR_CEN_MSS_SLEW_RATE_ADR_type);
        break;

    // Get size for ATTR_CEN_MSS_ALLOW_SINGLE_PORT attribute
    case ATTR_CEN_MSS_ALLOW_SINGLE_PORT:
        l_attrSize = sizeof(ATTR_CEN_MSS_ALLOW_SINGLE_PORT_type);
        break;

    // Get size for ATTR_CEN_MSS_DQS_SWIZZLE_TYPE attribute
    case ATTR_CEN_MSS_DQS_SWIZZLE_TYPE:
        l_attrSize = sizeof(ATTR_CEN_MSS_DQS_SWIZZLE_TYPE_type);
        break;

    // Get size for ATTR_CEN_MSS_PSRO attribute
    case ATTR_CEN_MSS_PSRO:
        l_attrSize = sizeof(ATTR_CEN_MSS_PSRO_type);
        break;

    // Get size for ATTR_CEN_MSS_NWELL_MISPLACEMENT attribute
    case ATTR_CEN_MSS_NWELL_MISPLACEMENT:
        l_attrSize = sizeof(ATTR_CEN_MSS_NWELL_MISPLACEMENT_type);
        break;

    // Get size for ATTR_CEN_MSS_BLUEWATERFALL_BROKEN attribute
    case ATTR_CEN_MSS_BLUEWATERFALL_BROKEN:
        l_attrSize = sizeof(ATTR_CEN_MSS_BLUEWATERFALL_BROKEN_type);
        break;

    // Get size for ATTR_CEN_MCBIST_PATTERN attribute
    case ATTR_CEN_MCBIST_PATTERN:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_PATTERN_type);
        break;

    // Get size for ATTR_CEN_MCBIST_TEST_TYPE attribute
    case ATTR_CEN_MCBIST_TEST_TYPE:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_TEST_TYPE_type);
        break;

    // Get size for ATTR_CEN_MCBIST_PRINTING_DISABLE attribute
    case ATTR_CEN_MCBIST_PRINTING_DISABLE:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_PRINTING_DISABLE_type);
        break;

    // Get size for ATTR_CEN_MCBIST_DATA_ENABLE attribute
    case ATTR_CEN_MCBIST_DATA_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_DATA_ENABLE_type);
        break;

    // Get size for ATTR_CEN_MCBIST_USER_RANK attribute
    case ATTR_CEN_MCBIST_USER_RANK:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_USER_RANK_type);
        break;

    // Get size for ATTR_CEN_MCBIST_USER_BANK attribute
    case ATTR_CEN_MCBIST_USER_BANK:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_USER_BANK_type);
        break;

    // Get size for ATTR_CEN_SCHMOO_MULTIPLE_SETUP_CALL attribute
    case ATTR_CEN_SCHMOO_MULTIPLE_SETUP_CALL:
        l_attrSize = sizeof(ATTR_CEN_SCHMOO_MULTIPLE_SETUP_CALL_type);
        break;

    // Get size for ATTR_CEN_LRDIMM_RANK_MULT_MODE attribute
    case ATTR_CEN_LRDIMM_RANK_MULT_MODE:
        l_attrSize = sizeof(ATTR_CEN_LRDIMM_RANK_MULT_MODE_type);
        break;

    // Get size for ATTR_CEN_MSS_THROTTLE_CONTROL_RAS_WEIGHT attribute
    case ATTR_CEN_MSS_THROTTLE_CONTROL_RAS_WEIGHT:
        l_attrSize = sizeof(ATTR_CEN_MSS_THROTTLE_CONTROL_RAS_WEIGHT_type);
        break;

    // Get size for ATTR_CEN_MSS_THROTTLE_CONTROL_CAS_WEIGHT attribute
    case ATTR_CEN_MSS_THROTTLE_CONTROL_CAS_WEIGHT:
        l_attrSize = sizeof(ATTR_CEN_MSS_THROTTLE_CONTROL_CAS_WEIGHT_type);
        break;

    // Get size for ATTR_CEN_MCBIST_RANDOM_SEED_VALUE attribute
    case ATTR_CEN_MCBIST_RANDOM_SEED_VALUE:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_RANDOM_SEED_VALUE_type);
        break;

    // Get size for ATTR_CEN_MCBIST_RANDOM_SEED_TYPE attribute
    case ATTR_CEN_MCBIST_RANDOM_SEED_TYPE:
        l_attrSize = sizeof(ATTR_CEN_MCBIST_RANDOM_SEED_TYPE_type);
        break;

    // Get size for ATTR_CEN_MSS_INIT_STATE attribute
    case ATTR_CEN_MSS_INIT_STATE:
        l_attrSize = sizeof(ATTR_CEN_MSS_INIT_STATE_type);
        break;

    // Get size for ATTR_CEN_MSS_EFF_VPD_VERSION attribute
    case ATTR_CEN_MSS_EFF_VPD_VERSION:
        l_attrSize = sizeof(ATTR_CEN_MSS_EFF_VPD_VERSION_type);
        break;

    // Get size for ATTR_CEN_MSS_NEST_CAPABLE_FREQUENCIES attribute
    case ATTR_CEN_MSS_NEST_CAPABLE_FREQUENCIES:
        l_attrSize = sizeof(ATTR_CEN_MSS_NEST_CAPABLE_FREQUENCIES_type);
        break;

    // Get size for ATTR_CEN_MRW_STRICT_MBA_PLUG_RULE_CHECKING attribute
    case ATTR_CEN_MRW_STRICT_MBA_PLUG_RULE_CHECKING:
        l_attrSize = sizeof(ATTR_CEN_MRW_STRICT_MBA_PLUG_RULE_CHECKING_type);
        break;

    // Get size for ATTR_CEN_MRW_CDIMM_MASTER_I2C_TEMP_SENSOR_ENABLE attribute
    case ATTR_CEN_MRW_CDIMM_MASTER_I2C_TEMP_SENSOR_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_MRW_CDIMM_MASTER_I2C_TEMP_SENSOR_ENABLE_type);
        break;

    // Get size for ATTR_CEN_MRW_CDIMM_SPARE_I2C_TEMP_SENSOR_ENABLE attribute
    case ATTR_CEN_MRW_CDIMM_SPARE_I2C_TEMP_SENSOR_ENABLE:
        l_attrSize = sizeof(ATTR_CEN_MRW_CDIMM_SPARE_I2C_TEMP_SENSOR_ENABLE_type);
        break;

    // Get size for ATTR_CEN_MRW_THERMAL_MEMORY_POWER_LIMIT attribute
    case ATTR_CEN_MRW_THERMAL_MEMORY_POWER_LIMIT:
        l_attrSize = sizeof(ATTR_CEN_MRW_THERMAL_MEMORY_POWER_LIMIT_type);
        break;

    // Get size for ATTR_CEN_MRW_POWER_CONTROL_REQUESTED attribute
    case ATTR_CEN_MRW_POWER_CONTROL_REQUESTED:
        l_attrSize = sizeof(ATTR_CEN_MRW_POWER_CONTROL_REQUESTED_type);
        break;

    // Get size for ATTR_CEN_MSS_DDR3_VDDR_SLOPE attribute
    case ATTR_CEN_MSS_DDR3_VDDR_SLOPE:
        l_attrSize = sizeof(ATTR_CEN_MSS_DDR3_VDDR_SLOPE_type);
        break;

    // Get size for ATTR_CEN_MSS_DDR3_VDDR_INTERCEPT attribute
    case ATTR_CEN_MSS_DDR3_VDDR_INTERCEPT:
        l_attrSize = sizeof(ATTR_CEN_MSS_DDR3_VDDR_INTERCEPT_type);
        break;

    // Get size for ATTR_CEN_MSS_DDR4_VDDR_SLOPE attribute
    case ATTR_CEN_MSS_DDR4_VDDR_SLOPE:
        l_attrSize = sizeof(ATTR_CEN_MSS_DDR4_VDDR_SLOPE_type);
        break;

    // Get size for ATTR_CEN_MSS_DDR4_VDDR_INTERCEPT attribute
    case ATTR_CEN_MSS_DDR4_VDDR_INTERCEPT:
        l_attrSize = sizeof(ATTR_CEN_MSS_DDR4_VDDR_INTERCEPT_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_OVERRIDE attribute
    case ATTR_CEN_MSS_VOLT_OVERRIDE:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_OVERRIDE_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_COMPLIANT_DIMMS attribute
    case ATTR_CEN_MSS_VOLT_COMPLIANT_DIMMS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_COMPLIANT_DIMMS_type);
        break;

    // Get size for ATTR_CEN_MRW_MCS_PREFETCH_RETRY_THRESHOLD attribute
    case ATTR_CEN_MRW_MCS_PREFETCH_RETRY_THRESHOLD:
        l_attrSize = sizeof(ATTR_CEN_MRW_MCS_PREFETCH_RETRY_THRESHOLD_type);
        break;

    // Get size for ATTR_CEN_MRW_DDR3_VDDR_MAX_LIMIT attribute
    case ATTR_CEN_MRW_DDR3_VDDR_MAX_LIMIT:
        l_attrSize = sizeof(ATTR_CEN_MRW_DDR3_VDDR_MAX_LIMIT_type);
        break;

    // Get size for ATTR_CEN_MRW_DDR4_VDDR_MAX_LIMIT attribute
    case ATTR_CEN_MRW_DDR4_VDDR_MAX_LIMIT:
        l_attrSize = sizeof(ATTR_CEN_MRW_DDR4_VDDR_MAX_LIMIT_type);
        break;

    // Get size for ATTR_MRW_MEM_SENSOR_CACHE_ADDR_MAP attribute
    case ATTR_MRW_MEM_SENSOR_CACHE_ADDR_MAP:
        l_attrSize = sizeof(ATTR_MRW_MEM_SENSOR_CACHE_ADDR_MAP_type);
        break;

    // Get size for ATTR_CEN_MRW_WR_VREF_CHECK_VREF_STEP_SIZE attribute
    case ATTR_CEN_MRW_WR_VREF_CHECK_VREF_STEP_SIZE:
        l_attrSize = sizeof(ATTR_CEN_MRW_WR_VREF_CHECK_VREF_STEP_SIZE_type);
        break;

    // Get size for ATTR_MEM_DRAM_CWL attribute
    case ATTR_MEM_DRAM_CWL:
        l_attrSize = sizeof(ATTR_MEM_DRAM_CWL_type);
        break;

    // Get size for ATTR_MEM_RDIMM_BUFFER_DELAY attribute
    case ATTR_MEM_RDIMM_BUFFER_DELAY:
        l_attrSize = sizeof(ATTR_MEM_RDIMM_BUFFER_DELAY_type);
        break;

    // Get size for ATTR_MEM_REORDER_QUEUE_SETTING attribute
    case ATTR_MEM_REORDER_QUEUE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_REORDER_QUEUE_SETTING_type);
        break;

    // Get size for ATTR_MEM_2N_MODE attribute
    case ATTR_MEM_2N_MODE:
        l_attrSize = sizeof(ATTR_MEM_2N_MODE_type);
        break;

    // Get size for ATTR_MEM_VPD_DQ_MAP attribute
    case ATTR_MEM_VPD_DQ_MAP:
        l_attrSize = sizeof(ATTR_MEM_VPD_DQ_MAP_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC0F attribute
    case ATTR_MEM_DIMM_DDR4_F0RC0F:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC0F_type);
        break;

    // Get size for ATTR_MEM_CS_CMD_LATENCY attribute
    case ATTR_MEM_CS_CMD_LATENCY:
        l_attrSize = sizeof(ATTR_MEM_CS_CMD_LATENCY_type);
        break;

    // Get size for ATTR_MEM_CA_PARITY_LATENCY attribute
    case ATTR_MEM_CA_PARITY_LATENCY:
        l_attrSize = sizeof(ATTR_MEM_CA_PARITY_LATENCY_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC02 attribute
    case ATTR_MEM_DIMM_DDR4_F0RC02:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC02_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC03 attribute
    case ATTR_MEM_DIMM_DDR4_F0RC03:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC03_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC04 attribute
    case ATTR_MEM_DIMM_DDR4_F0RC04:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC04_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC05 attribute
    case ATTR_MEM_DIMM_DDR4_F0RC05:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC05_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC0B attribute
    case ATTR_MEM_DIMM_DDR4_F0RC0B:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC0B_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC1X attribute
    case ATTR_MEM_DIMM_DDR4_F0RC1X:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC1X_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F0RC7X attribute
    case ATTR_MEM_DIMM_DDR4_F0RC7X:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F0RC7X_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F1RC00 attribute
    case ATTR_MEM_DIMM_DDR4_F1RC00:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F1RC00_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F1RC02 attribute
    case ATTR_MEM_DIMM_DDR4_F1RC02:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F1RC02_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F1RC03 attribute
    case ATTR_MEM_DIMM_DDR4_F1RC03:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F1RC03_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F1RC04 attribute
    case ATTR_MEM_DIMM_DDR4_F1RC04:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F1RC04_type);
        break;

    // Get size for ATTR_MEM_DIMM_DDR4_F1RC05 attribute
    case ATTR_MEM_DIMM_DDR4_F1RC05:
        l_attrSize = sizeof(ATTR_MEM_DIMM_DDR4_F1RC05_type);
        break;

    // Get size for ATTR_MEM_DIMM_POS_METADATA attribute
    case ATTR_MEM_DIMM_POS_METADATA:
        l_attrSize = sizeof(ATTR_MEM_DIMM_POS_METADATA_type);
        break;

    // Get size for ATTR_MEM_DRAM_GEN_METADATA attribute
    case ATTR_MEM_DRAM_GEN_METADATA:
        l_attrSize = sizeof(ATTR_MEM_DRAM_GEN_METADATA_type);
        break;

    // Get size for ATTR_MEM_DIMM_TYPE_METADATA attribute
    case ATTR_MEM_DIMM_TYPE_METADATA:
        l_attrSize = sizeof(ATTR_MEM_DIMM_TYPE_METADATA_type);
        break;

    // Get size for ATTR_MSS_OMI_EDPL_DISABLE attribute
    case ATTR_MSS_OMI_EDPL_DISABLE:
        l_attrSize = sizeof(ATTR_MSS_OMI_EDPL_DISABLE_type);
        break;

    // Get size for ATTR_MEM_IGNORE_PLUG_RULES attribute
    case ATTR_MEM_IGNORE_PLUG_RULES:
        l_attrSize = sizeof(ATTR_MEM_IGNORE_PLUG_RULES_type);
        break;

    // Get size for ATTR_OMI_DL_PREIPL_PRBS_TIME attribute
    case ATTR_OMI_DL_PREIPL_PRBS_TIME:
        l_attrSize = sizeof(ATTR_OMI_DL_PREIPL_PRBS_TIME_type);
        break;

    // Get size for ATTR_MEM_SI_SIGNATURE_HASH attribute
    case ATTR_MEM_SI_SIGNATURE_HASH:
        l_attrSize = sizeof(ATTR_MEM_SI_SIGNATURE_HASH_type);
        break;

    // Get size for ATTR_MEM_SI_DIMM_RCD_IBT_CA attribute
    case ATTR_MEM_SI_DIMM_RCD_IBT_CA:
        l_attrSize = sizeof(ATTR_MEM_SI_DIMM_RCD_IBT_CA_type);
        break;

    // Get size for ATTR_MEM_SI_DIMM_RCD_IBT_CKE attribute
    case ATTR_MEM_SI_DIMM_RCD_IBT_CKE:
        l_attrSize = sizeof(ATTR_MEM_SI_DIMM_RCD_IBT_CKE_type);
        break;

    // Get size for ATTR_MEM_SI_DIMM_RCD_IBT_CS attribute
    case ATTR_MEM_SI_DIMM_RCD_IBT_CS:
        l_attrSize = sizeof(ATTR_MEM_SI_DIMM_RCD_IBT_CS_type);
        break;

    // Get size for ATTR_MEM_SI_DIMM_RCD_IBT_ODT attribute
    case ATTR_MEM_SI_DIMM_RCD_IBT_ODT:
        l_attrSize = sizeof(ATTR_MEM_SI_DIMM_RCD_IBT_ODT_type);
        break;

    // Get size for ATTR_MEM_SI_DRAM_DRV_IMP_DQ_DQS attribute
    case ATTR_MEM_SI_DRAM_DRV_IMP_DQ_DQS:
        l_attrSize = sizeof(ATTR_MEM_SI_DRAM_DRV_IMP_DQ_DQS_type);
        break;

    // Get size for ATTR_MEM_SI_DRAM_PREAMBLE attribute
    case ATTR_MEM_SI_DRAM_PREAMBLE:
        l_attrSize = sizeof(ATTR_MEM_SI_DRAM_PREAMBLE_type);
        break;

    // Get size for ATTR_MEM_SI_DRAM_RTT_NOM attribute
    case ATTR_MEM_SI_DRAM_RTT_NOM:
        l_attrSize = sizeof(ATTR_MEM_SI_DRAM_RTT_NOM_type);
        break;

    // Get size for ATTR_MEM_SI_DRAM_RTT_PARK attribute
    case ATTR_MEM_SI_DRAM_RTT_PARK:
        l_attrSize = sizeof(ATTR_MEM_SI_DRAM_RTT_PARK_type);
        break;

    // Get size for ATTR_MEM_SI_DRAM_RTT_WR attribute
    case ATTR_MEM_SI_DRAM_RTT_WR:
        l_attrSize = sizeof(ATTR_MEM_SI_DRAM_RTT_WR_type);
        break;

    // Get size for ATTR_MEM_SI_GEARDOWN_MODE attribute
    case ATTR_MEM_SI_GEARDOWN_MODE:
        l_attrSize = sizeof(ATTR_MEM_SI_GEARDOWN_MODE_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_DQ_DQS attribute
    case ATTR_MEM_SI_MC_DRV_DQ_DQS:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_DQ_DQS_type);
        break;

    // Get size for ATTR_MEM_SI_MC_RCV_EQ_DQ_DQS attribute
    case ATTR_MEM_SI_MC_RCV_EQ_DQ_DQS:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_RCV_EQ_DQ_DQS_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_EQ_DQ_DQS attribute
    case ATTR_MEM_SI_MC_DRV_EQ_DQ_DQS:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_EQ_DQ_DQS_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_IMP_CLK attribute
    case ATTR_MEM_SI_MC_DRV_IMP_CLK:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_IMP_CLK_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_IMP_CMD_ADDR attribute
    case ATTR_MEM_SI_MC_DRV_IMP_CMD_ADDR:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_IMP_CMD_ADDR_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_IMP_CNTL attribute
    case ATTR_MEM_SI_MC_DRV_IMP_CNTL:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_IMP_CNTL_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_IMP_CSCID attribute
    case ATTR_MEM_SI_MC_DRV_IMP_CSCID:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_IMP_CSCID_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN attribute
    case ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_DOWN_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP attribute
    case ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_IMP_DQ_DQS_PULL_UP_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CLK attribute
    case ATTR_MEM_SI_MC_DRV_SLEW_RATE_CLK:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_SLEW_RATE_CLK_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR attribute
    case ATTR_MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_SLEW_RATE_CMD_ADDR_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CNTL attribute
    case ATTR_MEM_SI_MC_DRV_SLEW_RATE_CNTL:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_SLEW_RATE_CNTL_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_SLEW_RATE_CSCID attribute
    case ATTR_MEM_SI_MC_DRV_SLEW_RATE_CSCID:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_SLEW_RATE_CSCID_type);
        break;

    // Get size for ATTR_MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS attribute
    case ATTR_MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_DRV_SLEW_RATE_DQ_DQS_type);
        break;

    // Get size for ATTR_MEM_SI_MC_RCV_IMP_ALERT_N attribute
    case ATTR_MEM_SI_MC_RCV_IMP_ALERT_N:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_RCV_IMP_ALERT_N_type);
        break;

    // Get size for ATTR_MEM_SI_MC_RCV_IMP_DQ_DQS attribute
    case ATTR_MEM_SI_MC_RCV_IMP_DQ_DQS:
        l_attrSize = sizeof(ATTR_MEM_SI_MC_RCV_IMP_DQ_DQS_type);
        break;

    // Get size for ATTR_MEM_SI_ODT_RD attribute
    case ATTR_MEM_SI_ODT_RD:
        l_attrSize = sizeof(ATTR_MEM_SI_ODT_RD_type);
        break;

    // Get size for ATTR_MEM_SI_ODT_WR attribute
    case ATTR_MEM_SI_ODT_WR:
        l_attrSize = sizeof(ATTR_MEM_SI_ODT_WR_type);
        break;

    // Get size for ATTR_MEM_SI_VREF_DRAM_WR attribute
    case ATTR_MEM_SI_VREF_DRAM_WR:
        l_attrSize = sizeof(ATTR_MEM_SI_VREF_DRAM_WR_type);
        break;

    // Get size for ATTR_MEM_SI_VREF_MC_RD attribute
    case ATTR_MEM_SI_VREF_MC_RD:
        l_attrSize = sizeof(ATTR_MEM_SI_VREF_MC_RD_type);
        break;

    // Get size for ATTR_MEM_SI_WINDAGE_RD_CTR attribute
    case ATTR_MEM_SI_WINDAGE_RD_CTR:
        l_attrSize = sizeof(ATTR_MEM_SI_WINDAGE_RD_CTR_type);
        break;

    // Get size for ATTR_MSS_MRW_SUPPORTED_FREQ attribute
    case ATTR_MSS_MRW_SUPPORTED_FREQ:
        l_attrSize = sizeof(ATTR_MSS_MRW_SUPPORTED_FREQ_type);
        break;

    // Get size for ATTR_MEM_MRW_IS_PLANAR attribute
    case ATTR_MEM_MRW_IS_PLANAR:
        l_attrSize = sizeof(ATTR_MEM_MRW_IS_PLANAR_type);
        break;

    // Get size for ATTR_MSS_MRW_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
    case ATTR_MSS_MRW_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT:
        l_attrSize = sizeof(ATTR_MSS_MRW_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT_type);
        break;

    // Get size for ATTR_MSS_MRW_THERMAL_MEMORY_POWER_LIMIT attribute
    case ATTR_MSS_MRW_THERMAL_MEMORY_POWER_LIMIT:
        l_attrSize = sizeof(ATTR_MSS_MRW_THERMAL_MEMORY_POWER_LIMIT_type);
        break;

    // Get size for ATTR_MSS_MRW_PWR_INTERCEPT attribute
    case ATTR_MSS_MRW_PWR_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_MRW_PWR_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_MRW_PWR_SLOPE attribute
    case ATTR_MSS_MRW_PWR_SLOPE:
        l_attrSize = sizeof(ATTR_MSS_MRW_PWR_SLOPE_type);
        break;

    // Get size for ATTR_MSS_MRW_REFRESH_RATE_REQUEST attribute
    case ATTR_MSS_MRW_REFRESH_RATE_REQUEST:
        l_attrSize = sizeof(ATTR_MSS_MRW_REFRESH_RATE_REQUEST_type);
        break;

    // Get size for ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT attribute
    case ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT:
        l_attrSize = sizeof(ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_type);
        break;

    // Get size for ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_IDLE attribute
    case ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_IDLE:
        l_attrSize = sizeof(ATTR_MSS_MRW_DIMM_POWER_CURVE_PERCENT_UPLIFT_IDLE_type);
        break;

    // Get size for ATTR_MSS_MRW_MEM_M_DRAM_CLOCKS attribute
    case ATTR_MSS_MRW_MEM_M_DRAM_CLOCKS:
        l_attrSize = sizeof(ATTR_MSS_MRW_MEM_M_DRAM_CLOCKS_type);
        break;

    // Get size for ATTR_MSS_MRW_MAX_DRAM_DATABUS_UTIL attribute
    case ATTR_MSS_MRW_MAX_DRAM_DATABUS_UTIL:
        l_attrSize = sizeof(ATTR_MSS_MRW_MAX_DRAM_DATABUS_UTIL_type);
        break;

    // Get size for ATTR_MSS_MRW_POWER_CONTROL_REQUESTED attribute
    case ATTR_MSS_MRW_POWER_CONTROL_REQUESTED:
        l_attrSize = sizeof(ATTR_MSS_MRW_POWER_CONTROL_REQUESTED_type);
        break;

    // Get size for ATTR_MSS_MRW_IDLE_POWER_CONTROL_REQUESTED attribute
    case ATTR_MSS_MRW_IDLE_POWER_CONTROL_REQUESTED:
        l_attrSize = sizeof(ATTR_MSS_MRW_IDLE_POWER_CONTROL_REQUESTED_type);
        break;

    // Get size for ATTR_MSS_MRW_VMEM_REGULATOR_POWER_LIMIT_PER_DIMM_ADJ_ENABLE attribute
    case ATTR_MSS_MRW_VMEM_REGULATOR_POWER_LIMIT_PER_DIMM_ADJ_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_MRW_VMEM_REGULATOR_POWER_LIMIT_PER_DIMM_ADJ_ENABLE_type);
        break;

    // Get size for ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR3 attribute
    case ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR3:
        l_attrSize = sizeof(ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR3_type);
        break;

    // Get size for ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR4 attribute
    case ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR4:
        l_attrSize = sizeof(ATTR_MRW_VMEM_REGULATOR_MEMORY_POWER_LIMIT_PER_DIMM_DDR4_type);
        break;

    // Get size for ATTR_MSS_MRW_MAX_NUMBER_DIMMS_POSSIBLE_PER_VMEM_REGULATOR attribute
    case ATTR_MSS_MRW_MAX_NUMBER_DIMMS_POSSIBLE_PER_VMEM_REGULATOR:
        l_attrSize = sizeof(ATTR_MSS_MRW_MAX_NUMBER_DIMMS_POSSIBLE_PER_VMEM_REGULATOR_type);
        break;

    // Get size for ATTR_MSS_MRW_AVDD_OFFSET_ENABLE attribute
    case ATTR_MSS_MRW_AVDD_OFFSET_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_MRW_AVDD_OFFSET_ENABLE_type);
        break;

    // Get size for ATTR_MSS_MRW_VDD_OFFSET_ENABLE attribute
    case ATTR_MSS_MRW_VDD_OFFSET_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_MRW_VDD_OFFSET_ENABLE_type);
        break;

    // Get size for ATTR_MSS_MRW_VCS_OFFSET_ENABLE attribute
    case ATTR_MSS_MRW_VCS_OFFSET_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_MRW_VCS_OFFSET_ENABLE_type);
        break;

    // Get size for ATTR_MSS_MRW_VPP_OFFSET_ENABLE attribute
    case ATTR_MSS_MRW_VPP_OFFSET_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_MRW_VPP_OFFSET_ENABLE_type);
        break;

    // Get size for ATTR_MSS_MRW_VDDR_OFFSET_ENABLE attribute
    case ATTR_MSS_MRW_VDDR_OFFSET_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_MRW_VDDR_OFFSET_ENABLE_type);
        break;

    // Get size for ATTR_MSS_MRW_FINE_REFRESH_MODE attribute
    case ATTR_MSS_MRW_FINE_REFRESH_MODE:
        l_attrSize = sizeof(ATTR_MSS_MRW_FINE_REFRESH_MODE_type);
        break;

    // Get size for ATTR_MSS_MRW_TEMP_REFRESH_RANGE attribute
    case ATTR_MSS_MRW_TEMP_REFRESH_RANGE:
        l_attrSize = sizeof(ATTR_MSS_MRW_TEMP_REFRESH_RANGE_type);
        break;

    // Get size for ATTR_MSS_MRW_RESET_DELAY_BEFORE_CAL attribute
    case ATTR_MSS_MRW_RESET_DELAY_BEFORE_CAL:
        l_attrSize = sizeof(ATTR_MSS_MRW_RESET_DELAY_BEFORE_CAL_type);
        break;

    // Get size for ATTR_MSS_MRW_PERIODIC_MEMCAL_MODE_OPTIONS attribute
    case ATTR_MSS_MRW_PERIODIC_MEMCAL_MODE_OPTIONS:
        l_attrSize = sizeof(ATTR_MSS_MRW_PERIODIC_MEMCAL_MODE_OPTIONS_type);
        break;

    // Get size for ATTR_MSS_MRW_PERIODIC_ZQCAL_MODE_OPTIONS attribute
    case ATTR_MSS_MRW_PERIODIC_ZQCAL_MODE_OPTIONS:
        l_attrSize = sizeof(ATTR_MSS_MRW_PERIODIC_ZQCAL_MODE_OPTIONS_type);
        break;

    // Get size for ATTR_MSS_MRW_DRAM_2N_MODE attribute
    case ATTR_MSS_MRW_DRAM_2N_MODE:
        l_attrSize = sizeof(ATTR_MSS_MRW_DRAM_2N_MODE_type);
        break;

    // Get size for ATTR_MSS_MRW_DRAM_WRITE_CRC attribute
    case ATTR_MSS_MRW_DRAM_WRITE_CRC:
        l_attrSize = sizeof(ATTR_MSS_MRW_DRAM_WRITE_CRC_type);
        break;

    // Get size for ATTR_MSS_MRW_TEMP_REFRESH_MODE attribute
    case ATTR_MSS_MRW_TEMP_REFRESH_MODE:
        l_attrSize = sizeof(ATTR_MSS_MRW_TEMP_REFRESH_MODE_type);
        break;

    // Get size for ATTR_MSS_MRW_FORCE_BCMODE_OFF attribute
    case ATTR_MSS_MRW_FORCE_BCMODE_OFF:
        l_attrSize = sizeof(ATTR_MSS_MRW_FORCE_BCMODE_OFF_type);
        break;

    // Get size for ATTR_MSS_MRW_NVDIMM_PLUG_RULES attribute
    case ATTR_MSS_MRW_NVDIMM_PLUG_RULES:
        l_attrSize = sizeof(ATTR_MSS_MRW_NVDIMM_PLUG_RULES_type);
        break;

    // Get size for ATTR_MSS_MRW_ALLOW_UNSUPPORTED_RCW attribute
    case ATTR_MSS_MRW_ALLOW_UNSUPPORTED_RCW:
        l_attrSize = sizeof(ATTR_MSS_MRW_ALLOW_UNSUPPORTED_RCW_type);
        break;

    // Get size for ATTR_MSS_MRW_SUPPORTED_DRAM_WIDTH attribute
    case ATTR_MSS_MRW_SUPPORTED_DRAM_WIDTH:
        l_attrSize = sizeof(ATTR_MSS_MRW_SUPPORTED_DRAM_WIDTH_type);
        break;

    // Get size for ATTR_MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT attribute
    case ATTR_MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT:
        l_attrSize = sizeof(ATTR_MSS_MRW_OCMB_THERMAL_MEMORY_POWER_LIMIT_type);
        break;

    // Get size for ATTR_MSS_MRW_OCMB_PWR_SLOPE attribute
    case ATTR_MSS_MRW_OCMB_PWR_SLOPE:
        l_attrSize = sizeof(ATTR_MSS_MRW_OCMB_PWR_SLOPE_type);
        break;

    // Get size for ATTR_MSS_MRW_OCMB_PWR_INTERCEPT attribute
    case ATTR_MSS_MRW_OCMB_PWR_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_MRW_OCMB_PWR_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT attribute
    case ATTR_MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT:
        l_attrSize = sizeof(ATTR_MSS_MRW_OCMB_CURRENT_CURVE_WITH_LIMIT_type);
        break;

    // Get size for ATTR_MSS_MRW_SAFEMODE_DRAM_DATABUS_UTIL attribute
    case ATTR_MSS_MRW_SAFEMODE_DRAM_DATABUS_UTIL:
        l_attrSize = sizeof(ATTR_MSS_MRW_SAFEMODE_DRAM_DATABUS_UTIL_type);
        break;

    // Get size for ATTR_MEM_PORT_POS_OF_FAIL_THROTTLE attribute
    case ATTR_MEM_PORT_POS_OF_FAIL_THROTTLE:
        l_attrSize = sizeof(ATTR_MEM_PORT_POS_OF_FAIL_THROTTLE_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_GEN attribute
    case ATTR_MEM_EFF_DRAM_GEN:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_GEN_type);
        break;

    // Get size for ATTR_MEM_EFF_DIMM_TYPE attribute
    case ATTR_MEM_EFF_DIMM_TYPE:
        l_attrSize = sizeof(ATTR_MEM_EFF_DIMM_TYPE_type);
        break;

    // Get size for ATTR_MEM_EFF_HYBRID_MEMORY_TYPE attribute
    case ATTR_MEM_EFF_HYBRID_MEMORY_TYPE:
        l_attrSize = sizeof(ATTR_MEM_EFF_HYBRID_MEMORY_TYPE_type);
        break;

    // Get size for ATTR_MEM_EFF_HYBRID attribute
    case ATTR_MEM_EFF_HYBRID:
        l_attrSize = sizeof(ATTR_MEM_EFF_HYBRID_type);
        break;

    // Get size for ATTR_MEM_EFF_HOST_TO_DDR_SPEED_RATIO attribute
    case ATTR_MEM_EFF_HOST_TO_DDR_SPEED_RATIO:
        l_attrSize = sizeof(ATTR_MEM_EFF_HOST_TO_DDR_SPEED_RATIO_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_DENSITY attribute
    case ATTR_MEM_EFF_DRAM_DENSITY:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_DENSITY_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_BANK_BITS attribute
    case ATTR_MEM_EFF_DRAM_BANK_BITS:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_BANK_BITS_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_BANK_GROUP_BITS attribute
    case ATTR_MEM_EFF_DRAM_BANK_GROUP_BITS:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_BANK_GROUP_BITS_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_COLUMN_BITS attribute
    case ATTR_MEM_EFF_DRAM_COLUMN_BITS:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_COLUMN_BITS_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_ROW_BITS attribute
    case ATTR_MEM_EFF_DRAM_ROW_BITS:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_ROW_BITS_type);
        break;

    // Get size for ATTR_MEM_EFF_PRIM_DIE_COUNT attribute
    case ATTR_MEM_EFF_PRIM_DIE_COUNT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PRIM_DIE_COUNT_type);
        break;

    // Get size for ATTR_MEM_EFF_PRIM_STACK_TYPE attribute
    case ATTR_MEM_EFF_PRIM_STACK_TYPE:
        l_attrSize = sizeof(ATTR_MEM_EFF_PRIM_STACK_TYPE_type);
        break;

    // Get size for ATTR_MEM_EFF_PRIM_BUS_WIDTH attribute
    case ATTR_MEM_EFF_PRIM_BUS_WIDTH:
        l_attrSize = sizeof(ATTR_MEM_EFF_PRIM_BUS_WIDTH_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_PPR attribute
    case ATTR_MEM_EFF_DRAM_PPR:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_PPR_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_SOFT_PPR attribute
    case ATTR_MEM_EFF_DRAM_SOFT_PPR:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_SOFT_PPR_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRCD attribute
    case ATTR_MEM_EFF_DRAM_TRCD:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRCD_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRP attribute
    case ATTR_MEM_EFF_DRAM_TRP:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRP_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRAS attribute
    case ATTR_MEM_EFF_DRAM_TRAS:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRAS_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRC attribute
    case ATTR_MEM_EFF_DRAM_TRC:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRC_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRFC attribute
    case ATTR_MEM_EFF_DRAM_TRFC:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRFC_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TFAW attribute
    case ATTR_MEM_EFF_DRAM_TFAW:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TFAW_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRRD_S attribute
    case ATTR_MEM_EFF_DRAM_TRRD_S:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRRD_S_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRRD_L attribute
    case ATTR_MEM_EFF_DRAM_TRRD_L:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRRD_L_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRRD_DLR attribute
    case ATTR_MEM_EFF_DRAM_TRRD_DLR:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRRD_DLR_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TCCD_L attribute
    case ATTR_MEM_EFF_DRAM_TCCD_L:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TCCD_L_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TWR attribute
    case ATTR_MEM_EFF_DRAM_TWR:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TWR_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TWTR_S attribute
    case ATTR_MEM_EFF_DRAM_TWTR_S:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TWTR_S_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TWTR_L attribute
    case ATTR_MEM_EFF_DRAM_TWTR_L:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TWTR_L_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TMAW attribute
    case ATTR_MEM_EFF_DRAM_TMAW:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TMAW_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_WIDTH attribute
    case ATTR_MEM_EFF_DRAM_WIDTH:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_WIDTH_type);
        break;

    // Get size for ATTR_MEM_EFF_LOGICAL_RANKS_PER_DIMM attribute
    case ATTR_MEM_EFF_LOGICAL_RANKS_PER_DIMM:
        l_attrSize = sizeof(ATTR_MEM_EFF_LOGICAL_RANKS_PER_DIMM_type);
        break;

    // Get size for ATTR_MEM_3DS_HEIGHT attribute
    case ATTR_MEM_3DS_HEIGHT:
        l_attrSize = sizeof(ATTR_MEM_3DS_HEIGHT_type);
        break;

    // Get size for ATTR_MEM_EFF_REGISTER_TYPE attribute
    case ATTR_MEM_EFF_REGISTER_TYPE:
        l_attrSize = sizeof(ATTR_MEM_EFF_REGISTER_TYPE_type);
        break;

    // Get size for ATTR_MEM_EFF_MODULE_MFG_ID attribute
    case ATTR_MEM_EFF_MODULE_MFG_ID:
        l_attrSize = sizeof(ATTR_MEM_EFF_MODULE_MFG_ID_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_MFG_ID attribute
    case ATTR_MEM_EFF_DRAM_MFG_ID:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_MFG_ID_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_MODULE_HEIGHT attribute
    case ATTR_MEM_EFF_DRAM_MODULE_HEIGHT:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_MODULE_HEIGHT_type);
        break;

    // Get size for ATTR_MEM_EFF_RCD_MFG_ID attribute
    case ATTR_MEM_EFF_RCD_MFG_ID:
        l_attrSize = sizeof(ATTR_MEM_EFF_RCD_MFG_ID_type);
        break;

    // Get size for ATTR_MEM_EFF_REGISTER_REV attribute
    case ATTR_MEM_EFF_REGISTER_REV:
        l_attrSize = sizeof(ATTR_MEM_EFF_REGISTER_REV_type);
        break;

    // Get size for ATTR_MEM_EFF_PACKAGE_RANK_MAP attribute
    case ATTR_MEM_EFF_PACKAGE_RANK_MAP:
        l_attrSize = sizeof(ATTR_MEM_EFF_PACKAGE_RANK_MAP_type);
        break;

    // Get size for ATTR_MEM_EFF_NIBBLE_MAP attribute
    case ATTR_MEM_EFF_NIBBLE_MAP:
        l_attrSize = sizeof(ATTR_MEM_EFF_NIBBLE_MAP_type);
        break;

    // Get size for ATTR_MEM_EFF_DIMM_SIZE attribute
    case ATTR_MEM_EFF_DIMM_SIZE:
        l_attrSize = sizeof(ATTR_MEM_EFF_DIMM_SIZE_type);
        break;

    // Get size for ATTR_MEM_EFF_DIMM_SPARE attribute
    case ATTR_MEM_EFF_DIMM_SPARE:
        l_attrSize = sizeof(ATTR_MEM_EFF_DIMM_SPARE_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_CL attribute
    case ATTR_MEM_EFF_DRAM_CL:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_CL_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_MDS attribute
    case ATTR_MEM_EFF_DRAM_MDS:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_MDS_type);
        break;

    // Get size for ATTR_MEM_EFF_NUM_MASTER_RANKS_PER_DIMM attribute
    case ATTR_MEM_EFF_NUM_MASTER_RANKS_PER_DIMM:
        l_attrSize = sizeof(ATTR_MEM_EFF_NUM_MASTER_RANKS_PER_DIMM_type);
        break;

    // Get size for ATTR_MEM_EFF_DIMM_RANKS_CONFIGED attribute
    case ATTR_MEM_EFF_DIMM_RANKS_CONFIGED:
        l_attrSize = sizeof(ATTR_MEM_EFF_DIMM_RANKS_CONFIGED_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TREFI attribute
    case ATTR_MEM_EFF_DRAM_TREFI:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TREFI_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRTP attribute
    case ATTR_MEM_EFF_DRAM_TRTP:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRTP_type);
        break;

    // Get size for ATTR_MEM_EFF_DRAM_TRFC_DLR attribute
    case ATTR_MEM_EFF_DRAM_TRFC_DLR:
        l_attrSize = sizeof(ATTR_MEM_EFF_DRAM_TRFC_DLR_type);
        break;

    // Get size for ATTR_MEM_EFF_FREQ attribute
    case ATTR_MEM_EFF_FREQ:
        l_attrSize = sizeof(ATTR_MEM_EFF_FREQ_type);
        break;

    // Get size for ATTR_MEM_EFF_VOLT_VDDR attribute
    case ATTR_MEM_EFF_VOLT_VDDR:
        l_attrSize = sizeof(ATTR_MEM_EFF_VOLT_VDDR_type);
        break;

    // Get size for ATTR_MEM_EFF_VOLT_VPP attribute
    case ATTR_MEM_EFF_VOLT_VPP:
        l_attrSize = sizeof(ATTR_MEM_EFF_VOLT_VPP_type);
        break;

    // Get size for ATTR_MEM_EFF_SPD_REVISION attribute
    case ATTR_MEM_EFF_SPD_REVISION:
        l_attrSize = sizeof(ATTR_MEM_EFF_SPD_REVISION_type);
        break;

    // Get size for ATTR_EXPLR_ENABLE_US_TMPL_1 attribute
    case ATTR_EXPLR_ENABLE_US_TMPL_1:
        l_attrSize = sizeof(ATTR_EXPLR_ENABLE_US_TMPL_1_type);
        break;

    // Get size for ATTR_EXPLR_ENABLE_US_TMPL_5 attribute
    case ATTR_EXPLR_ENABLE_US_TMPL_5:
        l_attrSize = sizeof(ATTR_EXPLR_ENABLE_US_TMPL_5_type);
        break;

    // Get size for ATTR_EXPLR_ENABLE_US_TMPL_9 attribute
    case ATTR_EXPLR_ENABLE_US_TMPL_9:
        l_attrSize = sizeof(ATTR_EXPLR_ENABLE_US_TMPL_9_type);
        break;

    // Get size for ATTR_EXPLR_ENABLE_US_TMPL_A attribute
    case ATTR_EXPLR_ENABLE_US_TMPL_A:
        l_attrSize = sizeof(ATTR_EXPLR_ENABLE_US_TMPL_A_type);
        break;

    // Get size for ATTR_EXPLR_ENABLE_US_TMPL_B attribute
    case ATTR_EXPLR_ENABLE_US_TMPL_B:
        l_attrSize = sizeof(ATTR_EXPLR_ENABLE_US_TMPL_B_type);
        break;

    // Get size for ATTR_EXPLR_TMPL_0_PACING attribute
    case ATTR_EXPLR_TMPL_0_PACING:
        l_attrSize = sizeof(ATTR_EXPLR_TMPL_0_PACING_type);
        break;

    // Get size for ATTR_EXPLR_TMPL_1_PACING attribute
    case ATTR_EXPLR_TMPL_1_PACING:
        l_attrSize = sizeof(ATTR_EXPLR_TMPL_1_PACING_type);
        break;

    // Get size for ATTR_EXPLR_TMPL_5_PACING attribute
    case ATTR_EXPLR_TMPL_5_PACING:
        l_attrSize = sizeof(ATTR_EXPLR_TMPL_5_PACING_type);
        break;

    // Get size for ATTR_EXPLR_TMPL_9_PACING attribute
    case ATTR_EXPLR_TMPL_9_PACING:
        l_attrSize = sizeof(ATTR_EXPLR_TMPL_9_PACING_type);
        break;

    // Get size for ATTR_EXPLR_TMPL_B_PACING attribute
    case ATTR_EXPLR_TMPL_B_PACING:
        l_attrSize = sizeof(ATTR_EXPLR_TMPL_B_PACING_type);
        break;

    // Get size for ATTR_EXPLR_SHRT_BACKOFF_TIMER attribute
    case ATTR_EXPLR_SHRT_BACKOFF_TIMER:
        l_attrSize = sizeof(ATTR_EXPLR_SHRT_BACKOFF_TIMER_type);
        break;

    // Get size for ATTR_EXPLR_METADATA_ENABLE attribute
    case ATTR_EXPLR_METADATA_ENABLE:
        l_attrSize = sizeof(ATTR_EXPLR_METADATA_ENABLE_type);
        break;

    // Get size for ATTR_EXPLR_PASID_BASE attribute
    case ATTR_EXPLR_PASID_BASE:
        l_attrSize = sizeof(ATTR_EXPLR_PASID_BASE_type);
        break;

    // Get size for ATTR_EXPLR_ACTAG_BASE attribute
    case ATTR_EXPLR_ACTAG_BASE:
        l_attrSize = sizeof(ATTR_EXPLR_ACTAG_BASE_type);
        break;

    // Get size for ATTR_EXPLR_AFU_ACTAG_LEN attribute
    case ATTR_EXPLR_AFU_ACTAG_LEN:
        l_attrSize = sizeof(ATTR_EXPLR_AFU_ACTAG_LEN_type);
        break;

    // Get size for ATTR_EXPLR_PASID_LEN attribute
    case ATTR_EXPLR_PASID_LEN:
        l_attrSize = sizeof(ATTR_EXPLR_PASID_LEN_type);
        break;

    // Get size for ATTR_MEM_EFF_PSTATES attribute
    case ATTR_MEM_EFF_PSTATES:
        l_attrSize = sizeof(ATTR_MEM_EFF_PSTATES_type);
        break;

    // Get size for ATTR_MEM_EFF_BYTE_ENABLES attribute
    case ATTR_MEM_EFF_BYTE_ENABLES:
        l_attrSize = sizeof(ATTR_MEM_EFF_BYTE_ENABLES_type);
        break;

    // Get size for ATTR_MEM_EFF_NIBBLE_ENABLES attribute
    case ATTR_MEM_EFF_NIBBLE_ENABLES:
        l_attrSize = sizeof(ATTR_MEM_EFF_NIBBLE_ENABLES_type);
        break;

    // Get size for ATTR_MEM_EFF_FOUR_RANK_MODE attribute
    case ATTR_MEM_EFF_FOUR_RANK_MODE:
        l_attrSize = sizeof(ATTR_MEM_EFF_FOUR_RANK_MODE_type);
        break;

    // Get size for ATTR_MEM_EFF_MRAM_SUPPORT attribute
    case ATTR_MEM_EFF_MRAM_SUPPORT:
        l_attrSize = sizeof(ATTR_MEM_EFF_MRAM_SUPPORT_type);
        break;

    // Get size for ATTR_MEM_EFF_DDP_COMPATIBILITY attribute
    case ATTR_MEM_EFF_DDP_COMPATIBILITY:
        l_attrSize = sizeof(ATTR_MEM_EFF_DDP_COMPATIBILITY_type);
        break;

    // Get size for ATTR_MEM_EFF_TSV_8H_SUPPORT attribute
    case ATTR_MEM_EFF_TSV_8H_SUPPORT:
        l_attrSize = sizeof(ATTR_MEM_EFF_TSV_8H_SUPPORT_type);
        break;

    // Get size for ATTR_MEM_EFF_MODULE_THERMAL_SENSORS attribute
    case ATTR_MEM_EFF_MODULE_THERMAL_SENSORS:
        l_attrSize = sizeof(ATTR_MEM_EFF_MODULE_THERMAL_SENSORS_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_0_AVAIL attribute
    case ATTR_MEM_EFF_THERM_SENSOR_0_AVAIL:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_0_AVAIL_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_0_TYPE attribute
    case ATTR_MEM_EFF_THERM_SENSOR_0_TYPE:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_0_TYPE_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_0_USAGE attribute
    case ATTR_MEM_EFF_THERM_SENSOR_0_USAGE:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_0_USAGE_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_0_I2C_ADDR attribute
    case ATTR_MEM_EFF_THERM_SENSOR_0_I2C_ADDR:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_0_I2C_ADDR_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_1_AVAIL attribute
    case ATTR_MEM_EFF_THERM_SENSOR_1_AVAIL:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_1_AVAIL_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_1_TYPE attribute
    case ATTR_MEM_EFF_THERM_SENSOR_1_TYPE:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_1_TYPE_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_1_USAGE attribute
    case ATTR_MEM_EFF_THERM_SENSOR_1_USAGE:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_1_USAGE_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_1_I2C_ADDR attribute
    case ATTR_MEM_EFF_THERM_SENSOR_1_I2C_ADDR:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_1_I2C_ADDR_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_DIFF_AVAIL attribute
    case ATTR_MEM_EFF_THERM_SENSOR_DIFF_AVAIL:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_DIFF_AVAIL_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_DIFF_TYPE attribute
    case ATTR_MEM_EFF_THERM_SENSOR_DIFF_TYPE:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_DIFF_TYPE_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_DIFF_USAGE attribute
    case ATTR_MEM_EFF_THERM_SENSOR_DIFF_USAGE:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_DIFF_USAGE_type);
        break;

    // Get size for ATTR_MEM_EFF_THERM_SENSOR_DIFF_I2C_ADDR attribute
    case ATTR_MEM_EFF_THERM_SENSOR_DIFF_I2C_ADDR:
        l_attrSize = sizeof(ATTR_MEM_EFF_THERM_SENSOR_DIFF_I2C_ADDR_type);
        break;

    // Get size for ATTR_MEM_EFF_SUPPORTED_RCD attribute
    case ATTR_MEM_EFF_SUPPORTED_RCD:
        l_attrSize = sizeof(ATTR_MEM_EFF_SUPPORTED_RCD_type);
        break;

    // Get size for ATTR_MSS_OCMB_ENTERPRISE_MODE attribute
    case ATTR_MSS_OCMB_ENTERPRISE_MODE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_ENTERPRISE_MODE_type);
        break;

    // Get size for ATTR_MSS_OCMB_ENTERPRISE_POLICY attribute
    case ATTR_MSS_OCMB_ENTERPRISE_POLICY:
        l_attrSize = sizeof(ATTR_MSS_OCMB_ENTERPRISE_POLICY_type);
        break;

    // Get size for ATTR_MSS_OCMB_NONENTERPRISE_MODE_OVERRIDE attribute
    case ATTR_MSS_OCMB_NONENTERPRISE_MODE_OVERRIDE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_NONENTERPRISE_MODE_OVERRIDE_type);
        break;

    // Get size for ATTR_MSS_OCMB_HALF_DIMM_MODE attribute
    case ATTR_MSS_OCMB_HALF_DIMM_MODE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_HALF_DIMM_MODE_type);
        break;

    // Get size for ATTR_MSS_OCMB_HALF_DIMM_MODE_OVERRIDE attribute
    case ATTR_MSS_OCMB_HALF_DIMM_MODE_OVERRIDE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_HALF_DIMM_MODE_OVERRIDE_type);
        break;

    // Get size for ATTR_MSS_MEM_MVPD_FWMS attribute
    case ATTR_MSS_MEM_MVPD_FWMS:
        l_attrSize = sizeof(ATTR_MSS_MEM_MVPD_FWMS_type);
        break;

    // Get size for ATTR_MEM_EXP_INIT_VREF_DQ attribute
    case ATTR_MEM_EXP_INIT_VREF_DQ:
        l_attrSize = sizeof(ATTR_MEM_EXP_INIT_VREF_DQ_type);
        break;

    // Get size for ATTR_MEM_EXP_INIT_PHY_VREF attribute
    case ATTR_MEM_EXP_INIT_PHY_VREF:
        l_attrSize = sizeof(ATTR_MEM_EXP_INIT_PHY_VREF_type);
        break;

    // Get size for ATTR_MEM_EXP_RCD_DIC attribute
    case ATTR_MEM_EXP_RCD_DIC:
        l_attrSize = sizeof(ATTR_MEM_EXP_RCD_DIC_type);
        break;

    // Get size for ATTR_MEM_EXP_RCD_VOLTAGE_CTRL attribute
    case ATTR_MEM_EXP_RCD_VOLTAGE_CTRL:
        l_attrSize = sizeof(ATTR_MEM_EXP_RCD_VOLTAGE_CTRL_type);
        break;

    // Get size for ATTR_MEM_EXP_DRAM_ADDRESS_MIRRORING attribute
    case ATTR_MEM_EXP_DRAM_ADDRESS_MIRRORING:
        l_attrSize = sizeof(ATTR_MEM_EXP_DRAM_ADDRESS_MIRRORING_type);
        break;

    // Get size for ATTR_MEM_EXP_RCD_SLEW_RATE attribute
    case ATTR_MEM_EXP_RCD_SLEW_RATE:
        l_attrSize = sizeof(ATTR_MEM_EXP_RCD_SLEW_RATE_type);
        break;

    // Get size for ATTR_MEM_EXP_SPD_CL_SUPPORTED attribute
    case ATTR_MEM_EXP_SPD_CL_SUPPORTED:
        l_attrSize = sizeof(ATTR_MEM_EXP_SPD_CL_SUPPORTED_type);
        break;

    // Get size for ATTR_MEM_EXP_SPD_TAA_MIN attribute
    case ATTR_MEM_EXP_SPD_TAA_MIN:
        l_attrSize = sizeof(ATTR_MEM_EXP_SPD_TAA_MIN_type);
        break;

    // Get size for ATTR_MSS_EXP_REORDER_QUEUE_SETTING attribute
    case ATTR_MSS_EXP_REORDER_QUEUE_SETTING:
        l_attrSize = sizeof(ATTR_MSS_EXP_REORDER_QUEUE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EXP_FIRMWARE_EMULATION_MODE attribute
    case ATTR_MEM_EXP_FIRMWARE_EMULATION_MODE:
        l_attrSize = sizeof(ATTR_MEM_EXP_FIRMWARE_EMULATION_MODE_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_STRUCT_MMIO_ENDIAN_CTRL attribute
    case ATTR_MSS_OCMB_EXP_STRUCT_MMIO_ENDIAN_CTRL:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_STRUCT_MMIO_ENDIAN_CTRL_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_STRUCT_ENDIAN attribute
    case ATTR_MSS_OCMB_EXP_STRUCT_ENDIAN:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_STRUCT_ENDIAN_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_STRUCT_MMIO_WORD_SWAP attribute
    case ATTR_MSS_OCMB_EXP_STRUCT_MMIO_WORD_SWAP:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_STRUCT_MMIO_WORD_SWAP_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_OMI_CFG_ENDIAN_CTRL attribute
    case ATTR_MSS_OCMB_EXP_OMI_CFG_ENDIAN_CTRL:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_OMI_CFG_ENDIAN_CTRL_type);
        break;

    // Get size for ATTR_MSS_OCMB_ECID attribute
    case ATTR_MSS_OCMB_ECID:
        l_attrSize = sizeof(ATTR_MSS_OCMB_ECID_type);
        break;

    // Get size for ATTR_MEM_EXP_DFIMRL_CLK attribute
    case ATTR_MEM_EXP_DFIMRL_CLK:
        l_attrSize = sizeof(ATTR_MEM_EXP_DFIMRL_CLK_type);
        break;

    // Get size for ATTR_MEM_EFF_ATXDLY_A attribute
    case ATTR_MEM_EFF_ATXDLY_A:
        l_attrSize = sizeof(ATTR_MEM_EFF_ATXDLY_A_type);
        break;

    // Get size for ATTR_MEM_EFF_ATXDLY_B attribute
    case ATTR_MEM_EFF_ATXDLY_B:
        l_attrSize = sizeof(ATTR_MEM_EFF_ATXDLY_B_type);
        break;

    // Get size for ATTR_MSS_OCMB_PHY_INIT_MODE attribute
    case ATTR_MSS_OCMB_PHY_INIT_MODE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_PHY_INIT_MODE_type);
        break;

    // Get size for ATTR_MSS_OCMB_CHECKSTOP_OBJ_HANDLE attribute
    case ATTR_MSS_OCMB_CHECKSTOP_OBJ_HANDLE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_CHECKSTOP_OBJ_HANDLE_type);
        break;

    // Get size for ATTR_MSS_OCMB_RECOV_OBJ_HANDLE attribute
    case ATTR_MSS_OCMB_RECOV_OBJ_HANDLE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_RECOV_OBJ_HANDLE_type);
        break;

    // Get size for ATTR_MSS_OCMB_SPECATTN_OBJ_HANDLE attribute
    case ATTR_MSS_OCMB_SPECATTN_OBJ_HANDLE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_SPECATTN_OBJ_HANDLE_type);
        break;

    // Get size for ATTR_MSS_OCMB_APPINTR_OBJ_HANDLE attribute
    case ATTR_MSS_OCMB_APPINTR_OBJ_HANDLE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_APPINTR_OBJ_HANDLE_type);
        break;

    // Get size for ATTR_MSS_CHECK_FOR_READY_TIMEOUT attribute
    case ATTR_MSS_CHECK_FOR_READY_TIMEOUT:
        l_attrSize = sizeof(ATTR_MSS_CHECK_FOR_READY_TIMEOUT_type);
        break;

    // Get size for ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST attribute
    case ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST:
        l_attrSize = sizeof(ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_type);
        break;

    // Get size for ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_FAIL_BEHAVIOR attribute
    case ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_FAIL_BEHAVIOR:
        l_attrSize = sizeof(ATTR_MSS_POST_MEMDIAGS_READ_SUBTEST_FAIL_BEHAVIOR_type);
        break;

    // Get size for ATTR_MSS_OCMB_DISABLE_THERM_INIT_READ attribute
    case ATTR_MSS_OCMB_DISABLE_THERM_INIT_READ:
        l_attrSize = sizeof(ATTR_MSS_OCMB_DISABLE_THERM_INIT_READ_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC00 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC00:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC00_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC01 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC01:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC01_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC03 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC03:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC03_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC04 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC04:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC04_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC05 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC05:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC05_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC0B attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC0B:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC0B_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC0E attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC0E:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC0E_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC0F attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC0F:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC0F_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC1X attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC1X:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC1X_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F0RC7X attribute
    case ATTR_MSS_EXP_RESP_DDR4_F0RC7X:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F0RC7X_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F1RC00 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F1RC00:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F1RC00_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F1RC02 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F1RC02:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F1RC02_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F1RC03 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F1RC03:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F1RC03_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F1RC04 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F1RC04:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F1RC04_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DDR4_F1RC05 attribute
    case ATTR_MSS_EXP_RESP_DDR4_F1RC05:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DDR4_F1RC05_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_RBT attribute
    case ATTR_MSS_EXP_RESP_DRAM_RBT:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_RBT_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_TM attribute
    case ATTR_MSS_EXP_RESP_DRAM_TM:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_TM_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_DLL_RESET attribute
    case ATTR_MSS_EXP_RESP_DRAM_DLL_RESET:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_DLL_RESET_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_BURST_LENGTH attribute
    case ATTR_MSS_EXP_RESP_DRAM_BURST_LENGTH:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_BURST_LENGTH_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_DLL_ENABLE attribute
    case ATTR_MSS_EXP_RESP_DRAM_DLL_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_DLL_ENABLE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_ODIC attribute
    case ATTR_MSS_EXP_RESP_DRAM_ODIC:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_ODIC_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_AL attribute
    case ATTR_MSS_EXP_RESP_DRAM_AL:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_AL_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_WR_LVL_ENABLE attribute
    case ATTR_MSS_EXP_RESP_DRAM_WR_LVL_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_WR_LVL_ENABLE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_RTT_NOM attribute
    case ATTR_MSS_EXP_RESP_DRAM_RTT_NOM:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_RTT_NOM_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_RTT_PARK attribute
    case ATTR_MSS_EXP_RESP_DRAM_RTT_PARK:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_RTT_PARK_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_TDQS attribute
    case ATTR_MSS_EXP_RESP_DRAM_TDQS:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_TDQS_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_OUTPUT_BUFFER attribute
    case ATTR_MSS_EXP_RESP_DRAM_OUTPUT_BUFFER:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_OUTPUT_BUFFER_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_LPASR attribute
    case ATTR_MSS_EXP_RESP_DRAM_LPASR:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_LPASR_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DRAM_RTT_WR attribute
    case ATTR_MSS_EXP_RESP_DRAM_RTT_WR:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DRAM_RTT_WR_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_MPR_PAGE attribute
    case ATTR_MSS_EXP_RESP_MPR_PAGE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_MPR_PAGE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_MPR_MODE attribute
    case ATTR_MSS_EXP_RESP_MPR_MODE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_MPR_MODE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_GEARDOWN_MODE attribute
    case ATTR_MSS_EXP_RESP_GEARDOWN_MODE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_GEARDOWN_MODE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_PER_DRAM_ACCESS attribute
    case ATTR_MSS_EXP_RESP_PER_DRAM_ACCESS:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_PER_DRAM_ACCESS_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_TEMP_READOUT attribute
    case ATTR_MSS_EXP_RESP_TEMP_READOUT:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_TEMP_READOUT_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_CRC_WR_LATENCY attribute
    case ATTR_MSS_EXP_RESP_CRC_WR_LATENCY:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_CRC_WR_LATENCY_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_MPR_RD_FORMAT attribute
    case ATTR_MSS_EXP_RESP_MPR_RD_FORMAT:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_MPR_RD_FORMAT_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_MAX_POWERDOWN_MODE attribute
    case ATTR_MSS_EXP_RESP_MAX_POWERDOWN_MODE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_MAX_POWERDOWN_MODE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_INTERNAL_VREF_MONITOR attribute
    case ATTR_MSS_EXP_RESP_INTERNAL_VREF_MONITOR:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_INTERNAL_VREF_MONITOR_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_SELF_REF_ABORT attribute
    case ATTR_MSS_EXP_RESP_SELF_REF_ABORT:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_SELF_REF_ABORT_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_RD_PREAMBLE_TRAIN attribute
    case ATTR_MSS_EXP_RESP_RD_PREAMBLE_TRAIN:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_RD_PREAMBLE_TRAIN_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_RD_PREAMBLE attribute
    case ATTR_MSS_EXP_RESP_RD_PREAMBLE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_RD_PREAMBLE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_WR_PREAMBLE attribute
    case ATTR_MSS_EXP_RESP_WR_PREAMBLE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_WR_PREAMBLE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_CRC_ERROR_CLEAR attribute
    case ATTR_MSS_EXP_RESP_CRC_ERROR_CLEAR:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_CRC_ERROR_CLEAR_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_CA_PARITY_ERROR_STATUS attribute
    case ATTR_MSS_EXP_RESP_CA_PARITY_ERROR_STATUS:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_CA_PARITY_ERROR_STATUS_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_ODT_INPUT_BUFF attribute
    case ATTR_MSS_EXP_RESP_ODT_INPUT_BUFF:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_ODT_INPUT_BUFF_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_CA_PARITY attribute
    case ATTR_MSS_EXP_RESP_CA_PARITY:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_CA_PARITY_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_DATA_MASK attribute
    case ATTR_MSS_EXP_RESP_DATA_MASK:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_DATA_MASK_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_WRITE_DBI attribute
    case ATTR_MSS_EXP_RESP_WRITE_DBI:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_WRITE_DBI_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_READ_DBI attribute
    case ATTR_MSS_EXP_RESP_READ_DBI:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_READ_DBI_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE attribute
    case ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_VALUE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE attribute
    case ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_RANGE_type);
        break;

    // Get size for ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE attribute
    case ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE:
        l_attrSize = sizeof(ATTR_MSS_EXP_RESP_VREF_DQ_TRAIN_ENABLE_type);
        break;

    // Get size for ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
    case ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_PORT:
        l_attrSize = sizeof(ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_PORT_type);
        break;

    // Get size for ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
    case ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_SLOT:
        l_attrSize = sizeof(ATTR_EXP_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type);
        break;

    // Get size for ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
    case ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT:
        l_attrSize = sizeof(ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_PORT_type);
        break;

    // Get size for ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT attribute
    case ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT:
        l_attrSize = sizeof(ATTR_EXP_RUNTIME_MEM_THROTTLED_N_COMMANDS_PER_SLOT_type);
        break;

    // Get size for ATTR_EXP_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT attribute
    case ATTR_EXP_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT:
        l_attrSize = sizeof(ATTR_EXP_SAFEMODE_MEM_THROTTLED_N_COMMANDS_PER_PORT_type);
        break;

    // Get size for ATTR_EXP_RUNTIME_MEM_M_DRAM_CLOCKS attribute
    case ATTR_EXP_RUNTIME_MEM_M_DRAM_CLOCKS:
        l_attrSize = sizeof(ATTR_EXP_RUNTIME_MEM_M_DRAM_CLOCKS_type);
        break;

    // Get size for ATTR_EXP_MEM_PORT_POS_OF_FAIL_THROTTLE attribute
    case ATTR_EXP_MEM_PORT_POS_OF_FAIL_THROTTLE:
        l_attrSize = sizeof(ATTR_EXP_MEM_PORT_POS_OF_FAIL_THROTTLE_type);
        break;

    // Get size for ATTR_EXP_MEM_WATT_TARGET attribute
    case ATTR_EXP_MEM_WATT_TARGET:
        l_attrSize = sizeof(ATTR_EXP_MEM_WATT_TARGET_type);
        break;

    // Get size for ATTR_EXP_TOTAL_PWR_SLOPE attribute
    case ATTR_EXP_TOTAL_PWR_SLOPE:
        l_attrSize = sizeof(ATTR_EXP_TOTAL_PWR_SLOPE_type);
        break;

    // Get size for ATTR_EXP_TOTAL_PWR_INTERCEPT attribute
    case ATTR_EXP_TOTAL_PWR_INTERCEPT:
        l_attrSize = sizeof(ATTR_EXP_TOTAL_PWR_INTERCEPT_type);
        break;

    // Get size for ATTR_EXP_DATABUS_UTIL attribute
    case ATTR_EXP_DATABUS_UTIL:
        l_attrSize = sizeof(ATTR_EXP_DATABUS_UTIL_type);
        break;

    // Get size for ATTR_EXP_PORT_MAXPOWER attribute
    case ATTR_EXP_PORT_MAXPOWER:
        l_attrSize = sizeof(ATTR_EXP_PORT_MAXPOWER_type);
        break;

    // Get size for ATTR_EXP_DIMM_THERMAL_LIMIT attribute
    case ATTR_EXP_DIMM_THERMAL_LIMIT:
        l_attrSize = sizeof(ATTR_EXP_DIMM_THERMAL_LIMIT_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_FW_MODE attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_FW_MODE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_FW_MODE_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_OPENCAPI_LOOPBACK_TEST attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_OPENCAPI_LOOPBACK_TEST:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_OPENCAPI_LOOPBACK_TEST_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_TRANSPORT_LAYER attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_TRANSPORT_LAYER:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_TRANSPORT_LAYER_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DL_LAYER_BOOT_MODE attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DL_LAYER_BOOT_MODE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DL_LAYER_BOOT_MODE_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DFE_DISABLE attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DFE_DISABLE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_DFE_DISABLE_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_LANE_MODE attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_LANE_MODE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_LANE_MODE_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_SERDES_FREQUENCY attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_SERDES_FREQUENCY:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_SERDES_FREQUENCY_type);
        break;

    // Get size for ATTR_MSS_OCMB_EXP_BOOT_CONFIG_ADAPTATION_MODE attribute
    case ATTR_MSS_OCMB_EXP_BOOT_CONFIG_ADAPTATION_MODE:
        l_attrSize = sizeof(ATTR_MSS_OCMB_EXP_BOOT_CONFIG_ADAPTATION_MODE_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_MFG_ID attribute
    case ATTR_MEM_EFF_PMIC0_MFG_ID:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_MFG_ID_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWA_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWA_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWB_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWB_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWC_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWC_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWD_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SWD_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_MFG_ID attribute
    case ATTR_MEM_EFF_PMIC1_MFG_ID:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_MFG_ID_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWA_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWA_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWB_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWB_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWC_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWC_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_SETTING attribute
    case ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_SETTING:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_SETTING_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_RANGE_SELECT attribute
    case ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_RANGE_SELECT:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_RANGE_SELECT_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWD_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_DELAY attribute
    case ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_DELAY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_DELAY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_ORDER attribute
    case ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_ORDER:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SWD_SEQUENCE_ORDER_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_REDUNDANCY attribute
    case ATTR_MEM_EFF_PMIC0_REDUNDANCY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_REDUNDANCY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_PHASE_COMB attribute
    case ATTR_MEM_EFF_PMIC0_PHASE_COMB:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_PHASE_COMB_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_REDUNDANCY attribute
    case ATTR_MEM_EFF_PMIC1_REDUNDANCY:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_REDUNDANCY_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_REVISION attribute
    case ATTR_MEM_EFF_PMIC0_REVISION:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_REVISION_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_REVISION attribute
    case ATTR_MEM_EFF_PMIC1_REVISION:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_REVISION_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_PHASE_COMB attribute
    case ATTR_MEM_EFF_PMIC1_PHASE_COMB:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_PHASE_COMB_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC0_SEQUENCE attribute
    case ATTR_MEM_EFF_PMIC0_SEQUENCE:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC0_SEQUENCE_type);
        break;

    // Get size for ATTR_MEM_EFF_PMIC1_SEQUENCE attribute
    case ATTR_MEM_EFF_PMIC1_SEQUENCE:
        l_attrSize = sizeof(ATTR_MEM_EFF_PMIC1_SEQUENCE_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC0_SWA_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC0_SWA_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC0_SWA_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC0_SWB_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC0_SWB_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC0_SWB_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC0_SWC_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC0_SWC_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC0_SWC_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC0_SWD_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC0_SWD_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC0_SWD_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC1_SWA_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC1_SWA_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC1_SWA_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC1_SWB_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC1_SWB_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC1_SWB_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC1_SWC_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC1_SWC_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC1_SWC_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_MEM_EFF_EFD_PMIC1_SWD_VOLTAGE_OFFSET attribute
    case ATTR_MEM_EFF_EFD_PMIC1_SWD_VOLTAGE_OFFSET:
        l_attrSize = sizeof(ATTR_MEM_EFF_EFD_PMIC1_SWD_VOLTAGE_OFFSET_type);
        break;

    // Get size for ATTR_AFFINITY_PATH attribute
    case ATTR_AFFINITY_PATH:
        l_attrSize = sizeof(ATTR_AFFINITY_PATH_type);
        break;

    // Get size for ATTR_ALL_MCS_IN_INTERLEAVING_GROUP attribute
    case ATTR_ALL_MCS_IN_INTERLEAVING_GROUP:
        l_attrSize = sizeof(ATTR_ALL_MCS_IN_INTERLEAVING_GROUP_type);
        break;

    // Get size for ATTR_ALTFSI_MASTER_CHIP attribute
    case ATTR_ALTFSI_MASTER_CHIP:
        l_attrSize = sizeof(ATTR_ALTFSI_MASTER_CHIP_type);
        break;

    // Get size for ATTR_ALTFSI_MASTER_PORT attribute
    case ATTR_ALTFSI_MASTER_PORT:
        l_attrSize = sizeof(ATTR_ALTFSI_MASTER_PORT_type);
        break;

    // Get size for ATTR_ASYNC_NEST_FREQ_MHZ attribute
    case ATTR_ASYNC_NEST_FREQ_MHZ:
        l_attrSize = sizeof(ATTR_ASYNC_NEST_FREQ_MHZ_type);
        break;

    // Get size for ATTR_ATTN_AREA_1_ADDR attribute
    case ATTR_ATTN_AREA_1_ADDR:
        l_attrSize = sizeof(ATTR_ATTN_AREA_1_ADDR_type);
        break;

    // Get size for ATTR_ATTN_AREA_2_ADDR attribute
    case ATTR_ATTN_AREA_2_ADDR:
        l_attrSize = sizeof(ATTR_ATTN_AREA_2_ADDR_type);
        break;

    // Get size for ATTR_AVDD_ID attribute
    case ATTR_AVDD_ID:
        l_attrSize = sizeof(ATTR_AVDD_ID_type);
        break;

    // Get size for ATTR_BLOCK_SPEC_DECONFIG attribute
    case ATTR_BLOCK_SPEC_DECONFIG:
        l_attrSize = sizeof(ATTR_BLOCK_SPEC_DECONFIG_type);
        break;

    // Get size for ATTR_BMC_FRU_ID attribute
    case ATTR_BMC_FRU_ID:
        l_attrSize = sizeof(ATTR_BMC_FRU_ID_type);
        break;

    // Get size for ATTR_BPM_UPDATE_OVERRIDE attribute
    case ATTR_BPM_UPDATE_OVERRIDE:
        l_attrSize = sizeof(ATTR_BPM_UPDATE_OVERRIDE_type);
        break;

    // Get size for ATTR_BRAZOS_RX_FIFO_OVERRIDE attribute
    case ATTR_BRAZOS_RX_FIFO_OVERRIDE:
        l_attrSize = sizeof(ATTR_BRAZOS_RX_FIFO_OVERRIDE_type);
        break;

    // Get size for ATTR_CDM_DOMAIN attribute
    case ATTR_CDM_DOMAIN:
        l_attrSize = sizeof(ATTR_CDM_DOMAIN_type);
        break;

    // Get size for ATTR_CDM_POLICIES attribute
    case ATTR_CDM_POLICIES:
        l_attrSize = sizeof(ATTR_CDM_POLICIES_type);
        break;

    // Get size for ATTR_CEC_IPL_TYPE attribute
    case ATTR_CEC_IPL_TYPE:
        l_attrSize = sizeof(ATTR_CEC_IPL_TYPE_type);
        break;

    // Get size for ATTR_CENTAUR_ECID_FRU_ID attribute
    case ATTR_CENTAUR_ECID_FRU_ID:
        l_attrSize = sizeof(ATTR_CENTAUR_ECID_FRU_ID_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_AVDD_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_AVDD_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_AVDD_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VCS_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VCS_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VCS_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VCS_OFFSET_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VCS_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VCS_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VDDR_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VDDR_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VDDR_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VDD_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VDD_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VDD_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VDD_OFFSET_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VDD_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VDD_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VPP_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VPP_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VPP_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CEN_MSS_VOLT_VPP_OFFSET_MILLIVOLTS attribute
    case ATTR_CEN_MSS_VOLT_VPP_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_CEN_MSS_VOLT_VPP_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_CHIPLET_ID attribute
    case ATTR_CHIPLET_ID:
        l_attrSize = sizeof(ATTR_CHIPLET_ID_type);
        break;

    // Get size for ATTR_CHIP_ID attribute
    case ATTR_CHIP_ID:
        l_attrSize = sizeof(ATTR_CHIP_ID_type);
        break;

    // Get size for ATTR_CHIP_UNIT attribute
    case ATTR_CHIP_UNIT:
        l_attrSize = sizeof(ATTR_CHIP_UNIT_type);
        break;

    // Get size for ATTR_CLASS attribute
    case ATTR_CLASS:
        l_attrSize = sizeof(ATTR_CLASS_type);
        break;

    // Get size for ATTR_CPU_ATTR attribute
    case ATTR_CPU_ATTR:
        l_attrSize = sizeof(ATTR_CPU_ATTR_type);
        break;

    // Get size for ATTR_CVPD_MAX_SECTIONS attribute
    case ATTR_CVPD_MAX_SECTIONS:
        l_attrSize = sizeof(ATTR_CVPD_MAX_SECTIONS_type);
        break;

    // Get size for ATTR_CVPD_SIZE attribute
    case ATTR_CVPD_SIZE:
        l_attrSize = sizeof(ATTR_CVPD_SIZE_type);
        break;

    // Get size for ATTR_DATA_CACHE_LINE_SIZE attribute
    case ATTR_DATA_CACHE_LINE_SIZE:
        l_attrSize = sizeof(ATTR_DATA_CACHE_LINE_SIZE_type);
        break;

    // Get size for ATTR_DATA_CACHE_SIZE attribute
    case ATTR_DATA_CACHE_SIZE:
        l_attrSize = sizeof(ATTR_DATA_CACHE_SIZE_type);
        break;

    // Get size for ATTR_DCACHE_ASSOC_SETS attribute
    case ATTR_DCACHE_ASSOC_SETS:
        l_attrSize = sizeof(ATTR_DCACHE_ASSOC_SETS_type);
        break;

    // Get size for ATTR_DCACHE_LINE_SIZE attribute
    case ATTR_DCACHE_LINE_SIZE:
        l_attrSize = sizeof(ATTR_DCACHE_LINE_SIZE_type);
        break;

    // Get size for ATTR_DECONFIG_GARDABLE attribute
    case ATTR_DECONFIG_GARDABLE:
        l_attrSize = sizeof(ATTR_DECONFIG_GARDABLE_type);
        break;

    // Get size for ATTR_DISABLE_I2C_ENGINE2_PORT0_DIAG_MODE attribute
    case ATTR_DISABLE_I2C_ENGINE2_PORT0_DIAG_MODE:
        l_attrSize = sizeof(ATTR_DISABLE_I2C_ENGINE2_PORT0_DIAG_MODE_type);
        break;

    // Get size for ATTR_DO_ABUS_DECONFIG attribute
    case ATTR_DO_ABUS_DECONFIG:
        l_attrSize = sizeof(ATTR_DO_ABUS_DECONFIG_type);
        break;

    // Get size for ATTR_DPO_MIN_FREQ_PERCENT attribute
    case ATTR_DPO_MIN_FREQ_PERCENT:
        l_attrSize = sizeof(ATTR_DPO_MIN_FREQ_PERCENT_type);
        break;

    // Get size for ATTR_DUMMY_HEAP_ZERO_DEFAULT attribute
    case ATTR_DUMMY_HEAP_ZERO_DEFAULT:
        l_attrSize = sizeof(ATTR_DUMMY_HEAP_ZERO_DEFAULT_type);
        break;

    // Get size for ATTR_DUMMY_RO attribute
    case ATTR_DUMMY_RO:
        l_attrSize = sizeof(ATTR_DUMMY_RO_type);
        break;

    // Get size for ATTR_DUMMY_RW attribute
    case ATTR_DUMMY_RW:
        l_attrSize = sizeof(ATTR_DUMMY_RW_type);
        break;

    // Get size for ATTR_DUMMY_WO attribute
    case ATTR_DUMMY_WO:
        l_attrSize = sizeof(ATTR_DUMMY_WO_type);
        break;

    // Get size for ATTR_EC attribute
    case ATTR_EC:
        l_attrSize = sizeof(ATTR_EC_type);
        break;

    // Get size for ATTR_EEPROM_CONTENT_TYPE attribute
    case ATTR_EEPROM_CONTENT_TYPE:
        l_attrSize = sizeof(ATTR_EEPROM_CONTENT_TYPE_type);
        break;

    // Get size for ATTR_EEPROM_NV_INFO attribute
    case ATTR_EEPROM_NV_INFO:
        l_attrSize = sizeof(ATTR_EEPROM_NV_INFO_type);
        break;

    // Get size for ATTR_EEPROM_SBE_BACKUP_INFO attribute
    case ATTR_EEPROM_SBE_BACKUP_INFO:
        l_attrSize = sizeof(ATTR_EEPROM_SBE_BACKUP_INFO_type);
        break;

    // Get size for ATTR_EEPROM_SBE_PRIMARY_INFO attribute
    case ATTR_EEPROM_SBE_PRIMARY_INFO:
        l_attrSize = sizeof(ATTR_EEPROM_SBE_PRIMARY_INFO_type);
        break;

    // Get size for ATTR_EEPROM_VPD_BACKUP_INFO attribute
    case ATTR_EEPROM_VPD_BACKUP_INFO:
        l_attrSize = sizeof(ATTR_EEPROM_VPD_BACKUP_INFO_type);
        break;

    // Get size for ATTR_EEPROM_VPD_PRIMARY_INFO attribute
    case ATTR_EEPROM_VPD_PRIMARY_INFO:
        l_attrSize = sizeof(ATTR_EEPROM_VPD_PRIMARY_INFO_type);
        break;

    // Get size for ATTR_EFFECTIVE_EC attribute
    case ATTR_EFFECTIVE_EC:
        l_attrSize = sizeof(ATTR_EFFECTIVE_EC_type);
        break;

    // Get size for ATTR_ENABLED_THREADS attribute
    case ATTR_ENABLED_THREADS:
        l_attrSize = sizeof(ATTR_ENABLED_THREADS_type);
        break;

    // Get size for ATTR_ENGINE_TYPE attribute
    case ATTR_ENGINE_TYPE:
        l_attrSize = sizeof(ATTR_ENGINE_TYPE_type);
        break;

    // Get size for ATTR_FABRIC_CHIP_ID attribute
    case ATTR_FABRIC_CHIP_ID:
        l_attrSize = sizeof(ATTR_FABRIC_CHIP_ID_type);
        break;

    // Get size for ATTR_FABRIC_GROUP_ID attribute
    case ATTR_FABRIC_GROUP_ID:
        l_attrSize = sizeof(ATTR_FABRIC_GROUP_ID_type);
        break;

    // Get size for ATTR_FABRIC_TO_PHYSICAL_NODE_MAP attribute
    case ATTR_FABRIC_TO_PHYSICAL_NODE_MAP:
        l_attrSize = sizeof(ATTR_FABRIC_TO_PHYSICAL_NODE_MAP_type);
        break;

    // Get size for ATTR_FAPI_I2C_CONTROL_INFO attribute
    case ATTR_FAPI_I2C_CONTROL_INFO:
        l_attrSize = sizeof(ATTR_FAPI_I2C_CONTROL_INFO_type);
        break;

    // Get size for ATTR_FAPI_NAME attribute
    case ATTR_FAPI_NAME:
        l_attrSize = sizeof(ATTR_FAPI_NAME_type);
        break;

    // Get size for ATTR_FIELD_CORE_OVERRIDE attribute
    case ATTR_FIELD_CORE_OVERRIDE:
        l_attrSize = sizeof(ATTR_FIELD_CORE_OVERRIDE_type);
        break;

    // Get size for ATTR_FIELD_TH_P8EX_L2_COL_REPAIRS attribute
    case ATTR_FIELD_TH_P8EX_L2_COL_REPAIRS:
        l_attrSize = sizeof(ATTR_FIELD_TH_P8EX_L2_COL_REPAIRS_type);
        break;

    // Get size for ATTR_FIELD_TH_P8EX_L2_LINE_DELETES attribute
    case ATTR_FIELD_TH_P8EX_L2_LINE_DELETES:
        l_attrSize = sizeof(ATTR_FIELD_TH_P8EX_L2_LINE_DELETES_type);
        break;

    // Get size for ATTR_FIELD_TH_P8EX_L3_COL_REPAIRS attribute
    case ATTR_FIELD_TH_P8EX_L3_COL_REPAIRS:
        l_attrSize = sizeof(ATTR_FIELD_TH_P8EX_L3_COL_REPAIRS_type);
        break;

    // Get size for ATTR_FIELD_TH_P8EX_L3_LINE_DELETES attribute
    case ATTR_FIELD_TH_P8EX_L3_LINE_DELETES:
        l_attrSize = sizeof(ATTR_FIELD_TH_P8EX_L3_LINE_DELETES_type);
        break;

    // Get size for ATTR_FORCE_NVDIMM_RESET attribute
    case ATTR_FORCE_NVDIMM_RESET:
        l_attrSize = sizeof(ATTR_FORCE_NVDIMM_RESET_type);
        break;

    // Get size for ATTR_FORCE_SRAM_MMIO_OVER_I2C attribute
    case ATTR_FORCE_SRAM_MMIO_OVER_I2C:
        l_attrSize = sizeof(ATTR_FORCE_SRAM_MMIO_OVER_I2C_type);
        break;

    // Get size for ATTR_FREQ_CORE_CEILING_MHZ attribute
    case ATTR_FREQ_CORE_CEILING_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_CORE_CEILING_MHZ_type);
        break;

    // Get size for ATTR_FREQ_CORE_MAX attribute
    case ATTR_FREQ_CORE_MAX:
        l_attrSize = sizeof(ATTR_FREQ_CORE_MAX_type);
        break;

    // Get size for ATTR_FREQ_PROC_REFCLOCK attribute
    case ATTR_FREQ_PROC_REFCLOCK:
        l_attrSize = sizeof(ATTR_FREQ_PROC_REFCLOCK_type);
        break;

    // Get size for ATTR_FRU_ID attribute
    case ATTR_FRU_ID:
        l_attrSize = sizeof(ATTR_FRU_ID_type);
        break;

    // Get size for ATTR_FSI_MASTER_CHIP attribute
    case ATTR_FSI_MASTER_CHIP:
        l_attrSize = sizeof(ATTR_FSI_MASTER_CHIP_type);
        break;

    // Get size for ATTR_FSI_MASTER_PORT attribute
    case ATTR_FSI_MASTER_PORT:
        l_attrSize = sizeof(ATTR_FSI_MASTER_PORT_type);
        break;

    // Get size for ATTR_FSI_MASTER_TYPE attribute
    case ATTR_FSI_MASTER_TYPE:
        l_attrSize = sizeof(ATTR_FSI_MASTER_TYPE_type);
        break;

    // Get size for ATTR_FSI_OPTION_FLAGS attribute
    case ATTR_FSI_OPTION_FLAGS:
        l_attrSize = sizeof(ATTR_FSI_OPTION_FLAGS_type);
        break;

    // Get size for ATTR_FSI_SLAVE_CASCADE attribute
    case ATTR_FSI_SLAVE_CASCADE:
        l_attrSize = sizeof(ATTR_FSI_SLAVE_CASCADE_type);
        break;

    // Get size for ATTR_FSP_BAR_SIZE attribute
    case ATTR_FSP_BAR_SIZE:
        l_attrSize = sizeof(ATTR_FSP_BAR_SIZE_type);
        break;

    // Get size for ATTR_FSP_BASE_ADDR attribute
    case ATTR_FSP_BASE_ADDR:
        l_attrSize = sizeof(ATTR_FSP_BASE_ADDR_type);
        break;

    // Get size for ATTR_FUSED_CORE_MODE_HB attribute
    case ATTR_FUSED_CORE_MODE_HB:
        l_attrSize = sizeof(ATTR_FUSED_CORE_MODE_HB_type);
        break;

    // Get size for ATTR_FUSED_CORE_OPTION attribute
    case ATTR_FUSED_CORE_OPTION:
        l_attrSize = sizeof(ATTR_FUSED_CORE_OPTION_type);
        break;

    // Get size for ATTR_HB_HRMOR_NODAL_BASE attribute
    case ATTR_HB_HRMOR_NODAL_BASE:
        l_attrSize = sizeof(ATTR_HB_HRMOR_NODAL_BASE_type);
        break;

    // Get size for ATTR_HB_RSV_MEM_SIZE_MB attribute
    case ATTR_HB_RSV_MEM_SIZE_MB:
        l_attrSize = sizeof(ATTR_HB_RSV_MEM_SIZE_MB_type);
        break;

    // Get size for ATTR_HB_SETTINGS attribute
    case ATTR_HB_SETTINGS:
        l_attrSize = sizeof(ATTR_HB_SETTINGS_type);
        break;

    // Get size for ATTR_HDAT_EC attribute
    case ATTR_HDAT_EC:
        l_attrSize = sizeof(ATTR_HDAT_EC_type);
        break;

    // Get size for ATTR_HDAT_HBRT_NUM_SECTIONS attribute
    case ATTR_HDAT_HBRT_NUM_SECTIONS:
        l_attrSize = sizeof(ATTR_HDAT_HBRT_NUM_SECTIONS_type);
        break;

    // Get size for ATTR_HDAT_HBRT_SECTION_SIZE attribute
    case ATTR_HDAT_HBRT_SECTION_SIZE:
        l_attrSize = sizeof(ATTR_HDAT_HBRT_SECTION_SIZE_type);
        break;

    // Get size for ATTR_HDAT_RSV_MEM_NUM_SECTIONS attribute
    case ATTR_HDAT_RSV_MEM_NUM_SECTIONS:
        l_attrSize = sizeof(ATTR_HDAT_RSV_MEM_NUM_SECTIONS_type);
        break;

    // Get size for ATTR_HDDW_ORDER attribute
    case ATTR_HDDW_ORDER:
        l_attrSize = sizeof(ATTR_HDDW_ORDER_type);
        break;

    // Get size for ATTR_HOMER_PHYS_ADDR attribute
    case ATTR_HOMER_PHYS_ADDR:
        l_attrSize = sizeof(ATTR_HOMER_PHYS_ADDR_type);
        break;

    // Get size for ATTR_HOSTSVC_PLID attribute
    case ATTR_HOSTSVC_PLID:
        l_attrSize = sizeof(ATTR_HOSTSVC_PLID_type);
        break;

    // Get size for ATTR_HOT_PLUG_POWER_CONTROLLER_INFO attribute
    case ATTR_HOT_PLUG_POWER_CONTROLLER_INFO:
        l_attrSize = sizeof(ATTR_HOT_PLUG_POWER_CONTROLLER_INFO_type);
        break;

    // Get size for ATTR_HUID attribute
    case ATTR_HUID:
        l_attrSize = sizeof(ATTR_HUID_type);
        break;

    // Get size for ATTR_HWAS_STATE attribute
    case ATTR_HWAS_STATE:
        l_attrSize = sizeof(ATTR_HWAS_STATE_type);
        break;

    // Get size for ATTR_HWAS_STATE_CHANGED_FLAG attribute
    case ATTR_HWAS_STATE_CHANGED_FLAG:
        l_attrSize = sizeof(ATTR_HWAS_STATE_CHANGED_FLAG_type);
        break;

    // Get size for ATTR_HWAS_STATE_CHANGED_SUBSCRIPTION_MASK attribute
    case ATTR_HWAS_STATE_CHANGED_SUBSCRIPTION_MASK:
        l_attrSize = sizeof(ATTR_HWAS_STATE_CHANGED_SUBSCRIPTION_MASK_type);
        break;

    // Get size for ATTR_I2C_BUS_SPEED_ARRAY attribute
    case ATTR_I2C_BUS_SPEED_ARRAY:
        l_attrSize = sizeof(ATTR_I2C_BUS_SPEED_ARRAY_type);
        break;

    // Get size for ATTR_I2C_CONTROL_INFO attribute
    case ATTR_I2C_CONTROL_INFO:
        l_attrSize = sizeof(ATTR_I2C_CONTROL_INFO_type);
        break;

    // Get size for ATTR_I2C_SWITCHES attribute
    case ATTR_I2C_SWITCHES:
        l_attrSize = sizeof(ATTR_I2C_SWITCHES_type);
        break;

    // Get size for ATTR_IBSCOM_MCS_BASE_ADDR attribute
    case ATTR_IBSCOM_MCS_BASE_ADDR:
        l_attrSize = sizeof(ATTR_IBSCOM_MCS_BASE_ADDR_type);
        break;

    // Get size for ATTR_IBSCOM_PROC_BASE_ADDR attribute
    case ATTR_IBSCOM_PROC_BASE_ADDR:
        l_attrSize = sizeof(ATTR_IBSCOM_PROC_BASE_ADDR_type);
        break;

    // Get size for ATTR_ICACHE_ASSOC_SETS attribute
    case ATTR_ICACHE_ASSOC_SETS:
        l_attrSize = sizeof(ATTR_ICACHE_ASSOC_SETS_type);
        break;

    // Get size for ATTR_ICACHE_BLOCK_SIZE attribute
    case ATTR_ICACHE_BLOCK_SIZE:
        l_attrSize = sizeof(ATTR_ICACHE_BLOCK_SIZE_type);
        break;

    // Get size for ATTR_ICACHE_LINE_SIZE attribute
    case ATTR_ICACHE_LINE_SIZE:
        l_attrSize = sizeof(ATTR_ICACHE_LINE_SIZE_type);
        break;

    // Get size for ATTR_ICACHE_SIZE attribute
    case ATTR_ICACHE_SIZE:
        l_attrSize = sizeof(ATTR_ICACHE_SIZE_type);
        break;

    // Get size for ATTR_IMT_BAR_SIZE attribute
    case ATTR_IMT_BAR_SIZE:
        l_attrSize = sizeof(ATTR_IMT_BAR_SIZE_type);
        break;

    // Get size for ATTR_IMT_BASE_ADDR attribute
    case ATTR_IMT_BASE_ADDR:
        l_attrSize = sizeof(ATTR_IMT_BASE_ADDR_type);
        break;

    // Get size for ATTR_INIT_TO_AVAILABLE attribute
    case ATTR_INIT_TO_AVAILABLE:
        l_attrSize = sizeof(ATTR_INIT_TO_AVAILABLE_type);
        break;

    // Get size for ATTR_INTP_BASE_ADDR attribute
    case ATTR_INTP_BASE_ADDR:
        l_attrSize = sizeof(ATTR_INTP_BASE_ADDR_type);
        break;

    // Get size for ATTR_ISTEP_MODE attribute
    case ATTR_ISTEP_MODE:
        l_attrSize = sizeof(ATTR_ISTEP_MODE_type);
        break;

    // Get size for ATTR_ISTEP_PAUSE_CONFIG attribute
    case ATTR_ISTEP_PAUSE_CONFIG:
        l_attrSize = sizeof(ATTR_ISTEP_PAUSE_CONFIG_type);
        break;

    // Get size for ATTR_ISTEP_PAUSE_ENABLE attribute
    case ATTR_ISTEP_PAUSE_ENABLE:
        l_attrSize = sizeof(ATTR_ISTEP_PAUSE_ENABLE_type);
        break;

    // Get size for ATTR_IS_INTER_ENCLOSURE_BUS attribute
    case ATTR_IS_INTER_ENCLOSURE_BUS:
        l_attrSize = sizeof(ATTR_IS_INTER_ENCLOSURE_BUS_type);
        break;

    // Get size for ATTR_KEY_CLEAR_REQUEST attribute
    case ATTR_KEY_CLEAR_REQUEST:
        l_attrSize = sizeof(ATTR_KEY_CLEAR_REQUEST_type);
        break;

    // Get size for ATTR_KEY_TRANSITION_STATE attribute
    case ATTR_KEY_TRANSITION_STATE:
        l_attrSize = sizeof(ATTR_KEY_TRANSITION_STATE_type);
        break;

    // Get size for ATTR_L2_CACHE_ASSOC_SETS attribute
    case ATTR_L2_CACHE_ASSOC_SETS:
        l_attrSize = sizeof(ATTR_L2_CACHE_ASSOC_SETS_type);
        break;

    // Get size for ATTR_L2_CACHE_LINE_SIZE attribute
    case ATTR_L2_CACHE_LINE_SIZE:
        l_attrSize = sizeof(ATTR_L2_CACHE_LINE_SIZE_type);
        break;

    // Get size for ATTR_L2_CACHE_SIZE attribute
    case ATTR_L2_CACHE_SIZE:
        l_attrSize = sizeof(ATTR_L2_CACHE_SIZE_type);
        break;

    // Get size for ATTR_L3_CACHE_LINE_SIZE attribute
    case ATTR_L3_CACHE_LINE_SIZE:
        l_attrSize = sizeof(ATTR_L3_CACHE_LINE_SIZE_type);
        break;

    // Get size for ATTR_L3_CACHE_SIZE attribute
    case ATTR_L3_CACHE_SIZE:
        l_attrSize = sizeof(ATTR_L3_CACHE_SIZE_type);
        break;

    // Get size for ATTR_LOCATION_CODE attribute
    case ATTR_LOCATION_CODE:
        l_attrSize = sizeof(ATTR_LOCATION_CODE_type);
        break;

    // Get size for ATTR_LPC_BUS_ADDR attribute
    case ATTR_LPC_BUS_ADDR:
        l_attrSize = sizeof(ATTR_LPC_BUS_ADDR_type);
        break;

    // Get size for ATTR_MAX_CHIPLETS_PER_PROC attribute
    case ATTR_MAX_CHIPLETS_PER_PROC:
        l_attrSize = sizeof(ATTR_MAX_CHIPLETS_PER_PROC_type);
        break;

    // Get size for ATTR_MAX_COMPUTE_NODES_PER_SYSTEM attribute
    case ATTR_MAX_COMPUTE_NODES_PER_SYSTEM:
        l_attrSize = sizeof(ATTR_MAX_COMPUTE_NODES_PER_SYSTEM_type);
        break;

    // Get size for ATTR_MAX_DIMMS_PER_MBA_PORT attribute
    case ATTR_MAX_DIMMS_PER_MBA_PORT:
        l_attrSize = sizeof(ATTR_MAX_DIMMS_PER_MBA_PORT_type);
        break;

    // Get size for ATTR_MAX_DMI_PER_PROC attribute
    case ATTR_MAX_DMI_PER_PROC:
        l_attrSize = sizeof(ATTR_MAX_DMI_PER_PROC_type);
        break;

    // Get size for ATTR_MAX_EXS_PER_PROC_CHIP attribute
    case ATTR_MAX_EXS_PER_PROC_CHIP:
        l_attrSize = sizeof(ATTR_MAX_EXS_PER_PROC_CHIP_type);
        break;

    // Get size for ATTR_MAX_MBAS_PER_MEMBUF_CHIP attribute
    case ATTR_MAX_MBAS_PER_MEMBUF_CHIP:
        l_attrSize = sizeof(ATTR_MAX_MBAS_PER_MEMBUF_CHIP_type);
        break;

    // Get size for ATTR_MAX_MBA_PORTS_PER_MBA attribute
    case ATTR_MAX_MBA_PORTS_PER_MBA:
        l_attrSize = sizeof(ATTR_MAX_MBA_PORTS_PER_MBA_type);
        break;

    // Get size for ATTR_MAX_MCS_PER_SYSTEM attribute
    case ATTR_MAX_MCS_PER_SYSTEM:
        l_attrSize = sizeof(ATTR_MAX_MCS_PER_SYSTEM_type);
        break;

    // Get size for ATTR_MAX_POWER attribute
    case ATTR_MAX_POWER:
        l_attrSize = sizeof(ATTR_MAX_POWER_type);
        break;

    // Get size for ATTR_MAX_PROC_CHIPS_PER_NODE attribute
    case ATTR_MAX_PROC_CHIPS_PER_NODE:
        l_attrSize = sizeof(ATTR_MAX_PROC_CHIPS_PER_NODE_type);
        break;

    // Get size for ATTR_MAX_SBE_SEEPROM_SIZE attribute
    case ATTR_MAX_SBE_SEEPROM_SIZE:
        l_attrSize = sizeof(ATTR_MAX_SBE_SEEPROM_SIZE_type);
        break;

    // Get size for ATTR_MBA_DIMM attribute
    case ATTR_MBA_DIMM:
        l_attrSize = sizeof(ATTR_MBA_DIMM_type);
        break;

    // Get size for ATTR_MBA_PORT attribute
    case ATTR_MBA_PORT:
        l_attrSize = sizeof(ATTR_MBA_PORT_type);
        break;

    // Get size for ATTR_MC_PLL_BUCKET attribute
    case ATTR_MC_PLL_BUCKET:
        l_attrSize = sizeof(ATTR_MC_PLL_BUCKET_type);
        break;

    // Get size for ATTR_MEM_AVDD_OFFSET_MILLIVOLTS attribute
    case ATTR_MEM_AVDD_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MEM_AVDD_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MEM_BASE attribute
    case ATTR_MEM_BASE:
        l_attrSize = sizeof(ATTR_MEM_BASE_type);
        break;

    // Get size for ATTR_MEM_PORT attribute
    case ATTR_MEM_PORT:
        l_attrSize = sizeof(ATTR_MEM_PORT_type);
        break;

    // Get size for ATTR_MEM_VCS_OFFSET_MILLIVOLTS attribute
    case ATTR_MEM_VCS_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MEM_VCS_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MEM_VDDR_OFFSET_MILLIVOLTS attribute
    case ATTR_MEM_VDDR_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MEM_VDDR_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MEM_VDD_OFFSET_MILLIVOLTS attribute
    case ATTR_MEM_VDD_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MEM_VDD_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MEM_VPP_OFFSET_MILLIVOLTS attribute
    case ATTR_MEM_VPP_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MEM_VPP_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MFG_TRACE_ENABLE attribute
    case ATTR_MFG_TRACE_ENABLE:
        l_attrSize = sizeof(ATTR_MFG_TRACE_ENABLE_type);
        break;

    // Get size for ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET attribute
    case ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET:
        l_attrSize = sizeof(ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_type);
        break;

    // Get size for ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_ENABLE attribute
    case ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_ENABLE:
        l_attrSize = sizeof(ATTR_MFG_WRAP_TEST_ABUS_LINKS_SET_ENABLE_type);
        break;

    // Get size for ATTR_MGC_LOAD_SOURCE attribute
    case ATTR_MGC_LOAD_SOURCE:
        l_attrSize = sizeof(ATTR_MGC_LOAD_SOURCE_type);
        break;

    // Get size for ATTR_MIN_FREQ_MHZ attribute
    case ATTR_MIN_FREQ_MHZ:
        l_attrSize = sizeof(ATTR_MIN_FREQ_MHZ_type);
        break;

    // Get size for ATTR_MIRROR_BASE_ADDRESS attribute
    case ATTR_MIRROR_BASE_ADDRESS:
        l_attrSize = sizeof(ATTR_MIRROR_BASE_ADDRESS_type);
        break;

    // Get size for ATTR_MNFG_ABUS_MIN_EYE_HEIGHT attribute
    case ATTR_MNFG_ABUS_MIN_EYE_HEIGHT:
        l_attrSize = sizeof(ATTR_MNFG_ABUS_MIN_EYE_HEIGHT_type);
        break;

    // Get size for ATTR_MNFG_ABUS_MIN_EYE_WIDTH attribute
    case ATTR_MNFG_ABUS_MIN_EYE_WIDTH:
        l_attrSize = sizeof(ATTR_MNFG_ABUS_MIN_EYE_WIDTH_type);
        break;

    // Get size for ATTR_MNFG_DMI_MIN_EYE_HEIGHT attribute
    case ATTR_MNFG_DMI_MIN_EYE_HEIGHT:
        l_attrSize = sizeof(ATTR_MNFG_DMI_MIN_EYE_HEIGHT_type);
        break;

    // Get size for ATTR_MNFG_DMI_MIN_EYE_WIDTH attribute
    case ATTR_MNFG_DMI_MIN_EYE_WIDTH:
        l_attrSize = sizeof(ATTR_MNFG_DMI_MIN_EYE_WIDTH_type);
        break;

    // Get size for ATTR_MNFG_TH_CEN_L4_CACHE_CES attribute
    case ATTR_MNFG_TH_CEN_L4_CACHE_CES:
        l_attrSize = sizeof(ATTR_MNFG_TH_CEN_L4_CACHE_CES_type);
        break;

    // Get size for ATTR_MNFG_TH_CEN_MBA_IPL_SOFT_CE_TH_ALGO attribute
    case ATTR_MNFG_TH_CEN_MBA_IPL_SOFT_CE_TH_ALGO:
        l_attrSize = sizeof(ATTR_MNFG_TH_CEN_MBA_IPL_SOFT_CE_TH_ALGO_type);
        break;

    // Get size for ATTR_MNFG_TH_CEN_MBA_RT_RCE_PER_RANK attribute
    case ATTR_MNFG_TH_CEN_MBA_RT_RCE_PER_RANK:
        l_attrSize = sizeof(ATTR_MNFG_TH_CEN_MBA_RT_RCE_PER_RANK_type);
        break;

    // Get size for ATTR_MNFG_TH_CEN_MBA_RT_SOFT_CE_TH_ALGO attribute
    case ATTR_MNFG_TH_CEN_MBA_RT_SOFT_CE_TH_ALGO:
        l_attrSize = sizeof(ATTR_MNFG_TH_CEN_MBA_RT_SOFT_CE_TH_ALGO_type);
        break;

    // Get size for ATTR_MNFG_TH_MEMORY_IMPES attribute
    case ATTR_MNFG_TH_MEMORY_IMPES:
        l_attrSize = sizeof(ATTR_MNFG_TH_MEMORY_IMPES_type);
        break;

    // Get size for ATTR_MNFG_TH_MEMORY_IUES attribute
    case ATTR_MNFG_TH_MEMORY_IUES:
        l_attrSize = sizeof(ATTR_MNFG_TH_MEMORY_IUES_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L2_CACHE_CES attribute
    case ATTR_MNFG_TH_P8EX_L2_CACHE_CES:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L2_CACHE_CES_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L2_COL_REPAIRS attribute
    case ATTR_MNFG_TH_P8EX_L2_COL_REPAIRS:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L2_COL_REPAIRS_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L2_DIR_CES attribute
    case ATTR_MNFG_TH_P8EX_L2_DIR_CES:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L2_DIR_CES_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L2_LINE_DELETES attribute
    case ATTR_MNFG_TH_P8EX_L2_LINE_DELETES:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L2_LINE_DELETES_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L3_CACHE_CES attribute
    case ATTR_MNFG_TH_P8EX_L3_CACHE_CES:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L3_CACHE_CES_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L3_COL_REPAIRS attribute
    case ATTR_MNFG_TH_P8EX_L3_COL_REPAIRS:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L3_COL_REPAIRS_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L3_DIR_CES attribute
    case ATTR_MNFG_TH_P8EX_L3_DIR_CES:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L3_DIR_CES_type);
        break;

    // Get size for ATTR_MNFG_TH_P8EX_L3_LINE_DELETES attribute
    case ATTR_MNFG_TH_P8EX_L3_LINE_DELETES:
        l_attrSize = sizeof(ATTR_MNFG_TH_P8EX_L3_LINE_DELETES_type);
        break;

    // Get size for ATTR_MNFG_TH_RCD_PARITY_ERRORS attribute
    case ATTR_MNFG_TH_RCD_PARITY_ERRORS:
        l_attrSize = sizeof(ATTR_MNFG_TH_RCD_PARITY_ERRORS_type);
        break;

    // Get size for ATTR_MNFG_XBUS_MIN_EYE_WIDTH attribute
    case ATTR_MNFG_XBUS_MIN_EYE_WIDTH:
        l_attrSize = sizeof(ATTR_MNFG_XBUS_MIN_EYE_WIDTH_type);
        break;

    // Get size for ATTR_MODEL attribute
    case ATTR_MODEL:
        l_attrSize = sizeof(ATTR_MODEL_type);
        break;

    // Get size for ATTR_MRU_ID attribute
    case ATTR_MRU_ID:
        l_attrSize = sizeof(ATTR_MRU_ID_type);
        break;

    // Get size for ATTR_MRW_DDR3_VDDR_MAX_LIMIT attribute
    case ATTR_MRW_DDR3_VDDR_MAX_LIMIT:
        l_attrSize = sizeof(ATTR_MRW_DDR3_VDDR_MAX_LIMIT_type);
        break;

    // Get size for ATTR_MRW_DDR3_VDDR_MAX_LIMIT_POST_DRAM_INIT attribute
    case ATTR_MRW_DDR3_VDDR_MAX_LIMIT_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MRW_DDR3_VDDR_MAX_LIMIT_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MRW_DDR4_VDDR_MAX_LIMIT attribute
    case ATTR_MRW_DDR4_VDDR_MAX_LIMIT:
        l_attrSize = sizeof(ATTR_MRW_DDR4_VDDR_MAX_LIMIT_type);
        break;

    // Get size for ATTR_MRW_DDR4_VDDR_MAX_LIMIT_POST_DRAM_INIT attribute
    case ATTR_MRW_DDR4_VDDR_MAX_LIMIT_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MRW_DDR4_VDDR_MAX_LIMIT_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MRW_DEFAULT_RISK_LEVEL attribute
    case ATTR_MRW_DEFAULT_RISK_LEVEL:
        l_attrSize = sizeof(ATTR_MRW_DEFAULT_RISK_LEVEL_type);
        break;

    // Get size for ATTR_MRW_MEM_THROTTLE_DENOMINATOR attribute
    case ATTR_MRW_MEM_THROTTLE_DENOMINATOR:
        l_attrSize = sizeof(ATTR_MRW_MEM_THROTTLE_DENOMINATOR_type);
        break;

    // Get size for ATTR_MSS_AVDD_PROGRAM attribute
    case ATTR_MSS_AVDD_PROGRAM:
        l_attrSize = sizeof(ATTR_MSS_AVDD_PROGRAM_type);
        break;

    // Get size for ATTR_MSS_CENT_AVDD_INTERCEPT attribute
    case ATTR_MSS_CENT_AVDD_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_CENT_AVDD_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_CENT_AVDD_SLOPE_ACTIVE attribute
    case ATTR_MSS_CENT_AVDD_SLOPE_ACTIVE:
        l_attrSize = sizeof(ATTR_MSS_CENT_AVDD_SLOPE_ACTIVE_type);
        break;

    // Get size for ATTR_MSS_CENT_AVDD_SLOPE_INACTIVE attribute
    case ATTR_MSS_CENT_AVDD_SLOPE_INACTIVE:
        l_attrSize = sizeof(ATTR_MSS_CENT_AVDD_SLOPE_INACTIVE_type);
        break;

    // Get size for ATTR_MSS_CENT_VCS_INTERCEPT attribute
    case ATTR_MSS_CENT_VCS_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_CENT_VCS_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_CENT_VCS_SLOPE_ACTIVE attribute
    case ATTR_MSS_CENT_VCS_SLOPE_ACTIVE:
        l_attrSize = sizeof(ATTR_MSS_CENT_VCS_SLOPE_ACTIVE_type);
        break;

    // Get size for ATTR_MSS_CENT_VCS_SLOPE_INACTIVE attribute
    case ATTR_MSS_CENT_VCS_SLOPE_INACTIVE:
        l_attrSize = sizeof(ATTR_MSS_CENT_VCS_SLOPE_INACTIVE_type);
        break;

    // Get size for ATTR_MSS_CENT_VDD_INTERCEPT attribute
    case ATTR_MSS_CENT_VDD_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_CENT_VDD_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_CENT_VDD_SLOPE_ACTIVE attribute
    case ATTR_MSS_CENT_VDD_SLOPE_ACTIVE:
        l_attrSize = sizeof(ATTR_MSS_CENT_VDD_SLOPE_ACTIVE_type);
        break;

    // Get size for ATTR_MSS_CENT_VDD_SLOPE_INACTIVE attribute
    case ATTR_MSS_CENT_VDD_SLOPE_INACTIVE:
        l_attrSize = sizeof(ATTR_MSS_CENT_VDD_SLOPE_INACTIVE_type);
        break;

    // Get size for ATTR_MSS_DATABUS_UTIL_PER_MBA attribute
    case ATTR_MSS_DATABUS_UTIL_PER_MBA:
        l_attrSize = sizeof(ATTR_MSS_DATABUS_UTIL_PER_MBA_type);
        break;

    // Get size for ATTR_MSS_MBA_ADDR_INTERLEAVE_BIT attribute
    case ATTR_MSS_MBA_ADDR_INTERLEAVE_BIT:
        l_attrSize = sizeof(ATTR_MSS_MBA_ADDR_INTERLEAVE_BIT_type);
        break;

    // Get size for ATTR_MSS_MBA_CACHELINE_INTERLEAVE_MODE attribute
    case ATTR_MSS_MBA_CACHELINE_INTERLEAVE_MODE:
        l_attrSize = sizeof(ATTR_MSS_MBA_CACHELINE_INTERLEAVE_MODE_type);
        break;

    // Get size for ATTR_MSS_PHY_SEQ_REFRESH attribute
    case ATTR_MSS_PHY_SEQ_REFRESH:
        l_attrSize = sizeof(ATTR_MSS_PHY_SEQ_REFRESH_type);
        break;

    // Get size for ATTR_MSS_VCS_PROGRAM attribute
    case ATTR_MSS_VCS_PROGRAM:
        l_attrSize = sizeof(ATTR_MSS_VCS_PROGRAM_type);
        break;

    // Get size for ATTR_MSS_VDDR_PROGRAM attribute
    case ATTR_MSS_VDDR_PROGRAM:
        l_attrSize = sizeof(ATTR_MSS_VDDR_PROGRAM_type);
        break;

    // Get size for ATTR_MSS_VDD_PROGRAM attribute
    case ATTR_MSS_VDD_PROGRAM:
        l_attrSize = sizeof(ATTR_MSS_VDD_PROGRAM_type);
        break;

    // Get size for ATTR_MSS_VOLT_AVDD_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_AVDD_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_AVDD_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_AVDD_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT attribute
    case ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_POST_DRAM_INIT attribute
    case ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR3_VDDR_SLOPE attribute
    case ATTR_MSS_VOLT_DDR3_VDDR_SLOPE:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_POST_DRAM_INIT attribute
    case ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT attribute
    case ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_POST_DRAM_INIT attribute
    case ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR4_VDDR_SLOPE attribute
    case ATTR_MSS_VOLT_DDR4_VDDR_SLOPE:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_POST_DRAM_INIT attribute
    case ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MSS_VOLT_VCS_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VCS_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VCS_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VCS_OFFSET_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VCS_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VCS_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VDDR_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VDDR_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VDDR_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VDDR_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VDD_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VDD_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VDD_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VDD_OFFSET_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VDD_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VDD_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_INTERCEPT attribute
    case ATTR_MSS_VOLT_VPP_INTERCEPT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_INTERCEPT_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_INTERCEPT_POST_DRAM_INIT attribute
    case ATTR_MSS_VOLT_VPP_INTERCEPT_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_INTERCEPT_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VPP_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_OFFSET_MILLIVOLTS attribute
    case ATTR_MSS_VOLT_VPP_OFFSET_MILLIVOLTS:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_OFFSET_MILLIVOLTS_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_SLOPE attribute
    case ATTR_MSS_VOLT_VPP_SLOPE:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_SLOPE_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_SLOPE_POST_DRAM_INIT attribute
    case ATTR_MSS_VOLT_VPP_SLOPE_POST_DRAM_INIT:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_SLOPE_POST_DRAM_INIT_type);
        break;

    // Get size for ATTR_MSS_VPP_PROGRAM attribute
    case ATTR_MSS_VPP_PROGRAM:
        l_attrSize = sizeof(ATTR_MSS_VPP_PROGRAM_type);
        break;

    // Get size for ATTR_NEST_PLL_FREQ_BUCKETS attribute
    case ATTR_NEST_PLL_FREQ_BUCKETS:
        l_attrSize = sizeof(ATTR_NEST_PLL_FREQ_BUCKETS_type);
        break;

    // Get size for ATTR_NEST_PLL_FREQ_I2CDIV_LIST attribute
    case ATTR_NEST_PLL_FREQ_I2CDIV_LIST:
        l_attrSize = sizeof(ATTR_NEST_PLL_FREQ_I2CDIV_LIST_type);
        break;

    // Get size for ATTR_NEST_PLL_FREQ_LIST attribute
    case ATTR_NEST_PLL_FREQ_LIST:
        l_attrSize = sizeof(ATTR_NEST_PLL_FREQ_LIST_type);
        break;

    // Get size for ATTR_NEST_VCS_ID attribute
    case ATTR_NEST_VCS_ID:
        l_attrSize = sizeof(ATTR_NEST_VCS_ID_type);
        break;

    // Get size for ATTR_NEST_VDDR_ID attribute
    case ATTR_NEST_VDDR_ID:
        l_attrSize = sizeof(ATTR_NEST_VDDR_ID_type);
        break;

    // Get size for ATTR_NEST_VDD_ID attribute
    case ATTR_NEST_VDD_ID:
        l_attrSize = sizeof(ATTR_NEST_VDD_ID_type);
        break;

    // Get size for ATTR_NEST_VDN_ID attribute
    case ATTR_NEST_VDN_ID:
        l_attrSize = sizeof(ATTR_NEST_VDN_ID_type);
        break;

    // Get size for ATTR_NEST_VIO_ID attribute
    case ATTR_NEST_VIO_ID:
        l_attrSize = sizeof(ATTR_NEST_VIO_ID_type);
        break;

    // Get size for ATTR_NOMINAL_FREQ_MHZ attribute
    case ATTR_NOMINAL_FREQ_MHZ:
        l_attrSize = sizeof(ATTR_NOMINAL_FREQ_MHZ_type);
        break;

    // Get size for ATTR_NPU_MMIO_BAR_BASE_ADDR attribute
    case ATTR_NPU_MMIO_BAR_BASE_ADDR:
        l_attrSize = sizeof(ATTR_NPU_MMIO_BAR_BASE_ADDR_type);
        break;

    // Get size for ATTR_NPU_MMIO_BAR_SIZE attribute
    case ATTR_NPU_MMIO_BAR_SIZE:
        l_attrSize = sizeof(ATTR_NPU_MMIO_BAR_SIZE_type);
        break;

    // Get size for ATTR_NUMERIC_POD_TYPE_TEST attribute
    case ATTR_NUMERIC_POD_TYPE_TEST:
        l_attrSize = sizeof(ATTR_NUMERIC_POD_TYPE_TEST_type);
        break;

    // Get size for ATTR_NVDIMM_ARMED attribute
    case ATTR_NVDIMM_ARMED:
        l_attrSize = sizeof(ATTR_NVDIMM_ARMED_type);
        break;

    // Get size for ATTR_NVDIMM_AUTO_ARM attribute
    case ATTR_NVDIMM_AUTO_ARM:
        l_attrSize = sizeof(ATTR_NVDIMM_AUTO_ARM_type);
        break;

    // Get size for ATTR_NVDIMM_ENCRYPTION_ENABLE attribute
    case ATTR_NVDIMM_ENCRYPTION_ENABLE:
        l_attrSize = sizeof(ATTR_NVDIMM_ENCRYPTION_ENABLE_type);
        break;

    // Get size for ATTR_NVDIMM_ENCRYPTION_KEYS_ANCHOR attribute
    case ATTR_NVDIMM_ENCRYPTION_KEYS_ANCHOR:
        l_attrSize = sizeof(ATTR_NVDIMM_ENCRYPTION_KEYS_ANCHOR_type);
        break;

    // Get size for ATTR_NVDIMM_ENCRYPTION_KEYS_FW attribute
    case ATTR_NVDIMM_ENCRYPTION_KEYS_FW:
        l_attrSize = sizeof(ATTR_NVDIMM_ENCRYPTION_KEYS_FW_type);
        break;

    // Get size for ATTR_NVDIMM_FORCE_FW_UPDATE attribute
    case ATTR_NVDIMM_FORCE_FW_UPDATE:
        l_attrSize = sizeof(ATTR_NVDIMM_FORCE_FW_UPDATE_type);
        break;

    // Get size for ATTR_NVDIMM_UPDATE_I2C_BLOCK_SIZE attribute
    case ATTR_NVDIMM_UPDATE_I2C_BLOCK_SIZE:
        l_attrSize = sizeof(ATTR_NVDIMM_UPDATE_I2C_BLOCK_SIZE_type);
        break;

    // Get size for ATTR_NV_OPS_TIMEOUT_MSEC attribute
    case ATTR_NV_OPS_TIMEOUT_MSEC:
        l_attrSize = sizeof(ATTR_NV_OPS_TIMEOUT_MSEC_type);
        break;

    // Get size for ATTR_NV_STATUS_FLAG attribute
    case ATTR_NV_STATUS_FLAG:
        l_attrSize = sizeof(ATTR_NV_STATUS_FLAG_type);
        break;

    // Get size for ATTR_OBUS_BRICK_LANE_MASK attribute
    case ATTR_OBUS_BRICK_LANE_MASK:
        l_attrSize = sizeof(ATTR_OBUS_BRICK_LANE_MASK_type);
        break;

    // Get size for ATTR_OBUS_SLOT_INDEX attribute
    case ATTR_OBUS_SLOT_INDEX:
        l_attrSize = sizeof(ATTR_OBUS_SLOT_INDEX_type);
        break;

    // Get size for ATTR_OCC_MASTER_CAPABLE attribute
    case ATTR_OCC_MASTER_CAPABLE:
        l_attrSize = sizeof(ATTR_OCC_MASTER_CAPABLE_type);
        break;

    // Get size for ATTR_OCMB_FW_UPDATE_OVERRIDE attribute
    case ATTR_OCMB_FW_UPDATE_OVERRIDE:
        l_attrSize = sizeof(ATTR_OCMB_FW_UPDATE_OVERRIDE_type);
        break;

    // Get size for ATTR_ORDINAL_ID attribute
    case ATTR_ORDINAL_ID:
        l_attrSize = sizeof(ATTR_ORDINAL_ID_type);
        break;

    // Get size for ATTR_PARENT_DECONFIG_RULES attribute
    case ATTR_PARENT_DECONFIG_RULES:
        l_attrSize = sizeof(ATTR_PARENT_DECONFIG_RULES_type);
        break;

    // Get size for ATTR_PAYLOAD_BASE attribute
    case ATTR_PAYLOAD_BASE:
        l_attrSize = sizeof(ATTR_PAYLOAD_BASE_type);
        break;

    // Get size for ATTR_PAYLOAD_ENTRY attribute
    case ATTR_PAYLOAD_ENTRY:
        l_attrSize = sizeof(ATTR_PAYLOAD_ENTRY_type);
        break;

    // Get size for ATTR_PAYLOAD_IN_MIRROR_MEM attribute
    case ATTR_PAYLOAD_IN_MIRROR_MEM:
        l_attrSize = sizeof(ATTR_PAYLOAD_IN_MIRROR_MEM_type);
        break;

    // Get size for ATTR_PAYLOAD_KIND attribute
    case ATTR_PAYLOAD_KIND:
        l_attrSize = sizeof(ATTR_PAYLOAD_KIND_type);
        break;

    // Get size for ATTR_PCIE_32BIT_DMA_SIZE attribute
    case ATTR_PCIE_32BIT_DMA_SIZE:
        l_attrSize = sizeof(ATTR_PCIE_32BIT_DMA_SIZE_type);
        break;

    // Get size for ATTR_PCIE_32BIT_MMIO_SIZE attribute
    case ATTR_PCIE_32BIT_MMIO_SIZE:
        l_attrSize = sizeof(ATTR_PCIE_32BIT_MMIO_SIZE_type);
        break;

    // Get size for ATTR_PCIE_64BIT_DMA_SIZE attribute
    case ATTR_PCIE_64BIT_DMA_SIZE:
        l_attrSize = sizeof(ATTR_PCIE_64BIT_DMA_SIZE_type);
        break;

    // Get size for ATTR_PCIE_64BIT_MMIO_SIZE attribute
    case ATTR_PCIE_64BIT_MMIO_SIZE:
        l_attrSize = sizeof(ATTR_PCIE_64BIT_MMIO_SIZE_type);
        break;

    // Get size for ATTR_PCIE_CAPABILITES attribute
    case ATTR_PCIE_CAPABILITES:
        l_attrSize = sizeof(ATTR_PCIE_CAPABILITES_type);
        break;

    // Get size for ATTR_PCI_BASE_ADDRS_32 attribute
    case ATTR_PCI_BASE_ADDRS_32:
        l_attrSize = sizeof(ATTR_PCI_BASE_ADDRS_32_type);
        break;

    // Get size for ATTR_PCI_BASE_ADDRS_64 attribute
    case ATTR_PCI_BASE_ADDRS_64:
        l_attrSize = sizeof(ATTR_PCI_BASE_ADDRS_64_type);
        break;

    // Get size for ATTR_PCI_REFCLOCK_RCVR_TERM attribute
    case ATTR_PCI_REFCLOCK_RCVR_TERM:
        l_attrSize = sizeof(ATTR_PCI_REFCLOCK_RCVR_TERM_type);
        break;

    // Get size for ATTR_PEC_IS_BIFURCATABLE attribute
    case ATTR_PEC_IS_BIFURCATABLE:
        l_attrSize = sizeof(ATTR_PEC_IS_BIFURCATABLE_type);
        break;

    // Get size for ATTR_PEC_PCIE_HX_KEYWORD_DATA attribute
    case ATTR_PEC_PCIE_HX_KEYWORD_DATA:
        l_attrSize = sizeof(ATTR_PEC_PCIE_HX_KEYWORD_DATA_type);
        break;

    // Get size for ATTR_PEC_PCIE_IOP_REVERSAL attribute
    case ATTR_PEC_PCIE_IOP_REVERSAL:
        l_attrSize = sizeof(ATTR_PEC_PCIE_IOP_REVERSAL_type);
        break;

    // Get size for ATTR_PEC_PCIE_IOP_REVERSAL_BIFURCATED attribute
    case ATTR_PEC_PCIE_IOP_REVERSAL_BIFURCATED:
        l_attrSize = sizeof(ATTR_PEC_PCIE_IOP_REVERSAL_BIFURCATED_type);
        break;

    // Get size for ATTR_PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED attribute
    case ATTR_PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED:
        l_attrSize = sizeof(ATTR_PEC_PCIE_IOP_REVERSAL_NON_BIFURCATED_type);
        break;

    // Get size for ATTR_PEC_PCIE_IOP_SWAP_BIFURCATED attribute
    case ATTR_PEC_PCIE_IOP_SWAP_BIFURCATED:
        l_attrSize = sizeof(ATTR_PEC_PCIE_IOP_SWAP_BIFURCATED_type);
        break;

    // Get size for ATTR_PEC_PCIE_IOP_SWAP_NON_BIFURCATED attribute
    case ATTR_PEC_PCIE_IOP_SWAP_NON_BIFURCATED:
        l_attrSize = sizeof(ATTR_PEC_PCIE_IOP_SWAP_NON_BIFURCATED_type);
        break;

    // Get size for ATTR_PEC_PCIE_LANE_MASK_BIFURCATED attribute
    case ATTR_PEC_PCIE_LANE_MASK_BIFURCATED:
        l_attrSize = sizeof(ATTR_PEC_PCIE_LANE_MASK_BIFURCATED_type);
        break;

    // Get size for ATTR_PEC_PCIE_LANE_MASK_NON_BIFURCATED attribute
    case ATTR_PEC_PCIE_LANE_MASK_NON_BIFURCATED:
        l_attrSize = sizeof(ATTR_PEC_PCIE_LANE_MASK_NON_BIFURCATED_type);
        break;

    // Get size for ATTR_PEER_PATH attribute
    case ATTR_PEER_PATH:
        l_attrSize = sizeof(ATTR_PEER_PATH_type);
        break;

    // Get size for ATTR_PEER_TARGET attribute
    case ATTR_PEER_TARGET:
        l_attrSize = sizeof(ATTR_PEER_TARGET_type);
        break;

    // Get size for ATTR_PG attribute
    case ATTR_PG:
        l_attrSize = sizeof(ATTR_PG_type);
        break;

    // Get size for ATTR_PHB_BASE_ADDRS attribute
    case ATTR_PHB_BASE_ADDRS:
        l_attrSize = sizeof(ATTR_PHB_BASE_ADDRS_type);
        break;

    // Get size for ATTR_PHYS_PATH attribute
    case ATTR_PHYS_PATH:
        l_attrSize = sizeof(ATTR_PHYS_PATH_type);
        break;

    // Get size for ATTR_PIB_I2C_NEST_PLL attribute
    case ATTR_PIB_I2C_NEST_PLL:
        l_attrSize = sizeof(ATTR_PIB_I2C_NEST_PLL_type);
        break;

    // Get size for ATTR_PIB_I2C_REFCLOCK attribute
    case ATTR_PIB_I2C_REFCLOCK:
        l_attrSize = sizeof(ATTR_PIB_I2C_REFCLOCK_type);
        break;

    // Get size for ATTR_PLCK_IPL_ATTR_OVERRIDES_EXIST attribute
    case ATTR_PLCK_IPL_ATTR_OVERRIDES_EXIST:
        l_attrSize = sizeof(ATTR_PLCK_IPL_ATTR_OVERRIDES_EXIST_type);
        break;

    // Get size for ATTR_PNOR_FLASH_WORKAROUNDS attribute
    case ATTR_PNOR_FLASH_WORKAROUNDS:
        l_attrSize = sizeof(ATTR_PNOR_FLASH_WORKAROUNDS_type);
        break;

    // Get size for ATTR_POSITION attribute
    case ATTR_POSITION:
        l_attrSize = sizeof(ATTR_POSITION_type);
        break;

    // Get size for ATTR_POS_ON_MEM_PORT attribute
    case ATTR_POS_ON_MEM_PORT:
        l_attrSize = sizeof(ATTR_POS_ON_MEM_PORT_type);
        break;

    // Get size for ATTR_POWER_PATH attribute
    case ATTR_POWER_PATH:
        l_attrSize = sizeof(ATTR_POWER_PATH_type);
        break;

    // Get size for ATTR_PRD_HWP_PLID attribute
    case ATTR_PRD_HWP_PLID:
        l_attrSize = sizeof(ATTR_PRD_HWP_PLID_type);
        break;

    // Get size for ATTR_PREVIOUS_WAKEUP_FAIL attribute
    case ATTR_PREVIOUS_WAKEUP_FAIL:
        l_attrSize = sizeof(ATTR_PREVIOUS_WAKEUP_FAIL_type);
        break;

    // Get size for ATTR_PRIMARY_CAPABILITIES attribute
    case ATTR_PRIMARY_CAPABILITIES:
        l_attrSize = sizeof(ATTR_PRIMARY_CAPABILITIES_type);
        break;

    // Get size for ATTR_PROC_COMPATIBILITY_REQ attribute
    case ATTR_PROC_COMPATIBILITY_REQ:
        l_attrSize = sizeof(ATTR_PROC_COMPATIBILITY_REQ_type);
        break;

    // Get size for ATTR_PROC_DCM_INSTALLED attribute
    case ATTR_PROC_DCM_INSTALLED:
        l_attrSize = sizeof(ATTR_PROC_DCM_INSTALLED_type);
        break;

    // Get size for ATTR_PROC_MASTER_TYPE attribute
    case ATTR_PROC_MASTER_TYPE:
        l_attrSize = sizeof(ATTR_PROC_MASTER_TYPE_type);
        break;

    // Get size for ATTR_PROC_PCIE_IOP_CONFIG attribute
    case ATTR_PROC_PCIE_IOP_CONFIG:
        l_attrSize = sizeof(ATTR_PROC_PCIE_IOP_CONFIG_type);
        break;

    // Get size for ATTR_PROC_PCIE_IOVALID_ENABLE attribute
    case ATTR_PROC_PCIE_IOVALID_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_IOVALID_ENABLE_type);
        break;

    // Get size for ATTR_PROC_PCIE_IS_SLOT attribute
    case ATTR_PROC_PCIE_IS_SLOT:
        l_attrSize = sizeof(ATTR_PROC_PCIE_IS_SLOT_type);
        break;

    // Get size for ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN3 attribute
    case ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN3:
        l_attrSize = sizeof(ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN3_type);
        break;

    // Get size for ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN4 attribute
    case ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN4:
        l_attrSize = sizeof(ATTR_PROC_PCIE_LANE_EQUALIZATION_GEN4_type);
        break;

    // Get size for ATTR_PROC_PCIE_LANE_MASK attribute
    case ATTR_PROC_PCIE_LANE_MASK:
        l_attrSize = sizeof(ATTR_PROC_PCIE_LANE_MASK_type);
        break;

    // Get size for ATTR_PROC_PCIE_NUM_PEC attribute
    case ATTR_PROC_PCIE_NUM_PEC:
        l_attrSize = sizeof(ATTR_PROC_PCIE_NUM_PEC_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_LOFF_CONTROL attribute
    case ATTR_PROC_PCIE_PCS_RX_LOFF_CONTROL:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_LOFF_CONTROL_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_PHASE_ROTATOR_CNTL attribute
    case ATTR_PROC_PCIE_PCS_RX_PHASE_ROTATOR_CNTL:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_PHASE_ROTATOR_CNTL_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_SIGDET_CNTL attribute
    case ATTR_PROC_PCIE_PCS_RX_SIGDET_CNTL:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_SIGDET_CNTL_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3 attribute
    case ATTR_PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_RX_VGA_CONTRL_REGISTER3_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_SYSTEM_CNTL attribute
    case ATTR_PROC_PCIE_PCS_SYSTEM_CNTL:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_SYSTEM_CNTL_type);
        break;

    // Get size for ATTR_PROC_PCIE_PCS_TX_FIFO_CONFIG_OFFSET attribute
    case ATTR_PROC_PCIE_PCS_TX_FIFO_CONFIG_OFFSET:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PCS_TX_FIFO_CONFIG_OFFSET_type);
        break;

    // Get size for ATTR_PROC_PCIE_PHB_ACTIVE attribute
    case ATTR_PROC_PCIE_PHB_ACTIVE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_PHB_ACTIVE_type);
        break;

    // Get size for ATTR_PROC_PCIE_REFCLOCK_ENABLE attribute
    case ATTR_PROC_PCIE_REFCLOCK_ENABLE:
        l_attrSize = sizeof(ATTR_PROC_PCIE_REFCLOCK_ENABLE_type);
        break;

    // Get size for ATTR_PROC_REFCLOCK_RCVR_TERM attribute
    case ATTR_PROC_REFCLOCK_RCVR_TERM:
        l_attrSize = sizeof(ATTR_PROC_REFCLOCK_RCVR_TERM_type);
        break;

    // Get size for ATTR_PROC_SELECT_BOOT_SEEPROM_IMAGE attribute
    case ATTR_PROC_SELECT_BOOT_SEEPROM_IMAGE:
        l_attrSize = sizeof(ATTR_PROC_SELECT_BOOT_SEEPROM_IMAGE_type);
        break;

    // Get size for ATTR_PROC_SELECT_SEEPROM_IMAGE attribute
    case ATTR_PROC_SELECT_SEEPROM_IMAGE:
        l_attrSize = sizeof(ATTR_PROC_SELECT_SEEPROM_IMAGE_type);
        break;

    // Get size for ATTR_PSI_BRIDGE_BASE_ADDR attribute
    case ATTR_PSI_BRIDGE_BASE_ADDR:
        l_attrSize = sizeof(ATTR_PSI_BRIDGE_BASE_ADDR_type);
        break;

    // Get size for ATTR_PSI_HB_ESB_ADDR attribute
    case ATTR_PSI_HB_ESB_ADDR:
        l_attrSize = sizeof(ATTR_PSI_HB_ESB_ADDR_type);
        break;

    // Get size for ATTR_RAW_MTM attribute
    case ATTR_RAW_MTM:
        l_attrSize = sizeof(ATTR_RAW_MTM_type);
        break;

    // Get size for ATTR_RCD_PARITY_RECONFIG_LOOPS_ALLOWED attribute
    case ATTR_RCD_PARITY_RECONFIG_LOOPS_ALLOWED:
        l_attrSize = sizeof(ATTR_RCD_PARITY_RECONFIG_LOOPS_ALLOWED_type);
        break;

    // Get size for ATTR_RCD_PARITY_RECONFIG_LOOP_COUNT attribute
    case ATTR_RCD_PARITY_RECONFIG_LOOP_COUNT:
        l_attrSize = sizeof(ATTR_RCD_PARITY_RECONFIG_LOOP_COUNT_type);
        break;

    // Get size for ATTR_RECONFIG_LOOP_TESTS attribute
    case ATTR_RECONFIG_LOOP_TESTS:
        l_attrSize = sizeof(ATTR_RECONFIG_LOOP_TESTS_type);
        break;

    // Get size for ATTR_RECONFIG_LOOP_TESTS_ENABLE attribute
    case ATTR_RECONFIG_LOOP_TESTS_ENABLE:
        l_attrSize = sizeof(ATTR_RECONFIG_LOOP_TESTS_ENABLE_type);
        break;

    // Get size for ATTR_REDUNDANT_CLOCKS attribute
    case ATTR_REDUNDANT_CLOCKS:
        l_attrSize = sizeof(ATTR_REDUNDANT_CLOCKS_type);
        break;

    // Get size for ATTR_REL_POS attribute
    case ATTR_REL_POS:
        l_attrSize = sizeof(ATTR_REL_POS_type);
        break;

    // Get size for ATTR_RESOURCE_IS_CRITICAL attribute
    case ATTR_RESOURCE_IS_CRITICAL:
        l_attrSize = sizeof(ATTR_RESOURCE_IS_CRITICAL_type);
        break;

    // Get size for ATTR_RISK_LEVEL_ORIGIN attribute
    case ATTR_RISK_LEVEL_ORIGIN:
        l_attrSize = sizeof(ATTR_RISK_LEVEL_ORIGIN_type);
        break;

    // Get size for ATTR_RNG_BAR_SIZE attribute
    case ATTR_RNG_BAR_SIZE:
        l_attrSize = sizeof(ATTR_RNG_BAR_SIZE_type);
        break;

    // Get size for ATTR_RNG_BASE_ADDR attribute
    case ATTR_RNG_BASE_ADDR:
        l_attrSize = sizeof(ATTR_RNG_BASE_ADDR_type);
        break;

    // Get size for ATTR_RUN_MAX_MEM_PATTERNS attribute
    case ATTR_RUN_MAX_MEM_PATTERNS:
        l_attrSize = sizeof(ATTR_RUN_MAX_MEM_PATTERNS_type);
        break;

    // Get size for ATTR_SAVED_STATE_UINT8 attribute
    case ATTR_SAVED_STATE_UINT8:
        l_attrSize = sizeof(ATTR_SAVED_STATE_UINT8_type);
        break;

    // Get size for ATTR_SBE_COMMIT_ID attribute
    case ATTR_SBE_COMMIT_ID:
        l_attrSize = sizeof(ATTR_SBE_COMMIT_ID_type);
        break;

    // Get size for ATTR_SBE_IS_STARTED attribute
    case ATTR_SBE_IS_STARTED:
        l_attrSize = sizeof(ATTR_SBE_IS_STARTED_type);
        break;

    // Get size for ATTR_SBE_RELEASE_TAG attribute
    case ATTR_SBE_RELEASE_TAG:
        l_attrSize = sizeof(ATTR_SBE_RELEASE_TAG_type);
        break;

    // Get size for ATTR_SBE_UPDATE_DISABLE attribute
    case ATTR_SBE_UPDATE_DISABLE:
        l_attrSize = sizeof(ATTR_SBE_UPDATE_DISABLE_type);
        break;

    // Get size for ATTR_SBE_VERSION_INFO attribute
    case ATTR_SBE_VERSION_INFO:
        l_attrSize = sizeof(ATTR_SBE_VERSION_INFO_type);
        break;

    // Get size for ATTR_SCOM_SWITCHES attribute
    case ATTR_SCOM_SWITCHES:
        l_attrSize = sizeof(ATTR_SCOM_SWITCHES_type);
        break;

    // Get size for ATTR_SECUREBOOT_PROTECT_DECONFIGURED_TPM attribute
    case ATTR_SECUREBOOT_PROTECT_DECONFIGURED_TPM:
        l_attrSize = sizeof(ATTR_SECUREBOOT_PROTECT_DECONFIGURED_TPM_type);
        break;

    // Get size for ATTR_SLOT_INDEX attribute
    case ATTR_SLOT_INDEX:
        l_attrSize = sizeof(ATTR_SLOT_INDEX_type);
        break;

    // Get size for ATTR_SLOT_NAME attribute
    case ATTR_SLOT_NAME:
        l_attrSize = sizeof(ATTR_SLOT_NAME_type);
        break;

    // Get size for ATTR_SOCKET_POWER_NOMINAL attribute
    case ATTR_SOCKET_POWER_NOMINAL:
        l_attrSize = sizeof(ATTR_SOCKET_POWER_NOMINAL_type);
        break;

    // Get size for ATTR_SOCKET_POWER_TURBO attribute
    case ATTR_SOCKET_POWER_TURBO:
        l_attrSize = sizeof(ATTR_SOCKET_POWER_TURBO_type);
        break;

    // Get size for ATTR_SP_FUNCTIONS attribute
    case ATTR_SP_FUNCTIONS:
        l_attrSize = sizeof(ATTR_SP_FUNCTIONS_type);
        break;

    // Get size for ATTR_START_MEM_ADDRESS_FOR_PAYLOAD_TCE_TOKEN attribute
    case ATTR_START_MEM_ADDRESS_FOR_PAYLOAD_TCE_TOKEN:
        l_attrSize = sizeof(ATTR_START_MEM_ADDRESS_FOR_PAYLOAD_TCE_TOKEN_type);
        break;

    // Get size for ATTR_STOP5_DISABLE attribute
    case ATTR_STOP5_DISABLE:
        l_attrSize = sizeof(ATTR_STOP5_DISABLE_type);
        break;

    // Get size for ATTR_SUPPORTED_STOP_STATES attribute
    case ATTR_SUPPORTED_STOP_STATES:
        l_attrSize = sizeof(ATTR_SUPPORTED_STOP_STATES_type);
        break;

    // Get size for ATTR_SUPPORTS_DYNAMIC_MEM_VOLT attribute
    case ATTR_SUPPORTS_DYNAMIC_MEM_VOLT:
        l_attrSize = sizeof(ATTR_SUPPORTS_DYNAMIC_MEM_VOLT_type);
        break;

    // Get size for ATTR_SYNC_BETWEEN_STEPS attribute
    case ATTR_SYNC_BETWEEN_STEPS:
        l_attrSize = sizeof(ATTR_SYNC_BETWEEN_STEPS_type);
        break;

    // Get size for ATTR_SYSTEM_FAMILY attribute
    case ATTR_SYSTEM_FAMILY:
        l_attrSize = sizeof(ATTR_SYSTEM_FAMILY_type);
        break;

    // Get size for ATTR_SYSTEM_TYPE attribute
    case ATTR_SYSTEM_TYPE:
        l_attrSize = sizeof(ATTR_SYSTEM_TYPE_type);
        break;

    // Get size for ATTR_TCE_START_TOKEN_FOR_HDAT attribute
    case ATTR_TCE_START_TOKEN_FOR_HDAT:
        l_attrSize = sizeof(ATTR_TCE_START_TOKEN_FOR_HDAT_type);
        break;

    // Get size for ATTR_TCE_START_TOKEN_FOR_PAYLOAD attribute
    case ATTR_TCE_START_TOKEN_FOR_PAYLOAD:
        l_attrSize = sizeof(ATTR_TCE_START_TOKEN_FOR_PAYLOAD_type);
        break;

    // Get size for ATTR_TEMP_SENSOR_I2C_CONFIG attribute
    case ATTR_TEMP_SENSOR_I2C_CONFIG:
        l_attrSize = sizeof(ATTR_TEMP_SENSOR_I2C_CONFIG_type);
        break;

    // Get size for ATTR_TEST_MAX_STRING attribute
    case ATTR_TEST_MAX_STRING:
        l_attrSize = sizeof(ATTR_TEST_MAX_STRING_type);
        break;

    // Get size for ATTR_TEST_MIN_STRING attribute
    case ATTR_TEST_MIN_STRING:
        l_attrSize = sizeof(ATTR_TEST_MIN_STRING_type);
        break;

    // Get size for ATTR_TEST_NEGATIVE_FCN attribute
    case ATTR_TEST_NEGATIVE_FCN:
        l_attrSize = sizeof(ATTR_TEST_NEGATIVE_FCN_type);
        break;

    // Get size for ATTR_TEST_NO_DEFAULT_STRING attribute
    case ATTR_TEST_NO_DEFAULT_STRING:
        l_attrSize = sizeof(ATTR_TEST_NO_DEFAULT_STRING_type);
        break;

    // Get size for ATTR_TEST_NULL_STRING attribute
    case ATTR_TEST_NULL_STRING:
        l_attrSize = sizeof(ATTR_TEST_NULL_STRING_type);
        break;

    // Get size for ATTR_THREAD_COUNT attribute
    case ATTR_THREAD_COUNT:
        l_attrSize = sizeof(ATTR_THREAD_COUNT_type);
        break;

    // Get size for ATTR_TIME_BASE attribute
    case ATTR_TIME_BASE:
        l_attrSize = sizeof(ATTR_TIME_BASE_type);
        break;

    // Get size for ATTR_TLB_DATA_ASSOC_SETS attribute
    case ATTR_TLB_DATA_ASSOC_SETS:
        l_attrSize = sizeof(ATTR_TLB_DATA_ASSOC_SETS_type);
        break;

    // Get size for ATTR_TLB_DATA_ENTRIES attribute
    case ATTR_TLB_DATA_ENTRIES:
        l_attrSize = sizeof(ATTR_TLB_DATA_ENTRIES_type);
        break;

    // Get size for ATTR_TLB_INSTR_ASSOC_SETS attribute
    case ATTR_TLB_INSTR_ASSOC_SETS:
        l_attrSize = sizeof(ATTR_TLB_INSTR_ASSOC_SETS_type);
        break;

    // Get size for ATTR_TLB_INSTR_ENTRIES attribute
    case ATTR_TLB_INSTR_ENTRIES:
        l_attrSize = sizeof(ATTR_TLB_INSTR_ENTRIES_type);
        break;

    // Get size for ATTR_TLB_RESERVE_SIZE attribute
    case ATTR_TLB_RESERVE_SIZE:
        l_attrSize = sizeof(ATTR_TLB_RESERVE_SIZE_type);
        break;

    // Get size for ATTR_TOD_CPU_DATA attribute
    case ATTR_TOD_CPU_DATA:
        l_attrSize = sizeof(ATTR_TOD_CPU_DATA_type);
        break;

    // Get size for ATTR_TOD_ROLE attribute
    case ATTR_TOD_ROLE:
        l_attrSize = sizeof(ATTR_TOD_ROLE_type);
        break;

    // Get size for ATTR_TPM_INFO attribute
    case ATTR_TPM_INFO:
        l_attrSize = sizeof(ATTR_TPM_INFO_type);
        break;

    // Get size for ATTR_TPM_POISONED attribute
    case ATTR_TPM_POISONED:
        l_attrSize = sizeof(ATTR_TPM_POISONED_type);
        break;

    // Get size for ATTR_TPM_REQUIRED attribute
    case ATTR_TPM_REQUIRED:
        l_attrSize = sizeof(ATTR_TPM_REQUIRED_type);
        break;

    // Get size for ATTR_TPM_UNUSABLE attribute
    case ATTR_TPM_UNUSABLE:
        l_attrSize = sizeof(ATTR_TPM_UNUSABLE_type);
        break;

    // Get size for ATTR_TYPE attribute
    case ATTR_TYPE:
        l_attrSize = sizeof(ATTR_TYPE_type);
        break;

    // Get size for ATTR_UCD_MFR_REVISION_OVERRIDE attribute
    case ATTR_UCD_MFR_REVISION_OVERRIDE:
        l_attrSize = sizeof(ATTR_UCD_MFR_REVISION_OVERRIDE_type);
        break;

    // Get size for ATTR_ULTRA_TURBO_FREQ_MHZ attribute
    case ATTR_ULTRA_TURBO_FREQ_MHZ:
        l_attrSize = sizeof(ATTR_ULTRA_TURBO_FREQ_MHZ_type);
        break;

    // Get size for ATTR_USE_TCES_FOR_DMAS attribute
    case ATTR_USE_TCES_FOR_DMAS:
        l_attrSize = sizeof(ATTR_USE_TCES_FOR_DMAS_type);
        break;

    // Get size for ATTR_VCS_ID attribute
    case ATTR_VCS_ID:
        l_attrSize = sizeof(ATTR_VCS_ID_type);
        break;

    // Get size for ATTR_VDDR_ID attribute
    case ATTR_VDDR_ID:
        l_attrSize = sizeof(ATTR_VDDR_ID_type);
        break;

    // Get size for ATTR_VDD_ID attribute
    case ATTR_VDD_ID:
        l_attrSize = sizeof(ATTR_VDD_ID_type);
        break;

    // Get size for ATTR_VENDOR_ID attribute
    case ATTR_VENDOR_ID:
        l_attrSize = sizeof(ATTR_VENDOR_ID_type);
        break;

    // Get size for ATTR_VPD_REC_NUM attribute
    case ATTR_VPD_REC_NUM:
        l_attrSize = sizeof(ATTR_VPD_REC_NUM_type);
        break;

    // Get size for ATTR_VPP_ID attribute
    case ATTR_VPP_ID:
        l_attrSize = sizeof(ATTR_VPP_ID_type);
        break;

    // Get size for ATTR_WOF_POWER_LIMIT attribute
    case ATTR_WOF_POWER_LIMIT:
        l_attrSize = sizeof(ATTR_WOF_POWER_LIMIT_type);
        break;

    // Get size for ATTR_WOF_TABLE_LID_NUMBER attribute
    case ATTR_WOF_TABLE_LID_NUMBER:
        l_attrSize = sizeof(ATTR_WOF_TABLE_LID_NUMBER_type);
        break;

    // Get size for ATTR_XIVE_CONTROLLER_BAR_ADDR attribute
    case ATTR_XIVE_CONTROLLER_BAR_ADDR:
        l_attrSize = sizeof(ATTR_XIVE_CONTROLLER_BAR_ADDR_type);
        break;

    // Get size for ATTR_XIVE_THREAD_MGMT1_BAR_ADDR attribute
    case ATTR_XIVE_THREAD_MGMT1_BAR_ADDR:
        l_attrSize = sizeof(ATTR_XIVE_THREAD_MGMT1_BAR_ADDR_type);
        break;

    // Get size for ATTR_XSCOM_BASE_ADDRESS attribute
    case ATTR_XSCOM_BASE_ADDRESS:
        l_attrSize = sizeof(ATTR_XSCOM_BASE_ADDRESS_type);
        break;

    // Get size for ATTR_ALLOW_ATTR_OVERRIDES_IN_SECURE_MODE attribute
    case ATTR_ALLOW_ATTR_OVERRIDES_IN_SECURE_MODE:
        l_attrSize = sizeof(ATTR_ALLOW_ATTR_OVERRIDES_IN_SECURE_MODE_type);
        break;

    // Get size for ATTR_ASSUME_SBE_QUIESCED attribute
    case ATTR_ASSUME_SBE_QUIESCED:
        l_attrSize = sizeof(ATTR_ASSUME_SBE_QUIESCED_type);
        break;

    // Get size for ATTR_ATTN_CHK_ALL_PROCS attribute
    case ATTR_ATTN_CHK_ALL_PROCS:
        l_attrSize = sizeof(ATTR_ATTN_CHK_ALL_PROCS_type);
        break;

    // Get size for ATTR_CENTAUR_REGISTER_CACHE_PTR attribute
    case ATTR_CENTAUR_REGISTER_CACHE_PTR:
        l_attrSize = sizeof(ATTR_CENTAUR_REGISTER_CACHE_PTR_type);
        break;

    // Get size for ATTR_CHKSTOP_ANALYSIS_ON_STARTUP attribute
    case ATTR_CHKSTOP_ANALYSIS_ON_STARTUP:
        l_attrSize = sizeof(ATTR_CHKSTOP_ANALYSIS_ON_STARTUP_type);
        break;

    // Get size for ATTR_CLEAR_DIMM_SPD_ENABLE attribute
    case ATTR_CLEAR_DIMM_SPD_ENABLE:
        l_attrSize = sizeof(ATTR_CLEAR_DIMM_SPD_ENABLE_type);
        break;

    // Get size for ATTR_DIMM_SPD_BYTE_SIZE attribute
    case ATTR_DIMM_SPD_BYTE_SIZE:
        l_attrSize = sizeof(ATTR_DIMM_SPD_BYTE_SIZE_type);
        break;

    // Get size for ATTR_DRTM_PAYLOAD_ADDR_MB_HB attribute
    case ATTR_DRTM_PAYLOAD_ADDR_MB_HB:
        l_attrSize = sizeof(ATTR_DRTM_PAYLOAD_ADDR_MB_HB_type);
        break;

    // Get size for ATTR_DYNAMIC_I2C_DEVICE_ADDRESS attribute
    case ATTR_DYNAMIC_I2C_DEVICE_ADDRESS:
        l_attrSize = sizeof(ATTR_DYNAMIC_I2C_DEVICE_ADDRESS_type);
        break;

    // Get size for ATTR_EARLY_TESTCASES_ISTEP attribute
    case ATTR_EARLY_TESTCASES_ISTEP:
        l_attrSize = sizeof(ATTR_EARLY_TESTCASES_ISTEP_type);
        break;

    // Get size for ATTR_EEPROM_PAGE_ARRAY attribute
    case ATTR_EEPROM_PAGE_ARRAY:
        l_attrSize = sizeof(ATTR_EEPROM_PAGE_ARRAY_type);
        break;

    // Get size for ATTR_FORCE_PRE_PAYLOAD_DRTM attribute
    case ATTR_FORCE_PRE_PAYLOAD_DRTM:
        l_attrSize = sizeof(ATTR_FORCE_PRE_PAYLOAD_DRTM_type);
        break;

    // Get size for ATTR_FORCE_SBE_UPDATE attribute
    case ATTR_FORCE_SBE_UPDATE:
        l_attrSize = sizeof(ATTR_FORCE_SBE_UPDATE_type);
        break;

    // Get size for ATTR_FREQ_MCA_MHZ attribute
    case ATTR_FREQ_MCA_MHZ:
        l_attrSize = sizeof(ATTR_FREQ_MCA_MHZ_type);
        break;

    // Get size for ATTR_FSI_MASTER_MUTEX attribute
    case ATTR_FSI_MASTER_MUTEX:
        l_attrSize = sizeof(ATTR_FSI_MASTER_MUTEX_type);
        break;

    // Get size for ATTR_FSI_SCOM_MUTEX attribute
    case ATTR_FSI_SCOM_MUTEX:
        l_attrSize = sizeof(ATTR_FSI_SCOM_MUTEX_type);
        break;

    // Get size for ATTR_GPIO_INFO_PHYS_PRES attribute
    case ATTR_GPIO_INFO_PHYS_PRES:
        l_attrSize = sizeof(ATTR_GPIO_INFO_PHYS_PRES_type);
        break;

    // Get size for ATTR_HBRT_HYP_ID attribute
    case ATTR_HBRT_HYP_ID:
        l_attrSize = sizeof(ATTR_HBRT_HYP_ID_type);
        break;

    // Get size for ATTR_HB_EXISTING_IMAGE attribute
    case ATTR_HB_EXISTING_IMAGE:
        l_attrSize = sizeof(ATTR_HB_EXISTING_IMAGE_type);
        break;

    // Get size for ATTR_HB_INITIATED_PM_RESET attribute
    case ATTR_HB_INITIATED_PM_RESET:
        l_attrSize = sizeof(ATTR_HB_INITIATED_PM_RESET_type);
        break;

    // Get size for ATTR_HB_MUTEX_SERIALIZE_TEST_LOCK attribute
    case ATTR_HB_MUTEX_SERIALIZE_TEST_LOCK:
        l_attrSize = sizeof(ATTR_HB_MUTEX_SERIALIZE_TEST_LOCK_type);
        break;

    // Get size for ATTR_HB_MUTEX_TEST_LOCK attribute
    case ATTR_HB_MUTEX_TEST_LOCK:
        l_attrSize = sizeof(ATTR_HB_MUTEX_TEST_LOCK_type);
        break;

    // Get size for ATTR_HB_NODE_COMM_ABUS_MUTEX attribute
    case ATTR_HB_NODE_COMM_ABUS_MUTEX:
        l_attrSize = sizeof(ATTR_HB_NODE_COMM_ABUS_MUTEX_type);
        break;

    // Get size for ATTR_HB_NODE_COMM_XBUS_MUTEX attribute
    case ATTR_HB_NODE_COMM_XBUS_MUTEX:
        l_attrSize = sizeof(ATTR_HB_NODE_COMM_XBUS_MUTEX_type);
        break;

    // Get size for ATTR_HB_RECURSIVE_MUTEX_TEST_LOCK attribute
    case ATTR_HB_RECURSIVE_MUTEX_TEST_LOCK:
        l_attrSize = sizeof(ATTR_HB_RECURSIVE_MUTEX_TEST_LOCK_type);
        break;

    // Get size for ATTR_HB_RSV_MEM_NEXT_SECTION attribute
    case ATTR_HB_RSV_MEM_NEXT_SECTION:
        l_attrSize = sizeof(ATTR_HB_RSV_MEM_NEXT_SECTION_type);
        break;

    // Get size for ATTR_HB_SBE_SEEPROM_VERSION_MISMATCH attribute
    case ATTR_HB_SBE_SEEPROM_VERSION_MISMATCH:
        l_attrSize = sizeof(ATTR_HB_SBE_SEEPROM_VERSION_MISMATCH_type);
        break;

    // Get size for ATTR_HB_TARGET_SCOMABLE attribute
    case ATTR_HB_TARGET_SCOMABLE:
        l_attrSize = sizeof(ATTR_HB_TARGET_SCOMABLE_type);
        break;

    // Get size for ATTR_HB_TPM_INIT_ATTEMPTED attribute
    case ATTR_HB_TPM_INIT_ATTEMPTED:
        l_attrSize = sizeof(ATTR_HB_TPM_INIT_ATTEMPTED_type);
        break;

    // Get size for ATTR_HB_TPM_LOG_MGR_PTR attribute
    case ATTR_HB_TPM_LOG_MGR_PTR:
        l_attrSize = sizeof(ATTR_HB_TPM_LOG_MGR_PTR_type);
        break;

    // Get size for ATTR_HB_TPM_MUTEX attribute
    case ATTR_HB_TPM_MUTEX:
        l_attrSize = sizeof(ATTR_HB_TPM_MUTEX_type);
        break;

    // Get size for ATTR_HIDDEN_ERRLOGS_ENABLE attribute
    case ATTR_HIDDEN_ERRLOGS_ENABLE:
        l_attrSize = sizeof(ATTR_HIDDEN_ERRLOGS_ENABLE_type);
        break;

    // Get size for ATTR_HOMER_HCODE_LOADED attribute
    case ATTR_HOMER_HCODE_LOADED:
        l_attrSize = sizeof(ATTR_HOMER_HCODE_LOADED_type);
        break;

    // Get size for ATTR_HOMER_VIRT_ADDR attribute
    case ATTR_HOMER_VIRT_ADDR:
        l_attrSize = sizeof(ATTR_HOMER_VIRT_ADDR_type);
        break;

    // Get size for ATTR_I2C_ENGINE_MUTEX_0 attribute
    case ATTR_I2C_ENGINE_MUTEX_0:
        l_attrSize = sizeof(ATTR_I2C_ENGINE_MUTEX_0_type);
        break;

    // Get size for ATTR_I2C_ENGINE_MUTEX_1 attribute
    case ATTR_I2C_ENGINE_MUTEX_1:
        l_attrSize = sizeof(ATTR_I2C_ENGINE_MUTEX_1_type);
        break;

    // Get size for ATTR_I2C_ENGINE_MUTEX_2 attribute
    case ATTR_I2C_ENGINE_MUTEX_2:
        l_attrSize = sizeof(ATTR_I2C_ENGINE_MUTEX_2_type);
        break;

    // Get size for ATTR_I2C_ENGINE_MUTEX_3 attribute
    case ATTR_I2C_ENGINE_MUTEX_3:
        l_attrSize = sizeof(ATTR_I2C_ENGINE_MUTEX_3_type);
        break;

    // Get size for ATTR_I2C_PAGE_MUTEX_0 attribute
    case ATTR_I2C_PAGE_MUTEX_0:
        l_attrSize = sizeof(ATTR_I2C_PAGE_MUTEX_0_type);
        break;

    // Get size for ATTR_I2C_PAGE_MUTEX_1 attribute
    case ATTR_I2C_PAGE_MUTEX_1:
        l_attrSize = sizeof(ATTR_I2C_PAGE_MUTEX_1_type);
        break;

    // Get size for ATTR_I2C_PAGE_MUTEX_2 attribute
    case ATTR_I2C_PAGE_MUTEX_2:
        l_attrSize = sizeof(ATTR_I2C_PAGE_MUTEX_2_type);
        break;

    // Get size for ATTR_I2C_PAGE_MUTEX_3 attribute
    case ATTR_I2C_PAGE_MUTEX_3:
        l_attrSize = sizeof(ATTR_I2C_PAGE_MUTEX_3_type);
        break;

    // Get size for ATTR_IBSCOM_ENABLE_OVERRIDE attribute
    case ATTR_IBSCOM_ENABLE_OVERRIDE:
        l_attrSize = sizeof(ATTR_IBSCOM_ENABLE_OVERRIDE_type);
        break;

    // Get size for ATTR_IBSCOM_MUTEX attribute
    case ATTR_IBSCOM_MUTEX:
        l_attrSize = sizeof(ATTR_IBSCOM_MUTEX_type);
        break;

    // Get size for ATTR_IBSCOM_VIRTUAL_ADDR attribute
    case ATTR_IBSCOM_VIRTUAL_ADDR:
        l_attrSize = sizeof(ATTR_IBSCOM_VIRTUAL_ADDR_type);
        break;

    // Get size for ATTR_IPC_NODE_BUFFER_GLOBAL_ADDRESS attribute
    case ATTR_IPC_NODE_BUFFER_GLOBAL_ADDRESS:
        l_attrSize = sizeof(ATTR_IPC_NODE_BUFFER_GLOBAL_ADDRESS_type);
        break;

    // Get size for ATTR_IS_DRTM_MPIPL_HB attribute
    case ATTR_IS_DRTM_MPIPL_HB:
        l_attrSize = sizeof(ATTR_IS_DRTM_MPIPL_HB_type);
        break;

    // Get size for ATTR_IS_MPIPL_HB attribute
    case ATTR_IS_MPIPL_HB:
        l_attrSize = sizeof(ATTR_IS_MPIPL_HB_type);
        break;

    // Get size for ATTR_IS_SLAVE_DRAWER attribute
    case ATTR_IS_SLAVE_DRAWER:
        l_attrSize = sizeof(ATTR_IS_SLAVE_DRAWER_type);
        break;

    // Get size for ATTR_MASTER_MBOX_SCRATCH attribute
    case ATTR_MASTER_MBOX_SCRATCH:
        l_attrSize = sizeof(ATTR_MASTER_MBOX_SCRATCH_type);
        break;

    // Get size for ATTR_MEMD_OFFSET attribute
    case ATTR_MEMD_OFFSET:
        l_attrSize = sizeof(ATTR_MEMD_OFFSET_type);
        break;

    // Get size for ATTR_MMIO_PHYS_ADDR attribute
    case ATTR_MMIO_PHYS_ADDR:
        l_attrSize = sizeof(ATTR_MMIO_PHYS_ADDR_type);
        break;

    // Get size for ATTR_MMIO_VM_ADDR attribute
    case ATTR_MMIO_VM_ADDR:
        l_attrSize = sizeof(ATTR_MMIO_VM_ADDR_type);
        break;

    // Get size for ATTR_MPIPL_HB_MDRT_COUNT attribute
    case ATTR_MPIPL_HB_MDRT_COUNT:
        l_attrSize = sizeof(ATTR_MPIPL_HB_MDRT_COUNT_type);
        break;

    // Get size for ATTR_MRW_DDR3_VDDR_MAX_LIMIT_EFF_CONFIG attribute
    case ATTR_MRW_DDR3_VDDR_MAX_LIMIT_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MRW_DDR3_VDDR_MAX_LIMIT_EFF_CONFIG_type);
        break;

    // Get size for ATTR_MRW_DDR4_VDDR_MAX_LIMIT_EFF_CONFIG attribute
    case ATTR_MRW_DDR4_VDDR_MAX_LIMIT_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MRW_DDR4_VDDR_MAX_LIMIT_EFF_CONFIG_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_EFF_CONFIG attribute
    case ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR3_VDDR_INTERCEPT_EFF_CONFIG_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_EFF_CONFIG attribute
    case ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR3_VDDR_SLOPE_EFF_CONFIG_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_EFF_CONFIG attribute
    case ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR4_VDDR_INTERCEPT_EFF_CONFIG_type);
        break;

    // Get size for ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_EFF_CONFIG attribute
    case ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MSS_VOLT_DDR4_VDDR_SLOPE_EFF_CONFIG_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_INTERCEPT_EFF_CONFIG attribute
    case ATTR_MSS_VOLT_VPP_INTERCEPT_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_INTERCEPT_EFF_CONFIG_type);
        break;

    // Get size for ATTR_MSS_VOLT_VPP_SLOPE_EFF_CONFIG attribute
    case ATTR_MSS_VOLT_VPP_SLOPE_EFF_CONFIG:
        l_attrSize = sizeof(ATTR_MSS_VOLT_VPP_SLOPE_EFF_CONFIG_type);
        break;

    // Get size for ATTR_NVDIMM_READING_PAGE4 attribute
    case ATTR_NVDIMM_READING_PAGE4:
        l_attrSize = sizeof(ATTR_NVDIMM_READING_PAGE4_type);
        break;

    // Get size for ATTR_NVDIMM_READING_VENDOR_LOG attribute
    case ATTR_NVDIMM_READING_VENDOR_LOG:
        l_attrSize = sizeof(ATTR_NVDIMM_READING_VENDOR_LOG_type);
        break;

    // Get size for ATTR_OCC_COMMON_AREA_PHYS_ADDR attribute
    case ATTR_OCC_COMMON_AREA_PHYS_ADDR:
        l_attrSize = sizeof(ATTR_OCC_COMMON_AREA_PHYS_ADDR_type);
        break;

    // Get size for ATTR_OCMB_COUNTER_HB attribute
    case ATTR_OCMB_COUNTER_HB:
        l_attrSize = sizeof(ATTR_OCMB_COUNTER_HB_type);
        break;

    // Get size for ATTR_OVERRIDES_ATTEMPTED_FLAG attribute
    case ATTR_OVERRIDES_ATTEMPTED_FLAG:
        l_attrSize = sizeof(ATTR_OVERRIDES_ATTEMPTED_FLAG_type);
        break;

    // Get size for ATTR_PART_NUMBER attribute
    case ATTR_PART_NUMBER:
        l_attrSize = sizeof(ATTR_PART_NUMBER_type);
        break;

    // Get size for ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_ADDR attribute
    case ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_ADDR:
        l_attrSize = sizeof(ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_ADDR_type);
        break;

    // Get size for ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_SIZE attribute
    case ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_SIZE:
        l_attrSize = sizeof(ATTR_PDA_CAPTURED_THREAD_REG_ARRAY_SIZE_type);
        break;

    // Get size for ATTR_PDA_THREAD_REG_ENTRY_SIZE attribute
    case ATTR_PDA_THREAD_REG_ENTRY_SIZE:
        l_attrSize = sizeof(ATTR_PDA_THREAD_REG_ENTRY_SIZE_type);
        break;

    // Get size for ATTR_PDA_THREAD_REG_STATE_ENTRY_FORMAT attribute
    case ATTR_PDA_THREAD_REG_STATE_ENTRY_FORMAT:
        l_attrSize = sizeof(ATTR_PDA_THREAD_REG_STATE_ENTRY_FORMAT_type);
        break;

    // Get size for ATTR_PHYS_PRES_ASSERTED attribute
    case ATTR_PHYS_PRES_ASSERTED:
        l_attrSize = sizeof(ATTR_PHYS_PRES_ASSERTED_type);
        break;

    // Get size for ATTR_PHYS_PRES_FAKE_ASSERT attribute
    case ATTR_PHYS_PRES_FAKE_ASSERT:
        l_attrSize = sizeof(ATTR_PHYS_PRES_FAKE_ASSERT_type);
        break;

    // Get size for ATTR_PHYS_PRES_REIPL attribute
    case ATTR_PHYS_PRES_REIPL:
        l_attrSize = sizeof(ATTR_PHYS_PRES_REIPL_type);
        break;

    // Get size for ATTR_PHYS_PRES_REQUEST_OPEN_WINDOW attribute
    case ATTR_PHYS_PRES_REQUEST_OPEN_WINDOW:
        l_attrSize = sizeof(ATTR_PHYS_PRES_REQUEST_OPEN_WINDOW_type);
        break;

    // Get size for ATTR_SBE_ARCH_DUMP_ADDR attribute
    case ATTR_SBE_ARCH_DUMP_ADDR:
        l_attrSize = sizeof(ATTR_SBE_ARCH_DUMP_ADDR_type);
        break;

    // Get size for ATTR_SBE_COMM_ADDR attribute
    case ATTR_SBE_COMM_ADDR:
        l_attrSize = sizeof(ATTR_SBE_COMM_ADDR_type);
        break;

    // Get size for ATTR_SBE_FFDC_ADDR attribute
    case ATTR_SBE_FFDC_ADDR:
        l_attrSize = sizeof(ATTR_SBE_FFDC_ADDR_type);
        break;

    // Get size for ATTR_SCAN_MUTEX attribute
    case ATTR_SCAN_MUTEX:
        l_attrSize = sizeof(ATTR_SCAN_MUTEX_type);
        break;

    // Get size for ATTR_SCOM_IND_MUTEX attribute
    case ATTR_SCOM_IND_MUTEX:
        l_attrSize = sizeof(ATTR_SCOM_IND_MUTEX_type);
        break;

    // Get size for ATTR_SERIAL_NUMBER attribute
    case ATTR_SERIAL_NUMBER:
        l_attrSize = sizeof(ATTR_SERIAL_NUMBER_type);
        break;

    // Get size for ATTR_SPCWKUP_COUNT attribute
    case ATTR_SPCWKUP_COUNT:
        l_attrSize = sizeof(ATTR_SPCWKUP_COUNT_type);
        break;

    // Get size for ATTR_TPM_MODEL attribute
    case ATTR_TPM_MODEL:
        l_attrSize = sizeof(ATTR_TPM_MODEL_type);
        break;

    // Get size for ATTR_TPM_ROLE attribute
    case ATTR_TPM_ROLE:
        l_attrSize = sizeof(ATTR_TPM_ROLE_type);
        break;

    // Get size for ATTR_VPD_SWITCHES attribute
    case ATTR_VPD_SWITCHES:
        l_attrSize = sizeof(ATTR_VPD_SWITCHES_type);
        break;

    // Get size for ATTR_WOF_FREQUENCY_UPLIFT_SELECTED attribute
    case ATTR_WOF_FREQUENCY_UPLIFT_SELECTED:
        l_attrSize = sizeof(ATTR_WOF_FREQUENCY_UPLIFT_SELECTED_type);
        break;

    // Get size for ATTR_XSCOM_VIRTUAL_ADDR attribute
    case ATTR_XSCOM_VIRTUAL_ADDR:
        l_attrSize = sizeof(ATTR_XSCOM_VIRTUAL_ADDR_type);
        break;

    // Get size for ATTR_ASCII_VPD_LX_KEYWORD attribute
    case ATTR_ASCII_VPD_LX_KEYWORD:
        l_attrSize = sizeof(ATTR_ASCII_VPD_LX_KEYWORD_type);
        break;

    // Get size for ATTR_CHIP_VER attribute
    case ATTR_CHIP_VER:
        l_attrSize = sizeof(ATTR_CHIP_VER_type);
        break;

    // Get size for ATTR_CREATE_DEF_PARTITION attribute
    case ATTR_CREATE_DEF_PARTITION:
        l_attrSize = sizeof(ATTR_CREATE_DEF_PARTITION_type);
        break;

    // Get size for ATTR_ENLARGED_IO_SLOT_COUNT attribute
    case ATTR_ENLARGED_IO_SLOT_COUNT:
        l_attrSize = sizeof(ATTR_ENLARGED_IO_SLOT_COUNT_type);
        break;

    // Get size for ATTR_HUGE_PAGE_COUNT attribute
    case ATTR_HUGE_PAGE_COUNT:
        l_attrSize = sizeof(ATTR_HUGE_PAGE_COUNT_type);
        break;

    // Get size for ATTR_HUGE_PAGE_SIZE attribute
    case ATTR_HUGE_PAGE_SIZE:
        l_attrSize = sizeof(ATTR_HUGE_PAGE_SIZE_type);
        break;

    // Get size for ATTR_HW_VER attribute
    case ATTR_HW_VER:
        l_attrSize = sizeof(ATTR_HW_VER_type);
        break;

    // Get size for ATTR_HYPERVISOR_IPL_DESTINATION attribute
    case ATTR_HYPERVISOR_IPL_DESTINATION:
        l_attrSize = sizeof(ATTR_HYPERVISOR_IPL_DESTINATION_type);
        break;

    // Get size for ATTR_HYPERVISOR_IPL_SIDE attribute
    case ATTR_HYPERVISOR_IPL_SIDE:
        l_attrSize = sizeof(ATTR_HYPERVISOR_IPL_SIDE_type);
        break;

    // Get size for ATTR_HYP_DISPATCH_WHEEL attribute
    case ATTR_HYP_DISPATCH_WHEEL:
        l_attrSize = sizeof(ATTR_HYP_DISPATCH_WHEEL_type);
        break;

    // Get size for ATTR_IPL_ATTRIBUTES attribute
    case ATTR_IPL_ATTRIBUTES:
        l_attrSize = sizeof(ATTR_IPL_ATTRIBUTES_type);
        break;

    // Get size for ATTR_IPS_ENABLE attribute
    case ATTR_IPS_ENABLE:
        l_attrSize = sizeof(ATTR_IPS_ENABLE_type);
        break;

    // Get size for ATTR_IPS_ENTER_TIME_SECONDS attribute
    case ATTR_IPS_ENTER_TIME_SECONDS:
        l_attrSize = sizeof(ATTR_IPS_ENTER_TIME_SECONDS_type);
        break;

    // Get size for ATTR_IPS_ENTER_UTILIZATION_PERCENT attribute
    case ATTR_IPS_ENTER_UTILIZATION_PERCENT:
        l_attrSize = sizeof(ATTR_IPS_ENTER_UTILIZATION_PERCENT_type);
        break;

    // Get size for ATTR_IPS_EXIT_TIME_SECONDS attribute
    case ATTR_IPS_EXIT_TIME_SECONDS:
        l_attrSize = sizeof(ATTR_IPS_EXIT_TIME_SECONDS_type);
        break;

    // Get size for ATTR_IPS_EXIT_UTILIZATION_PERCENT attribute
    case ATTR_IPS_EXIT_UTILIZATION_PERCENT:
        l_attrSize = sizeof(ATTR_IPS_EXIT_UTILIZATION_PERCENT_type);
        break;

    // Get size for ATTR_LMB_SIZE attribute
    case ATTR_LMB_SIZE:
        l_attrSize = sizeof(ATTR_LMB_SIZE_type);
        break;

    // Get size for ATTR_MAX_HSL_OPTICONNECT_CONNECTIONS attribute
    case ATTR_MAX_HSL_OPTICONNECT_CONNECTIONS:
        l_attrSize = sizeof(ATTR_MAX_HSL_OPTICONNECT_CONNECTIONS_type);
        break;

    // Get size for ATTR_OPEN_POWER_PM_MODE attribute
    case ATTR_OPEN_POWER_PM_MODE:
        l_attrSize = sizeof(ATTR_OPEN_POWER_PM_MODE_type);
        break;

    // Get size for ATTR_OPEN_POWER_PM_MODE_FREQ_PERCENT attribute
    case ATTR_OPEN_POWER_PM_MODE_FREQ_PERCENT:
        l_attrSize = sizeof(ATTR_OPEN_POWER_PM_MODE_FREQ_PERCENT_type);
        break;

    // Get size for ATTR_OS_IPL_MODE attribute
    case ATTR_OS_IPL_MODE:
        l_attrSize = sizeof(ATTR_OS_IPL_MODE_type);
        break;

    // Get size for ATTR_PHYP_SYSTEM_TYPE attribute
    case ATTR_PHYP_SYSTEM_TYPE:
        l_attrSize = sizeof(ATTR_PHYP_SYSTEM_TYPE_type);
        break;

    // Get size for ATTR_PRESERVE_MDC_PARTITION_VPD attribute
    case ATTR_PRESERVE_MDC_PARTITION_VPD:
        l_attrSize = sizeof(ATTR_PRESERVE_MDC_PARTITION_VPD_type);
        break;

    // Get size for ATTR_PROC_HW_TOPOLOGY attribute
    case ATTR_PROC_HW_TOPOLOGY:
        l_attrSize = sizeof(ATTR_PROC_HW_TOPOLOGY_type);
        break;

    // Get size for ATTR_ROLE attribute
    case ATTR_ROLE:
        l_attrSize = sizeof(ATTR_ROLE_type);
        break;

    // Get size for ATTR_SLCA_INDEX attribute
    case ATTR_SLCA_INDEX:
        l_attrSize = sizeof(ATTR_SLCA_INDEX_type);
        break;

    // Get size for ATTR_SLCA_RID attribute
    case ATTR_SLCA_RID:
        l_attrSize = sizeof(ATTR_SLCA_RID_type);
        break;

    // Get size for ATTR_SW_VER attribute
    case ATTR_SW_VER:
        l_attrSize = sizeof(ATTR_SW_VER_type);
        break;

    // Get size for ATTR_SYSTEM_BRAND_NAME attribute
    case ATTR_SYSTEM_BRAND_NAME:
        l_attrSize = sizeof(ATTR_SYSTEM_BRAND_NAME_type);
        break;

    // Get size for ATTR_VLAN_SWITCHES attribute
    case ATTR_VLAN_SWITCHES:
        l_attrSize = sizeof(ATTR_VLAN_SWITCHES_type);
        break;

    // Get size for ATTR_VTPM_ENABLED attribute
    case ATTR_VTPM_ENABLED:
        l_attrSize = sizeof(ATTR_VTPM_ENABLED_type);
        break;

    // Get size for ATTR_ADC_CHANNEL_FUNC_IDS attribute
    case ATTR_ADC_CHANNEL_FUNC_IDS:
        l_attrSize = sizeof(ATTR_ADC_CHANNEL_FUNC_IDS_type);
        break;

    // Get size for ATTR_ADC_CHANNEL_GAINS attribute
    case ATTR_ADC_CHANNEL_GAINS:
        l_attrSize = sizeof(ATTR_ADC_CHANNEL_GAINS_type);
        break;

    // Get size for ATTR_ADC_CHANNEL_GNDS attribute
    case ATTR_ADC_CHANNEL_GNDS:
        l_attrSize = sizeof(ATTR_ADC_CHANNEL_GNDS_type);
        break;

    // Get size for ATTR_ADC_CHANNEL_OFFSETS attribute
    case ATTR_ADC_CHANNEL_OFFSETS:
        l_attrSize = sizeof(ATTR_ADC_CHANNEL_OFFSETS_type);
        break;

    // Get size for ATTR_ADC_CHANNEL_SENSOR_NUMBERS attribute
    case ATTR_ADC_CHANNEL_SENSOR_NUMBERS:
        l_attrSize = sizeof(ATTR_ADC_CHANNEL_SENSOR_NUMBERS_type);
        break;

    // Get size for ATTR_ALLOW_CALLHOME_ESELS_TO_BMC attribute
    case ATTR_ALLOW_CALLHOME_ESELS_TO_BMC:
        l_attrSize = sizeof(ATTR_ALLOW_CALLHOME_ESELS_TO_BMC_type);
        break;

    // Get size for ATTR_APSS_GPIO_PORT_MODES attribute
    case ATTR_APSS_GPIO_PORT_MODES:
        l_attrSize = sizeof(ATTR_APSS_GPIO_PORT_MODES_type);
        break;

    // Get size for ATTR_APSS_GPIO_PORT_PINS attribute
    case ATTR_APSS_GPIO_PORT_PINS:
        l_attrSize = sizeof(ATTR_APSS_GPIO_PORT_PINS_type);
        break;

    // Get size for ATTR_BMC_HW_CHIP_TYPE attribute
    case ATTR_BMC_HW_CHIP_TYPE:
        l_attrSize = sizeof(ATTR_BMC_HW_CHIP_TYPE_type);
        break;

    // Get size for ATTR_BMC_MANUFACTURER attribute
    case ATTR_BMC_MANUFACTURER:
        l_attrSize = sizeof(ATTR_BMC_MANUFACTURER_type);
        break;

    // Get size for ATTR_BMC_MAX_ERROR_LOG_SIZE attribute
    case ATTR_BMC_MAX_ERROR_LOG_SIZE:
        l_attrSize = sizeof(ATTR_BMC_MAX_ERROR_LOG_SIZE_type);
        break;

    // Get size for ATTR_BMC_SW_TYPE attribute
    case ATTR_BMC_SW_TYPE:
        l_attrSize = sizeof(ATTR_BMC_SW_TYPE_type);
        break;

    // Get size for ATTR_CALCULATED_MAX_SYS_POWER_EXCLUDING_GPUS attribute
    case ATTR_CALCULATED_MAX_SYS_POWER_EXCLUDING_GPUS:
        l_attrSize = sizeof(ATTR_CALCULATED_MAX_SYS_POWER_EXCLUDING_GPUS_type);
        break;

    // Get size for ATTR_CALCULATED_PROC_MEMORY_POWER_DROP attribute
    case ATTR_CALCULATED_PROC_MEMORY_POWER_DROP:
        l_attrSize = sizeof(ATTR_CALCULATED_PROC_MEMORY_POWER_DROP_type);
        break;

    // Get size for ATTR_CUMULATIVE_PMCOMPLEX_RESET_COUNT attribute
    case ATTR_CUMULATIVE_PMCOMPLEX_RESET_COUNT:
        l_attrSize = sizeof(ATTR_CUMULATIVE_PMCOMPLEX_RESET_COUNT_type);
        break;

    // Get size for ATTR_GPIO_INFO attribute
    case ATTR_GPIO_INFO:
        l_attrSize = sizeof(ATTR_GPIO_INFO_type);
        break;

    // Get size for ATTR_GPU_SENSORS attribute
    case ATTR_GPU_SENSORS:
        l_attrSize = sizeof(ATTR_GPU_SENSORS_type);
        break;

    // Get size for ATTR_HDAT_I2C_ADDR attribute
    case ATTR_HDAT_I2C_ADDR:
        l_attrSize = sizeof(ATTR_HDAT_I2C_ADDR_type);
        break;

    // Get size for ATTR_HDAT_I2C_BUS_FREQ attribute
    case ATTR_HDAT_I2C_BUS_FREQ:
        l_attrSize = sizeof(ATTR_HDAT_I2C_BUS_FREQ_type);
        break;

    // Get size for ATTR_HDAT_I2C_DEVICE_LABEL attribute
    case ATTR_HDAT_I2C_DEVICE_LABEL:
        l_attrSize = sizeof(ATTR_HDAT_I2C_DEVICE_LABEL_type);
        break;

    // Get size for ATTR_HDAT_I2C_DEVICE_PURPOSE attribute
    case ATTR_HDAT_I2C_DEVICE_PURPOSE:
        l_attrSize = sizeof(ATTR_HDAT_I2C_DEVICE_PURPOSE_type);
        break;

    // Get size for ATTR_HDAT_I2C_DEVICE_TYPE attribute
    case ATTR_HDAT_I2C_DEVICE_TYPE:
        l_attrSize = sizeof(ATTR_HDAT_I2C_DEVICE_TYPE_type);
        break;

    // Get size for ATTR_HDAT_I2C_ELEMENTS attribute
    case ATTR_HDAT_I2C_ELEMENTS:
        l_attrSize = sizeof(ATTR_HDAT_I2C_ELEMENTS_type);
        break;

    // Get size for ATTR_HDAT_I2C_ENGINE attribute
    case ATTR_HDAT_I2C_ENGINE:
        l_attrSize = sizeof(ATTR_HDAT_I2C_ENGINE_type);
        break;

    // Get size for ATTR_HDAT_I2C_MASTER_PORT attribute
    case ATTR_HDAT_I2C_MASTER_PORT:
        l_attrSize = sizeof(ATTR_HDAT_I2C_MASTER_PORT_type);
        break;

    // Get size for ATTR_HDAT_I2C_SLAVE_PORT attribute
    case ATTR_HDAT_I2C_SLAVE_PORT:
        l_attrSize = sizeof(ATTR_HDAT_I2C_SLAVE_PORT_type);
        break;

    // Get size for ATTR_HTMGT_INTERNAL_FLAGS attribute
    case ATTR_HTMGT_INTERNAL_FLAGS:
        l_attrSize = sizeof(ATTR_HTMGT_INTERNAL_FLAGS_type);
        break;

    // Get size for ATTR_HTMGT_PMCOMPLEX_RESET_COUNT_TIMER attribute
    case ATTR_HTMGT_PMCOMPLEX_RESET_COUNT_TIMER:
        l_attrSize = sizeof(ATTR_HTMGT_PMCOMPLEX_RESET_COUNT_TIMER_type);
        break;

    // Get size for ATTR_HTMGT_SAFEMODE attribute
    case ATTR_HTMGT_SAFEMODE:
        l_attrSize = sizeof(ATTR_HTMGT_SAFEMODE_type);
        break;

    // Get size for ATTR_HTMGT_SAVED_POWER_LIMIT attribute
    case ATTR_HTMGT_SAVED_POWER_LIMIT:
        l_attrSize = sizeof(ATTR_HTMGT_SAVED_POWER_LIMIT_type);
        break;

    // Get size for ATTR_IPMI_INSTANCE attribute
    case ATTR_IPMI_INSTANCE:
        l_attrSize = sizeof(ATTR_IPMI_INSTANCE_type);
        break;

    // Get size for ATTR_IPMI_MAX_BUFFER_SIZE attribute
    case ATTR_IPMI_MAX_BUFFER_SIZE:
        l_attrSize = sizeof(ATTR_IPMI_MAX_BUFFER_SIZE_type);
        break;

    // Get size for ATTR_IPMI_SENSORS attribute
    case ATTR_IPMI_SENSORS:
        l_attrSize = sizeof(ATTR_IPMI_SENSORS_type);
        break;

    // Get size for ATTR_IS_MPIPL_SUPPORTED attribute
    case ATTR_IS_MPIPL_SUPPORTED:
        l_attrSize = sizeof(ATTR_IS_MPIPL_SUPPORTED_type);
        break;

    // Get size for ATTR_MAX_VDD_CURRENT_READING attribute
    case ATTR_MAX_VDD_CURRENT_READING:
        l_attrSize = sizeof(ATTR_MAX_VDD_CURRENT_READING_type);
        break;

    // Get size for ATTR_MISC_SYSTEM_COMPONENTS_MAX_POWER_WATTS attribute
    case ATTR_MISC_SYSTEM_COMPONENTS_MAX_POWER_WATTS:
        l_attrSize = sizeof(ATTR_MISC_SYSTEM_COMPONENTS_MAX_POWER_WATTS_type);
        break;

    // Get size for ATTR_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP attribute
    case ATTR_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP:
        l_attrSize = sizeof(ATTR_MRW_SAFEMODE_MEM_THROTTLE_NUMERATOR_PER_CHIP_type);
        break;

    // Get size for ATTR_MSS_UTIL_N_PER_MBA attribute
    case ATTR_MSS_UTIL_N_PER_MBA:
        l_attrSize = sizeof(ATTR_MSS_UTIL_N_PER_MBA_type);
        break;

    // Get size for ATTR_NO_APSS_PROC_POWER_VCS_VIO_WATTS attribute
    case ATTR_NO_APSS_PROC_POWER_VCS_VIO_WATTS:
        l_attrSize = sizeof(ATTR_NO_APSS_PROC_POWER_VCS_VIO_WATTS_type);
        break;

    // Get size for ATTR_N_PLUS_ONE_MEM_POWER attribute
    case ATTR_N_PLUS_ONE_MEM_POWER:
        l_attrSize = sizeof(ATTR_N_PLUS_ONE_MEM_POWER_type);
        break;

    // Get size for ATTR_N_PLUS_ONE_N_PER_CHIP attribute
    case ATTR_N_PLUS_ONE_N_PER_CHIP:
        l_attrSize = sizeof(ATTR_N_PLUS_ONE_N_PER_CHIP_type);
        break;

    // Get size for ATTR_N_PLUS_ONE_N_PER_MBA attribute
    case ATTR_N_PLUS_ONE_N_PER_MBA:
        l_attrSize = sizeof(ATTR_N_PLUS_ONE_N_PER_MBA_type);
        break;

    // Get size for ATTR_OCC_CONTROL_DATA attribute
    case ATTR_OCC_CONTROL_DATA:
        l_attrSize = sizeof(ATTR_OCC_CONTROL_DATA_type);
        break;

    // Get size for ATTR_OPAL_MODEL attribute
    case ATTR_OPAL_MODEL:
        l_attrSize = sizeof(ATTR_OPAL_MODEL_type);
        break;

    // Get size for ATTR_OPEN_POWER_DIMM_ERROR_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_DIMM_ERROR_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_DIMM_ERROR_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_DIMM_READ_TIMEOUT_SEC attribute
    case ATTR_OPEN_POWER_DIMM_READ_TIMEOUT_SEC:
        l_attrSize = sizeof(ATTR_OPEN_POWER_DIMM_READ_TIMEOUT_SEC_type);
        break;

    // Get size for ATTR_OPEN_POWER_DIMM_THROTTLE_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_DIMM_THROTTLE_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_DIMM_THROTTLE_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_GPU_ERROR_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_GPU_ERROR_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_GPU_ERROR_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_GPU_MEM_ERROR_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_GPU_MEM_ERROR_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_GPU_MEM_ERROR_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_GPU_MEM_READ_TIMEOUT_SEC attribute
    case ATTR_OPEN_POWER_GPU_MEM_READ_TIMEOUT_SEC:
        l_attrSize = sizeof(ATTR_OPEN_POWER_GPU_MEM_READ_TIMEOUT_SEC_type);
        break;

    // Get size for ATTR_OPEN_POWER_GPU_READ_TIMEOUT_SEC attribute
    case ATTR_OPEN_POWER_GPU_READ_TIMEOUT_SEC:
        l_attrSize = sizeof(ATTR_OPEN_POWER_GPU_READ_TIMEOUT_SEC_type);
        break;

    // Get size for ATTR_OPEN_POWER_MEMCTRL_ERROR_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_MEMCTRL_ERROR_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_MEMCTRL_ERROR_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_MEMCTRL_READ_TIMEOUT_SEC attribute
    case ATTR_OPEN_POWER_MEMCTRL_READ_TIMEOUT_SEC:
        l_attrSize = sizeof(ATTR_OPEN_POWER_MEMCTRL_READ_TIMEOUT_SEC_type);
        break;

    // Get size for ATTR_OPEN_POWER_MEMCTRL_THROTTLE_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_MEMCTRL_THROTTLE_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_MEMCTRL_THROTTLE_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_POWER_CAP attribute
    case ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_POWER_CAP:
        l_attrSize = sizeof(ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_POWER_CAP_type);
        break;

    // Get size for ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_THROTTLING attribute
    case ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_THROTTLING:
        l_attrSize = sizeof(ATTR_OPEN_POWER_MIN_MEM_UTILIZATION_THROTTLING_type);
        break;

    // Get size for ATTR_OPEN_POWER_MIN_POWER_CAP_WATTS attribute
    case ATTR_OPEN_POWER_MIN_POWER_CAP_WATTS:
        l_attrSize = sizeof(ATTR_OPEN_POWER_MIN_POWER_CAP_WATTS_type);
        break;

    // Get size for ATTR_OPEN_POWER_N_BULK_POWER_LIMIT_WATTS attribute
    case ATTR_OPEN_POWER_N_BULK_POWER_LIMIT_WATTS:
        l_attrSize = sizeof(ATTR_OPEN_POWER_N_BULK_POWER_LIMIT_WATTS_type);
        break;

    // Get size for ATTR_OPEN_POWER_N_MAX_MEM_POWER_WATTS attribute
    case ATTR_OPEN_POWER_N_MAX_MEM_POWER_WATTS:
        l_attrSize = sizeof(ATTR_OPEN_POWER_N_MAX_MEM_POWER_WATTS_type);
        break;

    // Get size for ATTR_OPEN_POWER_N_PLUS_ONE_BULK_POWER_LIMIT_WATTS attribute
    case ATTR_OPEN_POWER_N_PLUS_ONE_BULK_POWER_LIMIT_WATTS:
        l_attrSize = sizeof(ATTR_OPEN_POWER_N_PLUS_ONE_BULK_POWER_LIMIT_WATTS_type);
        break;

    // Get size for ATTR_OPEN_POWER_N_PLUS_ONE_HPC_BULK_POWER_LIMIT_WATTS attribute
    case ATTR_OPEN_POWER_N_PLUS_ONE_HPC_BULK_POWER_LIMIT_WATTS:
        l_attrSize = sizeof(ATTR_OPEN_POWER_N_PLUS_ONE_HPC_BULK_POWER_LIMIT_WATTS_type);
        break;

    // Get size for ATTR_OPEN_POWER_N_PLUS_ONE_MAX_MEM_POWER_WATTS attribute
    case ATTR_OPEN_POWER_N_PLUS_ONE_MAX_MEM_POWER_WATTS:
        l_attrSize = sizeof(ATTR_OPEN_POWER_N_PLUS_ONE_MAX_MEM_POWER_WATTS_type);
        break;

    // Get size for ATTR_OPEN_POWER_PROC_DVFS_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_PROC_DVFS_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_PROC_DVFS_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_PROC_ERROR_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_PROC_ERROR_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_PROC_ERROR_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_PROC_READ_TIMEOUT_SEC attribute
    case ATTR_OPEN_POWER_PROC_READ_TIMEOUT_SEC:
        l_attrSize = sizeof(ATTR_OPEN_POWER_PROC_READ_TIMEOUT_SEC_type);
        break;

    // Get size for ATTR_OPEN_POWER_PROC_WEIGHT attribute
    case ATTR_OPEN_POWER_PROC_WEIGHT:
        l_attrSize = sizeof(ATTR_OPEN_POWER_PROC_WEIGHT_type);
        break;

    // Get size for ATTR_OPEN_POWER_QUAD_WEIGHT attribute
    case ATTR_OPEN_POWER_QUAD_WEIGHT:
        l_attrSize = sizeof(ATTR_OPEN_POWER_QUAD_WEIGHT_type);
        break;

    // Get size for ATTR_OPEN_POWER_REGULATOR_EFFICIENCY_FACTOR attribute
    case ATTR_OPEN_POWER_REGULATOR_EFFICIENCY_FACTOR:
        l_attrSize = sizeof(ATTR_OPEN_POWER_REGULATOR_EFFICIENCY_FACTOR_type);
        break;

    // Get size for ATTR_OPEN_POWER_SOFT_MIN_PCAP_WATTS attribute
    case ATTR_OPEN_POWER_SOFT_MIN_PCAP_WATTS:
        l_attrSize = sizeof(ATTR_OPEN_POWER_SOFT_MIN_PCAP_WATTS_type);
        break;

    // Get size for ATTR_OPEN_POWER_TURBO_MODE_SUPPORTED attribute
    case ATTR_OPEN_POWER_TURBO_MODE_SUPPORTED:
        l_attrSize = sizeof(ATTR_OPEN_POWER_TURBO_MODE_SUPPORTED_type);
        break;

    // Get size for ATTR_OPEN_POWER_VRM_READ_TIMEOUT_SEC attribute
    case ATTR_OPEN_POWER_VRM_READ_TIMEOUT_SEC:
        l_attrSize = sizeof(ATTR_OPEN_POWER_VRM_READ_TIMEOUT_SEC_type);
        break;

    // Get size for ATTR_OPEN_POWER_VRM_VDD_DVFS_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_VRM_VDD_DVFS_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_VRM_VDD_DVFS_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_VRM_VDD_ERROR_TEMP_DEG_C attribute
    case ATTR_OPEN_POWER_VRM_VDD_ERROR_TEMP_DEG_C:
        l_attrSize = sizeof(ATTR_OPEN_POWER_VRM_VDD_ERROR_TEMP_DEG_C_type);
        break;

    // Get size for ATTR_OPEN_POWER_VRM_VDD_READ_TIMEOUT_SEC attribute
    case ATTR_OPEN_POWER_VRM_VDD_READ_TIMEOUT_SEC:
        l_attrSize = sizeof(ATTR_OPEN_POWER_VRM_VDD_READ_TIMEOUT_SEC_type);
        break;

    // Get size for ATTR_OP_TRACE_LITE attribute
    case ATTR_OP_TRACE_LITE:
        l_attrSize = sizeof(ATTR_OP_TRACE_LITE_type);
        break;

    // Get size for ATTR_OT_MEM_POWER attribute
    case ATTR_OT_MEM_POWER:
        l_attrSize = sizeof(ATTR_OT_MEM_POWER_type);
        break;

    // Get size for ATTR_OT_MIN_N_PER_MBA attribute
    case ATTR_OT_MIN_N_PER_MBA:
        l_attrSize = sizeof(ATTR_OT_MIN_N_PER_MBA_type);
        break;

    // Get size for ATTR_POWERCAP_MEM_POWER attribute
    case ATTR_POWERCAP_MEM_POWER:
        l_attrSize = sizeof(ATTR_POWERCAP_MEM_POWER_type);
        break;

    // Get size for ATTR_POWERCAP_N_PER_CHIP attribute
    case ATTR_POWERCAP_N_PER_CHIP:
        l_attrSize = sizeof(ATTR_POWERCAP_N_PER_CHIP_type);
        break;

    // Get size for ATTR_POWERCAP_N_PER_MBA attribute
    case ATTR_POWERCAP_N_PER_MBA:
        l_attrSize = sizeof(ATTR_POWERCAP_N_PER_MBA_type);
        break;

    // Get size for ATTR_PROC_MHZ_PER_WATT attribute
    case ATTR_PROC_MHZ_PER_WATT:
        l_attrSize = sizeof(ATTR_PROC_MHZ_PER_WATT_type);
        break;

    // Get size for ATTR_PROC_SOCKET_POWER_WATTS attribute
    case ATTR_PROC_SOCKET_POWER_WATTS:
        l_attrSize = sizeof(ATTR_PROC_SOCKET_POWER_WATTS_type);
        break;

    // Get size for ATTR_REPORT_THROTTLE_BELOW_NOMINAL attribute
    case ATTR_REPORT_THROTTLE_BELOW_NOMINAL:
        l_attrSize = sizeof(ATTR_REPORT_THROTTLE_BELOW_NOMINAL_type);
        break;

    // Get size for ATTR_VDD_CURRENT_OVERFLOW_WORKAROUND_ENABLE attribute
    case ATTR_VDD_CURRENT_OVERFLOW_WORKAROUND_ENABLE:
        l_attrSize = sizeof(ATTR_VDD_CURRENT_OVERFLOW_WORKAROUND_ENABLE_type);
        break;

    default:
        break;
    }

    return l_attrSize;
}

} // End namespace TARGETING

#endif // TARG_ATTRIBUTETRAITS_H

